C  --------------------------------------------------------------
C (C) COPYRIGHT 1994, 1996 University of Chicago
C   
C   THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT
C   LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR
C   EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE
C   TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE
C   AND AVAILABLE AT:
C   http://www.mcs.anl.gov/autodiff/adifor_license.html, or
C   http://www.cs.rice.edu/~adifor/AdiforPublicLicense.html.
C   ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE
C   OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR
C   OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:
C   
C   Paul Hovland, Mathematics and Computer Science Division,
C   Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL
C   60439, hovland@mcs.anl.gov.
C   
C   Alan Carle, Center for Research on Parallel Computation, Rice
C   University MS 41, 6100 Main Street, Houston, TX 77005,
C   carle@cs.rice.edu.
C  --------------------------------------------------------------
C

      subroutine ehrptZ 
      integer FUNC_AINT, FUNC_ANINT, FUNC_DNINT, FUNC_ABS, FUNC_MOD
      integer FUNC_SIGN, FUNC_DIM, FUNC_MAX, FUNC_MIN
      integer FUNC_SQRT, FUNC_SQRT4CABS
      integer FUNC_POWERBASE, FUNC_POWEREXP, FUNC_POWERBOTH
      integer FUNC_ASIN, FUNC_ACOS

      parameter (FUNC_AINT = 1)
      parameter (FUNC_ANINT = 2)
      parameter (FUNC_DNINT = 2)
      parameter (FUNC_ABS = 3)
      parameter (FUNC_MOD = 4)      
      parameter (FUNC_SIGN = 5)
      parameter (FUNC_DIM = 6)
      parameter (FUNC_MAX = 7)
      parameter (FUNC_MIN = 8)
      parameter (FUNC_SQRT = 9)
      parameter (FUNC_POWERBASE = 10)      
      parameter (FUNC_POWEREXP = 11)      
      parameter (FUNC_POWERBOTH = 12)      
      parameter (FUNC_ASIN = 13)
      parameter (FUNC_ACOS = 14)
      parameter (FUNC_SQRT4CABS = 15)

      integer MAX_FN
      parameter (MAX_FN = FUNC_SQRT4CABS)

      integer ERR_AINT, ERR_ANINT, ERR_DNINT
      integer ERR_ABS, ERR_MOD, ERR_SIGN
      integer ERR_DIM, ERR_MAX, ERR_MIN
      integer ERR_SQRT
      integer ERR_POWER
      integer ERR_POWERBASE, ERR_POWEREXP
      integer ERR_POWERBOTH, ERR_ASIN, ERR_ACOS
      integer ERR_SQRT4CABS

      parameter (ERR_AINT = 1)
      parameter (ERR_ANINT = ERR_AINT + 1)
      parameter (ERR_DNINT = ERR_ANINT)
      parameter (ERR_ABS = ERR_DNINT + 1)
      parameter (ERR_MOD = ERR_ABS + 1)
      parameter (ERR_SIGN = ERR_MOD + 1)
      parameter (ERR_DIM = ERR_SIGN + 1)
      parameter (ERR_MAX = ERR_DIM + 1)
      parameter (ERR_MIN = ERR_MAX + 1)
      parameter (ERR_SQRT = ERR_MIN + 1)

      parameter (ERR_POWER = ERR_SQRT + 1)

      parameter (ERR_POWERBASE = ERR_POWER)
      parameter (ERR_POWEREXP = ERR_POWER)
      parameter (ERR_POWERBOTH = ERR_POWER)

      parameter (ERR_ASIN = ERR_POWER + 1)
      parameter (ERR_ACOS = ERR_ASIN + 1)
      parameter (ERR_SQRT4CABS = ERR_ACOS + 1)

      integer MAX_ERR
      parameter (MAX_ERR = ERR_SQRT4CABS)

      common /ehxcb/ ehxout
      integer ehxout
      common /ehxcbZ/ ehZres,
     +        ehZcnt
     
      integer MAX_CONDITIONS
      parameter (MAX_CONDITIONS = 5)

      DOUBLE COMPLEX ehZres(MAX_ERR, MAX_CONDITIONS)
      integer ehZcnt(MAX_ERR, MAX_CONDITIONS)
      
      integer i,j
      logical is_terse_mode
      logical is_not_terse_mode

      character*50 err
      parameter (err = 'INTERNAL ERROR')

      character*50 exceptions(MAX_CONDITIONS, MAX_ERR)

      integer FOUR
      parameter (FOUR = MAX_CONDITIONS - 1)
      
      integer THREE
      parameter (THREE = MAX_CONDITIONS - 2)



      data exceptions /
     +     'AINT',             FOUR * err,
     +     'ANINT or DNINT',   FOUR * err,
     +     'ABS',              FOUR * err,
     +     'MOD',              FOUR * err,
     +     'SIGN',             FOUR * err,
     +     'DIM',              FOUR * err,
     +     'MAX',              FOUR * err,
     +     'MIN',              FOUR * err,
     +     'SQRT first deriv', 'SQRT second deriv',   THREE * err,
     +     'Complex POWER: df/dx',
     +     'Complex POWER: df/dy',
     +     'Complex POWER second deriv: d2f/dxx',
     +     'Complex POWER second deriv: d2f/dxy',
     +     'Complex POWER second deriv: d2f/dyy',
     +     'ASIN first deriv', 'ASIN second deriv',   THREE * err,
     +     'ACOS first deriv', 'ACOS second deriv',   THREE * err,
     +     'SQRT4CABS first deriv', 'SQRT4CABS second deriv',
     +           THREE * err
     +     /



      is_terse_mode = .false.
      is_not_terse_mode = .false.

      do i = 1, MAX_ERR
         do j = 1, MAX_CONDITIONS

         if ( ehZcnt(i,j) .ne. 0 ) then

            if ( ehZcnt(i,j) .eq. -29999 ) then

               is_terse_mode = .true.
               write (ehxout,*) exceptions(j,i)

            else

               is_not_terse_mode = .true.
               write (ehxout,*) exceptions(j,i),
     +              ': ',ehZcnt(i,j),' times.'

            endif

            if ( (ehZcnt(i,j) .gt. -29999)
     +          .and. (ehZcnt(i,j) .lt. 0) ) then

C              -29999 < count < 0 means you mixed modes for sure
               is_terse_mode = .true.
               is_not_terse_mode = .true.

            endif

         endif

         enddo
      enddo

      if ( is_terse_mode .and. is_not_terse_mode ) then

         write (ehxout,*)
         write (ehxout,*)
     +'WARNING: Mixed terse and other modes! Counts may be inaccurate.' 
         write (ehxout,*)
     +'         Negative counts indicate known mixed modes.'

      endif   

      return
      end
