/*
  Customized part of the interface
  to the ADIFOR visualization library

  Vitaly Shmatikov, Summer 1994
*/

#include "PREC_advis_pd_include.h"
#include <sys/times.h>

/**************************************
*                                     *
*     ADVIS interface to spg1         *
*                                     *
**************************************/

void spg1_ca(destptrptr,alphaptr,vectoraptrptr)

  PREC_vectorstruct **vectoraptrptr,**destptrptr;
  FLOATING_DECLARE *alphaptr;

{

#ifdef _VTRACE
   int effect_arity = 0;      /* Effective arity */
   int no_of_zero_vecs = 0;   /* # of 0 vectors on the r.h.s. */
   int no_of_zero_mult = 0;   /* # of 0 multipliers on the r.h.s. */
   int no_of_zero_both = 0;   /* # of 0 vectors with 0 multipliers */
   int old_dstype = 0;        /* DS-type of the l.h.s. before spg call */
   int new_dstype = 0;        /*          ...          after     ...     */
   int old_nz = 0;            /* # of non-0 elements of the l.h.s. before */
   int new_nz = 0;            /*            ...                    after */
   int real_time = 0;         /* Time of spg call */
   struct tms tmrec;          /* Time record */
   short lhs_is_in_rhs = FALSE;  /* l.h.s. appears in r.h.s. */
   int rhs_nz_sum = 0;        /* Total # of nonzeroes on the r.h.s. */

   if ((*vectoraptrptr) != NULL) {

      if(!((*vectoraptrptr)->nonzeroes)) no_of_zero_vecs++;
      if(X_IS_ZERO(*alphaptr)) no_of_zero_mult++;
      if((*vectoraptrptr)->nonzeroes && !X_IS_ZERO(*alphaptr)) effect_arity++;
      if(!((*vectoraptrptr)->nonzeroes || !X_IS_ZERO(*alphaptr))) no_of_zero_both++;

      if ((*vectoraptrptr) == (*destptrptr)) lhs_is_in_rhs = TRUE;

      /* # of multiplications = # of nonzeroes on the r.h.s. */
      MULT_COUNT += (*vectoraptrptr)->nonzeroes;
      /* Calculate the sum of all nonzeroes on the r.h.s. */
      rhs_nz_sum += (*vectoraptrptr)->nonzeroes;
   }
   if ((*destptrptr) != NULL) {
      old_dstype = (*destptrptr)->dstype;
      old_nz = (*destptrptr)->nonzeroes;
   }
#endif   

adxpy1(destptrptr,alphaptr,vectoraptrptr);

#ifdef _VTRACE
   new_dstype = (*destptrptr)->dstype;
   new_nz = (*destptrptr)->nonzeroes;
   /* Call the timer */
   times(&tmrec); real_time = (long)tmrec.tms_utime;

   ADXPY_CALLS++;
   /* # of additions = (# nz on r.h.s - # nz on l.h.s.) / eff.arity */
   if(effect_arity)
     ADD_COUNT += ((rhs_nz_sum - new_nz) / effect_arity);

   /* Collect statistics */
   collect_visual(1, effect_arity, no_of_zero_vecs, no_of_zero_mult, no_of_zero_both, lhs_is_in_rhs, old_dstype, new_dstype, FALSE);  
   /* destptrptr has changed => update its chronology */
   update_vector((*destptrptr), real_time, 1, effect_arity, new_dstype);
#endif

}

/**************************************
*                                     *
*     ADVIS interface to spg2         *
*                                     *
***************************************/

void spg2_ca(destptrptr,alphaptr,vectoraptrptr,betaptr,vectorbptrptr)

  PREC_vectorstruct **destptrptr;
  FLOATING_DECLARE *alphaptr;
  PREC_vectorstruct **vectoraptrptr;
  FLOATING_DECLARE *betaptr;
  PREC_vectorstruct **vectorbptrptr;

{

#ifdef _VTRACE
   int effect_arity = 0;      /* Effective arity */
   int no_of_zero_vecs = 0;   /* # of 0 vectors on the r.h.s. */
   int no_of_zero_mult = 0;   /* # of 0 multipliers on the r.h.s. */
   int no_of_zero_both = 0;   /* # of 0 vectors that have 0 multipliers */
   int old_dstype = 0;        /* DS-type of the l.h.s. before spg call */
   int new_dstype = 0;        /*          ...          after     ...     */
   int old_nz = 0;            /* # of non-0 elements of the l.h.s. before */
   int new_nz = 0;            /*            ...                    after */
   int real_time = 0;         /* Time of spg call */
   struct tms tmrec;          /* Time record */
   short lhs_is_in_rhs = FALSE;  /* l.h.s. is also in r.h.s. */
   int rhs_nz_sum = 0;        /* Total # of nonzeroes on the r.h.s. */

   if ((*vectoraptrptr != NULL)) {

      if(!((*vectoraptrptr)->nonzeroes)) no_of_zero_vecs++;
      if(X_IS_ZERO(*alphaptr)) no_of_zero_mult++;
      if((*vectoraptrptr)->nonzeroes && !X_IS_ZERO(*alphaptr)) effect_arity++;
      if(!((*vectoraptrptr)->nonzeroes || !X_IS_ZERO(*alphaptr))) no_of_zero_both++;

      if ((*vectoraptrptr) == (*destptrptr)) lhs_is_in_rhs = TRUE;

      /* # of multiplications = # of nonzeroes on the r.h.s. */
      MULT_COUNT += (*vectoraptrptr)->nonzeroes;
      /* Calculate the sum of all nonzeroes on the r.h.s. */
      rhs_nz_sum += (*vectoraptrptr)->nonzeroes;
   }
   if ((*vectorbptrptr) != NULL) {

      if((*vectorbptrptr)->nonzeroes == 0) no_of_zero_vecs++;
      if(X_IS_ZERO(*betaptr)) no_of_zero_mult++;
      if((*vectorbptrptr)->nonzeroes && !X_IS_ZERO(*betaptr)) effect_arity++;
      if(!((*vectorbptrptr)->nonzeroes || !X_IS_ZERO(*betaptr))) no_of_zero_both++;

      if ((*vectorbptrptr) == (*destptrptr)) lhs_is_in_rhs = TRUE;

      /* # of multiplications = # of nonzeroes on the r.h.s. */
      MULT_COUNT += (*vectorbptrptr)->nonzeroes;
      /* Calculate the sum of all nonzeroes on the r.h.s. */
      rhs_nz_sum += (*vectorbptrptr)->nonzeroes;
   }
   if ((*destptrptr) != NULL) {
      old_dstype = (*destptrptr)->dstype;
      old_nz = (*destptrptr)->nonzeroes;
   }
#endif   

adxpy2(destptrptr,alphaptr,vectoraptrptr,betaptr,vectorbptrptr);

#ifdef _VTRACE
   new_dstype = (*destptrptr)->dstype;
   new_nz = (*destptrptr)->nonzeroes;
   /* Call the timer */
   times(&tmrec); real_time = (long)tmrec.tms_utime;

   ADXPY_CALLS++;
   /* # of additions = (# nz on r.h.s - # nz on l.h.s.) / eff.arity */
   if(effect_arity)
     ADD_COUNT += ((rhs_nz_sum - new_nz) / effect_arity);

   /* Collect statistics */
   collect_visual(2, effect_arity, no_of_zero_vecs, no_of_zero_mult, no_of_zero_both, lhs_is_in_rhs, old_dstype, new_dstype, FALSE);  
   /* destptrptr has changed => update its chronology */
   update_vector((*destptrptr), real_time, 2, effect_arity, new_dstype);
#endif

}

/**************************************
*                                     *
*     ADVIS interface to spgx         *
*                                     *
***************************************/

void spgx_ca(destptrptr,arity,alphaptrs,vectorptrptrs)

  PREC_vectorstruct **destptrptr;
  int arity;
  FLOATING_DECLARE *alphaptrs[];
  PREC_vectorstruct **vectorptrptrs[];

{
#ifdef _VTRACE

  /* code to be written later */

#endif

heapadxpy(destptrptr,arity,alphaptrs,vectorptrptrs);

#ifdef _VTRACE

  /* code to be written later */

#endif

}

/**************************************
*                                     *
*     ADVIS interface to spzro        *
*                                     *
***************************************/

void spzro_ca(destptrptr)
PREC_vectorstruct **destptrptr;
{

#ifdef _VTRACE
   int old_dstype = 0;  /* DS-type of the l.h.s. before spg call */
   int new_dstype = 0;  /*          ...          after     ...     */
   int old_nz = 0;      /* # of non-0 elements of the l.h.s. before */
   int new_nz = 0;      /*            ...                    after */
   int real_time = 0;   /* Time of spg call */
   struct tms tmrec;    /* Time record */

   if ((*destptrptr) != NULL) {
      old_dstype = (*destptrptr)->dstype;
      old_nz = (*destptrptr)->nonzeroes;
   }
#endif

adzero(destptrptr);

#ifdef _VTRACE
   new_dstype = (*destptrptr)->dstype;
   new_nz = (*destptrptr)->nonzeroes;
   /* Call the timer */
   times(&tmrec); real_time = (long)tmrec.tms_utime;

   /* destptrptr has changed => update its chronology */
   ADXPY_CALLS++;
   collect_visual(0, 0, 0, 0, 0, FALSE, old_dstype, new_dstype, FALSE);
   update_vector((*destptrptr), real_time, 0, 0, S1);
#endif

}

/***********************************
*                                  *
*     ADVIS interface to spcp      *
*                                  *
***********************************/

void spcp_ca(destptrptr,oldvectorptrptr)                     
PREC_vectorstruct **destptrptr,**oldvectorptrptr;
{

#ifdef _VTRACE

   int effect_arity = 0;      /* Effective arity */
   int no_of_zero_vecs = 0;   /* # of 0 vectors on the r.h.s. */
   int old_dstype = 0;        /* DS-type of the l.h.s. before spg call */
   int new_dstype = 0;        /*          ...          after     ...     */
   int old_nz = 0;            /* # of non-0 elements of the l.h.s. before */
   int new_nz = 0;            /*            ...                    after */
   int real_time = 0;         /* Time of spg call */
   struct tms tmrec;          /* Time record */
   short lhs_is_in_rhs = FALSE;  /* l.h.s. appears in r.h.s. */

   if ((*oldvectorptrptr) != NULL) {
      if(!((*oldvectorptrptr)->nonzeroes)) no_of_zero_vecs++;
      else effect_arity++;

      if ((*oldvectorptrptr) == (*destptrptr)) lhs_is_in_rhs = TRUE;

      /* # of multiplications = # of nonzeroes on the r.h.s. */
      MULT_COUNT += (*oldvectorptrptr)->nonzeroes;
   }

   if ((*destptrptr) != NULL) {
      old_dstype = (*destptrptr)->dstype;
      old_nz = (*destptrptr)->nonzeroes;
   }

#endif

adcopy(destptrptr,oldvectorptrptr);

#ifdef _VTRACE
   new_dstype = (*destptrptr)->dstype;
   new_nz = (*destptrptr)->nonzeroes;
   /* Call the timer */
   times(&tmrec); real_time = (long)tmrec.tms_utime;

   /* destptrptr has changed => update its chronology */
   ADXPY_CALLS++;
   collect_visual(1, effect_arity, no_of_zero_vecs, 0, 0, lhs_is_in_rhs, old_dstype, new_dstype, FALSE);
   update_vector((*destptrptr), real_time, 0, 0, (*oldvectorptrptr)->dstype);
#endif

}
