/*
  Precision-independent header file for
  the ADIFOR visualization library

  Vitaly Shmatikov, Summer 1994
*/

#ifndef ADVIS_PI_INCLUDED
#define ADVIS_PI_INCLUDED

#include <stdio.h>
#include "gaxpy_pi_macros_and_data_structures.h"

#define STRING_MAX 100
#define ADDR_MAX 15

/* For all histograms, the # of nonzeroes is clipped at this threshold */
#define NZ_THRESHOLD 30

/* 
 Exported routines 
*/

/* Set libraries parameters to default values */
void xvsdef _P(());
/* Initialize the visualization library */
void xvsini _P(());
/* 
  Set the parameters of the visualization library: 

  Value of the 1st argument        Possible values of the 2nd argument
  -------------------------        -----------------------------------
  1 (short tracking)               0 = off
				   Slice size > 0

  2 (tracking anonymous vectors)   0 = off
				   Anything else = on

  3 (output global history)        0 = off
				   Anything else = on

  4 (compression type)             0 = no compression
				   1 = compress
				   2 = gzip

  5 (data file name)               Name of the data file

  6 (history file name)            Name of the history file
*/
void xvscnf _P((int *, int *));
/* Put a numbered mark into the global history file */
void visual_mark _P((int *));
/* Close the visualization library */
void xvscls _P(());

#define SHORT_TRACKING 1
#define MAX_SLICE_SIZE 200

#define TRACK_ANONYMOUS 2

#define GLOBAL_TRACKING 3

#define COMPRESSION_TYPE 4
#define NO_COMPRESSION 0
#define COMP_COMPRESS 1
#define COMP_GZIP 2

#define DATAFILE_NAME 5

#define HISTORYFILE_NAME 6

/* Collect statistics */
void collect_visual _P((int, int, int, int, int, short, int, int, short));

/* Convert a string from Fortran to C */
char *f_to_c_str _P((char*));

/* Data structure for storing identifier sets */

typedef struct _id_elem {
   char* id;                /* Symbolic identifier */
   struct _id_elem *next;
} id_elem;
typedef id_elem *pid_elem;

/* Global declarations */

#ifdef ADVIS_GLOBALS
/* Actual declarations */
#define PREFIX 
#else
/* In all files that don't define ADVIS_GLOBALS,
   the declarations will be external */
#define PREFIX extern
#endif

/* Data file: name, file variable, buffer */
PREFIX char datafile_name[STRING_MAX];
PREFIX FILE *DF;
PREFIX char *data_buf;

/* History file: name, file variable, buffer */
PREFIX char historyfile_name[STRING_MAX];
PREFIX FILE *HF;
PREFIX char *history_buf;

/* Global parameters */

/* Is the library initialized already? */
PREFIX short is_initialized;
/* If 0 => no short tracking
    > 0 => use value to determine the slice size */
PREFIX int slice_size;
/* Record information about anonymous vectors */
PREFIX short track_anonymous;
/* Output global trace */
PREFIX short global_tracking;
/* Type of compression performed on the history file */
PREFIX short compression_type;

/* Adxpy call counter */
PREFIX long ADXPY_CALLS;
/* Global multiplication counter */
PREFIX unsigned long MULT_COUNT;
/* Global addition counter */
PREFIX unsigned long ADD_COUNT;
/* Global nonzeroes histogram */
PREFIX int nz_hist[NZ_THRESHOLD+1];
PREFIX int max_nz_global, min_nz_global;

#undef PREFIX

#endif
