/*
  Support routines for the ADIFOR visualization library
  (precision-independent part)

  Vitaly Shmatikov, Summer 1994
*/

#include "advis_pi_include.h"
#include <limits.h>

void xvsini()
/* 
 Set up the context for the visualization library routines
*/
{
#ifdef _VTRACE
 char compression_command[STRING_MAX];

 /* Open the data file for writing and allocate buffer */
 DF = fopen(datafile_name,"w");
 /* data_buf = NULL; setvbuf(DF, data_buf, _IOFBF, 16384); */

 /* Open the pipe to a compression program via which
    the history file will be written; allocate buffer */
 switch (compression_type) {
    case NO_COMPRESSION:
         strcpy(compression_command, "cat > ");
         strcat(compression_command, historyfile_name);
	 break;
    case COMP_COMPRESS:
         strcpy(compression_command, "compress -c > ");
         strcat(compression_command, historyfile_name);
         strcat(compression_command, ".Z");
	 break;
    case COMP_GZIP:
         strcpy(compression_command, "gzip -c > ");
         strcat(compression_command, historyfile_name);
         strcat(compression_command, ".gz");
	 break;
    default:
	 fprintf(stderr, "Error: unknown compression!\n"); exit(1);
 }

 HF = popen(compression_command,"w");
 /* history_buf = NULL; setvbuf(HF, history_buf, _IOFBF, 16384); */

 if((DF == NULL) || (HF == NULL)) {
   fprintf(stderr, "Error: could not open file!\n"); exit(1);
 }

 /* If short tracking, write the slice size into the history file */
 if(slice_size)
   fprintf(HF, "$ %d\n", slice_size);

 /* Raise the flag: the library has been initialized */
 is_initialized = TRUE;
#endif /* _VTRACE */
}

void xvsdef()
/*
 Sets library parameters to their default values
*/
{
#ifdef _VTRACE
 /* Data file: name, file variable, buffer */
 strcpy(datafile_name, "advis.dat");
 strcpy(historyfile_name, "/tmp/history.dat");

 is_initialized = FALSE;
 slice_size = 100;             /* Default slice size = 100 */
 track_anonymous = FALSE;
 global_tracking = TRUE;
 compression_type = COMP_GZIP;

 max_nz_global = -1;
 min_nz_global = INT_MAX;

 ADXPY_CALLS = 0;
 MULT_COUNT = 0;
 ADD_COUNT = 0;
#endif /* _VTRACE */
}

void xvscnf(param, value)
/*
 Set the value of one of the library parameters
*/
int *param;
int *value;
{
#ifdef _VTRACE
 /* Depending on what param is, treat value differently */
 switch ((*param)) {
    case SHORT_TRACKING:
	 slice_size = (*value); break;
    case TRACK_ANONYMOUS:
	 track_anonymous = (*value); break;
    case GLOBAL_TRACKING:
	 global_tracking = (*value); break;
    case COMPRESSION_TYPE:
	 /* Cannot change compression type if already initialized */
	 if(is_initialized) {
	   fprintf(stderr, "Error: change of compression type ");
	   fprintf(stderr, "after initialization!\n"); exit(1); }
         else
	   compression_type = (*value);
         break;
    case DATAFILE_NAME:
	 if(is_initialized) {
	   fprintf(stderr, "Error: change of compression type ");
	   fprintf(stderr, "after initialization!\n"); exit(1); }
         else
	   strcpy(datafile_name, f_to_c_str((char*)value));
	 break;
    case HISTORYFILE_NAME:
	 if(is_initialized) {
	   fprintf(stderr, "Error: change of compression type ");
	   fprintf(stderr, "after initialization!\n"); exit(1); }
         else
	   strcpy(historyfile_name, f_to_c_str((char*)value));
	 break;
    default:
	 fprintf(stderr, "Error: unknown visual parameter!\n"); exit(1);
 }
#endif /* _VTRACE */
}

void xvscls()
/*
 Clean up
*/
{
#ifdef _VTRACE
 char script_name[STRING_MAX];

 /* Force collect_visual to output */
 collect_visual(0, 0, 0, FALSE, 0, 0, 0, 0, TRUE);
 
 /* Close the data file */
 fclose(DF);
 /* Close the history pipe */
 pclose(HF);

 /* Copy the history file to the current directory */
 strcpy(script_name, "mv ");
 strcat(script_name,  historyfile_name); strcat(script_name, "* .");
 system(script_name);
#endif /* _VTRACE */
}

/*
  Converts a string from Fortran to C
*/
char *f_to_c_str(s)
char* s;
{
#ifdef _VTRACE
 return s;
#endif /* _VTRACE */
}
