#define COMPLEX
/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#include  <stdlib.h>
#include "cgaxpy_pd_macros_and_data_structures.h"
#include "cgaxpy_pd_global_variables.h"

/****************************************************************************
 *                                                                          *
 *     cspg1 performs the arity 1 gaxpy.                                   *
 *                                                                          *
 ****************************************************************************/

void cspg1(destptrptr,alphaptr,vectoraptrptr)
cvectorstruct **vectoraptrptr,**destptrptr;
complex *alphaptr;
{

  int arity = 1;

  cvectorstruct *rhscvectorptr[1];
  cvectorstruct *tempvectorptr;
  
  cbucketstruct *rhscbucketptr;
  cbucketstruct *lhscvectortailbucketptr;
  cbucketstruct *lhscbucketptr;
  cbucketstruct *tempbucketptr;
  
  int i,k;
  int bucketindex;
  
  cvbucketstruct *rhscvectorvbucketptr;
  cvbucketstruct *lhscvectortailvbucketptr;
  cvbucketstruct *tempvbucketptr;
  int vbucketindex;
  
  cibucketstruct *rhscvectoribucketptr;
  cibucketstruct *lhscvectortailibucketptr;
  cibucketstruct *tempibucketptr;
  cibucketstruct *lhscibucketptr;
  int ibucketindex;

  complex *alphaptrs[1];
  
#ifdef _TRACE
  int nnzb,nnza;
  cvectorstruct *lhscvectoraddress;
  
  
  cvectorstruct **vectorptrptrs[1];
  
  vectorptrptrs[0] = vectoraptrptr;
  
  if(*destptrptr==NULL){ 
    nnzb = 0; 
    lhscvectoraddress= (cvectorstruct *) 99900; 
  }
  else {
    nnzb = (*destptrptr)->nonzeroes; 
    lhscvectoraddress = *destptrptr;
  }
#endif
  
  
  /****************************************
   * Initialization of *destptrptr.       *
   ****************************************/
  
  if (*destptrptr == NULL)
    cgetvector(destptrptr);
  
  rhscvectorptr[0] = *vectoraptrptr;
  alphaptrs[0] = alphaptr;

#ifdef _PEEP
  if (toggle_adxpy_1_peep) {
    adpeep_header(1);
    cadpeep_pre(0,vectoraptrptr,alphaptr);
  }
#endif
  
  if (rhscvectorptr[0]->nonzeroes == 0 || X_IS_ZERO(*alphaptr)) {
    cspzronotca(destptrptr);
  }
  else /* if not ZERO AXPY */ {
    if (rhscvectorptr[0] == *destptrptr &&
	rhscvectorptr[0]->dstype == CS) {
      plusequalcspg1(destptrptr,alphaptr);
    }
    else /* if (not plusequalcspg1 or (dstype is S1 or SS)) */ {
      lhscvectorptr->nonzeroes = rhscvectorptr[0]->nonzeroes;
      lhscvectorptr->dstype = rhscvectorptr[0]->dstype;

      if (lhscvectorptr->dstype == S1) {
	
	/********************************************************************
	 *                       ONE_SUBSCRIPT AXPY                         *
	 ********************************************************************/


	lhscvectorptr->cds.s1.index =
	                rhscvectorptr[0]->cds.s1.index;
	GAXPY_EQUALS(lhscvectorptr->cds.s1.value,*alphaptr,
		        rhscvectorptr[0]->cds.s1.value);
      }
      else if (lhscvectorptr->dstype == SS) {
	
	/******************************************************************
	 *                 _SINGLE_SUBSCRIPTS AXPY                        *
	 ******************************************************************/
	
	rhscbucketptr = rhscvectorptr[0]->cds.ss.bucketptr;
	for (i=0;i<rhscvectorptr[0]->nonzeroes;i++) {
	  if ((bucketindex = MY_MOD(i,cBUCKET_DIM_minus1)) == 0) {
	    if (i == 0) {
	      cgetbucket(&lhscbucketptr);
	      lhscvectorptr->cds.ss.bucketptr = lhscbucketptr;
	    }
	    else {
	      rhscbucketptr = rhscbucketptr->nextbucket;
	      cgetbucket(&lhscbucketptr);
	      lhscvectortailbucketptr->nextbucket = lhscbucketptr;
	    }
	    lhscvectortailbucketptr = lhscbucketptr;
	  }
	  GAXPY_EQUALS(lhscvectortailbucketptr->value[bucketindex],
		       *alphaptr,rhscbucketptr->value[bucketindex]);
	  lhscvectortailbucketptr->index[bucketindex] =
	    rhscbucketptr->index[bucketindex];
	}
      }
      
      else /* if (lhscvectorptr->dstype == CS) */ {
	  
	/******************************************************************
	 *                  _COMPRESSED_SUBSCRIPTS AXPY                   *
	 ******************************************************************/
	
	lhscvectorptr->cds.cs.nonzeroranges =
	  rhscvectorptr[0]->cds.cs.nonzeroranges;
	rhscvectoribucketptr =
	  rhscvectorptr[0]->cds.cs.ibucketptr;
	rhscvectorvbucketptr =
	  rhscvectorptr[0]->cds.cs.vbucketptr;
	
	for (i=0;i<rhscvectorptr[0]->cds.cs.nonzeroranges;i++) {
	  if ((ibucketindex = MY_MOD(i,cIBUCKET_DIM_minus1)) == 0) {
	    if (i == 0) {
	      cgetibucket(&lhscibucketptr);
	      lhscvectorptr->cds.cs.ibucketptr = lhscibucketptr;
	    }
	    else {
	      rhscvectoribucketptr =
		rhscvectoribucketptr->nextibucket;
	      cgetibucket(&lhscibucketptr);
	      lhscvectortailibucketptr->nextibucket = lhscibucketptr;
	    }
	    lhscvectortailibucketptr = lhscibucketptr;
	  }
	  
	  lhscvectortailibucketptr->index_RANGE_BEGIN[ibucketindex] =
	    rhscvectoribucketptr->index_RANGE_BEGIN[ibucketindex];
	  lhscvectortailibucketptr->index_RANGE_END[ibucketindex] =
	    rhscvectoribucketptr->index_RANGE_END[ibucketindex];
	}
	
	for (i=0;i<rhscvectorptr[0]->nonzeroes;i++) {
	  if ((vbucketindex = MY_MOD(i,cVBUCKET_DIM_minus1)) == 0) {
	    if (i == 0) {
	      cgetvbucket(&lhscvbucketptr);
	      lhscvectorptr->cds.cs.vbucketptr = lhscvbucketptr;
	    }
	    else {
	      rhscvectorvbucketptr =
		rhscvectorvbucketptr->nextvbucket;
	      cgetvbucket(&lhscvbucketptr);
	      lhscvectortailvbucketptr->nextvbucket = lhscvbucketptr;
	    }
	    lhscvectortailvbucketptr = lhscvbucketptr;
	  }
	  GAXPY_EQUALS(lhscvectortailvbucketptr->value[vbucketindex],
		       *alphaptr,
		       rhscvectorvbucketptr->value[vbucketindex]);
	}
      } /* matches     else ** if (lhscvectorptr->dstype == CS) ** {  */
    
      caxpy_0_1_post_processing(destptrptr);

    } /* matches   else ** if (not plusequalcspg1) ** {                   */
  } /* matches   else ** if not ZERO AXPY ** {                             */
  

#ifdef _PEEP
  if (toggle_adxpy_1_peep) {
    cadpeep_post(destptrptr);
  }
#endif


#ifdef _TRACE 
  nnza = (*destptrptr)->nonzeroes; 
  cadinfo(arity,nnzb,lhscvectoraddress,nnza,alphaptrs,vectorptrptrs);
#endif
  
}
  
