#define COMPLEX
/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#include  <stdlib.h>
#include "cgaxpy_pd_macros_and_data_structures.h"
#include "cgaxpy_pd_global_variables.h"

/****************************************************************************
 *                                                                          *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/

void cspg2(destptrptr,alphaptr,vectoraptrptr,betaptr,vectorbptrptr)
cvectorstruct **destptrptr;
complex *alphaptr;
cvectorstruct **vectoraptrptr;
complex *betaptr;
cvectorstruct **vectorbptrptr;
{
  
  int arity = 2;
  
  cvectorstruct *rhscvectorptr[2];
  
  cbucketstruct *rhscbucketptr[2];
  
  cbucketstruct *lhscbucketptr;
  
  cibucketstruct *rhscibucketptr[2];
                                      /* Array of pointers to "current"     *
				       * buckets where "current" nonzeroes  *
				       * of the rhs_vectors reside.         */

  cibucketstruct *lhscibucketptr;  /* Pointer to current bucket of       *
				       * lhscvector.                         */

  complex alpha[2];          /* Dereferenced alphaptrs[].          */
  complex *alphaptrs[2];

  int rhscbucketindex[2];              /* Array of indices of current   *
				       * nonzeroes in the current buckets   *
				       * of rhs_vectors.                    */

  int lhscbucketindex;                 /* bucketindex of next nonzero to be  *
				       * added to lhscvector.                */

  int lastlhscbucketindex;             /* bucketindex of the latest added    *
				       * nonzero in the lhscvector.  We need *
				       * this additional index to leap-frog *
				       * buckets, if need be.               */


  int rhscibucketindex[2];             /* Array of indices of current        *
				       * nonzeroes in the current buckets   *
				       * of rhs_vectors.                    */

  int lhscibucketindex;                /* bucketindex of next nonzero to be  *
				       * added to lhscvector.                */

  int lastlhscibucketindex;            /* bucketindex of the latest added    *
				       * nonzero in the lhscvector.  We need *
				       * this additional index to leap-frog *
				       * buckets, if need be.               */

  int k;                              /* Counter for arity.                 */

  int j;                              /* Index for vectorsubscripts to be   *
				       * entered in (and later popped from) *
				       * the heap.                          */

  int rhscvectornonzeroes[2];          /* Array of total # of nonzeroes in   *
				       * each rhs_vector.                   */

  int rhscvectorconsumednonzeroes[2];  /* Array of # of nonzeroes consumed   *
				       * so far in each rhs_vector.         */

  int lhscvectornonzeroes;             /* Count of # of nonzeroes entered    *
				       * into the lhscvector so far.  By the *
				       * end of the saxpy operation, this   *
				       * index equals the total # of        *
				       * nonzeroes in the resultant vector. */

/*
 * These variables were originally for COMPRESSED_SUBSCRIPTS
 */

  cvbucketstruct *lastlhscvbucketptr; 
                                      /* Pointer to current vbucket of      *
				       * lhscvector.                    */

  int rhscvectornonzeroranges[2];
                                      /* Array of total # of nonzeroes      *
				       * ranges in each rhs_vector.         */

  int rhscvectorconsumednonzeroranges[2];
                                      /* Array of # of nonzeroes consumed   *
				       * so far in each rhs_vector.         */

  int rhscvbucketindex[2];             /* Array of indices of current        *
				       * nonzeroes in the current buckets   *
				       * of rhs_vectors.                    */

  int lhscvectornonzeroranges;         /* Count of # of nonzeroes entered    *
				       * into the lhscvector so far.  By the *
				       * end of the saxpy operation, this   *
				       * index equals the total # of        *
				       * nonzeroes in the resultant vector. */

  int lhscvbucketindex;                /* bucketindex of next nonzero to be  *
				       * added to lhscvector.                */


  int lastlhscvbucketindex;            /* bucketindex of the latest added    *
				       * nonzero in the lhscvector.  We need *
				       * this additional index to leap-frog *
				       * buckets, if need be.               */

  cvbucketstruct *templhscvbucketptr;
  int templhscvbucketindex;

  int j1;                             /* Index for vectorsubscripts to be   *
				       * entered in (and later popped from) *
				       * the heap.                          */

  int j2;                             /* Index for vectorsubscripts to be   *
				       * entered in (and later popped from) *
				       * the heap.                          */

  int h1;
  int h2;

  int totalcurrentnonzerorange;

#ifdef _DEBUG
  int unconsumedcurrentnonzerorange;
#endif

  int distancetooverlap;
  int distancetoendofvbucket;

  complex tmp_add,tmp_mult1,tmp_mult2;


  /*
   * These variables were added for union ds
   */
  
  cvectorstruct *restorerhscvectorptr[MAXARITY+1];

  int governingdstype;
  
  /*
   * These variables were added for plusequal code
   */
  
  cvectorstruct **rhscvectorptrptrs[MAXARITY];
  
  int condition[2];
  
#ifdef _PLUSEQUAL  
  int smallest_rhs_index; 
#endif /* _PLUSEQUAL */
  
  
#ifdef _TRACE
  int nnzb,nnza;
  cvectorstruct *lhscvectoraddress;
  
  cvectorstruct **vectorptrptrs[2];
  
  vectorptrptrs[0] = vectoraptrptr;
  
  vectorptrptrs[1] = vectorbptrptr;
  
  if(*destptrptr==NULL){ 
    nnzb = 0; 
    lhscvectoraddress= (cvectorstruct *) 99900; 
  }
  else {
    nnzb = (*destptrptr)->nonzeroes; 
    lhscvectoraddress = *destptrptr;
  }
#endif
  
  
  /****************************************
   * Initialization of *destptrptr.       *
   ****************************************/
  
  if (*destptrptr == NULL)
    cgetvector(destptrptr);
  
  /****************************************
   * Dereferencing.                       *
   ****************************************/
  
  alpha[0] = *alphaptr;
  rhscvectorptr[0] = *vectoraptrptr;
  alphaptrs[0] = alphaptr;
  
#ifdef _PEEP
  adpeep_header(2);
  cadpeep_pre(0,vectoraptrptr,alphaptr);
#endif    

  alpha[1] = *betaptr;
  rhscvectorptr[1] = *vectorbptrptr;
  alphaptrs[1] = betaptr;

#ifdef _PEEP
  cadpeep_pre(1,vectorbptrptr,betaptr);
#endif    
 
  for (k=0;k<arity;k++) {
    restorerhscvectorptr[k] = rhscvectorptr[k];
    rhscvectornonzeroes[k] = rhscvectorptr[k]->nonzeroes;
    if (rhscvectorptr[k]->dstype == CS)
      rhscvectornonzeroranges[k] =
	rhscvectorptr[k]->cds.cs.nonzeroranges;
  }
  
#ifdef _PLUSEQUAL
  smallest_rhs_index = DUMMYSUBSCRIPT;
  for (k=0;k<arity;k++) {
    if (rhscvectorptr[k]->dstype == S1) 
      smallest_rhs_index = min (smallest_rhs_index,
				rhscvectorptr[k]->cds.s1.index);
    else if (rhscvectorptr[k]->dstype == SS)
      smallest_rhs_index =
	min (smallest_rhs_index,
	     rhscvectorptr[k]->cds.ss.bucketptr->index[0]);
    else /* if (rhscvectorptr[k]->dstype == CS) */
      smallest_rhs_index =
	min (smallest_rhs_index,
	     rhscvectorptr[k]->cds.cs.ibucketptr->index_RANGE_BEGIN[0]);
  }
#endif /* _PLUSEQUAL */


  if ((rhscvectorptr[0]->nonzeroes == 0 || X_IS_ZERO(alpha[0])) &&
      (rhscvectorptr[1]->nonzeroes == 0 || X_IS_ZERO(alpha[1]))) {
    cspzronotca(destptrptr);
  }
  else if (rhscvectorptr[0]->nonzeroes == 0 || X_IS_ZERO(alpha[0])) {
    if (rhscvectorptr[1] == *destptrptr &&
	rhscvectorptr[1]->dstype == CS) {
      plusequalcspg1(destptrptr,betaptr);
    }
    /* if (rhscvectorptr[1] != *destptrptr ||
       rhscvectorptr[1]->dstype != CS)           */
    else {
#ifdef _PEEP
      toggle_adxpy_1_peep = 0;
      cspg1(destptrptr,betaptr,vectorbptrptr);
      toggle_adxpy_1_peep = 1;
#else
      cspg1(destptrptr,betaptr,vectorbptrptr);
#endif     
    }
  }
  else if (rhscvectorptr[1]->nonzeroes == 0 || X_IS_ZERO(alpha[1])) {
    if (rhscvectorptr[0] == *destptrptr &&
	rhscvectorptr[0]->dstype == CS) {
      plusequalcspg1(destptrptr,alphaptr);
    }
    /* else if (rhscvectorptr[0] != *destptrptr) ||
       rhscvectorptr[0]->dstype != CS)          */
    else {      
#ifdef _PEEP
      toggle_adxpy_1_peep = 0;
      cspg1(destptrptr,alphaptr,vectoraptrptr);
      toggle_adxpy_1_peep = 1;
#else
      cspg1(destptrptr,alphaptr,vectoraptrptr);
#endif
    }
  }
  else /* neither rhs_operation is a NO-OP */ {
    if (rhscvectorptr[0] == *destptrptr &&
	rhscvectorptr[1] == *destptrptr &&
	rhscvectorptr[0]->dstype == CS) {
#ifdef _DEBUG
      printf("cspg2 warning: lhscvector appears more than ");
      printf("once on rhs.\n");
#endif
      X_ADD(tmp_add,*alphaptr,*betaptr);
      plusequalcspg1(destptrptr,&tmp_add);
    }

#ifdef _PLUSEQUAL
    else if
      (rhscvectorptr[0] == *destptrptr &&
       rhscvectorptr[0]->dstype == CS &&
       rhscvectorptr[0]->cds.cs.ibucketptr->index_RANGE_BEGIN[0] <=
       smallest_rhs_index &&
       X_IS_ONE(alpha[0])) {
	rhscvectorptrptrs[1] = &rhscvectorptr[1];
	if ((*rhscvectorptrptrs[1])->dstype != CS &&
	    (*rhscvectorptrptrs[1])->nonzeroes !=0) {
	  cconvertdstype(rhscvectorptrptrs[1],1,CS);
	}
	plusequalcspg2(destptrptr,
			betaptr,rhscvectorptrptrs[1],ORIG_OTHER_VECTOR);
      }
    else if
      (rhscvectorptr[1] == *destptrptr &&
       rhscvectorptr[1]->dstype == CS &&
       rhscvectorptr[1]->cds.cs.ibucketptr->index_RANGE_BEGIN[0] <=
       smallest_rhs_index &&
       X_IS_ONE(alpha[1])) {
	rhscvectorptrptrs[0] = &rhscvectorptr[0];
	if ((*rhscvectorptrptrs[0])->dstype != CS &&
	    (*rhscvectorptrptrs[0])->nonzeroes !=0) {
	  cconvertdstype(rhscvectorptrptrs[0],0,CS);
	}
	plusequalcspg2(destptrptr,
			alphaptr,rhscvectorptrptrs[0],ORIG_OTHER_VECTOR);
	
      }
    
    /* 
     * else if neither rhs_operation is a NO-OP, and neither
     * rhscvector is (the same as the lhscvector AND has
     * multiplier equal to 1 AND is CS)
     */
    
    else if (rhscvectorptr[0] ==  rhscvectorptr[1]) {
#ifdef _DEBUG
      printf("cspg2 warning: lhscvector appears more than ");
      printf("once on rhs.\n");
#endif
      X_ADD(tmp_add,*alphaptr,*betaptr);
#ifdef _PEEP
      toggle_adxpy_1_peep = 0;
      cspg1(destptrptr,&tmp_add,vectoraptrptr);
      toggle_adxpy_1_peep = 1;
#else
      cspg1(destptrptr,&tmp_add,vectoraptrptr);
#endif
    }
#endif /* _PLUSEQUAL */
    
    else /* do an honest to goodness cspg2 */ {
      
      governingdstype = max(rhscvectorptr[0]->dstype,
			    rhscvectorptr[1]->dstype);
      if (governingdstype == S1 &&
	  rhscvectorptr[0]->cds.s1.index != rhscvectorptr[1]->cds.s1.index)
	governingdstype = SS;
      for (k=0;k<arity;k++)
	if (rhscvectorptr[k]->dstype != governingdstype) {
	  cconvertdstype((&rhscvectorptr[k]),k,governingdstype);
	  if (governingdstype == CS)
	    rhscvectornonzeroranges[k] =
	      rhscvectorptr[k]->cds.cs.nonzeroranges;
	}
      
      if (governingdstype == S1) {
	
	/******************************************************************
	 *                      ONE_SUBSCRIPT AXPY                        *
	 ******************************************************************/

	lhscvectorptr->nonzeroes =  1;
	lhscvectorptr->dstype = governingdstype;
	lhscvectorptr->cds.s1.index =
	  rhscvectorptr[0]->cds.s1.index;
        X_MULT(tmp_mult1,*alphaptr,rhscvectorptr[0]->cds.s1.value);
        X_MULT(tmp_mult2,*betaptr,rhscvectorptr[1]->cds.s1.value);
        X_ADD(lhscvectorptr->cds.s1.value,tmp_mult1,tmp_mult2);

	caxpy_0_1_post_processing(destptrptr);

      }
      else /* if we are going to use lhscvectorptr */ {
	if (governingdstype == SS) {
	  
	  /****************************************************************
	   *                 _SINGLE_SUBSCRIPTS AXPY                      *
	   ****************************************************************/
	  for (k=0;k<arity;k++) {
	    rhscvectorconsumednonzeroes[k]= 0;
	    rhscbucketindex[k] = 0;
	    rhscbucketptr[k] = rhscvectorptr[k]->cds.ss.bucketptr;
	  }
	  cgetbucket(&lhscbucketptr);
	  lhscvectorptr->dstype = governingdstype;
	  lhscvectorptr->cds.ss.bucketptr = lhscbucketptr;
	  lhscvectornonzeroes = 0;
	  
	  for (k=0;k<arity;k++)
	    condition[k] = rhscvectorconsumednonzeroes[k] <
	      rhscvectorptr[k]->nonzeroes;
	  
	  while (condition[0] || condition[1]) {
	    if (condition[0] && condition[1])
	      if (rhscbucketptr[0]->index[rhscbucketindex[0]] <=
		  rhscbucketptr[1]->index[rhscbucketindex[1]])
		k = 0;
	      else
		k = 1;
	    else if (condition[0])
	      k = 0;
	    else
	      k = 1;
	    
	    j = rhscbucketptr[k]->index[rhscbucketindex[k]];
	    
	    lhscbucketindex = MY_MOD(lhscvectornonzeroes,
					 cBUCKET_DIM_minus1);
	    
	    /****************************************
	     * Perform saxpy:                       *
	     * 3.1. Found 1st or new subscript for  *
	     *      lhscvector.  If needed, get new  *
	     *      bucket for lhscvector.           *
	     *      Consume nonzero in saxpy.       *
	     ****************************************/
	    if (lhscvectornonzeroes == 0 ||
		lhscbucketptr->index[lastlhscbucketindex] < j) {
	      if (lhscbucketindex == 0 && lhscvectornonzeroes != 0) {
		cgetbucket(&(lhscbucketptr->nextbucket));
		lhscbucketptr = lhscbucketptr->nextbucket;
	      }
	      lhscbucketptr->index[lhscbucketindex] = j;
	      GAXPY_EQUALS(
                lhscbucketptr->value[lhscbucketindex],
		alpha[k],
		rhscbucketptr[k]->value[rhscbucketindex[k]]);
	      lastlhscbucketindex = lhscbucketindex;
	      lhscvectornonzeroes++;
	    }
	    
	    /****************************************
	     * Perform saxpy:                       *
	     * 3.2. Found old subscript for         *
	     *      lhscvector.                      *
	     *      Consume nonzero in saxpy.       *
	     ****************************************/
	    else if (lhscbucketptr->index[lastlhscbucketindex] == j) {
	      GAXPY_PLUS_EQUALS(
                lhscbucketptr->value[lastlhscbucketindex],
		alpha[k],
		rhscbucketptr[k]->value[rhscbucketindex[k]]);
	    }
	    
	    /****************************************
	     * Perform saxpy:                       *
	     * 3.3. Found wrong subscript for       *
	     *      lhscvector.  Abort.               *
	     ****************************************/
	    
#ifdef _DEBUG
	    else {
	      printf("cspg2 error 6: lhscnz=%d, lastlhscindex[]=%d, j=%d\n",
		     lhscvectornonzeroes,
		     lhscbucketptr->index[lastlhscbucketindex],j);
	      abort();
	    }
#endif
	    
	    /****************************************
	     * Perform saxpy:                       *
	     * 3.4. If there is one, get new        *
	     *      nonzero from the rhs_vector     *
	     *      whose nonzero was just          *
	     *      consumed and replace the        *
	     *      heaproot with new entry;        *
	     *      Else, delete heaproot.          *
	     ****************************************/
	    rhscvectorconsumednonzeroes[k]++;
	    condition[k] = rhscvectorconsumednonzeroes[k] <
	      rhscvectorptr[k]->nonzeroes;
	    rhscbucketindex[k] = MY_MOD(rhscvectorconsumednonzeroes[k],
					    cBUCKET_DIM_minus1);
	    if (condition[k]) {
	      if (rhscbucketindex[k]  == 0) {
		if ((rhscbucketptr[k] = rhscbucketptr[k]->nextbucket) 
		    == NULL) {
#ifdef _DEBUG
		  printf("cspg2 err: unexpected NULL nextbucket: k=%d",k);
		  printf("rhscvectorconsumednonzeros=%d,rhscvecnonzeroes=%d,",
			 rhscvectorconsumednonzeroes[k],
			 rhscvectornonzeroes[k]);
		  abort();
#endif
		}	  
	      }
	    }
	  } /* matches     while loop     */
	} /* matches     else if (governingdstype == SS) {      */
	
	else /* if (governingdstype == CS) */ {
	  
	  /****************************************************************
	   *               _COMPRESSED_SUBSCRIPTS AXPY                    *
	   ****************************************************************/
	  
	  /****************************************
	   * Else perform saxpy:                  *
	   * 1. Put in heap the 1st nonzero of    *
	   *    each vector.                      *
	   ****************************************/
	  for (k=0;k<arity;k++) {
	    rhscvectorconsumednonzeroranges[k] = 0;
	    rhscibucketindex[k] = 0;
	    rhscvbucketindex[k] = 0;
	    rhscibucketptr[k] =
	      rhscvectorptr[k]->cds.cs.ibucketptr;
	    rhscvbucketptr[k] =
	      rhscvectorptr[k]->cds.cs.vbucketptr;
	    
#ifdef _DEBUG
	    if (rhscibucketptr[k] == NULL ||
		rhscvbucketptr[k] == NULL) {
	      printf("cspg2 error 10: Unexpected NULL rhscbucketptrs: ");
	      printf("vector # %d\n",k);
	      abort();
	    }
#endif
	  }	  
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 2. Initialize lhscvector.             *
	   ****************************************/
	  cgetibucket(&lhscibucketptr);
	  lhscvectorptr->dstype = governingdstype;
	  lhscvectorptr->cds.cs.ibucketptr = lhscibucketptr;
	  cgetvbucket(&lhscvbucketptr);
	  lhscvectorptr->cds.cs.vbucketptr = lhscvbucketptr;
	  lhscvectornonzeroranges = 0;
	  lhscvectornonzeroes = 0;
	  lhscvbucketindex = 0;
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3. Get nonzero by popping the        *
	   *    heaproot; consume it in saxpy;    *
	   *    either replace or delete heaproot.*
	   ****************************************/
	  
	  while (rhscvectorconsumednonzeroranges[0] <
		 rhscvectorptr[0]->cds.cs.nonzeroranges ||
		 rhscvectorconsumednonzeroranges[1] <
		 rhscvectorptr[1]->cds.cs.nonzeroranges) {
	    if
	      ((rhscvectorconsumednonzeroranges[1] ==
		rhscvectorptr[1]->cds.cs.nonzeroranges) ||
	       ((rhscvectorconsumednonzeroranges[0] <
		 rhscvectorptr[0]->cds.cs.nonzeroranges) &&
		(rhscibucketptr[0]->index_RANGE_BEGIN[rhscibucketindex[0]] <=
		 rhscibucketptr[1]->index_RANGE_BEGIN[rhscibucketindex[1]])))
		k = 0;
	    else if
	      ((rhscvectorconsumednonzeroranges[0] ==
		rhscvectorptr[0]->cds.cs.nonzeroranges) ||
	       ((rhscvectorconsumednonzeroranges[1] <
		 rhscvectorptr[1]->cds.cs.nonzeroranges) &&
		(rhscibucketptr[0]->index_RANGE_BEGIN[rhscibucketindex[0]] >
		 rhscibucketptr[1]->index_RANGE_BEGIN[rhscibucketindex[1]])))
		k = 1;
#ifdef _DEBUG
	    else {
	      printf("cspg2 error: BAD LOGIC.\n");
	      abort();
	    }
#endif
	    j1 = rhscibucketptr[k]->index_RANGE_BEGIN[rhscibucketindex[k]];
	    j2 = rhscibucketptr[k]->index_RANGE_END[rhscibucketindex[k]];
	    rhscibucketindex[k] = 
	      MY_MOD(rhscvectorconsumednonzeroranges[k],
		     cIBUCKET_DIM_minus1);
	    lhscibucketindex = MY_MOD(lhscvectornonzeroranges,
					  cIBUCKET_DIM_minus1);
	    
	    if ((rhscvbucketindex[k] = MY_MOD(rhscvbucketindex[k],
						  cVBUCKET_DIM_minus1))==0
		&& rhscvectorconsumednonzeroranges[k] !=0) {
	      cHOP_TO_NEXT_RHSVBUCKET(k,&rhscvbucketindex[k],
				     rhscvectorconsumednonzeroranges[k],
				     rhscvectornonzeroranges[k]);
	    }
	    lhscvbucketindex = MY_MOD(lhscvbucketindex,
					  cVBUCKET_DIM_minus1);
	    
#ifdef _DEBUG
	    if (rhscibucketptr[k]->errorflag != 9099 ||
		rhscvbucketptr[k]->errorflag != 9009) {
	      printf("cspg2 err 14: memory corruption of cibucketstore ");
	      printf("or cvbucketstore.\n");
	      abort();
	    }
#endif      
	    
	    /****************************************
	     * Perform saxpy:                       *
	     * 3.0. Check for startup errors.       *
	     ****************************************/
	    
#ifdef _DEBUG
	    if (lhscvectornonzeroranges == 0) {
	      if (rhscibucketindex[k] != 0 || rhscvbucketindex[k] != 0 ||
		  lhscibucketindex != 0 || lhscvbucketindex != 0) {
		printf("cspg2 error 15: Zeroes expected: \n\t");
		printf("rhscibucketindex[k]=%d, rhscvbucketindex[k]=%d\n\t",
		       rhscibucketindex[k], rhscvbucketindex[k]);
		printf("lhscibucketindex=%d, lhscvbucketindex=%d\n\t",
		       lhscibucketindex,lhscvbucketindex);
		abort();
	      }
	    }
#endif      
	    
	    /****************************************
	     * Perform saxpy:                       *
	     * 3.1. Found 1ST or NEW subscript      *
	     *      range (non-overlapping with     *
	     *      previous one, if any) for       *
	     *      lhscvector.  If needed, get new  *
	     *      ibuckets and vbuckets for       *
	     *      lhscvector.                      *
	     ****************************************/
	    if (lhscvectornonzeroranges == 0 ||
		j1 > lhscibucketptr->index_RANGE_END[lastlhscibucketindex]+1) {/*A.1*/
	      
	      if (lhscibucketindex == 0 && lhscvectornonzeroranges != 0) {
		cgetibucket(&(lhscibucketptr->nextibucket));
		lhscibucketptr = lhscibucketptr->nextibucket;
	      }
	      lhscibucketptr->index_RANGE_BEGIN[lhscibucketindex] = j1;
	      lhscibucketptr->index_RANGE_END[lhscibucketindex] = j2;
	      
	      if (lhscvbucketindex == 0 && lhscvectornonzeroranges != 0)
		cget_and_hop_to_next_vbucket(&lhscvbucketindex);
	      
	      lastlhscvbucketptr = lhscvbucketptr;
	      lastlhscvbucketindex = lhscvbucketindex;
	      
	      totalcurrentnonzerorange = j2-j1+1;
	      
	      lhscvectornonzeroes += totalcurrentnonzerorange;
	      lhscvectornonzeroranges++;
	      
	
	      cNON_OVER_LAPPING_CS_RANGE_AXPY(COMPLETELY_NEW_RANGE,
					     lhscibucketindex,
					     &lastlhscibucketindex,
					     &lhscvbucketindex,
					     k,
					     &rhscvbucketindex[k],
					     alpha[k],
					     totalcurrentnonzerorange,
					     &unconsumedcurrentnonzerorange,
					     rhscvectorconsumednonzeroranges[k],
					     rhscvectornonzeroranges[k],
					     lhscvectornonzeroranges);
	      
	            
	    } /*A.1*/
	    
	    
	    /****************************************
	     * Perform saxpy:                       *
	     * 3.2. Found old subscript for         *
	     *      lhscvector.  Consume  nonzero in *
	     *      saxpy.                          *
	     *                                      *
	     ****************************************/
	    else if (j1 <= lhscibucketptr->index_RANGE_END[lastlhscibucketindex]+1 &&
		     j1 >= lhscibucketptr->index_RANGE_BEGIN[lastlhscibucketindex]
		     ) { /*B.1*/
	      
	      templhscvbucketptr = lhscvbucketptr;
	      templhscvbucketindex = lhscvbucketindex;
	      lhscvbucketptr = lastlhscvbucketptr;
	      
	      distancetooverlap =
		j1 - lhscibucketptr->index_RANGE_BEGIN[lastlhscibucketindex];
	      distancetoendofvbucket = cVBUCKET_DIM -
		lastlhscvbucketindex;
	      
	      if (j1 ==
		  lhscibucketptr->index_RANGE_END[lastlhscibucketindex]+1) {
		h1 = 0;
		h2 = -1;
	      }
	      else if (j2 <= lhscibucketptr->index_RANGE_END[lastlhscibucketindex]
		       ) {
		h1 = j1;
		h2 = j2;
		j1 = 0;
		j2 = -1;
	      }
	      else {
		h1 = j1;
		h2 = lhscibucketptr->index_RANGE_END[lastlhscibucketindex];
		j1 = h2+1;
	      }
	      
	      if (distancetoendofvbucket > distancetooverlap) {
		lhscvbucketindex =
		  MY_MOD(lastlhscvbucketindex+distancetooverlap,
			 cVBUCKET_DIM_minus1);
	      }
	      else if (distancetoendofvbucket < distancetooverlap) {
		lhscvbucketindex =
		  MY_MOD(distancetooverlap - distancetoendofvbucket,
			 cVBUCKET_DIM_minus1);
	      }
	      else {
		lhscvbucketindex = 0;
	      }
	      
	      while (distancetooverlap >= distancetoendofvbucket &&
		     lhscvbucketptr->nextvbucket != NULL) {
		lhscvbucketptr = lhscvbucketptr->nextvbucket;
		distancetooverlap -= distancetoendofvbucket;
		distancetoendofvbucket = cVBUCKET_DIM;
	      }
	      
	      lhscibucketptr->index_RANGE_END[lastlhscibucketindex] =
		max(j2,
		    lhscibucketptr->index_RANGE_END[lastlhscibucketindex]);
	      
	      if ((totalcurrentnonzerorange = h2-h1+1) > 0) {
		

		cOVERLAPPING_CS_RANGE_AXPY(&lhscvbucketindex,
					  k,
					  &rhscvbucketindex[k],
					  alpha[k],
					  totalcurrentnonzerorange,
					  &unconsumedcurrentnonzerorange,
					  rhscvectorconsumednonzeroranges[k],
					  rhscvectornonzeroranges[k],
					  lhscvectornonzeroranges);
		
	      }
	     
	      /*
	       * This bit of logic must be performed, whether or not there
	       * is additional nonzeroes to process.
	       */
	      
	      if (j2 != -1 &&
		  (lhscvbucketptr != templhscvbucketptr ||
		   (lhscvbucketindex = MY_MOD(lhscvbucketindex,
						  cVBUCKET_DIM_minus1)) !=
		   templhscvbucketindex)) {
		
#ifdef _DEBUG
		printf("cspg2 err 50.0: Unexpectd value discrepancies\n");
		printf("lhscvbucketptr=%d, templhscvbucketptr=%d\n",
		       (int)lhscvbucketptr,(int)templhscvbucketptr);
		printf("lhscvbucketindex=%d, templhscvbucketindex=%d\n",
		       lhscvbucketindex,templhscvbucketindex);
		
		abort();
#endif
	      }
	      else {
		lhscvbucketptr = templhscvbucketptr;
		lhscvbucketindex = templhscvbucketindex;
	      }
	      
	      
	      if ((totalcurrentnonzerorange = j2-j1+1) > 0) {
		lhscvectornonzeroes += totalcurrentnonzerorange;
		
		/*
		 * adding new entries to vbucket
		 */
		
		if (lhscvbucketindex == 0 &&
		    lhscvectornonzeroranges != 0 &&
		    j2 != -1)
		  cget_and_hop_to_next_vbucket(&lhscvbucketindex);
		
		if(rhscvbucketindex[k] == cVBUCKET_DIM && j2 != -1) {

		  cHOP_TO_NEXT_RHSVBUCKET(k,&rhscvbucketindex[k],
					 rhscvectorconsumednonzeroranges[k],
					 rhscvectornonzeroranges[k]);
		}
		
		
		
		cNON_OVER_LAPPING_CS_RANGE_AXPY(NEW_ENTRIES_PADDED_TO_OLD_RANGE,
					       lhscibucketindex,
					       &lastlhscibucketindex,
					       &lhscvbucketindex,
					       k,
					       &rhscvbucketindex[k],
					       alpha[k],
					       totalcurrentnonzerorange,
					       &unconsumedcurrentnonzerorange,
					       rhscvectorconsumednonzeroranges[k],
					       rhscvectornonzeroranges[k],
					       lhscvectornonzeroranges);
		
	      } /* matches     if ((totalcurrent...) { */
	    } /*B.1*/
	    
	    
	    /****************************************	
	     * Perform saxpy:                       *
	     * 3.3. Found wrong subscript for       *
	     *      lhscvector.  Abort.               *
	     ****************************************/
	    
#ifdef _DEBUG
	    else {
	      printf("cspg2 error 60: lhscvectornonzeroranges=%d,",
		     lhscvectornonzeroranges);
	      printf("lastlhscibucketindex=%d, \n\t",lastlhscibucketindex);
	      printf("lhscibucketptr->index[%d] = (%d,%d), j1=%d\n",
		     lastlhscibucketindex,
		     lhscibucketptr->index_RANGE_BEGIN[lastlhscibucketindex],
		     lhscibucketptr->index_RANGE_END[lastlhscibucketindex],
		     j1);
	      abort();
	    }
#endif
	    
	    
	    /****************************************
	     * Perform saxpy:                       *
	     * 3.4. If there is one, get new        *
	     *      nonzero from the rhs_vector     *
	     *      whose nonzero was just          *
	     *      consumed and replace the        *
	     *      heaproot with new entry;        *
	     *      Else, delete heaproot.          *
	     ****************************************/
	    if (++rhscvectorconsumednonzeroranges[k] <
		rhscvectornonzeroranges[k]) {
	      if ((rhscibucketindex[k] = 
		   MY_MOD(rhscvectorconsumednonzeroranges[k],
			  cIBUCKET_DIM_minus1)) == 0) {
		if ((rhscibucketptr[k] =
		     rhscibucketptr[k]->nextibucket) == NULL) {
		  
#ifdef _DEBUG
		  printf("cspg2 err 61: unexpected NULL nextvbucket: \n");
		  printf("\t k=%d, rhscvectorconsumednonzeroranges=%d,\n",
			 k,rhscvectorconsumednonzeroranges[k]);
		  printf("\t rhscvectornonzeroranges[k]=%d\n",
			 rhscvectornonzeroranges[k]);
		  abort();
#endif
		  
		}
	      }
	    }   
	  } /* matches     while loop          */
	} /* matches     else ** if (governingdstype == CS) ** {     */
	

	
	
	
	rhscvectorptrptrs[0] = &rhscvectorptr[0];
	rhscvectorptrptrs[1] = &rhscvectorptr[1];
	
	caxpy_2_heap_post_processing(destptrptr,
				    rhscvectorptrptrs,
				    alphaptrs,
				    restorerhscvectorptr,
				    arity,
				    lhscvectornonzeroes,
				    lhscvectornonzeroranges,
				    governingdstype);
	
      } /* matches     if we are going to use lhscvectorptr           */
    } /* matches     else ** do an honest to goodness cspg2 **  {        */
  } /* matches     else ** neither rhs_operation is a NO-OP  ** {         */
    
  
#ifdef _PEEP
  cadpeep_post(destptrptr);
#endif
  
#ifdef _TRACE 
  nnza = (*destptrptr)->nonzeroes; 
  cadinfo(arity,nnzb,lhscvectoraddress,nnza,alphaptrs,vectorptrptrs);
#endif
  
}

