#define COMPLEX
/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#include  <stdlib.h>
#include "cgaxpy_pd_macros_and_data_structures.h"
#include "cgaxpy_pd_global_variables.h"

/****************************************************************************
 *                                                                          *
 *  cspgx(destptrptr,arity,alphaptrs,rhscvectorptrptrs)             *
 *                                                                          *
 *     destptrptr      is a doubly-indirected pointer to a cvectorstruct*
 *                     It is passed to cspgx by the adxpyN routines,    *
 *                     where 1 <= N <= MAXARITY.  In turn destptrptr is an  *
 *                     argument of the adxpyN functions which are called by *
 *                     ADIFOR-generated FORTRAN calls.  The purpose of the  *
 *                     double indirection is that FORTRAN calls by          *
 *                     reference, that is it puts the address of say g$x(1) *
 *                     on the stack.  In the FORTRAN code, the value of     *
 *                     g$x(1) represents a pointer to (address of) a sparse *
 *                     vector (i.e., cvectorstruct *).                  *
 *                     The FORTRAN call,                                    *
 *                     CALL adxpyN(g$x(1),...), causes inside C for an      *
 *                     internal copy, say destptrptr, to equal the address  *
 *                     of g$x(1), i.e., destptrptr == &g$x(1).  Therefore,  *
 *                     *destptrptr == g$x(1). But since the value contained *
 *                     in g$x(1) is (as interpreted by C) of type           *
 *                     (cvectorstruct *), destptrptr needs              *
 *                     one more level of                                    *
 *                     indirection, and is declared as                      *
 *                     (cvectorstruct **).                              *
 *                                                                          *
 *     arity           is the # of vectors on the rhs.                      *
 *                                                                          *
 *     alphaptrs       Array of pointers to the scalar multipliers.         *
 *                                                                          *
 *     rhscvectorptrptrs   Array of doubly-indirected pointers to       *
 *                     cvectorstructs.  In this case, since these values*
 *                     are not changed, there is no need for the double-    *
 *                     indirection; we simply want to maintain the g$x      *
 *                     references consistant.                               *
 *                                                                          *
 *     cspgx performs the saxpy operation for all aritys.               *
 *                                                                          *
 ****************************************************************************/

void cspgx(destptrptr,arity,alphaptrs,rhscvectorptrptrs)
cvectorstruct **destptrptr;
int arity;
complex *alphaptrs[];
cvectorstruct **rhscvectorptrptrs[];
{

  cvectorstruct *rhscvectorptr[MAXARITY];
                                 /* Dereferenced rhscvectorptrptrs[].   */

  cbucketstruct *rhscbucketptr[MAXARITY];
                                      /* Array of pointers to "current"     *
				       * buckets where "current" nonzeroes  *
				       * of the rhs_vectors reside.         */

  cbucketstruct *lhscbucketptr;
                                      /* Pointer to current bucket of       *
				       * lhscvector.                    */

  cibucketstruct *rhscibucketptr[MAXARITY];
                                      /* Array of pointers to "current"     *
				       * buckets where "current" nonzeroes  *
				       * of the rhs_vectors reside.         */

  cibucketstruct *lhscibucketptr;
                                      /* Pointer to current bucket of       *
				       * lhscvector.                    */

  complex alpha[MAXARITY];   /* Dereferenced alphaptrs[].          */

  int rhscbucketindex[MAXARITY];  /* Array of indices of current        *
				       * nonzeroes in the current buckets   *
				       * of rhs_vectors.                    */

  int lhscbucketindex;            /* bucketindex of next nonzero to be  *
				       * added to lhscvector.           */

  int lastlhscbucketindex;        /* bucketindex of the latest added    *
				       * nonzero in the lhscvector.  We *
				       * need this additional index to      *
				       * leap-frog buckets, if need be.     */

  int rhscibucketindex[MAXARITY]; /* Array of indices of current        *
				       * nonzeroes in the current buckets   *
				       * of rhs_vectors.                    */

  int lhscibucketindex;           /* bucketindex of next nonzero to be  *
				       * added to lhscvector.           */

  int lastlhscibucketindex;       /* bucketindex of the latest added    *
				       * nonzero in the lhscvector.  We *
				       * need this additional index to      *
				       * leap-frog buckets, if need be.     */

  int firstsaxpyoperand;              /* Order in the ennumeration of the   *
				       * rhs_vectors of the first non-null  *
				       * saxpy vectorop.                    */
  int k;                              /* Counter for arity.                 */

  int j;                              /* Index for vectorsubscripts to be   *
				       * entered in (and later popped from) *
				       * the heap.                          */

  int rhscvectornonzeroes[MAXARITY];
                                      /* Array of total # of nonzeroes in   *
				       * each rhs_vector.                   */

  int rhscvectorconsumednonzeroes[MAXARITY];
                                      /* Array of # of nonzeroes consumed   *
				       * so far in each rhs_vector.         */

  int lhscvectornonzeroes;        /* Count of # of nonzeroes entered    *
				       * into the lhscvector so far. By *
				       * the end of the saxpy operation     *
				       * this index equals the total # of   *
				       * nonzeroes in the resultant vector. */

/*
 * These variables were originally for COMPRESSED_SUBSCRIPTS
 */

  cvbucketstruct *lastlhscvbucketptr;
                                      /* Pointer to current vbucket of      *
				       * lhscvector.                    */

  int rhscvectornonzeroranges[MAXARITY];
                                      /* Array of total # of nonzeroes      *
				       * ranges in each rhs_vector.         */

  int rhscvectorconsumednonzeroranges[MAXARITY];
                                      /* Array of # of nonzeroes consumed   *
				       * so far in each rhs_vector.         */

  int rhscvbucketindex[MAXARITY]; /* Array of indices of current        *
				       * nonzeroes in the current buckets   *
				       * of rhs_vectors.                    */

  int lhscvectornonzeroranges;    /* Count of # of nonzeroes entered    *
				       * into the lhscvector so far. By *
				       * the end of the saxpy operation     *
				       * this index equals the total # of   *
				       * nonzeroes in the resultant vector. */

  int lhscvbucketindex;           /* bucketindex of next nonzero to be  *
				       * added to lhscvector.           */


  int lastlhscvbucketindex;       /* bucketindex of the latest added    *
				       * nonzero in the lhscvector.  We *
				       * need this additional index to      *
				       * leap-frog buckets, if need be.     */

  cvbucketstruct *templhscvbucketptr;
  int templhscvbucketindex;

  int j1;                             /* Index for vectorsubscripts to be   *
				       * entered in (and later popped from) *
				       * the heap.                          */

  int j2;                             /* Index for vectorsubscripts to be   *
				       * entered in (and later popped from) *
				       * the heap.                          */

  int h1;
  int h2;

  int totalcurrentnonzerorange;

#ifdef _DEBUG
  int unconsumedcurrentnonzerorange;
#endif

  int distancetooverlap;
  int distancetoendofvbucket;


  /*
   * These variables were added for union ds
   */
  
  cvectorstruct *restorerhscvectorptr[MAXARITY+1];

  int governingdstype;
  int rhs_largest_dstype = S1;        /* Largest dstype of any              *
				       * rhscvector ( CS > SS > S1 ).   */
  int s1_index;
  
  complex temp_gaxpy;

  /*
   * These variables were added for plusequal code
   */

#ifdef _PLUSEQUAL
  int plusequalsaxpy_flag = 0;
  int plusequalkorder;
  cvectorstruct **tempdestptrptr;
  complex *tempalphaptrs[MAXARITY-1];
  cvectorstruct **temprhscvectorptrptrs[MAXARITY-1];
  complex one; 
  int smallest_rhs_index;
#endif /* _PLUSEQUAL */

  /****************************************
   * Initialization of *destptrptr.       *
   ****************************************/
  
  if (*destptrptr == NULL)
    cgetvector(destptrptr);
  
  
  /****************************************
   * Dereferencing.                       *
   ****************************************/
  
  for (k=0;k<arity;k++) {
    alpha[k] = *alphaptrs[k];

    rhscvectorptr[k] = *rhscvectorptrptrs[k];
    restorerhscvectorptr[k] = rhscvectorptr[k];
    rhscvectornonzeroes[k] = rhscvectorptr[k]->nonzeroes;
    if (rhscvectorptr[k]->dstype == CS)
      rhscvectornonzeroranges[k] =
	rhscvectorptr[k]->cds.cs.nonzeroranges;
    if (rhscvectorptr[k]->dstype > rhs_largest_dstype)
      rhs_largest_dstype = rhscvectorptr[k]->dstype;

#ifdef _PLUSEQUAL    
    if (rhscvectorptr[k] == *destptrptr &&
	rhscvectorptr[k]->dstype == CS &&
	X_IS_ONE(alpha[k])) {
      plusequalsaxpy_flag++;
      plusequalkorder = k;
    }
#endif /* _PLUSEQUAL */

  }	
  
  /****************************************
   * Determination of 1st nonzero         *
   * rhscvector operand.              *
   ****************************************/
  
  firstsaxpyoperand = 0;

/*
 *   This bit of the code is obsolete, since the routines calling
 *   this routine already have stripped the gaxpy of zero vectors
 *   and zero multipliers.
 *

  while (nullsaxpyflag && firstsaxpyoperand<arity)
    
    if (rhscvectornonzeroes[firstsaxpyoperand] == 0 ||
	X_IS_ZERO(alpha[firstsaxpyoperand]))
      firstsaxpyoperand++;
    else
      nullsaxpyflag = 0;
  
 *
 * If no nonzero vectorops, return a null vector.
 *

  if (nullsaxpyflag) {
    
    (*destptrptr)->nonzeroes = 0;
    (*destptrptr)->dstype = S1;

  }
  else {

*/
    
    /****************************************
     * Predetermine governingdstype and     *
     * appropriate saxpy routine.           *
     ****************************************/

  if (rhs_largest_dstype == S1) {
    s1_index = rhscvectorptr[firstsaxpyoperand]->cds.s1.index;
    for (k=firstsaxpyoperand+1;k<arity;k++)
      if (rhscvectorptr[k]->cds.s1.index !=
	  rhscvectorptr[firstsaxpyoperand]->cds.s1.index &&
	  ! X_IS_ZERO(alpha[k]))
	s1_index = -1;
    if (s1_index > 0)
      governingdstype = S1;
    else
      governingdstype = SS;
  }
  else if (rhs_largest_dstype <= SS)
    governingdstype = SS;
  else /* if (rhs_largest_dstype == CS) */
    governingdstype = CS;

#ifdef _PLUSEQUAL
  smallest_rhs_index = DUMMYSUBSCRIPT;
  for (k=0;k<arity;k++) {
    if (rhscvectorptr[k]->dstype == S1) 
      smallest_rhs_index = min (smallest_rhs_index,
				rhscvectorptr[k]->cds.s1.index);
    else if (rhscvectorptr[k]->dstype == SS)
      smallest_rhs_index = min (smallest_rhs_index,
        rhscvectorptr[k]->cds.ss.bucketptr->index[0]);
    else /* if (rhscvectorptr[k]->dstype == CS) */
      smallest_rhs_index = min (smallest_rhs_index,
	rhscvectorptr[k]->cds.cs.ibucketptr->index_RANGE_BEGIN[0]);
  }

  if (plusequalsaxpy_flag &&
      (*destptrptr)->dstype == CS &&
      (*destptrptr)->cds.cs.ibucketptr->index_RANGE_BEGIN[0] <=
      smallest_rhs_index) {    
#ifdef _DEBUG
    if (plusequalsaxpy_flag > 1) {
      printf("cspgx warning: lhscvector appears more than ");
      printf("once on rhs.\n");
    }
#endif
    if (arity == 1) {
      plusequalcspg1(destptrptr,alphaptrs[plusequalkorder]);
    }
    else {
      for (k=0;k<plusequalkorder;k++) {
	tempalphaptrs[k] = alphaptrs[k];
	temprhscvectorptrptrs[k] = rhscvectorptrptrs[k];
      }
      for (k=plusequalkorder+1;k<arity;k++) {
	tempalphaptrs[k-1] = alphaptrs[k];
	temprhscvectorptrptrs[k-1] = rhscvectorptrptrs[k];
      }
      tempdestptrptr = &tempcdestptr;
      cspgx(tempdestptrptr,arity-1,
		tempalphaptrs,temprhscvectorptrptrs);
      restorerhscvectorptr[MAXARITY] = tempcdestptr;
      if ((*tempdestptrptr)->dstype != governingdstype &&
	  (*tempdestptrptr)->nonzeroes !=0) {
	cconvertdstype(tempdestptrptr,MAXARITY,governingdstype);
      }
      X_PUT_REAL(one,1.0);
      plusequalcspg2(destptrptr,
		      &one,tempdestptrptr,TEMP_OTHER_VECTOR);
      tempcdestptr = restorerhscvectorptr[MAXARITY];
    }    
  }
  else /* if (not plusequalsaxpy_flag) */ {
#endif /* _PLUSEQUAL */

    for (k=0;k<arity;k++) {
      if (rhscvectorptr[k]->dstype != governingdstype &&
	  rhscvectorptr[k]->nonzeroes !=0 && ! X_IS_ZERO(alpha[k])) {
	cconvertdstype((&rhscvectorptr[k]),k,governingdstype);
	if (governingdstype == CS)
	  rhscvectornonzeroranges[k] = 
	    rhscvectorptr[k]->cds.cs.nonzeroranges;
      }
    }
    
    if (governingdstype == S1) {
      
      /********************************************************************
       *                       ONE_SUBSCRIPT AXPY                         *
       ********************************************************************/

      lhscvectorptr->nonzeroes = 1;
      lhscvectorptr->dstype = S1;
      lhscvectorptr->cds.s1.index =
	rhscvectorptr[firstsaxpyoperand]->cds.s1.index;
      X_ZERO(temp_gaxpy);
      for (k=firstsaxpyoperand;k<arity;k++) {
	GAXPY_PLUS_EQUALS(temp_gaxpy,
			  alpha[k], rhscvectorptr[k]->cds.s1.value);
      }
      X_ASSIGN(lhscvectorptr->cds.s1.value,temp_gaxpy);

      caxpy_0_1_post_processing(destptrptr);

    }
    else /* if we are going to use lhscvectorptr */ {
      if (governingdstype == SS) {
	
	/******************************************************************
	 *                  _SINGLE_SUBSCRIPTS AXPY                       *
	 ******************************************************************/
	
	/****************************************
	 * Else perform saxpy:                  *
	 * 1. Put in heap the 1st nonzero of    *
	 *      each vector.                    *
	 ****************************************/
	rhscvectorconsumednonzeroes[firstsaxpyoperand] = 0;
	rhscbucketindex[firstsaxpyoperand] = 0;
	rhscbucketptr[firstsaxpyoperand] =
	  rhscvectorptr[firstsaxpyoperand]->cds.ss.bucketptr;
	
#ifdef _DEBUG
	if (rhscbucketptr[firstsaxpyoperand] == NULL) {
	  printf("cspgx error 1: Unexpected NULL rhscbucketptr: V# %d\n",
		 firstsaxpyoperand);
	  abort();
	}
#endif
	
	j = rhscbucketptr[firstsaxpyoperand]->
	  index[rhscbucketindex[firstsaxpyoperand]];
	
	if (j <= 0) {
#ifdef _DEBUG
	  printf("cspgx error 2: Illegal index: V # %d, subscript=%d\n",
		 firstsaxpyoperand,j);
	  abort();
#endif
	}
	
	else
	  heapinitialize(firstsaxpyoperand,j,j);
	
	for (k=firstsaxpyoperand+1;k<arity;k++) {
	  if (rhscvectornonzeroes[k] !=0 && ! X_IS_ZERO(alpha[k])) {
	    rhscvectorconsumednonzeroes[k] = 0;
	    rhscbucketindex[k] = 0;
	    rhscbucketptr[k] = rhscvectorptr[k]->cds.ss.bucketptr;
	    
#ifdef _DEBUG
	    if (rhscbucketptr[k] == NULL) {
	      printf("cspgx error 3: Unexpected NULL rhscbucketptr: ");
	      printf("vector # %d\n",k);
	      abort();
	    }
#endif
	    
	    j = rhscbucketptr[k]->index[rhscbucketindex[k]];
	    
#ifdef _DEBUG
	    if (j <= 0) {
	      printf("cspgx error 4: Illegal index: vec.=%d sub.=%d\n",
		     k,j);
	      abort();
	    }
#endif
	    
	    heapinsert(k,j,j);
	  }
	}
	
	/****************************************
	 * Perform saxpy:                       *
	 * 2. Initialize lhscvector.             *
	 ****************************************/
	cgetbucket(&lhscbucketptr);
	lhscvectorptr->dstype = governingdstype;
	lhscvectorptr->cds.ss.bucketptr = lhscbucketptr;
	lhscvectornonzeroes = 0;
	
	/****************************************
	 * Perform saxpy:                       *
	 * 3. Get nonzero by popping the        *
	 *    heaproot; consume it in saxpy;    *
	 *    replace / delete heaproot.        *
	 ****************************************/
	while (heaproot !=NULL) {
	  k = heaproot->vectorindex;
	  j = heaproot->rangebeginindex;
	  rhscbucketindex[k] = MY_MOD(rhscvectorconsumednonzeroes[k],
					  cBUCKET_DIM_minus1);
	  lhscbucketindex = MY_MOD(lhscvectornonzeroes,
				       cBUCKET_DIM_minus1);
	  
#ifdef _DEBUG
	  if (rhscbucketptr[k]->errorflag != 9999) {
	    printf("cspgx err 5: mem. corruption of cbucketstore\n");
	    abort();
	  }
#endif      
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.1. Found 1st or new subscript for  *
	   *      lhscvector.  If needed, get new  *
	   *      bucket for lhscvector.           *
	   *      Consume nonzero in saxpy.       *
	   ****************************************/
	  if (lhscvectornonzeroes == 0 ||
	      lhscbucketptr->index[lastlhscbucketindex] < j) {
	    if (lhscbucketindex == 0 && lhscvectornonzeroes != 0) {
	      cgetbucket(&(lhscbucketptr->nextbucket));
	      lhscbucketptr = lhscbucketptr->nextbucket;
	    }
	    lhscbucketptr->index[lhscbucketindex] = j;
	    GAXPY_EQUALS(lhscbucketptr->value[lhscbucketindex],
			 alpha[k],
			 rhscbucketptr[k]->value[rhscbucketindex[k]]);
	    lastlhscbucketindex = lhscbucketindex;
	    lhscvectornonzeroes++;
	  }
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.2. Found old subscript for         *
	   *      lhscvector.                      *
	   *      Consume nonzero in saxpy.       *
	   ****************************************/
	  else if (lhscbucketptr->index[lastlhscbucketindex] == j) {
	    GAXPY_PLUS_EQUALS(
	      lhscbucketptr->value[lastlhscbucketindex],
	      alpha[k],
	      rhscbucketptr[k]->value[rhscbucketindex[k]]);
	  }
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.3. Found wrong subscript for       *
	   *      lhscvector.  Abort.               *
	   ****************************************/
	  
#ifdef _DEBUG
	  else {
	    printf("cspgx error 6: lhscnz=%d, lastlhscindex[]=%d, j=%d\n",
		   lhscvectornonzeroes,
		   lhscbucketptr->index[lastlhscbucketindex],j);
	    abort();
	  }
#endif
	  
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.4. If there is one, get new        *
	   *      nonzero from the rhs_vector     *
	   *      whose nonzero was just          *
	   *      consumed and replace the        *
	   *      heaproot with new entry;        *
	   *      Else, delete heaproot.          *
	   ****************************************/
	  if (++rhscvectorconsumednonzeroes[k] < rhscvectornonzeroes[k]) {
	    if ((rhscbucketindex[k] =
		 MY_MOD(rhscvectorconsumednonzeroes[k],
			cBUCKET_DIM_minus1)) == 0) {
	      if ((rhscbucketptr[k] = rhscbucketptr[k]->nextbucket) == NULL) {
		
#ifdef _DEBUG
		printf("cspgx err 7: unexpected NULL nextbucket: k=%d",
		       k);
		printf("rhscvectorconsumednonzeros=%d, rhscvecnonzeroes=%d,",
		       rhscvectorconsumednonzeroes[k],rhscvectornonzeroes[k]);
		abort();
#endif
		
	      }
	    }     
	    j = rhscbucketptr[k]->index[rhscbucketindex[k]];
	    
	    if (j <= 0) {
#ifdef _DEBUG
	      printf("cspgx err 8: Illegal index: V=%d, subscript=%d\n"
		     ,k,j);
	      abort();
#endif
	      
	    }
	    
	    else
	      heapop(REPLACE,k,j,j);
	  }
	  else
	    heapop(DELETE,DUMMYVECTOR,DUMMYSUBSCRIPT,DUMMYSUBSCRIPT);
	  
	} /*matches     while (heaproot !=NULL) {     */
      } /*matches     else if (governingdstype == SS) {     */
      
      else /* if (governingdstype == CS) */ {
	
	/******************************************************************
	 *                 _COMPRESSED_SUBSCRIPTS AXPY                    *
	 ******************************************************************/
	
	
	/****************************************
	 * Else perform saxpy:                  *
	 * 1. Put in heap the 1st nonzero of    *
	 *    each vector.                      *
	 ****************************************/
	rhscvectorconsumednonzeroranges[firstsaxpyoperand] = 0;
	rhscibucketindex[firstsaxpyoperand] = 0;
	rhscvbucketindex[firstsaxpyoperand] = 0;
	rhscibucketptr[firstsaxpyoperand] =
	  rhscvectorptr[firstsaxpyoperand]->cds.cs.ibucketptr;
	rhscvbucketptr[firstsaxpyoperand] =
	  rhscvectorptr[firstsaxpyoperand]->cds.cs.vbucketptr;
	
#ifdef _DEBUG
	if (rhscibucketptr[firstsaxpyoperand] == NULL ||
	    rhscvbucketptr[firstsaxpyoperand] == NULL) {
	  printf("cspgx error 10: Unexpected NULL rhscbucketptrs: ");
	  printf("vector # %d\n",firstsaxpyoperand);
	  abort();
	}
#endif
	
	j1 = rhscibucketptr[firstsaxpyoperand]->
	  index_RANGE_BEGIN[rhscibucketindex[firstsaxpyoperand]];
	j2 = rhscibucketptr[firstsaxpyoperand]->
	  index_RANGE_END[rhscibucketindex[firstsaxpyoperand]];
	
	if (j1 <= 0 || j2 <= 0 || j2 < j1) {
#ifdef _DEBUG
	  printf("cspgx error 11: Nonsensical Nonzero Range; \n\t\t");
	  printf("vector # %d, subscript 1=%d subscript 2=%d\n",
		 firstsaxpyoperand,j1,j2);
	  abort();
#endif
	}
	
	else
	  heapinitialize(firstsaxpyoperand,j1,j2);
	
	for (k=firstsaxpyoperand+1;k<arity;k++) {

	  /*
	   * rhscvectornonzeroes[k] !=0 will be redundant and
	   *                           unnecessary most of the time
	   *                           but in the code preceding the
	   *                           addition of += module it was
	   *                           needed, so we leave it.
	   */
	
	  if (rhscvectornonzeroes[k] !=0 &&
	      rhscvectornonzeroranges[k] !=0 &&
	      ! X_IS_ZERO(alpha[k])) {
	    rhscvectorconsumednonzeroranges[k] = 0;
	    rhscibucketindex[k] = 0;
	    rhscibucketptr[k] =
	      rhscvectorptr[k]->cds.cs.ibucketptr;
	    rhscvbucketindex[k] = 0;
	    rhscvbucketptr[k] =
	      rhscvectorptr[k]->cds.cs.vbucketptr;
	    
#ifdef _DEBUG
	    if (rhscibucketptr[k] == NULL ||
		rhscvbucketptr[k] == NULL) {
	      printf("cspgx error 12: Unexpected ");
	      printf("NULL rhscibucketptr or rhscvbucketptr: vector # %d\n",k);
	      abort();
	    }
#endif
	    
	    j1 = rhscibucketptr[k]->index_RANGE_BEGIN[rhscibucketindex[k]];
	    j2 = rhscibucketptr[k]->index_RANGE_END[rhscibucketindex[k]];
	    
#ifdef _DEBUG
	    if (j1 <= 0 || j2 <= 0 || j2 < j1) {
	      printf("cspgx err 13: Nonsensical Nonzero Range; \n\t\t");
	      printf("vector # %d, subscript 1=%d subscript 2=%d\n",
		     firstsaxpyoperand,j1,j2);
	      abort();
	    }
#endif
	    
	    heapinsert(k,j1,j2);
	  }
	}
	
	/****************************************
	 * Perform saxpy:                       *
	 * 2. Initialize lhscvector.             *
	 ****************************************/
	cgetibucket(&lhscibucketptr);
	lhscvectorptr->dstype = governingdstype;
	lhscvectorptr->cds.cs.ibucketptr = lhscibucketptr;
	cgetvbucket(&lhscvbucketptr);
	lhscvectorptr->cds.cs.vbucketptr = lhscvbucketptr;
	lhscvectornonzeroranges = 0;
	lhscvectornonzeroes = 0;
	lhscvbucketindex = 0;
	
	/****************************************
	 * Perform saxpy:                       *
	 * 3. Get nonzero by popping the        *
	 *    heaproot; consume it in saxpy;    *
	 *    either replace or delete heaproot.*
	 ****************************************/

	while (heaproot !=NULL) {
	  k = heaproot->vectorindex;
	  j1 = heaproot->rangebeginindex;
	  j2 = heaproot->rangeendindex;
	  rhscibucketindex[k] =
	    MY_MOD(rhscvectorconsumednonzeroranges[k],
		   cIBUCKET_DIM_minus1);
	  lhscibucketindex = MY_MOD(lhscvectornonzeroranges,
					cIBUCKET_DIM_minus1);
	  if ((rhscvbucketindex[k] =
	       MY_MOD(rhscvbucketindex[k],cVBUCKET_DIM_minus1))==0
	      && rhscvectorconsumednonzeroranges[k] !=0) {
	    cHOP_TO_NEXT_RHSVBUCKET(k,&rhscvbucketindex[k],
				   rhscvectorconsumednonzeroranges[k],
				   rhscvectornonzeroranges[k]);
	  }
	  lhscvbucketindex = MY_MOD(lhscvbucketindex,
					cVBUCKET_DIM_minus1);
	  
	  
	  
#ifdef _DEBUG
	  if (rhscibucketptr[k]->errorflag != 9099 ||
	      rhscvbucketptr[k]->errorflag != 9009) {
	    printf("cspgx err 14: mem corruption of cibucketstore ");
	    printf("or cvbucketstore.\n");
	    abort();
	  }
#endif      
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.0. Check for startup errors.       *
	   ****************************************/
	  
#ifdef _DEBUG
	  if (lhscvectornonzeroranges == 0) {
	    if (rhscibucketindex[k] != 0 || rhscvbucketindex[k] != 0 ||
		lhscibucketindex != 0 || lhscvbucketindex != 0) {
	      printf("cspgx error 15: Zeroes expected: \n\t");
	      printf("rhscibucketindex[k]=%d, rhscvbucketindex[k]=%d\n\t",
		     rhscibucketindex[k], rhscvbucketindex[k]);
	      printf("lhscibucketindex=%d, lhscvbucketindex=%d\n\t",
		     lhscibucketindex,lhscvbucketindex);
	      abort();
	    }
	  }
#endif      
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.1. Found 1ST or NEW subscript      *
	   *      range (non-overlapping with     *
	   *      previous one, if any) for       *
	   *      lhscvector.  If needed, get new  *
	   *      ibuckets and vbuckets for        *
	   *      lhscvector.                      *
	   ****************************************/
	  if (lhscvectornonzeroranges == 0 ||
	      j1 > lhscibucketptr->index_RANGE_END[lastlhscibucketindex]+1) {/*A.1*/
	    
	    if (lhscibucketindex == 0 && 
		lhscvectornonzeroranges != 0) {
	      cgetibucket(&(lhscibucketptr->nextibucket));
	      lhscibucketptr = lhscibucketptr->nextibucket;
	    }
	    lhscibucketptr->index_RANGE_BEGIN[lhscibucketindex] = j1;
	    lhscibucketptr->index_RANGE_END[lhscibucketindex] = j2;
	    
	    if (lhscvbucketindex == 0 &&
		lhscvectornonzeroranges != 0)
	      cget_and_hop_to_next_vbucket(&lhscvbucketindex);
	    
	    lastlhscvbucketptr = lhscvbucketptr;
	    lastlhscvbucketindex = lhscvbucketindex;
	    
	    totalcurrentnonzerorange = j2-j1+1;
	    
	    lhscvectornonzeroes += totalcurrentnonzerorange;
	    lhscvectornonzeroranges++;
	    
	    
	    cNON_OVER_LAPPING_CS_RANGE_AXPY(COMPLETELY_NEW_RANGE,
					   lhscibucketindex,
					   &lastlhscibucketindex,
					   &lhscvbucketindex,
					   k,
					   &rhscvbucketindex[k],
					   alpha[k],
					   totalcurrentnonzerorange,
					   &unconsumedcurrentnonzerorange,
					   rhscvectorconsumednonzeroranges[k],
					   rhscvectornonzeroranges[k],
					   lhscvectornonzeroranges);
	    

	  } /*A.1*/
	  
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.2. Found old subscript for         *
	   *      lhscvector.  Consume  nonzero in *
	   *      saxpy.                          *
	   *                                      *
	   ****************************************/
	  else if (j1 <= 
		   lhscibucketptr->index_RANGE_END[lastlhscibucketindex]+1 &&
		   j1 >= 
		   lhscibucketptr->index_RANGE_BEGIN[lastlhscibucketindex]) {
	    /*B.1*/
	    
	    templhscvbucketptr = lhscvbucketptr;
	    templhscvbucketindex = lhscvbucketindex;
	    lhscvbucketptr = lastlhscvbucketptr;
	    
	    distancetooverlap =
	      j1 - lhscibucketptr->index_RANGE_BEGIN[lastlhscibucketindex];
	    distancetoendofvbucket = 
	      cVBUCKET_DIM - lastlhscvbucketindex;
	    
	    if (j1 == lhscibucketptr->index_RANGE_END[lastlhscibucketindex]+1) {
	      h1 = 0;
	      h2 = -1;
	    }
	    else if (j2 <= lhscibucketptr->index_RANGE_END[lastlhscibucketindex]) {
	      h1 = j1;
	      h2 = j2;
	      j1 = 0;
	      j2 = -1;
	    }
	    else {
	      h1 = j1;
	      h2 = lhscibucketptr->index_RANGE_END[lastlhscibucketindex];
	      j1 = h2+1;
	    }
	    
	    if (distancetoendofvbucket > distancetooverlap) {
	      lhscvbucketindex =
		MY_MOD(lastlhscvbucketindex + distancetooverlap,
		       cVBUCKET_DIM_minus1);
	    }
	    else if (distancetoendofvbucket < distancetooverlap) {
	      lhscvbucketindex =
		MY_MOD(distancetooverlap - distancetoendofvbucket,
		       cVBUCKET_DIM_minus1);
	    }
	    else {
	      lhscvbucketindex = 0;
	    }
	    
	    while (distancetooverlap >= distancetoendofvbucket &&
		   lhscvbucketptr->nextvbucket != NULL) {
	      lhscvbucketptr = lhscvbucketptr->nextvbucket;
	      distancetooverlap -= distancetoendofvbucket;
	      distancetoendofvbucket = cVBUCKET_DIM;
	    }
	    
	    lhscibucketptr->index_RANGE_END[lastlhscibucketindex] =
	      max(j2, lhscibucketptr->index_RANGE_END[lastlhscibucketindex]);
	    
	    if ((totalcurrentnonzerorange = h2-h1+1) > 0) {
	      
	      
	      cOVERLAPPING_CS_RANGE_AXPY(&lhscvbucketindex,
					k,
					&rhscvbucketindex[k],
					alpha[k],
					totalcurrentnonzerorange,
					&unconsumedcurrentnonzerorange,
					rhscvectorconsumednonzeroranges[k],
					rhscvectornonzeroranges[k],
					lhscvectornonzeroranges);
	
	    }
	      
	    /*
	     * This bit of logic must be performed, whether or not there
	     * is additional nonzeroes to process.
	     */
	    
	    if (j2 != -1 &&
		(lhscvbucketptr != templhscvbucketptr ||
		 (lhscvbucketindex = 
		  MY_MOD(lhscvbucketindex,cVBUCKET_DIM_minus1)) !=
		 templhscvbucketindex)) {
	      
#ifdef _DEBUG
	      printf("cspgx err 50.0: Unexpectd value discrepancies\n");
	      printf("lhscvbucketptr=%d, templhscvbucketptr=%d\n",
		     (int)lhscvbucketptr,(int)templhscvbucketptr);
	      printf("lhscvbucketindex=%d, templhscvbucketindex=%d\n",
		     lhscvbucketindex,templhscvbucketindex);
	      
	      abort();
#endif
	    }
	    else {
	      lhscvbucketptr = templhscvbucketptr;
	      lhscvbucketindex = templhscvbucketindex;
	    }
	    
	    
	    if ((totalcurrentnonzerorange = j2-j1+1) > 0) {
	      lhscvectornonzeroes += totalcurrentnonzerorange;
	      
	      /*
	       * adding new entries to vbucket
	       */
	      
	      if (lhscvbucketindex == 0 &&
		  lhscvectornonzeroranges != 0 &&
		  j2 != -1)
		cget_and_hop_to_next_vbucket(&lhscvbucketindex);
	      
	      if(rhscvbucketindex[k] == cVBUCKET_DIM && j2 != -1) {
		
		cHOP_TO_NEXT_RHSVBUCKET(k,&rhscvbucketindex[k],
				       rhscvectorconsumednonzeroranges[k],
				       rhscvectornonzeroranges[k]);
	      }

	      cNON_OVER_LAPPING_CS_RANGE_AXPY(NEW_ENTRIES_PADDED_TO_OLD_RANGE,
					     lhscibucketindex,
					     &lastlhscibucketindex,
					     &lhscvbucketindex,
					     k,
					     &rhscvbucketindex[k],
					     alpha[k],
					     totalcurrentnonzerorange,
					     &unconsumedcurrentnonzerorange,
					     rhscvectorconsumednonzeroranges[k],
					     rhscvectornonzeroranges[k],
					     lhscvectornonzeroranges);
	      
	    } /*  corresponds to the if ((totalcurrent...) { */
	    
	  } /*B.1*/
	  
	  
	  /****************************************	
	   * Perform saxpy:                       *
	   * 3.3. Found wrong subscript for       *
	   *      lhscvector.  Abort.               *
	   ****************************************/
	  
#ifdef _DEBUG
	  else {
	    printf("cspgx error 60: lhscvectornonzeroranges=%d,",
		   lhscvectornonzeroranges);
	    printf("lastlhscibucketindex=%d, \n\t",
		   lastlhscibucketindex);
	    printf("lhscibucketptr->index[%d] = (%d,%d), j1=%d\n",
		   lastlhscibucketindex,
		   lhscibucketptr->index_RANGE_BEGIN[lastlhscibucketindex],
		   lhscibucketptr->index_RANGE_END[lastlhscibucketindex],j1);
	    abort();
	  }
#endif
	  
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.4. If there is one, get new        *
	   *      nonzero from the rhs_vector     *
	   *      whose nonzero was just          *
	   *      consumed and replace the        *
	   *      heaproot with new entry;        *
	   *      Else, delete heaproot.          *
	   ****************************************/
	  if (++rhscvectorconsumednonzeroranges[k] <
	      rhscvectornonzeroranges[k]) {
	    if ((rhscibucketindex[k] =
		 MY_MOD(rhscvectorconsumednonzeroranges[k],
			cIBUCKET_DIM_minus1)) == 0) {
	      if ((rhscibucketptr[k] = rhscibucketptr[k]->nextibucket) ==
		  NULL) {
		
#ifdef _DEBUG
		printf("cspgx err 61: unexpected NULL nextvbucket: \n");
		printf("\t k=%d, rhscvectorconsumednonzeroranges=%d,\n",
		       k,rhscvectorconsumednonzeroranges[k]);
		printf("\t rhscvectornonzeroranges[k]=%d\n",
		       rhscvectornonzeroranges[k]);
		abort();
#endif
		
	      }
	    }     
	    j1 = rhscibucketptr[k]->index_RANGE_BEGIN[rhscibucketindex[k]];
	    j2 = rhscibucketptr[k]->index_RANGE_END[rhscibucketindex[k]];
	    
	    if (j1 <= 0 || j2 <= 0) {
#ifdef _DEBUG
	      printf("cspgx error 62: Illegal index: vector=%d, ",k);
	      printf("subscript1=%d, subscript2=%d \n",j1,j2);
	      abort();
#endif
	    }
	    
	    else
	      heapop(REPLACE,k,j1,j2);
	  }
	  else
	    heapop(DELETE,DUMMYVECTOR,DUMMYSUBSCRIPT,DUMMYSUBSCRIPT);
	} /*matches     while (heaproot !=NULL) {     */
      } /* matches  else ** if (governingdstype == CS) ** {     */	
      
      caxpy_2_heap_post_processing(destptrptr,
				  rhscvectorptrptrs,
				  alphaptrs,
				  restorerhscvectorptr,
				  arity,
				  lhscvectornonzeroes,
				  lhscvectornonzeroranges,
				  governingdstype);
      
      
    } /* matches    else ** if we are going to use lhscvectorptr ** {  */

#ifdef _PLUSEQUAL
  } /* matches    else ** if (not plusequalsaxpy_flag) ** {    */
#endif /* _PLUSEQUAL */

  /* }  matches    else ** if (not nullsaxpyflag) ** {    */

}

