#define COMPLEX
/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#include  <stdlib.h>
#include "cgaxpy_pd_macros_and_data_structures.h"
#include "cgaxpy_pd_global_variables.h"

/****************************************************************************
 *                                                                          *
 *    plusequalcspg2                                                       *
 *                                                                          *
 ****************************************************************************/

void plusequalcspg2(rhsanddestptrptr,
		     otheralphaptr,
                     otherrhscvectorptrptr,
                     status_otherrhscvector)
cvectorstruct **rhsanddestptrptr;
complex *otheralphaptr;
cvectorstruct **otherrhscvectorptrptr;
int status_otherrhscvector;
{
  complex *alphaptrs[2];
  
  int arity = 2;
  
  cvectorstruct *rhscvectorptr[2];
  cvectorstruct *tempvectorptr;
  
  int i;
  

  cibucketstruct *rhscibucketptr[2];
                                      /* Array of pointers to "current"     *
				       * buckets where "current" nonzeroes  *
				       * of the rhs_vectors reside.         */

  cibucketstruct *lhscibucketptr;
                                      /* Pointer to current bucket of       *
				       * lhscvector.                    */

  complex alpha[2];          /* Dereferenced alphaptrs[].          */

  int rhscibucketindex[2];        /* Array of indices of current        *
				       * nonzeroes in the current buckets   *
				       * of rhs_vectors.                    */

  int lhscibucketindex;           /* bucketindex of next nonzero to be  *
				       * added to lhscvector.           */

  int lastlhscibucketindex;       /* bucketindex of the latest added    *
				       * nonzero in the lhscvector.  We *
                                       * need this additional index to      *
				       * leap-frog buckets, if need be.     */

  int k;                              /* Counter for arity.                 */

  int lhscvectornonzeroes;        /* Count of # of nonzeroes entered    *
				       * into the lhscvector so far. By *
				       * the end of the saxpy operation,    *
				       * this index equals the total # of   *
				       * nonzeroes in the resultant vector. */

/*
 * These variables were originally for COMPRESSED_SUBSCRIPTS
 */

  cvbucketstruct *lastlhscvbucketptr;  
                                      /* Pointer to current vbucket of      *
				       * lhscvector.                    */

  int rhscvectornonzeroranges[2];
                                      /* Array of total # of nonzeroes      *
				       * ranges in each rhs_vector.         */

  int rhscvectorconsumednonzeroranges[2];
                                      /* Array of # of nonzeroes consumed   *
				       * so far in each rhs_vector.         */

  int rhscvbucketindex[2];        /* Array of indices of current        *
				       * nonzeroes in the current buckets   *
				       * of rhs_vectors.                    */

  int lhscvectornonzeroranges;    /* Count of # of nonzeroes entered    *
				       * into the lhscvector so far. By *
				       * the end of the saxpy operation,    *
				       * this index equals the total # of   *
				       * nonzeroes in the resultant vector. */

  int lhscvbucketindex;           /* bucketindex of next nonzero to be  *
				       * added to lhscvector.           */


  int lastlhscvbucketindex;       /* bucketindex of the latest added    *
				       * nonzero in the lhscvector.  We *
				       * need this additional index to frog *
				       * leap-frog buckets, if need be.     */

  cvbucketstruct *templhscvbucketptr;

  int templhscvbucketindex;

  int j1;                             /* Index for vectorsubscripts to be   *
				       * entered in (and later popped from) *
				       * the heap.                          */

  int j2;                             /* Index for vectorsubscripts to be   *
				       * entered in (and later popped from) *
				       * the heap.                          */

  int h1;
  int h2;

  int totalcurrentnonzerorange;

#ifdef _DEBUG
  int unconsumedcurrentnonzerorange;
#endif

  int distancetooverlap;
  int distancetoendofvbucket;
  int hopdistance;

  /*
   * These variables were added for union ds
   */
  
  cvectorstruct *restorerhscvectorptr[2+1];

  int governingdstype;
  
  /*
   * These variables were added for plusequal code
   */
  
  cvectorstruct **rhscvectorptrptrs[MAXARITY];

  int condition_1, condition_2;
  int lastrhscibucketindex[2];
  int lastrhscvbucketindex[2];

  cibucketstruct *lastrhscibucketptr[2];
  cvbucketstruct *lastrhscvbucketptr[2];
  cvbucketstruct *catchupto_rhsvbucketptr[2];

  cibucketstruct *temprhscibucketptr[2];
  cvbucketstruct *temprhscvbucketptr[2];
  int temprhscvbucketindex[2];

  cibucketstruct *restofplusequalibucketptr;
  cvbucketstruct *restofplusequalvbucketptr;
  int restofplusequalibucketindex;
  int restofplusequalvbucketindex;
  int restofplusequalvectorconsumednonzeroranges;

#ifdef _TRACE
  int nnzb,nnza;
  cvectorstruct *lhscvectoraddress;
  
  cvectorstruct **vectorptrptrs[2];
  
  alphaptrs[0] = alphaptr;
  vectorptrptrs[0] = vectoraptrptr;
  
  alphaptrs[1] = betaptr;
  vectorptrptrs[1] = vectorbptrptr;
  
  if(*destptrptr==NULL){ 
    nnzb = 0; 
    lhscvectoraddress= (cvectorstruct *) 99900; 
  }
  else {
    nnzb = (*destptrptr)->nonzeroes; 
    lhscvectoraddress = *destptrptr;
  }
#endif
  
  /****************************************
   * Dereferencing.                       *
   ****************************************/
  
  X_PUT_REAL(alpha[0],1.0);

#ifdef _DEBUG
  if (*rhsanddestptrptr == NULL) {
    printf("plusequalcspg2 error #1: Non_initialized lhscvector.\n");
    abort();
  }
#endif

  rhscvectorptr[0] = *rhsanddestptrptr;
  rhscvectornonzeroranges[0] =
    rhscvectorptr[0]->cds.cs.nonzeroranges;

  X_ASSIGN(alpha[1],(*otheralphaptr));

  alphaptrs[0] = &alpha[0];
  alphaptrs[1] = otheralphaptr;

  rhscvectorptr[1] = *otherrhscvectorptrptr;
  rhscvectornonzeroranges[1] =
    rhscvectorptr[1]->cds.cs.nonzeroranges;

  governingdstype = CS;
  k = 1;   
  restorerhscvectorptr[k] = rhscvectorptr[k];

  if (rhscvectorptr[k]->dstype != governingdstype) {
    cconvertdstype((&rhscvectorptr[k]),k,governingdstype);
    rhscvectornonzeroranges[k] =
      rhscvectorptr[k]->cds.cs.nonzeroranges;
  }  
 
  /******************************************************************
   *               _COMPRESSED_SUBSCRIPTS AXPY ONLY                 *
   ******************************************************************/
  
  /****************************************
   * perform saxpy:                       *
   * 1. Put in heap the 1st nonzero of    *
   *    each vector.                      *
   ****************************************/
  for (k=0;k<arity;k++) {
    rhscvectorconsumednonzeroranges[k] = 0;
    rhscibucketindex[k] = 0;
    rhscvbucketindex[k] = 0;
    rhscibucketptr[k] =
      rhscvectorptr[k]->cds.cs.ibucketptr;
    rhscvbucketptr[k] =
      rhscvectorptr[k]->cds.cs.vbucketptr;
    
#ifdef _DEBUG
    if (rhscibucketptr[k] == NULL ||
	rhscvbucketptr[k] == NULL) {
	      printf("plusequalcspg2 err 10: Unexpected NULL rhscbucketptr: ");
	      printf("vector # %d\n",k);
	      abort();
	    }
#endif
  }	  

  
  /****************************************
   * Perform saxpy:                       *
   * 2. Initialize lhscvector.             *
   ****************************************/

  lhscvectornonzeroranges = 0;
  lhscvectornonzeroes = 0;
  lhscvbucketindex = 0;


  /****************************************
   * Perform saxpy:                       *
   * 3. Get nonzero by popping the        *
   *    heaproot; consume it in saxpy;    *
   *    either replace or delete heaproot.*
   ****************************************/
  

  /* while (there are nonzeroranges to be processed AND                      *
   *         (condition_1 OR condition_2 ), where:                           *
   * condition_1 == TRUE, indicates that the range to be processed belongs   *
   *                to the plusequalvector,                                  *
   * condition_2 == TRUE, indicates that the range to be processed belongs   *
   *                to the otherrhscvector, and that this range is       *
   *                entirelyencapsulated in the last processed range of the  *
   *                plusequalvector.                                         */

  lastrhscibucketindex[0] = -1;
  lastrhscibucketptr[0] = NULL;
  
  while ((rhscvectorconsumednonzeroranges[0] <
	  rhscvectornonzeroranges[0] &&
	  rhscvectorconsumednonzeroranges[1] <
	  rhscvectornonzeroranges[1]) &&

	 ((condition_1 = (rhscibucketptr[0]->
			  index_RANGE_BEGIN[rhscibucketindex[0]] <=
			  rhscibucketptr[1]->
			  index_RANGE_BEGIN[rhscibucketindex[1]]))    ||
	  
	  (condition_2 = ((lastrhscibucketindex[0] >= 0) &&
			  (rhscibucketptr[1]->
			   index_RANGE_BEGIN[rhscibucketindex[1]] >=
			   lastrhscibucketptr[0]->
			   index_RANGE_BEGIN[lastrhscibucketindex[0]] &&
			   rhscibucketptr[1]->
			   index_RANGE_END[rhscibucketindex[1]] <=
			   lastrhscibucketptr[0]->
			   index_RANGE_END[lastrhscibucketindex[0]]))))) {

    if (condition_1) {
      k = 0;
      lastrhscibucketindex[0] = rhscibucketindex[0];
      lastrhscvbucketindex[0] = rhscvbucketindex[0];
      lastrhscibucketptr[0] = rhscibucketptr[0];
      lastrhscvbucketptr[0] = rhscvbucketptr[0];
      totalcurrentnonzerorange =
	rhscibucketptr[0]->index_RANGE_END[rhscibucketindex[0]] -
	  rhscibucketptr[0]->index_RANGE_BEGIN[rhscibucketindex[0]]+1;

#ifdef _DEBUG
      unconsumedcurrentnonzerorange = totalcurrentnonzerorange;
#endif

      hopdistance = totalcurrentnonzerorange;
      distancetoendofvbucket = cVBUCKET_DIM - rhscvbucketindex[0];
   
      if (distancetoendofvbucket > hopdistance) {
	rhscvbucketindex[0] = MY_MOD(rhscvbucketindex[0] + hopdistance,
					 cVBUCKET_DIM_minus1);
      }
      else if (distancetoendofvbucket < hopdistance) {
	rhscvbucketindex[0] = MY_MOD(hopdistance - distancetoendofvbucket,
					 cVBUCKET_DIM_minus1);
      }
      else {
	rhscvbucketindex[0] = 0;
      }
      while (hopdistance >= distancetoendofvbucket &&
	     rhscvbucketptr[0]->nextvbucket != NULL) {
	rhscvbucketptr[0] = rhscvbucketptr[0]->nextvbucket;
	hopdistance -= distancetoendofvbucket;
	distancetoendofvbucket = cVBUCKET_DIM;
      }

      lhscvectornonzeroes += totalcurrentnonzerorange;
    }
     
    else if (condition_2) {
      k = 1;
     temprhscibucketptr[0] = rhscibucketptr[0];
      temprhscvbucketptr[0] = rhscvbucketptr[0];
      rhscibucketptr[0] = lastrhscibucketptr[0];
      rhscvbucketptr[0] = lastrhscvbucketptr[0];

      /* The reason rhscvbucketindex[0] is not
       * set to lastrhscvbucketindex[0] is because
       * rhscvbucketindex[0] is computed below from
       * lastrhscvbucketindex[0].
       */

      temprhscvbucketindex[0] = rhscvbucketindex[0];
       
      distancetooverlap =
	rhscibucketptr[1]->index_RANGE_BEGIN[rhscibucketindex[1]] -
	rhscibucketptr[0]->index_RANGE_BEGIN[lastrhscibucketindex[0]];
      distancetoendofvbucket = cVBUCKET_DIM - lastrhscvbucketindex[0];
      
      if (distancetoendofvbucket > distancetooverlap) {
	rhscvbucketindex[0] =
	  MY_MOD(lastrhscvbucketindex[0] + distancetooverlap,
		 cVBUCKET_DIM_minus1);
      }
      else if (distancetoendofvbucket < distancetooverlap) {
	rhscvbucketindex[0] =
	  MY_MOD(distancetooverlap - distancetoendofvbucket,
		 cVBUCKET_DIM_minus1);
      }
      else {
	rhscvbucketindex[0] = 0;
      }
      
      while (distancetooverlap >= distancetoendofvbucket &&
	     rhscvbucketptr[0]->nextvbucket != NULL) {
	rhscvbucketptr[0] = rhscvbucketptr[0]->nextvbucket;
	distancetooverlap -= distancetoendofvbucket;
	distancetoendofvbucket = cVBUCKET_DIM;
      }

      totalcurrentnonzerorange = 
      	rhscibucketptr[1]->index_RANGE_END[rhscibucketindex[1]] -
	  rhscibucketptr[1]->index_RANGE_BEGIN[rhscibucketindex[1]]+1;

      /* Wrapping the call with the two assignments has the effect of
       * EQUIVALENCEing the data structures being pointed two by the
       * two pointers.
       */
    
      lhscvbucketptr = rhscvbucketptr[0];      
      cOVERLAPPING_CS_RANGE_AXPY(&rhscvbucketindex[0],
				1,
				&rhscvbucketindex[1],
				alpha[1],
				totalcurrentnonzerorange,
				&unconsumedcurrentnonzerorange,
				rhscvectorconsumednonzeroranges[1],
				rhscvectornonzeroranges[1],
				rhscvectornonzeroranges[0]);
      rhscvbucketptr[0] = lhscvbucketptr;
      
      rhscibucketptr[0] = temprhscibucketptr[0];
      rhscvbucketptr[0] = temprhscvbucketptr[0];
      rhscvbucketindex[0] = temprhscvbucketindex[0];
      
    } /* matches     else if (condition_2) {  */
  
    /*****************************************
     * Perform saxpy:                        *
     * 3.4. If there is one, get new         *
     *      nonzero from the rhs_vector      *
     *      whose nonzero was just consumed. *
     *****************************************/
    if (++rhscvectorconsumednonzeroranges[k] <
	rhscvectornonzeroranges[k]) {
      if ((rhscibucketindex[k] =
	   MY_MOD(rhscvectorconsumednonzeroranges[k],
		  cIBUCKET_DIM_minus1)) == 0) {
	if ((rhscibucketptr[k] = rhscibucketptr[k]->nextibucket) ==
	    NULL) {
	  
#ifdef _DEBUG
	  printf("plusqualcspg2 error #36: unexpected NULL nextvbucket: \n");
	  printf("\t k=%d, rhscvectorconsumednonzeroranges=%d,\n",
		 k,rhscvectorconsumednonzeroranges[k]);
	  printf("\t rhscvectornonzeroranges[k]=%d\n",
		 rhscvectornonzeroranges[k]);
	  abort();
#endif
	  
	}
      }
    }   

  } /* matches     1st while loop           */

  if (rhscvectorconsumednonzeroranges[1] ==
      rhscvectornonzeroranges[1]) {
    while (rhscvectorconsumednonzeroranges[0] <
	   rhscvectornonzeroranges[0]) {
      totalcurrentnonzerorange =
	rhscibucketptr[0]->index_RANGE_END[rhscibucketindex[0]] -
	  rhscibucketptr[0]->index_RANGE_BEGIN[rhscibucketindex[0]]+1;
      lhscvectornonzeroes += totalcurrentnonzerorange;

      if (++rhscvectorconsumednonzeroranges[0] <
	  rhscvectornonzeroranges[0]) {
	if ((rhscibucketindex[0] =
	     MY_MOD(rhscvectorconsumednonzeroranges[0],
		    cIBUCKET_DIM_minus1)) == 0) {
	  if ((rhscibucketptr[0] = rhscibucketptr[0]->nextibucket) ==
	      NULL) {
	    
#ifdef _DEBUG
	    printf("plusqualcspg2 err #58: unexpected NULL nextvbucket: \n");
	    printf("\t k=%d, rhscvectorconsumednonzeroranges=%d,\n",
		   0,rhscvectorconsumednonzeroranges[0]);
	    printf("\t rhscvectornonzeroranges[0]=%d\n",
		   rhscvectornonzeroranges[0]);
	    abort();
#endif
	    
	  }
	}
      }   
    }
   }

  else if (rhscvectorconsumednonzeroranges[0] ==
	   rhscvectornonzeroranges[0]) {
    while((rhscvectorconsumednonzeroranges[1] <
	   rhscvectornonzeroranges[1]) &&
	  ((lastrhscibucketindex[0] >= 0) &&
	   (rhscibucketptr[1]->
	    index_RANGE_BEGIN[rhscibucketindex[1]] >=
	    lastrhscibucketptr[0]->
	    index_RANGE_BEGIN[lastrhscibucketindex[0]] &&
	    rhscibucketptr[1]->
	    index_RANGE_END[rhscibucketindex[1]] <=
	    lastrhscibucketptr[0]->
	    index_RANGE_END[lastrhscibucketindex[0]]))) {
      
      /* Note: This part of the code is almost a copy of the part above
       * for the case where condition_2 is TRUE.  Therefore, when
       * modifying either code segment, care must be taken so that
       * both parts remain coherent.
       */

      k = 1;
      temprhscibucketptr[0] = rhscibucketptr[0];
      temprhscvbucketptr[0] = rhscvbucketptr[0];
      rhscibucketptr[0] = lastrhscibucketptr[0];
      rhscvbucketptr[0] = lastrhscvbucketptr[0];

      temprhscvbucketindex[0] = rhscvbucketindex[0];
      
      distancetooverlap =
	rhscibucketptr[1]->index_RANGE_BEGIN[rhscibucketindex[1]] -
	rhscibucketptr[0]->index_RANGE_BEGIN[lastrhscibucketindex[0]];
      distancetoendofvbucket = cVBUCKET_DIM - lastrhscvbucketindex[0];
      
      if (distancetoendofvbucket > distancetooverlap) {
	rhscvbucketindex[0] =
	  MY_MOD(lastrhscvbucketindex[0] + distancetooverlap,
		 cVBUCKET_DIM_minus1);
      }
      else if (distancetoendofvbucket < distancetooverlap) {
	rhscvbucketindex[0] =
	  MY_MOD(distancetooverlap - distancetoendofvbucket,
		 cVBUCKET_DIM_minus1);
      }
      else {
	rhscvbucketindex[0] = 0;
      }
      
      while (distancetooverlap >= distancetoendofvbucket &&
	     rhscvbucketptr[0]->nextvbucket != NULL) {
	rhscvbucketptr[0] = rhscvbucketptr[0]->nextvbucket;
	distancetooverlap -= distancetoendofvbucket;
	distancetoendofvbucket = cVBUCKET_DIM;
      }
      
      totalcurrentnonzerorange = 
      	rhscibucketptr[1]->index_RANGE_END[rhscibucketindex[1]] -
	  rhscibucketptr[1]->index_RANGE_BEGIN[rhscibucketindex[1]]+1;
      
      /* Wrapping the call with the two assignments has the effect of
       * EQUIVALENCEing the data structures being pointed two by the
       * two pointers.
       */
      
      lhscvbucketptr = rhscvbucketptr[0];      
      cOVERLAPPING_CS_RANGE_AXPY(&rhscvbucketindex[0],
 				1,
				&rhscvbucketindex[1],
				alpha[1],
				totalcurrentnonzerorange,
				&unconsumedcurrentnonzerorange,
				rhscvectorconsumednonzeroranges[1],
				rhscvectornonzeroranges[1],
				rhscvectornonzeroranges[0]);
      rhscvbucketptr[0] = lhscvbucketptr;
      
      rhscibucketptr[0] = temprhscibucketptr[0];
      rhscvbucketptr[0] = temprhscvbucketptr[0];
      rhscvbucketindex[0] = temprhscvbucketindex[0];
      
      /*****************************************
       * Perform saxpy:                        *
       * 3.4. If there is one, get new         *
       *      nonzero from the rhs_vector      *
       *      whose nonzero was just consumed. *
       *****************************************/

      if (++rhscvectorconsumednonzeroranges[k] <
	  rhscvectornonzeroranges[k]) {
	if ((rhscibucketindex[k] =
	     MY_MOD(rhscvectorconsumednonzeroranges[k],
		    cIBUCKET_DIM_minus1)) == 0) {
	  if ((rhscibucketptr[k] = rhscibucketptr[k]->nextibucket) ==
	      NULL) {
	    
#ifdef _DEBUG
	    printf("plusqualcspg2 err #75: unexpected NULL nextvbucket: \n");
	    printf("\t k=%d, rhscvectorconsumednonzeroranges=%d,\n",
		   k,rhscvectorconsumednonzeroranges[k]);
	    printf("\t rhscvectornonzeroranges[k]=%d\n",
		   rhscvectornonzeroranges[k]);
	    abort();
#endif
	    
	  }
	}
      }   
    }
    
  }

  if (rhscvectorconsumednonzeroranges[0] <
	 rhscvectornonzeroranges[0] ||
	 rhscvectorconsumednonzeroranges[1] <
	 rhscvectornonzeroranges[1]) {

    restofplusequalcvectorptr->nonzeroes = 0;
    restofplusequalcvectorptr->cds.cs.nonzeroranges =
      rhscvectornonzeroranges[0] - rhscvectorconsumednonzeroranges[0];
    restofplusequalvectorconsumednonzeroranges = 0;
      
    if (rhscvectorconsumednonzeroranges[0] <
	rhscvectornonzeroranges[0]) {

      if (rhscibucketindex[0] == 0 &&
	  rhscibucketptr[0] != lastrhscibucketptr[0]) {
	restofplusequalibucketptr = rhscibucketptr[0];
	restofplusequalibucketindex = rhscibucketindex[0];
	restofplusequalcvectorptr->cds.cs.ibucketptr =
	  restofplusequalibucketptr;
	rhscibucketptr[0] = lastrhscibucketptr[0];
      }
      else {
	cgetibucket(&restofplusequalibucketptr);
	restofplusequalcvectorptr->cds.cs.ibucketptr =
	  restofplusequalibucketptr;
	
	/* Note in the 'for' statement below, the remaining
	 * (not-yet-consumed-in-the saxpy) entries of rhscibucketptr[0]
	 * are placed into restofplusequalibucketptr.
	 */
	
	for (i=rhscibucketindex[0];i<cIBUCKET_DIM;i++) {
	  restofplusequalibucketptr->index_RANGE_BEGIN[i] =
	    rhscibucketptr[0]->index_RANGE_BEGIN[i];
	  restofplusequalibucketptr->index_RANGE_END[i] =
	    rhscibucketptr[0]->index_RANGE_END[i];
	}
	restofplusequalibucketindex = rhscibucketindex[0];
	restofplusequalibucketptr->nextibucket =
	  rhscibucketptr[0]->nextibucket;
	
	/* As distinct from the 'if' part above, there's no need here to set
	 * the values for rhscibucketptr[0] and rhscibucketindex[0],
	 * since they're already pointing where they should be pointing.
	 */
	
      }
      rhscibucketptr[0]->nextibucket = NULL;
      
      if (rhscvbucketindex[0] == 0 &&
	  rhscvbucketptr[0] != lastrhscvbucketptr[0]) {
	restofplusequalvbucketptr = rhscvbucketptr[0];
	restofplusequalvbucketindex = rhscvbucketindex[0];
	restofplusequalcvectorptr->cds.cs.vbucketptr =
	  restofplusequalvbucketptr;
	
	catchupto_rhsvbucketptr[0] = lastrhscvbucketptr[0];
	while (catchupto_rhsvbucketptr[0]->nextvbucket !=
	       rhscvbucketptr[0]) {
  
#ifdef _DEBUG
	  if (catchupto_rhsvbucketptr[0]->nextvbucket == NULL) {
	    printf("plusqualcspg2 error 18: unexpected NULL nextvbucket: \n");
	    abort();
	  }
#endif
	  
	  catchupto_rhsvbucketptr[0] = catchupto_rhsvbucketptr[0]->nextvbucket;
	}
	rhscvbucketptr[0] = catchupto_rhsvbucketptr[0];
	
	rhscvbucketindex[0] = cVBUCKET_DIM;
      }
      else {
	cgetvbucket(&restofplusequalvbucketptr);
	restofplusequalcvectorptr->cds.cs.vbucketptr =
	  restofplusequalvbucketptr;
	
	for (i=rhscvbucketindex[0];i<cVBUCKET_DIM;i++) {
	  restofplusequalvbucketptr->value[i] =
	    rhscvbucketptr[0]->value[i];
	}
	restofplusequalvbucketindex = rhscvbucketindex[0];
	restofplusequalvbucketptr->nextvbucket =
	  rhscvbucketptr[0]->nextvbucket;
      }
      rhscvbucketptr[0]->nextvbucket = NULL;
    }
    else {
      restofplusequalcvectorptr->cds.cs.ibucketptr = NULL;
      restofplusequalcvectorptr->cds.cs.vbucketptr = NULL;
    }
	  
    /* Switching vector names in order not to have to change names
     * in the code that follows.  Now, lhscvector will the new name
     * for the plusequalvector(rhscvector[0]), and rhscvector[0] will
     * be the new name for the restofplusequalvector.
     */
    
    tempvectorptr = lhscvectorptr;
    
    lhscvectorptr = rhscvectorptr[0];
    
    lhscvectornonzeroranges = rhscvectorconsumednonzeroranges[0];
    
    lhscibucketptr = rhscibucketptr[0];
    lhscvbucketptr = rhscvbucketptr[0];
    lhscvbucketindex = rhscvbucketindex[0];
    
    /*
     * lhscibucketindex is computed later using
     * lhscvectornonzeroranges.
     */
    
    lastlhscibucketindex = lastrhscibucketindex[0];
    lastlhscvbucketindex = lastrhscvbucketindex[0];
    lastlhscvbucketptr = lastrhscvbucketptr[0];
    
    /*
     * We never make use of lastlhscibucketptr.
     */

    if (rhscvectorconsumednonzeroranges[0] <
	rhscvectornonzeroranges[0]) {

      rhscvectorptr[0] = restofplusequalcvectorptr;
      rhscibucketptr[0] = rhscvectorptr[0]->cds.cs.ibucketptr;
      rhscvbucketptr[0] = rhscvectorptr[0]->cds.cs.vbucketptr;
      rhscibucketindex[0] = restofplusequalibucketindex;
      rhscvbucketindex[0] = restofplusequalvbucketindex;
      
    }

    while (rhscvectorconsumednonzeroranges[0] <
	   rhscvectornonzeroranges[0] ||
	   rhscvectorconsumednonzeroranges[1] <
	   rhscvectornonzeroranges[1]) {
      
      if
	((rhscvectorconsumednonzeroranges[1] ==
	  rhscvectornonzeroranges[1]) ||
	 ((rhscvectorconsumednonzeroranges[0] <
	   rhscvectornonzeroranges[0]) &&
	  (rhscibucketptr[0]->
	   index_RANGE_BEGIN[rhscibucketindex[0]] <=
	   rhscibucketptr[1]->
	   index_RANGE_BEGIN[rhscibucketindex[1]])))
	  k = 0;
      else if
	((rhscvectorconsumednonzeroranges[0] ==
	  rhscvectornonzeroranges[0]) ||
	 ((rhscvectorconsumednonzeroranges[1] <
	   rhscvectornonzeroranges[1]) &&
	  (rhscibucketptr[0]->
	   index_RANGE_BEGIN[rhscibucketindex[0]] >
	   rhscibucketptr[1]->
	   index_RANGE_BEGIN[rhscibucketindex[1]])))
	  k = 1;
#ifdef _DEBUG
      else {
	printf("plusequalcspg2 error #11: BAD LOGIC.\n");
	abort();
      }
#endif
      j1 = rhscibucketptr[k]->index_RANGE_BEGIN[rhscibucketindex[k]];
      j2 = rhscibucketptr[k]->index_RANGE_END[rhscibucketindex[k]];
      rhscibucketindex[k] =
	MY_MOD(rhscvectorconsumednonzeroranges[k],cIBUCKET_DIM_minus1);
      lhscibucketindex =
	MY_MOD(lhscvectornonzeroranges,cIBUCKET_DIM_minus1);
      
      if ((rhscvbucketindex[k] =
	  MY_MOD(rhscvbucketindex[k],cVBUCKET_DIM_minus1))==0
	  && rhscvectorconsumednonzeroranges[k] !=0
	  && (k !=0 || restofplusequalvectorconsumednonzeroranges !=0)) {
	cHOP_TO_NEXT_RHSVBUCKET(k,&rhscvbucketindex[k],
			       rhscvectorconsumednonzeroranges[k],
			       rhscvectornonzeroranges[k]);
      }
      lhscvbucketindex =
	MY_MOD(lhscvbucketindex,cVBUCKET_DIM_minus1);
      
#ifdef _DEBUG
      if (rhscibucketptr[k]->errorflag != 9099 ||
	  rhscvbucketptr[k]->errorflag != 9009) {
	printf("plusequalcspg2 err 14: memory corruption of ibucketstore ");
	printf("or vbucketstore.\n");
	abort();
      }
#endif      
      
      /****************************************
       * Perform saxpy:                       *
       * 3.0. Check for startup errors.       *
       ****************************************/
      
#ifdef _DEBUG
      if (lhscvectornonzeroranges == 0) {
	if (rhscibucketindex[k] != 0 || rhscvbucketindex[k] != 0 ||
	    lhscibucketindex != 0 || lhscvbucketindex != 0) {
	  printf("plusequalcspg2 error 15: Zeroes expected: \n\t");
	  printf("rhscibucketindex[k]=%d, rhscvbucketindex[k]=%d\n\t",
		 rhscibucketindex[k], rhscvbucketindex[k]);
	  printf("lhscibucketindex=%d, lhscvbucketindex=%d\n\t",
		 lhscibucketindex,lhscvbucketindex);
	  abort();
	}
      }
#endif      
      
      /****************************************
       * Perform saxpy:                       *
       * 3.1. Found 1ST or NEW subscript      *
       *      range (non-overlapping with     *
       *      previous one, if any) for       *
       *      lhscvector.  If needed, get new  *
       *      ibuckets and vbuckets for       *
       *      lhscvector.                      *
       ****************************************/
      if (lhscvectornonzeroranges == 0 ||
	  j1 > lhscibucketptr->index_RANGE_END[lastlhscibucketindex]+1) {/*A.1*/
	
	if (lhscibucketindex == 0 && lhscvectornonzeroranges != 0) {
	  cgetibucket(&(lhscibucketptr->nextibucket));
	  lhscibucketptr = lhscibucketptr->nextibucket;
	}
	lhscibucketptr->index_RANGE_BEGIN[lhscibucketindex] = j1;
	lhscibucketptr->index_RANGE_END[lhscibucketindex] = j2;
	
	if (lhscvbucketindex == 0 && lhscvectornonzeroranges != 0)
	  cget_and_hop_to_next_vbucket(&lhscvbucketindex);
	
	lastlhscvbucketptr = lhscvbucketptr;
	lastlhscvbucketindex = lhscvbucketindex;
	
	totalcurrentnonzerorange = j2-j1+1;
	
	lhscvectornonzeroes += totalcurrentnonzerorange;
	lhscvectornonzeroranges++;
	
	
	cNON_OVER_LAPPING_CS_RANGE_AXPY(COMPLETELY_NEW_RANGE,
				       lhscibucketindex,
				       &lastlhscibucketindex,
				       &lhscvbucketindex,
				       k,
				       &rhscvbucketindex[k],
				       alpha[k],
				       totalcurrentnonzerorange,
				       &unconsumedcurrentnonzerorange,
				       rhscvectorconsumednonzeroranges[k],
				       rhscvectornonzeroranges[k],
				       lhscvectornonzeroranges);
	
	
      } /*A.1*/
      
      
      /****************************************
       * Perform saxpy:                       *
       * 3.2. Found old subscript for         *
       *      lhscvector.  Consume  nonzero in *
       *      saxpy.                          *
       *                                      *
       ****************************************/
      else if (j1 <= 
	       lhscibucketptr->index_RANGE_END[lastlhscibucketindex]+1 &&
	       j1 >= 
	       lhscibucketptr->index_RANGE_BEGIN[lastlhscibucketindex]) { /*B.1*/
	
	templhscvbucketptr = lhscvbucketptr;
	templhscvbucketindex = lhscvbucketindex;
	lhscvbucketptr = lastlhscvbucketptr;
	
	distancetooverlap = j1 -
	  lhscibucketptr->index_RANGE_BEGIN[lastlhscibucketindex];
	distancetoendofvbucket = cVBUCKET_DIM -
	  lastlhscvbucketindex;
	
	if (j1 ==
	    lhscibucketptr->index_RANGE_END[lastlhscibucketindex]+1) {
	  h1 = 0;
	  h2 = -1;
	}
	else if (j2 <=
		 lhscibucketptr->index_RANGE_END[lastlhscibucketindex]) {
	  h1 = j1;
	  h2 = j2;
	  j1 = 0;
	  j2 = -1;
	}
	else {
	  h1 = j1;
	  h2 = lhscibucketptr->index_RANGE_END[lastlhscibucketindex];
	  j1 = h2+1;
	}
	
	if (distancetoendofvbucket > distancetooverlap) {
	  lhscvbucketindex =
	    MY_MOD(lastlhscvbucketindex + distancetooverlap,
		   cVBUCKET_DIM_minus1);
	}
	else if (distancetoendofvbucket < distancetooverlap) {
	  lhscvbucketindex =
	    MY_MOD(distancetooverlap - distancetoendofvbucket,
		   cVBUCKET_DIM_minus1);
	}
	else {
	  lhscvbucketindex = 0;
	}
	
	while (distancetooverlap >= distancetoendofvbucket &&
	       lhscvbucketptr->nextvbucket != NULL) {
	  lhscvbucketptr = lhscvbucketptr->nextvbucket;
	  distancetooverlap -= distancetoendofvbucket;
	  distancetoendofvbucket = cVBUCKET_DIM;
	}
	
	lhscibucketptr->index_RANGE_END[lastlhscibucketindex] =
	  max(j2,
	      lhscibucketptr->index_RANGE_END[lastlhscibucketindex]);
	
	if ((totalcurrentnonzerorange = h2-h1+1) > 0) {
	  
	  
	  cOVERLAPPING_CS_RANGE_AXPY(&lhscvbucketindex,
				    k,
				    &rhscvbucketindex[k],
				    alpha[k],
				    totalcurrentnonzerorange,
				    &unconsumedcurrentnonzerorange,
				    rhscvectorconsumednonzeroranges[k],
				    rhscvectornonzeroranges[k],
				    lhscvectornonzeroranges);
	  
	  
	}
	
	/*
	 * This bit of logic must be performed, whether or not there
	 * is additional nonzeroes to process.
	 */
	
	if (j2 != -1 &&
	    (lhscvbucketptr != templhscvbucketptr ||
	     (lhscvbucketindex =
	      MY_MOD(lhscvbucketindex,cVBUCKET_DIM_minus1)) !=
	     templhscvbucketindex)) {
	  
#ifdef _DEBUG
	  printf("plusequalcspg2 err 50.0: Unexpectd value discrepancies\n");
	  printf("lhscvbucketptr=%d, templhscvbucketptr=%d\n",
		 (int)lhscvbucketptr,(int)templhscvbucketptr);
	  printf("lhscvbucketindex=%d, templhscvbucketindex=%d\n",
		 lhscvbucketindex,templhscvbucketindex);
	  
	  abort();
#endif
	}
	else {
	  lhscvbucketptr = templhscvbucketptr;
	  lhscvbucketindex = templhscvbucketindex;
	}
	
	
	if ((totalcurrentnonzerorange = j2-j1+1) > 0) {
	  lhscvectornonzeroes += totalcurrentnonzerorange;
	  
	  /*
	   * adding new entries to vbucket
	   */
	  
	  if (lhscvbucketindex == 0 &&
	      lhscvectornonzeroranges != 0 &&
	      j2 != -1)
	    cget_and_hop_to_next_vbucket(&lhscvbucketindex);
	  
	  if(rhscvbucketindex[k] == cVBUCKET_DIM && j2 != -1) {
	    
	    cHOP_TO_NEXT_RHSVBUCKET(k,&rhscvbucketindex[k],
				   rhscvectorconsumednonzeroranges[k],
				   rhscvectornonzeroranges[k]);
	  }
	  
	  
	  
	  
	  cNON_OVER_LAPPING_CS_RANGE_AXPY(NEW_ENTRIES_PADDED_TO_OLD_RANGE,
					 lhscibucketindex,
					 &lastlhscibucketindex,
					 &lhscvbucketindex,
					 k,
					 &rhscvbucketindex[k],
					 alpha[k],
					 totalcurrentnonzerorange,
					 &unconsumedcurrentnonzerorange,
					 rhscvectorconsumednonzeroranges[k],
					 rhscvectornonzeroranges[k],
					 lhscvectornonzeroranges);
	  
	  
	} /* matches     if ((totalcurrent...) { */
      } /*B.1*/
      
      
      /****************************************	
       * Perform saxpy:                       *
       * 3.3. Found wrong subscript for       *
       *      lhscvector.  Exit.               *
       ****************************************/
      
#ifdef _DEBUG
      else {
	printf("plusequalcspg2 error 60: lhscvectornonzeroranges=%d,",
	       lhscvectornonzeroranges);
	printf("lastlhscibucketindex=%d, \n\t",lastlhscibucketindex);
	printf("lhscibucketptr->index[%d] = (%d,%d), j1=%d\n",
	       lastlhscibucketindex,
	       lhscibucketptr->
	       index_RANGE_BEGIN[lastlhscibucketindex],
	       lhscibucketptr->
	       index_RANGE_END[lastlhscibucketindex],
	       j1);
	abort();
      }
#endif
      
      
      /****************************************
       * Perform saxpy:                       *
       * 3.4. If there is one, get new        *
       *      nonzero from the rhs_vector     *
       *      whose nonzero was just          *
       *      consumed and replace the        *
       *      heaproot with new entry;        *
       *      Else, delete heaproot.          *
       ****************************************/

      if (k==0) restofplusequalvectorconsumednonzeroranges++;

      if (++rhscvectorconsumednonzeroranges[k] <
	  rhscvectornonzeroranges[k]) {
	if ((rhscibucketindex[k] =
	     MY_MOD(rhscvectorconsumednonzeroranges[k],
		    cIBUCKET_DIM_minus1)) == 0) {
	  if ((rhscibucketptr[k] = rhscibucketptr[k]->nextibucket) ==
	      NULL) {
	    
#ifdef _DEBUG
	    printf("plusequalcspg2 err 61: unexpected NULL nextvbucket: \n");
	    printf("\t k=%d, rhscvectorconsumednonzeroranges=%d,\n",
		   k,rhscvectorconsumednonzeroranges[k]);
	    printf("\t rhscvectornonzeroranges[k]=%d\n",
		   rhscvectornonzeroranges[k]);
	    abort();
#endif
	    
	  }
	}
      }  
      
    } /* matches     2nd while loop           */
  } 

  else {

    /* tempvectorptr is used in the if part as temporary storage
     * for the value of lhscvectorptr.  We echo that statement
     * here so that later the reinstatement of the value of
     * lhscvectorptr could be done regardless of the branch taken.
     */

    tempvectorptr = lhscvectorptr;
 
  }

  rhscvectorptrptrs[1] = &rhscvectorptr[1];
  
  cplusequal_axpy_postprocessing(rhsanddestptrptr,
				 rhscvectorptrptrs[1],
				 status_otherrhscvector,
				 &restofplusequalcvectorptr,
				 alphaptrs,
				 restorerhscvectorptr,
				 arity,
				 lhscvectornonzeroes,
				 lhscvectornonzeroranges,
				 CS);
  
  
  
  lhscvectorptr =  tempvectorptr;
  
  
#ifdef _TRACE 
  nnza = (*destptrptr)->nonzeroes; 
  cadinfo(arity,nnzb,lhscvectoraddress,nnza,alphaptrs,vectorptrptrs);
#endif

}

