#define COMPLEX
/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#include  <stdlib.h>
#include "cgaxpy_pd_macros_and_data_structures.h"
#include "cgaxpy_pd_global_variables.h"

/****************************************************************************
 *                                                                          *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/

void cnon_over_lapping_cs_range_axpy(rangedestination,
				    lhscibucketindex,
				    addressoflastlhscibucketindex,
				    addressoflhscvbucketindex,
				    k,
				    addressofrhscvbucketindex_k,
				    alpha_k,
				    totalcurrentnonzerorange
#ifdef _DEBUG
				    ,addressofunconsumedcurrentnonzerorange
				    ,rhscvectorconsumednonzeroranges_k
				    ,rhscvectornonzeroranges_k
				    ,lhscvectornonzeroranges
#endif
				    )
int rangedestination;
int lhscibucketindex;
int *addressoflastlhscibucketindex;
int *addressoflhscvbucketindex;
int k;
int *addressofrhscvbucketindex_k;
complex alpha_k;
int totalcurrentnonzerorange;
#ifdef _DEBUG
int *addressofunconsumedcurrentnonzerorange;
int rhscvectorconsumednonzeroranges_k;
int rhscvectornonzeroranges_k;
int lhscvectornonzeroranges;
#endif
{
  int firstloopsize;
  int numberofdoubleforloopiterations;
  int firstiterationloopsize;
  int seconditerationloopsize;
  int remainder1loopsize;
  int remainder2loopsize;

  int i;


  cDETERMINE_LOOP_BOUNDS(totalcurrentnonzerorange,
			*addressoflhscvbucketindex,
			*addressofrhscvbucketindex_k,
			&firstloopsize,
			&numberofdoubleforloopiterations,
			&firstiterationloopsize,
			&seconditerationloopsize,
			&remainder1loopsize,
			&remainder2loopsize,
			addressofunconsumedcurrentnonzerorange);
  cDENSE_AXPY_LOOP(NEW_RANGE,
		  addressoflhscvbucketindex,
		  addressofrhscvbucketindex_k,
		  k,
		  firstloopsize,
		  alpha_k,
		  addressofunconsumedcurrentnonzerorange);
  
    
  /*********************************************
   * Case 1: rhscvbucket has more spaces   *
   *         remaining than lhscvbucket.   *
   *********************************************/
  
  if (*addressofrhscvbucketindex_k <
      *addressoflhscvbucketindex) {
    
    for(i=0;i<numberofdoubleforloopiterations;i++) {
      cDEBUGGING_CODE_1(LEFT,
		       *addressoflhscvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      cget_and_hop_to_next_vbucket(addressoflhscvbucketindex);
      cDENSE_AXPY_LOOP(NEW_RANGE,
		      addressoflhscvbucketindex,
		      addressofrhscvbucketindex_k,
		      k,
		      firstiterationloopsize,alpha_k,
		      addressofunconsumedcurrentnonzerorange);
      cDEBUGGING_CODE_1(RIGHT,
		       *addressofrhscvbucketindex_k,
		       *addressofunconsumedcurrentnonzerorange);
      cHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhscvbucketindex_k,
			     rhscvectorconsumednonzeroranges_k,
			     rhscvectornonzeroranges_k);
      cDENSE_AXPY_LOOP(NEW_RANGE,
		      addressoflhscvbucketindex,
		      addressofrhscvbucketindex_k,
		      k,
		      seconditerationloopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    
    if (remainder1loopsize > 0) {
      cDEBUGGING_CODE_1(LEFT,
		       *addressoflhscvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      cget_and_hop_to_next_vbucket(addressoflhscvbucketindex);
      cDENSE_AXPY_LOOP(NEW_RANGE,
		      addressoflhscvbucketindex,
		      addressofrhscvbucketindex_k,
		      k,
		      remainder1loopsize,alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    if (remainder2loopsize > 0) {
      cDEBUGGING_CODE_1(RIGHT,
		       *addressofrhscvbucketindex_k,
		       *addressofunconsumedcurrentnonzerorange);
      cHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhscvbucketindex_k,
			     rhscvectorconsumednonzeroranges_k,
			     rhscvectornonzeroranges_k);
      cDENSE_AXPY_LOOP(NEW_RANGE,
		      addressoflhscvbucketindex,
		      addressofrhscvbucketindex_k,
		      k,
		      remainder2loopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    DEBUGGING_CODE_2(*addressofunconsumedcurrentnonzerorange);
    if (rangedestination == COMPLETELY_NEW_RANGE) {
      *addressoflastlhscibucketindex = lhscibucketindex;
    }
  }

  /*********************************************
   * Case 2: rhscvbucket has less spaces   *
   * remaining than lhscvbucket.           *
   *********************************************/	  
  else if (*addressofrhscvbucketindex_k >
	   *addressoflhscvbucketindex) {
    
    for(i=0;i<numberofdoubleforloopiterations;i++) {
      cDEBUGGING_CODE_1(RIGHT,
		       *addressofrhscvbucketindex_k,
		       *addressofunconsumedcurrentnonzerorange);
      cHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhscvbucketindex_k,
			     rhscvectorconsumednonzeroranges_k,
			     rhscvectornonzeroranges_k);
      cDENSE_AXPY_LOOP(NEW_RANGE,
		      addressoflhscvbucketindex,
		      addressofrhscvbucketindex_k,
		      k,
		      firstiterationloopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
      cDEBUGGING_CODE_1(LEFT,
		       *addressoflhscvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      cget_and_hop_to_next_vbucket(addressoflhscvbucketindex);
      cDENSE_AXPY_LOOP(NEW_RANGE,
		      addressoflhscvbucketindex,
		      addressofrhscvbucketindex_k,
		      k,
		      seconditerationloopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    
    if (remainder1loopsize > 0) {
      cDEBUGGING_CODE_1(RIGHT,
		       *addressofrhscvbucketindex_k,
		       *addressofunconsumedcurrentnonzerorange);
      cHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhscvbucketindex_k,
			     rhscvectorconsumednonzeroranges_k,
			     rhscvectornonzeroranges_k);
      cDENSE_AXPY_LOOP(NEW_RANGE,
		      addressoflhscvbucketindex,
		      addressofrhscvbucketindex_k,
		      k,
		      remainder1loopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    
    if (remainder2loopsize > 0) {
      cDEBUGGING_CODE_1(LEFT,
		       *addressoflhscvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      cget_and_hop_to_next_vbucket(addressoflhscvbucketindex);
      cDENSE_AXPY_LOOP(NEW_RANGE,
		      addressoflhscvbucketindex,
		      addressofrhscvbucketindex_k,
		      k,
		      remainder2loopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    DEBUGGING_CODE_2(*addressofunconsumedcurrentnonzerorange);
    if (rangedestination == COMPLETELY_NEW_RANGE) {
      *addressoflastlhscibucketindex = lhscibucketindex;
    }
  }
  
   
  /*********************************************
   * Case 3: rhscvbucket has same # of     *
   * spaces as lhscvbucket (i.e. 0).       *
   *********************************************/
  else if (*addressofrhscvbucketindex_k ==
	   *addressoflhscvbucketindex) {
    
    for(i=0;i<numberofdoubleforloopiterations;i++) {
      cDEBUGGING_CODE_1(LEFT,
		       *addressoflhscvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      cget_and_hop_to_next_vbucket(addressoflhscvbucketindex);
      cHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhscvbucketindex_k,
			     rhscvectorconsumednonzeroranges_k,
			     rhscvectornonzeroranges_k);
      cDENSE_AXPY_LOOP(NEW_RANGE,
		      addressoflhscvbucketindex,
		      addressofrhscvbucketindex_k,
		      k,
		      firstiterationloopsize+seconditerationloopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    if (remainder1loopsize+remainder2loopsize > 0) {
      cDEBUGGING_CODE_1(LEFT,
		       *addressoflhscvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      cget_and_hop_to_next_vbucket(addressoflhscvbucketindex);
      cHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhscvbucketindex_k,
			     rhscvectorconsumednonzeroranges_k,
			     rhscvectornonzeroranges_k);
      cDENSE_AXPY_LOOP(NEW_RANGE,
		      addressoflhscvbucketindex,
		      addressofrhscvbucketindex_k,
		      k,
		      remainder1loopsize+remainder2loopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    DEBUGGING_CODE_2(*addressofunconsumedcurrentnonzerorange);
    if (rangedestination == COMPLETELY_NEW_RANGE) {
      *addressoflastlhscibucketindex = lhscibucketindex;
    }
  }
}






/****************************************************************************
 *                                                                          *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/

void coverlapping_cs_range_axpy(addressoflhscvbucketindex,
			       k,
			       addressofrhscvbucketindex_k,
			       alpha_k,
			       totalcurrentnonzerorange
#ifdef _DEBUG
			       ,addressofunconsumedcurrentnonzerorange
			       ,rhscvectorconsumednonzeroranges_k
			       ,rhscvectornonzeroranges_k
			       ,lhscvectornonzeroranges
#endif
			       )
int *addressoflhscvbucketindex;
int k;
int *addressofrhscvbucketindex_k;
complex alpha_k;
int totalcurrentnonzerorange;
#ifdef _DEBUG
int *addressofunconsumedcurrentnonzerorange;
int rhscvectorconsumednonzeroranges_k;
int rhscvectornonzeroranges_k;
int lhscvectornonzeroranges;
#endif
{
  int firstloopsize;
  int numberofdoubleforloopiterations;
  int firstiterationloopsize;
  int seconditerationloopsize;
  int remainder1loopsize;
  int remainder2loopsize;

  int i;
  
  cDETERMINE_LOOP_BOUNDS(totalcurrentnonzerorange,
			*addressoflhscvbucketindex,
			*addressofrhscvbucketindex_k,
			&firstloopsize,
			&numberofdoubleforloopiterations,
			&firstiterationloopsize,
			&seconditerationloopsize,
			&remainder1loopsize,
			&remainder2loopsize,
			addressofunconsumedcurrentnonzerorange);
  cDENSE_AXPY_LOOP(OLD_RANGE,
		  addressoflhscvbucketindex,
		  addressofrhscvbucketindex_k,
		  k,
		  firstloopsize,
		  alpha_k,
		  addressofunconsumedcurrentnonzerorange);
  
  
  /*********************************************
   * Case 1: rhscvbucket has more spaces   *
   *         remaining than lhscvbucket.   *
   *********************************************/
  if (*addressofrhscvbucketindex_k <
      *addressoflhscvbucketindex) {
    /*B.2.1*/
    
    for(i=0;i<numberofdoubleforloopiterations;i++) { /*B.3.1*/
      cDEBUGGING_CODE_1(LEFT,
		       *addressoflhscvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      cHOP_TO_NEXT_LHSVBUCKET(addressoflhscvbucketindex,
			     lhscvectornonzeroranges);
      cDENSE_AXPY_LOOP(OLD_RANGE,
		      addressoflhscvbucketindex,
		      addressofrhscvbucketindex_k,
		      k,
		      firstiterationloopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
      cDEBUGGING_CODE_1(RIGHT,
		       *addressofrhscvbucketindex_k,
		       *addressofunconsumedcurrentnonzerorange);
      cHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhscvbucketindex_k,
			     rhscvectorconsumednonzeroranges_k,
			     rhscvectornonzeroranges_k);
      cDENSE_AXPY_LOOP(OLD_RANGE,
		      addressoflhscvbucketindex,
		      addressofrhscvbucketindex_k,
		      k,
		      seconditerationloopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    } /*B.3.1*/
    
    if (remainder1loopsize > 0) {    	    
      cDEBUGGING_CODE_1(LEFT,
		       *addressoflhscvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      cHOP_TO_NEXT_LHSVBUCKET(addressoflhscvbucketindex,
			     lhscvectornonzeroranges);
      cDENSE_AXPY_LOOP(OLD_RANGE,
		      addressoflhscvbucketindex,
		      addressofrhscvbucketindex_k,
		      k,
		      remainder1loopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    
    if (remainder2loopsize > 0) {    
      cDEBUGGING_CODE_1(RIGHT,
		       *addressofrhscvbucketindex_k,
		       *addressofunconsumedcurrentnonzerorange);
      cHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhscvbucketindex_k,
			     rhscvectorconsumednonzeroranges_k,
			     rhscvectornonzeroranges_k);
      cDENSE_AXPY_LOOP(OLD_RANGE,
		      addressoflhscvbucketindex,
		      addressofrhscvbucketindex_k,
		      k,
		      remainder2loopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    DEBUGGING_CODE_2(*addressofunconsumedcurrentnonzerorange);
  } /*B.2.1*/
  
  /*********************************************
   * Case 2: rhscvbucket has less spaces   *
   * remaining than lhscvbucket.           *
   *********************************************/
  else if (*addressofrhscvbucketindex_k >
	   *addressoflhscvbucketindex) {
 /*B.2.2*/
    
    for(i=0;i<numberofdoubleforloopiterations;i++) { /*B.3.2*/
      cDEBUGGING_CODE_1(RIGHT,
		       *addressofrhscvbucketindex_k,
		       *addressofunconsumedcurrentnonzerorange);
      cHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhscvbucketindex_k,
			     rhscvectorconsumednonzeroranges_k,
			     rhscvectornonzeroranges_k);
      cDENSE_AXPY_LOOP(OLD_RANGE,
		      addressoflhscvbucketindex,
		      addressofrhscvbucketindex_k,
		      k,
		      firstiterationloopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
      cDEBUGGING_CODE_1(LEFT,
		       *addressoflhscvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      cHOP_TO_NEXT_LHSVBUCKET(addressoflhscvbucketindex,
			     lhscvectornonzeroranges);
      cDENSE_AXPY_LOOP(OLD_RANGE,
		      addressoflhscvbucketindex,
		      addressofrhscvbucketindex_k,
		      k,
		      seconditerationloopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    } /*B.3.2*/
    
    if (remainder1loopsize > 0) {        
      cDEBUGGING_CODE_1(RIGHT,
		       *addressofrhscvbucketindex_k,
		       *addressofunconsumedcurrentnonzerorange);
      cHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhscvbucketindex_k,
			     rhscvectorconsumednonzeroranges_k,
			     rhscvectornonzeroranges_k);
      cDENSE_AXPY_LOOP(OLD_RANGE,
		      addressoflhscvbucketindex,
		      addressofrhscvbucketindex_k,
		      k,
		      remainder1loopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    if (remainder2loopsize > 0) {    
      cDEBUGGING_CODE_1(LEFT,
		       *addressoflhscvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      cHOP_TO_NEXT_LHSVBUCKET(addressoflhscvbucketindex,
			     lhscvectornonzeroranges);
      cDENSE_AXPY_LOOP(OLD_RANGE,
		      addressoflhscvbucketindex,
		      addressofrhscvbucketindex_k,
		      k,
		      remainder2loopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    DEBUGGING_CODE_2(*addressofunconsumedcurrentnonzerorange);
  } /*B.2.2*/
  
  
  /*********************************************
   * Case 3: rhscvbucket has same # of     *
   * spaces as lhscvbucket (i.e. 0).       *
   *********************************************/
  else if(*addressofrhscvbucketindex_k ==
	  *addressoflhscvbucketindex) {
    /*B.2.3*/
    
    for(i=0;i<numberofdoubleforloopiterations;i++) { /*B.3.3*/
      cDEBUGGING_CODE_1(LEFT,
		       *addressoflhscvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      cHOP_TO_NEXT_LHSVBUCKET(addressoflhscvbucketindex,
			     lhscvectornonzeroranges);
      cHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhscvbucketindex_k,
			     rhscvectorconsumednonzeroranges_k,
			     rhscvectornonzeroranges_k);
      cDENSE_AXPY_LOOP(OLD_RANGE,
		      addressoflhscvbucketindex,
		      addressofrhscvbucketindex_k,
		      k,
		      firstiterationloopsize+seconditerationloopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    } /*B.3.3*/
    
    if (remainder1loopsize+remainder2loopsize > 0) {    
      cDEBUGGING_CODE_1(LEFT,
		       *addressoflhscvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      cHOP_TO_NEXT_LHSVBUCKET(addressoflhscvbucketindex,
			     lhscvectornonzeroranges);
      cHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhscvbucketindex_k,
			     rhscvectorconsumednonzeroranges_k,
			     rhscvectornonzeroranges_k);
      cDENSE_AXPY_LOOP(OLD_RANGE,
		      addressoflhscvbucketindex,
		      addressofrhscvbucketindex_k,
		      k,
		      remainder1loopsize+remainder2loopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    DEBUGGING_CODE_2(*addressofunconsumedcurrentnonzerorange);
  } /*B.2.3*/
}



/****************************************************************************
 *                                                                          *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/

void chop_to_next_rhsvbucket(k,
			    vbucketindex
#ifdef _DEBUG
			    ,rhscvectorconsumednonzeroranges
			    ,rhscvectornonzeroranges
#endif
			    )
int k;
int *vbucketindex;
#ifdef _DEBUG
int rhscvectorconsumednonzeroranges;
int rhscvectornonzeroranges;
#endif
{
  if ((rhscvbucketptr[k] = rhscvbucketptr[k]->nextvbucket) == NULL) {
	      
#ifdef _DEBUG
    printf("chop_to_next_rhsvbucket error: unexpected NULL nextvbucket: \n");
    printf("\t k=%d, rhscvectorconsumednonzeroranges[k]=%d,\n",
	   k,rhscvectorconsumednonzeroranges);
    printf("\t rhscvectornonzeroranges[k]=%d\n",
	   rhscvectornonzeroranges);
    abort();
#endif
  }
  else {
    *vbucketindex = 0;
  }
}





/****************************************************************************
 *                                                                          *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/

void chop_to_next_lhsvbucket(vbucketindex
#ifdef _DEBUG
		      ,lhscvectornonzeroranges
#endif
		      )
int *vbucketindex;
#ifdef _DEBUG
int lhscvectornonzeroranges;
#endif
{
  if ((lhscvbucketptr = lhscvbucketptr->nextvbucket) == NULL) {
    
#ifdef _DEBUG
    printf("chop_to_next_lhsvbucket error: unexpected NULL nextvbucket: \n");
    printf("lhscvectornonzeroranges=%d\n",lhscvectornonzeroranges);
    abort();
#endif
  }
  else {
    *vbucketindex = 0;
  }
}  





/****************************************************************************
 *                                                                          *
 *      cget_and_hop_to_next_vbucket(lhscvbucketindex)                   *
 *                                                                          *
 ****************************************************************************/

void cget_and_hop_to_next_vbucket(lhscvbucketindex)
int *lhscvbucketindex;
{
  cgetvbucket(&(lhscvbucketptr->nextvbucket));
  lhscvbucketptr = lhscvbucketptr->nextvbucket;
  *lhscvbucketindex = 0;
}







/****************************************************************************
 *                                                                          *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/

void cconvertdstype(oldvectorptrptr,k,newvectordstype)
cvectorstruct **oldvectorptrptr;
int k;
int newvectordstype;
{
  cbucketstruct *holdingbinbucketptr;
  cbucketstruct *oldbucketptr;
  cibucketstruct *holdingbinibucketptr;
  cvbucketstruct *holdingbinvbucketptr;
  int bucketindex;
  int ibucketindex = 0;
  int vbucketindex;
  int j;
  int ij;
  int vj;
  int lastindex;
  cvectorstruct *oldvectorptr = (*oldvectorptrptr);
  int oldvectordstype = oldvectorptr->dstype;

  if (oldvectordstype >= newvectordstype || newvectordstype == S1) {
#ifdef _DEBUG
    printf("cconvertdstype error 1: Illegal Data Structure Conversion:\n\t");
    printf("newvectordstype=%d,    oldvectordstype=%d\n",
	   newvectordstype,oldvectordstype);
    abort();
#endif
  }
  else if (oldvectordstype == S1 &&  newvectordstype == SS) {
    holdingbincvectorptr[k]->nonzeroes = 1;
    holdingbincvectorptr[k]->dstype = SS;
    cgetbucket(&holdingbinbucketptr);
    holdingbincvectorptr[k]->cds.ss.bucketptr = holdingbinbucketptr;
    holdingbinbucketptr->index[0] = oldvectorptr->cds.s1.index;
    holdingbinbucketptr->value[0] = oldvectorptr->cds.s1.value;
  }
  else if (oldvectordstype == S1 &&  newvectordstype == CS) {
    holdingbincvectorptr[k]->nonzeroes = 1;
    holdingbincvectorptr[k]->dstype = CS;
    cgetibucket(&holdingbinibucketptr);
    cgetvbucket(&holdingbinvbucketptr);
    holdingbincvectorptr[k]->cds.cs.nonzeroranges = 1;
    holdingbincvectorptr[k]->cds.cs.ibucketptr = holdingbinibucketptr;
    holdingbincvectorptr[k]->cds.cs.vbucketptr = holdingbinvbucketptr;
    holdingbinibucketptr->index_RANGE_BEGIN[0] = oldvectorptr->cds.s1.index;
    holdingbinibucketptr->index_RANGE_END[0] = oldvectorptr->cds.s1.index;
    holdingbinvbucketptr->value[0] = oldvectorptr->cds.s1.value;
  }
  else if (oldvectordstype == SS &&  newvectordstype == CS) {
    oldbucketptr = oldvectorptr->cds.ss.bucketptr;
    holdingbincvectorptr[k]->nonzeroes = oldvectorptr->nonzeroes;
    holdingbincvectorptr[k]->dstype = CS;
    cgetibucket(&holdingbinibucketptr);
    holdingbincvectorptr[k]->cds.cs.ibucketptr = holdingbinibucketptr;
    cgetvbucket(&holdingbinvbucketptr);
    holdingbincvectorptr[k]->cds.cs.vbucketptr = holdingbinvbucketptr;
    holdingbinibucketptr->index_RANGE_BEGIN[0] = oldbucketptr->index[0];
    holdingbinvbucketptr->value[0] = oldbucketptr->value[0];
    lastindex = oldbucketptr->index[0];
    
    for (j=1,ij=0,vj=1;j<oldvectorptr->nonzeroes;j++,vj++) {
      if ((bucketindex = MY_MOD(j,cBUCKET_DIM_minus1)) == 0)
	  oldbucketptr = oldbucketptr->nextbucket;
      if ((vbucketindex = MY_MOD(vj,cVBUCKET_DIM_minus1)) == 0) {
	cgetvbucket(&holdingbinvbucketptr->nextvbucket);
	holdingbinvbucketptr = holdingbinvbucketptr->nextvbucket;
      }
      holdingbinvbucketptr->value[vbucketindex] =
	oldbucketptr->value[bucketindex];

      if (oldbucketptr->index[bucketindex] > lastindex+1) {
	holdingbinibucketptr->index_RANGE_END[ibucketindex] = lastindex;
	ij++;
	if ((ibucketindex = MY_MOD(ij,cIBUCKET_DIM_minus1)) == 0) {
	  cgetibucket(&holdingbinibucketptr->nextibucket);
	  holdingbinibucketptr = holdingbinibucketptr->nextibucket;
	}
	holdingbinibucketptr->index_RANGE_BEGIN[ibucketindex] =
	  oldbucketptr->index[bucketindex];
      }
      else if (oldbucketptr->index[bucketindex] < lastindex+1) {
#ifdef _DEBUG
	printf("cconvertdstype error 2: BAD LOGIC:\n\t");
	printf("newvectordstype=%d,    oldvectordstype=%d\n",
	       newvectordstype,oldvectordstype);
	abort();
#endif
      }
      lastindex = oldbucketptr->index[bucketindex];
    }
    holdingbinibucketptr->index_RANGE_END[ibucketindex] = lastindex;    
    holdingbincvectorptr[k]->cds.cs.nonzeroranges = ij+1;
  }
  else /* We should never go thru this branch */ {
#ifdef _DEBUG
    printf("cconvertdstype error 3: BAD LOGIC:\n\t");
    printf("newvectordstype=%d,    oldvectordstype=%d\n",
	   newvectordstype,oldvectordstype);
    abort();
#endif
  }
  *oldvectorptrptr =  holdingbincvectorptr[k];
}








/****************************************************************************
 *                                                                          *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/

void caxpy_0_1_post_processing(destptrptr)
cvectorstruct **destptrptr;
{
  cvectorstruct *tempvectorptr;
  cbucketstruct *tempbucketptr;
  cibucketstruct *tempibucketptr;
  cvbucketstruct *tempvbucketptr;
  
  /*********************************************
   * Switch lhscvectorptr and *destptrptr. *
   *********************************************/
  tempvectorptr = (*destptrptr);
  (*destptrptr) = lhscvectorptr;
  lhscvectorptr = tempvectorptr;
  
  if (tempvectorptr->dstype == SS) {
    /*********************************************
     * If *destptrptr was not empty,             *
     * add its buckets to freedcbucketslist. *
     *********************************************/
    tempbucketptr = tempvectorptr->cds.ss.bucketptr;
    if (tempbucketptr != NULL) {
      while (tempbucketptr->nextbucket != NULL)
	tempbucketptr = tempbucketptr->nextbucket;
      tempbucketptr->nextbucket = freedcbucketslist;
      RESET_freedcbucketslist(tempvectorptr->cds.ss.bucketptr);
      tempvectorptr->cds.ss.bucketptr = NULL;
    }
  }
  else if (tempvectorptr->dstype == CS) {
    /********************************************
     * If *destptrptr was not empty, add        *
     * its ibuckets to freedcibucketslist.  *
     ********************************************/
    
    tempibucketptr = tempvectorptr->cds.cs.ibucketptr;
    tempvbucketptr = tempvectorptr->cds.cs.vbucketptr;
    
    if (tempibucketptr != NULL) {
      while (tempibucketptr->nextibucket != NULL)
	tempibucketptr = tempibucketptr->nextibucket;
      tempibucketptr->nextibucket = freedcibucketslist;
      RESET_freedcibucketslist(tempvectorptr->cds.cs.ibucketptr);
      tempvectorptr->cds.cs.ibucketptr = NULL;
    }
    
    /********************************************
     * If *destptrptr was not empty, add        *
     * its vbuckets to freedcvbucketslist.  *
     ********************************************/
    
    if (tempvbucketptr != NULL) {
      while (tempvbucketptr->nextvbucket != NULL)
	tempvbucketptr = tempvbucketptr->nextvbucket;
      tempvbucketptr->nextvbucket = freedcvbucketslist;
      RESET_freedcvbucketslist(tempvectorptr->cds.cs.vbucketptr);
      tempvectorptr->cds.cs.vbucketptr = NULL;
    }
  } /* matches   else if (tempvectorptr->dstype == CS) {             */
}


/****************************************************************************
 *                                                                          *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/

void caxpy_2_heap_post_processing(destptrptr,
				 rhscvectorptrptrs,
				 alphaptrs,
				 restorerhscvectorptr,
				 arity,
				 lhscvectornonzeroes,
				 lhscvectornonzeroranges,
				 governingdstype)
cvectorstruct **destptrptr;
cvectorstruct **rhscvectorptrptrs[];
complex *alphaptrs[];
cvectorstruct *restorerhscvectorptr[MAXARITY+1];
int arity;
int lhscvectornonzeroes;
int lhscvectornonzeroranges;
int governingdstype;
{
  cvectorstruct *rhscvectorptr[MAXARITY];
  cvectorstruct *tempvectorptr;
  cvectorstruct *sstempvectorptr;
  cbucketstruct *tempbucketptr;
  cbucketstruct *sstempbucketptr;
  cibucketstruct *tempibucketptr;
  cvbucketstruct *tempvbucketptr;
  int k;

  for (k=0;k<arity;k++)
    rhscvectorptr[k] = *rhscvectorptrptrs[k];

  /****************************************
   * Switch lhscvectorptr and *destptrptr. *
   ****************************************/

  lhscvectorptr->nonzeroes = lhscvectornonzeroes;
  
  if (governingdstype == CS)
    lhscvectorptr->cds.cs.nonzeroranges = lhscvectornonzeroranges;
  
  tempvectorptr = (*destptrptr);
  (*destptrptr) = lhscvectorptr;
  lhscvectorptr = tempvectorptr;

  if (governingdstype == SS) {
    
    if ((INITIALIZED_MAXINDEX &&
	 ((float) (*destptrptr)->nonzeroes / (float) GLOBAL_MAXINDEX) >=
	 SWITCH_NZ_RATIO)     ||
	((*destptrptr)->nonzeroes >= SWITCH_NZ_VALUE)) {
     
       /****************************************
       * Convert SS to CS.                    *
       ****************************************/
      
      sstempvectorptr = (*destptrptr);
      cconvertdstype(destptrptr,MAXARITY,CS);
	  
      sstempbucketptr = sstempvectorptr->cds.ss.bucketptr;
      if (sstempbucketptr != NULL) {
	while (sstempbucketptr->nextbucket != NULL)
	  sstempbucketptr = sstempbucketptr->nextbucket;
	sstempbucketptr->nextbucket = freedcbucketslist;
	RESET_freedcbucketslist(sstempvectorptr->cds.ss.bucketptr);
	sstempvectorptr->cds.ss.bucketptr = NULL;
      }
      holdingbincvectorptr[MAXARITY] =  sstempvectorptr;
    }
  }
  
  /****************************************
   * If *destptrptr was not empty,        *
   * add its buckets to freedcbucketslist. *
   ****************************************/
  
  if (tempvectorptr->dstype == SS) {
    tempbucketptr = tempvectorptr->cds.ss.bucketptr;
    if (tempbucketptr != NULL) {
      while (tempbucketptr->nextbucket != NULL)
	tempbucketptr = tempbucketptr->nextbucket;
      tempbucketptr->nextbucket = freedcbucketslist;
      RESET_freedcbucketslist(tempvectorptr->cds.ss.bucketptr);
      tempvectorptr->cds.ss.bucketptr = NULL;
    }
  }
  else if (tempvectorptr->dstype == CS) {
    tempibucketptr = tempvectorptr->cds.cs.ibucketptr;
    if (tempibucketptr != NULL) {
      while (tempibucketptr->nextibucket != NULL)
	tempibucketptr = tempibucketptr->nextibucket;
      tempibucketptr->nextibucket = freedcibucketslist;
      RESET_freedcibucketslist(tempvectorptr->cds.cs.ibucketptr);
      tempvectorptr->cds.cs.ibucketptr = NULL;
    }
    tempvbucketptr = tempvectorptr->cds.cs.vbucketptr;
    if (tempvbucketptr != NULL) {
      while (tempvbucketptr->nextvbucket != NULL)
	tempvbucketptr = tempvbucketptr->nextvbucket;
      tempvbucketptr->nextvbucket = freedcvbucketslist;
      RESET_freedcvbucketslist(tempvectorptr->cds.cs.vbucketptr);
      tempvectorptr->cds.cs.vbucketptr = NULL;
    }
  }
  
  /*****************************************
   * If holdingbins were used, add         *
   * their buckets to freedcbucketslist, OR *
   * their ibuckets to freedcibucketslist & *
   * their vbuckets to freedcvbucketslist.  *
   *****************************************/
  for (k=0;k<arity;k++) {
    
    rhscvectorptr[k] = restorerhscvectorptr[k];
    
    if (holdingbincvectorptr[k]->dstype == SS) {
      tempbucketptr = holdingbincvectorptr[k]->cds.ss.bucketptr;
      if (tempbucketptr != NULL) {
#ifdef _DEBUG
	if (governingdstype == CS) {
	  printf("cspgx warning 7.4: Unexpcted non-NULL");
	  printf("tempbucketptr.\n");
	}
#endif
	while (tempbucketptr->nextbucket != NULL) {
#ifdef _DEBUG
	  if (governingdstype == SS) {
	    printf("cspgx warning 7.1: tempbucketptr should have");
	    printf(" no trailing nextbuckets\n");
	  }
#endif
	  tempbucketptr = tempbucketptr->nextbucket;
	}
	tempbucketptr->nextbucket = freedcbucketslist;
	RESET_freedcbucketslist(holdingbincvectorptr[k]->cds.ss.bucketptr);
	holdingbincvectorptr[k]->cds.ss.bucketptr = NULL;
      }
    }
    else if (holdingbincvectorptr[k]->dstype == CS) {
      tempibucketptr = holdingbincvectorptr[k]->cds.cs.ibucketptr;
      if (tempibucketptr != NULL) {
#ifdef _DEBUG
	if (governingdstype == SS) {
	  printf("cspgx warning 7.2: Unexpcted non-NULL");
	  printf(" ibucketptr.\n");
	}
#endif
	while (tempibucketptr->nextibucket != NULL)
	  tempibucketptr = tempibucketptr->nextibucket;
	tempibucketptr->nextibucket = freedcibucketslist;
	RESET_freedcibucketslist(holdingbincvectorptr[k]->cds.cs.ibucketptr);
	holdingbincvectorptr[k]->cds.cs.ibucketptr = NULL;
      }
      tempvbucketptr = holdingbincvectorptr[k]->cds.cs.vbucketptr;
      if (tempvbucketptr != NULL) {
#ifdef _DEBUG
	if (governingdstype == SS) {
	  printf("cspgx warning 7.3: Unexpcted non-NULL");
	  printf(" vbucketptr.\n");
	}
#endif
	while (tempvbucketptr->nextvbucket != NULL)
	  tempvbucketptr = tempvbucketptr->nextvbucket;
	tempvbucketptr->nextvbucket = freedcvbucketslist;
	RESET_freedcvbucketslist(holdingbincvectorptr[k]->cds.cs.vbucketptr);
	holdingbincvectorptr[k]->cds.cs.vbucketptr = NULL;
      }
    }
  } /* matches    for (k=0;k<arity;k++) {    */
}








/****************************************************************************
 *                                                                          *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/

void cplusequal_axpy_postprocessing(destptrptr,
				   rhscvectorptrptr,
				   status_otherrhscvector,
				   restofplusequalvectorptrptr,
				   alphaptrs,
				   restorerhscvectorptr,
				   arity,
				   lhscvectornonzeroes,
				   lhscvectornonzeroranges,
				   governingdstype)
cvectorstruct **destptrptr;
cvectorstruct **rhscvectorptrptr;
int status_otherrhscvector;
cvectorstruct **restofplusequalvectorptrptr;
complex *alphaptrs[2];
cvectorstruct *restorerhscvectorptr[2+1];
int arity;
int lhscvectornonzeroes;
int lhscvectornonzeroranges;
int governingdstype;
{
  cvectorstruct *rhscvectorptr[2];
  cvectorstruct *tempvectorptr;
  cbucketstruct *tempbucketptr;
  cibucketstruct *tempibucketptr;
  cvbucketstruct *tempvbucketptr;

  int k;

  k = 1;
    rhscvectorptr[k] = *rhscvectorptrptr;

    /****************************************
     * Switch lhscvectorptr and *destptrptr. *
     ****************************************/

    lhscvectorptr->nonzeroes = lhscvectornonzeroes;
    lhscvectorptr->cds.cs.nonzeroranges = lhscvectornonzeroranges;
  

    /*********************************************************
     * If *restofplusequalvectorptrptr was not empty,        *
     * add its buckets to freedcbucketslist.                  *
     *********************************************************/
  
    tempvectorptr = (*restofplusequalvectorptrptr);
    tempibucketptr = tempvectorptr->cds.cs.ibucketptr;
    if (tempibucketptr != NULL) {
      while (tempibucketptr->nextibucket != NULL)
	tempibucketptr = tempibucketptr->nextibucket;
      tempibucketptr->nextibucket = freedcibucketslist;
      RESET_freedcibucketslist(tempvectorptr->cds.cs.ibucketptr);
      tempvectorptr->cds.cs.ibucketptr = NULL;
    }
    tempvbucketptr = tempvectorptr->cds.cs.vbucketptr;
    if (tempvbucketptr != NULL) {
      while (tempvbucketptr->nextvbucket != NULL)
	tempvbucketptr = tempvbucketptr->nextvbucket;
      tempvbucketptr->nextvbucket = freedcvbucketslist;
      RESET_freedcvbucketslist(tempvectorptr->cds.cs.vbucketptr);
      tempvectorptr->cds.cs.vbucketptr = NULL;
    }

   
  /*****************************************
   * If status_otherrhscvector is TEMP,     *
   * add its buckets to freedcbucketslist,  *
   * and reassign pointers.                *
   *****************************************/

  if (status_otherrhscvector == TEMP_OTHER_VECTOR) {

    tempvectorptr = tempcdestptr;
    tempibucketptr = tempvectorptr->cds.cs.ibucketptr;
    if (tempibucketptr != NULL) {
      while (tempibucketptr->nextibucket != NULL)
	tempibucketptr = tempibucketptr->nextibucket;
      tempibucketptr->nextibucket = freedcibucketslist;
      RESET_freedcibucketslist(tempvectorptr->cds.cs.ibucketptr);
      tempvectorptr->cds.cs.ibucketptr = NULL;
    }
    tempvbucketptr = tempvectorptr->cds.cs.vbucketptr;
    if (tempvbucketptr != NULL) {
      while (tempvbucketptr->nextvbucket != NULL)
	tempvbucketptr = tempvbucketptr->nextvbucket;
      tempvbucketptr->nextvbucket = freedcvbucketslist;
      RESET_freedcvbucketslist(tempvectorptr->cds.cs.vbucketptr);
      tempvectorptr->cds.cs.vbucketptr = NULL;
    }

  }
 
  /*****************************************
   * If holdingbins were used, add         *
   * their buckets to freedcbucketslist, OR *
   * their ibuckets to freedcibucketslist & *
   * their vbuckets to freedcvbucketslist.  *
   *****************************************/

  /*  We could not have put rhscvectorptr[0] in a holdingbin;
   *  but there may be something in it from the routine that called
   *  plusequal (in particular cspg2 (cspgx does it's own 
   *  holdingbin cleaning).
   */

  for (k=0;k<arity;k++) {

    rhscvectorptr[k] = restorerhscvectorptr[k];
    
    if (holdingbincvectorptr[k]->dstype == SS) {
      tempbucketptr = holdingbincvectorptr[k]->cds.ss.bucketptr;
      if (tempbucketptr != NULL) {
#ifdef _DEBUG
	if (governingdstype == CS) {
	  printf("cspgx warning 7.4: Unexpcted non-NULL");
	  printf("tempbucketptr.\n");
	}
#endif
	while (tempbucketptr->nextbucket != NULL) {
#ifdef _DEBUG
	  if (governingdstype == SS) {
	    printf("cspgx warning 7.1: tempbucketptr should have");
	    printf(" no trailing nextbuckets\n");
	  }
#endif
	  tempbucketptr = tempbucketptr->nextbucket;
	}
	tempbucketptr->nextbucket = freedcbucketslist;
	RESET_freedcbucketslist(holdingbincvectorptr[k]->cds.ss.bucketptr);
	holdingbincvectorptr[k]->cds.ss.bucketptr = NULL;
      }
    }
    else if (holdingbincvectorptr[k]->dstype == CS) {
      tempibucketptr = holdingbincvectorptr[k]->cds.cs.ibucketptr;
      if (tempibucketptr != NULL) {
#ifdef _DEBUG
	if (governingdstype == SS) {
	  printf("cspgx warning 7.2: Unexpcted non-NULL");
	  printf(" ibucketptr.\n");
	}
#endif
	while (tempibucketptr->nextibucket != NULL)
	  tempibucketptr = tempibucketptr->nextibucket;
	tempibucketptr->nextibucket = freedcibucketslist;
	RESET_freedcibucketslist(holdingbincvectorptr[k]->cds.cs.ibucketptr);
	holdingbincvectorptr[k]->cds.cs.ibucketptr = NULL;
      }
      tempvbucketptr = holdingbincvectorptr[k]->cds.cs.vbucketptr;
      if (tempvbucketptr != NULL) {
#ifdef _DEBUG
	if (governingdstype == SS) {
	  printf("cspgx warning 7.3: Unexpcted non-NULL");
	  printf(" vbucketptr.\n");
	}
#endif
	while (tempvbucketptr->nextvbucket != NULL)
	  tempvbucketptr = tempvbucketptr->nextvbucket;
	tempvbucketptr->nextvbucket = freedcvbucketslist;
	RESET_freedcvbucketslist(holdingbincvectorptr[k]->cds.cs.vbucketptr);
	holdingbincvectorptr[k]->cds.cs.vbucketptr = NULL;
      }
    }
  }
}





/****************************************************************************
 *                                                                          *
 *  cdetermine_loop_bounds                                                   *
 *                                                                          *
 *     Based on how close lhsvbucketindex and rhsvbucketindex pointers are  *
 *     to the end of the vbucket, and based on the number of values left to *
 *     process in the axpy, cdetermine_loop_bounds figures out the various   *
 *     loop values.                                                         *
 *                                                                          *
 ****************************************************************************/


void cdetermine_loop_bounds(totalcurrentnonzerorange,
			   lhsvbucketindex,
			   rhsvbucketindex,
			   firstloopsize,
			   numberofdoubleforloopiterations,
			   firstiterationloopsize,
			   seconditerationloopsize,
			   remainder1loopsize,
			   remainder2loopsize
#ifdef _DEBUG
			   ,addressofunconsumedcurrentnonzerorange
#endif
			   )
int totalcurrentnonzerorange;
int lhsvbucketindex;
int rhsvbucketindex;
int *firstloopsize;
int *numberofdoubleforloopiterations;
int *firstiterationloopsize;
int *seconditerationloopsize;
int *remainder1loopsize;
int *remainder2loopsize;
#ifdef _DEBUG
int *addressofunconsumedcurrentnonzerorange;
#endif
{
  int temprange, tmpmod;

  *firstloopsize =
    min(((cVBUCKET_DIM - max(lhsvbucketindex,rhsvbucketindex))),
	totalcurrentnonzerorange);
  temprange = totalcurrentnonzerorange - *firstloopsize;
  *numberofdoubleforloopiterations = temprange / cVBUCKET_DIM;
  *firstiterationloopsize =
    min((cVBUCKET_DIM -
	 (*firstloopsize + min(lhsvbucketindex,rhsvbucketindex))),temprange);
  *seconditerationloopsize =
    min((cVBUCKET_DIM - *firstiterationloopsize),
	(temprange - *firstiterationloopsize));
  tmpmod = MY_MOD(temprange,cVBUCKET_DIM_minus1);
  *remainder1loopsize = min (tmpmod,*firstiterationloopsize);
  *remainder2loopsize = max ((tmpmod - (*firstiterationloopsize)),0);
  
#ifdef _DEBUG
  *addressofunconsumedcurrentnonzerorange =  totalcurrentnonzerorange;
#endif

}

