#define COMPLEX
/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#include  <stdlib.h>
#include "cgaxpy_pd_macros_and_data_structures.h"
#include "cgaxpy_pd_global_variables.h"


/****************************************************************************
 *                                                                          *
 *  cadallocate()                                                            *
 *                                                                          *
 *      cadallocate calls the bucket and vector store allocation functions.  *
 *      It should be called by the user once, before any calls to cspsd &  *
 *      adxpyNs.                                                            *
 *                                                                          *
 ****************************************************************************/

void cadallocate()
{
  int k;
  
  cinitialize_pd_global_variables();
  callocatevectorstore();
  callocatebucketstore();
  callocateibucketstore();
  callocatevbucketstore();

  cgetvector(&lhscvectorptr);
  for (k=0;k<MAXARITY+1;k++)
    cgetvector(&holdingbincvectorptr[k]);
  cgetvector(&tempcdestptr);
  cgetvector(&restofplusequalcvectorptr);
  restofplusequalcvectorptr->dstype = CS;
}



/****************************************************************************
 *                                                                          *
 *   cadfree()                                                               *
 *                                                                          *
 *      cadfree frees all the dynamically allocated space for a given        *
 *      percision.                                                          *
 *                                                                          *
 ****************************************************************************/

void cadfree()
{
  cvectorstorestruct  *remaindervectorstoreptr,  *tempvectorstoreptr;
  cbucketstorestruct  *remainderbucketstoreptr,  *tempbucketstoreptr;
  cibucketstorestruct *remainderibucketstoreptr, *tempibucketstoreptr;
  cvbucketstorestruct *remaindervbucketstoreptr, *tempvbucketstoreptr;

  cvectorstruct *tempvectorptr;
  cbucketstruct *tempbucketptr;
  cibucketstruct *tempibucketptr;
  cvbucketstruct *tempvbucketptr;

  int i;

  for(remaindervectorstoreptr=cvectorstorestartptr;
      remaindervectorstoreptr!=NULL;
      remaindervectorstoreptr=tempvectorstoreptr) {

    tempvectorstoreptr = remaindervectorstoreptr->nextstore;
    free(remaindervectorstoreptr);
  }

  for(remainderbucketstoreptr=cbucketstorestartptr;
      remainderbucketstoreptr!=NULL;
      remainderbucketstoreptr=tempbucketstoreptr) {
    tempbucketstoreptr = remainderbucketstoreptr->nextstore;

    /*
     * Freeing based on per BUCKET Dynamic Allocation.
     * Here we free the memory which has been allocated per bucket
     * in callocatebucketstore().
     */
    for (i=0; i<cBUCKETSTORE_DIM; i++) {
      tempbucketptr = &(remainderbucketstoreptr->bucket[i]);
      free(tempbucketptr->index);
      free(tempbucketptr->value);
    }
  
    /*
     * Now regardless of the allocation method above, we have freed
     * everything hanging off our bucketstore, and we can go ahead and
     * free the bucketstore itself.
     */
    free(remainderbucketstoreptr);
    
  }
  
  for(remainderibucketstoreptr=cibucketstorestartptr;
      remainderibucketstoreptr!=NULL;
      remainderibucketstoreptr=tempibucketstoreptr) {
    tempibucketstoreptr = remainderibucketstoreptr->nextstore;
    
    /*
     * Freeing based on per IBUCKET Dynamic Allocation.
     */
    for (i=0; i<cIBUCKETSTORE_DIM; i++) {
      tempibucketptr = &(remainderibucketstoreptr->ibucket[i]);
      free(tempibucketptr->index_RANGE_BEGIN);
      free(tempibucketptr->index_RANGE_END);
    }
    
    /*
     * Now regardless of the allocation method above, we have freed
     * everything hanging off our ibucketstore, and we can go ahead and
     * free the ibucketstore itself.
     */
    free(remainderibucketstoreptr);
  }
  
  for(remaindervbucketstoreptr=cvbucketstorestartptr;
      remaindervbucketstoreptr!=NULL;
      remaindervbucketstoreptr=tempvbucketstoreptr) {
    tempvbucketstoreptr = remaindervbucketstoreptr->nextstore;
    
    /*
     * Freeing based on per VBUCKET Dynamic Allocation.
     */
    for (i=0; i<cVBUCKETSTORE_DIM; i++) {
      tempvbucketptr = &(remaindervbucketstoreptr->vbucket[i]);
      free(tempvbucketptr->value);
    }
    
    /*
     * Now regardless of the allocation method above, we have freed
     * everything hanging off our vbucketstore, and we can go ahead and
     * free the vbucketstore itself.
     */
    free(remaindervbucketstoreptr);
  }
}



/****************************************************************************
 *                                                                          *
 *   cad_free_vectors_buckets(cvectorptrptr)                             *
 *                                                                          *
 *       cad_free_vectors_buckets frees the buckets of cvectorptrptr.    *
 *       BUT IT DOES NOT free cvectorptrptr, itself.                    *
 *       Note that freedcbucketslist, freedcibucketslist, and       *
 *       freedcvbucketslist are global variables included in            *
 *       $(SAXPY_FLAVOR)_global_variables.h.                                *
 *                                                                          *
 *       cvectorptrptr  : The vector to be freed.                       *
 *                                                                          *
 ****************************************************************************/

void cad_free_vectors_buckets(cvectorptrptr)
cvectorstruct **cvectorptrptr;
{

  cvectorstruct *cvectorptr;
  cbucketstruct *cbucketptr;
  cibucketstruct *cibucketptr;
  cvbucketstruct *cvbucketptr;

  cvectorptr = (cvectorstruct *) *cvectorptrptr;
  
  if (cvectorptr != NULL) {
    if (cvectorptr->dstype == S1) {
      /* do nothing */
    }
    else if (cvectorptr->dstype == SS) {
      /**********************************************************************
       *                        Single Subscript                            *
       **********************************************************************/
      
      cbucketptr = cvectorptr->cds.ss.bucketptr;
      if (cbucketptr != NULL) {
	while (cbucketptr->nextbucket != NULL)
	  cbucketptr = cbucketptr->nextbucket;
	cbucketptr->nextbucket = freedcbucketslist;
	RESET_freedcbucketslist(cvectorptr->cds.ss.bucketptr);
	cvectorptr->cds.ss.bucketptr = NULL;
      }
      
    } else /* cvectorptr->cdstype = CS */ {
      /**********************************************************************
       *                      Compressed Subscript                          *
       **********************************************************************/
      
      cibucketptr = cvectorptr->cds.cs.ibucketptr;
      if (cibucketptr != NULL) {
	while (cibucketptr->nextibucket != NULL)
	  cibucketptr = cibucketptr->nextibucket;
	cibucketptr->nextibucket = freedcibucketslist;
	RESET_freedcibucketslist(cvectorptr->cds.cs.ibucketptr);
	cvectorptr->cds.cs.ibucketptr = NULL;
      }
      cvbucketptr = cvectorptr->cds.cs.vbucketptr;
      if (cvbucketptr != NULL) {
	while (cvbucketptr->nextvbucket != NULL)
	  cvbucketptr = cvbucketptr->nextvbucket;
	cvbucketptr->nextvbucket = freedcvbucketslist;
	RESET_freedcvbucketslist(cvectorptr->cds.cs.vbucketptr);
	cvectorptr->cds.cs.vbucketptr = NULL;
      }
    }
  }

/*
 * PMK: begin 12/4/96 edits
 *
 * In order to ensure that cvectorptr points to a null vectorstruct,
 * i.e., a vectorstruct whose fields all have zero values, we assign to it
 * the contents of nullcvector which is a global initialized with all
 * zero members in gaxpy_pd_global_variables.h.
 *
 */

  *(cvectorptr) = nullcvector;

/*
 * PMK: end 12/4/96 edits
 */

}


/****************************************************************************
 *                                                                          *
 *    cinitialize_pd_global_variables()                                      *
 *                                                                          *
 *    cinitialize_pd_global_variables performs initializations for some      *
 *    precision dependent globals.                                          *
 *                                                                          *
 ****************************************************************************/

void cinitialize_pd_global_variables()
{
  cbucketstorestartptr=NULL;
  freedcbucketslist=NULL;
  cibucketstorestartptr=NULL;
  freedcibucketslist=NULL;
  cvbucketstorestartptr=NULL;
  freedcvbucketslist=NULL;
  cvectorstorestartptr=NULL;

}

