/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#ifndef cGAXPY_PD_MACROS_AND_DATA_STRUCTURES

#define cGAXPY_PD_MACROS_AND_DATA_STRUCTURES

#include "gaxpy_pi_macros_and_data_structures.h"


/***************************************************************************
 *                                                                         *
 *  Macro definitions for store dimensions                                 *
 *                                                                         *
 ***************************************************************************/

#define cBUCKETSTORE_DIM  128      /* # of buckets in a bucketstore.   */

#define cVBUCKETSTORE_DIM 128      /* # of vbuckets in a vbucketstore. */

#define cIBUCKETSTORE_DIM 128      /* # of ibuckets in a ibucketstore. */

#define cVECTORSTORE_DIM  128      /* # of vectors in a vectorstore.   */


/***************************************************************************
 *                                                                         *
 *  Macro definitions for subroutines;  DEBUG ON                           *
 *                                                                         *
 ***************************************************************************/

#ifdef _DEBUG

#define cHOP_TO_NEXT_RHSVBUCKET(a,b,c,d) \
        chop_to_next_rhsvbucket(a,b,c,d)

#define cHOP_TO_NEXT_LHSVBUCKET(a,b) \
        chop_to_next_lhsvbucket(a,b)

#define cNON_OVER_LAPPING_CS_RANGE_AXPY(a,b,c,d,e,f,g,h,i,j,k,l) \
        cnon_over_lapping_cs_range_axpy(a,b,c,d,e,f,g,h,i,j,k,l)

#define cOVERLAPPING_CS_RANGE_AXPY(a,b,c,d,e,f,g,h,i) \
        coverlapping_cs_range_axpy(a,b,c,d,e,f,g,h,i)

#define cDENSE_PLUSEQUAL_AXPY_LOOP(a,b,c,d,e) \
        cdense_plusequal_axpy_loop(a,b,c,d,e)

#define cNON_OVER_LAPPING_CS_RANGE_CONVERT_TO_c(a,b,c,d,e,f,g,h,i,j,k) \
        cnon_over_lapping_cs_range_convert_to_c(a,b,c,d,e,f,g,h,i,j,k)

#define cOVERLAPPING_CS_RANGE_CONVERT_TO_c(a,b,c,d,e,f,g,h) \
        coverlapping_cs_range_convert_to_c(a,b,c,d,e,f,g,h)

#define cDENSE_CONVERT_TO_c_LOOP(a,b,c,d,e,f) \
        cdense_convert_to_c_loop(a,b,c,d,e,f)

#define cNON_OVER_LAPPING_CS_RANGE_CONVERT_TO_z(a,b,c,d,e,f,g,h,i,j,k) \
        cnon_over_lapping_cs_range_convert_to_z(a,b,c,d,e,f,g,h,i,j,k)

#define cOVERLAPPING_CS_RANGE_CONVERT_TO_z(a,b,c,d,e,f,g,h) \
        coverlapping_cs_range_convert_to_z(a,b,c,d,e,f,g,h)

#define cDENSE_CONVERT_TO_z_LOOP(a,b,c,d,e,f) \
        cdense_convert_to_z_loop(a,b,c,d,e,f)

#define cDETERMINE_LOOP_BOUNDS(a,b,c,d,e,f,g,h,i,j) \
        cdetermine_loop_bounds(a,b,c,d,e,f,g,h,i,j)

#define cDEBUGGING_CODE_1(a,b,c) \
        cdebugging_code_1(a,b,c)

/*
 *  In-lining cdense_axpy_loop
 */

#define cDENSE_AXPY_LOOP(range,lhscvbucketindex,rhscvbucketindex,k,loopsize,alpha,addressofunconsumedcurrentnonzerorange) \
{ \
  int j; \
  if (range == OLD_RANGE) { \
    for (j=0;j<loopsize;j++) { \
      GAXPY_PLUS_EQUALS(lhscvbucketptr->value[(*lhscvbucketindex)+j], \
	      alpha,rhscvbucketptr[k]->value[(*rhscvbucketindex)+j]); \
    } \
  } \
  else { \
    for (j=0;j<loopsize;j++) { \
      GAXPY_EQUALS(lhscvbucketptr->value[(*lhscvbucketindex)+j], \
	      alpha,rhscvbucketptr[k]->value[(*rhscvbucketindex)+j]); \
    } \
  } \
  (*lhscvbucketindex) += loopsize; \
  (*rhscvbucketindex) += loopsize; \
  (*addressofunconsumedcurrentnonzerorange) -= loopsize; \
}


/***************************************************************************
 *                                                                         *
 *  Macro definitions for subroutines;  DEBUG OFF                          *
 *                                                                         *
 ***************************************************************************/

#else

#define cHOP_TO_NEXT_RHSVBUCKET(a,b,c,d) \
        chop_to_next_rhsvbucket(a,b)

#define cHOP_TO_NEXT_LHSVBUCKET(a,b) \
        chop_to_next_lhsvbucket(a)

#define cNON_OVER_LAPPING_CS_RANGE_AXPY(a,b,c,d,e,f,g,h,i,j,k,l) \
        cnon_over_lapping_cs_range_axpy(a,b,c,d,e,f,g,h)

#define cOVERLAPPING_CS_RANGE_AXPY(a,b,c,d,e,f,g,h,i) \
        coverlapping_cs_range_axpy(a,b,c,d,e)

#define cDENSE_PLUSEQUAL_AXPY_LOOP(a,b,c,d,e) \
        cdense_plusequal_axpy_loop(a,b,c,d)

#define cNON_OVER_LAPPING_CS_RANGE_CONVERT_TO_c(a,b,c,d,e,f,g,h,i,j,k) \
        cnon_over_lapping_cs_range_convert_to_c(a,b,c,d,e,f,g)

#define cOVERLAPPING_CS_RANGE_CONVERT_TO_c(a,b,c,d,e,f,g,h) \
        coverlapping_cs_range_convert_to_c(a,b,c,d)

#define cDENSE_CONVERT_TO_c_LOOP(a,b,c,d,e,f) \
        cdense_convert_to_c_loop(a,b,c,d,e)

#define cNON_OVER_LAPPING_CS_RANGE_CONVERT_TO_z(a,b,c,d,e,f,g,h,i,j,k) \
        cnon_over_lapping_cs_range_convert_to_z(a,b,c,d,e,f,g)

#define cOVERLAPPING_CS_RANGE_CONVERT_TO_z(a,b,c,d,e,f,g,h) \
        coverlapping_cs_range_convert_to_z(a,b,c,d)

#define cDENSE_CONVERT_TO_z_LOOP(a,b,c,d,e,f) \
        cdense_convert_to_z_loop(a,b,c,d,e)

#define cDETERMINE_LOOP_BOUNDS(a,b,c,d,e,f,g,h,i,j) \
        cdetermine_loop_bounds(a,b,c,d,e,f,g,h,i)

#define cDEBUGGING_CODE_1(a,b,c)

/*
 *  In-lining cdense_axpy_loop
 */

#define cDENSE_AXPY_LOOP(range,lhscvbucketindex,rhscvbucketindex,k,loopsize,alpha,addressofunconsumedcurrentnonzerorange) \
{ \
  int j; \
  if (range == OLD_RANGE) { \
    for (j=0;j<loopsize;j++) { \
      GAXPY_PLUS_EQUALS(lhscvbucketptr->value[(*lhscvbucketindex)+j], \
	      alpha,rhscvbucketptr[k]->value[(*rhscvbucketindex)+j]); \
    } \
  } \
  else { \
    for (j=0;j<loopsize;j++) { \
      GAXPY_EQUALS(lhscvbucketptr->value[(*lhscvbucketindex)+j], \
	      alpha,rhscvbucketptr[k]->value[(*rhscvbucketindex)+j]); \
    } \
  } \
  (*lhscvbucketindex) += loopsize; \
  (*rhscvbucketindex) += loopsize; \
}

#endif


/***************************************************************************
 *                                                                         *
 *  Macro definitions for subroutines with the same interface for          *
 * _DEBUG.                                                                 *
 *                                                                         *
 ***************************************************************************/

#define RESET_freedcbucketslist(new_head) \
{ freedcbucketslist = new_head; \
}

#define RESET_freedcibucketslist(new_head) \
{ freedcibucketslist = new_head; \
}

#define RESET_freedcvbucketslist(new_head) \
{ freedcvbucketslist = new_head; \
}



/****************************************************************************
 *                                                                          *
 *  Data Structures                                                         *
 *                                                                          *
 *     Two abstract structures need to be represented: sparse vector and    *
 *     heap.  We will first describe these for _SINGLE_SUBSCRIPTS and then  *
 *     point out the differences that exist between these data structures   *
 *     and an analog set of data structures used for _COMPRESSED_SUBSCRIPTS.*
 *                                                                          *
 *     To represent sparse vectors using _SINGLE_SUBSCRIPTS, we define 4    *
 *     structs:                                                             *
 *                                                                          *
 *                  cbucketstruct contains a max # of pairs of indices  *
 *                                    and the corresponding values of the   *
 *                                    of the nonzeroes in a vector, and a   *
 *                                    pointer to the next bucket if there   *
 *                                    are more nonzeroes in the vector. The *
 *                                    indices are always in ascending order *
 *                                    by virtue of the way they are input   *
 *                                    in cspsd and the fact that the       *
 *                                    heapsaxpy operations construct new    *
 *                                    vectors with the same ordering.       *
 *                                                                          *
 *                  cvectorstruct is a sparse representation of a       *
 *                                    vector.  It has a field for the total *
 *                                    # of nonzeroes in the vector and a    *
 *                                    linked list of buckets which contain  *
 *                                    the indices and values of the         *
 *                                    nonzeroes in the vector.              *
 *                                                                          *
 *                  cbucketstorestruct is a long array of buckets, and a*
 *                                    pointer to the next store, if needed. *
 *                                    The basic idea is to allocate space   *
 *                                    in large blocks so that dynamic       *
 *                                    memory allocations will be kept to a  *
 *                                    minimum, yet vectors will be able to  *
 *                                    "grow as they go" by having access to *
 *                                    a memory store of allocated buckets.  *
 *                                                                          *
 *                  cvectorstorestruct is a long array of vectors, and a*
 *                                    pointer to the next store, if needed. *
 *                                    The basic idea is the same as in      *
 *                                    cbucketstorestruct.               *
 *                                                                          *
 *     To represent a heap, we define 1 struct:                             *
 *                                                                          *
 *                  heapstruct        represents a node in the binary heap  *
 *                                    tree.  Each node of the heap is a     *
 *                                    placeholder for a nonzero vector      *
 *                                    element.  The nonzero itself is not   *
 *                                    stored in the heap; instead its       *
 *                                    vector index (order of its vector in  *
 *                                    the rhs of the saxpy),  and its       *
 *                                    subscript index (its array index in   *
 *                                    the non-sparse vector) are placed in  *
 *                                    the heap node.  Additionally, each    *
 *                                    heap node contains two pointer fields *
 *                                    containing addresses of the left and  *
 *                                    right children of the node.           *
 *                                    The subscript index serves as the key *
 *                                    for ordering the heap: if x and p(x)  *
 *                                    are a node and its parent, then       *
 *                                    key_p(x) <= key_x.  It follows that   *
 *                                    key_root is the smallest key in the   *
 *                                    tree at all times, and thus by always *
 *                                    poping the root and "consuming" it in *
 *                                    the saxpy, the lhs vector will be     *
 *                                    constructed in the correct (i.e.,     *
 *                                    ascending) order of subscripts.       *
 *                                                                          *
 *     Each typedef below is accompanied by a pictorial depiction.          *
 *                                                                          *
 *     A Note on C syntax: Several of the data structures are defined with  *
 *     the following template:                                              *
 *                                                                          *
 *          typedef struct _Astruct {                                       *
 *             ...                                                          *
 *             struct _Astruct *nextfield;                                  *
 *          } Astruct;                                                      *
 *                                                                          *
 *     _Astruct is the "structure tag".  It's declared explicitly so that   *
 *     the pointer field inside would recognize what it's pointing to as a  *
 *     defined structure.  Then, typedef defines a synonym, "Astruct", for  *
 *     "struct _Astruct".                                                   *
 *                                                                          *
 ****************************************************************************/


typedef struct _cbucketstruct {

  int *index;                        /* Subscript indices of nonzeroes.     */
  complex *value;           /* Values of nonzeroes.                */

#ifdef _DEBUG                        /* Debugging Flag Field:               */
  int errorflag;                     /*    to ensure integrity of mallocs.  */
#endif

  struct _cbucketstruct *nextbucket;  /* nsia.                          */
} cbucketstruct;

/*             ___________________bucketstruct___________________           *
 *             |                                                |           *
 *             |                   ___index___   ____value___   |           *
 *             |                   |         |   |          |   |           *
 *             |                 0 |  (int)  |   | (double) |   |           *
 *             |                   |_________|   |__________|   |           *
 *             |                   |         |   |          |   |           *
 *             |                 1 |  (int)  |   | (double) |   |           *
 *             |                   |_________|   |__________|   |           *
 *             |                   |    .    |   |     .    |   |           *
 *             |                   |    .    |   |     .    |   |           *
 *             |                   |    .    |   |     .    |   |           *
 *             |                   |_________|   |__________|   |           *
 *             |                   |         |   |          |   |           *
 *             | cBUCKET_DIM-1 |  (int)  |   | (double) |   |           *
 *             |                   |_________|   |__________|   |           *
 *             |                                                |           *
 *             |             nextbucket *                       |           *
 *             |                        |                       |           *
 *             |________________________|_______________________|           *
 *                                      |                                   *
 *                                      V                                   *
 *                               (cbucketstruct)                        */




typedef struct _cibucketstruct {
  int *index_RANGE_BEGIN;
  int *index_RANGE_END;
                                     /* Subscript index pairs of range of   *
				      * nonzeroes.  Note: since in C, 2-d   *
				      * array elements are stored by row,   *
				      * index_RANGE_BEGIN and               *
				      * index_RANGE_END,                    *
				      * represent a range and are stored    *
				      * contiguously, which is advantageous *
				      * from a cache memory point of view,  *
				      * since they are also accessed in     *
				      * in tandem.                          */

#ifdef _DEBUG                        /* Debugging Flag Field:               */
  int errorflag;                     /*    to ensure integrity of mallocs.  */
#endif

  struct _cibucketstruct *nextibucket;  /* nsia.                        */
} cibucketstruct;

/*     __________________________ibucketstruct__________________________    *
 *     |                                                               |    *
 *     |                       ______index______                       |    *
 *     |                       |       |       |                       |    *
 *     |                     0 | (int) | (int) |                       |    *
 *     |                       |_______|_______|                       |    *
 *     |                       |       |       |                       |    *
 *     |                     1 | (int) | (int) |                       |    *
 *     |                       |_______|_______|                       |    *
 *     |                       |   .   |   .   |                       |    *
 *     |                       |   .   |   .   |                       |    *
 *     |                       |   .   |   .   |                       |    *
 *     |                       |_______|_______|                       |    *
 *     |                       |       |       |                       |    *
 *     |    cIBUCKET_DIM-1 | (int) | (int) |                       |    *
 *     |                       |_______|_______|                       |    *
 *     |                                                               |    *
 *     |                   nextibucket *                               |    *
 *     |                               |                               |    *
 *     |_______________________________|_______________________________|    *
 *                                     |                                    *
 *                                     V                                    *
 *                              (cibucketstruct)                        */



typedef struct _cvbucketstruct {

  complex *value;           /* Values of nonzeroes.                */

#ifdef _DEBUG                        /* Debugging Flag Field:               */
  int errorflag;                     /*    to ensure integrity of mallocs.  */
#endif

  struct _cvbucketstruct *nextvbucket; /* nsia.                         */
} cvbucketstruct;

/*                   _____________vbucketstruct_____________                *
 *                   |                                     |                *
 *                   |                    ____value___     |                *
 *                   |                    |          |     |                *
 *                   |                  0 | (double) |     |                *
 *                   |                    |__________|     |                *
 *                   |                    |          |     |                *
 *                   |                  1 | (double) |     |                *
 *                   |                    |__________|     |                *
 *                   |                    |     .    |     |                *
 *                   |                    |     .    |     |                *
 *                   |                    |     .    |     |                *
 *                   |                    |__________|     |                *
 *                   |                    |          |     |                *
 *                   | cVBUCKET_DIM-1 | (double) |     |                *
 *                   |                    |__________|     |                *
 *                   |                                     |                *
 *                   |      nextvbucket *                  |                *
 *                   |                  |                  |                *
 *                   |__________________|__________________|                *
 *                                      |                                   *
 *                                      V                                   *
 *                               (cvbucketstruct)                       */


/*
 * Sparse Vector Internal Data Structure (DS) Types: S1, SS, CS
 */

typedef struct _cS1struct {
  int index;

  complex value;

} cS1struct;

typedef struct _cSSstruct {
  cbucketstruct *bucketptr;      /* Pointer to linked list of buckets.  */
} cSSstruct;

typedef struct _cCSstruct {
  int nonzeroranges;                 /* Number of contiguous nonzero ranges.*/
  cibucketstruct *ibucketptr;    /* Pointer to linked list of ibuckets. */
  cvbucketstruct *vbucketptr;    /* Pointer to linked list of vbuckets. */
} cCSstruct;


typedef struct {
  int nonzeroes;                    /* The total # of                       *
				     * nonzeroes in the vector.             */
  int dstype;                       /* The data-structure type of the       *
				     * cvectorstruct will be one of the *
				     * following:
				     *          S1 = 1 SUBSCRIPT            *
				     *          SS = SINGLE SUBSCRIPTS      *
				     *          CS = COMPRESSED SUBSCRIPTS  */
  union _cds {
    cS1struct s1;
    cSSstruct ss;
    cCSstruct cs;
  } cds;

#ifdef _VTRACE

  /* Symbolic identifier */
  char* id;
  /* Numeric identifier */
  int num_id;

  /* Tracked? */
  unsigned char is_tracked;
  /* Named? */
  unsigned char is_named;
  /* Visualized? */
  unsigned char is_visual;

  /* List of all updates over the vector's lifetime */
  pu_struct u_list;

#endif

} cvectorstruct;

/*     ___________________________vectorstruct___________________________   *
 *     |                                                                |   *
 *     |          ____nonzeroes____         _____dstype______           |   *
 *     |          |               |         |               |           |   *
 *     |          |     (int)     |         |     (int)     |           |   *
 *     |          |_______________|         |_______________|           |   *
 *     |                                                                |   *
 *     |                               ds                               |   *
 *     |    {~~~~~~~~~~~~~~~~~~~~~~~~~~--~~~~~~~~~~~~~~~~~~~~~~~~~~}    |   *
 *     |    {                                                      }    |   *
 *     |    { EITHER:                                              }    |   *
 *     |    {     _____________________S1_____________________     }    |   *
 *     |    {     |   _____index_____     _____value_____    |     }    |   *
 *     |    {     |   |             |     |             |    |     }    |   *
 *     |    {     |   |    (int)    |     |    (int)    |    |     }    |   *
 *     |    {     |   |_____________|     |_____________|    |     }    |   *
 *     |    {     |__________________________________________|     }    |   *
 *     |    {                                                      }    |   *
 *     |    { OR:                                                  }    |   *
 *     |    {     _____________________SS_____________________     }    |   *
 *     |    {     |                                          |     }    |   *
 *     |    {     |           bucketptr *                    |     }    |   *
 *     |    {     |                     |                    |     }    |   *
 *     |    {     |_____________________|____________________|     }    |   *
 *     |    {                           |                          }    |   *
 *     |    {                           V                          }    |   *
 *     |    {                     (cbucketstruct)              }    |   *
 *     |    {                                                      }    |   *
 *     |    { OR:                                                  }    |   *
 *     |    {     _____________________CS_____________________     }    |   *
 *     |    {     |                                          |     }    |   *
 *     |    {     |              _nonzeroranges_             |     }    |   *
 *     |    {     |              |             |             |     }    |   *
 *     |    {     |              |    (int)    |             |     }    |   *
 *     |    {     |              |_____________|             |     }    |   *
 *     |    {     |                                          |     }    |   *
 *     |    {     |     ibucketptr *        vbucketptr *     |     }    |   *
 *     |    {     |                |                   |     |     }    |   *
 *     |    {     |________________|___________________|_____|     }    |   *
 *     |    {                      |                   |           }    |   *
 *     |    {                      V                   V           }    |   *
 *     |    {          (cibucketstruct)  (cvbucketstruct)  }    |   *
 *     |    {                                                      }    |   *
 *     |    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~    |   *
 *     |                                                                |   *
 *     |________________________________________________________________|   */




typedef struct _cvectorstorestruct {
  cvectorstruct vector[cVECTORSTORE_DIM];  /* Long array of vectors */
  struct _cvectorstorestruct *nextstore;       /* nsia.                 */
} cvectorstorestruct;

/*             ________________vectorstorestruct________________            *
 *             |                                               |            *
 *             |                        ______vector______     |            *
 *             |                        |                |     |            *
 *             |                      0 | (cvectorstruct) ||            *
 *             |                        |________________|     |            *
 *             |                        |                |     |            *
 *             |                      1 | (cvectorstruct) ||            *
 *             |                        |________________|     |            *
 *             |                        |        .       |     |            *
 *             |                        |        .       |     |            *
 *             |                        |        .       |     |            *
 *             |                        |________________|     |            *
 *             |                        |                |     |            *
 *             | cVECTORSTORE_DIM-1 | (cvectorstruct) ||            *
 *             |                        |________________|     |            *
 *             |                                               |            *
 *             |             nextstore *                       |            *
 *             |                       |                       |            *
 *             |_______________________|_______________________|            *
 *                                     |                                    *
 *                                     V                                    *
 *                            (cvectorstorestruct)                      */






typedef struct _cbucketstorestruct {
  cbucketstruct bucket[cBUCKETSTORE_DIM];  /* Long array of buckets.*/
  struct _cbucketstorestruct *nextstore;       /* nsia.                 */
} cbucketstorestruct;

/*             ________________bucketstorestruct________________            *
 *             |                                               |            *
 *             |                        ______bucket______     |            *
 *             |                        |                |     |            *
 *             |                      0 | (cbucketstruct) ||            *
 *             |                        |________________|     |            *
 *             |                        |                |     |            *
 *             |                      1 | (cbucketstruct) ||            *
 *             |                        |________________|     |            *
 *             |                        |        .       |     |            *
 *             |                        |        .       |     |            *
 *             |                        |        .       |     |            *
 *             |                        |________________|     |            *
 *             |                        |                |     |            *
 *             | cBUCKETSTORE_DIM-1 | (cbucketstruct) ||            *
 *             |                        |________________|     |            *
 *             |                                               |            *
 *             |             nextstore *                       |            *
 *             |                       |                       |            *
 *             |_______________________|_______________________|            *
 *                                     |                                    *
 *                                     V                                    *
 *                            (cbucketstorestruct)                      */


typedef struct _cibucketstorestruct {
  cibucketstruct ibucket[cIBUCKETSTORE_DIM];
                                                 /* Long array of ibuckets. */
  struct _cibucketstorestruct *nextstore;    /* nsia.                   */
} cibucketstorestruct;

/*             ________________ibucketstorestruct________________           *
 *             |                                               |            *
 *             |                        ______bucket______     |            *
 *             |                        |                |     |            *
 *             |                      0 |(cibucketstruct) ||            *
 *             |                        |________________|     |            *
 *             |                        |                |     |            *
 *             |                      1 |(cibucketstruct) ||            *
 *             |                        |________________|     |            *
 *             |                        |        .       |     |            *
 *             |                        |        .       |     |            *
 *             |                        |        .       |     |            *
 *             |                        |________________|     |            *
 *             |                        |                |     |            *
 *             | cIBUCKETSTORE_DIM-1| (cibucketstruct)||            *
 *             |                        |________________|     |            *
 *             |                                               |            *
 *             |             nextstore *                       |            *
 *             |                       |                       |            *
 *             |_______________________|_______________________|            *
 *                                     |                                    *
 *                                     V                                    *
 *                            (cibucketstorestruct)                     */



typedef struct _cvbucketstorestruct {
  cvbucketstruct vbucket[cVBUCKETSTORE_DIM];
                                                 /* Long array of vbuckets. */
  struct _cvbucketstorestruct *nextstore;    /* nsia.                   */
} cvbucketstorestruct;

/*             ________________vbucketstorestruct_________________          *
 *             |                                                 |          *
 *             |                         ______vbucket______     |          *
 *             |                         |                 |     |          *
 *             |                       0 | (cvbucketstruct) ||          *
 *             |                         |_________________|     |          *
 *             |                         |                 |     |          *
 *             |                       1 | (cvbucketstruct) ||          *
 *             |                         |_________________|     |          *
 *             |                         |         .       |     |          *
 *             |                         |         .       |     |          *
 *             |                         |         .       |     |          *
 *             |                         |_________________|     |          *
 *             |                         |                 |     |          *
 *             | cVBUCKETSTORE_DIM-1 | (cvbucketstruct) ||          *
 *             |                         |_________________|     |          *
 *             |                                                 |          *
 *             |              nextstore *                        |          *
 *             |                        |                        |          *
 *             |________________________|________________________|          *
 *                                      |                                   *
 *                                      V                                   *
 *                            (cvbucketstorestruct)                     */





/****************************************************************************
 *                                                                          *
 *  Function Prototyping                                                    *
 *                                                                          *
 ****************************************************************************/

#define _P(x) ()

void callocatebucketstore _P(());
void callocateibucketstore _P(());
void callocatevbucketstore _P(());
void callocatevectorstore _P(());

void cgetbucket _P((cbucketstruct **bucketptrptr));
void cgetibucket _P((cibucketstruct **ibucketptrptr));
void cgetvbucket _P((cvbucketstruct **vbucketptrptr));
void cgetvector _P((cvectorstruct **vectorptrptr));

void cspgx _P((cvectorstruct **destptrptr,
		   int arity,
		   complex *alphaptrs[],
		   cvectorstruct **rhsvectorptrptrs[]));

void cnon_over_lapping_cs_range_axpy _P((int rangedestination,
				      int lhscibucketindex,
				      int *addressoflastlhscibucketindex,
				      int *addressoflhscvbucketindex,
				      int k,
				      int *addressofrhscvbucketindex_k,
				      complex alpha_k,
				      int totalcurrentnonzerorange
#ifdef _DEBUG
				      ,int *addressofunconsumedcurrentnonzerorange
				      ,int rhscvectorconsumednonzeroranges_k
				      ,int rhscvectornonzeroranges_k
				      ,int lhscvectornonzeroranges
#endif
				      ));

void coverlapping_cs_range_axpy _P((int *addressoflhscvbucketindex,
				   int k,
				   int *addressofrhscvbucketindex_k,
				   complex alpha_k,
				   int totalcurrentnonzerorange
#ifdef _DEBUG
				   ,int *addressofunconsumedcurrentnonzerorange
				   ,int rhscvectorconsumednonzeroranges_k
				   ,int rhscvectornonzeroranges_k
				   ,int lhscvectornonzeroranges
#endif
				   ));


void cdetermine_loop_bounds _P((int totalcurrentnonzerorange,
			       int lhsvbucketindex,
			       int rhsvbucketindex,
			       int *firstloopsize,
			       int *numberofdoubleforloopiterations,
			       int *firstiterationloopsize,
			       int *seconditerationloopsize,
			       int *remainder1loopsize,
			       int *remainder2loopsize
#ifdef _DEBUG
			       ,int *addressofunconsumedcurrentnonzerorange
#endif
			       ));

void cdebugging_code_1 _P((char *RIGHT_OR_LEFT,
			  int vbucketindex,
			  int unconsumedcurrentnonzerorange));

void chop_to_next_rhsvbucket _P((int k,
				int *vbucketindex
#ifdef _DEBUG
				,int rhscvectorconsumednonzeroranges
				,int rhscvectornonzeroranges
#endif
				));


void chop_to_next_lhsvbucket _P((int *vbucketindex
#ifdef _DEBUG
				,int lhscvectornonzeroranges
#endif
				));

void cget_and_hop_to_next_vbucket _P((int *lhscvbucketindex));

void cconvertdstype _P((cvectorstruct **oldvectorptrptr,
		       int k,
		       int newvectordstype));

void caxpy_0_1_post_processing _P((cvectorstruct **destptrptr));

void caxpy_2_heap_post_processing _P((cvectorstruct **destptrptr,
                                     cvectorstruct **rhscvectorptrptrs[],
                                     complex *alphaptrs[],
                                     cvectorstruct *restorerhscvectorptr[MAXARITY+1],
                                     int arity,
                                     int lhscvectornonzeroes,
                                     int lhscvectornonzeroranges,
                                     int governingdstype));

void cplusequal_axpy_postprocessing _P((cvectorstruct **destptrptr,
				       cvectorstruct **rhscvectorptrptr,
				       int status_otherrhscvector,
				       cvectorstruct **restofplusequalvectorptrptr,
				       complex *alphaptrs[2],
				       cvectorstruct *restorerhscvectorptr[2+1],
				       int arity,
				       int lhscvectornonzeroes,
				       int lhscvectornonzeroranges,
				       int governingdstype
				       ));

#ifdef _PEEP
void cadpeep_pre _P((int vectorindex,
		    cvectorstruct **rhsvectorptrptr,
		    complex *alphaptr));

void cadpeep_post _P((cvectorstruct **destptrptr));
#endif

void cadallocate _P(());
void cadfree _P(());
void cad_free_vectors_buckets _P((cvectorstruct **vectorptrptr));
void cinitialize_pd_global_variables _P(());

void cspsd _P((cvectorstruct **destptrptr,
		int *indexvector,
		complex *valuevector,
		int *lengthptr));

void cspxd _P((complex *vector,
			 int *lengthptr,
			 cvectorstruct **objectptrptr,
			 int *returned_len,
			 int *len_info));

void cspxm _P((complex *vector,
			       int *lengthptr,
			       complex *multiplier,
			       cvectorstruct **objectptrptr,
			       int *returned_len,
			       int *len_info));

void cspxa _P((complex *vector,
				   int *lengthptr,
				   cvectorstruct **objectptrptr,
				   int *returned_len,
				   int *len_info));

void cspxs _P((int *indvector,
			  complex *valvector,
			  int *lengthptr,
			  cvectorstruct **objectptrptr,
			  int *returned_len,
			  int *len_info)); 

void csppr _P((cvectorstruct **objectptrptr, int *unit_number));

void cspzronotca _P((cvectorstruct **destptrptr));

void cspcp _P((cvectorstruct **destptrptr,
		cvectorstruct **oldvectorptrptr));

void cadinfo _P((int arity,
		int nnzb,
		cvectorstruct *lhsvectoraddress,
		int nnza,
		complex *alphaptrs[],
		cvectorstruct **vectorptrptrs[]));

void cspg1 _P((cvectorstruct **destptrptr,
		complex *alphaptr,
		cvectorstruct **vectoraptrptr));
void cspg2 _P((cvectorstruct **destptrptr,
		complex *alphaptr,
		cvectorstruct **vectoraptrptr,
		complex *betaptr,
		cvectorstruct **vectorbptrptr));
void plusequalcspg1 _P((cvectorstruct **rhsanddestptrptr,
			 complex *alphaptr));
void plusequalcspg2 _P((cvectorstruct **rhsanddestptrptr,
			 complex *otheralphaptr,
			 cvectorstruct **otherrhsvectorptrptr,
			 int status_otherrhsvector));



#endif /*  #ifndef cGAXPY_PD_MACROS_AND_DATA_STRUCTURES  */
