#define COMPLEX
/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#include  <stdlib.h>
#include "cgaxpy_pd_macros_and_data_structures.h"
#include "cgaxpy_pd_global_variables.h"


/****************************************************************************
 *                                                                          *
 *  callocatebucketstore()                                                   *
 *                                                                          *
 *     1 malloc call gets all the memory for a bucketstore.  If malloc does *
 *     not return NULL, one of the following 2 scenarios will result:       *
 *                                                                          *
 *     A: outcome of callocatebucketstore the 1st time it's called:          *
 *                                                                          *
 *     bucketstoreptr -----------+                                          *
 *                               |   store                                  *
 *                              _V_____1_____                               *
 *     bucketstorestartptr ---->|    _____  |                               *
 *     bucketstorecounter ------+->0 |___|  |                               *
 *                              |    | . |  |                               *
 *                              |    | . |  |                               *
 *                              |    | . |  |                               *
 *                              |    |___|  |                               *
 *                              |    |___|  |                               *
 *                              |           |                               *
 *                              | nextstore |                               *
 *                              |     *     |                               *
 *                              |_____|_____|                               *
 *                                    |                                     *
 *                                    V                                     *
 *                                   NULL                                   *
 *                                                                          *
 *     B: outcome of callocatebucketstore the Nth time it's called (N > 1):  *
 *                                                                          *
 *     bucketstorecounter --------------------------------------+           *
 *                                                              |           *
 *     bucketstoreptr ----------------------------------------+ |           *
 *                               store        store           | | store     *
 *                              ___1___      ___2___ ...     _V_|___N_____  *
 *     bucketstorestartptr ---->|     |  +-->|     |     +-->|  V _____  |  *
 *                              |     |  |   |     |     |   |  0 |___|  |  *
 *                              |     |  |   |     |     |   |    | . |  |  *
 *                              |     |  |   |     |     |   |    | . |  |  *
 *                              |     |  |   |     |     |   |    | . |  |  *
 *                              |     |  |   |     |     |   |    |___|  |  *
 *                              |     |  |   |     |     |   |    |___|  |  *
 *                              |     |  |   |     |     |   |           |  *
 *                              |     |  |   |     |     |   | nextstore |  *
 *                              |  *  |  |   |  *  |     |   |     *     |  *
 *                              |__|__|  |   |__|__|     |   |_____|_____|  *
 *                                 |     |      |        |         |        *
 *                                 +-----+      +---...--+         V        *
 *                                                                NULL      *
 *                                                                          *
 ****************************************************************************/

void callocatebucketstore()
{
  int i,k;
  
  cbucketstorestruct *tempbucketstoreptr;
  cbucketstruct *tempbucketptr;
  
  int *index_store;
  complex *value_store;

  int temp_size;

  /*
   * Per BUCKET Dynamic Allocation: constant cBUCKETSTORE_DIM
   *                                variable cBUCKET_DIM
   */
  
  if ((tempbucketstoreptr = (cbucketstorestruct *)
       palloc ((temp_size = (sizeof(cbucketstorestruct)))))==NULL) {
    printf("callocatebucketstore error: malloc returned NULL.\n");
    abort();
  }
  memory_count(C_EVAL,MEM_BUCKET,temp_size);
  
  for (k=0; k<cBUCKETSTORE_DIM; k++) {
    tempbucketptr = &(tempbucketstoreptr->bucket[k]);
    if ((tempbucketptr->index = (int *)
	 palloc ((temp_size = (cBUCKET_DIM * sizeof(int)))))==NULL) {
      printf("callocatebucketstore error: malloc returned NULL.\n");
      abort();
    }
    memory_count(C_EVAL,MEM_BUCKET,temp_size);
    if ((tempbucketptr->value = (complex *)
	 palloc ((temp_size = (cBUCKET_DIM *
			       sizeof(complex)))))==NULL) {
      printf("callocatebucketstore error: malloc returned NULL.\n");
      abort();
    }
    memory_count(C_EVAL,MEM_BUCKET,temp_size);
  }
      
  if (cbucketstorestartptr == NULL)
    cbucketstorestartptr = tempbucketstoreptr;
  else
    cbucketstoreptr->nextstore = tempbucketstoreptr;
  
  cbucketstoreptr = tempbucketstoreptr;
  cbucketstoreptr->nextstore = NULL;
  cbucketstorecounter = 0;
  
  for (i=0;i<cBUCKETSTORE_DIM;i++) {
    (&(cbucketstoreptr->bucket[i]))->nextbucket = NULL;
#ifdef _DEBUG  
    (&(cbucketstoreptr->bucket[i]))->errorflag = 9999;
#endif
  }
}



/****************************************************************************
 *                                                                          *
 *  callocateibucketstore()                                                  *
 *                                                                          *
 *     Identical to callocatebucketstore in terms of control flow.           *
 *                                                                          *
 ****************************************************************************/

void callocateibucketstore()
{
  int i,k;

  cibucketstorestruct *tempibucketstoreptr;
  cibucketstruct *tempibucketptr;
  
  int *index_RANGE_BEGIN_store;
  int *index_RANGE_END_store;

  int temp_size;
 
  /*
   * Per IBUCKET Dynamic Allocation: constant cIBUCKETSTORE_DIM
   *                                 variable cIBUCKET_DIM
   */
  
  if ((tempibucketstoreptr = (cibucketstorestruct *)
       palloc ((temp_size = (sizeof(cibucketstorestruct)))))==NULL) {
    printf("callocateibucketstore error: malloc returned NULL.\n");
    abort();
  }
  memory_count(C_EVAL,MEM_IBUCKET,temp_size);
  
  for (k=0; k<cIBUCKETSTORE_DIM; k++) {
    tempibucketptr = &(tempibucketstoreptr->ibucket[k]);
    if ((tempibucketptr->index_RANGE_BEGIN = (int *)
	 palloc ((temp_size = (cIBUCKET_DIM * sizeof(int)))))==NULL) {
      printf("callocateibucketstore error: malloc returned NULL.\n");
      abort();
    }
    memory_count(C_EVAL,MEM_IBUCKET,temp_size);
    if ((tempibucketptr->index_RANGE_END = (int *)
	 palloc ((temp_size = (cIBUCKET_DIM * sizeof(int)))))==NULL) {
      printf("callocateibucketstore error: malloc returned NULL.\n");
      abort();
    }
    memory_count(C_EVAL,MEM_IBUCKET,temp_size);
  }
  
  if (cibucketstorestartptr == NULL)
    cibucketstorestartptr = tempibucketstoreptr;
  else
    cibucketstoreptr->nextstore = tempibucketstoreptr;
  
  cibucketstoreptr = tempibucketstoreptr;
  cibucketstoreptr->nextstore = NULL;
  cibucketstorecounter = 0;
  
  for (i=0;i<cIBUCKETSTORE_DIM;i++) {
    (&(cibucketstoreptr->ibucket[i]))->nextibucket = NULL;
#ifdef _DEBUG  
    (&(cibucketstoreptr->ibucket[i]))->errorflag = 9099;
#endif
  }
}




/****************************************************************************
 *                                                                          *
 *  callocatevbucketstore()                                                  *
 *                                                                          *
 *     Identical to callocatebucketstore in terms of control flow.           *
 *                                                                          *
 ****************************************************************************/

void callocatevbucketstore()
{
  int i,k;

  cvbucketstorestruct *tempvbucketstoreptr;
  cvbucketstruct *tempvbucketptr;
  
  complex *value_store;

  int temp_size;
  
  /*
   * Per VBUCKET Dynamic Allocation: constant cVBUCKETSTORE_DIM
   *                                 variable cVBUCKET_DIM
   */
  
  if ((tempvbucketstoreptr = (cvbucketstorestruct *)
       palloc ((temp_size = (sizeof(cvbucketstorestruct)))))==NULL) {
    printf("callocatevbucketstore error: malloc returned NULL.\n");
    abort();
  }
  memory_count(C_EVAL,MEM_VBUCKET,temp_size);
  
  for (k=0; k<cVBUCKETSTORE_DIM; k++) {
    tempvbucketptr = &(tempvbucketstoreptr->vbucket[k]);
    if ((tempvbucketptr->value = (complex *)
	 palloc ((temp_size = (cVBUCKET_DIM *
			       sizeof(complex)))))==NULL) {
      printf("callocatevbucketstore error: malloc returned NULL.\n");
      abort();
    }
    memory_count(C_EVAL,MEM_VBUCKET,temp_size);
  }
  
  if (cvbucketstorestartptr == NULL)
    cvbucketstorestartptr = tempvbucketstoreptr;
  else
    cvbucketstoreptr->nextstore = tempvbucketstoreptr;
  
  cvbucketstoreptr = tempvbucketstoreptr;
  cvbucketstoreptr->nextstore = NULL;
  cvbucketstorecounter = 0;
  
  for (i=0;i<cVBUCKETSTORE_DIM;i++) {
    (&(cvbucketstoreptr->vbucket[i]))->nextvbucket = NULL;
#ifdef _DEBUG  
    (&(cvbucketstoreptr->vbucket[i]))->errorflag = 9009;
#endif
  }
}




/****************************************************************************
 *                                                                          *
 *  callocatevectorstore()                                                   *
 *                                                                          *
 *     cVECTORSTORE_DIM is a constant and in contrast to the buckets    *
 *     there is no VECTOR_DIM.  Therefore there is no need for any          *
 *     allocation scheme requiring the per STORE vs per VECTOR methods.     *
 *                                                                          *
 ****************************************************************************/

void callocatevectorstore()
{
  int i;
  cvectorstorestruct *tempvectorstoreptr;
  cvectorstruct *tempvectorptr;

  int temp_size;

  /*
   * Wholesale Dynamic Allocation: constant cVECTORSTORE_DIM
   * This allocation is not obsolete unlike the bucket type stores
   * above, since there is no variable-sized bucket in a vectorstruct.
   */
  if ((tempvectorstoreptr = (cvectorstorestruct *)
       palloc ((temp_size = (sizeof(cvectorstorestruct)))))==NULL ) {
    printf("callocatevectorstore error: malloc returned NULL.\n");
    abort();
  }
  memory_count(C_EVAL,MEM_VECTOR,temp_size);
  
  /*
   * End of Wholesale Dynamic Allocation
   */
  
  if (cvectorstorestartptr == NULL)
    cvectorstorestartptr = tempvectorstoreptr;
  else
    cvectorstoreptr->nextstore = tempvectorstoreptr;
  
  cvectorstoreptr = tempvectorstoreptr;
  cvectorstoreptr->nextstore = NULL;
  cvectorstorecounter = 0;
  for (i=0;i<cVECTORSTORE_DIM;i++) {
    tempvectorptr = &(cvectorstoreptr->vector[i]);

/*
 *  PMK: 12/4/96 edits
 *
 *  The assignment of the S0 "non-yet-specified" data structure type is
 *  extraneous for now, since we never have checks on S0 (i.e., nowhere
 *  in the code do we say, if (somevectorptr->dstype == S0) ...
 *  Therefore, the following code line is commented out:

    tempvectorptr->dstype = S0;

 *
 */

  }
}



/****************************************************************************
 *                                                                          *
 *  cgetbucket(bucketptrptr)                                                 *
 *                                                                          *
 *     bucketptrptr    is a doubly-indirected pointer to a cbucketstruct*
 *                     Since we want the value of "bucketptr" to be changed *
 *                     by cgetbucket (i.e., *bucketptrptr to get assigned a  *
 *                     pointer to a cbucketstruct), we need to make its *
 *                     address (bucketptrptr) the function argument.  Thus, *
 *                     the need for the double indirection (**).            *
 *                                                                          *
 *     cgetbucket assigns to *bucketptrptr the pointer to the next available *
 *     cbucketstruct from the freedbucketslist or if none there from the*
 *     bucketstore.  In the latter case, if we're at the end of the current *
 *     bucketstore, a new store is allocated before doing the "get".        *
 *                                                                          *
 ****************************************************************************/

void cgetbucket(bucketptrptr)
cbucketstruct **bucketptrptr;
{
  if (freedcbucketslist != NULL) {
    (*bucketptrptr) = freedcbucketslist;
    freedcbucketslist = freedcbucketslist->nextbucket;
  }
  else {
    if (cbucketstorecounter == cBUCKETSTORE_DIM) {
      callocatebucketstore();
    }
    
#ifdef _DEBUG
    else if (cbucketstorecounter > cBUCKETSTORE_DIM) {
      printf("cgetbucket error: cbucketstorecounter=%d,  STOREDIMENSION=%d\n",
	     cbucketstorecounter,cBUCKETSTORE_DIM);
      abort();
    }
#endif

    /*
     *  The following assignment of a store entry to *bucketptrptr
     *  works regardless of whether we've used per BUCKET or per STORE
     *  allocation method.
     */

    *bucketptrptr = (cbucketstruct *)
                     &(cbucketstoreptr->bucket[cbucketstorecounter++]);
  }
  (*bucketptrptr)->nextbucket = NULL;
  
#ifdef _DEBUG
  if ((*bucketptrptr)->errorflag != 9999) {
    printf("cgetbucket error: memory corruption of cbucketstore\n");
    abort();
  }
#endif
  
}

/****************************************************************************
 *                                                                          *
 *  cgetibucket(ibucketptrptr)                                               *
 *                                                                          *
 *     ibucketptrptr   is a doubly-indirected pointer to a cibucketstruc*
 *                     Since we want the value of "ibucketptr" to be        *
 *                     changed by cgetibucket (i.e., *ibucketptrptr to get   *
 *                     assigned a  pointer to a cibucketstruct), we need*
 *                     to make its address (ibucketptrptr) the function     *
 *                     argument.  Thus, the need for the double indirection *
 *                     (**).                                                *
 *                                                                          *
 *     cgetibucket assigns to *ibucketptrptr the pointer to the next         *
 *     available cibucketstruct from the freedibucketslist or if none   *
 *     from the ibucketstore.  In the latter case, if we're at the end      *
 *     of the current ibucketstore, a new store is allocated before doing   *
 *     the "get".                                                           *
 *                                                                          *
 ****************************************************************************/

void cgetibucket(ibucketptrptr)
cibucketstruct **ibucketptrptr;
{
  if (freedcibucketslist != NULL) {
    (*ibucketptrptr) = freedcibucketslist;
    freedcibucketslist = freedcibucketslist->nextibucket;
  }
  else {
    if (cibucketstorecounter == cIBUCKETSTORE_DIM) {
      callocateibucketstore();
    }
    
#ifdef _DEBUG
    else if (cibucketstorecounter > cIBUCKETSTORE_DIM) {
      printf("cgetibucket error: cibucketstorecounter=%d,  STOREDIMENSION=%d\n",
	     cibucketstorecounter,cIBUCKETSTORE_DIM);
      abort();
    }
#endif

    *ibucketptrptr = (cibucketstruct *)
              &(cibucketstoreptr->ibucket[cibucketstorecounter++]);
  }
  (*ibucketptrptr)->nextibucket = NULL;
  
#ifdef _DEBUG
  if ((*ibucketptrptr)->errorflag != 9099) {
    printf("cgetibucket error: memory corruption of cibucketstore\n");
    abort();
  }
#endif
  
}


/****************************************************************************
 *                                                                          *
 *  cgetvbucket(vbucketptrptr)                                               *
 *                                                                          *
 *     vbucketptrptr   is a doubly-indirected pointer to a cvbucketstruc*
 *                     Since we want the value of "vbucketptr" to be        *
 *                     changed by cgetvbucket (i.e., *vbucketptrptr to get   *
 *                     assigned a  pointer to a cvbucketstruct), we need*
 *                     to make its address (vbucketptrptr) the function     *
 *                     argument.  Thus, the need for the double indirection *
 *                     (**).                                                *
 *                                                                          *
 *     cgetvbucket assigns to *vbucketptrptr the pointer to the next         *
 *     available cvbucketstruct from the freedvbucketslist or if none   *
 *     there from the vbucketstore.  In the latter case, if we're at the end*
 *     of the current vbucketstore, a new store is allocated before doing   *
 *     the "get".                                                           *
 *                                                                          *
 ****************************************************************************/

void cgetvbucket(vbucketptrptr)
cvbucketstruct **vbucketptrptr;
{
  if (freedcvbucketslist != NULL) {
    (*vbucketptrptr) = freedcvbucketslist;
    freedcvbucketslist = freedcvbucketslist->nextvbucket;
  }
  else {
    if (cvbucketstorecounter == cVBUCKETSTORE_DIM) {
      callocatevbucketstore();
    }
    
#ifdef _DEBUG
    else if (cvbucketstorecounter > cVBUCKETSTORE_DIM) {
      printf("cgetvbucket error: cvbucketstorecounter=%d,  STOREDIMENSION=%d\n",
	     cvbucketstorecounter,cVBUCKETSTORE_DIM);
      abort();
    }
#endif

    *vbucketptrptr = (cvbucketstruct *)
              &(cvbucketstoreptr->vbucket[cvbucketstorecounter++]);
  }
  (*vbucketptrptr)->nextvbucket = NULL;
  
#ifdef _DEBUG
  if ((*vbucketptrptr)->errorflag != 9009) {
    printf("cgetvbucket error: memory corruption of cvbucketstore\n");
    abort();
  }
#endif
  
}



/****************************************************************************
 *                                                                          *
 *  cgetvector(vectorptrptr)                                                 *
 *                                                                          *
 *     vectorptrptr  is a doubly-indirected pointer to a cvectorstruct. *
 *                                                                          *
 *     cgetvector works in the same way as cgetvector, except that there is   *
 *     no freed list of vectors.  It replaces the functionality of "adnew". *
 *                                                                          *
 ****************************************************************************/

void cgetvector(vectorptrptr)
cvectorstruct **vectorptrptr;
{
  if (cvectorstorecounter == cVECTORSTORE_DIM) {
    callocatevectorstore();
  }

#ifdef _DEBUG
  else if (cvectorstorecounter > cVECTORSTORE_DIM) {
    printf("cgetvector error: cvectorstorecount=%d, cVECTORSTORE_DIM=%d\n",
	   cvectorstorecounter,cVECTORSTORE_DIM);
    abort();
  }
#endif

  *vectorptrptr = (cvectorstruct *)
                    &(cvectorstoreptr->vector[cvectorstorecounter++]);

#ifdef _TRACE
  cadinfo(-1,0,*vectorptrptr,0,NULL,NULL);
#endif

}

