#define COMPLEX
/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#include  <stdlib.h>
#include "cgaxpy_pd_macros_and_data_structures.h"
#include "sgaxpy_pd_macros_and_data_structures.h"
#include "cgaxpy_pd_global_variables.h"
#include "sgaxpy_pd_global_variables.h"

/*
 *  Function Prototyping done locally to grab "TO_pREc_" coorectly.
 */

#define _P(x) ()

void csp2s _P((svectorstruct **destptrptr,
				    cvectorstruct **oldvectorptrptr
				    ));

/****************************************************************************
 *                                                                          *
 *                                                                          *
 *                                                                          *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/

void csp2s(destptrptr,oldvectorptrptr)
svectorstruct **destptrptr;
cvectorstruct **oldvectorptrptr;
{
  svectorstruct *newvectorptr;
  cvectorstruct *oldvectorptr;
  sbucketstruct *newvectortailbucketptr,*newbucketptr;
  cbucketstruct *oldvectorbucketptr;
  sibucketstruct *newvectortailibucketptr,*newibucketptr;
  cibucketstruct *oldvectoribucketptr;
  svbucketstruct *newvectortailvbucketptr,*newvbucketptr;
  cvbucketstruct *oldvectorvbucketptr;
  int i,bucketindex,ibucketindex,vbucketindex;

  oldvectorptr = *oldvectorptrptr;

  if (*destptrptr == NULL)
     sgetvector(destptrptr);

  newvectorptr = lhssvectorptr;

  newvectorptr->nonzeroes =  oldvectorptr->nonzeroes;
  newvectorptr->dstype =  oldvectorptr->dstype;

  if (newvectorptr->dstype == S1) {
    newvectorptr->sds.s1.index =  oldvectorptr->cds.s1.index;
    X_GET_AND_CAST_REAL(newvectorptr->sds.s1.value,
			oldvectorptr->cds.s1.value,float);
  }
  else if (newvectorptr->dstype == SS) {
    oldvectorbucketptr = oldvectorptr->cds.ss.bucketptr;

    for (i=0;i<newvectorptr->nonzeroes;i++) {
      if ((bucketindex = MY_MOD(i,cBUCKET_DIM_minus1)) == 0) {
	if (i == 0) {
	  sgetbucket(&newbucketptr);
	  newvectorptr->sds.ss.bucketptr = newbucketptr;
	}
	else {
	  oldvectorbucketptr = oldvectorbucketptr->nextbucket;
	  sgetbucket(&newbucketptr);
	  newvectortailbucketptr->nextbucket = newbucketptr;
	}
	newvectortailbucketptr = newbucketptr;
      }
      newvectortailbucketptr->index[bucketindex] =
        oldvectorbucketptr->index[bucketindex];
      X_GET_AND_CAST_REAL(newvectortailbucketptr->value[bucketindex],
			  oldvectorbucketptr->value[bucketindex],float);
    }
  }
  else /* if (newvectorptr->dstype == CS */ {
    oldvectoribucketptr = oldvectorptr->cds.cs.ibucketptr;
    oldvectorvbucketptr = oldvectorptr->cds.cs.vbucketptr;
    newvectorptr->sds.cs.nonzeroranges = oldvectorptr->cds.cs.nonzeroranges;
    
    for (i=0;i<newvectorptr->sds.cs.nonzeroranges;i++) {
      if ((ibucketindex = MY_MOD(i,cIBUCKET_DIM_minus1)) == 0) {
	if (i == 0) {
	  sgetibucket(&newibucketptr);
	  newvectorptr->sds.cs.ibucketptr = newibucketptr;
	}
	else {
	  oldvectoribucketptr = oldvectoribucketptr->nextibucket;
	  sgetibucket(&newibucketptr);
	  newvectortailibucketptr->nextibucket = newibucketptr;
	}
	newvectortailibucketptr = newibucketptr;
      }

      newvectortailibucketptr->index_RANGE_BEGIN[ibucketindex] =
        oldvectoribucketptr->index_RANGE_BEGIN[ibucketindex];
      newvectortailibucketptr->index_RANGE_END[ibucketindex] =
        oldvectoribucketptr->index_RANGE_END[ibucketindex];
    }
    
    for (i=0;i<newvectorptr->nonzeroes;i++) {
      if ((vbucketindex = MY_MOD(i,cVBUCKET_DIM_minus1)) == 0) {
	if (i == 0) {
	  sgetvbucket(&newvbucketptr);
	  newvectorptr->sds.cs.vbucketptr = newvbucketptr;
	}
	else {
	  oldvectorvbucketptr = oldvectorvbucketptr->nextvbucket;
	  sgetvbucket(&newvbucketptr);
	  newvectortailvbucketptr->nextvbucket = newvbucketptr;
	}
	newvectortailvbucketptr = newvbucketptr;
      }
      X_GET_AND_CAST_REAL(newvectortailvbucketptr->value[vbucketindex],
			  oldvectorvbucketptr->value[vbucketindex],float);
    }
  }

  saxpy_0_1_post_processing(destptrptr);

}

