#define COMPLEX
/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#include  <stdlib.h>
#include "cgaxpy_pd_macros_and_data_structures.h"
#include "cgaxpy_pd_global_variables.h"

/****************************************************************************
 *                                                                          *
 *     cspzronotca assigns a zero vector to its argument.                        *
 *                                                                          *
 ****************************************************************************/

void cspzronotca(destptrptr)
cvectorstruct **destptrptr;
{
  if (*destptrptr == NULL)
    cgetvector(destptrptr);
  
  lhscvectorptr->nonzeroes = 0;
  lhscvectorptr->dstype = S1;
  lhscvectorptr->cds.s1.index = DUMMYSUBSCRIPT;
  X_ZERO(lhscvectorptr->cds.s1.value);

  caxpy_0_1_post_processing(destptrptr);
}



/****************************************************************************
 *                                                                          *
 *     cspcp assigns to its first vector argument a vector whose values    *
 *     are identical to its second vectpr argument.                         *
 *                                                                          *
 ****************************************************************************/

void cspcp(destptrptr,oldvectorptrptr)
cvectorstruct **destptrptr,**oldvectorptrptr;
{
  cvectorstruct *oldvectorptr,*newvectorptr;
  cbucketstruct *oldvectorbucketptr,
                    *newvectortailbucketptr,*newbucketptr;
  cibucketstruct *oldvectoribucketptr,
                     *newvectortailibucketptr,*newibucketptr;
  cvbucketstruct *oldvectorvbucketptr,
                     *newvectortailvbucketptr,*newvbucketptr;
  int i,bucketindex,ibucketindex,vbucketindex;

#ifdef _PEEP
  complex one;
  X_PUT_REAL(one,1.0);
#endif    

  oldvectorptr = *oldvectorptrptr;

  if (*destptrptr == NULL)
     cgetvector(destptrptr);

#ifdef _PEEP
  adpeep_header(0);
  cadpeep_pre(0,oldvectorptrptr,&one);
#endif    
  newvectorptr = lhscvectorptr;

  newvectorptr->nonzeroes =  oldvectorptr->nonzeroes;
  newvectorptr->dstype =  oldvectorptr->dstype;

  if (newvectorptr->dstype == S1) {
    newvectorptr->cds.s1.index =  oldvectorptr->cds.s1.index;
    newvectorptr->cds.s1.value =  oldvectorptr->cds.s1.value;
  }
  else if (newvectorptr->dstype == SS) {
    oldvectorbucketptr = oldvectorptr->cds.ss.bucketptr;

    for (i=0;i<newvectorptr->nonzeroes;i++) {
      if ((bucketindex = MY_MOD(i,cBUCKET_DIM_minus1)) == 0) {
	if (i == 0) {
	  cgetbucket(&newbucketptr);
	  newvectorptr->cds.ss.bucketptr = newbucketptr;
	}
	else {
	  oldvectorbucketptr = oldvectorbucketptr->nextbucket;
	  cgetbucket(&newbucketptr);
	  newvectortailbucketptr->nextbucket = newbucketptr;
	}
	newvectortailbucketptr = newbucketptr;
      }
      newvectortailbucketptr->value[bucketindex] =
        oldvectorbucketptr->value[bucketindex];
      newvectortailbucketptr->index[bucketindex] =
        oldvectorbucketptr->index[bucketindex];
    }
  }
  else /* if (newvectorptr->dstype == CS */ {
    oldvectoribucketptr = oldvectorptr->cds.cs.ibucketptr;
    oldvectorvbucketptr = oldvectorptr->cds.cs.vbucketptr;
    newvectorptr->cds.cs.nonzeroranges =
      oldvectorptr->cds.cs.nonzeroranges;
    
    for (i=0;i<newvectorptr->cds.cs.nonzeroranges;i++) {
      if ((ibucketindex = MY_MOD(i,cIBUCKET_DIM_minus1)) == 0) {
	if (i == 0) {
	  cgetibucket(&newibucketptr);
	  newvectorptr->cds.cs.ibucketptr = newibucketptr;
	}
	else {
	  oldvectoribucketptr = oldvectoribucketptr->nextibucket;
	  cgetibucket(&newibucketptr);
	  newvectortailibucketptr->nextibucket = newibucketptr;
	}
	newvectortailibucketptr = newibucketptr;
      }

      newvectortailibucketptr->index_RANGE_BEGIN[ibucketindex] =
        oldvectoribucketptr->index_RANGE_BEGIN[ibucketindex];
      newvectortailibucketptr->index_RANGE_END[ibucketindex] =
        oldvectoribucketptr->index_RANGE_END[ibucketindex];
    }
    
    for (i=0;i<newvectorptr->nonzeroes;i++) {
      if ((vbucketindex = MY_MOD(i,cVBUCKET_DIM_minus1)) == 0) {
	if (i == 0) {
	  cgetvbucket(&newvbucketptr);
	  newvectorptr->cds.cs.vbucketptr = newvbucketptr;
	}
	else {
	  oldvectorvbucketptr = oldvectorvbucketptr->nextvbucket;
	  cgetvbucket(&newvbucketptr);
	  newvectortailvbucketptr->nextvbucket = newvbucketptr;
	}
	newvectortailvbucketptr = newvbucketptr;
      }
      newvectortailvbucketptr->value[vbucketindex] =
        oldvectorvbucketptr->value[vbucketindex];
    }
  }

  caxpy_0_1_post_processing(destptrptr);

#ifdef _PEEP
  cadpeep_post(destptrptr);
#endif

}

