#define DOUBLE
/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#include  <stdlib.h>
#include "dgaxpy_pd_macros_and_data_structures.h"
#include "dgaxpy_pd_global_variables.h"

/****************************************************************************
 *                                                                          *
 *     dspg1 performs the arity 1 gaxpy.                                   *
 *                                                                          *
 ****************************************************************************/

void dspg1(destptrptr,alphaptr,vectoraptrptr)
dvectorstruct **vectoraptrptr,**destptrptr;
double *alphaptr;
{

  int arity = 1;

  dvectorstruct *rhsdvectorptr[1];
  dvectorstruct *tempvectorptr;
  
  dbucketstruct *rhsdbucketptr;
  dbucketstruct *lhsdvectortailbucketptr;
  dbucketstruct *lhsdbucketptr;
  dbucketstruct *tempbucketptr;
  
  int i,k;
  int bucketindex;
  
  dvbucketstruct *rhsdvectorvbucketptr;
  dvbucketstruct *lhsdvectortailvbucketptr;
  dvbucketstruct *tempvbucketptr;
  int vbucketindex;
  
  dibucketstruct *rhsdvectoribucketptr;
  dibucketstruct *lhsdvectortailibucketptr;
  dibucketstruct *tempibucketptr;
  dibucketstruct *lhsdibucketptr;
  int ibucketindex;

  double *alphaptrs[1];
  
#ifdef _TRACE
  int nnzb,nnza;
  dvectorstruct *lhsdvectoraddress;
  
  
  dvectorstruct **vectorptrptrs[1];
  
  vectorptrptrs[0] = vectoraptrptr;
  
  if(*destptrptr==NULL){ 
    nnzb = 0; 
    lhsdvectoraddress= (dvectorstruct *) 99900; 
  }
  else {
    nnzb = (*destptrptr)->nonzeroes; 
    lhsdvectoraddress = *destptrptr;
  }
#endif
  
  
  /****************************************
   * Initialization of *destptrptr.       *
   ****************************************/
  
  if (*destptrptr == NULL)
    dgetvector(destptrptr);
  
  rhsdvectorptr[0] = *vectoraptrptr;
  alphaptrs[0] = alphaptr;

#ifdef _PEEP
  if (toggle_adxpy_1_peep) {
    adpeep_header(1);
    dadpeep_pre(0,vectoraptrptr,alphaptr);
  }
#endif
  
  if (rhsdvectorptr[0]->nonzeroes == 0 || X_IS_ZERO(*alphaptr)) {
    dspzronotca(destptrptr);
  }
  else /* if not ZERO AXPY */ {
    if (rhsdvectorptr[0] == *destptrptr &&
	rhsdvectorptr[0]->dstype == CS) {
      plusequaldspg1(destptrptr,alphaptr);
    }
    else /* if (not plusequaldspg1 or (dstype is S1 or SS)) */ {
      lhsdvectorptr->nonzeroes = rhsdvectorptr[0]->nonzeroes;
      lhsdvectorptr->dstype = rhsdvectorptr[0]->dstype;

      if (lhsdvectorptr->dstype == S1) {
	
	/********************************************************************
	 *                       ONE_SUBSCRIPT AXPY                         *
	 ********************************************************************/


	lhsdvectorptr->dds.s1.index =
	                rhsdvectorptr[0]->dds.s1.index;
	GAXPY_EQUALS(lhsdvectorptr->dds.s1.value,*alphaptr,
		        rhsdvectorptr[0]->dds.s1.value);
      }
      else if (lhsdvectorptr->dstype == SS) {
	
	/******************************************************************
	 *                 _SINGLE_SUBSCRIPTS AXPY                        *
	 ******************************************************************/
	
	rhsdbucketptr = rhsdvectorptr[0]->dds.ss.bucketptr;
	for (i=0;i<rhsdvectorptr[0]->nonzeroes;i++) {
	  if ((bucketindex = MY_MOD(i,dBUCKET_DIM_minus1)) == 0) {
	    if (i == 0) {
	      dgetbucket(&lhsdbucketptr);
	      lhsdvectorptr->dds.ss.bucketptr = lhsdbucketptr;
	    }
	    else {
	      rhsdbucketptr = rhsdbucketptr->nextbucket;
	      dgetbucket(&lhsdbucketptr);
	      lhsdvectortailbucketptr->nextbucket = lhsdbucketptr;
	    }
	    lhsdvectortailbucketptr = lhsdbucketptr;
	  }
	  GAXPY_EQUALS(lhsdvectortailbucketptr->value[bucketindex],
		       *alphaptr,rhsdbucketptr->value[bucketindex]);
	  lhsdvectortailbucketptr->index[bucketindex] =
	    rhsdbucketptr->index[bucketindex];
	}
      }
      
      else /* if (lhsdvectorptr->dstype == CS) */ {
	  
	/******************************************************************
	 *                  _COMPRESSED_SUBSCRIPTS AXPY                   *
	 ******************************************************************/
	
	lhsdvectorptr->dds.cs.nonzeroranges =
	  rhsdvectorptr[0]->dds.cs.nonzeroranges;
	rhsdvectoribucketptr =
	  rhsdvectorptr[0]->dds.cs.ibucketptr;
	rhsdvectorvbucketptr =
	  rhsdvectorptr[0]->dds.cs.vbucketptr;
	
	for (i=0;i<rhsdvectorptr[0]->dds.cs.nonzeroranges;i++) {
	  if ((ibucketindex = MY_MOD(i,dIBUCKET_DIM_minus1)) == 0) {
	    if (i == 0) {
	      dgetibucket(&lhsdibucketptr);
	      lhsdvectorptr->dds.cs.ibucketptr = lhsdibucketptr;
	    }
	    else {
	      rhsdvectoribucketptr =
		rhsdvectoribucketptr->nextibucket;
	      dgetibucket(&lhsdibucketptr);
	      lhsdvectortailibucketptr->nextibucket = lhsdibucketptr;
	    }
	    lhsdvectortailibucketptr = lhsdibucketptr;
	  }
	  
	  lhsdvectortailibucketptr->index_RANGE_BEGIN[ibucketindex] =
	    rhsdvectoribucketptr->index_RANGE_BEGIN[ibucketindex];
	  lhsdvectortailibucketptr->index_RANGE_END[ibucketindex] =
	    rhsdvectoribucketptr->index_RANGE_END[ibucketindex];
	}
	
	for (i=0;i<rhsdvectorptr[0]->nonzeroes;i++) {
	  if ((vbucketindex = MY_MOD(i,dVBUCKET_DIM_minus1)) == 0) {
	    if (i == 0) {
	      dgetvbucket(&lhsdvbucketptr);
	      lhsdvectorptr->dds.cs.vbucketptr = lhsdvbucketptr;
	    }
	    else {
	      rhsdvectorvbucketptr =
		rhsdvectorvbucketptr->nextvbucket;
	      dgetvbucket(&lhsdvbucketptr);
	      lhsdvectortailvbucketptr->nextvbucket = lhsdvbucketptr;
	    }
	    lhsdvectortailvbucketptr = lhsdvbucketptr;
	  }
	  GAXPY_EQUALS(lhsdvectortailvbucketptr->value[vbucketindex],
		       *alphaptr,
		       rhsdvectorvbucketptr->value[vbucketindex]);
	}
      } /* matches     else ** if (lhsdvectorptr->dstype == CS) ** {  */
    
      daxpy_0_1_post_processing(destptrptr);

    } /* matches   else ** if (not plusequaldspg1) ** {                   */
  } /* matches   else ** if not ZERO AXPY ** {                             */
  

#ifdef _PEEP
  if (toggle_adxpy_1_peep) {
    dadpeep_post(destptrptr);
  }
#endif


#ifdef _TRACE 
  nnza = (*destptrptr)->nonzeroes; 
  dadinfo(arity,nnzb,lhsdvectoraddress,nnza,alphaptrs,vectorptrptrs);
#endif
  
}
  
