#define DOUBLE
/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#include  <stdlib.h>
#include "dgaxpy_pd_macros_and_data_structures.h"
#include "dgaxpy_pd_global_variables.h"

/****************************************************************************
 *                                                                          *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/

void dspg2(destptrptr,alphaptr,vectoraptrptr,betaptr,vectorbptrptr)
dvectorstruct **destptrptr;
double *alphaptr;
dvectorstruct **vectoraptrptr;
double *betaptr;
dvectorstruct **vectorbptrptr;
{
  
  int arity = 2;
  
  dvectorstruct *rhsdvectorptr[2];
  
  dbucketstruct *rhsdbucketptr[2];
  
  dbucketstruct *lhsdbucketptr;
  
  dibucketstruct *rhsdibucketptr[2];
                                      /* Array of pointers to "current"     *
				       * buckets where "current" nonzeroes  *
				       * of the rhs_vectors reside.         */

  dibucketstruct *lhsdibucketptr;  /* Pointer to current bucket of       *
				       * lhsdvector.                         */

  double alpha[2];          /* Dereferenced alphaptrs[].          */
  double *alphaptrs[2];

  int rhsdbucketindex[2];              /* Array of indices of current   *
				       * nonzeroes in the current buckets   *
				       * of rhs_vectors.                    */

  int lhsdbucketindex;                 /* bucketindex of next nonzero to be  *
				       * added to lhsdvector.                */

  int lastlhsdbucketindex;             /* bucketindex of the latest added    *
				       * nonzero in the lhsdvector.  We need *
				       * this additional index to leap-frog *
				       * buckets, if need be.               */


  int rhsdibucketindex[2];             /* Array of indices of current        *
				       * nonzeroes in the current buckets   *
				       * of rhs_vectors.                    */

  int lhsdibucketindex;                /* bucketindex of next nonzero to be  *
				       * added to lhsdvector.                */

  int lastlhsdibucketindex;            /* bucketindex of the latest added    *
				       * nonzero in the lhsdvector.  We need *
				       * this additional index to leap-frog *
				       * buckets, if need be.               */

  int k;                              /* Counter for arity.                 */

  int j;                              /* Index for vectorsubscripts to be   *
				       * entered in (and later popped from) *
				       * the heap.                          */

  int rhsdvectornonzeroes[2];          /* Array of total # of nonzeroes in   *
				       * each rhs_vector.                   */

  int rhsdvectorconsumednonzeroes[2];  /* Array of # of nonzeroes consumed   *
				       * so far in each rhs_vector.         */

  int lhsdvectornonzeroes;             /* Count of # of nonzeroes entered    *
				       * into the lhsdvector so far.  By the *
				       * end of the saxpy operation, this   *
				       * index equals the total # of        *
				       * nonzeroes in the resultant vector. */

/*
 * These variables were originally for COMPRESSED_SUBSCRIPTS
 */

  dvbucketstruct *lastlhsdvbucketptr; 
                                      /* Pointer to current vbucket of      *
				       * lhsdvector.                    */

  int rhsdvectornonzeroranges[2];
                                      /* Array of total # of nonzeroes      *
				       * ranges in each rhs_vector.         */

  int rhsdvectorconsumednonzeroranges[2];
                                      /* Array of # of nonzeroes consumed   *
				       * so far in each rhs_vector.         */

  int rhsdvbucketindex[2];             /* Array of indices of current        *
				       * nonzeroes in the current buckets   *
				       * of rhs_vectors.                    */

  int lhsdvectornonzeroranges;         /* Count of # of nonzeroes entered    *
				       * into the lhsdvector so far.  By the *
				       * end of the saxpy operation, this   *
				       * index equals the total # of        *
				       * nonzeroes in the resultant vector. */

  int lhsdvbucketindex;                /* bucketindex of next nonzero to be  *
				       * added to lhsdvector.                */


  int lastlhsdvbucketindex;            /* bucketindex of the latest added    *
				       * nonzero in the lhsdvector.  We need *
				       * this additional index to leap-frog *
				       * buckets, if need be.               */

  dvbucketstruct *templhsdvbucketptr;
  int templhsdvbucketindex;

  int j1;                             /* Index for vectorsubscripts to be   *
				       * entered in (and later popped from) *
				       * the heap.                          */

  int j2;                             /* Index for vectorsubscripts to be   *
				       * entered in (and later popped from) *
				       * the heap.                          */

  int h1;
  int h2;

  int totalcurrentnonzerorange;

#ifdef _DEBUG
  int unconsumedcurrentnonzerorange;
#endif

  int distancetooverlap;
  int distancetoendofvbucket;

  double tmp_add,tmp_mult1,tmp_mult2;


  /*
   * These variables were added for union ds
   */
  
  dvectorstruct *restorerhsdvectorptr[MAXARITY+1];

  int governingdstype;
  
  /*
   * These variables were added for plusequal code
   */
  
  dvectorstruct **rhsdvectorptrptrs[MAXARITY];
  
  int condition[2];
  
#ifdef _PLUSEQUAL  
  int smallest_rhs_index; 
#endif /* _PLUSEQUAL */
  
  
#ifdef _TRACE
  int nnzb,nnza;
  dvectorstruct *lhsdvectoraddress;
  
  dvectorstruct **vectorptrptrs[2];
  
  vectorptrptrs[0] = vectoraptrptr;
  
  vectorptrptrs[1] = vectorbptrptr;
  
  if(*destptrptr==NULL){ 
    nnzb = 0; 
    lhsdvectoraddress= (dvectorstruct *) 99900; 
  }
  else {
    nnzb = (*destptrptr)->nonzeroes; 
    lhsdvectoraddress = *destptrptr;
  }
#endif
  
  
  /****************************************
   * Initialization of *destptrptr.       *
   ****************************************/
  
  if (*destptrptr == NULL)
    dgetvector(destptrptr);
  
  /****************************************
   * Dereferencing.                       *
   ****************************************/
  
  alpha[0] = *alphaptr;
  rhsdvectorptr[0] = *vectoraptrptr;
  alphaptrs[0] = alphaptr;
  
#ifdef _PEEP
  adpeep_header(2);
  dadpeep_pre(0,vectoraptrptr,alphaptr);
#endif    

  alpha[1] = *betaptr;
  rhsdvectorptr[1] = *vectorbptrptr;
  alphaptrs[1] = betaptr;

#ifdef _PEEP
  dadpeep_pre(1,vectorbptrptr,betaptr);
#endif    
 
  for (k=0;k<arity;k++) {
    restorerhsdvectorptr[k] = rhsdvectorptr[k];
    rhsdvectornonzeroes[k] = rhsdvectorptr[k]->nonzeroes;
    if (rhsdvectorptr[k]->dstype == CS)
      rhsdvectornonzeroranges[k] =
	rhsdvectorptr[k]->dds.cs.nonzeroranges;
  }
  
#ifdef _PLUSEQUAL
  smallest_rhs_index = DUMMYSUBSCRIPT;
  for (k=0;k<arity;k++) {
    if (rhsdvectorptr[k]->dstype == S1) 
      smallest_rhs_index = min (smallest_rhs_index,
				rhsdvectorptr[k]->dds.s1.index);
    else if (rhsdvectorptr[k]->dstype == SS)
      smallest_rhs_index =
	min (smallest_rhs_index,
	     rhsdvectorptr[k]->dds.ss.bucketptr->index[0]);
    else /* if (rhsdvectorptr[k]->dstype == CS) */
      smallest_rhs_index =
	min (smallest_rhs_index,
	     rhsdvectorptr[k]->dds.cs.ibucketptr->index_RANGE_BEGIN[0]);
  }
#endif /* _PLUSEQUAL */


  if ((rhsdvectorptr[0]->nonzeroes == 0 || X_IS_ZERO(alpha[0])) &&
      (rhsdvectorptr[1]->nonzeroes == 0 || X_IS_ZERO(alpha[1]))) {
    dspzronotca(destptrptr);
  }
  else if (rhsdvectorptr[0]->nonzeroes == 0 || X_IS_ZERO(alpha[0])) {
    if (rhsdvectorptr[1] == *destptrptr &&
	rhsdvectorptr[1]->dstype == CS) {
      plusequaldspg1(destptrptr,betaptr);
    }
    /* if (rhsdvectorptr[1] != *destptrptr ||
       rhsdvectorptr[1]->dstype != CS)           */
    else {
#ifdef _PEEP
      toggle_adxpy_1_peep = 0;
      dspg1(destptrptr,betaptr,vectorbptrptr);
      toggle_adxpy_1_peep = 1;
#else
      dspg1(destptrptr,betaptr,vectorbptrptr);
#endif     
    }
  }
  else if (rhsdvectorptr[1]->nonzeroes == 0 || X_IS_ZERO(alpha[1])) {
    if (rhsdvectorptr[0] == *destptrptr &&
	rhsdvectorptr[0]->dstype == CS) {
      plusequaldspg1(destptrptr,alphaptr);
    }
    /* else if (rhsdvectorptr[0] != *destptrptr) ||
       rhsdvectorptr[0]->dstype != CS)          */
    else {      
#ifdef _PEEP
      toggle_adxpy_1_peep = 0;
      dspg1(destptrptr,alphaptr,vectoraptrptr);
      toggle_adxpy_1_peep = 1;
#else
      dspg1(destptrptr,alphaptr,vectoraptrptr);
#endif
    }
  }
  else /* neither rhs_operation is a NO-OP */ {
    if (rhsdvectorptr[0] == *destptrptr &&
	rhsdvectorptr[1] == *destptrptr &&
	rhsdvectorptr[0]->dstype == CS) {
#ifdef _DEBUG
      printf("dspg2 warning: lhsdvector appears more than ");
      printf("once on rhs.\n");
#endif
      X_ADD(tmp_add,*alphaptr,*betaptr);
      plusequaldspg1(destptrptr,&tmp_add);
    }

#ifdef _PLUSEQUAL
    else if
      (rhsdvectorptr[0] == *destptrptr &&
       rhsdvectorptr[0]->dstype == CS &&
       rhsdvectorptr[0]->dds.cs.ibucketptr->index_RANGE_BEGIN[0] <=
       smallest_rhs_index &&
       X_IS_ONE(alpha[0])) {
	rhsdvectorptrptrs[1] = &rhsdvectorptr[1];
	if ((*rhsdvectorptrptrs[1])->dstype != CS &&
	    (*rhsdvectorptrptrs[1])->nonzeroes !=0) {
	  dconvertdstype(rhsdvectorptrptrs[1],1,CS);
	}
	plusequaldspg2(destptrptr,
			betaptr,rhsdvectorptrptrs[1],ORIG_OTHER_VECTOR);
      }
    else if
      (rhsdvectorptr[1] == *destptrptr &&
       rhsdvectorptr[1]->dstype == CS &&
       rhsdvectorptr[1]->dds.cs.ibucketptr->index_RANGE_BEGIN[0] <=
       smallest_rhs_index &&
       X_IS_ONE(alpha[1])) {
	rhsdvectorptrptrs[0] = &rhsdvectorptr[0];
	if ((*rhsdvectorptrptrs[0])->dstype != CS &&
	    (*rhsdvectorptrptrs[0])->nonzeroes !=0) {
	  dconvertdstype(rhsdvectorptrptrs[0],0,CS);
	}
	plusequaldspg2(destptrptr,
			alphaptr,rhsdvectorptrptrs[0],ORIG_OTHER_VECTOR);
	
      }
    
    /* 
     * else if neither rhs_operation is a NO-OP, and neither
     * rhsdvector is (the same as the lhsdvector AND has
     * multiplier equal to 1 AND is CS)
     */
    
    else if (rhsdvectorptr[0] ==  rhsdvectorptr[1]) {
#ifdef _DEBUG
      printf("dspg2 warning: lhsdvector appears more than ");
      printf("once on rhs.\n");
#endif
      X_ADD(tmp_add,*alphaptr,*betaptr);
#ifdef _PEEP
      toggle_adxpy_1_peep = 0;
      dspg1(destptrptr,&tmp_add,vectoraptrptr);
      toggle_adxpy_1_peep = 1;
#else
      dspg1(destptrptr,&tmp_add,vectoraptrptr);
#endif
    }
#endif /* _PLUSEQUAL */
    
    else /* do an honest to goodness dspg2 */ {
      
      governingdstype = max(rhsdvectorptr[0]->dstype,
			    rhsdvectorptr[1]->dstype);
      if (governingdstype == S1 &&
	  rhsdvectorptr[0]->dds.s1.index != rhsdvectorptr[1]->dds.s1.index)
	governingdstype = SS;
      for (k=0;k<arity;k++)
	if (rhsdvectorptr[k]->dstype != governingdstype) {
	  dconvertdstype((&rhsdvectorptr[k]),k,governingdstype);
	  if (governingdstype == CS)
	    rhsdvectornonzeroranges[k] =
	      rhsdvectorptr[k]->dds.cs.nonzeroranges;
	}
      
      if (governingdstype == S1) {
	
	/******************************************************************
	 *                      ONE_SUBSCRIPT AXPY                        *
	 ******************************************************************/

	lhsdvectorptr->nonzeroes =  1;
	lhsdvectorptr->dstype = governingdstype;
	lhsdvectorptr->dds.s1.index =
	  rhsdvectorptr[0]->dds.s1.index;
        X_MULT(tmp_mult1,*alphaptr,rhsdvectorptr[0]->dds.s1.value);
        X_MULT(tmp_mult2,*betaptr,rhsdvectorptr[1]->dds.s1.value);
        X_ADD(lhsdvectorptr->dds.s1.value,tmp_mult1,tmp_mult2);

	daxpy_0_1_post_processing(destptrptr);

      }
      else /* if we are going to use lhsdvectorptr */ {
	if (governingdstype == SS) {
	  
	  /****************************************************************
	   *                 _SINGLE_SUBSCRIPTS AXPY                      *
	   ****************************************************************/
	  for (k=0;k<arity;k++) {
	    rhsdvectorconsumednonzeroes[k]= 0;
	    rhsdbucketindex[k] = 0;
	    rhsdbucketptr[k] = rhsdvectorptr[k]->dds.ss.bucketptr;
	  }
	  dgetbucket(&lhsdbucketptr);
	  lhsdvectorptr->dstype = governingdstype;
	  lhsdvectorptr->dds.ss.bucketptr = lhsdbucketptr;
	  lhsdvectornonzeroes = 0;
	  
	  for (k=0;k<arity;k++)
	    condition[k] = rhsdvectorconsumednonzeroes[k] <
	      rhsdvectorptr[k]->nonzeroes;
	  
	  while (condition[0] || condition[1]) {
	    if (condition[0] && condition[1])
	      if (rhsdbucketptr[0]->index[rhsdbucketindex[0]] <=
		  rhsdbucketptr[1]->index[rhsdbucketindex[1]])
		k = 0;
	      else
		k = 1;
	    else if (condition[0])
	      k = 0;
	    else
	      k = 1;
	    
	    j = rhsdbucketptr[k]->index[rhsdbucketindex[k]];
	    
	    lhsdbucketindex = MY_MOD(lhsdvectornonzeroes,
					 dBUCKET_DIM_minus1);
	    
	    /****************************************
	     * Perform saxpy:                       *
	     * 3.1. Found 1st or new subscript for  *
	     *      lhsdvector.  If needed, get new  *
	     *      bucket for lhsdvector.           *
	     *      Consume nonzero in saxpy.       *
	     ****************************************/
	    if (lhsdvectornonzeroes == 0 ||
		lhsdbucketptr->index[lastlhsdbucketindex] < j) {
	      if (lhsdbucketindex == 0 && lhsdvectornonzeroes != 0) {
		dgetbucket(&(lhsdbucketptr->nextbucket));
		lhsdbucketptr = lhsdbucketptr->nextbucket;
	      }
	      lhsdbucketptr->index[lhsdbucketindex] = j;
	      GAXPY_EQUALS(
                lhsdbucketptr->value[lhsdbucketindex],
		alpha[k],
		rhsdbucketptr[k]->value[rhsdbucketindex[k]]);
	      lastlhsdbucketindex = lhsdbucketindex;
	      lhsdvectornonzeroes++;
	    }
	    
	    /****************************************
	     * Perform saxpy:                       *
	     * 3.2. Found old subscript for         *
	     *      lhsdvector.                      *
	     *      Consume nonzero in saxpy.       *
	     ****************************************/
	    else if (lhsdbucketptr->index[lastlhsdbucketindex] == j) {
	      GAXPY_PLUS_EQUALS(
                lhsdbucketptr->value[lastlhsdbucketindex],
		alpha[k],
		rhsdbucketptr[k]->value[rhsdbucketindex[k]]);
	    }
	    
	    /****************************************
	     * Perform saxpy:                       *
	     * 3.3. Found wrong subscript for       *
	     *      lhsdvector.  Abort.               *
	     ****************************************/
	    
#ifdef _DEBUG
	    else {
	      printf("dspg2 error 6: lhsdnz=%d, lastlhsdindex[]=%d, j=%d\n",
		     lhsdvectornonzeroes,
		     lhsdbucketptr->index[lastlhsdbucketindex],j);
	      abort();
	    }
#endif
	    
	    /****************************************
	     * Perform saxpy:                       *
	     * 3.4. If there is one, get new        *
	     *      nonzero from the rhs_vector     *
	     *      whose nonzero was just          *
	     *      consumed and replace the        *
	     *      heaproot with new entry;        *
	     *      Else, delete heaproot.          *
	     ****************************************/
	    rhsdvectorconsumednonzeroes[k]++;
	    condition[k] = rhsdvectorconsumednonzeroes[k] <
	      rhsdvectorptr[k]->nonzeroes;
	    rhsdbucketindex[k] = MY_MOD(rhsdvectorconsumednonzeroes[k],
					    dBUCKET_DIM_minus1);
	    if (condition[k]) {
	      if (rhsdbucketindex[k]  == 0) {
		if ((rhsdbucketptr[k] = rhsdbucketptr[k]->nextbucket) 
		    == NULL) {
#ifdef _DEBUG
		  printf("dspg2 err: unexpected NULL nextbucket: k=%d",k);
		  printf("rhsdvectorconsumednonzeros=%d,rhsdvecnonzeroes=%d,",
			 rhsdvectorconsumednonzeroes[k],
			 rhsdvectornonzeroes[k]);
		  abort();
#endif
		}	  
	      }
	    }
	  } /* matches     while loop     */
	} /* matches     else if (governingdstype == SS) {      */
	
	else /* if (governingdstype == CS) */ {
	  
	  /****************************************************************
	   *               _COMPRESSED_SUBSCRIPTS AXPY                    *
	   ****************************************************************/
	  
	  /****************************************
	   * Else perform saxpy:                  *
	   * 1. Put in heap the 1st nonzero of    *
	   *    each vector.                      *
	   ****************************************/
	  for (k=0;k<arity;k++) {
	    rhsdvectorconsumednonzeroranges[k] = 0;
	    rhsdibucketindex[k] = 0;
	    rhsdvbucketindex[k] = 0;
	    rhsdibucketptr[k] =
	      rhsdvectorptr[k]->dds.cs.ibucketptr;
	    rhsdvbucketptr[k] =
	      rhsdvectorptr[k]->dds.cs.vbucketptr;
	    
#ifdef _DEBUG
	    if (rhsdibucketptr[k] == NULL ||
		rhsdvbucketptr[k] == NULL) {
	      printf("dspg2 error 10: Unexpected NULL rhsdbucketptrs: ");
	      printf("vector # %d\n",k);
	      abort();
	    }
#endif
	  }	  
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 2. Initialize lhsdvector.             *
	   ****************************************/
	  dgetibucket(&lhsdibucketptr);
	  lhsdvectorptr->dstype = governingdstype;
	  lhsdvectorptr->dds.cs.ibucketptr = lhsdibucketptr;
	  dgetvbucket(&lhsdvbucketptr);
	  lhsdvectorptr->dds.cs.vbucketptr = lhsdvbucketptr;
	  lhsdvectornonzeroranges = 0;
	  lhsdvectornonzeroes = 0;
	  lhsdvbucketindex = 0;
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3. Get nonzero by popping the        *
	   *    heaproot; consume it in saxpy;    *
	   *    either replace or delete heaproot.*
	   ****************************************/
	  
	  while (rhsdvectorconsumednonzeroranges[0] <
		 rhsdvectorptr[0]->dds.cs.nonzeroranges ||
		 rhsdvectorconsumednonzeroranges[1] <
		 rhsdvectorptr[1]->dds.cs.nonzeroranges) {
	    if
	      ((rhsdvectorconsumednonzeroranges[1] ==
		rhsdvectorptr[1]->dds.cs.nonzeroranges) ||
	       ((rhsdvectorconsumednonzeroranges[0] <
		 rhsdvectorptr[0]->dds.cs.nonzeroranges) &&
		(rhsdibucketptr[0]->index_RANGE_BEGIN[rhsdibucketindex[0]] <=
		 rhsdibucketptr[1]->index_RANGE_BEGIN[rhsdibucketindex[1]])))
		k = 0;
	    else if
	      ((rhsdvectorconsumednonzeroranges[0] ==
		rhsdvectorptr[0]->dds.cs.nonzeroranges) ||
	       ((rhsdvectorconsumednonzeroranges[1] <
		 rhsdvectorptr[1]->dds.cs.nonzeroranges) &&
		(rhsdibucketptr[0]->index_RANGE_BEGIN[rhsdibucketindex[0]] >
		 rhsdibucketptr[1]->index_RANGE_BEGIN[rhsdibucketindex[1]])))
		k = 1;
#ifdef _DEBUG
	    else {
	      printf("dspg2 error: BAD LOGIC.\n");
	      abort();
	    }
#endif
	    j1 = rhsdibucketptr[k]->index_RANGE_BEGIN[rhsdibucketindex[k]];
	    j2 = rhsdibucketptr[k]->index_RANGE_END[rhsdibucketindex[k]];
	    rhsdibucketindex[k] = 
	      MY_MOD(rhsdvectorconsumednonzeroranges[k],
		     dIBUCKET_DIM_minus1);
	    lhsdibucketindex = MY_MOD(lhsdvectornonzeroranges,
					  dIBUCKET_DIM_minus1);
	    
	    if ((rhsdvbucketindex[k] = MY_MOD(rhsdvbucketindex[k],
						  dVBUCKET_DIM_minus1))==0
		&& rhsdvectorconsumednonzeroranges[k] !=0) {
	      dHOP_TO_NEXT_RHSVBUCKET(k,&rhsdvbucketindex[k],
				     rhsdvectorconsumednonzeroranges[k],
				     rhsdvectornonzeroranges[k]);
	    }
	    lhsdvbucketindex = MY_MOD(lhsdvbucketindex,
					  dVBUCKET_DIM_minus1);
	    
#ifdef _DEBUG
	    if (rhsdibucketptr[k]->errorflag != 9099 ||
		rhsdvbucketptr[k]->errorflag != 9009) {
	      printf("dspg2 err 14: memory corruption of dibucketstore ");
	      printf("or dvbucketstore.\n");
	      abort();
	    }
#endif      
	    
	    /****************************************
	     * Perform saxpy:                       *
	     * 3.0. Check for startup errors.       *
	     ****************************************/
	    
#ifdef _DEBUG
	    if (lhsdvectornonzeroranges == 0) {
	      if (rhsdibucketindex[k] != 0 || rhsdvbucketindex[k] != 0 ||
		  lhsdibucketindex != 0 || lhsdvbucketindex != 0) {
		printf("dspg2 error 15: Zeroes expected: \n\t");
		printf("rhsdibucketindex[k]=%d, rhsdvbucketindex[k]=%d\n\t",
		       rhsdibucketindex[k], rhsdvbucketindex[k]);
		printf("lhsdibucketindex=%d, lhsdvbucketindex=%d\n\t",
		       lhsdibucketindex,lhsdvbucketindex);
		abort();
	      }
	    }
#endif      
	    
	    /****************************************
	     * Perform saxpy:                       *
	     * 3.1. Found 1ST or NEW subscript      *
	     *      range (non-overlapping with     *
	     *      previous one, if any) for       *
	     *      lhsdvector.  If needed, get new  *
	     *      ibuckets and vbuckets for       *
	     *      lhsdvector.                      *
	     ****************************************/
	    if (lhsdvectornonzeroranges == 0 ||
		j1 > lhsdibucketptr->index_RANGE_END[lastlhsdibucketindex]+1) {/*A.1*/
	      
	      if (lhsdibucketindex == 0 && lhsdvectornonzeroranges != 0) {
		dgetibucket(&(lhsdibucketptr->nextibucket));
		lhsdibucketptr = lhsdibucketptr->nextibucket;
	      }
	      lhsdibucketptr->index_RANGE_BEGIN[lhsdibucketindex] = j1;
	      lhsdibucketptr->index_RANGE_END[lhsdibucketindex] = j2;
	      
	      if (lhsdvbucketindex == 0 && lhsdvectornonzeroranges != 0)
		dget_and_hop_to_next_vbucket(&lhsdvbucketindex);
	      
	      lastlhsdvbucketptr = lhsdvbucketptr;
	      lastlhsdvbucketindex = lhsdvbucketindex;
	      
	      totalcurrentnonzerorange = j2-j1+1;
	      
	      lhsdvectornonzeroes += totalcurrentnonzerorange;
	      lhsdvectornonzeroranges++;
	      
	
	      dNON_OVER_LAPPING_CS_RANGE_AXPY(COMPLETELY_NEW_RANGE,
					     lhsdibucketindex,
					     &lastlhsdibucketindex,
					     &lhsdvbucketindex,
					     k,
					     &rhsdvbucketindex[k],
					     alpha[k],
					     totalcurrentnonzerorange,
					     &unconsumedcurrentnonzerorange,
					     rhsdvectorconsumednonzeroranges[k],
					     rhsdvectornonzeroranges[k],
					     lhsdvectornonzeroranges);
	      
	            
	    } /*A.1*/
	    
	    
	    /****************************************
	     * Perform saxpy:                       *
	     * 3.2. Found old subscript for         *
	     *      lhsdvector.  Consume  nonzero in *
	     *      saxpy.                          *
	     *                                      *
	     ****************************************/
	    else if (j1 <= lhsdibucketptr->index_RANGE_END[lastlhsdibucketindex]+1 &&
		     j1 >= lhsdibucketptr->index_RANGE_BEGIN[lastlhsdibucketindex]
		     ) { /*B.1*/
	      
	      templhsdvbucketptr = lhsdvbucketptr;
	      templhsdvbucketindex = lhsdvbucketindex;
	      lhsdvbucketptr = lastlhsdvbucketptr;
	      
	      distancetooverlap =
		j1 - lhsdibucketptr->index_RANGE_BEGIN[lastlhsdibucketindex];
	      distancetoendofvbucket = dVBUCKET_DIM -
		lastlhsdvbucketindex;
	      
	      if (j1 ==
		  lhsdibucketptr->index_RANGE_END[lastlhsdibucketindex]+1) {
		h1 = 0;
		h2 = -1;
	      }
	      else if (j2 <= lhsdibucketptr->index_RANGE_END[lastlhsdibucketindex]
		       ) {
		h1 = j1;
		h2 = j2;
		j1 = 0;
		j2 = -1;
	      }
	      else {
		h1 = j1;
		h2 = lhsdibucketptr->index_RANGE_END[lastlhsdibucketindex];
		j1 = h2+1;
	      }
	      
	      if (distancetoendofvbucket > distancetooverlap) {
		lhsdvbucketindex =
		  MY_MOD(lastlhsdvbucketindex+distancetooverlap,
			 dVBUCKET_DIM_minus1);
	      }
	      else if (distancetoendofvbucket < distancetooverlap) {
		lhsdvbucketindex =
		  MY_MOD(distancetooverlap - distancetoendofvbucket,
			 dVBUCKET_DIM_minus1);
	      }
	      else {
		lhsdvbucketindex = 0;
	      }
	      
	      while (distancetooverlap >= distancetoendofvbucket &&
		     lhsdvbucketptr->nextvbucket != NULL) {
		lhsdvbucketptr = lhsdvbucketptr->nextvbucket;
		distancetooverlap -= distancetoendofvbucket;
		distancetoendofvbucket = dVBUCKET_DIM;
	      }
	      
	      lhsdibucketptr->index_RANGE_END[lastlhsdibucketindex] =
		max(j2,
		    lhsdibucketptr->index_RANGE_END[lastlhsdibucketindex]);
	      
	      if ((totalcurrentnonzerorange = h2-h1+1) > 0) {
		

		dOVERLAPPING_CS_RANGE_AXPY(&lhsdvbucketindex,
					  k,
					  &rhsdvbucketindex[k],
					  alpha[k],
					  totalcurrentnonzerorange,
					  &unconsumedcurrentnonzerorange,
					  rhsdvectorconsumednonzeroranges[k],
					  rhsdvectornonzeroranges[k],
					  lhsdvectornonzeroranges);
		
	      }
	     
	      /*
	       * This bit of logic must be performed, whether or not there
	       * is additional nonzeroes to process.
	       */
	      
	      if (j2 != -1 &&
		  (lhsdvbucketptr != templhsdvbucketptr ||
		   (lhsdvbucketindex = MY_MOD(lhsdvbucketindex,
						  dVBUCKET_DIM_minus1)) !=
		   templhsdvbucketindex)) {
		
#ifdef _DEBUG
		printf("dspg2 err 50.0: Unexpectd value discrepancies\n");
		printf("lhsdvbucketptr=%d, templhsdvbucketptr=%d\n",
		       (int)lhsdvbucketptr,(int)templhsdvbucketptr);
		printf("lhsdvbucketindex=%d, templhsdvbucketindex=%d\n",
		       lhsdvbucketindex,templhsdvbucketindex);
		
		abort();
#endif
	      }
	      else {
		lhsdvbucketptr = templhsdvbucketptr;
		lhsdvbucketindex = templhsdvbucketindex;
	      }
	      
	      
	      if ((totalcurrentnonzerorange = j2-j1+1) > 0) {
		lhsdvectornonzeroes += totalcurrentnonzerorange;
		
		/*
		 * adding new entries to vbucket
		 */
		
		if (lhsdvbucketindex == 0 &&
		    lhsdvectornonzeroranges != 0 &&
		    j2 != -1)
		  dget_and_hop_to_next_vbucket(&lhsdvbucketindex);
		
		if(rhsdvbucketindex[k] == dVBUCKET_DIM && j2 != -1) {

		  dHOP_TO_NEXT_RHSVBUCKET(k,&rhsdvbucketindex[k],
					 rhsdvectorconsumednonzeroranges[k],
					 rhsdvectornonzeroranges[k]);
		}
		
		
		
		dNON_OVER_LAPPING_CS_RANGE_AXPY(NEW_ENTRIES_PADDED_TO_OLD_RANGE,
					       lhsdibucketindex,
					       &lastlhsdibucketindex,
					       &lhsdvbucketindex,
					       k,
					       &rhsdvbucketindex[k],
					       alpha[k],
					       totalcurrentnonzerorange,
					       &unconsumedcurrentnonzerorange,
					       rhsdvectorconsumednonzeroranges[k],
					       rhsdvectornonzeroranges[k],
					       lhsdvectornonzeroranges);
		
	      } /* matches     if ((totalcurrent...) { */
	    } /*B.1*/
	    
	    
	    /****************************************	
	     * Perform saxpy:                       *
	     * 3.3. Found wrong subscript for       *
	     *      lhsdvector.  Abort.               *
	     ****************************************/
	    
#ifdef _DEBUG
	    else {
	      printf("dspg2 error 60: lhsdvectornonzeroranges=%d,",
		     lhsdvectornonzeroranges);
	      printf("lastlhsdibucketindex=%d, \n\t",lastlhsdibucketindex);
	      printf("lhsdibucketptr->index[%d] = (%d,%d), j1=%d\n",
		     lastlhsdibucketindex,
		     lhsdibucketptr->index_RANGE_BEGIN[lastlhsdibucketindex],
		     lhsdibucketptr->index_RANGE_END[lastlhsdibucketindex],
		     j1);
	      abort();
	    }
#endif
	    
	    
	    /****************************************
	     * Perform saxpy:                       *
	     * 3.4. If there is one, get new        *
	     *      nonzero from the rhs_vector     *
	     *      whose nonzero was just          *
	     *      consumed and replace the        *
	     *      heaproot with new entry;        *
	     *      Else, delete heaproot.          *
	     ****************************************/
	    if (++rhsdvectorconsumednonzeroranges[k] <
		rhsdvectornonzeroranges[k]) {
	      if ((rhsdibucketindex[k] = 
		   MY_MOD(rhsdvectorconsumednonzeroranges[k],
			  dIBUCKET_DIM_minus1)) == 0) {
		if ((rhsdibucketptr[k] =
		     rhsdibucketptr[k]->nextibucket) == NULL) {
		  
#ifdef _DEBUG
		  printf("dspg2 err 61: unexpected NULL nextvbucket: \n");
		  printf("\t k=%d, rhsdvectorconsumednonzeroranges=%d,\n",
			 k,rhsdvectorconsumednonzeroranges[k]);
		  printf("\t rhsdvectornonzeroranges[k]=%d\n",
			 rhsdvectornonzeroranges[k]);
		  abort();
#endif
		  
		}
	      }
	    }   
	  } /* matches     while loop          */
	} /* matches     else ** if (governingdstype == CS) ** {     */
	

	
	
	
	rhsdvectorptrptrs[0] = &rhsdvectorptr[0];
	rhsdvectorptrptrs[1] = &rhsdvectorptr[1];
	
	daxpy_2_heap_post_processing(destptrptr,
				    rhsdvectorptrptrs,
				    alphaptrs,
				    restorerhsdvectorptr,
				    arity,
				    lhsdvectornonzeroes,
				    lhsdvectornonzeroranges,
				    governingdstype);
	
      } /* matches     if we are going to use lhsdvectorptr           */
    } /* matches     else ** do an honest to goodness dspg2 **  {        */
  } /* matches     else ** neither rhs_operation is a NO-OP  ** {         */
    
  
#ifdef _PEEP
  dadpeep_post(destptrptr);
#endif
  
#ifdef _TRACE 
  nnza = (*destptrptr)->nonzeroes; 
  dadinfo(arity,nnzb,lhsdvectoraddress,nnza,alphaptrs,vectorptrptrs);
#endif
  
}

