#define DOUBLE
/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,         */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#include "dgaxpy_pd_macros_and_data_structures.h"

 
void dspg3(destptrptr,aaptr,aaptrptr,baptr,baptrptr,captr,captrptr,lineno,filename) 

  dvectorstruct **destptrptr, **aaptrptr,**baptrptr,**captrptr;
  double *aaptr,*baptr,*captr;

{
  int arity = 0;

#ifdef _TRACE
   int nnzb,nnza;
   dvectorstruct *lhsdvectoraddress;
#endif
 
  double *alphaptrs[3];
  dvectorstruct **rhsdvectorptrptrs[3];

#ifdef _PEEP
  adpeep_header(3);
#endif
  if (( !X_IS_ZERO(*aaptr)) && ((*aaptrptr)->nonzeroes > 0)) /* vector op */
     {
        alphaptrs[arity] = aaptr;
        rhsdvectorptrptrs[arity++] = aaptrptr;
     }
#ifdef _PEEP
  dadpeep_pre(0,aaptrptr,aaptr);
#endif

  if (( !X_IS_ZERO(*baptr)) && ((*baptrptr)->nonzeroes > 0)) /* vector op */
     {
        alphaptrs[arity] = baptr;
        rhsdvectorptrptrs[arity++] = baptrptr;
     }
#ifdef _PEEP
  dadpeep_pre(1,baptrptr,baptr);
#endif

  if (( !X_IS_ZERO(*captr)) && ((*captrptr)->nonzeroes > 0)) /* vector op */
     {
        alphaptrs[arity] = captr;
        rhsdvectorptrptrs[arity++] = captrptr;
     }
#ifdef _PEEP
  dadpeep_pre(2,captrptr,captr);
#endif

#ifdef _TRACE 
  if(*destptrptr==NULL){ 
     nnzb = 0; 
     lhsdvectoraddress= (dvectorstruct *) 99900; 
  }
  else {
     nnzb = (*destptrptr)->nonzeroes; 
     lhsdvectoraddress = *destptrptr;
  }
#endif
 
  if (arity == 0){ 
     dspzronotca(destptrptr); 
  }
  else if (arity == 1)
     dspg1(destptrptr,alphaptrs[0],rhsdvectorptrptrs[0]); 
  else if (arity == 2)
     dspg2(destptrptr,alphaptrs[0],rhsdvectorptrptrs[0],alphaptrs[1],rhsdvectorptrptrs[1]); 
  else 
     dspgx(destptrptr,arity,alphaptrs,rhsdvectorptrptrs); 
#ifdef _PEEP
  if (arity > 2){ 
     dadpeep_post(destptrptr);
  }
#endif

#ifdef _TRACE 
   nnza = (*destptrptr)->nonzeroes; 
   dadinfo(arity,nnzb,lhsdvectoraddress,nnza,alphaptrs,rhsdvectorptrptrs);
#endif

} 

 
void dspg4(destptrptr,aaptr,aaptrptr,baptr,baptrptr,captr,captrptr,daptr,daptrptr,lineno,filename) 

  dvectorstruct **destptrptr, **aaptrptr,**baptrptr,**captrptr,**daptrptr;
  double *aaptr,*baptr,*captr,*daptr;

{
  int arity = 0;

#ifdef _TRACE
   int nnzb,nnza;
   dvectorstruct *lhsdvectoraddress;
#endif
 
  double *alphaptrs[4];
  dvectorstruct **rhsdvectorptrptrs[4];

#ifdef _PEEP
  adpeep_header(4);
#endif
  if (( !X_IS_ZERO(*aaptr)) && ((*aaptrptr)->nonzeroes > 0)) /* vector op */
     {
        alphaptrs[arity] = aaptr;
        rhsdvectorptrptrs[arity++] = aaptrptr;
     }
#ifdef _PEEP
  dadpeep_pre(0,aaptrptr,aaptr);
#endif

  if (( !X_IS_ZERO(*baptr)) && ((*baptrptr)->nonzeroes > 0)) /* vector op */
     {
        alphaptrs[arity] = baptr;
        rhsdvectorptrptrs[arity++] = baptrptr;
     }
#ifdef _PEEP
  dadpeep_pre(1,baptrptr,baptr);
#endif

  if (( !X_IS_ZERO(*captr)) && ((*captrptr)->nonzeroes > 0)) /* vector op */
     {
        alphaptrs[arity] = captr;
        rhsdvectorptrptrs[arity++] = captrptr;
     }
#ifdef _PEEP
  dadpeep_pre(2,captrptr,captr);
#endif

  if (( !X_IS_ZERO(*daptr)) && ((*daptrptr)->nonzeroes > 0)) /* vector op */
     {
        alphaptrs[arity] = daptr;
        rhsdvectorptrptrs[arity++] = daptrptr;
     }
#ifdef _PEEP
  dadpeep_pre(3,daptrptr,daptr);
#endif

#ifdef _TRACE 
  if(*destptrptr==NULL){ 
     nnzb = 0; 
     lhsdvectoraddress= (dvectorstruct *) 99900; 
  }
  else {
     nnzb = (*destptrptr)->nonzeroes; 
     lhsdvectoraddress = *destptrptr;
  }
#endif
 
  if (arity == 0){ 
     dspzronotca(destptrptr); 
  }
  else if (arity == 1)
     dspg1(destptrptr,alphaptrs[0],rhsdvectorptrptrs[0]); 
  else if (arity == 2)
     dspg2(destptrptr,alphaptrs[0],rhsdvectorptrptrs[0],alphaptrs[1],rhsdvectorptrptrs[1]); 
  else 
     dspgx(destptrptr,arity,alphaptrs,rhsdvectorptrptrs); 
#ifdef _PEEP
  if (arity > 2){ 
     dadpeep_post(destptrptr);
  }
#endif

#ifdef _TRACE 
   nnza = (*destptrptr)->nonzeroes; 
   dadinfo(arity,nnzb,lhsdvectoraddress,nnza,alphaptrs,rhsdvectorptrptrs);
#endif

} 

 
void dspg5(destptrptr,aaptr,aaptrptr,baptr,baptrptr,captr,captrptr,daptr,daptrptr,eaptr,eaptrptr,lineno,filename) 

  dvectorstruct **destptrptr, **aaptrptr,**baptrptr,**captrptr,**daptrptr,**eaptrptr;
  double *aaptr,*baptr,*captr,*daptr,*eaptr;

{
  int arity = 0;

#ifdef _TRACE
   int nnzb,nnza;
   dvectorstruct *lhsdvectoraddress;
#endif
 
  double *alphaptrs[5];
  dvectorstruct **rhsdvectorptrptrs[5];

#ifdef _PEEP
  adpeep_header(5);
#endif
  if (( !X_IS_ZERO(*aaptr)) && ((*aaptrptr)->nonzeroes > 0)) /* vector op */
     {
        alphaptrs[arity] = aaptr;
        rhsdvectorptrptrs[arity++] = aaptrptr;
     }
#ifdef _PEEP
  dadpeep_pre(0,aaptrptr,aaptr);
#endif

  if (( !X_IS_ZERO(*baptr)) && ((*baptrptr)->nonzeroes > 0)) /* vector op */
     {
        alphaptrs[arity] = baptr;
        rhsdvectorptrptrs[arity++] = baptrptr;
     }
#ifdef _PEEP
  dadpeep_pre(1,baptrptr,baptr);
#endif

  if (( !X_IS_ZERO(*captr)) && ((*captrptr)->nonzeroes > 0)) /* vector op */
     {
        alphaptrs[arity] = captr;
        rhsdvectorptrptrs[arity++] = captrptr;
     }
#ifdef _PEEP
  dadpeep_pre(2,captrptr,captr);
#endif

  if (( !X_IS_ZERO(*daptr)) && ((*daptrptr)->nonzeroes > 0)) /* vector op */
     {
        alphaptrs[arity] = daptr;
        rhsdvectorptrptrs[arity++] = daptrptr;
     }
#ifdef _PEEP
  dadpeep_pre(3,daptrptr,daptr);
#endif

  if (( !X_IS_ZERO(*eaptr)) && ((*eaptrptr)->nonzeroes > 0)) /* vector op */
     {
        alphaptrs[arity] = eaptr;
        rhsdvectorptrptrs[arity++] = eaptrptr;
     }
#ifdef _PEEP
  dadpeep_pre(4,eaptrptr,eaptr);
#endif

#ifdef _TRACE 
  if(*destptrptr==NULL){ 
     nnzb = 0; 
     lhsdvectoraddress= (dvectorstruct *) 99900; 
  }
  else {
     nnzb = (*destptrptr)->nonzeroes; 
     lhsdvectoraddress = *destptrptr;
  }
#endif
 
  if (arity == 0){ 
     dspzronotca(destptrptr); 
  }
  else if (arity == 1)
     dspg1(destptrptr,alphaptrs[0],rhsdvectorptrptrs[0]); 
  else if (arity == 2)
     dspg2(destptrptr,alphaptrs[0],rhsdvectorptrptrs[0],alphaptrs[1],rhsdvectorptrptrs[1]); 
  else 
     dspgx(destptrptr,arity,alphaptrs,rhsdvectorptrptrs); 
#ifdef _PEEP
  if (arity > 2){ 
     dadpeep_post(destptrptr);
  }
#endif

#ifdef _TRACE 
   nnza = (*destptrptr)->nonzeroes; 
   dadinfo(arity,nnzb,lhsdvectoraddress,nnza,alphaptrs,rhsdvectorptrptrs);
#endif

} 
