#define DOUBLE
/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#include  <stdlib.h>
#include "dgaxpy_pd_macros_and_data_structures.h"
#include "dgaxpy_pd_global_variables.h"

/****************************************************************************
 *                                                                          *
 *  dspgx(destptrptr,arity,alphaptrs,rhsdvectorptrptrs)             *
 *                                                                          *
 *     destptrptr      is a doubly-indirected pointer to a dvectorstruct*
 *                     It is passed to dspgx by the adxpyN routines,    *
 *                     where 1 <= N <= MAXARITY.  In turn destptrptr is an  *
 *                     argument of the adxpyN functions which are called by *
 *                     ADIFOR-generated FORTRAN calls.  The purpose of the  *
 *                     double indirection is that FORTRAN calls by          *
 *                     reference, that is it puts the address of say g$x(1) *
 *                     on the stack.  In the FORTRAN code, the value of     *
 *                     g$x(1) represents a pointer to (address of) a sparse *
 *                     vector (i.e., dvectorstruct *).                  *
 *                     The FORTRAN call,                                    *
 *                     CALL adxpyN(g$x(1),...), causes inside C for an      *
 *                     internal copy, say destptrptr, to equal the address  *
 *                     of g$x(1), i.e., destptrptr == &g$x(1).  Therefore,  *
 *                     *destptrptr == g$x(1). But since the value contained *
 *                     in g$x(1) is (as interpreted by C) of type           *
 *                     (dvectorstruct *), destptrptr needs              *
 *                     one more level of                                    *
 *                     indirection, and is declared as                      *
 *                     (dvectorstruct **).                              *
 *                                                                          *
 *     arity           is the # of vectors on the rhs.                      *
 *                                                                          *
 *     alphaptrs       Array of pointers to the scalar multipliers.         *
 *                                                                          *
 *     rhsdvectorptrptrs   Array of doubly-indirected pointers to       *
 *                     dvectorstructs.  In this case, since these values*
 *                     are not changed, there is no need for the double-    *
 *                     indirection; we simply want to maintain the g$x      *
 *                     references consistant.                               *
 *                                                                          *
 *     dspgx performs the saxpy operation for all aritys.               *
 *                                                                          *
 ****************************************************************************/

void dspgx(destptrptr,arity,alphaptrs,rhsdvectorptrptrs)
dvectorstruct **destptrptr;
int arity;
double *alphaptrs[];
dvectorstruct **rhsdvectorptrptrs[];
{

  dvectorstruct *rhsdvectorptr[MAXARITY];
                                 /* Dereferenced rhsdvectorptrptrs[].   */

  dbucketstruct *rhsdbucketptr[MAXARITY];
                                      /* Array of pointers to "current"     *
				       * buckets where "current" nonzeroes  *
				       * of the rhs_vectors reside.         */

  dbucketstruct *lhsdbucketptr;
                                      /* Pointer to current bucket of       *
				       * lhsdvector.                    */

  dibucketstruct *rhsdibucketptr[MAXARITY];
                                      /* Array of pointers to "current"     *
				       * buckets where "current" nonzeroes  *
				       * of the rhs_vectors reside.         */

  dibucketstruct *lhsdibucketptr;
                                      /* Pointer to current bucket of       *
				       * lhsdvector.                    */

  double alpha[MAXARITY];   /* Dereferenced alphaptrs[].          */

  int rhsdbucketindex[MAXARITY];  /* Array of indices of current        *
				       * nonzeroes in the current buckets   *
				       * of rhs_vectors.                    */

  int lhsdbucketindex;            /* bucketindex of next nonzero to be  *
				       * added to lhsdvector.           */

  int lastlhsdbucketindex;        /* bucketindex of the latest added    *
				       * nonzero in the lhsdvector.  We *
				       * need this additional index to      *
				       * leap-frog buckets, if need be.     */

  int rhsdibucketindex[MAXARITY]; /* Array of indices of current        *
				       * nonzeroes in the current buckets   *
				       * of rhs_vectors.                    */

  int lhsdibucketindex;           /* bucketindex of next nonzero to be  *
				       * added to lhsdvector.           */

  int lastlhsdibucketindex;       /* bucketindex of the latest added    *
				       * nonzero in the lhsdvector.  We *
				       * need this additional index to      *
				       * leap-frog buckets, if need be.     */

  int firstsaxpyoperand;              /* Order in the ennumeration of the   *
				       * rhs_vectors of the first non-null  *
				       * saxpy vectorop.                    */
  int k;                              /* Counter for arity.                 */

  int j;                              /* Index for vectorsubscripts to be   *
				       * entered in (and later popped from) *
				       * the heap.                          */

  int rhsdvectornonzeroes[MAXARITY];
                                      /* Array of total # of nonzeroes in   *
				       * each rhs_vector.                   */

  int rhsdvectorconsumednonzeroes[MAXARITY];
                                      /* Array of # of nonzeroes consumed   *
				       * so far in each rhs_vector.         */

  int lhsdvectornonzeroes;        /* Count of # of nonzeroes entered    *
				       * into the lhsdvector so far. By *
				       * the end of the saxpy operation     *
				       * this index equals the total # of   *
				       * nonzeroes in the resultant vector. */

/*
 * These variables were originally for COMPRESSED_SUBSCRIPTS
 */

  dvbucketstruct *lastlhsdvbucketptr;
                                      /* Pointer to current vbucket of      *
				       * lhsdvector.                    */

  int rhsdvectornonzeroranges[MAXARITY];
                                      /* Array of total # of nonzeroes      *
				       * ranges in each rhs_vector.         */

  int rhsdvectorconsumednonzeroranges[MAXARITY];
                                      /* Array of # of nonzeroes consumed   *
				       * so far in each rhs_vector.         */

  int rhsdvbucketindex[MAXARITY]; /* Array of indices of current        *
				       * nonzeroes in the current buckets   *
				       * of rhs_vectors.                    */

  int lhsdvectornonzeroranges;    /* Count of # of nonzeroes entered    *
				       * into the lhsdvector so far. By *
				       * the end of the saxpy operation     *
				       * this index equals the total # of   *
				       * nonzeroes in the resultant vector. */

  int lhsdvbucketindex;           /* bucketindex of next nonzero to be  *
				       * added to lhsdvector.           */


  int lastlhsdvbucketindex;       /* bucketindex of the latest added    *
				       * nonzero in the lhsdvector.  We *
				       * need this additional index to      *
				       * leap-frog buckets, if need be.     */

  dvbucketstruct *templhsdvbucketptr;
  int templhsdvbucketindex;

  int j1;                             /* Index for vectorsubscripts to be   *
				       * entered in (and later popped from) *
				       * the heap.                          */

  int j2;                             /* Index for vectorsubscripts to be   *
				       * entered in (and later popped from) *
				       * the heap.                          */

  int h1;
  int h2;

  int totalcurrentnonzerorange;

#ifdef _DEBUG
  int unconsumedcurrentnonzerorange;
#endif

  int distancetooverlap;
  int distancetoendofvbucket;


  /*
   * These variables were added for union ds
   */
  
  dvectorstruct *restorerhsdvectorptr[MAXARITY+1];

  int governingdstype;
  int rhs_largest_dstype = S1;        /* Largest dstype of any              *
				       * rhsdvector ( CS > SS > S1 ).   */
  int s1_index;
  
  double temp_gaxpy;

  /*
   * These variables were added for plusequal code
   */

#ifdef _PLUSEQUAL
  int plusequalsaxpy_flag = 0;
  int plusequalkorder;
  dvectorstruct **tempdestptrptr;
  double *tempalphaptrs[MAXARITY-1];
  dvectorstruct **temprhsdvectorptrptrs[MAXARITY-1];
  double one; 
  int smallest_rhs_index;
#endif /* _PLUSEQUAL */

  /****************************************
   * Initialization of *destptrptr.       *
   ****************************************/
  
  if (*destptrptr == NULL)
    dgetvector(destptrptr);
  
  
  /****************************************
   * Dereferencing.                       *
   ****************************************/
  
  for (k=0;k<arity;k++) {
    alpha[k] = *alphaptrs[k];

    rhsdvectorptr[k] = *rhsdvectorptrptrs[k];
    restorerhsdvectorptr[k] = rhsdvectorptr[k];
    rhsdvectornonzeroes[k] = rhsdvectorptr[k]->nonzeroes;
    if (rhsdvectorptr[k]->dstype == CS)
      rhsdvectornonzeroranges[k] =
	rhsdvectorptr[k]->dds.cs.nonzeroranges;
    if (rhsdvectorptr[k]->dstype > rhs_largest_dstype)
      rhs_largest_dstype = rhsdvectorptr[k]->dstype;

#ifdef _PLUSEQUAL    
    if (rhsdvectorptr[k] == *destptrptr &&
	rhsdvectorptr[k]->dstype == CS &&
	X_IS_ONE(alpha[k])) {
      plusequalsaxpy_flag++;
      plusequalkorder = k;
    }
#endif /* _PLUSEQUAL */

  }	
  
  /****************************************
   * Determination of 1st nonzero         *
   * rhsdvector operand.              *
   ****************************************/
  
  firstsaxpyoperand = 0;

/*
 *   This bit of the code is obsolete, since the routines calling
 *   this routine already have stripped the gaxpy of zero vectors
 *   and zero multipliers.
 *

  while (nullsaxpyflag && firstsaxpyoperand<arity)
    
    if (rhsdvectornonzeroes[firstsaxpyoperand] == 0 ||
	X_IS_ZERO(alpha[firstsaxpyoperand]))
      firstsaxpyoperand++;
    else
      nullsaxpyflag = 0;
  
 *
 * If no nonzero vectorops, return a null vector.
 *

  if (nullsaxpyflag) {
    
    (*destptrptr)->nonzeroes = 0;
    (*destptrptr)->dstype = S1;

  }
  else {

*/
    
    /****************************************
     * Predetermine governingdstype and     *
     * appropriate saxpy routine.           *
     ****************************************/

  if (rhs_largest_dstype == S1) {
    s1_index = rhsdvectorptr[firstsaxpyoperand]->dds.s1.index;
    for (k=firstsaxpyoperand+1;k<arity;k++)
      if (rhsdvectorptr[k]->dds.s1.index !=
	  rhsdvectorptr[firstsaxpyoperand]->dds.s1.index &&
	  ! X_IS_ZERO(alpha[k]))
	s1_index = -1;
    if (s1_index > 0)
      governingdstype = S1;
    else
      governingdstype = SS;
  }
  else if (rhs_largest_dstype <= SS)
    governingdstype = SS;
  else /* if (rhs_largest_dstype == CS) */
    governingdstype = CS;

#ifdef _PLUSEQUAL
  smallest_rhs_index = DUMMYSUBSCRIPT;
  for (k=0;k<arity;k++) {
    if (rhsdvectorptr[k]->dstype == S1) 
      smallest_rhs_index = min (smallest_rhs_index,
				rhsdvectorptr[k]->dds.s1.index);
    else if (rhsdvectorptr[k]->dstype == SS)
      smallest_rhs_index = min (smallest_rhs_index,
        rhsdvectorptr[k]->dds.ss.bucketptr->index[0]);
    else /* if (rhsdvectorptr[k]->dstype == CS) */
      smallest_rhs_index = min (smallest_rhs_index,
	rhsdvectorptr[k]->dds.cs.ibucketptr->index_RANGE_BEGIN[0]);
  }

  if (plusequalsaxpy_flag &&
      (*destptrptr)->dstype == CS &&
      (*destptrptr)->dds.cs.ibucketptr->index_RANGE_BEGIN[0] <=
      smallest_rhs_index) {    
#ifdef _DEBUG
    if (plusequalsaxpy_flag > 1) {
      printf("dspgx warning: lhsdvector appears more than ");
      printf("once on rhs.\n");
    }
#endif
    if (arity == 1) {
      plusequaldspg1(destptrptr,alphaptrs[plusequalkorder]);
    }
    else {
      for (k=0;k<plusequalkorder;k++) {
	tempalphaptrs[k] = alphaptrs[k];
	temprhsdvectorptrptrs[k] = rhsdvectorptrptrs[k];
      }
      for (k=plusequalkorder+1;k<arity;k++) {
	tempalphaptrs[k-1] = alphaptrs[k];
	temprhsdvectorptrptrs[k-1] = rhsdvectorptrptrs[k];
      }
      tempdestptrptr = &tempddestptr;
      dspgx(tempdestptrptr,arity-1,
		tempalphaptrs,temprhsdvectorptrptrs);
      restorerhsdvectorptr[MAXARITY] = tempddestptr;
      if ((*tempdestptrptr)->dstype != governingdstype &&
	  (*tempdestptrptr)->nonzeroes !=0) {
	dconvertdstype(tempdestptrptr,MAXARITY,governingdstype);
      }
      X_PUT_REAL(one,1.0);
      plusequaldspg2(destptrptr,
		      &one,tempdestptrptr,TEMP_OTHER_VECTOR);
      tempddestptr = restorerhsdvectorptr[MAXARITY];
    }    
  }
  else /* if (not plusequalsaxpy_flag) */ {
#endif /* _PLUSEQUAL */

    for (k=0;k<arity;k++) {
      if (rhsdvectorptr[k]->dstype != governingdstype &&
	  rhsdvectorptr[k]->nonzeroes !=0 && ! X_IS_ZERO(alpha[k])) {
	dconvertdstype((&rhsdvectorptr[k]),k,governingdstype);
	if (governingdstype == CS)
	  rhsdvectornonzeroranges[k] = 
	    rhsdvectorptr[k]->dds.cs.nonzeroranges;
      }
    }
    
    if (governingdstype == S1) {
      
      /********************************************************************
       *                       ONE_SUBSCRIPT AXPY                         *
       ********************************************************************/

      lhsdvectorptr->nonzeroes = 1;
      lhsdvectorptr->dstype = S1;
      lhsdvectorptr->dds.s1.index =
	rhsdvectorptr[firstsaxpyoperand]->dds.s1.index;
      X_ZERO(temp_gaxpy);
      for (k=firstsaxpyoperand;k<arity;k++) {
	GAXPY_PLUS_EQUALS(temp_gaxpy,
			  alpha[k], rhsdvectorptr[k]->dds.s1.value);
      }
      X_ASSIGN(lhsdvectorptr->dds.s1.value,temp_gaxpy);

      daxpy_0_1_post_processing(destptrptr);

    }
    else /* if we are going to use lhsdvectorptr */ {
      if (governingdstype == SS) {
	
	/******************************************************************
	 *                  _SINGLE_SUBSCRIPTS AXPY                       *
	 ******************************************************************/
	
	/****************************************
	 * Else perform saxpy:                  *
	 * 1. Put in heap the 1st nonzero of    *
	 *      each vector.                    *
	 ****************************************/
	rhsdvectorconsumednonzeroes[firstsaxpyoperand] = 0;
	rhsdbucketindex[firstsaxpyoperand] = 0;
	rhsdbucketptr[firstsaxpyoperand] =
	  rhsdvectorptr[firstsaxpyoperand]->dds.ss.bucketptr;
	
#ifdef _DEBUG
	if (rhsdbucketptr[firstsaxpyoperand] == NULL) {
	  printf("dspgx error 1: Unexpected NULL rhsdbucketptr: V# %d\n",
		 firstsaxpyoperand);
	  abort();
	}
#endif
	
	j = rhsdbucketptr[firstsaxpyoperand]->
	  index[rhsdbucketindex[firstsaxpyoperand]];
	
	if (j <= 0) {
#ifdef _DEBUG
	  printf("dspgx error 2: Illegal index: V # %d, subscript=%d\n",
		 firstsaxpyoperand,j);
	  abort();
#endif
	}
	
	else
	  heapinitialize(firstsaxpyoperand,j,j);
	
	for (k=firstsaxpyoperand+1;k<arity;k++) {
	  if (rhsdvectornonzeroes[k] !=0 && ! X_IS_ZERO(alpha[k])) {
	    rhsdvectorconsumednonzeroes[k] = 0;
	    rhsdbucketindex[k] = 0;
	    rhsdbucketptr[k] = rhsdvectorptr[k]->dds.ss.bucketptr;
	    
#ifdef _DEBUG
	    if (rhsdbucketptr[k] == NULL) {
	      printf("dspgx error 3: Unexpected NULL rhsdbucketptr: ");
	      printf("vector # %d\n",k);
	      abort();
	    }
#endif
	    
	    j = rhsdbucketptr[k]->index[rhsdbucketindex[k]];
	    
#ifdef _DEBUG
	    if (j <= 0) {
	      printf("dspgx error 4: Illegal index: vec.=%d sub.=%d\n",
		     k,j);
	      abort();
	    }
#endif
	    
	    heapinsert(k,j,j);
	  }
	}
	
	/****************************************
	 * Perform saxpy:                       *
	 * 2. Initialize lhsdvector.             *
	 ****************************************/
	dgetbucket(&lhsdbucketptr);
	lhsdvectorptr->dstype = governingdstype;
	lhsdvectorptr->dds.ss.bucketptr = lhsdbucketptr;
	lhsdvectornonzeroes = 0;
	
	/****************************************
	 * Perform saxpy:                       *
	 * 3. Get nonzero by popping the        *
	 *    heaproot; consume it in saxpy;    *
	 *    replace / delete heaproot.        *
	 ****************************************/
	while (heaproot !=NULL) {
	  k = heaproot->vectorindex;
	  j = heaproot->rangebeginindex;
	  rhsdbucketindex[k] = MY_MOD(rhsdvectorconsumednonzeroes[k],
					  dBUCKET_DIM_minus1);
	  lhsdbucketindex = MY_MOD(lhsdvectornonzeroes,
				       dBUCKET_DIM_minus1);
	  
#ifdef _DEBUG
	  if (rhsdbucketptr[k]->errorflag != 9999) {
	    printf("dspgx err 5: mem. corruption of dbucketstore\n");
	    abort();
	  }
#endif      
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.1. Found 1st or new subscript for  *
	   *      lhsdvector.  If needed, get new  *
	   *      bucket for lhsdvector.           *
	   *      Consume nonzero in saxpy.       *
	   ****************************************/
	  if (lhsdvectornonzeroes == 0 ||
	      lhsdbucketptr->index[lastlhsdbucketindex] < j) {
	    if (lhsdbucketindex == 0 && lhsdvectornonzeroes != 0) {
	      dgetbucket(&(lhsdbucketptr->nextbucket));
	      lhsdbucketptr = lhsdbucketptr->nextbucket;
	    }
	    lhsdbucketptr->index[lhsdbucketindex] = j;
	    GAXPY_EQUALS(lhsdbucketptr->value[lhsdbucketindex],
			 alpha[k],
			 rhsdbucketptr[k]->value[rhsdbucketindex[k]]);
	    lastlhsdbucketindex = lhsdbucketindex;
	    lhsdvectornonzeroes++;
	  }
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.2. Found old subscript for         *
	   *      lhsdvector.                      *
	   *      Consume nonzero in saxpy.       *
	   ****************************************/
	  else if (lhsdbucketptr->index[lastlhsdbucketindex] == j) {
	    GAXPY_PLUS_EQUALS(
	      lhsdbucketptr->value[lastlhsdbucketindex],
	      alpha[k],
	      rhsdbucketptr[k]->value[rhsdbucketindex[k]]);
	  }
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.3. Found wrong subscript for       *
	   *      lhsdvector.  Abort.               *
	   ****************************************/
	  
#ifdef _DEBUG
	  else {
	    printf("dspgx error 6: lhsdnz=%d, lastlhsdindex[]=%d, j=%d\n",
		   lhsdvectornonzeroes,
		   lhsdbucketptr->index[lastlhsdbucketindex],j);
	    abort();
	  }
#endif
	  
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.4. If there is one, get new        *
	   *      nonzero from the rhs_vector     *
	   *      whose nonzero was just          *
	   *      consumed and replace the        *
	   *      heaproot with new entry;        *
	   *      Else, delete heaproot.          *
	   ****************************************/
	  if (++rhsdvectorconsumednonzeroes[k] < rhsdvectornonzeroes[k]) {
	    if ((rhsdbucketindex[k] =
		 MY_MOD(rhsdvectorconsumednonzeroes[k],
			dBUCKET_DIM_minus1)) == 0) {
	      if ((rhsdbucketptr[k] = rhsdbucketptr[k]->nextbucket) == NULL) {
		
#ifdef _DEBUG
		printf("dspgx err 7: unexpected NULL nextbucket: k=%d",
		       k);
		printf("rhsdvectorconsumednonzeros=%d, rhsdvecnonzeroes=%d,",
		       rhsdvectorconsumednonzeroes[k],rhsdvectornonzeroes[k]);
		abort();
#endif
		
	      }
	    }     
	    j = rhsdbucketptr[k]->index[rhsdbucketindex[k]];
	    
	    if (j <= 0) {
#ifdef _DEBUG
	      printf("dspgx err 8: Illegal index: V=%d, subscript=%d\n"
		     ,k,j);
	      abort();
#endif
	      
	    }
	    
	    else
	      heapop(REPLACE,k,j,j);
	  }
	  else
	    heapop(DELETE,DUMMYVECTOR,DUMMYSUBSCRIPT,DUMMYSUBSCRIPT);
	  
	} /*matches     while (heaproot !=NULL) {     */
      } /*matches     else if (governingdstype == SS) {     */
      
      else /* if (governingdstype == CS) */ {
	
	/******************************************************************
	 *                 _COMPRESSED_SUBSCRIPTS AXPY                    *
	 ******************************************************************/
	
	
	/****************************************
	 * Else perform saxpy:                  *
	 * 1. Put in heap the 1st nonzero of    *
	 *    each vector.                      *
	 ****************************************/
	rhsdvectorconsumednonzeroranges[firstsaxpyoperand] = 0;
	rhsdibucketindex[firstsaxpyoperand] = 0;
	rhsdvbucketindex[firstsaxpyoperand] = 0;
	rhsdibucketptr[firstsaxpyoperand] =
	  rhsdvectorptr[firstsaxpyoperand]->dds.cs.ibucketptr;
	rhsdvbucketptr[firstsaxpyoperand] =
	  rhsdvectorptr[firstsaxpyoperand]->dds.cs.vbucketptr;
	
#ifdef _DEBUG
	if (rhsdibucketptr[firstsaxpyoperand] == NULL ||
	    rhsdvbucketptr[firstsaxpyoperand] == NULL) {
	  printf("dspgx error 10: Unexpected NULL rhsdbucketptrs: ");
	  printf("vector # %d\n",firstsaxpyoperand);
	  abort();
	}
#endif
	
	j1 = rhsdibucketptr[firstsaxpyoperand]->
	  index_RANGE_BEGIN[rhsdibucketindex[firstsaxpyoperand]];
	j2 = rhsdibucketptr[firstsaxpyoperand]->
	  index_RANGE_END[rhsdibucketindex[firstsaxpyoperand]];
	
	if (j1 <= 0 || j2 <= 0 || j2 < j1) {
#ifdef _DEBUG
	  printf("dspgx error 11: Nonsensical Nonzero Range; \n\t\t");
	  printf("vector # %d, subscript 1=%d subscript 2=%d\n",
		 firstsaxpyoperand,j1,j2);
	  abort();
#endif
	}
	
	else
	  heapinitialize(firstsaxpyoperand,j1,j2);
	
	for (k=firstsaxpyoperand+1;k<arity;k++) {

	  /*
	   * rhsdvectornonzeroes[k] !=0 will be redundant and
	   *                           unnecessary most of the time
	   *                           but in the code preceding the
	   *                           addition of += module it was
	   *                           needed, so we leave it.
	   */
	
	  if (rhsdvectornonzeroes[k] !=0 &&
	      rhsdvectornonzeroranges[k] !=0 &&
	      ! X_IS_ZERO(alpha[k])) {
	    rhsdvectorconsumednonzeroranges[k] = 0;
	    rhsdibucketindex[k] = 0;
	    rhsdibucketptr[k] =
	      rhsdvectorptr[k]->dds.cs.ibucketptr;
	    rhsdvbucketindex[k] = 0;
	    rhsdvbucketptr[k] =
	      rhsdvectorptr[k]->dds.cs.vbucketptr;
	    
#ifdef _DEBUG
	    if (rhsdibucketptr[k] == NULL ||
		rhsdvbucketptr[k] == NULL) {
	      printf("dspgx error 12: Unexpected ");
	      printf("NULL rhsdibucketptr or rhsdvbucketptr: vector # %d\n",k);
	      abort();
	    }
#endif
	    
	    j1 = rhsdibucketptr[k]->index_RANGE_BEGIN[rhsdibucketindex[k]];
	    j2 = rhsdibucketptr[k]->index_RANGE_END[rhsdibucketindex[k]];
	    
#ifdef _DEBUG
	    if (j1 <= 0 || j2 <= 0 || j2 < j1) {
	      printf("dspgx err 13: Nonsensical Nonzero Range; \n\t\t");
	      printf("vector # %d, subscript 1=%d subscript 2=%d\n",
		     firstsaxpyoperand,j1,j2);
	      abort();
	    }
#endif
	    
	    heapinsert(k,j1,j2);
	  }
	}
	
	/****************************************
	 * Perform saxpy:                       *
	 * 2. Initialize lhsdvector.             *
	 ****************************************/
	dgetibucket(&lhsdibucketptr);
	lhsdvectorptr->dstype = governingdstype;
	lhsdvectorptr->dds.cs.ibucketptr = lhsdibucketptr;
	dgetvbucket(&lhsdvbucketptr);
	lhsdvectorptr->dds.cs.vbucketptr = lhsdvbucketptr;
	lhsdvectornonzeroranges = 0;
	lhsdvectornonzeroes = 0;
	lhsdvbucketindex = 0;
	
	/****************************************
	 * Perform saxpy:                       *
	 * 3. Get nonzero by popping the        *
	 *    heaproot; consume it in saxpy;    *
	 *    either replace or delete heaproot.*
	 ****************************************/

	while (heaproot !=NULL) {
	  k = heaproot->vectorindex;
	  j1 = heaproot->rangebeginindex;
	  j2 = heaproot->rangeendindex;
	  rhsdibucketindex[k] =
	    MY_MOD(rhsdvectorconsumednonzeroranges[k],
		   dIBUCKET_DIM_minus1);
	  lhsdibucketindex = MY_MOD(lhsdvectornonzeroranges,
					dIBUCKET_DIM_minus1);
	  if ((rhsdvbucketindex[k] =
	       MY_MOD(rhsdvbucketindex[k],dVBUCKET_DIM_minus1))==0
	      && rhsdvectorconsumednonzeroranges[k] !=0) {
	    dHOP_TO_NEXT_RHSVBUCKET(k,&rhsdvbucketindex[k],
				   rhsdvectorconsumednonzeroranges[k],
				   rhsdvectornonzeroranges[k]);
	  }
	  lhsdvbucketindex = MY_MOD(lhsdvbucketindex,
					dVBUCKET_DIM_minus1);
	  
	  
	  
#ifdef _DEBUG
	  if (rhsdibucketptr[k]->errorflag != 9099 ||
	      rhsdvbucketptr[k]->errorflag != 9009) {
	    printf("dspgx err 14: mem corruption of dibucketstore ");
	    printf("or dvbucketstore.\n");
	    abort();
	  }
#endif      
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.0. Check for startup errors.       *
	   ****************************************/
	  
#ifdef _DEBUG
	  if (lhsdvectornonzeroranges == 0) {
	    if (rhsdibucketindex[k] != 0 || rhsdvbucketindex[k] != 0 ||
		lhsdibucketindex != 0 || lhsdvbucketindex != 0) {
	      printf("dspgx error 15: Zeroes expected: \n\t");
	      printf("rhsdibucketindex[k]=%d, rhsdvbucketindex[k]=%d\n\t",
		     rhsdibucketindex[k], rhsdvbucketindex[k]);
	      printf("lhsdibucketindex=%d, lhsdvbucketindex=%d\n\t",
		     lhsdibucketindex,lhsdvbucketindex);
	      abort();
	    }
	  }
#endif      
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.1. Found 1ST or NEW subscript      *
	   *      range (non-overlapping with     *
	   *      previous one, if any) for       *
	   *      lhsdvector.  If needed, get new  *
	   *      ibuckets and vbuckets for        *
	   *      lhsdvector.                      *
	   ****************************************/
	  if (lhsdvectornonzeroranges == 0 ||
	      j1 > lhsdibucketptr->index_RANGE_END[lastlhsdibucketindex]+1) {/*A.1*/
	    
	    if (lhsdibucketindex == 0 && 
		lhsdvectornonzeroranges != 0) {
	      dgetibucket(&(lhsdibucketptr->nextibucket));
	      lhsdibucketptr = lhsdibucketptr->nextibucket;
	    }
	    lhsdibucketptr->index_RANGE_BEGIN[lhsdibucketindex] = j1;
	    lhsdibucketptr->index_RANGE_END[lhsdibucketindex] = j2;
	    
	    if (lhsdvbucketindex == 0 &&
		lhsdvectornonzeroranges != 0)
	      dget_and_hop_to_next_vbucket(&lhsdvbucketindex);
	    
	    lastlhsdvbucketptr = lhsdvbucketptr;
	    lastlhsdvbucketindex = lhsdvbucketindex;
	    
	    totalcurrentnonzerorange = j2-j1+1;
	    
	    lhsdvectornonzeroes += totalcurrentnonzerorange;
	    lhsdvectornonzeroranges++;
	    
	    
	    dNON_OVER_LAPPING_CS_RANGE_AXPY(COMPLETELY_NEW_RANGE,
					   lhsdibucketindex,
					   &lastlhsdibucketindex,
					   &lhsdvbucketindex,
					   k,
					   &rhsdvbucketindex[k],
					   alpha[k],
					   totalcurrentnonzerorange,
					   &unconsumedcurrentnonzerorange,
					   rhsdvectorconsumednonzeroranges[k],
					   rhsdvectornonzeroranges[k],
					   lhsdvectornonzeroranges);
	    

	  } /*A.1*/
	  
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.2. Found old subscript for         *
	   *      lhsdvector.  Consume  nonzero in *
	   *      saxpy.                          *
	   *                                      *
	   ****************************************/
	  else if (j1 <= 
		   lhsdibucketptr->index_RANGE_END[lastlhsdibucketindex]+1 &&
		   j1 >= 
		   lhsdibucketptr->index_RANGE_BEGIN[lastlhsdibucketindex]) {
	    /*B.1*/
	    
	    templhsdvbucketptr = lhsdvbucketptr;
	    templhsdvbucketindex = lhsdvbucketindex;
	    lhsdvbucketptr = lastlhsdvbucketptr;
	    
	    distancetooverlap =
	      j1 - lhsdibucketptr->index_RANGE_BEGIN[lastlhsdibucketindex];
	    distancetoendofvbucket = 
	      dVBUCKET_DIM - lastlhsdvbucketindex;
	    
	    if (j1 == lhsdibucketptr->index_RANGE_END[lastlhsdibucketindex]+1) {
	      h1 = 0;
	      h2 = -1;
	    }
	    else if (j2 <= lhsdibucketptr->index_RANGE_END[lastlhsdibucketindex]) {
	      h1 = j1;
	      h2 = j2;
	      j1 = 0;
	      j2 = -1;
	    }
	    else {
	      h1 = j1;
	      h2 = lhsdibucketptr->index_RANGE_END[lastlhsdibucketindex];
	      j1 = h2+1;
	    }
	    
	    if (distancetoendofvbucket > distancetooverlap) {
	      lhsdvbucketindex =
		MY_MOD(lastlhsdvbucketindex + distancetooverlap,
		       dVBUCKET_DIM_minus1);
	    }
	    else if (distancetoendofvbucket < distancetooverlap) {
	      lhsdvbucketindex =
		MY_MOD(distancetooverlap - distancetoendofvbucket,
		       dVBUCKET_DIM_minus1);
	    }
	    else {
	      lhsdvbucketindex = 0;
	    }
	    
	    while (distancetooverlap >= distancetoendofvbucket &&
		   lhsdvbucketptr->nextvbucket != NULL) {
	      lhsdvbucketptr = lhsdvbucketptr->nextvbucket;
	      distancetooverlap -= distancetoendofvbucket;
	      distancetoendofvbucket = dVBUCKET_DIM;
	    }
	    
	    lhsdibucketptr->index_RANGE_END[lastlhsdibucketindex] =
	      max(j2, lhsdibucketptr->index_RANGE_END[lastlhsdibucketindex]);
	    
	    if ((totalcurrentnonzerorange = h2-h1+1) > 0) {
	      
	      
	      dOVERLAPPING_CS_RANGE_AXPY(&lhsdvbucketindex,
					k,
					&rhsdvbucketindex[k],
					alpha[k],
					totalcurrentnonzerorange,
					&unconsumedcurrentnonzerorange,
					rhsdvectorconsumednonzeroranges[k],
					rhsdvectornonzeroranges[k],
					lhsdvectornonzeroranges);
	
	    }
	      
	    /*
	     * This bit of logic must be performed, whether or not there
	     * is additional nonzeroes to process.
	     */
	    
	    if (j2 != -1 &&
		(lhsdvbucketptr != templhsdvbucketptr ||
		 (lhsdvbucketindex = 
		  MY_MOD(lhsdvbucketindex,dVBUCKET_DIM_minus1)) !=
		 templhsdvbucketindex)) {
	      
#ifdef _DEBUG
	      printf("dspgx err 50.0: Unexpectd value discrepancies\n");
	      printf("lhsdvbucketptr=%d, templhsdvbucketptr=%d\n",
		     (int)lhsdvbucketptr,(int)templhsdvbucketptr);
	      printf("lhsdvbucketindex=%d, templhsdvbucketindex=%d\n",
		     lhsdvbucketindex,templhsdvbucketindex);
	      
	      abort();
#endif
	    }
	    else {
	      lhsdvbucketptr = templhsdvbucketptr;
	      lhsdvbucketindex = templhsdvbucketindex;
	    }
	    
	    
	    if ((totalcurrentnonzerorange = j2-j1+1) > 0) {
	      lhsdvectornonzeroes += totalcurrentnonzerorange;
	      
	      /*
	       * adding new entries to vbucket
	       */
	      
	      if (lhsdvbucketindex == 0 &&
		  lhsdvectornonzeroranges != 0 &&
		  j2 != -1)
		dget_and_hop_to_next_vbucket(&lhsdvbucketindex);
	      
	      if(rhsdvbucketindex[k] == dVBUCKET_DIM && j2 != -1) {
		
		dHOP_TO_NEXT_RHSVBUCKET(k,&rhsdvbucketindex[k],
				       rhsdvectorconsumednonzeroranges[k],
				       rhsdvectornonzeroranges[k]);
	      }

	      dNON_OVER_LAPPING_CS_RANGE_AXPY(NEW_ENTRIES_PADDED_TO_OLD_RANGE,
					     lhsdibucketindex,
					     &lastlhsdibucketindex,
					     &lhsdvbucketindex,
					     k,
					     &rhsdvbucketindex[k],
					     alpha[k],
					     totalcurrentnonzerorange,
					     &unconsumedcurrentnonzerorange,
					     rhsdvectorconsumednonzeroranges[k],
					     rhsdvectornonzeroranges[k],
					     lhsdvectornonzeroranges);
	      
	    } /*  corresponds to the if ((totalcurrent...) { */
	    
	  } /*B.1*/
	  
	  
	  /****************************************	
	   * Perform saxpy:                       *
	   * 3.3. Found wrong subscript for       *
	   *      lhsdvector.  Abort.               *
	   ****************************************/
	  
#ifdef _DEBUG
	  else {
	    printf("dspgx error 60: lhsdvectornonzeroranges=%d,",
		   lhsdvectornonzeroranges);
	    printf("lastlhsdibucketindex=%d, \n\t",
		   lastlhsdibucketindex);
	    printf("lhsdibucketptr->index[%d] = (%d,%d), j1=%d\n",
		   lastlhsdibucketindex,
		   lhsdibucketptr->index_RANGE_BEGIN[lastlhsdibucketindex],
		   lhsdibucketptr->index_RANGE_END[lastlhsdibucketindex],j1);
	    abort();
	  }
#endif
	  
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.4. If there is one, get new        *
	   *      nonzero from the rhs_vector     *
	   *      whose nonzero was just          *
	   *      consumed and replace the        *
	   *      heaproot with new entry;        *
	   *      Else, delete heaproot.          *
	   ****************************************/
	  if (++rhsdvectorconsumednonzeroranges[k] <
	      rhsdvectornonzeroranges[k]) {
	    if ((rhsdibucketindex[k] =
		 MY_MOD(rhsdvectorconsumednonzeroranges[k],
			dIBUCKET_DIM_minus1)) == 0) {
	      if ((rhsdibucketptr[k] = rhsdibucketptr[k]->nextibucket) ==
		  NULL) {
		
#ifdef _DEBUG
		printf("dspgx err 61: unexpected NULL nextvbucket: \n");
		printf("\t k=%d, rhsdvectorconsumednonzeroranges=%d,\n",
		       k,rhsdvectorconsumednonzeroranges[k]);
		printf("\t rhsdvectornonzeroranges[k]=%d\n",
		       rhsdvectornonzeroranges[k]);
		abort();
#endif
		
	      }
	    }     
	    j1 = rhsdibucketptr[k]->index_RANGE_BEGIN[rhsdibucketindex[k]];
	    j2 = rhsdibucketptr[k]->index_RANGE_END[rhsdibucketindex[k]];
	    
	    if (j1 <= 0 || j2 <= 0) {
#ifdef _DEBUG
	      printf("dspgx error 62: Illegal index: vector=%d, ",k);
	      printf("subscript1=%d, subscript2=%d \n",j1,j2);
	      abort();
#endif
	    }
	    
	    else
	      heapop(REPLACE,k,j1,j2);
	  }
	  else
	    heapop(DELETE,DUMMYVECTOR,DUMMYSUBSCRIPT,DUMMYSUBSCRIPT);
	} /*matches     while (heaproot !=NULL) {     */
      } /* matches  else ** if (governingdstype == CS) ** {     */	
      
      daxpy_2_heap_post_processing(destptrptr,
				  rhsdvectorptrptrs,
				  alphaptrs,
				  restorerhsdvectorptr,
				  arity,
				  lhsdvectornonzeroes,
				  lhsdvectornonzeroranges,
				  governingdstype);
      
      
    } /* matches    else ** if we are going to use lhsdvectorptr ** {  */

#ifdef _PLUSEQUAL
  } /* matches    else ** if (not plusequalsaxpy_flag) ** {    */
#endif /* _PLUSEQUAL */

  /* }  matches    else ** if (not nullsaxpyflag) ** {    */

}

