#define DOUBLE
/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#include  <stdlib.h>
#include "dgaxpy_pd_macros_and_data_structures.h"
#include "dgaxpy_pd_global_variables.h"

/****************************************************************************
 *                                                                          *
 *    plusequaldspg2                                                       *
 *                                                                          *
 ****************************************************************************/

void plusequaldspg2(rhsanddestptrptr,
		     otheralphaptr,
                     otherrhsdvectorptrptr,
                     status_otherrhsdvector)
dvectorstruct **rhsanddestptrptr;
double *otheralphaptr;
dvectorstruct **otherrhsdvectorptrptr;
int status_otherrhsdvector;
{
  double *alphaptrs[2];
  
  int arity = 2;
  
  dvectorstruct *rhsdvectorptr[2];
  dvectorstruct *tempvectorptr;
  
  int i;
  

  dibucketstruct *rhsdibucketptr[2];
                                      /* Array of pointers to "current"     *
				       * buckets where "current" nonzeroes  *
				       * of the rhs_vectors reside.         */

  dibucketstruct *lhsdibucketptr;
                                      /* Pointer to current bucket of       *
				       * lhsdvector.                    */

  double alpha[2];          /* Dereferenced alphaptrs[].          */

  int rhsdibucketindex[2];        /* Array of indices of current        *
				       * nonzeroes in the current buckets   *
				       * of rhs_vectors.                    */

  int lhsdibucketindex;           /* bucketindex of next nonzero to be  *
				       * added to lhsdvector.           */

  int lastlhsdibucketindex;       /* bucketindex of the latest added    *
				       * nonzero in the lhsdvector.  We *
                                       * need this additional index to      *
				       * leap-frog buckets, if need be.     */

  int k;                              /* Counter for arity.                 */

  int lhsdvectornonzeroes;        /* Count of # of nonzeroes entered    *
				       * into the lhsdvector so far. By *
				       * the end of the saxpy operation,    *
				       * this index equals the total # of   *
				       * nonzeroes in the resultant vector. */

/*
 * These variables were originally for COMPRESSED_SUBSCRIPTS
 */

  dvbucketstruct *lastlhsdvbucketptr;  
                                      /* Pointer to current vbucket of      *
				       * lhsdvector.                    */

  int rhsdvectornonzeroranges[2];
                                      /* Array of total # of nonzeroes      *
				       * ranges in each rhs_vector.         */

  int rhsdvectorconsumednonzeroranges[2];
                                      /* Array of # of nonzeroes consumed   *
				       * so far in each rhs_vector.         */

  int rhsdvbucketindex[2];        /* Array of indices of current        *
				       * nonzeroes in the current buckets   *
				       * of rhs_vectors.                    */

  int lhsdvectornonzeroranges;    /* Count of # of nonzeroes entered    *
				       * into the lhsdvector so far. By *
				       * the end of the saxpy operation,    *
				       * this index equals the total # of   *
				       * nonzeroes in the resultant vector. */

  int lhsdvbucketindex;           /* bucketindex of next nonzero to be  *
				       * added to lhsdvector.           */


  int lastlhsdvbucketindex;       /* bucketindex of the latest added    *
				       * nonzero in the lhsdvector.  We *
				       * need this additional index to frog *
				       * leap-frog buckets, if need be.     */

  dvbucketstruct *templhsdvbucketptr;

  int templhsdvbucketindex;

  int j1;                             /* Index for vectorsubscripts to be   *
				       * entered in (and later popped from) *
				       * the heap.                          */

  int j2;                             /* Index for vectorsubscripts to be   *
				       * entered in (and later popped from) *
				       * the heap.                          */

  int h1;
  int h2;

  int totalcurrentnonzerorange;

#ifdef _DEBUG
  int unconsumedcurrentnonzerorange;
#endif

  int distancetooverlap;
  int distancetoendofvbucket;
  int hopdistance;

  /*
   * These variables were added for union ds
   */
  
  dvectorstruct *restorerhsdvectorptr[2+1];

  int governingdstype;
  
  /*
   * These variables were added for plusequal code
   */
  
  dvectorstruct **rhsdvectorptrptrs[MAXARITY];

  int condition_1, condition_2;
  int lastrhsdibucketindex[2];
  int lastrhsdvbucketindex[2];

  dibucketstruct *lastrhsdibucketptr[2];
  dvbucketstruct *lastrhsdvbucketptr[2];
  dvbucketstruct *catchupto_rhsvbucketptr[2];

  dibucketstruct *temprhsdibucketptr[2];
  dvbucketstruct *temprhsdvbucketptr[2];
  int temprhsdvbucketindex[2];

  dibucketstruct *restofplusequalibucketptr;
  dvbucketstruct *restofplusequalvbucketptr;
  int restofplusequalibucketindex;
  int restofplusequalvbucketindex;
  int restofplusequalvectorconsumednonzeroranges;

#ifdef _TRACE
  int nnzb,nnza;
  dvectorstruct *lhsdvectoraddress;
  
  dvectorstruct **vectorptrptrs[2];
  
  alphaptrs[0] = alphaptr;
  vectorptrptrs[0] = vectoraptrptr;
  
  alphaptrs[1] = betaptr;
  vectorptrptrs[1] = vectorbptrptr;
  
  if(*destptrptr==NULL){ 
    nnzb = 0; 
    lhsdvectoraddress= (dvectorstruct *) 99900; 
  }
  else {
    nnzb = (*destptrptr)->nonzeroes; 
    lhsdvectoraddress = *destptrptr;
  }
#endif
  
  /****************************************
   * Dereferencing.                       *
   ****************************************/
  
  X_PUT_REAL(alpha[0],1.0);

#ifdef _DEBUG
  if (*rhsanddestptrptr == NULL) {
    printf("plusequaldspg2 error #1: Non_initialized lhsdvector.\n");
    abort();
  }
#endif

  rhsdvectorptr[0] = *rhsanddestptrptr;
  rhsdvectornonzeroranges[0] =
    rhsdvectorptr[0]->dds.cs.nonzeroranges;

  X_ASSIGN(alpha[1],(*otheralphaptr));

  alphaptrs[0] = &alpha[0];
  alphaptrs[1] = otheralphaptr;

  rhsdvectorptr[1] = *otherrhsdvectorptrptr;
  rhsdvectornonzeroranges[1] =
    rhsdvectorptr[1]->dds.cs.nonzeroranges;

  governingdstype = CS;
  k = 1;   
  restorerhsdvectorptr[k] = rhsdvectorptr[k];

  if (rhsdvectorptr[k]->dstype != governingdstype) {
    dconvertdstype((&rhsdvectorptr[k]),k,governingdstype);
    rhsdvectornonzeroranges[k] =
      rhsdvectorptr[k]->dds.cs.nonzeroranges;
  }  
 
  /******************************************************************
   *               _COMPRESSED_SUBSCRIPTS AXPY ONLY                 *
   ******************************************************************/
  
  /****************************************
   * perform saxpy:                       *
   * 1. Put in heap the 1st nonzero of    *
   *    each vector.                      *
   ****************************************/
  for (k=0;k<arity;k++) {
    rhsdvectorconsumednonzeroranges[k] = 0;
    rhsdibucketindex[k] = 0;
    rhsdvbucketindex[k] = 0;
    rhsdibucketptr[k] =
      rhsdvectorptr[k]->dds.cs.ibucketptr;
    rhsdvbucketptr[k] =
      rhsdvectorptr[k]->dds.cs.vbucketptr;
    
#ifdef _DEBUG
    if (rhsdibucketptr[k] == NULL ||
	rhsdvbucketptr[k] == NULL) {
	      printf("plusequaldspg2 err 10: Unexpected NULL rhsdbucketptr: ");
	      printf("vector # %d\n",k);
	      abort();
	    }
#endif
  }	  

  
  /****************************************
   * Perform saxpy:                       *
   * 2. Initialize lhsdvector.             *
   ****************************************/

  lhsdvectornonzeroranges = 0;
  lhsdvectornonzeroes = 0;
  lhsdvbucketindex = 0;


  /****************************************
   * Perform saxpy:                       *
   * 3. Get nonzero by popping the        *
   *    heaproot; consume it in saxpy;    *
   *    either replace or delete heaproot.*
   ****************************************/
  

  /* while (there are nonzeroranges to be processed AND                      *
   *         (condition_1 OR condition_2 ), where:                           *
   * condition_1 == TRUE, indicates that the range to be processed belongs   *
   *                to the plusequalvector,                                  *
   * condition_2 == TRUE, indicates that the range to be processed belongs   *
   *                to the otherrhsdvector, and that this range is       *
   *                entirelyencapsulated in the last processed range of the  *
   *                plusequalvector.                                         */

  lastrhsdibucketindex[0] = -1;
  lastrhsdibucketptr[0] = NULL;
  
  while ((rhsdvectorconsumednonzeroranges[0] <
	  rhsdvectornonzeroranges[0] &&
	  rhsdvectorconsumednonzeroranges[1] <
	  rhsdvectornonzeroranges[1]) &&

	 ((condition_1 = (rhsdibucketptr[0]->
			  index_RANGE_BEGIN[rhsdibucketindex[0]] <=
			  rhsdibucketptr[1]->
			  index_RANGE_BEGIN[rhsdibucketindex[1]]))    ||
	  
	  (condition_2 = ((lastrhsdibucketindex[0] >= 0) &&
			  (rhsdibucketptr[1]->
			   index_RANGE_BEGIN[rhsdibucketindex[1]] >=
			   lastrhsdibucketptr[0]->
			   index_RANGE_BEGIN[lastrhsdibucketindex[0]] &&
			   rhsdibucketptr[1]->
			   index_RANGE_END[rhsdibucketindex[1]] <=
			   lastrhsdibucketptr[0]->
			   index_RANGE_END[lastrhsdibucketindex[0]]))))) {

    if (condition_1) {
      k = 0;
      lastrhsdibucketindex[0] = rhsdibucketindex[0];
      lastrhsdvbucketindex[0] = rhsdvbucketindex[0];
      lastrhsdibucketptr[0] = rhsdibucketptr[0];
      lastrhsdvbucketptr[0] = rhsdvbucketptr[0];
      totalcurrentnonzerorange =
	rhsdibucketptr[0]->index_RANGE_END[rhsdibucketindex[0]] -
	  rhsdibucketptr[0]->index_RANGE_BEGIN[rhsdibucketindex[0]]+1;

#ifdef _DEBUG
      unconsumedcurrentnonzerorange = totalcurrentnonzerorange;
#endif

      hopdistance = totalcurrentnonzerorange;
      distancetoendofvbucket = dVBUCKET_DIM - rhsdvbucketindex[0];
   
      if (distancetoendofvbucket > hopdistance) {
	rhsdvbucketindex[0] = MY_MOD(rhsdvbucketindex[0] + hopdistance,
					 dVBUCKET_DIM_minus1);
      }
      else if (distancetoendofvbucket < hopdistance) {
	rhsdvbucketindex[0] = MY_MOD(hopdistance - distancetoendofvbucket,
					 dVBUCKET_DIM_minus1);
      }
      else {
	rhsdvbucketindex[0] = 0;
      }
      while (hopdistance >= distancetoendofvbucket &&
	     rhsdvbucketptr[0]->nextvbucket != NULL) {
	rhsdvbucketptr[0] = rhsdvbucketptr[0]->nextvbucket;
	hopdistance -= distancetoendofvbucket;
	distancetoendofvbucket = dVBUCKET_DIM;
      }

      lhsdvectornonzeroes += totalcurrentnonzerorange;
    }
     
    else if (condition_2) {
      k = 1;
     temprhsdibucketptr[0] = rhsdibucketptr[0];
      temprhsdvbucketptr[0] = rhsdvbucketptr[0];
      rhsdibucketptr[0] = lastrhsdibucketptr[0];
      rhsdvbucketptr[0] = lastrhsdvbucketptr[0];

      /* The reason rhsdvbucketindex[0] is not
       * set to lastrhsdvbucketindex[0] is because
       * rhsdvbucketindex[0] is computed below from
       * lastrhsdvbucketindex[0].
       */

      temprhsdvbucketindex[0] = rhsdvbucketindex[0];
       
      distancetooverlap =
	rhsdibucketptr[1]->index_RANGE_BEGIN[rhsdibucketindex[1]] -
	rhsdibucketptr[0]->index_RANGE_BEGIN[lastrhsdibucketindex[0]];
      distancetoendofvbucket = dVBUCKET_DIM - lastrhsdvbucketindex[0];
      
      if (distancetoendofvbucket > distancetooverlap) {
	rhsdvbucketindex[0] =
	  MY_MOD(lastrhsdvbucketindex[0] + distancetooverlap,
		 dVBUCKET_DIM_minus1);
      }
      else if (distancetoendofvbucket < distancetooverlap) {
	rhsdvbucketindex[0] =
	  MY_MOD(distancetooverlap - distancetoendofvbucket,
		 dVBUCKET_DIM_minus1);
      }
      else {
	rhsdvbucketindex[0] = 0;
      }
      
      while (distancetooverlap >= distancetoendofvbucket &&
	     rhsdvbucketptr[0]->nextvbucket != NULL) {
	rhsdvbucketptr[0] = rhsdvbucketptr[0]->nextvbucket;
	distancetooverlap -= distancetoendofvbucket;
	distancetoendofvbucket = dVBUCKET_DIM;
      }

      totalcurrentnonzerorange = 
      	rhsdibucketptr[1]->index_RANGE_END[rhsdibucketindex[1]] -
	  rhsdibucketptr[1]->index_RANGE_BEGIN[rhsdibucketindex[1]]+1;

      /* Wrapping the call with the two assignments has the effect of
       * EQUIVALENCEing the data structures being pointed two by the
       * two pointers.
       */
    
      lhsdvbucketptr = rhsdvbucketptr[0];      
      dOVERLAPPING_CS_RANGE_AXPY(&rhsdvbucketindex[0],
				1,
				&rhsdvbucketindex[1],
				alpha[1],
				totalcurrentnonzerorange,
				&unconsumedcurrentnonzerorange,
				rhsdvectorconsumednonzeroranges[1],
				rhsdvectornonzeroranges[1],
				rhsdvectornonzeroranges[0]);
      rhsdvbucketptr[0] = lhsdvbucketptr;
      
      rhsdibucketptr[0] = temprhsdibucketptr[0];
      rhsdvbucketptr[0] = temprhsdvbucketptr[0];
      rhsdvbucketindex[0] = temprhsdvbucketindex[0];
      
    } /* matches     else if (condition_2) {  */
  
    /*****************************************
     * Perform saxpy:                        *
     * 3.4. If there is one, get new         *
     *      nonzero from the rhs_vector      *
     *      whose nonzero was just consumed. *
     *****************************************/
    if (++rhsdvectorconsumednonzeroranges[k] <
	rhsdvectornonzeroranges[k]) {
      if ((rhsdibucketindex[k] =
	   MY_MOD(rhsdvectorconsumednonzeroranges[k],
		  dIBUCKET_DIM_minus1)) == 0) {
	if ((rhsdibucketptr[k] = rhsdibucketptr[k]->nextibucket) ==
	    NULL) {
	  
#ifdef _DEBUG
	  printf("plusqualdspg2 error #36: unexpected NULL nextvbucket: \n");
	  printf("\t k=%d, rhsdvectorconsumednonzeroranges=%d,\n",
		 k,rhsdvectorconsumednonzeroranges[k]);
	  printf("\t rhsdvectornonzeroranges[k]=%d\n",
		 rhsdvectornonzeroranges[k]);
	  abort();
#endif
	  
	}
      }
    }   

  } /* matches     1st while loop           */

  if (rhsdvectorconsumednonzeroranges[1] ==
      rhsdvectornonzeroranges[1]) {
    while (rhsdvectorconsumednonzeroranges[0] <
	   rhsdvectornonzeroranges[0]) {
      totalcurrentnonzerorange =
	rhsdibucketptr[0]->index_RANGE_END[rhsdibucketindex[0]] -
	  rhsdibucketptr[0]->index_RANGE_BEGIN[rhsdibucketindex[0]]+1;
      lhsdvectornonzeroes += totalcurrentnonzerorange;

      if (++rhsdvectorconsumednonzeroranges[0] <
	  rhsdvectornonzeroranges[0]) {
	if ((rhsdibucketindex[0] =
	     MY_MOD(rhsdvectorconsumednonzeroranges[0],
		    dIBUCKET_DIM_minus1)) == 0) {
	  if ((rhsdibucketptr[0] = rhsdibucketptr[0]->nextibucket) ==
	      NULL) {
	    
#ifdef _DEBUG
	    printf("plusqualdspg2 err #58: unexpected NULL nextvbucket: \n");
	    printf("\t k=%d, rhsdvectorconsumednonzeroranges=%d,\n",
		   0,rhsdvectorconsumednonzeroranges[0]);
	    printf("\t rhsdvectornonzeroranges[0]=%d\n",
		   rhsdvectornonzeroranges[0]);
	    abort();
#endif
	    
	  }
	}
      }   
    }
   }

  else if (rhsdvectorconsumednonzeroranges[0] ==
	   rhsdvectornonzeroranges[0]) {
    while((rhsdvectorconsumednonzeroranges[1] <
	   rhsdvectornonzeroranges[1]) &&
	  ((lastrhsdibucketindex[0] >= 0) &&
	   (rhsdibucketptr[1]->
	    index_RANGE_BEGIN[rhsdibucketindex[1]] >=
	    lastrhsdibucketptr[0]->
	    index_RANGE_BEGIN[lastrhsdibucketindex[0]] &&
	    rhsdibucketptr[1]->
	    index_RANGE_END[rhsdibucketindex[1]] <=
	    lastrhsdibucketptr[0]->
	    index_RANGE_END[lastrhsdibucketindex[0]]))) {
      
      /* Note: This part of the code is almost a copy of the part above
       * for the case where condition_2 is TRUE.  Therefore, when
       * modifying either code segment, care must be taken so that
       * both parts remain coherent.
       */

      k = 1;
      temprhsdibucketptr[0] = rhsdibucketptr[0];
      temprhsdvbucketptr[0] = rhsdvbucketptr[0];
      rhsdibucketptr[0] = lastrhsdibucketptr[0];
      rhsdvbucketptr[0] = lastrhsdvbucketptr[0];

      temprhsdvbucketindex[0] = rhsdvbucketindex[0];
      
      distancetooverlap =
	rhsdibucketptr[1]->index_RANGE_BEGIN[rhsdibucketindex[1]] -
	rhsdibucketptr[0]->index_RANGE_BEGIN[lastrhsdibucketindex[0]];
      distancetoendofvbucket = dVBUCKET_DIM - lastrhsdvbucketindex[0];
      
      if (distancetoendofvbucket > distancetooverlap) {
	rhsdvbucketindex[0] =
	  MY_MOD(lastrhsdvbucketindex[0] + distancetooverlap,
		 dVBUCKET_DIM_minus1);
      }
      else if (distancetoendofvbucket < distancetooverlap) {
	rhsdvbucketindex[0] =
	  MY_MOD(distancetooverlap - distancetoendofvbucket,
		 dVBUCKET_DIM_minus1);
      }
      else {
	rhsdvbucketindex[0] = 0;
      }
      
      while (distancetooverlap >= distancetoendofvbucket &&
	     rhsdvbucketptr[0]->nextvbucket != NULL) {
	rhsdvbucketptr[0] = rhsdvbucketptr[0]->nextvbucket;
	distancetooverlap -= distancetoendofvbucket;
	distancetoendofvbucket = dVBUCKET_DIM;
      }
      
      totalcurrentnonzerorange = 
      	rhsdibucketptr[1]->index_RANGE_END[rhsdibucketindex[1]] -
	  rhsdibucketptr[1]->index_RANGE_BEGIN[rhsdibucketindex[1]]+1;
      
      /* Wrapping the call with the two assignments has the effect of
       * EQUIVALENCEing the data structures being pointed two by the
       * two pointers.
       */
      
      lhsdvbucketptr = rhsdvbucketptr[0];      
      dOVERLAPPING_CS_RANGE_AXPY(&rhsdvbucketindex[0],
 				1,
				&rhsdvbucketindex[1],
				alpha[1],
				totalcurrentnonzerorange,
				&unconsumedcurrentnonzerorange,
				rhsdvectorconsumednonzeroranges[1],
				rhsdvectornonzeroranges[1],
				rhsdvectornonzeroranges[0]);
      rhsdvbucketptr[0] = lhsdvbucketptr;
      
      rhsdibucketptr[0] = temprhsdibucketptr[0];
      rhsdvbucketptr[0] = temprhsdvbucketptr[0];
      rhsdvbucketindex[0] = temprhsdvbucketindex[0];
      
      /*****************************************
       * Perform saxpy:                        *
       * 3.4. If there is one, get new         *
       *      nonzero from the rhs_vector      *
       *      whose nonzero was just consumed. *
       *****************************************/

      if (++rhsdvectorconsumednonzeroranges[k] <
	  rhsdvectornonzeroranges[k]) {
	if ((rhsdibucketindex[k] =
	     MY_MOD(rhsdvectorconsumednonzeroranges[k],
		    dIBUCKET_DIM_minus1)) == 0) {
	  if ((rhsdibucketptr[k] = rhsdibucketptr[k]->nextibucket) ==
	      NULL) {
	    
#ifdef _DEBUG
	    printf("plusqualdspg2 err #75: unexpected NULL nextvbucket: \n");
	    printf("\t k=%d, rhsdvectorconsumednonzeroranges=%d,\n",
		   k,rhsdvectorconsumednonzeroranges[k]);
	    printf("\t rhsdvectornonzeroranges[k]=%d\n",
		   rhsdvectornonzeroranges[k]);
	    abort();
#endif
	    
	  }
	}
      }   
    }
    
  }

  if (rhsdvectorconsumednonzeroranges[0] <
	 rhsdvectornonzeroranges[0] ||
	 rhsdvectorconsumednonzeroranges[1] <
	 rhsdvectornonzeroranges[1]) {

    restofplusequaldvectorptr->nonzeroes = 0;
    restofplusequaldvectorptr->dds.cs.nonzeroranges =
      rhsdvectornonzeroranges[0] - rhsdvectorconsumednonzeroranges[0];
    restofplusequalvectorconsumednonzeroranges = 0;
      
    if (rhsdvectorconsumednonzeroranges[0] <
	rhsdvectornonzeroranges[0]) {

      if (rhsdibucketindex[0] == 0 &&
	  rhsdibucketptr[0] != lastrhsdibucketptr[0]) {
	restofplusequalibucketptr = rhsdibucketptr[0];
	restofplusequalibucketindex = rhsdibucketindex[0];
	restofplusequaldvectorptr->dds.cs.ibucketptr =
	  restofplusequalibucketptr;
	rhsdibucketptr[0] = lastrhsdibucketptr[0];
      }
      else {
	dgetibucket(&restofplusequalibucketptr);
	restofplusequaldvectorptr->dds.cs.ibucketptr =
	  restofplusequalibucketptr;
	
	/* Note in the 'for' statement below, the remaining
	 * (not-yet-consumed-in-the saxpy) entries of rhsdibucketptr[0]
	 * are placed into restofplusequalibucketptr.
	 */
	
	for (i=rhsdibucketindex[0];i<dIBUCKET_DIM;i++) {
	  restofplusequalibucketptr->index_RANGE_BEGIN[i] =
	    rhsdibucketptr[0]->index_RANGE_BEGIN[i];
	  restofplusequalibucketptr->index_RANGE_END[i] =
	    rhsdibucketptr[0]->index_RANGE_END[i];
	}
	restofplusequalibucketindex = rhsdibucketindex[0];
	restofplusequalibucketptr->nextibucket =
	  rhsdibucketptr[0]->nextibucket;
	
	/* As distinct from the 'if' part above, there's no need here to set
	 * the values for rhsdibucketptr[0] and rhsdibucketindex[0],
	 * since they're already pointing where they should be pointing.
	 */
	
      }
      rhsdibucketptr[0]->nextibucket = NULL;
      
      if (rhsdvbucketindex[0] == 0 &&
	  rhsdvbucketptr[0] != lastrhsdvbucketptr[0]) {
	restofplusequalvbucketptr = rhsdvbucketptr[0];
	restofplusequalvbucketindex = rhsdvbucketindex[0];
	restofplusequaldvectorptr->dds.cs.vbucketptr =
	  restofplusequalvbucketptr;
	
	catchupto_rhsvbucketptr[0] = lastrhsdvbucketptr[0];
	while (catchupto_rhsvbucketptr[0]->nextvbucket !=
	       rhsdvbucketptr[0]) {
  
#ifdef _DEBUG
	  if (catchupto_rhsvbucketptr[0]->nextvbucket == NULL) {
	    printf("plusqualdspg2 error 18: unexpected NULL nextvbucket: \n");
	    abort();
	  }
#endif
	  
	  catchupto_rhsvbucketptr[0] = catchupto_rhsvbucketptr[0]->nextvbucket;
	}
	rhsdvbucketptr[0] = catchupto_rhsvbucketptr[0];
	
	rhsdvbucketindex[0] = dVBUCKET_DIM;
      }
      else {
	dgetvbucket(&restofplusequalvbucketptr);
	restofplusequaldvectorptr->dds.cs.vbucketptr =
	  restofplusequalvbucketptr;
	
	for (i=rhsdvbucketindex[0];i<dVBUCKET_DIM;i++) {
	  restofplusequalvbucketptr->value[i] =
	    rhsdvbucketptr[0]->value[i];
	}
	restofplusequalvbucketindex = rhsdvbucketindex[0];
	restofplusequalvbucketptr->nextvbucket =
	  rhsdvbucketptr[0]->nextvbucket;
      }
      rhsdvbucketptr[0]->nextvbucket = NULL;
    }
    else {
      restofplusequaldvectorptr->dds.cs.ibucketptr = NULL;
      restofplusequaldvectorptr->dds.cs.vbucketptr = NULL;
    }
	  
    /* Switching vector names in order not to have to change names
     * in the code that follows.  Now, lhsdvector will the new name
     * for the plusequalvector(rhsdvector[0]), and rhsdvector[0] will
     * be the new name for the restofplusequalvector.
     */
    
    tempvectorptr = lhsdvectorptr;
    
    lhsdvectorptr = rhsdvectorptr[0];
    
    lhsdvectornonzeroranges = rhsdvectorconsumednonzeroranges[0];
    
    lhsdibucketptr = rhsdibucketptr[0];
    lhsdvbucketptr = rhsdvbucketptr[0];
    lhsdvbucketindex = rhsdvbucketindex[0];
    
    /*
     * lhsdibucketindex is computed later using
     * lhsdvectornonzeroranges.
     */
    
    lastlhsdibucketindex = lastrhsdibucketindex[0];
    lastlhsdvbucketindex = lastrhsdvbucketindex[0];
    lastlhsdvbucketptr = lastrhsdvbucketptr[0];
    
    /*
     * We never make use of lastlhsdibucketptr.
     */

    if (rhsdvectorconsumednonzeroranges[0] <
	rhsdvectornonzeroranges[0]) {

      rhsdvectorptr[0] = restofplusequaldvectorptr;
      rhsdibucketptr[0] = rhsdvectorptr[0]->dds.cs.ibucketptr;
      rhsdvbucketptr[0] = rhsdvectorptr[0]->dds.cs.vbucketptr;
      rhsdibucketindex[0] = restofplusequalibucketindex;
      rhsdvbucketindex[0] = restofplusequalvbucketindex;
      
    }

    while (rhsdvectorconsumednonzeroranges[0] <
	   rhsdvectornonzeroranges[0] ||
	   rhsdvectorconsumednonzeroranges[1] <
	   rhsdvectornonzeroranges[1]) {
      
      if
	((rhsdvectorconsumednonzeroranges[1] ==
	  rhsdvectornonzeroranges[1]) ||
	 ((rhsdvectorconsumednonzeroranges[0] <
	   rhsdvectornonzeroranges[0]) &&
	  (rhsdibucketptr[0]->
	   index_RANGE_BEGIN[rhsdibucketindex[0]] <=
	   rhsdibucketptr[1]->
	   index_RANGE_BEGIN[rhsdibucketindex[1]])))
	  k = 0;
      else if
	((rhsdvectorconsumednonzeroranges[0] ==
	  rhsdvectornonzeroranges[0]) ||
	 ((rhsdvectorconsumednonzeroranges[1] <
	   rhsdvectornonzeroranges[1]) &&
	  (rhsdibucketptr[0]->
	   index_RANGE_BEGIN[rhsdibucketindex[0]] >
	   rhsdibucketptr[1]->
	   index_RANGE_BEGIN[rhsdibucketindex[1]])))
	  k = 1;
#ifdef _DEBUG
      else {
	printf("plusequaldspg2 error #11: BAD LOGIC.\n");
	abort();
      }
#endif
      j1 = rhsdibucketptr[k]->index_RANGE_BEGIN[rhsdibucketindex[k]];
      j2 = rhsdibucketptr[k]->index_RANGE_END[rhsdibucketindex[k]];
      rhsdibucketindex[k] =
	MY_MOD(rhsdvectorconsumednonzeroranges[k],dIBUCKET_DIM_minus1);
      lhsdibucketindex =
	MY_MOD(lhsdvectornonzeroranges,dIBUCKET_DIM_minus1);
      
      if ((rhsdvbucketindex[k] =
	  MY_MOD(rhsdvbucketindex[k],dVBUCKET_DIM_minus1))==0
	  && rhsdvectorconsumednonzeroranges[k] !=0
	  && (k !=0 || restofplusequalvectorconsumednonzeroranges !=0)) {
	dHOP_TO_NEXT_RHSVBUCKET(k,&rhsdvbucketindex[k],
			       rhsdvectorconsumednonzeroranges[k],
			       rhsdvectornonzeroranges[k]);
      }
      lhsdvbucketindex =
	MY_MOD(lhsdvbucketindex,dVBUCKET_DIM_minus1);
      
#ifdef _DEBUG
      if (rhsdibucketptr[k]->errorflag != 9099 ||
	  rhsdvbucketptr[k]->errorflag != 9009) {
	printf("plusequaldspg2 err 14: memory corruption of ibucketstore ");
	printf("or vbucketstore.\n");
	abort();
      }
#endif      
      
      /****************************************
       * Perform saxpy:                       *
       * 3.0. Check for startup errors.       *
       ****************************************/
      
#ifdef _DEBUG
      if (lhsdvectornonzeroranges == 0) {
	if (rhsdibucketindex[k] != 0 || rhsdvbucketindex[k] != 0 ||
	    lhsdibucketindex != 0 || lhsdvbucketindex != 0) {
	  printf("plusequaldspg2 error 15: Zeroes expected: \n\t");
	  printf("rhsdibucketindex[k]=%d, rhsdvbucketindex[k]=%d\n\t",
		 rhsdibucketindex[k], rhsdvbucketindex[k]);
	  printf("lhsdibucketindex=%d, lhsdvbucketindex=%d\n\t",
		 lhsdibucketindex,lhsdvbucketindex);
	  abort();
	}
      }
#endif      
      
      /****************************************
       * Perform saxpy:                       *
       * 3.1. Found 1ST or NEW subscript      *
       *      range (non-overlapping with     *
       *      previous one, if any) for       *
       *      lhsdvector.  If needed, get new  *
       *      ibuckets and vbuckets for       *
       *      lhsdvector.                      *
       ****************************************/
      if (lhsdvectornonzeroranges == 0 ||
	  j1 > lhsdibucketptr->index_RANGE_END[lastlhsdibucketindex]+1) {/*A.1*/
	
	if (lhsdibucketindex == 0 && lhsdvectornonzeroranges != 0) {
	  dgetibucket(&(lhsdibucketptr->nextibucket));
	  lhsdibucketptr = lhsdibucketptr->nextibucket;
	}
	lhsdibucketptr->index_RANGE_BEGIN[lhsdibucketindex] = j1;
	lhsdibucketptr->index_RANGE_END[lhsdibucketindex] = j2;
	
	if (lhsdvbucketindex == 0 && lhsdvectornonzeroranges != 0)
	  dget_and_hop_to_next_vbucket(&lhsdvbucketindex);
	
	lastlhsdvbucketptr = lhsdvbucketptr;
	lastlhsdvbucketindex = lhsdvbucketindex;
	
	totalcurrentnonzerorange = j2-j1+1;
	
	lhsdvectornonzeroes += totalcurrentnonzerorange;
	lhsdvectornonzeroranges++;
	
	
	dNON_OVER_LAPPING_CS_RANGE_AXPY(COMPLETELY_NEW_RANGE,
				       lhsdibucketindex,
				       &lastlhsdibucketindex,
				       &lhsdvbucketindex,
				       k,
				       &rhsdvbucketindex[k],
				       alpha[k],
				       totalcurrentnonzerorange,
				       &unconsumedcurrentnonzerorange,
				       rhsdvectorconsumednonzeroranges[k],
				       rhsdvectornonzeroranges[k],
				       lhsdvectornonzeroranges);
	
	
      } /*A.1*/
      
      
      /****************************************
       * Perform saxpy:                       *
       * 3.2. Found old subscript for         *
       *      lhsdvector.  Consume  nonzero in *
       *      saxpy.                          *
       *                                      *
       ****************************************/
      else if (j1 <= 
	       lhsdibucketptr->index_RANGE_END[lastlhsdibucketindex]+1 &&
	       j1 >= 
	       lhsdibucketptr->index_RANGE_BEGIN[lastlhsdibucketindex]) { /*B.1*/
	
	templhsdvbucketptr = lhsdvbucketptr;
	templhsdvbucketindex = lhsdvbucketindex;
	lhsdvbucketptr = lastlhsdvbucketptr;
	
	distancetooverlap = j1 -
	  lhsdibucketptr->index_RANGE_BEGIN[lastlhsdibucketindex];
	distancetoendofvbucket = dVBUCKET_DIM -
	  lastlhsdvbucketindex;
	
	if (j1 ==
	    lhsdibucketptr->index_RANGE_END[lastlhsdibucketindex]+1) {
	  h1 = 0;
	  h2 = -1;
	}
	else if (j2 <=
		 lhsdibucketptr->index_RANGE_END[lastlhsdibucketindex]) {
	  h1 = j1;
	  h2 = j2;
	  j1 = 0;
	  j2 = -1;
	}
	else {
	  h1 = j1;
	  h2 = lhsdibucketptr->index_RANGE_END[lastlhsdibucketindex];
	  j1 = h2+1;
	}
	
	if (distancetoendofvbucket > distancetooverlap) {
	  lhsdvbucketindex =
	    MY_MOD(lastlhsdvbucketindex + distancetooverlap,
		   dVBUCKET_DIM_minus1);
	}
	else if (distancetoendofvbucket < distancetooverlap) {
	  lhsdvbucketindex =
	    MY_MOD(distancetooverlap - distancetoendofvbucket,
		   dVBUCKET_DIM_minus1);
	}
	else {
	  lhsdvbucketindex = 0;
	}
	
	while (distancetooverlap >= distancetoendofvbucket &&
	       lhsdvbucketptr->nextvbucket != NULL) {
	  lhsdvbucketptr = lhsdvbucketptr->nextvbucket;
	  distancetooverlap -= distancetoendofvbucket;
	  distancetoendofvbucket = dVBUCKET_DIM;
	}
	
	lhsdibucketptr->index_RANGE_END[lastlhsdibucketindex] =
	  max(j2,
	      lhsdibucketptr->index_RANGE_END[lastlhsdibucketindex]);
	
	if ((totalcurrentnonzerorange = h2-h1+1) > 0) {
	  
	  
	  dOVERLAPPING_CS_RANGE_AXPY(&lhsdvbucketindex,
				    k,
				    &rhsdvbucketindex[k],
				    alpha[k],
				    totalcurrentnonzerorange,
				    &unconsumedcurrentnonzerorange,
				    rhsdvectorconsumednonzeroranges[k],
				    rhsdvectornonzeroranges[k],
				    lhsdvectornonzeroranges);
	  
	  
	}
	
	/*
	 * This bit of logic must be performed, whether or not there
	 * is additional nonzeroes to process.
	 */
	
	if (j2 != -1 &&
	    (lhsdvbucketptr != templhsdvbucketptr ||
	     (lhsdvbucketindex =
	      MY_MOD(lhsdvbucketindex,dVBUCKET_DIM_minus1)) !=
	     templhsdvbucketindex)) {
	  
#ifdef _DEBUG
	  printf("plusequaldspg2 err 50.0: Unexpectd value discrepancies\n");
	  printf("lhsdvbucketptr=%d, templhsdvbucketptr=%d\n",
		 (int)lhsdvbucketptr,(int)templhsdvbucketptr);
	  printf("lhsdvbucketindex=%d, templhsdvbucketindex=%d\n",
		 lhsdvbucketindex,templhsdvbucketindex);
	  
	  abort();
#endif
	}
	else {
	  lhsdvbucketptr = templhsdvbucketptr;
	  lhsdvbucketindex = templhsdvbucketindex;
	}
	
	
	if ((totalcurrentnonzerorange = j2-j1+1) > 0) {
	  lhsdvectornonzeroes += totalcurrentnonzerorange;
	  
	  /*
	   * adding new entries to vbucket
	   */
	  
	  if (lhsdvbucketindex == 0 &&
	      lhsdvectornonzeroranges != 0 &&
	      j2 != -1)
	    dget_and_hop_to_next_vbucket(&lhsdvbucketindex);
	  
	  if(rhsdvbucketindex[k] == dVBUCKET_DIM && j2 != -1) {
	    
	    dHOP_TO_NEXT_RHSVBUCKET(k,&rhsdvbucketindex[k],
				   rhsdvectorconsumednonzeroranges[k],
				   rhsdvectornonzeroranges[k]);
	  }
	  
	  
	  
	  
	  dNON_OVER_LAPPING_CS_RANGE_AXPY(NEW_ENTRIES_PADDED_TO_OLD_RANGE,
					 lhsdibucketindex,
					 &lastlhsdibucketindex,
					 &lhsdvbucketindex,
					 k,
					 &rhsdvbucketindex[k],
					 alpha[k],
					 totalcurrentnonzerorange,
					 &unconsumedcurrentnonzerorange,
					 rhsdvectorconsumednonzeroranges[k],
					 rhsdvectornonzeroranges[k],
					 lhsdvectornonzeroranges);
	  
	  
	} /* matches     if ((totalcurrent...) { */
      } /*B.1*/
      
      
      /****************************************	
       * Perform saxpy:                       *
       * 3.3. Found wrong subscript for       *
       *      lhsdvector.  Exit.               *
       ****************************************/
      
#ifdef _DEBUG
      else {
	printf("plusequaldspg2 error 60: lhsdvectornonzeroranges=%d,",
	       lhsdvectornonzeroranges);
	printf("lastlhsdibucketindex=%d, \n\t",lastlhsdibucketindex);
	printf("lhsdibucketptr->index[%d] = (%d,%d), j1=%d\n",
	       lastlhsdibucketindex,
	       lhsdibucketptr->
	       index_RANGE_BEGIN[lastlhsdibucketindex],
	       lhsdibucketptr->
	       index_RANGE_END[lastlhsdibucketindex],
	       j1);
	abort();
      }
#endif
      
      
      /****************************************
       * Perform saxpy:                       *
       * 3.4. If there is one, get new        *
       *      nonzero from the rhs_vector     *
       *      whose nonzero was just          *
       *      consumed and replace the        *
       *      heaproot with new entry;        *
       *      Else, delete heaproot.          *
       ****************************************/

      if (k==0) restofplusequalvectorconsumednonzeroranges++;

      if (++rhsdvectorconsumednonzeroranges[k] <
	  rhsdvectornonzeroranges[k]) {
	if ((rhsdibucketindex[k] =
	     MY_MOD(rhsdvectorconsumednonzeroranges[k],
		    dIBUCKET_DIM_minus1)) == 0) {
	  if ((rhsdibucketptr[k] = rhsdibucketptr[k]->nextibucket) ==
	      NULL) {
	    
#ifdef _DEBUG
	    printf("plusequaldspg2 err 61: unexpected NULL nextvbucket: \n");
	    printf("\t k=%d, rhsdvectorconsumednonzeroranges=%d,\n",
		   k,rhsdvectorconsumednonzeroranges[k]);
	    printf("\t rhsdvectornonzeroranges[k]=%d\n",
		   rhsdvectornonzeroranges[k]);
	    abort();
#endif
	    
	  }
	}
      }  
      
    } /* matches     2nd while loop           */
  } 

  else {

    /* tempvectorptr is used in the if part as temporary storage
     * for the value of lhsdvectorptr.  We echo that statement
     * here so that later the reinstatement of the value of
     * lhsdvectorptr could be done regardless of the branch taken.
     */

    tempvectorptr = lhsdvectorptr;
 
  }

  rhsdvectorptrptrs[1] = &rhsdvectorptr[1];
  
  dplusequal_axpy_postprocessing(rhsanddestptrptr,
				 rhsdvectorptrptrs[1],
				 status_otherrhsdvector,
				 &restofplusequaldvectorptr,
				 alphaptrs,
				 restorerhsdvectorptr,
				 arity,
				 lhsdvectornonzeroes,
				 lhsdvectornonzeroranges,
				 CS);
  
  
  
  lhsdvectorptr =  tempvectorptr;
  
  
#ifdef _TRACE 
  nnza = (*destptrptr)->nonzeroes; 
  dadinfo(arity,nnzb,lhsdvectoraddress,nnza,alphaptrs,vectorptrptrs);
#endif

}

