#define DOUBLE
/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#include  <stdlib.h>
#include "dgaxpy_pd_macros_and_data_structures.h"
#include "dgaxpy_pd_global_variables.h"

/****************************************************************************
 *                                                                          *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/

void dnon_over_lapping_cs_range_axpy(rangedestination,
				    lhsdibucketindex,
				    addressoflastlhsdibucketindex,
				    addressoflhsdvbucketindex,
				    k,
				    addressofrhsdvbucketindex_k,
				    alpha_k,
				    totalcurrentnonzerorange
#ifdef _DEBUG
				    ,addressofunconsumedcurrentnonzerorange
				    ,rhsdvectorconsumednonzeroranges_k
				    ,rhsdvectornonzeroranges_k
				    ,lhsdvectornonzeroranges
#endif
				    )
int rangedestination;
int lhsdibucketindex;
int *addressoflastlhsdibucketindex;
int *addressoflhsdvbucketindex;
int k;
int *addressofrhsdvbucketindex_k;
double alpha_k;
int totalcurrentnonzerorange;
#ifdef _DEBUG
int *addressofunconsumedcurrentnonzerorange;
int rhsdvectorconsumednonzeroranges_k;
int rhsdvectornonzeroranges_k;
int lhsdvectornonzeroranges;
#endif
{
  int firstloopsize;
  int numberofdoubleforloopiterations;
  int firstiterationloopsize;
  int seconditerationloopsize;
  int remainder1loopsize;
  int remainder2loopsize;

  int i;


  dDETERMINE_LOOP_BOUNDS(totalcurrentnonzerorange,
			*addressoflhsdvbucketindex,
			*addressofrhsdvbucketindex_k,
			&firstloopsize,
			&numberofdoubleforloopiterations,
			&firstiterationloopsize,
			&seconditerationloopsize,
			&remainder1loopsize,
			&remainder2loopsize,
			addressofunconsumedcurrentnonzerorange);
  dDENSE_AXPY_LOOP(NEW_RANGE,
		  addressoflhsdvbucketindex,
		  addressofrhsdvbucketindex_k,
		  k,
		  firstloopsize,
		  alpha_k,
		  addressofunconsumedcurrentnonzerorange);
  
    
  /*********************************************
   * Case 1: rhsdvbucket has more spaces   *
   *         remaining than lhsdvbucket.   *
   *********************************************/
  
  if (*addressofrhsdvbucketindex_k <
      *addressoflhsdvbucketindex) {
    
    for(i=0;i<numberofdoubleforloopiterations;i++) {
      dDEBUGGING_CODE_1(LEFT,
		       *addressoflhsdvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      dget_and_hop_to_next_vbucket(addressoflhsdvbucketindex);
      dDENSE_AXPY_LOOP(NEW_RANGE,
		      addressoflhsdvbucketindex,
		      addressofrhsdvbucketindex_k,
		      k,
		      firstiterationloopsize,alpha_k,
		      addressofunconsumedcurrentnonzerorange);
      dDEBUGGING_CODE_1(RIGHT,
		       *addressofrhsdvbucketindex_k,
		       *addressofunconsumedcurrentnonzerorange);
      dHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhsdvbucketindex_k,
			     rhsdvectorconsumednonzeroranges_k,
			     rhsdvectornonzeroranges_k);
      dDENSE_AXPY_LOOP(NEW_RANGE,
		      addressoflhsdvbucketindex,
		      addressofrhsdvbucketindex_k,
		      k,
		      seconditerationloopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    
    if (remainder1loopsize > 0) {
      dDEBUGGING_CODE_1(LEFT,
		       *addressoflhsdvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      dget_and_hop_to_next_vbucket(addressoflhsdvbucketindex);
      dDENSE_AXPY_LOOP(NEW_RANGE,
		      addressoflhsdvbucketindex,
		      addressofrhsdvbucketindex_k,
		      k,
		      remainder1loopsize,alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    if (remainder2loopsize > 0) {
      dDEBUGGING_CODE_1(RIGHT,
		       *addressofrhsdvbucketindex_k,
		       *addressofunconsumedcurrentnonzerorange);
      dHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhsdvbucketindex_k,
			     rhsdvectorconsumednonzeroranges_k,
			     rhsdvectornonzeroranges_k);
      dDENSE_AXPY_LOOP(NEW_RANGE,
		      addressoflhsdvbucketindex,
		      addressofrhsdvbucketindex_k,
		      k,
		      remainder2loopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    DEBUGGING_CODE_2(*addressofunconsumedcurrentnonzerorange);
    if (rangedestination == COMPLETELY_NEW_RANGE) {
      *addressoflastlhsdibucketindex = lhsdibucketindex;
    }
  }

  /*********************************************
   * Case 2: rhsdvbucket has less spaces   *
   * remaining than lhsdvbucket.           *
   *********************************************/	  
  else if (*addressofrhsdvbucketindex_k >
	   *addressoflhsdvbucketindex) {
    
    for(i=0;i<numberofdoubleforloopiterations;i++) {
      dDEBUGGING_CODE_1(RIGHT,
		       *addressofrhsdvbucketindex_k,
		       *addressofunconsumedcurrentnonzerorange);
      dHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhsdvbucketindex_k,
			     rhsdvectorconsumednonzeroranges_k,
			     rhsdvectornonzeroranges_k);
      dDENSE_AXPY_LOOP(NEW_RANGE,
		      addressoflhsdvbucketindex,
		      addressofrhsdvbucketindex_k,
		      k,
		      firstiterationloopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
      dDEBUGGING_CODE_1(LEFT,
		       *addressoflhsdvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      dget_and_hop_to_next_vbucket(addressoflhsdvbucketindex);
      dDENSE_AXPY_LOOP(NEW_RANGE,
		      addressoflhsdvbucketindex,
		      addressofrhsdvbucketindex_k,
		      k,
		      seconditerationloopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    
    if (remainder1loopsize > 0) {
      dDEBUGGING_CODE_1(RIGHT,
		       *addressofrhsdvbucketindex_k,
		       *addressofunconsumedcurrentnonzerorange);
      dHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhsdvbucketindex_k,
			     rhsdvectorconsumednonzeroranges_k,
			     rhsdvectornonzeroranges_k);
      dDENSE_AXPY_LOOP(NEW_RANGE,
		      addressoflhsdvbucketindex,
		      addressofrhsdvbucketindex_k,
		      k,
		      remainder1loopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    
    if (remainder2loopsize > 0) {
      dDEBUGGING_CODE_1(LEFT,
		       *addressoflhsdvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      dget_and_hop_to_next_vbucket(addressoflhsdvbucketindex);
      dDENSE_AXPY_LOOP(NEW_RANGE,
		      addressoflhsdvbucketindex,
		      addressofrhsdvbucketindex_k,
		      k,
		      remainder2loopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    DEBUGGING_CODE_2(*addressofunconsumedcurrentnonzerorange);
    if (rangedestination == COMPLETELY_NEW_RANGE) {
      *addressoflastlhsdibucketindex = lhsdibucketindex;
    }
  }
  
   
  /*********************************************
   * Case 3: rhsdvbucket has same # of     *
   * spaces as lhsdvbucket (i.e. 0).       *
   *********************************************/
  else if (*addressofrhsdvbucketindex_k ==
	   *addressoflhsdvbucketindex) {
    
    for(i=0;i<numberofdoubleforloopiterations;i++) {
      dDEBUGGING_CODE_1(LEFT,
		       *addressoflhsdvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      dget_and_hop_to_next_vbucket(addressoflhsdvbucketindex);
      dHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhsdvbucketindex_k,
			     rhsdvectorconsumednonzeroranges_k,
			     rhsdvectornonzeroranges_k);
      dDENSE_AXPY_LOOP(NEW_RANGE,
		      addressoflhsdvbucketindex,
		      addressofrhsdvbucketindex_k,
		      k,
		      firstiterationloopsize+seconditerationloopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    if (remainder1loopsize+remainder2loopsize > 0) {
      dDEBUGGING_CODE_1(LEFT,
		       *addressoflhsdvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      dget_and_hop_to_next_vbucket(addressoflhsdvbucketindex);
      dHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhsdvbucketindex_k,
			     rhsdvectorconsumednonzeroranges_k,
			     rhsdvectornonzeroranges_k);
      dDENSE_AXPY_LOOP(NEW_RANGE,
		      addressoflhsdvbucketindex,
		      addressofrhsdvbucketindex_k,
		      k,
		      remainder1loopsize+remainder2loopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    DEBUGGING_CODE_2(*addressofunconsumedcurrentnonzerorange);
    if (rangedestination == COMPLETELY_NEW_RANGE) {
      *addressoflastlhsdibucketindex = lhsdibucketindex;
    }
  }
}






/****************************************************************************
 *                                                                          *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/

void doverlapping_cs_range_axpy(addressoflhsdvbucketindex,
			       k,
			       addressofrhsdvbucketindex_k,
			       alpha_k,
			       totalcurrentnonzerorange
#ifdef _DEBUG
			       ,addressofunconsumedcurrentnonzerorange
			       ,rhsdvectorconsumednonzeroranges_k
			       ,rhsdvectornonzeroranges_k
			       ,lhsdvectornonzeroranges
#endif
			       )
int *addressoflhsdvbucketindex;
int k;
int *addressofrhsdvbucketindex_k;
double alpha_k;
int totalcurrentnonzerorange;
#ifdef _DEBUG
int *addressofunconsumedcurrentnonzerorange;
int rhsdvectorconsumednonzeroranges_k;
int rhsdvectornonzeroranges_k;
int lhsdvectornonzeroranges;
#endif
{
  int firstloopsize;
  int numberofdoubleforloopiterations;
  int firstiterationloopsize;
  int seconditerationloopsize;
  int remainder1loopsize;
  int remainder2loopsize;

  int i;
  
  dDETERMINE_LOOP_BOUNDS(totalcurrentnonzerorange,
			*addressoflhsdvbucketindex,
			*addressofrhsdvbucketindex_k,
			&firstloopsize,
			&numberofdoubleforloopiterations,
			&firstiterationloopsize,
			&seconditerationloopsize,
			&remainder1loopsize,
			&remainder2loopsize,
			addressofunconsumedcurrentnonzerorange);
  dDENSE_AXPY_LOOP(OLD_RANGE,
		  addressoflhsdvbucketindex,
		  addressofrhsdvbucketindex_k,
		  k,
		  firstloopsize,
		  alpha_k,
		  addressofunconsumedcurrentnonzerorange);
  
  
  /*********************************************
   * Case 1: rhsdvbucket has more spaces   *
   *         remaining than lhsdvbucket.   *
   *********************************************/
  if (*addressofrhsdvbucketindex_k <
      *addressoflhsdvbucketindex) {
    /*B.2.1*/
    
    for(i=0;i<numberofdoubleforloopiterations;i++) { /*B.3.1*/
      dDEBUGGING_CODE_1(LEFT,
		       *addressoflhsdvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      dHOP_TO_NEXT_LHSVBUCKET(addressoflhsdvbucketindex,
			     lhsdvectornonzeroranges);
      dDENSE_AXPY_LOOP(OLD_RANGE,
		      addressoflhsdvbucketindex,
		      addressofrhsdvbucketindex_k,
		      k,
		      firstiterationloopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
      dDEBUGGING_CODE_1(RIGHT,
		       *addressofrhsdvbucketindex_k,
		       *addressofunconsumedcurrentnonzerorange);
      dHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhsdvbucketindex_k,
			     rhsdvectorconsumednonzeroranges_k,
			     rhsdvectornonzeroranges_k);
      dDENSE_AXPY_LOOP(OLD_RANGE,
		      addressoflhsdvbucketindex,
		      addressofrhsdvbucketindex_k,
		      k,
		      seconditerationloopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    } /*B.3.1*/
    
    if (remainder1loopsize > 0) {    	    
      dDEBUGGING_CODE_1(LEFT,
		       *addressoflhsdvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      dHOP_TO_NEXT_LHSVBUCKET(addressoflhsdvbucketindex,
			     lhsdvectornonzeroranges);
      dDENSE_AXPY_LOOP(OLD_RANGE,
		      addressoflhsdvbucketindex,
		      addressofrhsdvbucketindex_k,
		      k,
		      remainder1loopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    
    if (remainder2loopsize > 0) {    
      dDEBUGGING_CODE_1(RIGHT,
		       *addressofrhsdvbucketindex_k,
		       *addressofunconsumedcurrentnonzerorange);
      dHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhsdvbucketindex_k,
			     rhsdvectorconsumednonzeroranges_k,
			     rhsdvectornonzeroranges_k);
      dDENSE_AXPY_LOOP(OLD_RANGE,
		      addressoflhsdvbucketindex,
		      addressofrhsdvbucketindex_k,
		      k,
		      remainder2loopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    DEBUGGING_CODE_2(*addressofunconsumedcurrentnonzerorange);
  } /*B.2.1*/
  
  /*********************************************
   * Case 2: rhsdvbucket has less spaces   *
   * remaining than lhsdvbucket.           *
   *********************************************/
  else if (*addressofrhsdvbucketindex_k >
	   *addressoflhsdvbucketindex) {
 /*B.2.2*/
    
    for(i=0;i<numberofdoubleforloopiterations;i++) { /*B.3.2*/
      dDEBUGGING_CODE_1(RIGHT,
		       *addressofrhsdvbucketindex_k,
		       *addressofunconsumedcurrentnonzerorange);
      dHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhsdvbucketindex_k,
			     rhsdvectorconsumednonzeroranges_k,
			     rhsdvectornonzeroranges_k);
      dDENSE_AXPY_LOOP(OLD_RANGE,
		      addressoflhsdvbucketindex,
		      addressofrhsdvbucketindex_k,
		      k,
		      firstiterationloopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
      dDEBUGGING_CODE_1(LEFT,
		       *addressoflhsdvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      dHOP_TO_NEXT_LHSVBUCKET(addressoflhsdvbucketindex,
			     lhsdvectornonzeroranges);
      dDENSE_AXPY_LOOP(OLD_RANGE,
		      addressoflhsdvbucketindex,
		      addressofrhsdvbucketindex_k,
		      k,
		      seconditerationloopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    } /*B.3.2*/
    
    if (remainder1loopsize > 0) {        
      dDEBUGGING_CODE_1(RIGHT,
		       *addressofrhsdvbucketindex_k,
		       *addressofunconsumedcurrentnonzerorange);
      dHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhsdvbucketindex_k,
			     rhsdvectorconsumednonzeroranges_k,
			     rhsdvectornonzeroranges_k);
      dDENSE_AXPY_LOOP(OLD_RANGE,
		      addressoflhsdvbucketindex,
		      addressofrhsdvbucketindex_k,
		      k,
		      remainder1loopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    if (remainder2loopsize > 0) {    
      dDEBUGGING_CODE_1(LEFT,
		       *addressoflhsdvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      dHOP_TO_NEXT_LHSVBUCKET(addressoflhsdvbucketindex,
			     lhsdvectornonzeroranges);
      dDENSE_AXPY_LOOP(OLD_RANGE,
		      addressoflhsdvbucketindex,
		      addressofrhsdvbucketindex_k,
		      k,
		      remainder2loopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    DEBUGGING_CODE_2(*addressofunconsumedcurrentnonzerorange);
  } /*B.2.2*/
  
  
  /*********************************************
   * Case 3: rhsdvbucket has same # of     *
   * spaces as lhsdvbucket (i.e. 0).       *
   *********************************************/
  else if(*addressofrhsdvbucketindex_k ==
	  *addressoflhsdvbucketindex) {
    /*B.2.3*/
    
    for(i=0;i<numberofdoubleforloopiterations;i++) { /*B.3.3*/
      dDEBUGGING_CODE_1(LEFT,
		       *addressoflhsdvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      dHOP_TO_NEXT_LHSVBUCKET(addressoflhsdvbucketindex,
			     lhsdvectornonzeroranges);
      dHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhsdvbucketindex_k,
			     rhsdvectorconsumednonzeroranges_k,
			     rhsdvectornonzeroranges_k);
      dDENSE_AXPY_LOOP(OLD_RANGE,
		      addressoflhsdvbucketindex,
		      addressofrhsdvbucketindex_k,
		      k,
		      firstiterationloopsize+seconditerationloopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    } /*B.3.3*/
    
    if (remainder1loopsize+remainder2loopsize > 0) {    
      dDEBUGGING_CODE_1(LEFT,
		       *addressoflhsdvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      dHOP_TO_NEXT_LHSVBUCKET(addressoflhsdvbucketindex,
			     lhsdvectornonzeroranges);
      dHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhsdvbucketindex_k,
			     rhsdvectorconsumednonzeroranges_k,
			     rhsdvectornonzeroranges_k);
      dDENSE_AXPY_LOOP(OLD_RANGE,
		      addressoflhsdvbucketindex,
		      addressofrhsdvbucketindex_k,
		      k,
		      remainder1loopsize+remainder2loopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    DEBUGGING_CODE_2(*addressofunconsumedcurrentnonzerorange);
  } /*B.2.3*/
}



/****************************************************************************
 *                                                                          *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/

void dhop_to_next_rhsvbucket(k,
			    vbucketindex
#ifdef _DEBUG
			    ,rhsdvectorconsumednonzeroranges
			    ,rhsdvectornonzeroranges
#endif
			    )
int k;
int *vbucketindex;
#ifdef _DEBUG
int rhsdvectorconsumednonzeroranges;
int rhsdvectornonzeroranges;
#endif
{
  if ((rhsdvbucketptr[k] = rhsdvbucketptr[k]->nextvbucket) == NULL) {
	      
#ifdef _DEBUG
    printf("dhop_to_next_rhsvbucket error: unexpected NULL nextvbucket: \n");
    printf("\t k=%d, rhsdvectorconsumednonzeroranges[k]=%d,\n",
	   k,rhsdvectorconsumednonzeroranges);
    printf("\t rhsdvectornonzeroranges[k]=%d\n",
	   rhsdvectornonzeroranges);
    abort();
#endif
  }
  else {
    *vbucketindex = 0;
  }
}





/****************************************************************************
 *                                                                          *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/

void dhop_to_next_lhsvbucket(vbucketindex
#ifdef _DEBUG
		      ,lhsdvectornonzeroranges
#endif
		      )
int *vbucketindex;
#ifdef _DEBUG
int lhsdvectornonzeroranges;
#endif
{
  if ((lhsdvbucketptr = lhsdvbucketptr->nextvbucket) == NULL) {
    
#ifdef _DEBUG
    printf("dhop_to_next_lhsvbucket error: unexpected NULL nextvbucket: \n");
    printf("lhsdvectornonzeroranges=%d\n",lhsdvectornonzeroranges);
    abort();
#endif
  }
  else {
    *vbucketindex = 0;
  }
}  





/****************************************************************************
 *                                                                          *
 *      dget_and_hop_to_next_vbucket(lhsdvbucketindex)                   *
 *                                                                          *
 ****************************************************************************/

void dget_and_hop_to_next_vbucket(lhsdvbucketindex)
int *lhsdvbucketindex;
{
  dgetvbucket(&(lhsdvbucketptr->nextvbucket));
  lhsdvbucketptr = lhsdvbucketptr->nextvbucket;
  *lhsdvbucketindex = 0;
}







/****************************************************************************
 *                                                                          *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/

void dconvertdstype(oldvectorptrptr,k,newvectordstype)
dvectorstruct **oldvectorptrptr;
int k;
int newvectordstype;
{
  dbucketstruct *holdingbinbucketptr;
  dbucketstruct *oldbucketptr;
  dibucketstruct *holdingbinibucketptr;
  dvbucketstruct *holdingbinvbucketptr;
  int bucketindex;
  int ibucketindex = 0;
  int vbucketindex;
  int j;
  int ij;
  int vj;
  int lastindex;
  dvectorstruct *oldvectorptr = (*oldvectorptrptr);
  int oldvectordstype = oldvectorptr->dstype;

  if (oldvectordstype >= newvectordstype || newvectordstype == S1) {
#ifdef _DEBUG
    printf("dconvertdstype error 1: Illegal Data Structure Conversion:\n\t");
    printf("newvectordstype=%d,    oldvectordstype=%d\n",
	   newvectordstype,oldvectordstype);
    abort();
#endif
  }
  else if (oldvectordstype == S1 &&  newvectordstype == SS) {
    holdingbindvectorptr[k]->nonzeroes = 1;
    holdingbindvectorptr[k]->dstype = SS;
    dgetbucket(&holdingbinbucketptr);
    holdingbindvectorptr[k]->dds.ss.bucketptr = holdingbinbucketptr;
    holdingbinbucketptr->index[0] = oldvectorptr->dds.s1.index;
    holdingbinbucketptr->value[0] = oldvectorptr->dds.s1.value;
  }
  else if (oldvectordstype == S1 &&  newvectordstype == CS) {
    holdingbindvectorptr[k]->nonzeroes = 1;
    holdingbindvectorptr[k]->dstype = CS;
    dgetibucket(&holdingbinibucketptr);
    dgetvbucket(&holdingbinvbucketptr);
    holdingbindvectorptr[k]->dds.cs.nonzeroranges = 1;
    holdingbindvectorptr[k]->dds.cs.ibucketptr = holdingbinibucketptr;
    holdingbindvectorptr[k]->dds.cs.vbucketptr = holdingbinvbucketptr;
    holdingbinibucketptr->index_RANGE_BEGIN[0] = oldvectorptr->dds.s1.index;
    holdingbinibucketptr->index_RANGE_END[0] = oldvectorptr->dds.s1.index;
    holdingbinvbucketptr->value[0] = oldvectorptr->dds.s1.value;
  }
  else if (oldvectordstype == SS &&  newvectordstype == CS) {
    oldbucketptr = oldvectorptr->dds.ss.bucketptr;
    holdingbindvectorptr[k]->nonzeroes = oldvectorptr->nonzeroes;
    holdingbindvectorptr[k]->dstype = CS;
    dgetibucket(&holdingbinibucketptr);
    holdingbindvectorptr[k]->dds.cs.ibucketptr = holdingbinibucketptr;
    dgetvbucket(&holdingbinvbucketptr);
    holdingbindvectorptr[k]->dds.cs.vbucketptr = holdingbinvbucketptr;
    holdingbinibucketptr->index_RANGE_BEGIN[0] = oldbucketptr->index[0];
    holdingbinvbucketptr->value[0] = oldbucketptr->value[0];
    lastindex = oldbucketptr->index[0];
    
    for (j=1,ij=0,vj=1;j<oldvectorptr->nonzeroes;j++,vj++) {
      if ((bucketindex = MY_MOD(j,dBUCKET_DIM_minus1)) == 0)
	  oldbucketptr = oldbucketptr->nextbucket;
      if ((vbucketindex = MY_MOD(vj,dVBUCKET_DIM_minus1)) == 0) {
	dgetvbucket(&holdingbinvbucketptr->nextvbucket);
	holdingbinvbucketptr = holdingbinvbucketptr->nextvbucket;
      }
      holdingbinvbucketptr->value[vbucketindex] =
	oldbucketptr->value[bucketindex];

      if (oldbucketptr->index[bucketindex] > lastindex+1) {
	holdingbinibucketptr->index_RANGE_END[ibucketindex] = lastindex;
	ij++;
	if ((ibucketindex = MY_MOD(ij,dIBUCKET_DIM_minus1)) == 0) {
	  dgetibucket(&holdingbinibucketptr->nextibucket);
	  holdingbinibucketptr = holdingbinibucketptr->nextibucket;
	}
	holdingbinibucketptr->index_RANGE_BEGIN[ibucketindex] =
	  oldbucketptr->index[bucketindex];
      }
      else if (oldbucketptr->index[bucketindex] < lastindex+1) {
#ifdef _DEBUG
	printf("dconvertdstype error 2: BAD LOGIC:\n\t");
	printf("newvectordstype=%d,    oldvectordstype=%d\n",
	       newvectordstype,oldvectordstype);
	abort();
#endif
      }
      lastindex = oldbucketptr->index[bucketindex];
    }
    holdingbinibucketptr->index_RANGE_END[ibucketindex] = lastindex;    
    holdingbindvectorptr[k]->dds.cs.nonzeroranges = ij+1;
  }
  else /* We should never go thru this branch */ {
#ifdef _DEBUG
    printf("dconvertdstype error 3: BAD LOGIC:\n\t");
    printf("newvectordstype=%d,    oldvectordstype=%d\n",
	   newvectordstype,oldvectordstype);
    abort();
#endif
  }
  *oldvectorptrptr =  holdingbindvectorptr[k];
}








/****************************************************************************
 *                                                                          *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/

void daxpy_0_1_post_processing(destptrptr)
dvectorstruct **destptrptr;
{
  dvectorstruct *tempvectorptr;
  dbucketstruct *tempbucketptr;
  dibucketstruct *tempibucketptr;
  dvbucketstruct *tempvbucketptr;
  
  /*********************************************
   * Switch lhsdvectorptr and *destptrptr. *
   *********************************************/
  tempvectorptr = (*destptrptr);
  (*destptrptr) = lhsdvectorptr;
  lhsdvectorptr = tempvectorptr;
  
  if (tempvectorptr->dstype == SS) {
    /*********************************************
     * If *destptrptr was not empty,             *
     * add its buckets to freeddbucketslist. *
     *********************************************/
    tempbucketptr = tempvectorptr->dds.ss.bucketptr;
    if (tempbucketptr != NULL) {
      while (tempbucketptr->nextbucket != NULL)
	tempbucketptr = tempbucketptr->nextbucket;
      tempbucketptr->nextbucket = freeddbucketslist;
      RESET_freeddbucketslist(tempvectorptr->dds.ss.bucketptr);
      tempvectorptr->dds.ss.bucketptr = NULL;
    }
  }
  else if (tempvectorptr->dstype == CS) {
    /********************************************
     * If *destptrptr was not empty, add        *
     * its ibuckets to freeddibucketslist.  *
     ********************************************/
    
    tempibucketptr = tempvectorptr->dds.cs.ibucketptr;
    tempvbucketptr = tempvectorptr->dds.cs.vbucketptr;
    
    if (tempibucketptr != NULL) {
      while (tempibucketptr->nextibucket != NULL)
	tempibucketptr = tempibucketptr->nextibucket;
      tempibucketptr->nextibucket = freeddibucketslist;
      RESET_freeddibucketslist(tempvectorptr->dds.cs.ibucketptr);
      tempvectorptr->dds.cs.ibucketptr = NULL;
    }
    
    /********************************************
     * If *destptrptr was not empty, add        *
     * its vbuckets to freeddvbucketslist.  *
     ********************************************/
    
    if (tempvbucketptr != NULL) {
      while (tempvbucketptr->nextvbucket != NULL)
	tempvbucketptr = tempvbucketptr->nextvbucket;
      tempvbucketptr->nextvbucket = freeddvbucketslist;
      RESET_freeddvbucketslist(tempvectorptr->dds.cs.vbucketptr);
      tempvectorptr->dds.cs.vbucketptr = NULL;
    }
  } /* matches   else if (tempvectorptr->dstype == CS) {             */
}


/****************************************************************************
 *                                                                          *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/

void daxpy_2_heap_post_processing(destptrptr,
				 rhsdvectorptrptrs,
				 alphaptrs,
				 restorerhsdvectorptr,
				 arity,
				 lhsdvectornonzeroes,
				 lhsdvectornonzeroranges,
				 governingdstype)
dvectorstruct **destptrptr;
dvectorstruct **rhsdvectorptrptrs[];
double *alphaptrs[];
dvectorstruct *restorerhsdvectorptr[MAXARITY+1];
int arity;
int lhsdvectornonzeroes;
int lhsdvectornonzeroranges;
int governingdstype;
{
  dvectorstruct *rhsdvectorptr[MAXARITY];
  dvectorstruct *tempvectorptr;
  dvectorstruct *sstempvectorptr;
  dbucketstruct *tempbucketptr;
  dbucketstruct *sstempbucketptr;
  dibucketstruct *tempibucketptr;
  dvbucketstruct *tempvbucketptr;
  int k;

  for (k=0;k<arity;k++)
    rhsdvectorptr[k] = *rhsdvectorptrptrs[k];

  /****************************************
   * Switch lhsdvectorptr and *destptrptr. *
   ****************************************/

  lhsdvectorptr->nonzeroes = lhsdvectornonzeroes;
  
  if (governingdstype == CS)
    lhsdvectorptr->dds.cs.nonzeroranges = lhsdvectornonzeroranges;
  
  tempvectorptr = (*destptrptr);
  (*destptrptr) = lhsdvectorptr;
  lhsdvectorptr = tempvectorptr;

  if (governingdstype == SS) {
    
    if ((INITIALIZED_MAXINDEX &&
	 ((float) (*destptrptr)->nonzeroes / (float) GLOBAL_MAXINDEX) >=
	 SWITCH_NZ_RATIO)     ||
	((*destptrptr)->nonzeroes >= SWITCH_NZ_VALUE)) {
     
       /****************************************
       * Convert SS to CS.                    *
       ****************************************/
      
      sstempvectorptr = (*destptrptr);
      dconvertdstype(destptrptr,MAXARITY,CS);
	  
      sstempbucketptr = sstempvectorptr->dds.ss.bucketptr;
      if (sstempbucketptr != NULL) {
	while (sstempbucketptr->nextbucket != NULL)
	  sstempbucketptr = sstempbucketptr->nextbucket;
	sstempbucketptr->nextbucket = freeddbucketslist;
	RESET_freeddbucketslist(sstempvectorptr->dds.ss.bucketptr);
	sstempvectorptr->dds.ss.bucketptr = NULL;
      }
      holdingbindvectorptr[MAXARITY] =  sstempvectorptr;
    }
  }
  
  /****************************************
   * If *destptrptr was not empty,        *
   * add its buckets to freeddbucketslist. *
   ****************************************/
  
  if (tempvectorptr->dstype == SS) {
    tempbucketptr = tempvectorptr->dds.ss.bucketptr;
    if (tempbucketptr != NULL) {
      while (tempbucketptr->nextbucket != NULL)
	tempbucketptr = tempbucketptr->nextbucket;
      tempbucketptr->nextbucket = freeddbucketslist;
      RESET_freeddbucketslist(tempvectorptr->dds.ss.bucketptr);
      tempvectorptr->dds.ss.bucketptr = NULL;
    }
  }
  else if (tempvectorptr->dstype == CS) {
    tempibucketptr = tempvectorptr->dds.cs.ibucketptr;
    if (tempibucketptr != NULL) {
      while (tempibucketptr->nextibucket != NULL)
	tempibucketptr = tempibucketptr->nextibucket;
      tempibucketptr->nextibucket = freeddibucketslist;
      RESET_freeddibucketslist(tempvectorptr->dds.cs.ibucketptr);
      tempvectorptr->dds.cs.ibucketptr = NULL;
    }
    tempvbucketptr = tempvectorptr->dds.cs.vbucketptr;
    if (tempvbucketptr != NULL) {
      while (tempvbucketptr->nextvbucket != NULL)
	tempvbucketptr = tempvbucketptr->nextvbucket;
      tempvbucketptr->nextvbucket = freeddvbucketslist;
      RESET_freeddvbucketslist(tempvectorptr->dds.cs.vbucketptr);
      tempvectorptr->dds.cs.vbucketptr = NULL;
    }
  }
  
  /*****************************************
   * If holdingbins were used, add         *
   * their buckets to freeddbucketslist, OR *
   * their ibuckets to freeddibucketslist & *
   * their vbuckets to freeddvbucketslist.  *
   *****************************************/
  for (k=0;k<arity;k++) {
    
    rhsdvectorptr[k] = restorerhsdvectorptr[k];
    
    if (holdingbindvectorptr[k]->dstype == SS) {
      tempbucketptr = holdingbindvectorptr[k]->dds.ss.bucketptr;
      if (tempbucketptr != NULL) {
#ifdef _DEBUG
	if (governingdstype == CS) {
	  printf("dspgx warning 7.4: Unexpcted non-NULL");
	  printf("tempbucketptr.\n");
	}
#endif
	while (tempbucketptr->nextbucket != NULL) {
#ifdef _DEBUG
	  if (governingdstype == SS) {
	    printf("dspgx warning 7.1: tempbucketptr should have");
	    printf(" no trailing nextbuckets\n");
	  }
#endif
	  tempbucketptr = tempbucketptr->nextbucket;
	}
	tempbucketptr->nextbucket = freeddbucketslist;
	RESET_freeddbucketslist(holdingbindvectorptr[k]->dds.ss.bucketptr);
	holdingbindvectorptr[k]->dds.ss.bucketptr = NULL;
      }
    }
    else if (holdingbindvectorptr[k]->dstype == CS) {
      tempibucketptr = holdingbindvectorptr[k]->dds.cs.ibucketptr;
      if (tempibucketptr != NULL) {
#ifdef _DEBUG
	if (governingdstype == SS) {
	  printf("dspgx warning 7.2: Unexpcted non-NULL");
	  printf(" ibucketptr.\n");
	}
#endif
	while (tempibucketptr->nextibucket != NULL)
	  tempibucketptr = tempibucketptr->nextibucket;
	tempibucketptr->nextibucket = freeddibucketslist;
	RESET_freeddibucketslist(holdingbindvectorptr[k]->dds.cs.ibucketptr);
	holdingbindvectorptr[k]->dds.cs.ibucketptr = NULL;
      }
      tempvbucketptr = holdingbindvectorptr[k]->dds.cs.vbucketptr;
      if (tempvbucketptr != NULL) {
#ifdef _DEBUG
	if (governingdstype == SS) {
	  printf("dspgx warning 7.3: Unexpcted non-NULL");
	  printf(" vbucketptr.\n");
	}
#endif
	while (tempvbucketptr->nextvbucket != NULL)
	  tempvbucketptr = tempvbucketptr->nextvbucket;
	tempvbucketptr->nextvbucket = freeddvbucketslist;
	RESET_freeddvbucketslist(holdingbindvectorptr[k]->dds.cs.vbucketptr);
	holdingbindvectorptr[k]->dds.cs.vbucketptr = NULL;
      }
    }
  } /* matches    for (k=0;k<arity;k++) {    */
}








/****************************************************************************
 *                                                                          *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/

void dplusequal_axpy_postprocessing(destptrptr,
				   rhsdvectorptrptr,
				   status_otherrhsdvector,
				   restofplusequalvectorptrptr,
				   alphaptrs,
				   restorerhsdvectorptr,
				   arity,
				   lhsdvectornonzeroes,
				   lhsdvectornonzeroranges,
				   governingdstype)
dvectorstruct **destptrptr;
dvectorstruct **rhsdvectorptrptr;
int status_otherrhsdvector;
dvectorstruct **restofplusequalvectorptrptr;
double *alphaptrs[2];
dvectorstruct *restorerhsdvectorptr[2+1];
int arity;
int lhsdvectornonzeroes;
int lhsdvectornonzeroranges;
int governingdstype;
{
  dvectorstruct *rhsdvectorptr[2];
  dvectorstruct *tempvectorptr;
  dbucketstruct *tempbucketptr;
  dibucketstruct *tempibucketptr;
  dvbucketstruct *tempvbucketptr;

  int k;

  k = 1;
    rhsdvectorptr[k] = *rhsdvectorptrptr;

    /****************************************
     * Switch lhsdvectorptr and *destptrptr. *
     ****************************************/

    lhsdvectorptr->nonzeroes = lhsdvectornonzeroes;
    lhsdvectorptr->dds.cs.nonzeroranges = lhsdvectornonzeroranges;
  

    /*********************************************************
     * If *restofplusequalvectorptrptr was not empty,        *
     * add its buckets to freeddbucketslist.                  *
     *********************************************************/
  
    tempvectorptr = (*restofplusequalvectorptrptr);
    tempibucketptr = tempvectorptr->dds.cs.ibucketptr;
    if (tempibucketptr != NULL) {
      while (tempibucketptr->nextibucket != NULL)
	tempibucketptr = tempibucketptr->nextibucket;
      tempibucketptr->nextibucket = freeddibucketslist;
      RESET_freeddibucketslist(tempvectorptr->dds.cs.ibucketptr);
      tempvectorptr->dds.cs.ibucketptr = NULL;
    }
    tempvbucketptr = tempvectorptr->dds.cs.vbucketptr;
    if (tempvbucketptr != NULL) {
      while (tempvbucketptr->nextvbucket != NULL)
	tempvbucketptr = tempvbucketptr->nextvbucket;
      tempvbucketptr->nextvbucket = freeddvbucketslist;
      RESET_freeddvbucketslist(tempvectorptr->dds.cs.vbucketptr);
      tempvectorptr->dds.cs.vbucketptr = NULL;
    }

   
  /*****************************************
   * If status_otherrhsdvector is TEMP,     *
   * add its buckets to freeddbucketslist,  *
   * and reassign pointers.                *
   *****************************************/

  if (status_otherrhsdvector == TEMP_OTHER_VECTOR) {

    tempvectorptr = tempddestptr;
    tempibucketptr = tempvectorptr->dds.cs.ibucketptr;
    if (tempibucketptr != NULL) {
      while (tempibucketptr->nextibucket != NULL)
	tempibucketptr = tempibucketptr->nextibucket;
      tempibucketptr->nextibucket = freeddibucketslist;
      RESET_freeddibucketslist(tempvectorptr->dds.cs.ibucketptr);
      tempvectorptr->dds.cs.ibucketptr = NULL;
    }
    tempvbucketptr = tempvectorptr->dds.cs.vbucketptr;
    if (tempvbucketptr != NULL) {
      while (tempvbucketptr->nextvbucket != NULL)
	tempvbucketptr = tempvbucketptr->nextvbucket;
      tempvbucketptr->nextvbucket = freeddvbucketslist;
      RESET_freeddvbucketslist(tempvectorptr->dds.cs.vbucketptr);
      tempvectorptr->dds.cs.vbucketptr = NULL;
    }

  }
 
  /*****************************************
   * If holdingbins were used, add         *
   * their buckets to freeddbucketslist, OR *
   * their ibuckets to freeddibucketslist & *
   * their vbuckets to freeddvbucketslist.  *
   *****************************************/

  /*  We could not have put rhsdvectorptr[0] in a holdingbin;
   *  but there may be something in it from the routine that called
   *  plusequal (in particular dspg2 (dspgx does it's own 
   *  holdingbin cleaning).
   */

  for (k=0;k<arity;k++) {

    rhsdvectorptr[k] = restorerhsdvectorptr[k];
    
    if (holdingbindvectorptr[k]->dstype == SS) {
      tempbucketptr = holdingbindvectorptr[k]->dds.ss.bucketptr;
      if (tempbucketptr != NULL) {
#ifdef _DEBUG
	if (governingdstype == CS) {
	  printf("dspgx warning 7.4: Unexpcted non-NULL");
	  printf("tempbucketptr.\n");
	}
#endif
	while (tempbucketptr->nextbucket != NULL) {
#ifdef _DEBUG
	  if (governingdstype == SS) {
	    printf("dspgx warning 7.1: tempbucketptr should have");
	    printf(" no trailing nextbuckets\n");
	  }
#endif
	  tempbucketptr = tempbucketptr->nextbucket;
	}
	tempbucketptr->nextbucket = freeddbucketslist;
	RESET_freeddbucketslist(holdingbindvectorptr[k]->dds.ss.bucketptr);
	holdingbindvectorptr[k]->dds.ss.bucketptr = NULL;
      }
    }
    else if (holdingbindvectorptr[k]->dstype == CS) {
      tempibucketptr = holdingbindvectorptr[k]->dds.cs.ibucketptr;
      if (tempibucketptr != NULL) {
#ifdef _DEBUG
	if (governingdstype == SS) {
	  printf("dspgx warning 7.2: Unexpcted non-NULL");
	  printf(" ibucketptr.\n");
	}
#endif
	while (tempibucketptr->nextibucket != NULL)
	  tempibucketptr = tempibucketptr->nextibucket;
	tempibucketptr->nextibucket = freeddibucketslist;
	RESET_freeddibucketslist(holdingbindvectorptr[k]->dds.cs.ibucketptr);
	holdingbindvectorptr[k]->dds.cs.ibucketptr = NULL;
      }
      tempvbucketptr = holdingbindvectorptr[k]->dds.cs.vbucketptr;
      if (tempvbucketptr != NULL) {
#ifdef _DEBUG
	if (governingdstype == SS) {
	  printf("dspgx warning 7.3: Unexpcted non-NULL");
	  printf(" vbucketptr.\n");
	}
#endif
	while (tempvbucketptr->nextvbucket != NULL)
	  tempvbucketptr = tempvbucketptr->nextvbucket;
	tempvbucketptr->nextvbucket = freeddvbucketslist;
	RESET_freeddvbucketslist(holdingbindvectorptr[k]->dds.cs.vbucketptr);
	holdingbindvectorptr[k]->dds.cs.vbucketptr = NULL;
      }
    }
  }
}





/****************************************************************************
 *                                                                          *
 *  ddetermine_loop_bounds                                                   *
 *                                                                          *
 *     Based on how close lhsvbucketindex and rhsvbucketindex pointers are  *
 *     to the end of the vbucket, and based on the number of values left to *
 *     process in the axpy, ddetermine_loop_bounds figures out the various   *
 *     loop values.                                                         *
 *                                                                          *
 ****************************************************************************/


void ddetermine_loop_bounds(totalcurrentnonzerorange,
			   lhsvbucketindex,
			   rhsvbucketindex,
			   firstloopsize,
			   numberofdoubleforloopiterations,
			   firstiterationloopsize,
			   seconditerationloopsize,
			   remainder1loopsize,
			   remainder2loopsize
#ifdef _DEBUG
			   ,addressofunconsumedcurrentnonzerorange
#endif
			   )
int totalcurrentnonzerorange;
int lhsvbucketindex;
int rhsvbucketindex;
int *firstloopsize;
int *numberofdoubleforloopiterations;
int *firstiterationloopsize;
int *seconditerationloopsize;
int *remainder1loopsize;
int *remainder2loopsize;
#ifdef _DEBUG
int *addressofunconsumedcurrentnonzerorange;
#endif
{
  int temprange, tmpmod;

  *firstloopsize =
    min(((dVBUCKET_DIM - max(lhsvbucketindex,rhsvbucketindex))),
	totalcurrentnonzerorange);
  temprange = totalcurrentnonzerorange - *firstloopsize;
  *numberofdoubleforloopiterations = temprange / dVBUCKET_DIM;
  *firstiterationloopsize =
    min((dVBUCKET_DIM -
	 (*firstloopsize + min(lhsvbucketindex,rhsvbucketindex))),temprange);
  *seconditerationloopsize =
    min((dVBUCKET_DIM - *firstiterationloopsize),
	(temprange - *firstiterationloopsize));
  tmpmod = MY_MOD(temprange,dVBUCKET_DIM_minus1);
  *remainder1loopsize = min (tmpmod,*firstiterationloopsize);
  *remainder2loopsize = max ((tmpmod - (*firstiterationloopsize)),0);
  
#ifdef _DEBUG
  *addressofunconsumedcurrentnonzerorange =  totalcurrentnonzerorange;
#endif

}

