#define COMPLEX
/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#include  <stdlib.h>
#include "dgaxpy_pd_macros_and_data_structures.h"
#include "cgaxpy_pd_macros_and_data_structures.h"
#include "dgaxpy_pd_global_variables.h"
#include "cgaxpy_pd_global_variables.h"


/*
 *  Function Prototyping done locally to grab "TO_pREc_" coorectly.
 */

#define _P(x) ()

void dsp2c _P((cvectorstruct **destptrptr,
				       dvectorstruct **realvectorptrptr,
				       dvectorstruct **imagvectorptrptr
				       ));

/****************************************************************************
 *                                                                          *
 *    Patterned after dspg2                                                *
 *                                                                          *
 ****************************************************************************/

void dsp2c(destptrptr,realvectorptrptr,imagvectorptrptr)
cvectorstruct **destptrptr;
dvectorstruct **realvectorptrptr;
dvectorstruct **imagvectorptrptr;
{

  int arity = 2;
  
  dvectorstruct *rhsvectorptr[2];
  
  dbucketstruct *rhsbucketptr[2];
  
  cbucketstruct *lhscbucketptr;
  
  int i;
  
  
  dibucketstruct *rhsibucketptr[2];
                                      /* Array of pointers to "current"     *
				       * buckets where "current" nonzeroes  *
				       * of the rhs vectors reside.         */

  cibucketstruct *lhscibucketptr; 
                                      /* Pointer to current bucket of       *
				       * lhscvector.                 */

  int rhsbucketindex[2];              /* Array of indices of current        *
				       * nonzeroes in the current buckets   *
				       * of rhs vectors.                    */

  int lhscbucketindex;         /* bucketindex of next nonzero to be  *
				       * added to lhscvector.        */

  int lastlhscbucketindex;     /* bucketindex of the latest added    *
				       * nonzero in the lhscvector.  *
				       * We need this additional index to   *
				       * leap-frogbuckets, if need be.      */


  int rhsibucketindex[2];             /* Array of indices of current        *
				       * nonzeroes in the current buckets   *
				       * of rhs vectors.                    */

  int lhscibucketindex;        /* bucketindex of next nonzero to be  *
				       * added to lhscvector.        */

  int lastlhscibucketindex;    /* bucketindex of the latest added    *
				       * nonzero in the lhscvector.  *
				       * We need this additional index to   *
				       * leap-frogbuckets, if need be.      */

  int firstsaxpyoperand;              /* Order in the ennumeration of the   *
				       * rhs vectors of the first non-null  *
				       * saxpy vectorop.                    */

  int nullsaxpyflag = 1;              /* == 1 means no non-null vectorops.  */

  int k;                              /* Counter for arity.                 */

  int j;                              /* Index for vectorsubscripts to be   *
				       * entered in (and later popped from) *
				       * the heap.                          */

  int rhsvectornonzeroes[2];          /* Array of total # of nonzeroes in   *
				       * each rhs vector.                   */

  int rhsvectorconsumednonzeroes[2];  /* Array of # of nonzeroes consumed   *
				       * so far in each rhs vector.         */

  int lhscvectornonzeroes;     /* Count of # of nonzeroes entered    *
				       * into the lhscvector so far. *
				       * By the end of the saxpy operation, *
				       * this index equals the total # of   *
				       * nonzeroes in the resultant vector. */

/*
 * These variables were originally for COMPRESSED_SUBSCRIPTS
 */

  dvbucketstruct *rhsvbucketptr[2];
                                      /* Array of pointers to "current"     *
				       * vbuckets where "current" nonzeroes *
				       * of the rhs vectors reside.         */


  cvbucketstruct *lastlhscvbucketptr;
                                      /* Pointer to current vbucket of      *
				       * lhscvector.                 */


  int rhsvectornonzeroranges[2];
                                      /* Array of total # of nonzeroes      *
				       * ranges in each rhs vector.         */

  int rhsvectorconsumednonzeroranges[2];
                                      /* Array of # of nonzeroes consumed   *
				       * so far in each rhs vector.         */

  int rhsvbucketindex[2];             /* Array of indices of current        *
				       * nonzeroes in the current buckets   *
				       * of rhs vectors.                    */

  int lhscvectornonzeroranges; /* Count of # of nonzeroes entered    *
				       * into the lhscvector so far. *
				       * By the end of the saxpy operation, *
				       * this index equals the total # of   *
				       * nonzeroes in the resultant vector. */

  int lhscvbucketindex;        /* bucketindex of next nonzero to be  *
				       * added to lhscvector.        */


  int lastlhscvbucketindex;    /* bucketindex of the latest added    *
				       * nonzero in the lhscvector.  *
				       * We need this additional index to   *
				       * leap-frog buckets, if need be.     */

  cvbucketstruct *templhscvbucketptr;
  int templhscvbucketindex;

  int j1;                             /* Index for vectorsubscripts to be   *
				       * entered in (and later popped from) *
				       * the heap.                          */

  int j2;                             /* Index for vectorsubscripts to be   *
				       * entered in (and later popped from) *
				       * the heap.                          */

  int h1;
  int h2;

  int totalcurrentnonzerorange;

#ifdef _DEBUG
  int unconsumedcurrentnonzerorange;
#endif

  int temp1range;
  int distancetooverlap;
  int distancetoendofvbucket;


  /*
   * These variables were added for union ds
   */
  
  dvectorstruct *restorerhsvectorptr[2];
  int governingdstype;
  
  /*
   * These variables were added for plusequal code
   */
  
  dvectorstruct **temprhsvectorptrptrs[MAXARITY-1];
  
  
  dvectorstruct **rhsvectorptrptrs[MAXARITY];
  
  int condition[2];
  
  
  
  /****************************************
   * Initialization of *destptrptr.       *
   ****************************************/
  
  if (*destptrptr == NULL)
    cgetvector(destptrptr);
  
  /****************************************
   * Dereferencing.                       *
   ****************************************/

  rhsvectorptr[0] = *realvectorptrptr;
  rhsvectorptr[1] = *imagvectorptrptr;
 
  for (k=0;k<arity;k++) {
    restorerhsvectorptr[k] = rhsvectorptr[k];
    rhsvectornonzeroes[k] = rhsvectorptr[k]->nonzeroes;
    if (rhsvectorptr[k]->dstype == CS)
      rhsvectornonzeroranges[k] = rhsvectorptr[k]->dds.cs.nonzeroranges;
  }
  
  if (rhsvectorptr[0]->nonzeroes == 0 &&
      rhsvectorptr[1]->nonzeroes == 0) {
    cspzronotca(destptrptr);
  }
  else {
    governingdstype = max(rhsvectorptr[0]->dstype,
			  rhsvectorptr[1]->dstype);
    if (governingdstype == S1 &&
	rhsvectorptr[0]->dds.s1.index != rhsvectorptr[1]->dds.s1.index)
      governingdstype = SS;
    for (k=0;k<arity;k++)
      if (rhsvectorptr[k]->dstype != governingdstype) {
	dconvertdstype((&rhsvectorptr[k]),k,governingdstype);
	if (governingdstype == CS)
	  rhsvectornonzeroranges[k] =
	    rhsvectorptr[k]->dds.cs.nonzeroranges;
      }
    
    if (governingdstype == S1) {
      
      /******************************************************************
       *                      ONE_SUBSCRIPT AXPY                        *
       ******************************************************************/
      
      lhscvectorptr->nonzeroes =  1;
      lhscvectorptr->dstype = governingdstype;
      lhscvectorptr->cds.s1.index =
	rhsvectorptr[0]->dds.s1.index;
      XY_PUT_AND_CAST(lhscvectorptr->cds.s1.value,
		      rhsvectorptr[0]->dds.s1.value,
		      rhsvectorptr[1]->dds.s1.value,
		      float);

      caxpy_0_1_post_processing(destptrptr);

    }
    else /* if we are going to use lhscvectorptr */ {
      if (governingdstype == SS) {
	
	/****************************************************************
	 *                 _SINGLE_SUBSCRIPTS AXPY                      *
	 ****************************************************************/
	for (k=0;k<arity;k++) {
	  rhsvectorconsumednonzeroes[k]= 0;
	  rhsbucketindex[k] = 0;
	  rhsbucketptr[k] = rhsvectorptr[k]->dds.ss.bucketptr;
	}
	cgetbucket(&lhscbucketptr);
	lhscvectorptr->dstype = governingdstype;
	lhscvectorptr->cds.ss.bucketptr = lhscbucketptr;
	lhscvectornonzeroes = 0;
	
	for (k=0;k<arity;k++)
	  condition[k] = rhsvectorconsumednonzeroes[k] <
	      rhsvectorptr[k]->nonzeroes;
	
	while (condition[0] || condition[1]) {
	  if (condition[0] && condition[1])
	    if (rhsbucketptr[0]->index[rhsbucketindex[0]] <=
		rhsbucketptr[1]->index[rhsbucketindex[1]])
	      k = REAL_PART;
	    else
	      k = IMAG_PART;
	  else if (condition[0])
	    k = REAL_PART;
	  else
	    k = IMAG_PART;
	  
	  j = rhsbucketptr[k]->index[rhsbucketindex[k]];
	  
	  lhscbucketindex = MY_MOD(lhscvectornonzeroes,
					  dBUCKET_DIM_minus1);
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.1. Found 1st or new subscript for  *
	   *      lhscvector.  If needed,  *
	   *      get new bucket for              *
	   *      lhscvector.              *
	   *      Consume nonzero in saxpy.       *
	   ****************************************/
	  if (lhscvectornonzeroes == 0 ||
	      lhscbucketptr->index[lastlhscbucketindex] < j) {
	    if (lhscbucketindex == 0 &&
		lhscvectornonzeroes != 0) {
	      cgetbucket(&(lhscbucketptr->nextbucket));
	      lhscbucketptr = lhscbucketptr->nextbucket;
	    }
	    lhscbucketptr->index[lhscbucketindex] = j;
	    if (k == REAL_PART) {
	      XY_PUT_AND_CAST(lhscbucketptr->value[lhscbucketindex],
			      rhsbucketptr[k]->value[rhsbucketindex[k]],
			      0,
			      float);
	    }
	    else {
	      XY_PUT_AND_CAST(lhscbucketptr->value[lhscbucketindex],
			      0,
			      rhsbucketptr[k]->value[rhsbucketindex[k]],
			      float);
	    }
	    lastlhscbucketindex = lhscbucketindex;
	    lhscvectornonzeroes++;
	  }
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.2. Found old subscript for         *
	   *      lhscvector.                      *
	   *      Consume nonzero in saxpy.       *
	   ****************************************/
	  else if (lhscbucketptr->index[lastlhscbucketindex] == j) {
	    if (k == REAL_PART) {
	      XY_PUT_AND_CAST_PLUS_EQUALS(lhscbucketptr->value[lastlhscbucketindex],
					  rhsbucketptr[k]->value[rhsbucketindex[k]],
					  0,
					  float);
	    }
	    else {
	      XY_PUT_AND_CAST_PLUS_EQUALS(lhscbucketptr->value[lastlhscbucketindex],
					  0,
					  rhsbucketptr[k]->value[rhsbucketindex[k]],
					  float);
	    }

	  }
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.3. Found wrong subscript for       *
	   *      lhscvector.  Abort.       *
	   ****************************************/
	  
#ifdef _DEBUG
	  else {
	    printf("dspg2 error 6: lhscnz=%d, lastlhscindex[]=%d, j=%d\n",
		   lhscvectornonzeroes,
		   lhscbucketptr->index[lastlhscbucketindex],j);
	    abort();
	  }
#endif
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.4. If there is one, get new        *
	   *      nonzero from the rhs vector     *
	   *      whose nonzero was just          *
	   *      consumed and replace the        *
	   *      heaproot with new entry;        *
	   *      Else, delete heaproot.          *
	   ****************************************/
	  rhsvectorconsumednonzeroes[k]++;
	  condition[k] = rhsvectorconsumednonzeroes[k] <
	    rhsvectorptr[k]->nonzeroes;
	  rhsbucketindex[k] =
	    MY_MOD(rhsvectorconsumednonzeroes[k],dBUCKET_DIM_minus1);
	  if (condition[k]) {
	    if (rhsbucketindex[k]  == 0) {
	      if ((rhsbucketptr[k] = rhsbucketptr[k]->nextbucket) 
		  == NULL) {
#ifdef _DEBUG
		printf("dspg2 err: unexpected NULL nextbucket: k=%d",k);
		printf("rhsvectorconsumednonzeros=%d,rhsvecnonzeroes=%d,",
		       rhsvectorconsumednonzeroes[k],
		       rhsvectornonzeroes[k]);
		abort();
#endif
	      }	  
	    }
	  }
	} /* matches     while loop     */
      } /* matches     else if (governingdstype == SS) {      */
      
      else /* if (governingdstype == CS) */ {
	
	/****************************************************************
	 *               _COMPRESSED_SUBSCRIPTS AXPY                    *
	 ****************************************************************/
	
	/****************************************
	 * Else perform saxpy:                  *
	 * 1. Put in heap the 1st nonzero of    *
	 *    each vector.                      *
	 ****************************************/
	for (k=0;k<arity;k++) {
	  rhsvectorconsumednonzeroranges[k] = 0;
	  rhsibucketindex[k] = 0;
	  rhsvbucketindex[k] = 0;
	  rhsibucketptr[k] =
	    rhsvectorptr[k]->dds.cs.ibucketptr;
	  rhsvbucketptr[k] =
	    rhsvectorptr[k]->dds.cs.vbucketptr;
	  
#ifdef _DEBUG
	  if (rhsibucketptr[k] == NULL ||
	      rhsvbucketptr[k] == NULL) {
	    printf("dspg2 error 10: Unexpected NULL rhsbucketptrs: ");
	    printf("vector # %d\n",k);
	    abort();
	  }
#endif
	}	  
	
	/****************************************
	 * Perform saxpy:                       *
	 * 2. Initialize lhscvector.             *
	 ****************************************/
	cgetibucket(&lhscibucketptr);
	lhscvectorptr->dstype = governingdstype;
	lhscvectorptr->cds.cs.ibucketptr = lhscibucketptr;
	cgetvbucket(&lhscvbucketptr);
	lhscvectorptr->cds.cs.vbucketptr = lhscvbucketptr;
	lhscvectornonzeroranges = 0;
	lhscvectornonzeroes = 0;
	lhscvbucketindex = 0;
	
	/****************************************
	 * Perform saxpy:                       *
	 * 3. Get nonzero by popping the        *
	 *    heaproot; consume it in saxpy;    *
	 *    either replace or delete heaproot.*
	 ****************************************/
	
	while (rhsvectorconsumednonzeroranges[0] <
	       rhsvectorptr[0]->dds.cs.nonzeroranges ||
	       rhsvectorconsumednonzeroranges[1] <
	       rhsvectorptr[1]->dds.cs.nonzeroranges) {
	  if
	    ((rhsvectorconsumednonzeroranges[1] ==
	      rhsvectorptr[1]->dds.cs.nonzeroranges) ||
	     ((rhsvectorconsumednonzeroranges[0] <
	       rhsvectorptr[0]->dds.cs.nonzeroranges) &&
	      (rhsibucketptr[0]->index_RANGE_BEGIN[rhsibucketindex[0]] <=
	       rhsibucketptr[1]->index_RANGE_BEGIN[rhsibucketindex[1]])))
	      k = REAL_PART;
	  else if
	    ((rhsvectorconsumednonzeroranges[0] ==
	      rhsvectorptr[0]->dds.cs.nonzeroranges) ||
	     ((rhsvectorconsumednonzeroranges[1] <
	       rhsvectorptr[1]->dds.cs.nonzeroranges) &&
	      (rhsibucketptr[0]->index_RANGE_BEGIN[rhsibucketindex[0]] >
	       rhsibucketptr[1]->index_RANGE_BEGIN[rhsibucketindex[1]])))
	      k = IMAG_PART;
#ifdef _DEBUG
	  else {
	    printf("dspg2 error: BAD LOGIC.\n");
	    abort();
	  }
#endif
	  j1 = rhsibucketptr[k]->index_RANGE_BEGIN[rhsibucketindex[k]];
	  j2 = rhsibucketptr[k]->index_RANGE_END[rhsibucketindex[k]];
	  rhsibucketindex[k] =
	    MY_MOD(rhsvectorconsumednonzeroranges[k],dIBUCKET_DIM_minus1);
	  lhscibucketindex =
	    MY_MOD(lhscvectornonzeroranges,dIBUCKET_DIM_minus1);
	  
	  if ((rhsvbucketindex[k] =
	       MY_MOD(rhsvbucketindex[k],dVBUCKET_DIM_minus1))==0
	      && rhsvectorconsumednonzeroranges[k] !=0) {
	    dHOP_TO_NEXT_RHSVBUCKET(k,&rhsvbucketindex[k],
				   rhsvectorconsumednonzeroranges[k],
				   rhsvectornonzeroranges[k]);
	  }
	  lhscvbucketindex = MY_MOD(lhscvbucketindex,
					   dVBUCKET_DIM_minus1);
	  
#ifdef _DEBUG
	  if (rhsibucketptr[k]->errorflag != 9099 ||
	      rhsvbucketptr[k]->errorflag != 9009) {
	    printf("dspg2 err 14: memory corruption of ibucketstore ");
	    printf("or vbucketstore.\n");
	    abort();
	  }
#endif      
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.0. Check for startup errors.       *
	   ****************************************/
	  
#ifdef _DEBUG
	  if (lhscvectornonzeroranges == 0) {
	    if (rhsibucketindex[k] != 0 || rhsvbucketindex[k] != 0 ||
		lhscibucketindex != 0 || lhscvbucketindex != 0) {
	      printf("dspg2 error 15: Zeroes expected: \n\t");
	      printf("rhsibucketindex[k]=%d, rhsvbucketindex[k]=%d\n\t",
		     rhsibucketindex[k], rhsvbucketindex[k]);
	      printf("lhscibucketindex=%d, lhscvbucketindex=%d\n\t",
		     lhscibucketindex,lhscvbucketindex);
	      abort();
	    }
	  }
#endif      
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.1. Found 1ST or NEW subscript      *
	   *      range (non-overlapping with     *
	   *      previous one, if any) for       *
	   *      lhscvector.  If needed, get new  *
	   *      ibuckets and vbuckets for       *
	   *      lhscvector.                      *
	   ****************************************/
	  if (lhscvectornonzeroranges == 0 ||
	      j1 > lhscibucketptr->index_RANGE_END[lastlhscibucketindex]+1) {/*A.1*/
	    
	    if (lhscibucketindex == 0 &&
		lhscvectornonzeroranges != 0) {
	      cgetibucket(&(lhscibucketptr->nextibucket));
	      lhscibucketptr = lhscibucketptr->nextibucket;
	    }
	    lhscibucketptr->index_RANGE_BEGIN[lhscibucketindex] = j1;
	    lhscibucketptr->index_RANGE_END[lhscibucketindex] = j2;
	    
	    if (lhscvbucketindex == 0 &&
		lhscvectornonzeroranges != 0)
	      dget_and_hop_to_next_to_c_vbucket(&lhscvbucketindex);
	    
	    lastlhscvbucketptr = lhscvbucketptr;
	    lastlhscvbucketindex = lhscvbucketindex;
	    
	    totalcurrentnonzerorange = j2-j1+1;
	    
	    lhscvectornonzeroes += totalcurrentnonzerorange;
	    lhscvectornonzeroranges++;
	    
	    
	    dNON_OVER_LAPPING_CS_RANGE_CONVERT_TO_c(COMPLETELY_NEW_RANGE,
					      lhscibucketindex,
					      &lastlhscibucketindex,
					      &lhscvbucketindex,
					      k,
					      &rhsvbucketindex[k],
					      totalcurrentnonzerorange,
					      &unconsumedcurrentnonzerorange,
					      rhsvectorconsumednonzeroranges[k],
					      rhsvectornonzeroranges[k],
					      lhscvectornonzeroranges);
	    
	    
	  } /*A.1*/
	  
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.2. Found old subscript for         *
	   *      lhscvector.  Consume  nonzero in *
	   *      saxpy.                          *
	   *                                      *
	   ****************************************/
	  else if (j1 <= 
		   lhscibucketptr->index_RANGE_END[lastlhscibucketindex]+1 &&
		   j1 >= 
		   lhscibucketptr->index_RANGE_BEGIN[lastlhscibucketindex]) { /*B.1*/
	    
	    templhscvbucketptr = lhscvbucketptr;
	    templhscvbucketindex = lhscvbucketindex;
	    lhscvbucketptr = lastlhscvbucketptr;
	    
	    distancetooverlap =
	      j1 - lhscibucketptr->index_RANGE_BEGIN[lastlhscibucketindex];
	    distancetoendofvbucket = dVBUCKET_DIM -
	      lastlhscvbucketindex;
	    
	    if (j1 ==
		lhscibucketptr->index_RANGE_END[lastlhscibucketindex]+1) {
	      h1 = 0;
	      h2 = -1;
	    }
	    else if (j2 <=
		     lhscibucketptr->index_RANGE_END[lastlhscibucketindex]) {
	      h1 = j1;
	      h2 = j2;
	      j1 = 0;
	      j2 = -1;
	    }
	    else {
	      h1 = j1;
	      h2 = lhscibucketptr->index_RANGE_END[lastlhscibucketindex];
	      j1 = h2+1;
	    }
	    
	    if (distancetoendofvbucket > distancetooverlap) {
	      lhscvbucketindex =
		MY_MOD(lastlhscvbucketindex + distancetooverlap,
		       dVBUCKET_DIM_minus1);
	    }
	    else if (distancetoendofvbucket < distancetooverlap) {
	      lhscvbucketindex =
		MY_MOD(distancetooverlap - distancetoendofvbucket,
		       dVBUCKET_DIM_minus1);
	    }
	    else {
	      lhscvbucketindex = 0;
	    }
	    
	    while (distancetooverlap >= distancetoendofvbucket &&
		   lhscvbucketptr->nextvbucket != NULL) {
	      lhscvbucketptr = lhscvbucketptr->nextvbucket;
	      distancetooverlap -= distancetoendofvbucket;
	      distancetoendofvbucket = dVBUCKET_DIM;
	    }
	    
	    lhscibucketptr->index_RANGE_END[lastlhscibucketindex] =
	      max(j2,lhscibucketptr->index_RANGE_END[lastlhscibucketindex]);
	    
	    if ((totalcurrentnonzerorange = h2-h1+1) > 0) {
	      dOVERLAPPING_CS_RANGE_CONVERT_TO_c(&lhscvbucketindex,
					   k,
					   &rhsvbucketindex[k],
					   totalcurrentnonzerorange,
					   &unconsumedcurrentnonzerorange,
					   rhsvectorconsumednonzeroranges[k],
					   rhsvectornonzeroranges[k],
					   lhscvectornonzeroranges);
	    }
	    
	    /*
	      
	      This bit of logic must be performed, whether or not there
	      is additional nonzeroes to process.
	      */
	    
	    if (j2 != -1 &&
		(lhscvbucketptr != templhscvbucketptr ||
		 (lhscvbucketindex =
		  MY_MOD(lhscvbucketindex,dVBUCKET_DIM_minus1)) !=
		 templhscvbucketindex)) {
	      
#ifdef _DEBUG
	      printf("dspg2 err 50.0: Unexpectd value discrepancies\n");
	      printf("lhscvbucketptr=%d, templhscvbucketptr=%d\n",
		     (int)lhscvbucketptr,
		     (int)templhscvbucketptr);
	      printf("lhscvbucketindex=%d, templhscvbucketindex=%d\n",
		     lhscvbucketindex,templhscvbucketindex);
	      
	      abort();
#endif
	    }
	    else {
	      lhscvbucketptr = templhscvbucketptr;
	      lhscvbucketindex = templhscvbucketindex;
	    }
	    
	    
	    if ((totalcurrentnonzerorange = j2-j1+1) > 0) {
	      lhscvectornonzeroes += totalcurrentnonzerorange;
	      
	      /*
	       * adding new entries to vbucket
	       */
	      
	      if (lhscvbucketindex == 0 &&
		  lhscvectornonzeroranges != 0 &&
		  j2 != -1)
		dget_and_hop_to_next_to_c_vbucket(&lhscvbucketindex);
	      
	      if(rhsvbucketindex[k] == dVBUCKET_DIM && j2 != -1) {
		
		dHOP_TO_NEXT_RHSVBUCKET(k,&rhsvbucketindex[k],
				       rhsvectorconsumednonzeroranges[k],
				       rhsvectornonzeroranges[k]);
	      }
	      
	      
	      
	      
	      dNON_OVER_LAPPING_CS_RANGE_CONVERT_TO_c(NEW_ENTRIES_PADDED_TO_OLD_RANGE,
						lhscibucketindex,
						&lastlhscibucketindex,
						&lhscvbucketindex,
						k,
						&rhsvbucketindex[k],
						totalcurrentnonzerorange,
						&unconsumedcurrentnonzerorange,
						rhsvectorconsumednonzeroranges[k],
						rhsvectornonzeroranges[k],
						lhscvectornonzeroranges);
	      
	      
	    } /* matches     if ((totalcurrent...) { */
	  } /*B.1*/
	  
	  
	  /****************************************	
	   * Perform saxpy:                       *
	   * 3.3. Found wrong subscript for       *
	   *      lhscvector.  Abort.               *
	   ****************************************/
	  
#ifdef _DEBUG
	  else {
	    printf("dspg2 error 60: lhscvectornonzeroranges=%d,",
		   lhscvectornonzeroranges);
	    printf("lastlhscibucketindex=%d, \n\t",lastlhscibucketindex);
	    printf("lhscibucketptr->index[%d] = (%d,%d), j1=%d\n",
		   lastlhscibucketindex,
		   lhscibucketptr->index_RANGE_BEGIN[lastlhscibucketindex],
		   lhscibucketptr->index_RANGE_END[lastlhscibucketindex],
		   j1);
	    abort();
	  }
#endif
	  
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.4. If there is one, get new        *
	   *      nonzero from the rhs vector     *
	   *      whose nonzero was just          *
	   *      consumed and replace the        *
	   *      heaproot with new entry;        *
	   *      Else, delete heaproot.          *
	   ****************************************/
	  if (++rhsvectorconsumednonzeroranges[k] <
	      rhsvectornonzeroranges[k]) {
	    if ((rhsibucketindex[k] =
		 MY_MOD(rhsvectorconsumednonzeroranges[k],
			dIBUCKET_DIM_minus1)) == 0) {
	      if ((rhsibucketptr[k] = rhsibucketptr[k]->nextibucket) ==
		  NULL) {
		
#ifdef _DEBUG
		printf("dspg2 err 61: unexpected NULL nextvbucket: \n");
		printf("\t k=%d, rhsvectorconsumednonzeroranges=%d,\n",
		       k,rhsvectorconsumednonzeroranges[k]);
		printf("\t rhsvectornonzeroranges[k]=%d\n",
		       rhsvectornonzeroranges[k]);
		abort();
#endif
		
	      }
	    }
	  }   
	} /* matches     while loop          */
      } /* matches     else ** if (governingdstype == CS) ** {     */


      rhsvectorptrptrs[0] = &rhsvectorptr[0];
      rhsvectorptrptrs[1] = &rhsvectorptr[1];
      
      daxpy_2_heap_to_c_post_processing(destptrptr,
					 rhsvectorptrptrs,
					 restorerhsvectorptr,
					 arity,
					 lhscvectornonzeroes,
					 lhscvectornonzeroranges,
					 governingdstype);
      
    } /* matches     if we are going to use lhscvectorptr       */
  }
}





