#define DOUBLE_COMPLEX
/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#include  <stdlib.h>
#include "dgaxpy_pd_macros_and_data_structures.h"
#include "zgaxpy_pd_macros_and_data_structures.h"
#include "dgaxpy_pd_global_variables.h"
#include "zgaxpy_pd_global_variables.h"


/*
 *  Function Prototyping done locally to grab "TO_pREc_" correctly.
 */

#define _P(x) ()

void dnon_over_lapping_cs_range_convert_to_z _P((
				  int rangedestination,
				  int lhszibucketindex,
				  int *addressoflastlhszibucketindex,
				  int *addressoflhszvbucketindex,
				  int k,
				  int *addressofrhsvbucketindex_k,
				  int totalcurrentnonzerorange
#ifdef _DEBUG
				  ,int *addressofunconsumedcurrentnonzerorange
				  ,int rhsvectorconsumednonzeroranges_k
				  ,int rhsvectornonzeroranges_k
				  ,int lhszvectornonzeroranges
#endif
				  ));

void doverlapping_cs_range_convert_to_z _P((
			     int *addressoflhszvbucketindex,
			     int k,
			     int *addressofrhsvbucketindex_k,
			     int totalcurrentnonzerorange
#ifdef _DEBUG
			     ,int *addressofunconsumedcurrentnonzerorange
			     ,int rhsvectorconsumednonzeroranges_k
			     ,int rhsvectornonzeroranges_k
			     ,int lhszvectornonzeroranges
#endif
			     ));


void ddense_convert_to_z_loop _P((int range,
					int *lhszvbucketindex,
					int *rhsdvbucketindex,
					int k,
					int loopsize
#ifdef _DEBUG
					,int *unconsumedcurrentnonzerorange
#endif
					));

void dget_and_hop_to_next_to_z_vbucket _P((int* lhszvbucketindex));

/*
 * Unnecessary code prototype
 *
void axpy_0_1_to_z_post_processing _P((
		    zvectorstruct **destptrptr));
 *
 * End of Unnecessary code prototype
 */

void daxpy_2_heap_to_z_post_processing _P((
		    zvectorstruct **destptrptr,
		    dvectorstruct **rhsdvectorptrptrs[],
		    dvectorstruct *restorerhsdvectorptr[MAXARITY+1],
		    int arity,
		    int lhsdvectornonzeroes,
		    int lhsdvectornonzeroranges,
		    int governingdstype));

/****************************************************************************
 *                                                                          *
 *      All rhs objects are d objects, whereas all lhs objects are      *
 *      z objects.  We omit the prefix d for the rhs objects     *
 *      everywhere except in the "dense-convert" routine, since every       *
 *      other rhs object passed is int except rhsdvbucketptrptr, which  *
 *      is a global and which is only used in the "dense-convert" routine.  *
 *                                                                          *
 ****************************************************************************/

void dnon_over_lapping_cs_range_convert_to_z(
				  rangedestination,
				  lhszibucketindex,
				  addressoflastlhszibucketindex,
				  addressoflhszvbucketindex,
				  k,
				  addressofrhsvbucketindex_k,
				  totalcurrentnonzerorange
#ifdef _DEBUG
				  ,addressofunconsumedcurrentnonzerorange
				  ,rhsvectorconsumednonzeroranges_k
				  ,rhsvectornonzeroranges_k
				  ,lhszvectornonzeroranges
#endif
				  )
int rangedestination;
int lhszibucketindex;
int *addressoflastlhszibucketindex;
int *addressoflhszvbucketindex;
int k;
int *addressofrhsvbucketindex_k;
int totalcurrentnonzerorange;
#ifdef _DEBUG
int *addressofunconsumedcurrentnonzerorange;
int rhsvectorconsumednonzeroranges_k;
int rhsvectornonzeroranges_k;
int lhszvectornonzeroranges;
#endif
{
  int firstloopsize;
  int numberofdoubleforloopiterations;
  int firstiterationloopsize;
  int seconditerationloopsize;
  int remainder1loopsize;
  int remainder2loopsize;

  int i;


  dDETERMINE_LOOP_BOUNDS(totalcurrentnonzerorange,
			*addressoflhszvbucketindex,
			*addressofrhsvbucketindex_k,
			&firstloopsize,
			&numberofdoubleforloopiterations,
			&firstiterationloopsize,
			&seconditerationloopsize,
			&remainder1loopsize,
			&remainder2loopsize,
			addressofunconsumedcurrentnonzerorange);
  dDENSE_CONVERT_TO_z_LOOP(NEW_RANGE,
				 addressoflhszvbucketindex,
				 addressofrhsvbucketindex_k,
				 k,
				 firstloopsize,
				 addressofunconsumedcurrentnonzerorange);
  
  
  /************************************************
   * Case 1: rhsdvbucket has more spaces      *
   *         remaining than lhszvbucket.   *
   ************************************************/
  
  if (*addressofrhsvbucketindex_k <
      *addressoflhszvbucketindex) {
    
    for(i=0;i<numberofdoubleforloopiterations;i++) {
      dDEBUGGING_CODE_1(LEFT,
		       *addressoflhszvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      dget_and_hop_to_next_vbucket(addressoflhszvbucketindex);
      dDENSE_CONVERT_TO_z_LOOP(NEW_RANGE,
				     addressoflhszvbucketindex,
				     addressofrhsvbucketindex_k,
				     k,
				     firstiterationloopsize,
				     addressofunconsumedcurrentnonzerorange);
      dDEBUGGING_CODE_1(RIGHT,
		       *addressofrhsvbucketindex_k,
		       *addressofunconsumedcurrentnonzerorange);
      dHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhsvbucketindex_k,
			     rhsvectorconsumednonzeroranges_k,
			     rhsvectornonzeroranges_k);
      dDENSE_CONVERT_TO_z_LOOP(NEW_RANGE,
				     addressoflhszvbucketindex,
				     addressofrhsvbucketindex_k,
				     k,
				     seconditerationloopsize,
				     addressofunconsumedcurrentnonzerorange);
    }
    
    if (remainder1loopsize > 0) {
      dDEBUGGING_CODE_1(LEFT,
		       *addressoflhszvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      dget_and_hop_to_next_vbucket(addressoflhszvbucketindex);
      dDENSE_CONVERT_TO_z_LOOP(NEW_RANGE,
				     addressoflhszvbucketindex,
				     addressofrhsvbucketindex_k,
				     k,
				     remainder1loopsize,
				     addressofunconsumedcurrentnonzerorange);
    }
    if (remainder2loopsize > 0) {
      dDEBUGGING_CODE_1(RIGHT,
		       *addressofrhsvbucketindex_k,
		       *addressofunconsumedcurrentnonzerorange);
      dHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhsvbucketindex_k,
			     rhsvectorconsumednonzeroranges_k,
			     rhsvectornonzeroranges_k);
      dDENSE_CONVERT_TO_z_LOOP(NEW_RANGE,
				     addressoflhszvbucketindex,
				     addressofrhsvbucketindex_k,
				     k,
				     remainder2loopsize,
				     addressofunconsumedcurrentnonzerorange);
    }
    DEBUGGING_CODE_2(*addressofunconsumedcurrentnonzerorange);
    if (rangedestination == COMPLETELY_NEW_RANGE) {
      *addressoflastlhszibucketindex = lhszibucketindex;
    }
  }
  
  /*********************************************
   * Case 2: rhsdvbucket has less spaces   *
   * remaining than lhszvbucket.        *
   *********************************************/	  
  else if (*addressofrhsvbucketindex_k >
	   *addressoflhszvbucketindex) {
    
    for(i=0;i<numberofdoubleforloopiterations;i++) {
      dDEBUGGING_CODE_1(RIGHT,
		       *addressofrhsvbucketindex_k,
		       *addressofunconsumedcurrentnonzerorange);
      dHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhsvbucketindex_k,
			     rhsvectorconsumednonzeroranges_k,
			     rhsvectornonzeroranges_k);
      dDENSE_CONVERT_TO_z_LOOP(NEW_RANGE,
				     addressoflhszvbucketindex,
				     addressofrhsvbucketindex_k,
				     k,
				     firstiterationloopsize,
				     addressofunconsumedcurrentnonzerorange);
      dDEBUGGING_CODE_1(LEFT,
		       *addressoflhszvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      dget_and_hop_to_next_vbucket(addressoflhszvbucketindex);
      dDENSE_CONVERT_TO_z_LOOP(NEW_RANGE,
				     addressoflhszvbucketindex,
				     addressofrhsvbucketindex_k,
				     k,
				     seconditerationloopsize,
				     addressofunconsumedcurrentnonzerorange);
    }
    
    if (remainder1loopsize > 0) {
      dDEBUGGING_CODE_1(RIGHT,
		       *addressofrhsvbucketindex_k,
		       *addressofunconsumedcurrentnonzerorange);
      dHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhsvbucketindex_k,
			     rhsvectorconsumednonzeroranges_k,
			     rhsvectornonzeroranges_k);
      dDENSE_CONVERT_TO_z_LOOP(NEW_RANGE,
				     addressoflhszvbucketindex,
				     addressofrhsvbucketindex_k,
				     k,
				     remainder1loopsize,
				     addressofunconsumedcurrentnonzerorange);
    }
    
    if (remainder2loopsize > 0) {
      dDEBUGGING_CODE_1(LEFT,
		       *addressoflhszvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      dget_and_hop_to_next_vbucket(addressoflhszvbucketindex);
      dDENSE_CONVERT_TO_z_LOOP(NEW_RANGE,
				     addressoflhszvbucketindex,
				     addressofrhsvbucketindex_k,
				     k,
				     remainder2loopsize,
				     addressofunconsumedcurrentnonzerorange);
    }
    DEBUGGING_CODE_2(*addressofunconsumedcurrentnonzerorange);
    if (rangedestination == COMPLETELY_NEW_RANGE) {
      *addressoflastlhszibucketindex = lhszibucketindex;
    }
  }
  
  
  /*********************************************
   * Case 3: rhsdvbucket has same # of     *
   * spaces as lhszvbucket (i.e. 0).    *
   *********************************************/
  else if (*addressofrhsvbucketindex_k ==
	   *addressoflhszvbucketindex) {
    
    for(i=0;i<numberofdoubleforloopiterations;i++) {
      dDEBUGGING_CODE_1(LEFT,
		       *addressoflhszvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      dget_and_hop_to_next_vbucket(addressoflhszvbucketindex);
      dHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhsvbucketindex_k,
			     rhsvectorconsumednonzeroranges_k,
			     rhsvectornonzeroranges_k);
      dDENSE_CONVERT_TO_z_LOOP(NEW_RANGE,
				     addressoflhszvbucketindex,
				     addressofrhsvbucketindex_k,
				     k,
				     firstiterationloopsize+seconditerationloopsize,
				     addressofunconsumedcurrentnonzerorange);
    }
    if (remainder1loopsize+remainder2loopsize > 0) {
      dDEBUGGING_CODE_1(LEFT,
		       *addressoflhszvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      dget_and_hop_to_next_vbucket(addressoflhszvbucketindex);
      dHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhsvbucketindex_k,
			     rhsvectorconsumednonzeroranges_k,
			     rhsvectornonzeroranges_k);
      dDENSE_CONVERT_TO_z_LOOP(NEW_RANGE,
				     addressoflhszvbucketindex,
				     addressofrhsvbucketindex_k,
				     k,
				     remainder1loopsize+remainder2loopsize,
				     addressofunconsumedcurrentnonzerorange);
    }
    DEBUGGING_CODE_2(*addressofunconsumedcurrentnonzerorange);
    if (rangedestination == COMPLETELY_NEW_RANGE) {
      *addressoflastlhszibucketindex = lhszibucketindex;
    }
  }
}







/****************************************************************************
 *                                                                          *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/

void doverlapping_cs_range_convert_to_z(
			     addressoflhszvbucketindex,
			     k,
			     addressofrhsvbucketindex_k,
			     totalcurrentnonzerorange
#ifdef _DEBUG
			     ,addressofunconsumedcurrentnonzerorange
			     ,rhsvectorconsumednonzeroranges_k
			     ,rhsvectornonzeroranges_k
			     ,lhszvectornonzeroranges
#endif
			  )
int *addressoflhszvbucketindex;
int k;
int *addressofrhsvbucketindex_k;
int totalcurrentnonzerorange;
#ifdef _DEBUG
int *addressofunconsumedcurrentnonzerorange;
int rhsvectorconsumednonzeroranges_k;
int rhsvectornonzeroranges_k;
int lhszvectornonzeroranges;
#endif
{
  int firstloopsize;
  int numberofdoubleforloopiterations;
  int firstiterationloopsize;
  int seconditerationloopsize;
  int remainder1loopsize;
  int remainder2loopsize;

  int i;
  
  dDETERMINE_LOOP_BOUNDS(totalcurrentnonzerorange,
			*addressoflhszvbucketindex,
			*addressofrhsvbucketindex_k,
			&firstloopsize,
			&numberofdoubleforloopiterations,
			&firstiterationloopsize,
			&seconditerationloopsize,
			&remainder1loopsize,
			&remainder2loopsize,
			addressofunconsumedcurrentnonzerorange);
  dDENSE_CONVERT_TO_z_LOOP(OLD_RANGE,
				 addressoflhszvbucketindex,
				 addressofrhsvbucketindex_k,
				 k,
				 firstloopsize,
				 addressofunconsumedcurrentnonzerorange);
  
  
  /*********************************************
   * Case 1: rhsdvbucket has more spaces   *
   *         remaining than lhszvbucket.*
   *********************************************/
  if (*addressofrhsvbucketindex_k <
      *addressoflhszvbucketindex) {
    /*B.2.1*/
    
    for(i=0;i<numberofdoubleforloopiterations;i++) { /*B.3.1*/
      dDEBUGGING_CODE_1(LEFT,
		       *addressoflhszvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      dHOP_TO_NEXT_LHSVBUCKET(addressoflhszvbucketindex,
			     lhszvectornonzeroranges);
      dDENSE_CONVERT_TO_z_LOOP(OLD_RANGE,
				     addressoflhszvbucketindex,
				     addressofrhsvbucketindex_k,
				     k,
				     firstiterationloopsize,
				     addressofunconsumedcurrentnonzerorange);
      dDEBUGGING_CODE_1(RIGHT,
		       *addressofrhsvbucketindex_k,
		       *addressofunconsumedcurrentnonzerorange);
      dHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhsvbucketindex_k,
			     rhsvectorconsumednonzeroranges_k,
			     rhsvectornonzeroranges_k);
      dDENSE_CONVERT_TO_z_LOOP(OLD_RANGE,
				     addressoflhszvbucketindex,
				     addressofrhsvbucketindex_k,
				     k,
				     seconditerationloopsize,
				     addressofunconsumedcurrentnonzerorange);
    } /*B.3.1*/
    
    if (remainder1loopsize > 0) {    	    
      dDEBUGGING_CODE_1(LEFT,
		       *addressoflhszvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      dHOP_TO_NEXT_LHSVBUCKET(addressoflhszvbucketindex,
			     lhszvectornonzeroranges);
      dDENSE_CONVERT_TO_z_LOOP(OLD_RANGE,
				     addressoflhszvbucketindex,
				     addressofrhsvbucketindex_k,
				     k,
				     remainder1loopsize,
				     addressofunconsumedcurrentnonzerorange);
    }
    
    if (remainder2loopsize > 0) {    
      dDEBUGGING_CODE_1(RIGHT,
		       *addressofrhsvbucketindex_k,
		       *addressofunconsumedcurrentnonzerorange);
      dHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhsvbucketindex_k,
			     rhsvectorconsumednonzeroranges_k,
			     rhsvectornonzeroranges_k);
      dDENSE_CONVERT_TO_z_LOOP(OLD_RANGE,
				     addressoflhszvbucketindex,
				     addressofrhsvbucketindex_k,
				     k,
				     remainder2loopsize,
				     addressofunconsumedcurrentnonzerorange);
    }
    DEBUGGING_CODE_2(*addressofunconsumedcurrentnonzerorange);
  } /*B.2.1*/
  
  /*********************************************
   * Case 2: rhsdvbucket has less spaces   *
   * remaining than lhszvbucket.        *
   *********************************************/
  else if (*addressofrhsvbucketindex_k >
	   *addressoflhszvbucketindex) {
    /*B.2.2*/
    
    for(i=0;i<numberofdoubleforloopiterations;i++) { /*B.3.2*/
      dDEBUGGING_CODE_1(RIGHT,
		       *addressofrhsvbucketindex_k,
		       *addressofunconsumedcurrentnonzerorange);
      dHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhsvbucketindex_k,
			     rhsvectorconsumednonzeroranges_k,
			     rhsvectornonzeroranges_k);
      dDENSE_CONVERT_TO_z_LOOP(OLD_RANGE,
				     addressoflhszvbucketindex,
				     addressofrhsvbucketindex_k,
				     k,
				     firstiterationloopsize,
				     addressofunconsumedcurrentnonzerorange);
      dDEBUGGING_CODE_1(LEFT,
		       *addressoflhszvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      dHOP_TO_NEXT_LHSVBUCKET(addressoflhszvbucketindex,
			     lhszvectornonzeroranges);
      dDENSE_CONVERT_TO_z_LOOP(OLD_RANGE,
				     addressoflhszvbucketindex,
				     addressofrhsvbucketindex_k,
				     k,
				     seconditerationloopsize,
				     addressofunconsumedcurrentnonzerorange);
    } /*B.3.2*/
    
    if (remainder1loopsize > 0) {        
      dDEBUGGING_CODE_1(RIGHT,
		       *addressofrhsvbucketindex_k,
		       *addressofunconsumedcurrentnonzerorange);
      dHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhsvbucketindex_k,
			     rhsvectorconsumednonzeroranges_k,
			     rhsvectornonzeroranges_k);
      dDENSE_CONVERT_TO_z_LOOP(OLD_RANGE,
				     addressoflhszvbucketindex,
				     addressofrhsvbucketindex_k,
				     k,
				     remainder1loopsize,
				     addressofunconsumedcurrentnonzerorange);
    }
    if (remainder2loopsize > 0) {    
      dDEBUGGING_CODE_1(LEFT,
		       *addressoflhszvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      dHOP_TO_NEXT_LHSVBUCKET(addressoflhszvbucketindex,
			     lhszvectornonzeroranges);
      dDENSE_CONVERT_TO_z_LOOP(OLD_RANGE,
				     addressoflhszvbucketindex,
				     addressofrhsvbucketindex_k,
				     k,
				     remainder2loopsize,
				     addressofunconsumedcurrentnonzerorange);
    }
    DEBUGGING_CODE_2(*addressofunconsumedcurrentnonzerorange);
  } /*B.2.2*/
  
  
  /*********************************************
   * Case 3: rhsdvbucket has same # of     *
   * spaces as lhszvbucket (i.e. 0).    *
   *********************************************/
  else if (*addressofrhsvbucketindex_k ==
	   *addressoflhszvbucketindex) {
    /*B.2.3*/
    
    for(i=0;i<numberofdoubleforloopiterations;i++) { /*B.3.3*/
      dDEBUGGING_CODE_1(LEFT,
		       *addressoflhszvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      dHOP_TO_NEXT_LHSVBUCKET(addressoflhszvbucketindex,
			     lhszvectornonzeroranges);
      dHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhsvbucketindex_k,
			     rhsvectorconsumednonzeroranges_k,
			     rhsvectornonzeroranges_k);
      dDENSE_CONVERT_TO_z_LOOP(OLD_RANGE,
				     addressoflhszvbucketindex,
				     addressofrhsvbucketindex_k,
				     k,
				     firstiterationloopsize+seconditerationloopsize,
				     addressofunconsumedcurrentnonzerorange);
    } /*B.3.3*/
    
    if (remainder1loopsize+remainder2loopsize > 0) {    
      dDEBUGGING_CODE_1(LEFT,
		       *addressoflhszvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      dHOP_TO_NEXT_LHSVBUCKET(addressoflhszvbucketindex,
			     lhszvectornonzeroranges);
      dHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhsvbucketindex_k,
			     rhsvectorconsumednonzeroranges_k,
			     rhsvectornonzeroranges_k);
      dDENSE_CONVERT_TO_z_LOOP(OLD_RANGE,
				     addressoflhszvbucketindex,
				     addressofrhsvbucketindex_k,
				     k,
				     remainder1loopsize+remainder2loopsize,
				     addressofunconsumedcurrentnonzerorange);
    }
    DEBUGGING_CODE_2(*addressofunconsumedcurrentnonzerorange);
  } /*B.2.3*/
}




/****************************************************************************
 *                                                                          *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/

void ddense_convert_to_z_loop(range,
				    lhszvbucketindex,
				    rhsdvbucketindex,
				    k,
				    loopsize
#ifdef _DEBUG
				    ,unconsumedcurrentnonzerorange
#endif
				    )
int range;
int *lhszvbucketindex;
int *rhsdvbucketindex;
int k;
int loopsize;
#ifdef _DEBUG
int *unconsumedcurrentnonzerorange;
#endif
{
  int j;
  
  if (range == OLD_RANGE) {
    if (k == REAL_PART) {
      for (j=0;j<loopsize;j++)
	XY_PUT_AND_CAST_PLUS_EQUALS(lhszvbucketptr->
				    value[(*lhszvbucketindex)+j],
				    rhsdvbucketptr[k]->
				    value[(*rhsdvbucketindex)+j],
				    0,
				    double);
    }
    else {
      for (j=0;j<loopsize;j++)
	XY_PUT_AND_CAST_PLUS_EQUALS(lhszvbucketptr->
				    value[(*lhszvbucketindex)+j],
				    0,
				    rhsdvbucketptr[k]->
				    value[(*rhsdvbucketindex)+j],
				    double);
    }
  }
  else {
    if (k == REAL_PART) {
      for (j=0;j<loopsize;j++)
	XY_PUT_AND_CAST(lhszvbucketptr->
			value[(*lhszvbucketindex)+j],
			rhsdvbucketptr[k]->
			value[(*rhsdvbucketindex)+j],
			0,
			double);
    }
    else {
      for (j=0;j<loopsize;j++)
	XY_PUT_AND_CAST(lhszvbucketptr->
			value[(*lhszvbucketindex)+j],
			0,
			rhsdvbucketptr[k]->
			value[(*rhsdvbucketindex)+j],
			double);
    }
  }
  
  *lhszvbucketindex += loopsize;
  *rhsdvbucketindex += loopsize;
  
#ifdef _DEBUG
  *unconsumedcurrentnonzerorange -= loopsize;
#endif
  
}



/****************************************************************************
 *                                                                          *
 *      dget_and_hop_to_next_to_z_vbucket(lhszvbucketindex)    *
 *                                                                          *
 *      Question: Why do we need this routine since a very similar routine  *
 *                exists called 'dget_and_hop_to_next_vbucket' which we      *
 *                instantiate with each precision?                          *
 *      Answer  : Because in the CALLs to these routines the globals        *
 *                lhsTO_pREc_vbucketptr and lhspREc_vbucketptr will be      *
 *                instantiated distinctly, even though the code for both    *
 *                these routines will be identical once instantiated.       *
 *                Note that once instantiated, both these names acquire     *
 *                a precision marker [s,d,c,z] in their name, depending on  *
 *                'TO_pREc_' and 'pREc_', respectively. (Uppercase and      *
 *                lowercase letters in both these names have been           *
 *                permutated so that this comment would be preserved as is  *
 *                after instantiation.  Here, by instantiation we mean the  *
 *                result of macro expansion.)                               *  
 *                                                                          *
 ****************************************************************************/

void dget_and_hop_to_next_to_z_vbucket(lhszvbucketindex)
int *lhszvbucketindex;
{
  zgetvbucket(&(lhszvbucketptr->nextvbucket));
  lhszvbucketptr = lhszvbucketptr->nextvbucket;
  *lhszvbucketindex = 0;
}




/****************************************************************************
 *                                                                          *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/
void daxpy_2_heap_to_z_post_processing(destptrptr,
					     rhsdvectorptrptrs,
					     restorerhsdvectorptr,
					     arity,
					     lhszvectornonzeroes,
					     lhszvectornonzeroranges,
					     governingdstype)
zvectorstruct **destptrptr;
dvectorstruct **rhsdvectorptrptrs[];
dvectorstruct *restorerhsdvectorptr[MAXARITY+1];
int arity;
int lhszvectornonzeroes;
int lhszvectornonzeroranges;
int governingdstype;
{
  dvectorstruct *rhsdvectorptr[MAXARITY];

  zvectorstruct *tempzvectorptr;
  zvectorstruct *sstempzvectorptr;
  zbucketstruct *tempzbucketptr;
  zbucketstruct *sstempzbucketptr;
  zibucketstruct *tempzibucketptr;
  zvbucketstruct *tempzvbucketptr;

  dvectorstruct *tempdvectorptr;
  dbucketstruct *tempdbucketptr;
  dibucketstruct *tempdibucketptr;
  dvbucketstruct *tempdvbucketptr;

  int k;

  for (k=0;k<arity;k++)
    rhsdvectorptr[k] = *rhsdvectorptrptrs[k];

  /****************************************
   * Switch lhszvectorptr and *destptrptr. *
   ****************************************/

  lhszvectorptr->nonzeroes = lhszvectornonzeroes;
  
  if (governingdstype == CS)
    lhszvectorptr->zds.cs.nonzeroranges =
      lhszvectornonzeroranges;
  
  tempzvectorptr = (*destptrptr);
  (*destptrptr) = lhszvectorptr;
  lhszvectorptr = tempzvectorptr;

  if (governingdstype == SS) {
    
    if ((INITIALIZED_MAXINDEX &&
	 ((float) (*destptrptr)->nonzeroes / (float) GLOBAL_MAXINDEX) >=
	 SWITCH_NZ_RATIO)     ||
	((*destptrptr)->nonzeroes >= SWITCH_NZ_VALUE)) {
             
      /****************************************
       * Convert SS to CS.                    *
       ****************************************/
      
      sstempzvectorptr = (*destptrptr);
      zconvertdstype(destptrptr,MAXARITY,CS);
	  
      sstempzbucketptr =
	sstempzvectorptr->zds.ss.bucketptr;
      if (sstempzbucketptr != NULL) {
	while (sstempzbucketptr->nextbucket != NULL)
	  sstempzbucketptr = sstempzbucketptr->nextbucket;
	sstempzbucketptr->nextbucket = freedzbucketslist;
	RESET_freedzbucketslist(sstempzvectorptr->zds.ss.bucketptr);
	sstempzvectorptr->zds.ss.bucketptr = NULL;
      }
      holdingbinzvectorptr[MAXARITY] =  sstempzvectorptr;
    }
  }
  
  /****************************************
   * If *destptrptr was not empty,        *
   * add its buckets to freedzbucketslist. *
   ****************************************/
  
  if (tempzvectorptr->dstype == SS) {
    tempzbucketptr = tempzvectorptr->zds.ss.bucketptr;
    if (tempzbucketptr != NULL) {
      while (tempzbucketptr->nextbucket != NULL)
	tempzbucketptr = tempzbucketptr->nextbucket;
      tempzbucketptr->nextbucket = freedzbucketslist;
      RESET_freedzbucketslist(tempzvectorptr->zds.ss.bucketptr);
      tempzvectorptr->zds.ss.bucketptr = NULL;
    }
  }
  else if (tempzvectorptr->dstype == CS) {
    tempzibucketptr = tempzvectorptr->zds.cs.ibucketptr;
    if (tempzibucketptr != NULL) {
      while (tempzibucketptr->nextibucket != NULL)
	tempzibucketptr = tempzibucketptr->nextibucket;
      tempzibucketptr->nextibucket = freedzibucketslist;
      RESET_freedzibucketslist(tempzvectorptr->zds.cs.ibucketptr);
      tempzvectorptr->zds.cs.ibucketptr = NULL;
    }
    tempzvbucketptr = tempzvectorptr->zds.cs.vbucketptr;
    if (tempzvbucketptr != NULL) {
      while (tempzvbucketptr->nextvbucket != NULL)
	tempzvbucketptr = tempzvbucketptr->nextvbucket;
      tempzvbucketptr->nextvbucket = freedzvbucketslist;
      RESET_freedzvbucketslist(tempzvectorptr->zds.cs.vbucketptr);
      tempzvectorptr->zds.cs.vbucketptr = NULL;
    }
  }
  
  /*****************************************
   * If holdingbins were used, add         *
   * their buckets to freeddbucketslist, OR *
   * their ibuckets to freeddibucketslist & *
   * their vbuckets to freeddvbucketslist.  *
   *****************************************/
  for (k=0;k<arity;k++) {
    
    rhsdvectorptr[k] = restorerhsdvectorptr[k];
    
    if (holdingbindvectorptr[k]->dstype == SS) {
      tempdbucketptr = holdingbindvectorptr[k]->dds.ss.bucketptr;
      if (tempdbucketptr != NULL) {
#ifdef _DEBUG
	if (governingdstype == CS) {
	  printf("dspgx warning 7.4: Unexpcted non-NULL");
	  printf("tempdbucketptr.\n");
	}
#endif
	while (tempdbucketptr->nextbucket != NULL) {
#ifdef _DEBUG
	  if (governingdstype == SS) {
	    printf("dspgx warning 7.1: tempdbucketptr should have");
	    printf(" no trailing nextbuckets\n");
	  }
#endif
	  tempdbucketptr = tempdbucketptr->nextbucket;
	}
	tempdbucketptr->nextbucket = freeddbucketslist;
	RESET_freeddbucketslist(holdingbindvectorptr[k]->dds.ss.bucketptr);
	holdingbindvectorptr[k]->dds.ss.bucketptr = NULL;
      }
    }
    else if (holdingbindvectorptr[k]->dstype == CS) {
      tempdibucketptr =
	holdingbindvectorptr[k]->dds.cs.ibucketptr;
      if (tempdibucketptr != NULL) {
#ifdef _DEBUG
	if (governingdstype == SS) {
	  printf("dspgx warning 7.2: Unexpcted non-NULL");
	  printf(" ibucketptr.\n");
	}
#endif
	while (tempdibucketptr->nextibucket != NULL)
	  tempdibucketptr = tempdibucketptr->nextibucket;
	tempdibucketptr->nextibucket = freeddibucketslist;
	RESET_freeddibucketslist(holdingbindvectorptr[k]->dds.cs.ibucketptr);
	holdingbindvectorptr[k]->dds.cs.ibucketptr = NULL;
      }
      tempdvbucketptr =
	holdingbindvectorptr[k]->dds.cs.vbucketptr;
      if (tempdvbucketptr != NULL) {
#ifdef _DEBUG
	if (governingdstype == SS) {
	  printf("dspgx warning 7.3: Unexpcted non-NULL");
	  printf(" vbucketptr.\n");
	}
#endif
	while (tempdvbucketptr->nextvbucket != NULL)
	  tempdvbucketptr = tempdvbucketptr->nextvbucket;
	tempdvbucketptr->nextvbucket = freeddvbucketslist;
	RESET_freeddvbucketslist(holdingbindvectorptr[k]->dds.cs.vbucketptr);
	holdingbindvectorptr[k]->dds.cs.vbucketptr = NULL;
      }
    }
  } /* matches    for (k=0;k<arity;k++) {    */
}

