#define DOUBLE
/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#include  <stdlib.h>
#include "dgaxpy_pd_macros_and_data_structures.h"
#include "dgaxpy_pd_global_variables.h"


/****************************************************************************
 *                                                                          *
 *  dadallocate()                                                            *
 *                                                                          *
 *      dadallocate calls the bucket and vector store allocation functions.  *
 *      It should be called by the user once, before any calls to dspsd &  *
 *      adxpyNs.                                                            *
 *                                                                          *
 ****************************************************************************/

void dadallocate()
{
  int k;
  
  dinitialize_pd_global_variables();
  dallocatevectorstore();
  dallocatebucketstore();
  dallocateibucketstore();
  dallocatevbucketstore();

  dgetvector(&lhsdvectorptr);
  for (k=0;k<MAXARITY+1;k++)
    dgetvector(&holdingbindvectorptr[k]);
  dgetvector(&tempddestptr);
  dgetvector(&restofplusequaldvectorptr);
  restofplusequaldvectorptr->dstype = CS;
}



/****************************************************************************
 *                                                                          *
 *   dadfree()                                                               *
 *                                                                          *
 *      dadfree frees all the dynamically allocated space for a given        *
 *      percision.                                                          *
 *                                                                          *
 ****************************************************************************/

void dadfree()
{
  dvectorstorestruct  *remaindervectorstoreptr,  *tempvectorstoreptr;
  dbucketstorestruct  *remainderbucketstoreptr,  *tempbucketstoreptr;
  dibucketstorestruct *remainderibucketstoreptr, *tempibucketstoreptr;
  dvbucketstorestruct *remaindervbucketstoreptr, *tempvbucketstoreptr;

  dvectorstruct *tempvectorptr;
  dbucketstruct *tempbucketptr;
  dibucketstruct *tempibucketptr;
  dvbucketstruct *tempvbucketptr;

  int i;

  for(remaindervectorstoreptr=dvectorstorestartptr;
      remaindervectorstoreptr!=NULL;
      remaindervectorstoreptr=tempvectorstoreptr) {

    tempvectorstoreptr = remaindervectorstoreptr->nextstore;
    free(remaindervectorstoreptr);
  }

  for(remainderbucketstoreptr=dbucketstorestartptr;
      remainderbucketstoreptr!=NULL;
      remainderbucketstoreptr=tempbucketstoreptr) {
    tempbucketstoreptr = remainderbucketstoreptr->nextstore;

    /*
     * Freeing based on per BUCKET Dynamic Allocation.
     * Here we free the memory which has been allocated per bucket
     * in dallocatebucketstore().
     */
    for (i=0; i<dBUCKETSTORE_DIM; i++) {
      tempbucketptr = &(remainderbucketstoreptr->bucket[i]);
      free(tempbucketptr->index);
      free(tempbucketptr->value);
    }
  
    /*
     * Now regardless of the allocation method above, we have freed
     * everything hanging off our bucketstore, and we can go ahead and
     * free the bucketstore itself.
     */
    free(remainderbucketstoreptr);
    
  }
  
  for(remainderibucketstoreptr=dibucketstorestartptr;
      remainderibucketstoreptr!=NULL;
      remainderibucketstoreptr=tempibucketstoreptr) {
    tempibucketstoreptr = remainderibucketstoreptr->nextstore;
    
    /*
     * Freeing based on per IBUCKET Dynamic Allocation.
     */
    for (i=0; i<dIBUCKETSTORE_DIM; i++) {
      tempibucketptr = &(remainderibucketstoreptr->ibucket[i]);
      free(tempibucketptr->index_RANGE_BEGIN);
      free(tempibucketptr->index_RANGE_END);
    }
    
    /*
     * Now regardless of the allocation method above, we have freed
     * everything hanging off our ibucketstore, and we can go ahead and
     * free the ibucketstore itself.
     */
    free(remainderibucketstoreptr);
  }
  
  for(remaindervbucketstoreptr=dvbucketstorestartptr;
      remaindervbucketstoreptr!=NULL;
      remaindervbucketstoreptr=tempvbucketstoreptr) {
    tempvbucketstoreptr = remaindervbucketstoreptr->nextstore;
    
    /*
     * Freeing based on per VBUCKET Dynamic Allocation.
     */
    for (i=0; i<dVBUCKETSTORE_DIM; i++) {
      tempvbucketptr = &(remaindervbucketstoreptr->vbucket[i]);
      free(tempvbucketptr->value);
    }
    
    /*
     * Now regardless of the allocation method above, we have freed
     * everything hanging off our vbucketstore, and we can go ahead and
     * free the vbucketstore itself.
     */
    free(remaindervbucketstoreptr);
  }
}



/****************************************************************************
 *                                                                          *
 *   dad_free_vectors_buckets(dvectorptrptr)                             *
 *                                                                          *
 *       dad_free_vectors_buckets frees the buckets of dvectorptrptr.    *
 *       BUT IT DOES NOT free dvectorptrptr, itself.                    *
 *       Note that freeddbucketslist, freeddibucketslist, and       *
 *       freeddvbucketslist are global variables included in            *
 *       $(SAXPY_FLAVOR)_global_variables.h.                                *
 *                                                                          *
 *       dvectorptrptr  : The vector to be freed.                       *
 *                                                                          *
 ****************************************************************************/

void dad_free_vectors_buckets(dvectorptrptr)
dvectorstruct **dvectorptrptr;
{

  dvectorstruct *dvectorptr;
  dbucketstruct *dbucketptr;
  dibucketstruct *dibucketptr;
  dvbucketstruct *dvbucketptr;

  dvectorptr = (dvectorstruct *) *dvectorptrptr;
  
  if (dvectorptr != NULL) {
    if (dvectorptr->dstype == S1) {
      /* do nothing */
    }
    else if (dvectorptr->dstype == SS) {
      /**********************************************************************
       *                        Single Subscript                            *
       **********************************************************************/
      
      dbucketptr = dvectorptr->dds.ss.bucketptr;
      if (dbucketptr != NULL) {
	while (dbucketptr->nextbucket != NULL)
	  dbucketptr = dbucketptr->nextbucket;
	dbucketptr->nextbucket = freeddbucketslist;
	RESET_freeddbucketslist(dvectorptr->dds.ss.bucketptr);
	dvectorptr->dds.ss.bucketptr = NULL;
      }
      
    } else /* dvectorptr->ddstype = CS */ {
      /**********************************************************************
       *                      Compressed Subscript                          *
       **********************************************************************/
      
      dibucketptr = dvectorptr->dds.cs.ibucketptr;
      if (dibucketptr != NULL) {
	while (dibucketptr->nextibucket != NULL)
	  dibucketptr = dibucketptr->nextibucket;
	dibucketptr->nextibucket = freeddibucketslist;
	RESET_freeddibucketslist(dvectorptr->dds.cs.ibucketptr);
	dvectorptr->dds.cs.ibucketptr = NULL;
      }
      dvbucketptr = dvectorptr->dds.cs.vbucketptr;
      if (dvbucketptr != NULL) {
	while (dvbucketptr->nextvbucket != NULL)
	  dvbucketptr = dvbucketptr->nextvbucket;
	dvbucketptr->nextvbucket = freeddvbucketslist;
	RESET_freeddvbucketslist(dvectorptr->dds.cs.vbucketptr);
	dvectorptr->dds.cs.vbucketptr = NULL;
      }
    }
  }

/*
 * PMK: begin 12/4/96 edits
 *
 * In order to ensure that dvectorptr points to a null vectorstruct,
 * i.e., a vectorstruct whose fields all have zero values, we assign to it
 * the contents of nulldvector which is a global initialized with all
 * zero members in gaxpy_pd_global_variables.h.
 *
 */

  *(dvectorptr) = nulldvector;

/*
 * PMK: end 12/4/96 edits
 */

}


/****************************************************************************
 *                                                                          *
 *    dinitialize_pd_global_variables()                                      *
 *                                                                          *
 *    dinitialize_pd_global_variables performs initializations for some      *
 *    precision dependent globals.                                          *
 *                                                                          *
 ****************************************************************************/

void dinitialize_pd_global_variables()
{
  dbucketstorestartptr=NULL;
  freeddbucketslist=NULL;
  dibucketstorestartptr=NULL;
  freeddibucketslist=NULL;
  dvbucketstorestartptr=NULL;
  freeddvbucketslist=NULL;
  dvectorstorestartptr=NULL;

}

