/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#ifndef dGAXPY_PD_MACROS_AND_DATA_STRUCTURES

#define dGAXPY_PD_MACROS_AND_DATA_STRUCTURES

#include "gaxpy_pi_macros_and_data_structures.h"


/***************************************************************************
 *                                                                         *
 *  Macro definitions for store dimensions                                 *
 *                                                                         *
 ***************************************************************************/

#define dBUCKETSTORE_DIM  128      /* # of buckets in a bucketstore.   */

#define dVBUCKETSTORE_DIM 128      /* # of vbuckets in a vbucketstore. */

#define dIBUCKETSTORE_DIM 128      /* # of ibuckets in a ibucketstore. */

#define dVECTORSTORE_DIM  128      /* # of vectors in a vectorstore.   */


/***************************************************************************
 *                                                                         *
 *  Macro definitions for subroutines;  DEBUG ON                           *
 *                                                                         *
 ***************************************************************************/

#ifdef _DEBUG

#define dHOP_TO_NEXT_RHSVBUCKET(a,b,c,d) \
        dhop_to_next_rhsvbucket(a,b,c,d)

#define dHOP_TO_NEXT_LHSVBUCKET(a,b) \
        dhop_to_next_lhsvbucket(a,b)

#define dNON_OVER_LAPPING_CS_RANGE_AXPY(a,b,c,d,e,f,g,h,i,j,k,l) \
        dnon_over_lapping_cs_range_axpy(a,b,c,d,e,f,g,h,i,j,k,l)

#define dOVERLAPPING_CS_RANGE_AXPY(a,b,c,d,e,f,g,h,i) \
        doverlapping_cs_range_axpy(a,b,c,d,e,f,g,h,i)

#define dDENSE_PLUSEQUAL_AXPY_LOOP(a,b,c,d,e) \
        ddense_plusequal_axpy_loop(a,b,c,d,e)

#define dNON_OVER_LAPPING_CS_RANGE_CONVERT_TO_c(a,b,c,d,e,f,g,h,i,j,k) \
        dnon_over_lapping_cs_range_convert_to_c(a,b,c,d,e,f,g,h,i,j,k)

#define dOVERLAPPING_CS_RANGE_CONVERT_TO_c(a,b,c,d,e,f,g,h) \
        doverlapping_cs_range_convert_to_c(a,b,c,d,e,f,g,h)

#define dDENSE_CONVERT_TO_c_LOOP(a,b,c,d,e,f) \
        ddense_convert_to_c_loop(a,b,c,d,e,f)

#define dNON_OVER_LAPPING_CS_RANGE_CONVERT_TO_z(a,b,c,d,e,f,g,h,i,j,k) \
        dnon_over_lapping_cs_range_convert_to_z(a,b,c,d,e,f,g,h,i,j,k)

#define dOVERLAPPING_CS_RANGE_CONVERT_TO_z(a,b,c,d,e,f,g,h) \
        doverlapping_cs_range_convert_to_z(a,b,c,d,e,f,g,h)

#define dDENSE_CONVERT_TO_z_LOOP(a,b,c,d,e,f) \
        ddense_convert_to_z_loop(a,b,c,d,e,f)

#define dDETERMINE_LOOP_BOUNDS(a,b,c,d,e,f,g,h,i,j) \
        ddetermine_loop_bounds(a,b,c,d,e,f,g,h,i,j)

#define dDEBUGGING_CODE_1(a,b,c) \
        ddebugging_code_1(a,b,c)

/*
 *  In-lining ddense_axpy_loop
 */

#define dDENSE_AXPY_LOOP(range,lhsdvbucketindex,rhsdvbucketindex,k,loopsize,alpha,addressofunconsumedcurrentnonzerorange) \
{ \
  int j; \
  if (range == OLD_RANGE) { \
    for (j=0;j<loopsize;j++) { \
      GAXPY_PLUS_EQUALS(lhsdvbucketptr->value[(*lhsdvbucketindex)+j], \
	      alpha,rhsdvbucketptr[k]->value[(*rhsdvbucketindex)+j]); \
    } \
  } \
  else { \
    for (j=0;j<loopsize;j++) { \
      GAXPY_EQUALS(lhsdvbucketptr->value[(*lhsdvbucketindex)+j], \
	      alpha,rhsdvbucketptr[k]->value[(*rhsdvbucketindex)+j]); \
    } \
  } \
  (*lhsdvbucketindex) += loopsize; \
  (*rhsdvbucketindex) += loopsize; \
  (*addressofunconsumedcurrentnonzerorange) -= loopsize; \
}


/***************************************************************************
 *                                                                         *
 *  Macro definitions for subroutines;  DEBUG OFF                          *
 *                                                                         *
 ***************************************************************************/

#else

#define dHOP_TO_NEXT_RHSVBUCKET(a,b,c,d) \
        dhop_to_next_rhsvbucket(a,b)

#define dHOP_TO_NEXT_LHSVBUCKET(a,b) \
        dhop_to_next_lhsvbucket(a)

#define dNON_OVER_LAPPING_CS_RANGE_AXPY(a,b,c,d,e,f,g,h,i,j,k,l) \
        dnon_over_lapping_cs_range_axpy(a,b,c,d,e,f,g,h)

#define dOVERLAPPING_CS_RANGE_AXPY(a,b,c,d,e,f,g,h,i) \
        doverlapping_cs_range_axpy(a,b,c,d,e)

#define dDENSE_PLUSEQUAL_AXPY_LOOP(a,b,c,d,e) \
        ddense_plusequal_axpy_loop(a,b,c,d)

#define dNON_OVER_LAPPING_CS_RANGE_CONVERT_TO_c(a,b,c,d,e,f,g,h,i,j,k) \
        dnon_over_lapping_cs_range_convert_to_c(a,b,c,d,e,f,g)

#define dOVERLAPPING_CS_RANGE_CONVERT_TO_c(a,b,c,d,e,f,g,h) \
        doverlapping_cs_range_convert_to_c(a,b,c,d)

#define dDENSE_CONVERT_TO_c_LOOP(a,b,c,d,e,f) \
        ddense_convert_to_c_loop(a,b,c,d,e)

#define dNON_OVER_LAPPING_CS_RANGE_CONVERT_TO_z(a,b,c,d,e,f,g,h,i,j,k) \
        dnon_over_lapping_cs_range_convert_to_z(a,b,c,d,e,f,g)

#define dOVERLAPPING_CS_RANGE_CONVERT_TO_z(a,b,c,d,e,f,g,h) \
        doverlapping_cs_range_convert_to_z(a,b,c,d)

#define dDENSE_CONVERT_TO_z_LOOP(a,b,c,d,e,f) \
        ddense_convert_to_z_loop(a,b,c,d,e)

#define dDETERMINE_LOOP_BOUNDS(a,b,c,d,e,f,g,h,i,j) \
        ddetermine_loop_bounds(a,b,c,d,e,f,g,h,i)

#define dDEBUGGING_CODE_1(a,b,c)

/*
 *  In-lining ddense_axpy_loop
 */

#define dDENSE_AXPY_LOOP(range,lhsdvbucketindex,rhsdvbucketindex,k,loopsize,alpha,addressofunconsumedcurrentnonzerorange) \
{ \
  int j; \
  if (range == OLD_RANGE) { \
    for (j=0;j<loopsize;j++) { \
      GAXPY_PLUS_EQUALS(lhsdvbucketptr->value[(*lhsdvbucketindex)+j], \
	      alpha,rhsdvbucketptr[k]->value[(*rhsdvbucketindex)+j]); \
    } \
  } \
  else { \
    for (j=0;j<loopsize;j++) { \
      GAXPY_EQUALS(lhsdvbucketptr->value[(*lhsdvbucketindex)+j], \
	      alpha,rhsdvbucketptr[k]->value[(*rhsdvbucketindex)+j]); \
    } \
  } \
  (*lhsdvbucketindex) += loopsize; \
  (*rhsdvbucketindex) += loopsize; \
}

#endif


/***************************************************************************
 *                                                                         *
 *  Macro definitions for subroutines with the same interface for          *
 * _DEBUG.                                                                 *
 *                                                                         *
 ***************************************************************************/

#define RESET_freeddbucketslist(new_head) \
{ freeddbucketslist = new_head; \
}

#define RESET_freeddibucketslist(new_head) \
{ freeddibucketslist = new_head; \
}

#define RESET_freeddvbucketslist(new_head) \
{ freeddvbucketslist = new_head; \
}



/****************************************************************************
 *                                                                          *
 *  Data Structures                                                         *
 *                                                                          *
 *     Two abstract structures need to be represented: sparse vector and    *
 *     heap.  We will first describe these for _SINGLE_SUBSCRIPTS and then  *
 *     point out the differences that exist between these data structures   *
 *     and an analog set of data structures used for _COMPRESSED_SUBSCRIPTS.*
 *                                                                          *
 *     To represent sparse vectors using _SINGLE_SUBSCRIPTS, we define 4    *
 *     structs:                                                             *
 *                                                                          *
 *                  dbucketstruct contains a max # of pairs of indices  *
 *                                    and the corresponding values of the   *
 *                                    of the nonzeroes in a vector, and a   *
 *                                    pointer to the next bucket if there   *
 *                                    are more nonzeroes in the vector. The *
 *                                    indices are always in ascending order *
 *                                    by virtue of the way they are input   *
 *                                    in dspsd and the fact that the       *
 *                                    heapsaxpy operations construct new    *
 *                                    vectors with the same ordering.       *
 *                                                                          *
 *                  dvectorstruct is a sparse representation of a       *
 *                                    vector.  It has a field for the total *
 *                                    # of nonzeroes in the vector and a    *
 *                                    linked list of buckets which contain  *
 *                                    the indices and values of the         *
 *                                    nonzeroes in the vector.              *
 *                                                                          *
 *                  dbucketstorestruct is a long array of buckets, and a*
 *                                    pointer to the next store, if needed. *
 *                                    The basic idea is to allocate space   *
 *                                    in large blocks so that dynamic       *
 *                                    memory allocations will be kept to a  *
 *                                    minimum, yet vectors will be able to  *
 *                                    "grow as they go" by having access to *
 *                                    a memory store of allocated buckets.  *
 *                                                                          *
 *                  dvectorstorestruct is a long array of vectors, and a*
 *                                    pointer to the next store, if needed. *
 *                                    The basic idea is the same as in      *
 *                                    dbucketstorestruct.               *
 *                                                                          *
 *     To represent a heap, we define 1 struct:                             *
 *                                                                          *
 *                  heapstruct        represents a node in the binary heap  *
 *                                    tree.  Each node of the heap is a     *
 *                                    placeholder for a nonzero vector      *
 *                                    element.  The nonzero itself is not   *
 *                                    stored in the heap; instead its       *
 *                                    vector index (order of its vector in  *
 *                                    the rhs of the saxpy),  and its       *
 *                                    subscript index (its array index in   *
 *                                    the non-sparse vector) are placed in  *
 *                                    the heap node.  Additionally, each    *
 *                                    heap node contains two pointer fields *
 *                                    containing addresses of the left and  *
 *                                    right children of the node.           *
 *                                    The subscript index serves as the key *
 *                                    for ordering the heap: if x and p(x)  *
 *                                    are a node and its parent, then       *
 *                                    key_p(x) <= key_x.  It follows that   *
 *                                    key_root is the smallest key in the   *
 *                                    tree at all times, and thus by always *
 *                                    poping the root and "consuming" it in *
 *                                    the saxpy, the lhs vector will be     *
 *                                    constructed in the correct (i.e.,     *
 *                                    ascending) order of subscripts.       *
 *                                                                          *
 *     Each typedef below is accompanied by a pictorial depiction.          *
 *                                                                          *
 *     A Note on C syntax: Several of the data structures are defined with  *
 *     the following template:                                              *
 *                                                                          *
 *          typedef struct _Astruct {                                       *
 *             ...                                                          *
 *             struct _Astruct *nextfield;                                  *
 *          } Astruct;                                                      *
 *                                                                          *
 *     _Astruct is the "structure tag".  It's declared explicitly so that   *
 *     the pointer field inside would recognize what it's pointing to as a  *
 *     defined structure.  Then, typedef defines a synonym, "Astruct", for  *
 *     "struct _Astruct".                                                   *
 *                                                                          *
 ****************************************************************************/


typedef struct _dbucketstruct {

  int *index;                        /* Subscript indices of nonzeroes.     */
  double *value;           /* Values of nonzeroes.                */

#ifdef _DEBUG                        /* Debugging Flag Field:               */
  int errorflag;                     /*    to ensure integrity of mallocs.  */
#endif

  struct _dbucketstruct *nextbucket;  /* nsia.                          */
} dbucketstruct;

/*             ___________________bucketstruct___________________           *
 *             |                                                |           *
 *             |                   ___index___   ____value___   |           *
 *             |                   |         |   |          |   |           *
 *             |                 0 |  (int)  |   | (double) |   |           *
 *             |                   |_________|   |__________|   |           *
 *             |                   |         |   |          |   |           *
 *             |                 1 |  (int)  |   | (double) |   |           *
 *             |                   |_________|   |__________|   |           *
 *             |                   |    .    |   |     .    |   |           *
 *             |                   |    .    |   |     .    |   |           *
 *             |                   |    .    |   |     .    |   |           *
 *             |                   |_________|   |__________|   |           *
 *             |                   |         |   |          |   |           *
 *             | dBUCKET_DIM-1 |  (int)  |   | (double) |   |           *
 *             |                   |_________|   |__________|   |           *
 *             |                                                |           *
 *             |             nextbucket *                       |           *
 *             |                        |                       |           *
 *             |________________________|_______________________|           *
 *                                      |                                   *
 *                                      V                                   *
 *                               (dbucketstruct)                        */




typedef struct _dibucketstruct {
  int *index_RANGE_BEGIN;
  int *index_RANGE_END;
                                     /* Subscript index pairs of range of   *
				      * nonzeroes.  Note: since in C, 2-d   *
				      * array elements are stored by row,   *
				      * index_RANGE_BEGIN and               *
				      * index_RANGE_END,                    *
				      * represent a range and are stored    *
				      * contiguously, which is advantageous *
				      * from a cache memory point of view,  *
				      * since they are also accessed in     *
				      * in tandem.                          */

#ifdef _DEBUG                        /* Debugging Flag Field:               */
  int errorflag;                     /*    to ensure integrity of mallocs.  */
#endif

  struct _dibucketstruct *nextibucket;  /* nsia.                        */
} dibucketstruct;

/*     __________________________ibucketstruct__________________________    *
 *     |                                                               |    *
 *     |                       ______index______                       |    *
 *     |                       |       |       |                       |    *
 *     |                     0 | (int) | (int) |                       |    *
 *     |                       |_______|_______|                       |    *
 *     |                       |       |       |                       |    *
 *     |                     1 | (int) | (int) |                       |    *
 *     |                       |_______|_______|                       |    *
 *     |                       |   .   |   .   |                       |    *
 *     |                       |   .   |   .   |                       |    *
 *     |                       |   .   |   .   |                       |    *
 *     |                       |_______|_______|                       |    *
 *     |                       |       |       |                       |    *
 *     |    dIBUCKET_DIM-1 | (int) | (int) |                       |    *
 *     |                       |_______|_______|                       |    *
 *     |                                                               |    *
 *     |                   nextibucket *                               |    *
 *     |                               |                               |    *
 *     |_______________________________|_______________________________|    *
 *                                     |                                    *
 *                                     V                                    *
 *                              (dibucketstruct)                        */



typedef struct _dvbucketstruct {

  double *value;           /* Values of nonzeroes.                */

#ifdef _DEBUG                        /* Debugging Flag Field:               */
  int errorflag;                     /*    to ensure integrity of mallocs.  */
#endif

  struct _dvbucketstruct *nextvbucket; /* nsia.                         */
} dvbucketstruct;

/*                   _____________vbucketstruct_____________                *
 *                   |                                     |                *
 *                   |                    ____value___     |                *
 *                   |                    |          |     |                *
 *                   |                  0 | (double) |     |                *
 *                   |                    |__________|     |                *
 *                   |                    |          |     |                *
 *                   |                  1 | (double) |     |                *
 *                   |                    |__________|     |                *
 *                   |                    |     .    |     |                *
 *                   |                    |     .    |     |                *
 *                   |                    |     .    |     |                *
 *                   |                    |__________|     |                *
 *                   |                    |          |     |                *
 *                   | dVBUCKET_DIM-1 | (double) |     |                *
 *                   |                    |__________|     |                *
 *                   |                                     |                *
 *                   |      nextvbucket *                  |                *
 *                   |                  |                  |                *
 *                   |__________________|__________________|                *
 *                                      |                                   *
 *                                      V                                   *
 *                               (dvbucketstruct)                       */


/*
 * Sparse Vector Internal Data Structure (DS) Types: S1, SS, CS
 */

typedef struct _dS1struct {
  int index;

  double value;

} dS1struct;

typedef struct _dSSstruct {
  dbucketstruct *bucketptr;      /* Pointer to linked list of buckets.  */
} dSSstruct;

typedef struct _dCSstruct {
  int nonzeroranges;                 /* Number of contiguous nonzero ranges.*/
  dibucketstruct *ibucketptr;    /* Pointer to linked list of ibuckets. */
  dvbucketstruct *vbucketptr;    /* Pointer to linked list of vbuckets. */
} dCSstruct;


typedef struct {
  int nonzeroes;                    /* The total # of                       *
				     * nonzeroes in the vector.             */
  int dstype;                       /* The data-structure type of the       *
				     * dvectorstruct will be one of the *
				     * following:
				     *          S1 = 1 SUBSCRIPT            *
				     *          SS = SINGLE SUBSCRIPTS      *
				     *          CS = COMPRESSED SUBSCRIPTS  */
  union _dds {
    dS1struct s1;
    dSSstruct ss;
    dCSstruct cs;
  } dds;

#ifdef _VTRACE

  /* Symbolic identifier */
  char* id;
  /* Numeric identifier */
  int num_id;

  /* Tracked? */
  unsigned char is_tracked;
  /* Named? */
  unsigned char is_named;
  /* Visualized? */
  unsigned char is_visual;

  /* List of all updates over the vector's lifetime */
  pu_struct u_list;

#endif

} dvectorstruct;

/*     ___________________________vectorstruct___________________________   *
 *     |                                                                |   *
 *     |          ____nonzeroes____         _____dstype______           |   *
 *     |          |               |         |               |           |   *
 *     |          |     (int)     |         |     (int)     |           |   *
 *     |          |_______________|         |_______________|           |   *
 *     |                                                                |   *
 *     |                               ds                               |   *
 *     |    {~~~~~~~~~~~~~~~~~~~~~~~~~~--~~~~~~~~~~~~~~~~~~~~~~~~~~}    |   *
 *     |    {                                                      }    |   *
 *     |    { EITHER:                                              }    |   *
 *     |    {     _____________________S1_____________________     }    |   *
 *     |    {     |   _____index_____     _____value_____    |     }    |   *
 *     |    {     |   |             |     |             |    |     }    |   *
 *     |    {     |   |    (int)    |     |    (int)    |    |     }    |   *
 *     |    {     |   |_____________|     |_____________|    |     }    |   *
 *     |    {     |__________________________________________|     }    |   *
 *     |    {                                                      }    |   *
 *     |    { OR:                                                  }    |   *
 *     |    {     _____________________SS_____________________     }    |   *
 *     |    {     |                                          |     }    |   *
 *     |    {     |           bucketptr *                    |     }    |   *
 *     |    {     |                     |                    |     }    |   *
 *     |    {     |_____________________|____________________|     }    |   *
 *     |    {                           |                          }    |   *
 *     |    {                           V                          }    |   *
 *     |    {                     (dbucketstruct)              }    |   *
 *     |    {                                                      }    |   *
 *     |    { OR:                                                  }    |   *
 *     |    {     _____________________CS_____________________     }    |   *
 *     |    {     |                                          |     }    |   *
 *     |    {     |              _nonzeroranges_             |     }    |   *
 *     |    {     |              |             |             |     }    |   *
 *     |    {     |              |    (int)    |             |     }    |   *
 *     |    {     |              |_____________|             |     }    |   *
 *     |    {     |                                          |     }    |   *
 *     |    {     |     ibucketptr *        vbucketptr *     |     }    |   *
 *     |    {     |                |                   |     |     }    |   *
 *     |    {     |________________|___________________|_____|     }    |   *
 *     |    {                      |                   |           }    |   *
 *     |    {                      V                   V           }    |   *
 *     |    {          (dibucketstruct)  (dvbucketstruct)  }    |   *
 *     |    {                                                      }    |   *
 *     |    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~    |   *
 *     |                                                                |   *
 *     |________________________________________________________________|   */




typedef struct _dvectorstorestruct {
  dvectorstruct vector[dVECTORSTORE_DIM];  /* Long array of vectors */
  struct _dvectorstorestruct *nextstore;       /* nsia.                 */
} dvectorstorestruct;

/*             ________________vectorstorestruct________________            *
 *             |                                               |            *
 *             |                        ______vector______     |            *
 *             |                        |                |     |            *
 *             |                      0 | (dvectorstruct) ||            *
 *             |                        |________________|     |            *
 *             |                        |                |     |            *
 *             |                      1 | (dvectorstruct) ||            *
 *             |                        |________________|     |            *
 *             |                        |        .       |     |            *
 *             |                        |        .       |     |            *
 *             |                        |        .       |     |            *
 *             |                        |________________|     |            *
 *             |                        |                |     |            *
 *             | dVECTORSTORE_DIM-1 | (dvectorstruct) ||            *
 *             |                        |________________|     |            *
 *             |                                               |            *
 *             |             nextstore *                       |            *
 *             |                       |                       |            *
 *             |_______________________|_______________________|            *
 *                                     |                                    *
 *                                     V                                    *
 *                            (dvectorstorestruct)                      */






typedef struct _dbucketstorestruct {
  dbucketstruct bucket[dBUCKETSTORE_DIM];  /* Long array of buckets.*/
  struct _dbucketstorestruct *nextstore;       /* nsia.                 */
} dbucketstorestruct;

/*             ________________bucketstorestruct________________            *
 *             |                                               |            *
 *             |                        ______bucket______     |            *
 *             |                        |                |     |            *
 *             |                      0 | (dbucketstruct) ||            *
 *             |                        |________________|     |            *
 *             |                        |                |     |            *
 *             |                      1 | (dbucketstruct) ||            *
 *             |                        |________________|     |            *
 *             |                        |        .       |     |            *
 *             |                        |        .       |     |            *
 *             |                        |        .       |     |            *
 *             |                        |________________|     |            *
 *             |                        |                |     |            *
 *             | dBUCKETSTORE_DIM-1 | (dbucketstruct) ||            *
 *             |                        |________________|     |            *
 *             |                                               |            *
 *             |             nextstore *                       |            *
 *             |                       |                       |            *
 *             |_______________________|_______________________|            *
 *                                     |                                    *
 *                                     V                                    *
 *                            (dbucketstorestruct)                      */


typedef struct _dibucketstorestruct {
  dibucketstruct ibucket[dIBUCKETSTORE_DIM];
                                                 /* Long array of ibuckets. */
  struct _dibucketstorestruct *nextstore;    /* nsia.                   */
} dibucketstorestruct;

/*             ________________ibucketstorestruct________________           *
 *             |                                               |            *
 *             |                        ______bucket______     |            *
 *             |                        |                |     |            *
 *             |                      0 |(dibucketstruct) ||            *
 *             |                        |________________|     |            *
 *             |                        |                |     |            *
 *             |                      1 |(dibucketstruct) ||            *
 *             |                        |________________|     |            *
 *             |                        |        .       |     |            *
 *             |                        |        .       |     |            *
 *             |                        |        .       |     |            *
 *             |                        |________________|     |            *
 *             |                        |                |     |            *
 *             | dIBUCKETSTORE_DIM-1| (dibucketstruct)||            *
 *             |                        |________________|     |            *
 *             |                                               |            *
 *             |             nextstore *                       |            *
 *             |                       |                       |            *
 *             |_______________________|_______________________|            *
 *                                     |                                    *
 *                                     V                                    *
 *                            (dibucketstorestruct)                     */



typedef struct _dvbucketstorestruct {
  dvbucketstruct vbucket[dVBUCKETSTORE_DIM];
                                                 /* Long array of vbuckets. */
  struct _dvbucketstorestruct *nextstore;    /* nsia.                   */
} dvbucketstorestruct;

/*             ________________vbucketstorestruct_________________          *
 *             |                                                 |          *
 *             |                         ______vbucket______     |          *
 *             |                         |                 |     |          *
 *             |                       0 | (dvbucketstruct) ||          *
 *             |                         |_________________|     |          *
 *             |                         |                 |     |          *
 *             |                       1 | (dvbucketstruct) ||          *
 *             |                         |_________________|     |          *
 *             |                         |         .       |     |          *
 *             |                         |         .       |     |          *
 *             |                         |         .       |     |          *
 *             |                         |_________________|     |          *
 *             |                         |                 |     |          *
 *             | dVBUCKETSTORE_DIM-1 | (dvbucketstruct) ||          *
 *             |                         |_________________|     |          *
 *             |                                                 |          *
 *             |              nextstore *                        |          *
 *             |                        |                        |          *
 *             |________________________|________________________|          *
 *                                      |                                   *
 *                                      V                                   *
 *                            (dvbucketstorestruct)                     */





/****************************************************************************
 *                                                                          *
 *  Function Prototyping                                                    *
 *                                                                          *
 ****************************************************************************/

#define _P(x) ()

void dallocatebucketstore _P(());
void dallocateibucketstore _P(());
void dallocatevbucketstore _P(());
void dallocatevectorstore _P(());

void dgetbucket _P((dbucketstruct **bucketptrptr));
void dgetibucket _P((dibucketstruct **ibucketptrptr));
void dgetvbucket _P((dvbucketstruct **vbucketptrptr));
void dgetvector _P((dvectorstruct **vectorptrptr));

void dspgx _P((dvectorstruct **destptrptr,
		   int arity,
		   double *alphaptrs[],
		   dvectorstruct **rhsvectorptrptrs[]));

void dnon_over_lapping_cs_range_axpy _P((int rangedestination,
				      int lhsdibucketindex,
				      int *addressoflastlhsdibucketindex,
				      int *addressoflhsdvbucketindex,
				      int k,
				      int *addressofrhsdvbucketindex_k,
				      double alpha_k,
				      int totalcurrentnonzerorange
#ifdef _DEBUG
				      ,int *addressofunconsumedcurrentnonzerorange
				      ,int rhsdvectorconsumednonzeroranges_k
				      ,int rhsdvectornonzeroranges_k
				      ,int lhsdvectornonzeroranges
#endif
				      ));

void doverlapping_cs_range_axpy _P((int *addressoflhsdvbucketindex,
				   int k,
				   int *addressofrhsdvbucketindex_k,
				   double alpha_k,
				   int totalcurrentnonzerorange
#ifdef _DEBUG
				   ,int *addressofunconsumedcurrentnonzerorange
				   ,int rhsdvectorconsumednonzeroranges_k
				   ,int rhsdvectornonzeroranges_k
				   ,int lhsdvectornonzeroranges
#endif
				   ));


void ddetermine_loop_bounds _P((int totalcurrentnonzerorange,
			       int lhsvbucketindex,
			       int rhsvbucketindex,
			       int *firstloopsize,
			       int *numberofdoubleforloopiterations,
			       int *firstiterationloopsize,
			       int *seconditerationloopsize,
			       int *remainder1loopsize,
			       int *remainder2loopsize
#ifdef _DEBUG
			       ,int *addressofunconsumedcurrentnonzerorange
#endif
			       ));

void ddebugging_code_1 _P((char *RIGHT_OR_LEFT,
			  int vbucketindex,
			  int unconsumedcurrentnonzerorange));

void dhop_to_next_rhsvbucket _P((int k,
				int *vbucketindex
#ifdef _DEBUG
				,int rhsdvectorconsumednonzeroranges
				,int rhsdvectornonzeroranges
#endif
				));


void dhop_to_next_lhsvbucket _P((int *vbucketindex
#ifdef _DEBUG
				,int lhsdvectornonzeroranges
#endif
				));

void dget_and_hop_to_next_vbucket _P((int *lhsdvbucketindex));

void dconvertdstype _P((dvectorstruct **oldvectorptrptr,
		       int k,
		       int newvectordstype));

void daxpy_0_1_post_processing _P((dvectorstruct **destptrptr));

void daxpy_2_heap_post_processing _P((dvectorstruct **destptrptr,
                                     dvectorstruct **rhsdvectorptrptrs[],
                                     double *alphaptrs[],
                                     dvectorstruct *restorerhsdvectorptr[MAXARITY+1],
                                     int arity,
                                     int lhsdvectornonzeroes,
                                     int lhsdvectornonzeroranges,
                                     int governingdstype));

void dplusequal_axpy_postprocessing _P((dvectorstruct **destptrptr,
				       dvectorstruct **rhsdvectorptrptr,
				       int status_otherrhsdvector,
				       dvectorstruct **restofplusequalvectorptrptr,
				       double *alphaptrs[2],
				       dvectorstruct *restorerhsdvectorptr[2+1],
				       int arity,
				       int lhsdvectornonzeroes,
				       int lhsdvectornonzeroranges,
				       int governingdstype
				       ));

#ifdef _PEEP
void dadpeep_pre _P((int vectorindex,
		    dvectorstruct **rhsvectorptrptr,
		    double *alphaptr));

void dadpeep_post _P((dvectorstruct **destptrptr));
#endif

void dadallocate _P(());
void dadfree _P(());
void dad_free_vectors_buckets _P((dvectorstruct **vectorptrptr));
void dinitialize_pd_global_variables _P(());

void dspsd _P((dvectorstruct **destptrptr,
		int *indexvector,
		double *valuevector,
		int *lengthptr));

void dspxd _P((double *vector,
			 int *lengthptr,
			 dvectorstruct **objectptrptr,
			 int *returned_len,
			 int *len_info));

void dspxm _P((double *vector,
			       int *lengthptr,
			       double *multiplier,
			       dvectorstruct **objectptrptr,
			       int *returned_len,
			       int *len_info));

void dspxa _P((double *vector,
				   int *lengthptr,
				   dvectorstruct **objectptrptr,
				   int *returned_len,
				   int *len_info));

void dspxs _P((int *indvector,
			  double *valvector,
			  int *lengthptr,
			  dvectorstruct **objectptrptr,
			  int *returned_len,
			  int *len_info)); 

void dsppr _P((dvectorstruct **objectptrptr, int *unit_number));

void dspzronotca _P((dvectorstruct **destptrptr));

void dspcp _P((dvectorstruct **destptrptr,
		dvectorstruct **oldvectorptrptr));

void dadinfo _P((int arity,
		int nnzb,
		dvectorstruct *lhsvectoraddress,
		int nnza,
		double *alphaptrs[],
		dvectorstruct **vectorptrptrs[]));

void dspg1 _P((dvectorstruct **destptrptr,
		double *alphaptr,
		dvectorstruct **vectoraptrptr));
void dspg2 _P((dvectorstruct **destptrptr,
		double *alphaptr,
		dvectorstruct **vectoraptrptr,
		double *betaptr,
		dvectorstruct **vectorbptrptr));
void plusequaldspg1 _P((dvectorstruct **rhsanddestptrptr,
			 double *alphaptr));
void plusequaldspg2 _P((dvectorstruct **rhsanddestptrptr,
			 double *otheralphaptr,
			 dvectorstruct **otherrhsvectorptrptr,
			 int status_otherrhsvector));



#endif /*  #ifndef dGAXPY_PD_MACROS_AND_DATA_STRUCTURES  */
