#define DOUBLE
/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#include  <stdlib.h>
#include "dgaxpy_pd_macros_and_data_structures.h"
#include "dgaxpy_pd_global_variables.h"


/****************************************************************************
 *                                                                          *
 *  dallocatebucketstore()                                                   *
 *                                                                          *
 *     1 malloc call gets all the memory for a bucketstore.  If malloc does *
 *     not return NULL, one of the following 2 scenarios will result:       *
 *                                                                          *
 *     A: outcome of dallocatebucketstore the 1st time it's called:          *
 *                                                                          *
 *     bucketstoreptr -----------+                                          *
 *                               |   store                                  *
 *                              _V_____1_____                               *
 *     bucketstorestartptr ---->|    _____  |                               *
 *     bucketstorecounter ------+->0 |___|  |                               *
 *                              |    | . |  |                               *
 *                              |    | . |  |                               *
 *                              |    | . |  |                               *
 *                              |    |___|  |                               *
 *                              |    |___|  |                               *
 *                              |           |                               *
 *                              | nextstore |                               *
 *                              |     *     |                               *
 *                              |_____|_____|                               *
 *                                    |                                     *
 *                                    V                                     *
 *                                   NULL                                   *
 *                                                                          *
 *     B: outcome of dallocatebucketstore the Nth time it's called (N > 1):  *
 *                                                                          *
 *     bucketstorecounter --------------------------------------+           *
 *                                                              |           *
 *     bucketstoreptr ----------------------------------------+ |           *
 *                               store        store           | | store     *
 *                              ___1___      ___2___ ...     _V_|___N_____  *
 *     bucketstorestartptr ---->|     |  +-->|     |     +-->|  V _____  |  *
 *                              |     |  |   |     |     |   |  0 |___|  |  *
 *                              |     |  |   |     |     |   |    | . |  |  *
 *                              |     |  |   |     |     |   |    | . |  |  *
 *                              |     |  |   |     |     |   |    | . |  |  *
 *                              |     |  |   |     |     |   |    |___|  |  *
 *                              |     |  |   |     |     |   |    |___|  |  *
 *                              |     |  |   |     |     |   |           |  *
 *                              |     |  |   |     |     |   | nextstore |  *
 *                              |  *  |  |   |  *  |     |   |     *     |  *
 *                              |__|__|  |   |__|__|     |   |_____|_____|  *
 *                                 |     |      |        |         |        *
 *                                 +-----+      +---...--+         V        *
 *                                                                NULL      *
 *                                                                          *
 ****************************************************************************/

void dallocatebucketstore()
{
  int i,k;
  
  dbucketstorestruct *tempbucketstoreptr;
  dbucketstruct *tempbucketptr;
  
  int *index_store;
  double *value_store;

  int temp_size;

  /*
   * Per BUCKET Dynamic Allocation: constant dBUCKETSTORE_DIM
   *                                variable dBUCKET_DIM
   */
  
  if ((tempbucketstoreptr = (dbucketstorestruct *)
       palloc ((temp_size = (sizeof(dbucketstorestruct)))))==NULL) {
    printf("dallocatebucketstore error: malloc returned NULL.\n");
    abort();
  }
  memory_count(D_EVAL,MEM_BUCKET,temp_size);
  
  for (k=0; k<dBUCKETSTORE_DIM; k++) {
    tempbucketptr = &(tempbucketstoreptr->bucket[k]);
    if ((tempbucketptr->index = (int *)
	 palloc ((temp_size = (dBUCKET_DIM * sizeof(int)))))==NULL) {
      printf("dallocatebucketstore error: malloc returned NULL.\n");
      abort();
    }
    memory_count(D_EVAL,MEM_BUCKET,temp_size);
    if ((tempbucketptr->value = (double *)
	 palloc ((temp_size = (dBUCKET_DIM *
			       sizeof(double)))))==NULL) {
      printf("dallocatebucketstore error: malloc returned NULL.\n");
      abort();
    }
    memory_count(D_EVAL,MEM_BUCKET,temp_size);
  }
      
  if (dbucketstorestartptr == NULL)
    dbucketstorestartptr = tempbucketstoreptr;
  else
    dbucketstoreptr->nextstore = tempbucketstoreptr;
  
  dbucketstoreptr = tempbucketstoreptr;
  dbucketstoreptr->nextstore = NULL;
  dbucketstorecounter = 0;
  
  for (i=0;i<dBUCKETSTORE_DIM;i++) {
    (&(dbucketstoreptr->bucket[i]))->nextbucket = NULL;
#ifdef _DEBUG  
    (&(dbucketstoreptr->bucket[i]))->errorflag = 9999;
#endif
  }
}



/****************************************************************************
 *                                                                          *
 *  dallocateibucketstore()                                                  *
 *                                                                          *
 *     Identical to dallocatebucketstore in terms of control flow.           *
 *                                                                          *
 ****************************************************************************/

void dallocateibucketstore()
{
  int i,k;

  dibucketstorestruct *tempibucketstoreptr;
  dibucketstruct *tempibucketptr;
  
  int *index_RANGE_BEGIN_store;
  int *index_RANGE_END_store;

  int temp_size;
 
  /*
   * Per IBUCKET Dynamic Allocation: constant dIBUCKETSTORE_DIM
   *                                 variable dIBUCKET_DIM
   */
  
  if ((tempibucketstoreptr = (dibucketstorestruct *)
       palloc ((temp_size = (sizeof(dibucketstorestruct)))))==NULL) {
    printf("dallocateibucketstore error: malloc returned NULL.\n");
    abort();
  }
  memory_count(D_EVAL,MEM_IBUCKET,temp_size);
  
  for (k=0; k<dIBUCKETSTORE_DIM; k++) {
    tempibucketptr = &(tempibucketstoreptr->ibucket[k]);
    if ((tempibucketptr->index_RANGE_BEGIN = (int *)
	 palloc ((temp_size = (dIBUCKET_DIM * sizeof(int)))))==NULL) {
      printf("dallocateibucketstore error: malloc returned NULL.\n");
      abort();
    }
    memory_count(D_EVAL,MEM_IBUCKET,temp_size);
    if ((tempibucketptr->index_RANGE_END = (int *)
	 palloc ((temp_size = (dIBUCKET_DIM * sizeof(int)))))==NULL) {
      printf("dallocateibucketstore error: malloc returned NULL.\n");
      abort();
    }
    memory_count(D_EVAL,MEM_IBUCKET,temp_size);
  }
  
  if (dibucketstorestartptr == NULL)
    dibucketstorestartptr = tempibucketstoreptr;
  else
    dibucketstoreptr->nextstore = tempibucketstoreptr;
  
  dibucketstoreptr = tempibucketstoreptr;
  dibucketstoreptr->nextstore = NULL;
  dibucketstorecounter = 0;
  
  for (i=0;i<dIBUCKETSTORE_DIM;i++) {
    (&(dibucketstoreptr->ibucket[i]))->nextibucket = NULL;
#ifdef _DEBUG  
    (&(dibucketstoreptr->ibucket[i]))->errorflag = 9099;
#endif
  }
}




/****************************************************************************
 *                                                                          *
 *  dallocatevbucketstore()                                                  *
 *                                                                          *
 *     Identical to dallocatebucketstore in terms of control flow.           *
 *                                                                          *
 ****************************************************************************/

void dallocatevbucketstore()
{
  int i,k;

  dvbucketstorestruct *tempvbucketstoreptr;
  dvbucketstruct *tempvbucketptr;
  
  double *value_store;

  int temp_size;
  
  /*
   * Per VBUCKET Dynamic Allocation: constant dVBUCKETSTORE_DIM
   *                                 variable dVBUCKET_DIM
   */
  
  if ((tempvbucketstoreptr = (dvbucketstorestruct *)
       palloc ((temp_size = (sizeof(dvbucketstorestruct)))))==NULL) {
    printf("dallocatevbucketstore error: malloc returned NULL.\n");
    abort();
  }
  memory_count(D_EVAL,MEM_VBUCKET,temp_size);
  
  for (k=0; k<dVBUCKETSTORE_DIM; k++) {
    tempvbucketptr = &(tempvbucketstoreptr->vbucket[k]);
    if ((tempvbucketptr->value = (double *)
	 palloc ((temp_size = (dVBUCKET_DIM *
			       sizeof(double)))))==NULL) {
      printf("dallocatevbucketstore error: malloc returned NULL.\n");
      abort();
    }
    memory_count(D_EVAL,MEM_VBUCKET,temp_size);
  }
  
  if (dvbucketstorestartptr == NULL)
    dvbucketstorestartptr = tempvbucketstoreptr;
  else
    dvbucketstoreptr->nextstore = tempvbucketstoreptr;
  
  dvbucketstoreptr = tempvbucketstoreptr;
  dvbucketstoreptr->nextstore = NULL;
  dvbucketstorecounter = 0;
  
  for (i=0;i<dVBUCKETSTORE_DIM;i++) {
    (&(dvbucketstoreptr->vbucket[i]))->nextvbucket = NULL;
#ifdef _DEBUG  
    (&(dvbucketstoreptr->vbucket[i]))->errorflag = 9009;
#endif
  }
}




/****************************************************************************
 *                                                                          *
 *  dallocatevectorstore()                                                   *
 *                                                                          *
 *     dVECTORSTORE_DIM is a constant and in contrast to the buckets    *
 *     there is no VECTOR_DIM.  Therefore there is no need for any          *
 *     allocation scheme requiring the per STORE vs per VECTOR methods.     *
 *                                                                          *
 ****************************************************************************/

void dallocatevectorstore()
{
  int i;
  dvectorstorestruct *tempvectorstoreptr;
  dvectorstruct *tempvectorptr;

  int temp_size;

  /*
   * Wholesale Dynamic Allocation: constant dVECTORSTORE_DIM
   * This allocation is not obsolete unlike the bucket type stores
   * above, since there is no variable-sized bucket in a vectorstruct.
   */
  if ((tempvectorstoreptr = (dvectorstorestruct *)
       palloc ((temp_size = (sizeof(dvectorstorestruct)))))==NULL ) {
    printf("dallocatevectorstore error: malloc returned NULL.\n");
    abort();
  }
  memory_count(D_EVAL,MEM_VECTOR,temp_size);
  
  /*
   * End of Wholesale Dynamic Allocation
   */
  
  if (dvectorstorestartptr == NULL)
    dvectorstorestartptr = tempvectorstoreptr;
  else
    dvectorstoreptr->nextstore = tempvectorstoreptr;
  
  dvectorstoreptr = tempvectorstoreptr;
  dvectorstoreptr->nextstore = NULL;
  dvectorstorecounter = 0;
  for (i=0;i<dVECTORSTORE_DIM;i++) {
    tempvectorptr = &(dvectorstoreptr->vector[i]);

/*
 *  PMK: 12/4/96 edits
 *
 *  The assignment of the S0 "non-yet-specified" data structure type is
 *  extraneous for now, since we never have checks on S0 (i.e., nowhere
 *  in the code do we say, if (somevectorptr->dstype == S0) ...
 *  Therefore, the following code line is commented out:

    tempvectorptr->dstype = S0;

 *
 */

  }
}



/****************************************************************************
 *                                                                          *
 *  dgetbucket(bucketptrptr)                                                 *
 *                                                                          *
 *     bucketptrptr    is a doubly-indirected pointer to a dbucketstruct*
 *                     Since we want the value of "bucketptr" to be changed *
 *                     by dgetbucket (i.e., *bucketptrptr to get assigned a  *
 *                     pointer to a dbucketstruct), we need to make its *
 *                     address (bucketptrptr) the function argument.  Thus, *
 *                     the need for the double indirection (**).            *
 *                                                                          *
 *     dgetbucket assigns to *bucketptrptr the pointer to the next available *
 *     dbucketstruct from the freedbucketslist or if none there from the*
 *     bucketstore.  In the latter case, if we're at the end of the current *
 *     bucketstore, a new store is allocated before doing the "get".        *
 *                                                                          *
 ****************************************************************************/

void dgetbucket(bucketptrptr)
dbucketstruct **bucketptrptr;
{
  if (freeddbucketslist != NULL) {
    (*bucketptrptr) = freeddbucketslist;
    freeddbucketslist = freeddbucketslist->nextbucket;
  }
  else {
    if (dbucketstorecounter == dBUCKETSTORE_DIM) {
      dallocatebucketstore();
    }
    
#ifdef _DEBUG
    else if (dbucketstorecounter > dBUCKETSTORE_DIM) {
      printf("dgetbucket error: dbucketstorecounter=%d,  STOREDIMENSION=%d\n",
	     dbucketstorecounter,dBUCKETSTORE_DIM);
      abort();
    }
#endif

    /*
     *  The following assignment of a store entry to *bucketptrptr
     *  works regardless of whether we've used per BUCKET or per STORE
     *  allocation method.
     */

    *bucketptrptr = (dbucketstruct *)
                     &(dbucketstoreptr->bucket[dbucketstorecounter++]);
  }
  (*bucketptrptr)->nextbucket = NULL;
  
#ifdef _DEBUG
  if ((*bucketptrptr)->errorflag != 9999) {
    printf("dgetbucket error: memory corruption of dbucketstore\n");
    abort();
  }
#endif
  
}

/****************************************************************************
 *                                                                          *
 *  dgetibucket(ibucketptrptr)                                               *
 *                                                                          *
 *     ibucketptrptr   is a doubly-indirected pointer to a dibucketstruc*
 *                     Since we want the value of "ibucketptr" to be        *
 *                     changed by dgetibucket (i.e., *ibucketptrptr to get   *
 *                     assigned a  pointer to a dibucketstruct), we need*
 *                     to make its address (ibucketptrptr) the function     *
 *                     argument.  Thus, the need for the double indirection *
 *                     (**).                                                *
 *                                                                          *
 *     dgetibucket assigns to *ibucketptrptr the pointer to the next         *
 *     available dibucketstruct from the freedibucketslist or if none   *
 *     from the ibucketstore.  In the latter case, if we're at the end      *
 *     of the current ibucketstore, a new store is allocated before doing   *
 *     the "get".                                                           *
 *                                                                          *
 ****************************************************************************/

void dgetibucket(ibucketptrptr)
dibucketstruct **ibucketptrptr;
{
  if (freeddibucketslist != NULL) {
    (*ibucketptrptr) = freeddibucketslist;
    freeddibucketslist = freeddibucketslist->nextibucket;
  }
  else {
    if (dibucketstorecounter == dIBUCKETSTORE_DIM) {
      dallocateibucketstore();
    }
    
#ifdef _DEBUG
    else if (dibucketstorecounter > dIBUCKETSTORE_DIM) {
      printf("dgetibucket error: dibucketstorecounter=%d,  STOREDIMENSION=%d\n",
	     dibucketstorecounter,dIBUCKETSTORE_DIM);
      abort();
    }
#endif

    *ibucketptrptr = (dibucketstruct *)
              &(dibucketstoreptr->ibucket[dibucketstorecounter++]);
  }
  (*ibucketptrptr)->nextibucket = NULL;
  
#ifdef _DEBUG
  if ((*ibucketptrptr)->errorflag != 9099) {
    printf("dgetibucket error: memory corruption of dibucketstore\n");
    abort();
  }
#endif
  
}


/****************************************************************************
 *                                                                          *
 *  dgetvbucket(vbucketptrptr)                                               *
 *                                                                          *
 *     vbucketptrptr   is a doubly-indirected pointer to a dvbucketstruc*
 *                     Since we want the value of "vbucketptr" to be        *
 *                     changed by dgetvbucket (i.e., *vbucketptrptr to get   *
 *                     assigned a  pointer to a dvbucketstruct), we need*
 *                     to make its address (vbucketptrptr) the function     *
 *                     argument.  Thus, the need for the double indirection *
 *                     (**).                                                *
 *                                                                          *
 *     dgetvbucket assigns to *vbucketptrptr the pointer to the next         *
 *     available dvbucketstruct from the freedvbucketslist or if none   *
 *     there from the vbucketstore.  In the latter case, if we're at the end*
 *     of the current vbucketstore, a new store is allocated before doing   *
 *     the "get".                                                           *
 *                                                                          *
 ****************************************************************************/

void dgetvbucket(vbucketptrptr)
dvbucketstruct **vbucketptrptr;
{
  if (freeddvbucketslist != NULL) {
    (*vbucketptrptr) = freeddvbucketslist;
    freeddvbucketslist = freeddvbucketslist->nextvbucket;
  }
  else {
    if (dvbucketstorecounter == dVBUCKETSTORE_DIM) {
      dallocatevbucketstore();
    }
    
#ifdef _DEBUG
    else if (dvbucketstorecounter > dVBUCKETSTORE_DIM) {
      printf("dgetvbucket error: dvbucketstorecounter=%d,  STOREDIMENSION=%d\n",
	     dvbucketstorecounter,dVBUCKETSTORE_DIM);
      abort();
    }
#endif

    *vbucketptrptr = (dvbucketstruct *)
              &(dvbucketstoreptr->vbucket[dvbucketstorecounter++]);
  }
  (*vbucketptrptr)->nextvbucket = NULL;
  
#ifdef _DEBUG
  if ((*vbucketptrptr)->errorflag != 9009) {
    printf("dgetvbucket error: memory corruption of dvbucketstore\n");
    abort();
  }
#endif
  
}



/****************************************************************************
 *                                                                          *
 *  dgetvector(vectorptrptr)                                                 *
 *                                                                          *
 *     vectorptrptr  is a doubly-indirected pointer to a dvectorstruct. *
 *                                                                          *
 *     dgetvector works in the same way as dgetvector, except that there is   *
 *     no freed list of vectors.  It replaces the functionality of "adnew". *
 *                                                                          *
 ****************************************************************************/

void dgetvector(vectorptrptr)
dvectorstruct **vectorptrptr;
{
  if (dvectorstorecounter == dVECTORSTORE_DIM) {
    dallocatevectorstore();
  }

#ifdef _DEBUG
  else if (dvectorstorecounter > dVECTORSTORE_DIM) {
    printf("dgetvector error: dvectorstorecount=%d, dVECTORSTORE_DIM=%d\n",
	   dvectorstorecounter,dVECTORSTORE_DIM);
    abort();
  }
#endif

  *vectorptrptr = (dvectorstruct *)
                    &(dvectorstoreptr->vector[dvectorstorecounter++]);

#ifdef _TRACE
  dadinfo(-1,0,*vectorptrptr,0,NULL,NULL);
#endif

}

