/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#include  <stdlib.h>
#include "sgaxpy_pd_macros_and_data_structures.h"
#include "dgaxpy_pd_macros_and_data_structures.h"
#include "cgaxpy_pd_macros_and_data_structures.h"
#include "zgaxpy_pd_macros_and_data_structures.h"
#include "sgaxpy_pd_global_variables.h"
#include "dgaxpy_pd_global_variables.h"
#include "cgaxpy_pd_global_variables.h"
#include "zgaxpy_pd_global_variables.h"

#ifdef _CHECK_FLOATING_PT
#include <floatingpoint.h>
int ieee_handler _P((char action[], 
		     char exception[],
		     sigfpe_handler_type hdl));
#endif

#define NOT_SET  0
#define SET      1

static int ALLOCATION_FLAG = NOT_SET;  /* When the SparsLinC library is     *
					* first invoked,                    *
					* this flag is NOT_SET.  After the  *
					* first call to xspini, and from    *
					* then on, it's SET.                */

static int OPT_S_BUCK_SZ_FLAG      = NOT_SET;
                                       /* This and the following flags      *
					* ensure that only those parameters *
					* not already defined by xspcnx     *
					* will acquire default values.      */
static int OPT_D_BUCK_SZ_FLAG      = NOT_SET;
static int OPT_C_BUCK_SZ_FLAG      = NOT_SET;
static int OPT_Z_BUCK_SZ_FLAG      = NOT_SET;
static int OPT_S_IBUCK_SZ_FLAG     = NOT_SET;
static int OPT_D_IBUCK_SZ_FLAG     = NOT_SET;
static int OPT_C_IBUCK_SZ_FLAG     = NOT_SET;
static int OPT_Z_IBUCK_SZ_FLAG     = NOT_SET;
static int OPT_S_VBUCK_SZ_FLAG     = NOT_SET;
static int OPT_D_VBUCK_SZ_FLAG     = NOT_SET;
static int OPT_C_VBUCK_SZ_FLAG     = NOT_SET;
static int OPT_Z_VBUCK_SZ_FLAG     = NOT_SET;
static int OPT_G_P_FLAG            = NOT_SET;
static int OPT_SWITCH_PERCENT_FLAG = NOT_SET;
static int OPT_SWITCH_NONZ_FLAG    = NOT_SET;


/****************************************************************************
 *                                                                          *
 *    xspini()                                                              *
 *                                                                          *
 *    xspini (used to be called all_allocate) performs all precision-       *
 *    dependent allocates.                                                  *
 *                                                                          *
 ****************************************************************************/

void xspini()
{

#ifdef _CHECK_FLOATING_PT
  int ieeer;
  ieeer = ieee_handler("set","common",SIGFPE_ABORT);
#endif

  if ( ALLOCATION_FLAG == NOT_SET ) {
    initialize_pi_global_variables();
    sadallocate();
    dadallocate();
    cadallocate();
    zadallocate();
    
    allocateheapstore();

    ALLOCATION_FLAG = SET;
  }
}


/****************************************************************************
 *                                                                          *
 *    xspfra()                                                              *
 *                                                                          *
 *    xspfra frees the dynamically-allocated memory for all precisions.     *
 *                                                                          *
 ****************************************************************************/

void xspfra()
{
  if ( ALLOCATION_FLAG == SET ) {
    sadfree();
    dadfree();
    cadfree();
    zadfree();
  }
  else {
    printf("xspfra Error: Nothing has been allocated.\n");
    abort();
  }
}


/****************************************************************************
 *                                                                          *
 *    xspcnx(opt,value)                                                     *
 *                                                                          *
 *       opt          Option number associated with the parameter.          *
 *                                                                          *
 *       value        New value to which the parameter is set.              *
 *                                                                          *
 *    xspcnx          allows the user to customize SparsLinC for each run.  *
 *                                                                          *
 ****************************************************************************/

void xspcnx(opt,value)
int *opt,*value;
{

  if ( (*opt) >= OPT_S_BUCK_SZ && (*opt) <= OPT_ALL_VBUCK_SZ ) {
    if ( ALLOCATION_FLAG == SET) {
      printf("xspcnx Error: Can't set [iv]bucket size after calling xspini\n");
      printf("              OPT = %d\n",(*opt));
      abort();
    }
    if ( (*value) <= 1 || ! ispower2(*value) ) {
      printf("xspcnx Error: Parameter value must be > 1 and a power of 2.\n");
      printf("              value = %d\n",(*value));
      abort();
    }
  }

  switch ((*opt)) {
    case OPT_S_BUCK_SZ:
	 sBUCKET_DIM = (*value);
	 sBUCKET_DIM_minus1 = sBUCKET_DIM - 1;
	 OPT_S_BUCK_SZ_FLAG = SET;
	 break;
    case OPT_D_BUCK_SZ:
	 dBUCKET_DIM = (*value);
	 dBUCKET_DIM_minus1 = dBUCKET_DIM - 1;
	 OPT_D_BUCK_SZ_FLAG = SET;
	 break;
    case OPT_C_BUCK_SZ:
	 cBUCKET_DIM = (*value);
	 cBUCKET_DIM_minus1 = cBUCKET_DIM - 1;
	 OPT_C_BUCK_SZ_FLAG = SET;
	 break;
    case OPT_Z_BUCK_SZ:
	 zBUCKET_DIM = (*value);
	 zBUCKET_DIM_minus1 = zBUCKET_DIM - 1;
	 OPT_Z_BUCK_SZ_FLAG = SET;
	 break;
    case OPT_ALL_BUCK_SZ:
	 sBUCKET_DIM = (*value);
	 sBUCKET_DIM_minus1 = sBUCKET_DIM - 1;
	 dBUCKET_DIM = (*value);
	 dBUCKET_DIM_minus1 = dBUCKET_DIM - 1;
	 cBUCKET_DIM = (*value);
	 cBUCKET_DIM_minus1 = cBUCKET_DIM - 1;
	 zBUCKET_DIM = (*value);
	 zBUCKET_DIM_minus1 = zBUCKET_DIM - 1;
	 OPT_S_BUCK_SZ_FLAG = SET;
	 OPT_D_BUCK_SZ_FLAG = SET;
	 OPT_C_BUCK_SZ_FLAG = SET;
	 OPT_Z_BUCK_SZ_FLAG = SET;
	 break;
    case OPT_S_IBUCK_SZ:
	 sIBUCKET_DIM = (*value);
	 sIBUCKET_DIM_minus1 = sIBUCKET_DIM - 1;
	 OPT_S_IBUCK_SZ_FLAG = SET;
	 break;
    case OPT_D_IBUCK_SZ:
	 dIBUCKET_DIM = (*value);
	 dIBUCKET_DIM_minus1 = dIBUCKET_DIM - 1;
	 OPT_D_IBUCK_SZ_FLAG = SET;
	 break;
    case OPT_C_IBUCK_SZ:
	 cIBUCKET_DIM = (*value);
	 cIBUCKET_DIM_minus1 = cIBUCKET_DIM - 1;
	 OPT_C_IBUCK_SZ_FLAG = SET;
	 break;
    case OPT_Z_IBUCK_SZ:
	 zIBUCKET_DIM = (*value);
	 zIBUCKET_DIM_minus1 = zIBUCKET_DIM - 1;
	 OPT_Z_IBUCK_SZ_FLAG = SET;
	 break;
    case OPT_ALL_IBUCK_SZ:
	 sIBUCKET_DIM = (*value);
	 sIBUCKET_DIM_minus1 = sIBUCKET_DIM - 1;
	 dIBUCKET_DIM = (*value);
	 dIBUCKET_DIM_minus1 = dIBUCKET_DIM - 1;
	 cIBUCKET_DIM = (*value);
	 cIBUCKET_DIM_minus1 = cIBUCKET_DIM - 1;
	 zIBUCKET_DIM = (*value);
	 zIBUCKET_DIM_minus1 = zIBUCKET_DIM - 1;
	 OPT_S_IBUCK_SZ_FLAG = SET;
	 OPT_D_IBUCK_SZ_FLAG = SET;
	 OPT_C_IBUCK_SZ_FLAG = SET;
	 OPT_Z_IBUCK_SZ_FLAG = SET;
	 break;
    case OPT_S_VBUCK_SZ:
	 sVBUCKET_DIM = (*value);
	 sVBUCKET_DIM_minus1 = sVBUCKET_DIM - 1;
	 OPT_S_VBUCK_SZ_FLAG = SET;
	 break;
    case OPT_D_VBUCK_SZ:
	 dVBUCKET_DIM = (*value);
	 dVBUCKET_DIM_minus1 = dVBUCKET_DIM - 1;
	 OPT_D_VBUCK_SZ_FLAG = SET;
	 break;
    case OPT_C_VBUCK_SZ:
	 cVBUCKET_DIM = (*value);
	 cVBUCKET_DIM_minus1 = cVBUCKET_DIM - 1;
	 OPT_C_VBUCK_SZ_FLAG = SET;
	 break;
    case OPT_Z_VBUCK_SZ:
	 zVBUCKET_DIM = (*value);
	 zVBUCKET_DIM_minus1 = zVBUCKET_DIM - 1;
	 OPT_Z_VBUCK_SZ_FLAG = SET;
	 break;
    case OPT_ALL_VBUCK_SZ:
	 sVBUCKET_DIM = (*value);
	 sVBUCKET_DIM_minus1 = sVBUCKET_DIM - 1;
	 dVBUCKET_DIM = (*value);
	 dVBUCKET_DIM_minus1 = dVBUCKET_DIM - 1;
	 cVBUCKET_DIM = (*value);
	 cVBUCKET_DIM_minus1 = cVBUCKET_DIM - 1;
	 zVBUCKET_DIM = (*value);
	 zVBUCKET_DIM_minus1 = zVBUCKET_DIM - 1;
	 OPT_S_VBUCK_SZ_FLAG = SET;
	 OPT_D_VBUCK_SZ_FLAG = SET;
	 OPT_C_VBUCK_SZ_FLAG = SET;
	 OPT_Z_VBUCK_SZ_FLAG = SET;
	 break;
    case OPT_G_P:
	 GLOBAL_MAXINDEX = (*value);
	 OPT_G_P_FLAG = SET;
	 INITIALIZED_MAXINDEX = TRUE;
	 break;
    case OPT_SWITCH_PERCENT:
	 SWITCH_NZ_RATIO = (float) (*value) / 100.0;
	 OPT_SWITCH_PERCENT_FLAG = SET;
	 break;
    case OPT_SWITCH_NONZ:
	 SWITCH_NZ_VALUE = (*value);
	 OPT_SWITCH_NONZ_FLAG = SET;
	 break;
    default:
	 printf("xspcnx Error: unknown option %d\n",(*opt));
	 abort();
  }
}


/****************************************************************************
 *                                                                          *
 *    xspcnf(terse_opt,value)                                               *
 *                                                                          *
 *       terse_opt    Option number associated with the parameter.          *
 *                                                                          *
 *       value        New value to which the parameter is set.              *
 *                                                                          *
 *    xspcnf          gives the user some SparsLinC customization capability*
 *                    It's the terse version of xspcnx.                     *
 *                    Note: the terse_option numbers are DIFFERENT than the *
 *                          the option numbers in xspcnx                    *
 *                                                                          *
 ****************************************************************************/

void xspcnf(terse_opt,value)
int *terse_opt,*value;
{

  if ( (*terse_opt) >= TERSE_OPT_ALL_BUCK_SZ &&
       (*terse_opt) <= TERSE_OPT_ALL_VBUCK_SZ ) {
    if ( ALLOCATION_FLAG == SET) {
      printf("xspcnf Error: Can't set [iv]bucket size after calling xspini\n");
      printf("              TERSE_OPT = %d\n",(*terse_opt));
      abort();
    }
    if ( (*value) <= 1 || ! ispower2(*value) ) {
      printf("xspcnf Error: Parameter value must be > 1 and a power of 2.\n");
      printf("              value = %d\n",(*value));
      abort();
    }
  }



  switch ((*terse_opt)) {
    case TERSE_OPT_ALL_BUCK_SZ:
	 sBUCKET_DIM = (*value);
	 sBUCKET_DIM_minus1 = sBUCKET_DIM - 1;
	 dBUCKET_DIM = (*value);
	 dBUCKET_DIM_minus1 = dBUCKET_DIM - 1;
	 cBUCKET_DIM = (*value);
	 cBUCKET_DIM_minus1 = cBUCKET_DIM - 1;
	 zBUCKET_DIM = (*value);
	 zBUCKET_DIM_minus1 = zBUCKET_DIM - 1;
	 OPT_S_BUCK_SZ_FLAG = SET;
	 OPT_D_BUCK_SZ_FLAG = SET;
	 OPT_C_BUCK_SZ_FLAG = SET;
	 OPT_Z_BUCK_SZ_FLAG = SET;
	 break;
    case TERSE_OPT_ALL_VBUCK_SZ:
	 sVBUCKET_DIM = (*value);
	 sVBUCKET_DIM_minus1 = sVBUCKET_DIM - 1;
	 dVBUCKET_DIM = (*value);
	 dVBUCKET_DIM_minus1 = dVBUCKET_DIM - 1;
	 cVBUCKET_DIM = (*value);
	 cVBUCKET_DIM_minus1 = cVBUCKET_DIM - 1;
	 zVBUCKET_DIM = (*value);
	 zVBUCKET_DIM_minus1 = zVBUCKET_DIM - 1;
	 OPT_S_VBUCK_SZ_FLAG = SET;
	 OPT_D_VBUCK_SZ_FLAG = SET;
	 OPT_C_VBUCK_SZ_FLAG = SET;
	 OPT_Z_VBUCK_SZ_FLAG = SET;
	 break;
    case TERSE_OPT_SWITCH_NONZ:

      /*** value of switchover must be > 2 ***/
      /*** not clear why this is the case  - CHB 12/12/96 ***/
if ( (*value) <= 2 ) {
      printf("xspcnf Error: Parameter value must be > 2.\n");
      printf("              value = %d\n",(*value));
      abort();
    }

	 SWITCH_NZ_VALUE = (*value);
	 OPT_SWITCH_NONZ_FLAG = SET;
	 break;
    default:
	 printf("xspcnf Error: unknown terse_option %d\n",(*terse_opt));
	 abort();
  }
}




/****************************************************************************
 *                                                                          *
 *    ispower2(n)                                                           *
 *                                                                          *
 *    returns 1 (true)     if n is a power of 2;                            *
 *            0 (false     otherwise.                                       *
 *                                                                          *
 ****************************************************************************/

int ispower2(n)
int n;
{
 /*
  * repeatedly divide by 2 until we get a final answer
  * that is less than or equal to 2, or we get a remainder
  * after one of the divisions...
  *
  * this returns 0 for all n < 0 
  */

  while (n > 2)
  {
    int r = n % 2;
    if (r != 0)
      return 1;

    n = n / 2;
  }

  /* n was a power of 2 iff n == 2 now */
  return (n == 2);
}


/****************************************************************************
 *                                                                          *
 *    initialize_pi_global_variables()                                      *
 *                                                                          *
 *    initialize_pi_global_variables initializes some initial global values.*
 *                                                                          *
 *                                                                          *
 ****************************************************************************/

void initialize_pi_global_variables()
{
  int i, j;

  if (OPT_S_BUCK_SZ_FLAG      == NOT_SET) {
    sBUCKET_DIM          = 8;
    sBUCKET_DIM_minus1   = sBUCKET_DIM - 1;
  }
  if (OPT_D_BUCK_SZ_FLAG      == NOT_SET) {
    dBUCKET_DIM          = 8;
    dBUCKET_DIM_minus1   = dBUCKET_DIM - 1;
  }
  if (OPT_C_BUCK_SZ_FLAG      == NOT_SET) {
    cBUCKET_DIM          = 8;
    cBUCKET_DIM_minus1   = cBUCKET_DIM - 1;
  }
  if (OPT_Z_BUCK_SZ_FLAG      == NOT_SET) {
    zBUCKET_DIM          = 8;
    zBUCKET_DIM_minus1   = zBUCKET_DIM - 1;
  }
  if (OPT_S_IBUCK_SZ_FLAG     == NOT_SET) {
    sIBUCKET_DIM         = 8;
    sIBUCKET_DIM_minus1  = sIBUCKET_DIM - 1;
  }
  if (OPT_D_IBUCK_SZ_FLAG     == NOT_SET) {
    dIBUCKET_DIM         = 8;
    dIBUCKET_DIM_minus1  = dIBUCKET_DIM - 1;
  }
  if (OPT_C_IBUCK_SZ_FLAG     == NOT_SET) {
    cIBUCKET_DIM         = 8;
    cIBUCKET_DIM_minus1  = cIBUCKET_DIM - 1;
  }
  if (OPT_Z_IBUCK_SZ_FLAG     == NOT_SET) {
    zIBUCKET_DIM         = 8;
    zIBUCKET_DIM_minus1  = zIBUCKET_DIM - 1;
  }
  if (OPT_S_VBUCK_SZ_FLAG     == NOT_SET) {
    sVBUCKET_DIM         = 32;
    sVBUCKET_DIM_minus1  = sVBUCKET_DIM - 1;
  }
  if (OPT_D_VBUCK_SZ_FLAG     == NOT_SET) {
    dVBUCKET_DIM          = 32;
    dVBUCKET_DIM_minus1   = dVBUCKET_DIM - 1;
  }
  if (OPT_C_VBUCK_SZ_FLAG     == NOT_SET) {
    cVBUCKET_DIM          = 32;
    cVBUCKET_DIM_minus1   = cVBUCKET_DIM - 1;
  }
  if (OPT_Z_VBUCK_SZ_FLAG     == NOT_SET) {
    zVBUCKET_DIM          = 32;
    zVBUCKET_DIM_minus1   = zVBUCKET_DIM - 1;
  }
  if (OPT_G_P_FLAG            == NOT_SET) INITIALIZED_MAXINDEX = FALSE;
  if (OPT_SWITCH_PERCENT_FLAG == NOT_SET) SWITCH_NZ_VALUE      = 16;
  if (OPT_SWITCH_NONZ_FLAG    == NOT_SET) SWITCH_NZ_RATIO      = 0.2;

  toggle_adxpy_1_peep = 1;

  for (i=0;i<5;i++)
    for (j=0;j<5;j++)
      MEM_ALLOCATED[i][j] = 0.0;
  
#ifdef _PEEP
  fileopenflag = 1;
#endif

#ifdef _MEMORY_CHECK
  fileopenflag2 = 1;
  total_mallocs_size = 0;
#endif
}




/****************************************************************************
 *                                                                          *
 *   my_itoa (n, s)                                                         *
 *                                                                          *
 *       n            Integer ( n > 0) to be converted.                     *
 *                                                                          *
 *       s            Ascii string representation of n.                     *
 *                                                                          *
 *       my_itoa converts n to characters in s.  I adopted  the code from   *
 *       the Kernighan-Ritchie C Programming language book.                 *
 *                                                                          *
 ****************************************************************************/

void my_itoa (n, s)
int n;
char *s;
{
  int c, i, j;

  /* generate digits in reverse order and store in s */

  i= 0;
  do {
    s[i++] = n % 10 + '0';
  } while ((n /= 10) > 0);
  s[i] = '\0';

  /* reverse s */

  for (i=0,j=strlen(s)-1; i<j; i++,j--) {
    c    = s[i];
    s[i] = s[j];
    s[j] = c   ;
  }

}



/****************************************************************************
 *                                                                          *
 *    xspmex(usedmem,unit_number)                                           *
 *                                                                          *
 *       usedmem       Array where memory information is stored.            *
 *                                                                          *
 *       unit_number  A pointer to an integer containing information about  *
 *                    where to write output to. If unit_number == 0, stdout *
 *                    is output.  If unit_number == -1, no output is        *
 *                    generated (i.e., info is only stored in usedmem.  If  *
 *                    unit_number is in the range [1,999], output is to a   *
 *                    file whose name has unit_number appended to 'SPMEX.'  *
 *                    For example, if unit_number = 16, then the output     *
 *                    is appended to the file 'SPMEX.16' in the current     *
 *                    directory, i.e., if 'SPMEX.16' did not already exist  *
 *                    in the directory it is created and written to, else   *
 *                    the output is appended to the end of the 'SPMEX.16'.  *
 *                                                                          *
 *    xspmex reports how many bytes have been allocated.                    *
 *                                                                          *
 ****************************************************************************/

void xspmex(usedmem,unit_number)
float *usedmem;
int *unit_number;
{
  int i,j;
  
  FILE *fp;
  char *ext;
  char *filename;
  
  for (i=0;i<5;i++)
    for (j=0;j<5;j++)
      usedmem[5*i+j] = MEM_ALLOCATED[i][j];
  
  if ( (*unit_number) > 0 && (*unit_number) < 1000 ) {
    ext = (char *) malloc(4*sizeof(char));
    filename = (char *) malloc(7*sizeof(char));
    strcpy(filename, "SPMEX.");
    my_itoa ( (*unit_number), ext );
    strcat (filename, ext);
    if ( (fp = fopen (filename,"a")) == NULL ) {
      printf("xspmex error: Couldn't open file.\n");
      abort();
    }
  }
  else if  ( (*unit_number) < -1 ||  (*unit_number) >= 1000 ) {
    printf("xspmex error: Invalid Unit Number = %d  (must be in [-1,999])\n",
	   (*unit_number) );
    abort();
  }

  if ( (*unit_number) != -1 ) {
    if ( (*unit_number) == 0 ) {

      printf("\n\nDynamic Memory Allocation Report (numbers are in KBytes):\n\n");
      printf("                 BUCKETS     IBUCKETS    VBUCKETS    VECTORS         TOTAL\n");
      printf("                 ========    ========    ========    ========       ========\n");
      
      printf("REAL            %9.2e   %9.2e   %9.2e   %9.2e      %9.2e\n",
	     usedmem[0],usedmem[1],usedmem[2],usedmem[3],usedmem[4]);
      printf("DOUBLE          %9.2e   %9.2e   %9.2e   %9.2e      %9.2e\n",
	     usedmem[5],usedmem[6],usedmem[7],usedmem[8],usedmem[9]);
      printf("COMPLEX         %9.2e   %9.2e   %9.2e   %9.2e      %9.2e\n",
	     usedmem[10],usedmem[11],usedmem[12],usedmem[13],usedmem[14]);
      printf("DOUBLE COMPLEX  %9.2e   %9.2e   %9.2e   %9.2e      %9.2e\n\n",
	     usedmem[15],usedmem[16],usedmem[17],usedmem[18],usedmem[19]);
      printf("TOTAL           %9.2e   %9.2e   %9.2e   %9.2e      %9.2e\n\n",
	     usedmem[20],usedmem[21],usedmem[22],usedmem[23],usedmem[24]);
    }
    else {
      fprintf(fp,"\n\nDynamic Memory Allocation Report (numbers are in KBytes):\n\n");
      fprintf(fp,"                 BUCKETS     IBUCKETS    VBUCKETS    VECTORS         TOTAL\n");
      fprintf(fp,"                 ========    ========    ========    ========       ========\n");
      
      fprintf(fp,"REAL            %9.2e   %9.2e   %9.2e   %9.2e      %9.2e\n",
	      usedmem[0],usedmem[1],usedmem[2],usedmem[3],usedmem[4]);
      fprintf(fp,"DOUBLE          %9.2e   %9.2e   %9.2e   %9.2e      %9.2e\n",
	      usedmem[5],usedmem[6],usedmem[7],usedmem[8],usedmem[9]);
      fprintf(fp,"COMPLEX         %9.2e   %9.2e   %9.2e   %9.2e      %9.2e\n",
	      usedmem[10],usedmem[11],usedmem[12],usedmem[13],usedmem[14]);
      fprintf(fp,"DOUBLE COMPLEX  %9.2e   %9.2e   %9.2e   %9.2e      %9.2e\n\n",
	      usedmem[15],usedmem[16],usedmem[17],usedmem[18],usedmem[19]);
      fprintf(fp,"TOTAL           %9.2e   %9.2e   %9.2e   %9.2e      %9.2e\n\n",
	      usedmem[20],usedmem[21],usedmem[22],usedmem[23],usedmem[24]);
      fclose(fp);
    }
  }
}



/****************************************************************************
 *                                                                          *
 *    xspmem(total_dma)                                                     *
 *                                                                          *
 *       total_dma    Total count in Kbytes of Dynamically-allocated Memory *
 *                    allocated by SparsLinC.                               *
 *                                                                          *
 *    xspmem          returns how many Kbytes have been allocated. It's the *
 *                    terse version of xspmex.                              *
 *                                                                          *
 ****************************************************************************/

void xspmem(total_dma)
float *total_dma;
{
  *total_dma = MEM_ALLOCATED[ALL][ALL];
}





/****************************************************************************
 *                                                                          *
 *    memory_count(precision,storage,size);                                 *
 *                                                                          *
 *    keep count of memory allocated.                                       *
 *                                                                          *
 ****************************************************************************/

void memory_count(precision,storage,size)
int precision, storage, size;
{
  float real_size;
  
  real_size = (float) size / 1000.0;

  MEM_ALLOCATED [precision] [storage] += real_size;
  MEM_ALLOCATED [ALL]       [storage] += real_size;
  MEM_ALLOCATED [precision] [ALL]     += real_size;
  MEM_ALLOCATED [ALL]       [ALL]     += real_size;
}
