/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#include  <stdlib.h>
#include "gaxpy_pi_macros_and_data_structures.h"
#include "gaxpy_pi_global_variables.h"

#ifdef _PEEP


/****************************************************************************
 *                                                                          *
 *  adpeep_header(arity)                                                    *
 *                                                                          *
 *       adpeep_header prints the adpeep header for a right-hand side.      *
 *       Printing is in the following format:                               *
 *                                                                          *
 *       ****************  adxpy<n> ****************                        *
 *                                                                          *
 *       arity == 0 indicates the caller is adcopy.  We print as follows:   *
 *                                                                          *
 *       ****************  adcopy   ****************                        *
 *                                                                          *
 ****************************************************************************/

void adpeep_header (arity)
int arity;
{
  if (fileopenflag) {
    fp = fopen("gaxpy_adxpy.dump","w");     
    fileopenflag = 0;
  }

#ifdef _VERBOSE_DUMP
    
  if (arity == 0) {
    fprintf(fp,"\n****************  adcopy  ****************\n");
#ifdef _SCREEN_PEEP
    printf("\n****************  adcopy  ****************\n");
#endif
  }

  else if (arity > 0 && arity <= MAXARITY) {
    fprintf(fp,"\n****************  adxpy%d  ****************\n",arity);
#ifdef _SCREEN_PEEP
    printf("\n****************  adxpy%d  ****************\n",arity);
#endif
  }
  
  else {
    printf("adpeep_header error: illegal arity = %d\n",arity);
    abort();
  }

#endif

}


#endif






#ifdef _MEMORY_CHECK

/****************************************************************************
 *                                                                          *
 *  mymalloc(malloc_size)                                                   *
 *                                                                          *
 *       mymalloc is the same as malloc, except it keeps a running count of *
 *       how much dynamic memory has been allocated.                        *
 *                                                                          *
 *       malloc_size  :  size of dynamic memory block to be allocated.      *
 *                                                                          *
 ****************************************************************************/

void *mymalloc(malloc_size)
int malloc_size;
{
  if (fileopenflag2) {
    fp2 = fopen("gaxpy_memory_check","w");     
    fileopenflag2 = 0;
  }

  total_mallocs_size += malloc_size;
  fprintf(fp2,"\tmalloc_size = %d      total_malloc_size =  %d\n",
	  malloc_size,total_mallocs_size);
  return((void *)calloc(1,malloc_size));      /* PMK: 12/4/96 edit:         *
                                               *                            *
                                               * malloc changed to calloc   *
                                               * to ensure zeroed out       *
                                               * dynamic memory.            */
}


#endif


/****************************************************************************
 *                                                                          *
 *  debugging_code_2(unconsumedcurrentnonzerorange)                         *
 *                                                                          *
 *       debugging_code_2 prints error messages if a quantity expected      *
 *       to be zero at some point of the computation is not zero;           *
 *       it then aborts.                                                    *
 *                                                                          *
 *       unconsumedcurrentnonzerorange  : nsia                              *
 *                                                                          *
 ****************************************************************************/

void debugging_code_2(unconsumedcurrentnonzerorange)
int unconsumedcurrentnonzerorange;
{
  if (unconsumedcurrentnonzerorange != 0) {
    printf("debugging_code_2 error: Quantity should be zero\n\t");
    printf("unconsumedcurrentnonzerorange=%d\n",
	   unconsumedcurrentnonzerorange);
    abort();
  }
}

