/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#ifndef GAXPY_PI_MACROS_AND_DATA_STRUCTURES

#define GAXPY_PI_MACROS_AND_DATA_STRUCTURES

#include <stdio.h>

#ifndef _WIN32
#include <values.h>
#endif

#include "options.h"
#include "gaxpy_pi_axpy_auto_gen.h"
#include "multitype.h"
#include "renaming.h"

/****************************************************************************
 *                                                                          *
 *  Macro Declarations                                                      *
 *                                                                          *
 ****************************************************************************/


#define REPLACE 0                  /* Operation label used in heapop.       */

#define DELETE 1                   /* Operation label used in heapop.       */

#define S0 0                       /* As-yet Undefined data structure label.*/
#define S1 1                       /* One-Subscript data structure label.   */
#define SS 2                       /* Single-Subscript data structure       *
				    * label.                                */
#define CS 3                       /* Compressed-Subscript data structure   *
				    * label.                                */
#define FALSE 0
#define TRUE 1

#define COMPLETELY_NEW_RANGE 1
#define NEW_ENTRIES_PADDED_TO_OLD_RANGE 2

#define ORIG_OTHER_VECTOR 0
#define TEMP_OTHER_VECTOR 1


#define OPT_S_BUCK_SZ       1      /* These OPT macros are used in xspcnx  */
#define OPT_D_BUCK_SZ       2
#define OPT_C_BUCK_SZ       3
#define OPT_Z_BUCK_SZ       4
#define OPT_ALL_BUCK_SZ     5
#define OPT_S_IBUCK_SZ      6
#define OPT_D_IBUCK_SZ      7
#define OPT_C_IBUCK_SZ      8
#define OPT_Z_IBUCK_SZ      9
#define OPT_ALL_IBUCK_SZ   10
#define OPT_S_VBUCK_SZ     11
#define OPT_D_VBUCK_SZ     12
#define OPT_C_VBUCK_SZ     13
#define OPT_Z_VBUCK_SZ     14
#define OPT_ALL_VBUCK_SZ   15
#define OPT_G_P            16
#define OPT_SWITCH_PERCENT 17
#define OPT_SWITCH_NONZ    18

#define TERSE_OPT_ALL_BUCK_SZ   1  /* These OPT macros are used in xspcnf  */
#define TERSE_OPT_ALL_VBUCK_SZ  2
#define TERSE_OPT_SWITCH_NONZ   3


#define S_EVAL 0
#define D_EVAL 1
#define C_EVAL 2
#define Z_EVAL 3

#define MEM_BUCKET  0
#define MEM_IBUCKET 1
#define MEM_VBUCKET 2
#define MEM_VECTOR  3

#define ALL 4

#ifdef  _WIN32
#define HBIT    ((unsigned)1 << 32 - 1)
#define MAXINT  ((int)(~HBIT))
#endif

#define DUMMYSUBSCRIPT MAXINT      /* == largest Integer from values.h.     *
                                    * Used as argument                      *
                                    * when calling heapop(DELETE,...) to    *
				    * ensure root node "sinks".             */

#define DUMMYVECTOR MAXARITY       /* Larger than a valid vectorindex. Used *
				    * as a dummy argument when calling      *
				    * heapop(DELETE,...).                   */

/*
 *  Obsolete Constants: now treated as variables that are precision-dependent
 *

#define BUCKETDIMENSION 10         ** Max. # of (index,value) pairs in an   *
				    * SS bucket.                            **

#define VBUCKETDIMENSION 10        ** Max. # of values in a                 *
                                    * CS vbucket.                           **

#define IBUCKETDIMENSION 5         ** Max. # of                             *
				    * (index_range_begin,index_range_end)   *
				    * pairs in a                            *
                                    * CS ibucket.                           **

#define BUCKETSTOREDIMENSION 500
#define VBUCKETSTOREDIMENSION 500
#define IBUCKETSTOREDIMENSION 500
#define VECTORSTOREDIMENSION 500

*
*  End of Obsolete Constants
*/


#ifndef min
#define min(A,B) ((A)<(B)?(A):(B))
#endif

#ifndef max
#define max(A,B) ((A)>(B)?(A):(B))
#endif

#define REAL_PART 0
#define IMAG_PART 1

#define GAXPY_EQUALS(y,a,b)      X_MULT(y,a,b)
#define GAXPY_PLUS_EQUALS(y,a,b) X_PLUS_MULT(y,a,b)


#ifdef _MOD_MACRO
#define MY_MOD(a,b_minus1) (a) & (b_minus1)
#else
#define MY_MOD(a,b_minus1) (a) % ((b_minus1)+1)
#endif


#define OLD_RANGE 1
#define NEW_RANGE 2
#define CONTINUING_RANGE 3

#define RIGHT "rhs"
#define LEFT  "lhs"

/*
 *  Macro definitions for subroutines;  DEBUG ON
 */

#ifdef _DEBUG

#define DEBUGGING_CODE_2(a)             debugging_code_2(a)

/*
 *  Macro definitions for subroutines;  DEBUG OFF
 */

#else

#define DEBUGGING_CODE_2(a)

#endif


#ifdef _MEMORY_CHECK
#define palloc mymalloc
#else
#define palloc(x) calloc(1,x)                 /* PMK: 12/4/96 edit:         *
                                               *                            *
                                               * malloc changed to calloc   *
                                               * to ensure zeroed out       *
                                               * dynamic memory.            */
#endif


typedef struct _heapstruct {
  int vectorindex;                            /* Order of vector in the rhs *
                                               * of the saxpy.              */
  int rangebeginindex;                        /* Array index of the nonzero *
                                               * in the non-sparse vector.  */
  int rangeendindex;

  struct _heapstruct *left;                   /* Address of left child.     */
  struct _heapstruct *right;                  /* Address of right child.    */
} heapstruct;

/*     ____________________________heapstruct___________________________    *
 *     |                                                               |    *
 *     |   ___vectorindex___   _rangebeginindex_   ___rangeendindex__  |    *
 *     |   |               |   |               |   |                |  |    *
 *     |   |     (int)     |   |     (int)     |   |     (int)      |  |    *
 *     |   |_______________|   |_______________|   |________________|  |    *
 *     |                                                               |    *
 *     |                left *             right *                     |    *
 *     |                    /                     \                    |    *
 *     |___________________/_______________________\___________________|    *
 *                        /                         \                       *
 *                      |/_                         _\|                     *
 *                (heapstruct)                   (heapstruct)               */




/****************************************************************************
 *                                                                          *
 *  Function Prototyping                                                    *
 *                                                                          *
 ****************************************************************************/

#define _P(x) ()

void allocateheapstore _P(());
void heapinitialize _P((int vectorindex,
			int rangebeginindex,
			int rangeendindex));
void heapinsert _P((int vectorindex,
		    int rangebeginindex,
		    int rangeendindex));
void heapop _P((int operationlabel,
		int vectorindex,
		int rangebeginindex,
		int rangeendindex));

#ifdef _PEEP
void adpeep_header _P((int arity));
#endif

void debugging_code_2 _P((int unconsumedcurrentnonzerorange));

#ifdef _MEMORY_CHECK
void *mymalloc _P((int malloc_size));
#endif

void xspcnx _P((int *opt, int* value));

void xspcnf _P((int *terse_opt, int* value));

int ispower2 _P((int n));

int intpower _P ((int base, int n));

double log2 _P ((double x));

void xspini _P(());

void xspfra _P(());

void initialize_pi_global_variables _P(());

void my_itoa _P((int n, char *s));

void xspmex _P((float *usedmem, int *unit_number));

void xspmem _P((float *total_dma));

void memory_count _P((int precision, int storage, int size));


/*
 * Vitaly's data structure additions (pu_struct, u_struct) and
 *                         modification (vectorstruct)
 * for the visulaization functionality.
 */

#ifdef _VTRACE

typedef struct _u_struct {
   int adxpy_time;     /* Time of the update (measured in adxpy calls) */
   long real_time;     /* Time of the update (measured as CPU time) */
   int nz;             /* Number of non-zeroes after the update */
   int arity;          /* Arity of the r.h.s. */
   int effect_arity;   /* Effective arity of the r.h.s. */
   int dstype;         /* Data structure type */
   struct _u_struct *next;
} u_struct;

typedef u_struct *pu_struct;

#endif



#endif  /*     #ifndef GAXPY_PI_MACROS_AND_DATA_STRUCTURES     */

