#define SINGLE
/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#include  <stdlib.h>
#include "sgaxpy_pd_macros_and_data_structures.h"
#include "sgaxpy_pd_global_variables.h"

/****************************************************************************
 *                                                                          *
 *     sspg1 performs the arity 1 gaxpy.                                   *
 *                                                                          *
 ****************************************************************************/

void sspg1(destptrptr,alphaptr,vectoraptrptr)
svectorstruct **vectoraptrptr,**destptrptr;
float *alphaptr;
{

  int arity = 1;

  svectorstruct *rhssvectorptr[1];
  svectorstruct *tempvectorptr;
  
  sbucketstruct *rhssbucketptr;
  sbucketstruct *lhssvectortailbucketptr;
  sbucketstruct *lhssbucketptr;
  sbucketstruct *tempbucketptr;
  
  int i,k;
  int bucketindex;
  
  svbucketstruct *rhssvectorvbucketptr;
  svbucketstruct *lhssvectortailvbucketptr;
  svbucketstruct *tempvbucketptr;
  int vbucketindex;
  
  sibucketstruct *rhssvectoribucketptr;
  sibucketstruct *lhssvectortailibucketptr;
  sibucketstruct *tempibucketptr;
  sibucketstruct *lhssibucketptr;
  int ibucketindex;

  float *alphaptrs[1];
  
#ifdef _TRACE
  int nnzb,nnza;
  svectorstruct *lhssvectoraddress;
  
  
  svectorstruct **vectorptrptrs[1];
  
  vectorptrptrs[0] = vectoraptrptr;
  
  if(*destptrptr==NULL){ 
    nnzb = 0; 
    lhssvectoraddress= (svectorstruct *) 99900; 
  }
  else {
    nnzb = (*destptrptr)->nonzeroes; 
    lhssvectoraddress = *destptrptr;
  }
#endif
  
  
  /****************************************
   * Initialization of *destptrptr.       *
   ****************************************/
  
  if (*destptrptr == NULL)
    sgetvector(destptrptr);
  
  rhssvectorptr[0] = *vectoraptrptr;
  alphaptrs[0] = alphaptr;

#ifdef _PEEP
  if (toggle_adxpy_1_peep) {
    adpeep_header(1);
    sadpeep_pre(0,vectoraptrptr,alphaptr);
  }
#endif
  
  if (rhssvectorptr[0]->nonzeroes == 0 || X_IS_ZERO(*alphaptr)) {
    sspzronotca(destptrptr);
  }
  else /* if not ZERO AXPY */ {
    if (rhssvectorptr[0] == *destptrptr &&
	rhssvectorptr[0]->dstype == CS) {
      plusequalsspg1(destptrptr,alphaptr);
    }
    else /* if (not plusequalsspg1 or (dstype is S1 or SS)) */ {
      lhssvectorptr->nonzeroes = rhssvectorptr[0]->nonzeroes;
      lhssvectorptr->dstype = rhssvectorptr[0]->dstype;

      if (lhssvectorptr->dstype == S1) {
	
	/********************************************************************
	 *                       ONE_SUBSCRIPT AXPY                         *
	 ********************************************************************/


	lhssvectorptr->sds.s1.index =
	                rhssvectorptr[0]->sds.s1.index;
	GAXPY_EQUALS(lhssvectorptr->sds.s1.value,*alphaptr,
		        rhssvectorptr[0]->sds.s1.value);
      }
      else if (lhssvectorptr->dstype == SS) {
	
	/******************************************************************
	 *                 _SINGLE_SUBSCRIPTS AXPY                        *
	 ******************************************************************/
	
	rhssbucketptr = rhssvectorptr[0]->sds.ss.bucketptr;
	for (i=0;i<rhssvectorptr[0]->nonzeroes;i++) {
	  if ((bucketindex = MY_MOD(i,sBUCKET_DIM_minus1)) == 0) {
	    if (i == 0) {
	      sgetbucket(&lhssbucketptr);
	      lhssvectorptr->sds.ss.bucketptr = lhssbucketptr;
	    }
	    else {
	      rhssbucketptr = rhssbucketptr->nextbucket;
	      sgetbucket(&lhssbucketptr);
	      lhssvectortailbucketptr->nextbucket = lhssbucketptr;
	    }
	    lhssvectortailbucketptr = lhssbucketptr;
	  }
	  GAXPY_EQUALS(lhssvectortailbucketptr->value[bucketindex],
		       *alphaptr,rhssbucketptr->value[bucketindex]);
	  lhssvectortailbucketptr->index[bucketindex] =
	    rhssbucketptr->index[bucketindex];
	}
      }
      
      else /* if (lhssvectorptr->dstype == CS) */ {
	  
	/******************************************************************
	 *                  _COMPRESSED_SUBSCRIPTS AXPY                   *
	 ******************************************************************/
	
	lhssvectorptr->sds.cs.nonzeroranges =
	  rhssvectorptr[0]->sds.cs.nonzeroranges;
	rhssvectoribucketptr =
	  rhssvectorptr[0]->sds.cs.ibucketptr;
	rhssvectorvbucketptr =
	  rhssvectorptr[0]->sds.cs.vbucketptr;
	
	for (i=0;i<rhssvectorptr[0]->sds.cs.nonzeroranges;i++) {
	  if ((ibucketindex = MY_MOD(i,sIBUCKET_DIM_minus1)) == 0) {
	    if (i == 0) {
	      sgetibucket(&lhssibucketptr);
	      lhssvectorptr->sds.cs.ibucketptr = lhssibucketptr;
	    }
	    else {
	      rhssvectoribucketptr =
		rhssvectoribucketptr->nextibucket;
	      sgetibucket(&lhssibucketptr);
	      lhssvectortailibucketptr->nextibucket = lhssibucketptr;
	    }
	    lhssvectortailibucketptr = lhssibucketptr;
	  }
	  
	  lhssvectortailibucketptr->index_RANGE_BEGIN[ibucketindex] =
	    rhssvectoribucketptr->index_RANGE_BEGIN[ibucketindex];
	  lhssvectortailibucketptr->index_RANGE_END[ibucketindex] =
	    rhssvectoribucketptr->index_RANGE_END[ibucketindex];
	}
	
	for (i=0;i<rhssvectorptr[0]->nonzeroes;i++) {
	  if ((vbucketindex = MY_MOD(i,sVBUCKET_DIM_minus1)) == 0) {
	    if (i == 0) {
	      sgetvbucket(&lhssvbucketptr);
	      lhssvectorptr->sds.cs.vbucketptr = lhssvbucketptr;
	    }
	    else {
	      rhssvectorvbucketptr =
		rhssvectorvbucketptr->nextvbucket;
	      sgetvbucket(&lhssvbucketptr);
	      lhssvectortailvbucketptr->nextvbucket = lhssvbucketptr;
	    }
	    lhssvectortailvbucketptr = lhssvbucketptr;
	  }
	  GAXPY_EQUALS(lhssvectortailvbucketptr->value[vbucketindex],
		       *alphaptr,
		       rhssvectorvbucketptr->value[vbucketindex]);
	}
      } /* matches     else ** if (lhssvectorptr->dstype == CS) ** {  */
    
      saxpy_0_1_post_processing(destptrptr);

    } /* matches   else ** if (not plusequalsspg1) ** {                   */
  } /* matches   else ** if not ZERO AXPY ** {                             */
  

#ifdef _PEEP
  if (toggle_adxpy_1_peep) {
    sadpeep_post(destptrptr);
  }
#endif


#ifdef _TRACE 
  nnza = (*destptrptr)->nonzeroes; 
  sadinfo(arity,nnzb,lhssvectoraddress,nnza,alphaptrs,vectorptrptrs);
#endif
  
}
  
