#define SINGLE
/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#include  <stdlib.h>
#include "sgaxpy_pd_macros_and_data_structures.h"
#include "sgaxpy_pd_global_variables.h"

/****************************************************************************
 *                                                                          *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/

void sspg2(destptrptr,alphaptr,vectoraptrptr,betaptr,vectorbptrptr)
svectorstruct **destptrptr;
float *alphaptr;
svectorstruct **vectoraptrptr;
float *betaptr;
svectorstruct **vectorbptrptr;
{
  
  int arity = 2;
  
  svectorstruct *rhssvectorptr[2];
  
  sbucketstruct *rhssbucketptr[2];
  
  sbucketstruct *lhssbucketptr;
  
  sibucketstruct *rhssibucketptr[2];
                                      /* Array of pointers to "current"     *
				       * buckets where "current" nonzeroes  *
				       * of the rhs_vectors reside.         */

  sibucketstruct *lhssibucketptr;  /* Pointer to current bucket of       *
				       * lhssvector.                         */

  float alpha[2];          /* Dereferenced alphaptrs[].          */
  float *alphaptrs[2];

  int rhssbucketindex[2];              /* Array of indices of current   *
				       * nonzeroes in the current buckets   *
				       * of rhs_vectors.                    */

  int lhssbucketindex;                 /* bucketindex of next nonzero to be  *
				       * added to lhssvector.                */

  int lastlhssbucketindex;             /* bucketindex of the latest added    *
				       * nonzero in the lhssvector.  We need *
				       * this additional index to leap-frog *
				       * buckets, if need be.               */


  int rhssibucketindex[2];             /* Array of indices of current        *
				       * nonzeroes in the current buckets   *
				       * of rhs_vectors.                    */

  int lhssibucketindex;                /* bucketindex of next nonzero to be  *
				       * added to lhssvector.                */

  int lastlhssibucketindex;            /* bucketindex of the latest added    *
				       * nonzero in the lhssvector.  We need *
				       * this additional index to leap-frog *
				       * buckets, if need be.               */

  int k;                              /* Counter for arity.                 */

  int j;                              /* Index for vectorsubscripts to be   *
				       * entered in (and later popped from) *
				       * the heap.                          */

  int rhssvectornonzeroes[2];          /* Array of total # of nonzeroes in   *
				       * each rhs_vector.                   */

  int rhssvectorconsumednonzeroes[2];  /* Array of # of nonzeroes consumed   *
				       * so far in each rhs_vector.         */

  int lhssvectornonzeroes;             /* Count of # of nonzeroes entered    *
				       * into the lhssvector so far.  By the *
				       * end of the saxpy operation, this   *
				       * index equals the total # of        *
				       * nonzeroes in the resultant vector. */

/*
 * These variables were originally for COMPRESSED_SUBSCRIPTS
 */

  svbucketstruct *lastlhssvbucketptr; 
                                      /* Pointer to current vbucket of      *
				       * lhssvector.                    */

  int rhssvectornonzeroranges[2];
                                      /* Array of total # of nonzeroes      *
				       * ranges in each rhs_vector.         */

  int rhssvectorconsumednonzeroranges[2];
                                      /* Array of # of nonzeroes consumed   *
				       * so far in each rhs_vector.         */

  int rhssvbucketindex[2];             /* Array of indices of current        *
				       * nonzeroes in the current buckets   *
				       * of rhs_vectors.                    */

  int lhssvectornonzeroranges;         /* Count of # of nonzeroes entered    *
				       * into the lhssvector so far.  By the *
				       * end of the saxpy operation, this   *
				       * index equals the total # of        *
				       * nonzeroes in the resultant vector. */

  int lhssvbucketindex;                /* bucketindex of next nonzero to be  *
				       * added to lhssvector.                */


  int lastlhssvbucketindex;            /* bucketindex of the latest added    *
				       * nonzero in the lhssvector.  We need *
				       * this additional index to leap-frog *
				       * buckets, if need be.               */

  svbucketstruct *templhssvbucketptr;
  int templhssvbucketindex;

  int j1;                             /* Index for vectorsubscripts to be   *
				       * entered in (and later popped from) *
				       * the heap.                          */

  int j2;                             /* Index for vectorsubscripts to be   *
				       * entered in (and later popped from) *
				       * the heap.                          */

  int h1;
  int h2;

  int totalcurrentnonzerorange;

#ifdef _DEBUG
  int unconsumedcurrentnonzerorange;
#endif

  int distancetooverlap;
  int distancetoendofvbucket;

  float tmp_add,tmp_mult1,tmp_mult2;


  /*
   * These variables were added for union ds
   */
  
  svectorstruct *restorerhssvectorptr[MAXARITY+1];

  int governingdstype;
  
  /*
   * These variables were added for plusequal code
   */
  
  svectorstruct **rhssvectorptrptrs[MAXARITY];
  
  int condition[2];
  
#ifdef _PLUSEQUAL  
  int smallest_rhs_index; 
#endif /* _PLUSEQUAL */
  
  
#ifdef _TRACE
  int nnzb,nnza;
  svectorstruct *lhssvectoraddress;
  
  svectorstruct **vectorptrptrs[2];
  
  vectorptrptrs[0] = vectoraptrptr;
  
  vectorptrptrs[1] = vectorbptrptr;
  
  if(*destptrptr==NULL){ 
    nnzb = 0; 
    lhssvectoraddress= (svectorstruct *) 99900; 
  }
  else {
    nnzb = (*destptrptr)->nonzeroes; 
    lhssvectoraddress = *destptrptr;
  }
#endif
  
  
  /****************************************
   * Initialization of *destptrptr.       *
   ****************************************/
  
  if (*destptrptr == NULL)
    sgetvector(destptrptr);
  
  /****************************************
   * Dereferencing.                       *
   ****************************************/
  
  alpha[0] = *alphaptr;
  rhssvectorptr[0] = *vectoraptrptr;
  alphaptrs[0] = alphaptr;
  
#ifdef _PEEP
  adpeep_header(2);
  sadpeep_pre(0,vectoraptrptr,alphaptr);
#endif    

  alpha[1] = *betaptr;
  rhssvectorptr[1] = *vectorbptrptr;
  alphaptrs[1] = betaptr;

#ifdef _PEEP
  sadpeep_pre(1,vectorbptrptr,betaptr);
#endif    
 
  for (k=0;k<arity;k++) {
    restorerhssvectorptr[k] = rhssvectorptr[k];
    rhssvectornonzeroes[k] = rhssvectorptr[k]->nonzeroes;
    if (rhssvectorptr[k]->dstype == CS)
      rhssvectornonzeroranges[k] =
	rhssvectorptr[k]->sds.cs.nonzeroranges;
  }
  
#ifdef _PLUSEQUAL
  smallest_rhs_index = DUMMYSUBSCRIPT;
  for (k=0;k<arity;k++) {
    if (rhssvectorptr[k]->dstype == S1) 
      smallest_rhs_index = min (smallest_rhs_index,
				rhssvectorptr[k]->sds.s1.index);
    else if (rhssvectorptr[k]->dstype == SS)
      smallest_rhs_index =
	min (smallest_rhs_index,
	     rhssvectorptr[k]->sds.ss.bucketptr->index[0]);
    else /* if (rhssvectorptr[k]->dstype == CS) */
      smallest_rhs_index =
	min (smallest_rhs_index,
	     rhssvectorptr[k]->sds.cs.ibucketptr->index_RANGE_BEGIN[0]);
  }
#endif /* _PLUSEQUAL */


  if ((rhssvectorptr[0]->nonzeroes == 0 || X_IS_ZERO(alpha[0])) &&
      (rhssvectorptr[1]->nonzeroes == 0 || X_IS_ZERO(alpha[1]))) {
    sspzronotca(destptrptr);
  }
  else if (rhssvectorptr[0]->nonzeroes == 0 || X_IS_ZERO(alpha[0])) {
    if (rhssvectorptr[1] == *destptrptr &&
	rhssvectorptr[1]->dstype == CS) {
      plusequalsspg1(destptrptr,betaptr);
    }
    /* if (rhssvectorptr[1] != *destptrptr ||
       rhssvectorptr[1]->dstype != CS)           */
    else {
#ifdef _PEEP
      toggle_adxpy_1_peep = 0;
      sspg1(destptrptr,betaptr,vectorbptrptr);
      toggle_adxpy_1_peep = 1;
#else
      sspg1(destptrptr,betaptr,vectorbptrptr);
#endif     
    }
  }
  else if (rhssvectorptr[1]->nonzeroes == 0 || X_IS_ZERO(alpha[1])) {
    if (rhssvectorptr[0] == *destptrptr &&
	rhssvectorptr[0]->dstype == CS) {
      plusequalsspg1(destptrptr,alphaptr);
    }
    /* else if (rhssvectorptr[0] != *destptrptr) ||
       rhssvectorptr[0]->dstype != CS)          */
    else {      
#ifdef _PEEP
      toggle_adxpy_1_peep = 0;
      sspg1(destptrptr,alphaptr,vectoraptrptr);
      toggle_adxpy_1_peep = 1;
#else
      sspg1(destptrptr,alphaptr,vectoraptrptr);
#endif
    }
  }
  else /* neither rhs_operation is a NO-OP */ {
    if (rhssvectorptr[0] == *destptrptr &&
	rhssvectorptr[1] == *destptrptr &&
	rhssvectorptr[0]->dstype == CS) {
#ifdef _DEBUG
      printf("sspg2 warning: lhssvector appears more than ");
      printf("once on rhs.\n");
#endif
      X_ADD(tmp_add,*alphaptr,*betaptr);
      plusequalsspg1(destptrptr,&tmp_add);
    }

#ifdef _PLUSEQUAL
    else if
      (rhssvectorptr[0] == *destptrptr &&
       rhssvectorptr[0]->dstype == CS &&
       rhssvectorptr[0]->sds.cs.ibucketptr->index_RANGE_BEGIN[0] <=
       smallest_rhs_index &&
       X_IS_ONE(alpha[0])) {
	rhssvectorptrptrs[1] = &rhssvectorptr[1];
	if ((*rhssvectorptrptrs[1])->dstype != CS &&
	    (*rhssvectorptrptrs[1])->nonzeroes !=0) {
	  sconvertdstype(rhssvectorptrptrs[1],1,CS);
	}
	plusequalsspg2(destptrptr,
			betaptr,rhssvectorptrptrs[1],ORIG_OTHER_VECTOR);
      }
    else if
      (rhssvectorptr[1] == *destptrptr &&
       rhssvectorptr[1]->dstype == CS &&
       rhssvectorptr[1]->sds.cs.ibucketptr->index_RANGE_BEGIN[0] <=
       smallest_rhs_index &&
       X_IS_ONE(alpha[1])) {
	rhssvectorptrptrs[0] = &rhssvectorptr[0];
	if ((*rhssvectorptrptrs[0])->dstype != CS &&
	    (*rhssvectorptrptrs[0])->nonzeroes !=0) {
	  sconvertdstype(rhssvectorptrptrs[0],0,CS);
	}
	plusequalsspg2(destptrptr,
			alphaptr,rhssvectorptrptrs[0],ORIG_OTHER_VECTOR);
	
      }
    
    /* 
     * else if neither rhs_operation is a NO-OP, and neither
     * rhssvector is (the same as the lhssvector AND has
     * multiplier equal to 1 AND is CS)
     */
    
    else if (rhssvectorptr[0] ==  rhssvectorptr[1]) {
#ifdef _DEBUG
      printf("sspg2 warning: lhssvector appears more than ");
      printf("once on rhs.\n");
#endif
      X_ADD(tmp_add,*alphaptr,*betaptr);
#ifdef _PEEP
      toggle_adxpy_1_peep = 0;
      sspg1(destptrptr,&tmp_add,vectoraptrptr);
      toggle_adxpy_1_peep = 1;
#else
      sspg1(destptrptr,&tmp_add,vectoraptrptr);
#endif
    }
#endif /* _PLUSEQUAL */
    
    else /* do an honest to goodness sspg2 */ {
      
      governingdstype = max(rhssvectorptr[0]->dstype,
			    rhssvectorptr[1]->dstype);
      if (governingdstype == S1 &&
	  rhssvectorptr[0]->sds.s1.index != rhssvectorptr[1]->sds.s1.index)
	governingdstype = SS;
      for (k=0;k<arity;k++)
	if (rhssvectorptr[k]->dstype != governingdstype) {
	  sconvertdstype((&rhssvectorptr[k]),k,governingdstype);
	  if (governingdstype == CS)
	    rhssvectornonzeroranges[k] =
	      rhssvectorptr[k]->sds.cs.nonzeroranges;
	}
      
      if (governingdstype == S1) {
	
	/******************************************************************
	 *                      ONE_SUBSCRIPT AXPY                        *
	 ******************************************************************/

	lhssvectorptr->nonzeroes =  1;
	lhssvectorptr->dstype = governingdstype;
	lhssvectorptr->sds.s1.index =
	  rhssvectorptr[0]->sds.s1.index;
        X_MULT(tmp_mult1,*alphaptr,rhssvectorptr[0]->sds.s1.value);
        X_MULT(tmp_mult2,*betaptr,rhssvectorptr[1]->sds.s1.value);
        X_ADD(lhssvectorptr->sds.s1.value,tmp_mult1,tmp_mult2);

	saxpy_0_1_post_processing(destptrptr);

      }
      else /* if we are going to use lhssvectorptr */ {
	if (governingdstype == SS) {
	  
	  /****************************************************************
	   *                 _SINGLE_SUBSCRIPTS AXPY                      *
	   ****************************************************************/
	  for (k=0;k<arity;k++) {
	    rhssvectorconsumednonzeroes[k]= 0;
	    rhssbucketindex[k] = 0;
	    rhssbucketptr[k] = rhssvectorptr[k]->sds.ss.bucketptr;
	  }
	  sgetbucket(&lhssbucketptr);
	  lhssvectorptr->dstype = governingdstype;
	  lhssvectorptr->sds.ss.bucketptr = lhssbucketptr;
	  lhssvectornonzeroes = 0;
	  
	  for (k=0;k<arity;k++)
	    condition[k] = rhssvectorconsumednonzeroes[k] <
	      rhssvectorptr[k]->nonzeroes;
	  
	  while (condition[0] || condition[1]) {
	    if (condition[0] && condition[1])
	      if (rhssbucketptr[0]->index[rhssbucketindex[0]] <=
		  rhssbucketptr[1]->index[rhssbucketindex[1]])
		k = 0;
	      else
		k = 1;
	    else if (condition[0])
	      k = 0;
	    else
	      k = 1;
	    
	    j = rhssbucketptr[k]->index[rhssbucketindex[k]];
	    
	    lhssbucketindex = MY_MOD(lhssvectornonzeroes,
					 sBUCKET_DIM_minus1);
	    
	    /****************************************
	     * Perform saxpy:                       *
	     * 3.1. Found 1st or new subscript for  *
	     *      lhssvector.  If needed, get new  *
	     *      bucket for lhssvector.           *
	     *      Consume nonzero in saxpy.       *
	     ****************************************/
	    if (lhssvectornonzeroes == 0 ||
		lhssbucketptr->index[lastlhssbucketindex] < j) {
	      if (lhssbucketindex == 0 && lhssvectornonzeroes != 0) {
		sgetbucket(&(lhssbucketptr->nextbucket));
		lhssbucketptr = lhssbucketptr->nextbucket;
	      }
	      lhssbucketptr->index[lhssbucketindex] = j;
	      GAXPY_EQUALS(
                lhssbucketptr->value[lhssbucketindex],
		alpha[k],
		rhssbucketptr[k]->value[rhssbucketindex[k]]);
	      lastlhssbucketindex = lhssbucketindex;
	      lhssvectornonzeroes++;
	    }
	    
	    /****************************************
	     * Perform saxpy:                       *
	     * 3.2. Found old subscript for         *
	     *      lhssvector.                      *
	     *      Consume nonzero in saxpy.       *
	     ****************************************/
	    else if (lhssbucketptr->index[lastlhssbucketindex] == j) {
	      GAXPY_PLUS_EQUALS(
                lhssbucketptr->value[lastlhssbucketindex],
		alpha[k],
		rhssbucketptr[k]->value[rhssbucketindex[k]]);
	    }
	    
	    /****************************************
	     * Perform saxpy:                       *
	     * 3.3. Found wrong subscript for       *
	     *      lhssvector.  Abort.               *
	     ****************************************/
	    
#ifdef _DEBUG
	    else {
	      printf("sspg2 error 6: lhssnz=%d, lastlhssindex[]=%d, j=%d\n",
		     lhssvectornonzeroes,
		     lhssbucketptr->index[lastlhssbucketindex],j);
	      abort();
	    }
#endif
	    
	    /****************************************
	     * Perform saxpy:                       *
	     * 3.4. If there is one, get new        *
	     *      nonzero from the rhs_vector     *
	     *      whose nonzero was just          *
	     *      consumed and replace the        *
	     *      heaproot with new entry;        *
	     *      Else, delete heaproot.          *
	     ****************************************/
	    rhssvectorconsumednonzeroes[k]++;
	    condition[k] = rhssvectorconsumednonzeroes[k] <
	      rhssvectorptr[k]->nonzeroes;
	    rhssbucketindex[k] = MY_MOD(rhssvectorconsumednonzeroes[k],
					    sBUCKET_DIM_minus1);
	    if (condition[k]) {
	      if (rhssbucketindex[k]  == 0) {
		if ((rhssbucketptr[k] = rhssbucketptr[k]->nextbucket) 
		    == NULL) {
#ifdef _DEBUG
		  printf("sspg2 err: unexpected NULL nextbucket: k=%d",k);
		  printf("rhssvectorconsumednonzeros=%d,rhssvecnonzeroes=%d,",
			 rhssvectorconsumednonzeroes[k],
			 rhssvectornonzeroes[k]);
		  abort();
#endif
		}	  
	      }
	    }
	  } /* matches     while loop     */
	} /* matches     else if (governingdstype == SS) {      */
	
	else /* if (governingdstype == CS) */ {
	  
	  /****************************************************************
	   *               _COMPRESSED_SUBSCRIPTS AXPY                    *
	   ****************************************************************/
	  
	  /****************************************
	   * Else perform saxpy:                  *
	   * 1. Put in heap the 1st nonzero of    *
	   *    each vector.                      *
	   ****************************************/
	  for (k=0;k<arity;k++) {
	    rhssvectorconsumednonzeroranges[k] = 0;
	    rhssibucketindex[k] = 0;
	    rhssvbucketindex[k] = 0;
	    rhssibucketptr[k] =
	      rhssvectorptr[k]->sds.cs.ibucketptr;
	    rhssvbucketptr[k] =
	      rhssvectorptr[k]->sds.cs.vbucketptr;
	    
#ifdef _DEBUG
	    if (rhssibucketptr[k] == NULL ||
		rhssvbucketptr[k] == NULL) {
	      printf("sspg2 error 10: Unexpected NULL rhssbucketptrs: ");
	      printf("vector # %d\n",k);
	      abort();
	    }
#endif
	  }	  
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 2. Initialize lhssvector.             *
	   ****************************************/
	  sgetibucket(&lhssibucketptr);
	  lhssvectorptr->dstype = governingdstype;
	  lhssvectorptr->sds.cs.ibucketptr = lhssibucketptr;
	  sgetvbucket(&lhssvbucketptr);
	  lhssvectorptr->sds.cs.vbucketptr = lhssvbucketptr;
	  lhssvectornonzeroranges = 0;
	  lhssvectornonzeroes = 0;
	  lhssvbucketindex = 0;
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3. Get nonzero by popping the        *
	   *    heaproot; consume it in saxpy;    *
	   *    either replace or delete heaproot.*
	   ****************************************/
	  
	  while (rhssvectorconsumednonzeroranges[0] <
		 rhssvectorptr[0]->sds.cs.nonzeroranges ||
		 rhssvectorconsumednonzeroranges[1] <
		 rhssvectorptr[1]->sds.cs.nonzeroranges) {
	    if
	      ((rhssvectorconsumednonzeroranges[1] ==
		rhssvectorptr[1]->sds.cs.nonzeroranges) ||
	       ((rhssvectorconsumednonzeroranges[0] <
		 rhssvectorptr[0]->sds.cs.nonzeroranges) &&
		(rhssibucketptr[0]->index_RANGE_BEGIN[rhssibucketindex[0]] <=
		 rhssibucketptr[1]->index_RANGE_BEGIN[rhssibucketindex[1]])))
		k = 0;
	    else if
	      ((rhssvectorconsumednonzeroranges[0] ==
		rhssvectorptr[0]->sds.cs.nonzeroranges) ||
	       ((rhssvectorconsumednonzeroranges[1] <
		 rhssvectorptr[1]->sds.cs.nonzeroranges) &&
		(rhssibucketptr[0]->index_RANGE_BEGIN[rhssibucketindex[0]] >
		 rhssibucketptr[1]->index_RANGE_BEGIN[rhssibucketindex[1]])))
		k = 1;
#ifdef _DEBUG
	    else {
	      printf("sspg2 error: BAD LOGIC.\n");
	      abort();
	    }
#endif
	    j1 = rhssibucketptr[k]->index_RANGE_BEGIN[rhssibucketindex[k]];
	    j2 = rhssibucketptr[k]->index_RANGE_END[rhssibucketindex[k]];
	    rhssibucketindex[k] = 
	      MY_MOD(rhssvectorconsumednonzeroranges[k],
		     sIBUCKET_DIM_minus1);
	    lhssibucketindex = MY_MOD(lhssvectornonzeroranges,
					  sIBUCKET_DIM_minus1);
	    
	    if ((rhssvbucketindex[k] = MY_MOD(rhssvbucketindex[k],
						  sVBUCKET_DIM_minus1))==0
		&& rhssvectorconsumednonzeroranges[k] !=0) {
	      sHOP_TO_NEXT_RHSVBUCKET(k,&rhssvbucketindex[k],
				     rhssvectorconsumednonzeroranges[k],
				     rhssvectornonzeroranges[k]);
	    }
	    lhssvbucketindex = MY_MOD(lhssvbucketindex,
					  sVBUCKET_DIM_minus1);
	    
#ifdef _DEBUG
	    if (rhssibucketptr[k]->errorflag != 9099 ||
		rhssvbucketptr[k]->errorflag != 9009) {
	      printf("sspg2 err 14: memory corruption of sibucketstore ");
	      printf("or svbucketstore.\n");
	      abort();
	    }
#endif      
	    
	    /****************************************
	     * Perform saxpy:                       *
	     * 3.0. Check for startup errors.       *
	     ****************************************/
	    
#ifdef _DEBUG
	    if (lhssvectornonzeroranges == 0) {
	      if (rhssibucketindex[k] != 0 || rhssvbucketindex[k] != 0 ||
		  lhssibucketindex != 0 || lhssvbucketindex != 0) {
		printf("sspg2 error 15: Zeroes expected: \n\t");
		printf("rhssibucketindex[k]=%d, rhssvbucketindex[k]=%d\n\t",
		       rhssibucketindex[k], rhssvbucketindex[k]);
		printf("lhssibucketindex=%d, lhssvbucketindex=%d\n\t",
		       lhssibucketindex,lhssvbucketindex);
		abort();
	      }
	    }
#endif      
	    
	    /****************************************
	     * Perform saxpy:                       *
	     * 3.1. Found 1ST or NEW subscript      *
	     *      range (non-overlapping with     *
	     *      previous one, if any) for       *
	     *      lhssvector.  If needed, get new  *
	     *      ibuckets and vbuckets for       *
	     *      lhssvector.                      *
	     ****************************************/
	    if (lhssvectornonzeroranges == 0 ||
		j1 > lhssibucketptr->index_RANGE_END[lastlhssibucketindex]+1) {/*A.1*/
	      
	      if (lhssibucketindex == 0 && lhssvectornonzeroranges != 0) {
		sgetibucket(&(lhssibucketptr->nextibucket));
		lhssibucketptr = lhssibucketptr->nextibucket;
	      }
	      lhssibucketptr->index_RANGE_BEGIN[lhssibucketindex] = j1;
	      lhssibucketptr->index_RANGE_END[lhssibucketindex] = j2;
	      
	      if (lhssvbucketindex == 0 && lhssvectornonzeroranges != 0)
		sget_and_hop_to_next_vbucket(&lhssvbucketindex);
	      
	      lastlhssvbucketptr = lhssvbucketptr;
	      lastlhssvbucketindex = lhssvbucketindex;
	      
	      totalcurrentnonzerorange = j2-j1+1;
	      
	      lhssvectornonzeroes += totalcurrentnonzerorange;
	      lhssvectornonzeroranges++;
	      
	
	      sNON_OVER_LAPPING_CS_RANGE_AXPY(COMPLETELY_NEW_RANGE,
					     lhssibucketindex,
					     &lastlhssibucketindex,
					     &lhssvbucketindex,
					     k,
					     &rhssvbucketindex[k],
					     alpha[k],
					     totalcurrentnonzerorange,
					     &unconsumedcurrentnonzerorange,
					     rhssvectorconsumednonzeroranges[k],
					     rhssvectornonzeroranges[k],
					     lhssvectornonzeroranges);
	      
	            
	    } /*A.1*/
	    
	    
	    /****************************************
	     * Perform saxpy:                       *
	     * 3.2. Found old subscript for         *
	     *      lhssvector.  Consume  nonzero in *
	     *      saxpy.                          *
	     *                                      *
	     ****************************************/
	    else if (j1 <= lhssibucketptr->index_RANGE_END[lastlhssibucketindex]+1 &&
		     j1 >= lhssibucketptr->index_RANGE_BEGIN[lastlhssibucketindex]
		     ) { /*B.1*/
	      
	      templhssvbucketptr = lhssvbucketptr;
	      templhssvbucketindex = lhssvbucketindex;
	      lhssvbucketptr = lastlhssvbucketptr;
	      
	      distancetooverlap =
		j1 - lhssibucketptr->index_RANGE_BEGIN[lastlhssibucketindex];
	      distancetoendofvbucket = sVBUCKET_DIM -
		lastlhssvbucketindex;
	      
	      if (j1 ==
		  lhssibucketptr->index_RANGE_END[lastlhssibucketindex]+1) {
		h1 = 0;
		h2 = -1;
	      }
	      else if (j2 <= lhssibucketptr->index_RANGE_END[lastlhssibucketindex]
		       ) {
		h1 = j1;
		h2 = j2;
		j1 = 0;
		j2 = -1;
	      }
	      else {
		h1 = j1;
		h2 = lhssibucketptr->index_RANGE_END[lastlhssibucketindex];
		j1 = h2+1;
	      }
	      
	      if (distancetoendofvbucket > distancetooverlap) {
		lhssvbucketindex =
		  MY_MOD(lastlhssvbucketindex+distancetooverlap,
			 sVBUCKET_DIM_minus1);
	      }
	      else if (distancetoendofvbucket < distancetooverlap) {
		lhssvbucketindex =
		  MY_MOD(distancetooverlap - distancetoendofvbucket,
			 sVBUCKET_DIM_minus1);
	      }
	      else {
		lhssvbucketindex = 0;
	      }
	      
	      while (distancetooverlap >= distancetoendofvbucket &&
		     lhssvbucketptr->nextvbucket != NULL) {
		lhssvbucketptr = lhssvbucketptr->nextvbucket;
		distancetooverlap -= distancetoendofvbucket;
		distancetoendofvbucket = sVBUCKET_DIM;
	      }
	      
	      lhssibucketptr->index_RANGE_END[lastlhssibucketindex] =
		max(j2,
		    lhssibucketptr->index_RANGE_END[lastlhssibucketindex]);
	      
	      if ((totalcurrentnonzerorange = h2-h1+1) > 0) {
		

		sOVERLAPPING_CS_RANGE_AXPY(&lhssvbucketindex,
					  k,
					  &rhssvbucketindex[k],
					  alpha[k],
					  totalcurrentnonzerorange,
					  &unconsumedcurrentnonzerorange,
					  rhssvectorconsumednonzeroranges[k],
					  rhssvectornonzeroranges[k],
					  lhssvectornonzeroranges);
		
	      }
	     
	      /*
	       * This bit of logic must be performed, whether or not there
	       * is additional nonzeroes to process.
	       */
	      
	      if (j2 != -1 &&
		  (lhssvbucketptr != templhssvbucketptr ||
		   (lhssvbucketindex = MY_MOD(lhssvbucketindex,
						  sVBUCKET_DIM_minus1)) !=
		   templhssvbucketindex)) {
		
#ifdef _DEBUG
		printf("sspg2 err 50.0: Unexpectd value discrepancies\n");
		printf("lhssvbucketptr=%d, templhssvbucketptr=%d\n",
		       (int)lhssvbucketptr,(int)templhssvbucketptr);
		printf("lhssvbucketindex=%d, templhssvbucketindex=%d\n",
		       lhssvbucketindex,templhssvbucketindex);
		
		abort();
#endif
	      }
	      else {
		lhssvbucketptr = templhssvbucketptr;
		lhssvbucketindex = templhssvbucketindex;
	      }
	      
	      
	      if ((totalcurrentnonzerorange = j2-j1+1) > 0) {
		lhssvectornonzeroes += totalcurrentnonzerorange;
		
		/*
		 * adding new entries to vbucket
		 */
		
		if (lhssvbucketindex == 0 &&
		    lhssvectornonzeroranges != 0 &&
		    j2 != -1)
		  sget_and_hop_to_next_vbucket(&lhssvbucketindex);
		
		if(rhssvbucketindex[k] == sVBUCKET_DIM && j2 != -1) {

		  sHOP_TO_NEXT_RHSVBUCKET(k,&rhssvbucketindex[k],
					 rhssvectorconsumednonzeroranges[k],
					 rhssvectornonzeroranges[k]);
		}
		
		
		
		sNON_OVER_LAPPING_CS_RANGE_AXPY(NEW_ENTRIES_PADDED_TO_OLD_RANGE,
					       lhssibucketindex,
					       &lastlhssibucketindex,
					       &lhssvbucketindex,
					       k,
					       &rhssvbucketindex[k],
					       alpha[k],
					       totalcurrentnonzerorange,
					       &unconsumedcurrentnonzerorange,
					       rhssvectorconsumednonzeroranges[k],
					       rhssvectornonzeroranges[k],
					       lhssvectornonzeroranges);
		
	      } /* matches     if ((totalcurrent...) { */
	    } /*B.1*/
	    
	    
	    /****************************************	
	     * Perform saxpy:                       *
	     * 3.3. Found wrong subscript for       *
	     *      lhssvector.  Abort.               *
	     ****************************************/
	    
#ifdef _DEBUG
	    else {
	      printf("sspg2 error 60: lhssvectornonzeroranges=%d,",
		     lhssvectornonzeroranges);
	      printf("lastlhssibucketindex=%d, \n\t",lastlhssibucketindex);
	      printf("lhssibucketptr->index[%d] = (%d,%d), j1=%d\n",
		     lastlhssibucketindex,
		     lhssibucketptr->index_RANGE_BEGIN[lastlhssibucketindex],
		     lhssibucketptr->index_RANGE_END[lastlhssibucketindex],
		     j1);
	      abort();
	    }
#endif
	    
	    
	    /****************************************
	     * Perform saxpy:                       *
	     * 3.4. If there is one, get new        *
	     *      nonzero from the rhs_vector     *
	     *      whose nonzero was just          *
	     *      consumed and replace the        *
	     *      heaproot with new entry;        *
	     *      Else, delete heaproot.          *
	     ****************************************/
	    if (++rhssvectorconsumednonzeroranges[k] <
		rhssvectornonzeroranges[k]) {
	      if ((rhssibucketindex[k] = 
		   MY_MOD(rhssvectorconsumednonzeroranges[k],
			  sIBUCKET_DIM_minus1)) == 0) {
		if ((rhssibucketptr[k] =
		     rhssibucketptr[k]->nextibucket) == NULL) {
		  
#ifdef _DEBUG
		  printf("sspg2 err 61: unexpected NULL nextvbucket: \n");
		  printf("\t k=%d, rhssvectorconsumednonzeroranges=%d,\n",
			 k,rhssvectorconsumednonzeroranges[k]);
		  printf("\t rhssvectornonzeroranges[k]=%d\n",
			 rhssvectornonzeroranges[k]);
		  abort();
#endif
		  
		}
	      }
	    }   
	  } /* matches     while loop          */
	} /* matches     else ** if (governingdstype == CS) ** {     */
	

	
	
	
	rhssvectorptrptrs[0] = &rhssvectorptr[0];
	rhssvectorptrptrs[1] = &rhssvectorptr[1];
	
	saxpy_2_heap_post_processing(destptrptr,
				    rhssvectorptrptrs,
				    alphaptrs,
				    restorerhssvectorptr,
				    arity,
				    lhssvectornonzeroes,
				    lhssvectornonzeroranges,
				    governingdstype);
	
      } /* matches     if we are going to use lhssvectorptr           */
    } /* matches     else ** do an honest to goodness sspg2 **  {        */
  } /* matches     else ** neither rhs_operation is a NO-OP  ** {         */
    
  
#ifdef _PEEP
  sadpeep_post(destptrptr);
#endif
  
#ifdef _TRACE 
  nnza = (*destptrptr)->nonzeroes; 
  sadinfo(arity,nnzb,lhssvectoraddress,nnza,alphaptrs,vectorptrptrs);
#endif
  
}

