#define SINGLE
/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#include  <stdlib.h>
#include "sgaxpy_pd_macros_and_data_structures.h"
#include "sgaxpy_pd_global_variables.h"

/****************************************************************************
 *                                                                          *
 *  sspgx(destptrptr,arity,alphaptrs,rhssvectorptrptrs)             *
 *                                                                          *
 *     destptrptr      is a doubly-indirected pointer to a svectorstruct*
 *                     It is passed to sspgx by the adxpyN routines,    *
 *                     where 1 <= N <= MAXARITY.  In turn destptrptr is an  *
 *                     argument of the adxpyN functions which are called by *
 *                     ADIFOR-generated FORTRAN calls.  The purpose of the  *
 *                     double indirection is that FORTRAN calls by          *
 *                     reference, that is it puts the address of say g$x(1) *
 *                     on the stack.  In the FORTRAN code, the value of     *
 *                     g$x(1) represents a pointer to (address of) a sparse *
 *                     vector (i.e., svectorstruct *).                  *
 *                     The FORTRAN call,                                    *
 *                     CALL adxpyN(g$x(1),...), causes inside C for an      *
 *                     internal copy, say destptrptr, to equal the address  *
 *                     of g$x(1), i.e., destptrptr == &g$x(1).  Therefore,  *
 *                     *destptrptr == g$x(1). But since the value contained *
 *                     in g$x(1) is (as interpreted by C) of type           *
 *                     (svectorstruct *), destptrptr needs              *
 *                     one more level of                                    *
 *                     indirection, and is declared as                      *
 *                     (svectorstruct **).                              *
 *                                                                          *
 *     arity           is the # of vectors on the rhs.                      *
 *                                                                          *
 *     alphaptrs       Array of pointers to the scalar multipliers.         *
 *                                                                          *
 *     rhssvectorptrptrs   Array of doubly-indirected pointers to       *
 *                     svectorstructs.  In this case, since these values*
 *                     are not changed, there is no need for the double-    *
 *                     indirection; we simply want to maintain the g$x      *
 *                     references consistant.                               *
 *                                                                          *
 *     sspgx performs the saxpy operation for all aritys.               *
 *                                                                          *
 ****************************************************************************/

void sspgx(destptrptr,arity,alphaptrs,rhssvectorptrptrs)
svectorstruct **destptrptr;
int arity;
float *alphaptrs[];
svectorstruct **rhssvectorptrptrs[];
{

  svectorstruct *rhssvectorptr[MAXARITY];
                                 /* Dereferenced rhssvectorptrptrs[].   */

  sbucketstruct *rhssbucketptr[MAXARITY];
                                      /* Array of pointers to "current"     *
				       * buckets where "current" nonzeroes  *
				       * of the rhs_vectors reside.         */

  sbucketstruct *lhssbucketptr;
                                      /* Pointer to current bucket of       *
				       * lhssvector.                    */

  sibucketstruct *rhssibucketptr[MAXARITY];
                                      /* Array of pointers to "current"     *
				       * buckets where "current" nonzeroes  *
				       * of the rhs_vectors reside.         */

  sibucketstruct *lhssibucketptr;
                                      /* Pointer to current bucket of       *
				       * lhssvector.                    */

  float alpha[MAXARITY];   /* Dereferenced alphaptrs[].          */

  int rhssbucketindex[MAXARITY];  /* Array of indices of current        *
				       * nonzeroes in the current buckets   *
				       * of rhs_vectors.                    */

  int lhssbucketindex;            /* bucketindex of next nonzero to be  *
				       * added to lhssvector.           */

  int lastlhssbucketindex;        /* bucketindex of the latest added    *
				       * nonzero in the lhssvector.  We *
				       * need this additional index to      *
				       * leap-frog buckets, if need be.     */

  int rhssibucketindex[MAXARITY]; /* Array of indices of current        *
				       * nonzeroes in the current buckets   *
				       * of rhs_vectors.                    */

  int lhssibucketindex;           /* bucketindex of next nonzero to be  *
				       * added to lhssvector.           */

  int lastlhssibucketindex;       /* bucketindex of the latest added    *
				       * nonzero in the lhssvector.  We *
				       * need this additional index to      *
				       * leap-frog buckets, if need be.     */

  int firstsaxpyoperand;              /* Order in the ennumeration of the   *
				       * rhs_vectors of the first non-null  *
				       * saxpy vectorop.                    */
  int k;                              /* Counter for arity.                 */

  int j;                              /* Index for vectorsubscripts to be   *
				       * entered in (and later popped from) *
				       * the heap.                          */

  int rhssvectornonzeroes[MAXARITY];
                                      /* Array of total # of nonzeroes in   *
				       * each rhs_vector.                   */

  int rhssvectorconsumednonzeroes[MAXARITY];
                                      /* Array of # of nonzeroes consumed   *
				       * so far in each rhs_vector.         */

  int lhssvectornonzeroes;        /* Count of # of nonzeroes entered    *
				       * into the lhssvector so far. By *
				       * the end of the saxpy operation     *
				       * this index equals the total # of   *
				       * nonzeroes in the resultant vector. */

/*
 * These variables were originally for COMPRESSED_SUBSCRIPTS
 */

  svbucketstruct *lastlhssvbucketptr;
                                      /* Pointer to current vbucket of      *
				       * lhssvector.                    */

  int rhssvectornonzeroranges[MAXARITY];
                                      /* Array of total # of nonzeroes      *
				       * ranges in each rhs_vector.         */

  int rhssvectorconsumednonzeroranges[MAXARITY];
                                      /* Array of # of nonzeroes consumed   *
				       * so far in each rhs_vector.         */

  int rhssvbucketindex[MAXARITY]; /* Array of indices of current        *
				       * nonzeroes in the current buckets   *
				       * of rhs_vectors.                    */

  int lhssvectornonzeroranges;    /* Count of # of nonzeroes entered    *
				       * into the lhssvector so far. By *
				       * the end of the saxpy operation     *
				       * this index equals the total # of   *
				       * nonzeroes in the resultant vector. */

  int lhssvbucketindex;           /* bucketindex of next nonzero to be  *
				       * added to lhssvector.           */


  int lastlhssvbucketindex;       /* bucketindex of the latest added    *
				       * nonzero in the lhssvector.  We *
				       * need this additional index to      *
				       * leap-frog buckets, if need be.     */

  svbucketstruct *templhssvbucketptr;
  int templhssvbucketindex;

  int j1;                             /* Index for vectorsubscripts to be   *
				       * entered in (and later popped from) *
				       * the heap.                          */

  int j2;                             /* Index for vectorsubscripts to be   *
				       * entered in (and later popped from) *
				       * the heap.                          */

  int h1;
  int h2;

  int totalcurrentnonzerorange;

#ifdef _DEBUG
  int unconsumedcurrentnonzerorange;
#endif

  int distancetooverlap;
  int distancetoendofvbucket;


  /*
   * These variables were added for union ds
   */
  
  svectorstruct *restorerhssvectorptr[MAXARITY+1];

  int governingdstype;
  int rhs_largest_dstype = S1;        /* Largest dstype of any              *
				       * rhssvector ( CS > SS > S1 ).   */
  int s1_index;
  
  float temp_gaxpy;

  /*
   * These variables were added for plusequal code
   */

#ifdef _PLUSEQUAL
  int plusequalsaxpy_flag = 0;
  int plusequalkorder;
  svectorstruct **tempdestptrptr;
  float *tempalphaptrs[MAXARITY-1];
  svectorstruct **temprhssvectorptrptrs[MAXARITY-1];
  float one; 
  int smallest_rhs_index;
#endif /* _PLUSEQUAL */

  /****************************************
   * Initialization of *destptrptr.       *
   ****************************************/
  
  if (*destptrptr == NULL)
    sgetvector(destptrptr);
  
  
  /****************************************
   * Dereferencing.                       *
   ****************************************/
  
  for (k=0;k<arity;k++) {
    alpha[k] = *alphaptrs[k];

    rhssvectorptr[k] = *rhssvectorptrptrs[k];
    restorerhssvectorptr[k] = rhssvectorptr[k];
    rhssvectornonzeroes[k] = rhssvectorptr[k]->nonzeroes;
    if (rhssvectorptr[k]->dstype == CS)
      rhssvectornonzeroranges[k] =
	rhssvectorptr[k]->sds.cs.nonzeroranges;
    if (rhssvectorptr[k]->dstype > rhs_largest_dstype)
      rhs_largest_dstype = rhssvectorptr[k]->dstype;

#ifdef _PLUSEQUAL    
    if (rhssvectorptr[k] == *destptrptr &&
	rhssvectorptr[k]->dstype == CS &&
	X_IS_ONE(alpha[k])) {
      plusequalsaxpy_flag++;
      plusequalkorder = k;
    }
#endif /* _PLUSEQUAL */

  }	
  
  /****************************************
   * Determination of 1st nonzero         *
   * rhssvector operand.              *
   ****************************************/
  
  firstsaxpyoperand = 0;

/*
 *   This bit of the code is obsolete, since the routines calling
 *   this routine already have stripped the gaxpy of zero vectors
 *   and zero multipliers.
 *

  while (nullsaxpyflag && firstsaxpyoperand<arity)
    
    if (rhssvectornonzeroes[firstsaxpyoperand] == 0 ||
	X_IS_ZERO(alpha[firstsaxpyoperand]))
      firstsaxpyoperand++;
    else
      nullsaxpyflag = 0;
  
 *
 * If no nonzero vectorops, return a null vector.
 *

  if (nullsaxpyflag) {
    
    (*destptrptr)->nonzeroes = 0;
    (*destptrptr)->dstype = S1;

  }
  else {

*/
    
    /****************************************
     * Predetermine governingdstype and     *
     * appropriate saxpy routine.           *
     ****************************************/

  if (rhs_largest_dstype == S1) {
    s1_index = rhssvectorptr[firstsaxpyoperand]->sds.s1.index;
    for (k=firstsaxpyoperand+1;k<arity;k++)
      if (rhssvectorptr[k]->sds.s1.index !=
	  rhssvectorptr[firstsaxpyoperand]->sds.s1.index &&
	  ! X_IS_ZERO(alpha[k]))
	s1_index = -1;
    if (s1_index > 0)
      governingdstype = S1;
    else
      governingdstype = SS;
  }
  else if (rhs_largest_dstype <= SS)
    governingdstype = SS;
  else /* if (rhs_largest_dstype == CS) */
    governingdstype = CS;

#ifdef _PLUSEQUAL
  smallest_rhs_index = DUMMYSUBSCRIPT;
  for (k=0;k<arity;k++) {
    if (rhssvectorptr[k]->dstype == S1) 
      smallest_rhs_index = min (smallest_rhs_index,
				rhssvectorptr[k]->sds.s1.index);
    else if (rhssvectorptr[k]->dstype == SS)
      smallest_rhs_index = min (smallest_rhs_index,
        rhssvectorptr[k]->sds.ss.bucketptr->index[0]);
    else /* if (rhssvectorptr[k]->dstype == CS) */
      smallest_rhs_index = min (smallest_rhs_index,
	rhssvectorptr[k]->sds.cs.ibucketptr->index_RANGE_BEGIN[0]);
  }

  if (plusequalsaxpy_flag &&
      (*destptrptr)->dstype == CS &&
      (*destptrptr)->sds.cs.ibucketptr->index_RANGE_BEGIN[0] <=
      smallest_rhs_index) {    
#ifdef _DEBUG
    if (plusequalsaxpy_flag > 1) {
      printf("sspgx warning: lhssvector appears more than ");
      printf("once on rhs.\n");
    }
#endif
    if (arity == 1) {
      plusequalsspg1(destptrptr,alphaptrs[plusequalkorder]);
    }
    else {
      for (k=0;k<plusequalkorder;k++) {
	tempalphaptrs[k] = alphaptrs[k];
	temprhssvectorptrptrs[k] = rhssvectorptrptrs[k];
      }
      for (k=plusequalkorder+1;k<arity;k++) {
	tempalphaptrs[k-1] = alphaptrs[k];
	temprhssvectorptrptrs[k-1] = rhssvectorptrptrs[k];
      }
      tempdestptrptr = &tempsdestptr;
      sspgx(tempdestptrptr,arity-1,
		tempalphaptrs,temprhssvectorptrptrs);
      restorerhssvectorptr[MAXARITY] = tempsdestptr;
      if ((*tempdestptrptr)->dstype != governingdstype &&
	  (*tempdestptrptr)->nonzeroes !=0) {
	sconvertdstype(tempdestptrptr,MAXARITY,governingdstype);
      }
      X_PUT_REAL(one,1.0);
      plusequalsspg2(destptrptr,
		      &one,tempdestptrptr,TEMP_OTHER_VECTOR);
      tempsdestptr = restorerhssvectorptr[MAXARITY];
    }    
  }
  else /* if (not plusequalsaxpy_flag) */ {
#endif /* _PLUSEQUAL */

    for (k=0;k<arity;k++) {
      if (rhssvectorptr[k]->dstype != governingdstype &&
	  rhssvectorptr[k]->nonzeroes !=0 && ! X_IS_ZERO(alpha[k])) {
	sconvertdstype((&rhssvectorptr[k]),k,governingdstype);
	if (governingdstype == CS)
	  rhssvectornonzeroranges[k] = 
	    rhssvectorptr[k]->sds.cs.nonzeroranges;
      }
    }
    
    if (governingdstype == S1) {
      
      /********************************************************************
       *                       ONE_SUBSCRIPT AXPY                         *
       ********************************************************************/

      lhssvectorptr->nonzeroes = 1;
      lhssvectorptr->dstype = S1;
      lhssvectorptr->sds.s1.index =
	rhssvectorptr[firstsaxpyoperand]->sds.s1.index;
      X_ZERO(temp_gaxpy);
      for (k=firstsaxpyoperand;k<arity;k++) {
	GAXPY_PLUS_EQUALS(temp_gaxpy,
			  alpha[k], rhssvectorptr[k]->sds.s1.value);
      }
      X_ASSIGN(lhssvectorptr->sds.s1.value,temp_gaxpy);

      saxpy_0_1_post_processing(destptrptr);

    }
    else /* if we are going to use lhssvectorptr */ {
      if (governingdstype == SS) {
	
	/******************************************************************
	 *                  _SINGLE_SUBSCRIPTS AXPY                       *
	 ******************************************************************/
	
	/****************************************
	 * Else perform saxpy:                  *
	 * 1. Put in heap the 1st nonzero of    *
	 *      each vector.                    *
	 ****************************************/
	rhssvectorconsumednonzeroes[firstsaxpyoperand] = 0;
	rhssbucketindex[firstsaxpyoperand] = 0;
	rhssbucketptr[firstsaxpyoperand] =
	  rhssvectorptr[firstsaxpyoperand]->sds.ss.bucketptr;
	
#ifdef _DEBUG
	if (rhssbucketptr[firstsaxpyoperand] == NULL) {
	  printf("sspgx error 1: Unexpected NULL rhssbucketptr: V# %d\n",
		 firstsaxpyoperand);
	  abort();
	}
#endif
	
	j = rhssbucketptr[firstsaxpyoperand]->
	  index[rhssbucketindex[firstsaxpyoperand]];
	
	if (j <= 0) {
#ifdef _DEBUG
	  printf("sspgx error 2: Illegal index: V # %d, subscript=%d\n",
		 firstsaxpyoperand,j);
	  abort();
#endif
	}
	
	else
	  heapinitialize(firstsaxpyoperand,j,j);
	
	for (k=firstsaxpyoperand+1;k<arity;k++) {
	  if (rhssvectornonzeroes[k] !=0 && ! X_IS_ZERO(alpha[k])) {
	    rhssvectorconsumednonzeroes[k] = 0;
	    rhssbucketindex[k] = 0;
	    rhssbucketptr[k] = rhssvectorptr[k]->sds.ss.bucketptr;
	    
#ifdef _DEBUG
	    if (rhssbucketptr[k] == NULL) {
	      printf("sspgx error 3: Unexpected NULL rhssbucketptr: ");
	      printf("vector # %d\n",k);
	      abort();
	    }
#endif
	    
	    j = rhssbucketptr[k]->index[rhssbucketindex[k]];
	    
#ifdef _DEBUG
	    if (j <= 0) {
	      printf("sspgx error 4: Illegal index: vec.=%d sub.=%d\n",
		     k,j);
	      abort();
	    }
#endif
	    
	    heapinsert(k,j,j);
	  }
	}
	
	/****************************************
	 * Perform saxpy:                       *
	 * 2. Initialize lhssvector.             *
	 ****************************************/
	sgetbucket(&lhssbucketptr);
	lhssvectorptr->dstype = governingdstype;
	lhssvectorptr->sds.ss.bucketptr = lhssbucketptr;
	lhssvectornonzeroes = 0;
	
	/****************************************
	 * Perform saxpy:                       *
	 * 3. Get nonzero by popping the        *
	 *    heaproot; consume it in saxpy;    *
	 *    replace / delete heaproot.        *
	 ****************************************/
	while (heaproot !=NULL) {
	  k = heaproot->vectorindex;
	  j = heaproot->rangebeginindex;
	  rhssbucketindex[k] = MY_MOD(rhssvectorconsumednonzeroes[k],
					  sBUCKET_DIM_minus1);
	  lhssbucketindex = MY_MOD(lhssvectornonzeroes,
				       sBUCKET_DIM_minus1);
	  
#ifdef _DEBUG
	  if (rhssbucketptr[k]->errorflag != 9999) {
	    printf("sspgx err 5: mem. corruption of sbucketstore\n");
	    abort();
	  }
#endif      
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.1. Found 1st or new subscript for  *
	   *      lhssvector.  If needed, get new  *
	   *      bucket for lhssvector.           *
	   *      Consume nonzero in saxpy.       *
	   ****************************************/
	  if (lhssvectornonzeroes == 0 ||
	      lhssbucketptr->index[lastlhssbucketindex] < j) {
	    if (lhssbucketindex == 0 && lhssvectornonzeroes != 0) {
	      sgetbucket(&(lhssbucketptr->nextbucket));
	      lhssbucketptr = lhssbucketptr->nextbucket;
	    }
	    lhssbucketptr->index[lhssbucketindex] = j;
	    GAXPY_EQUALS(lhssbucketptr->value[lhssbucketindex],
			 alpha[k],
			 rhssbucketptr[k]->value[rhssbucketindex[k]]);
	    lastlhssbucketindex = lhssbucketindex;
	    lhssvectornonzeroes++;
	  }
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.2. Found old subscript for         *
	   *      lhssvector.                      *
	   *      Consume nonzero in saxpy.       *
	   ****************************************/
	  else if (lhssbucketptr->index[lastlhssbucketindex] == j) {
	    GAXPY_PLUS_EQUALS(
	      lhssbucketptr->value[lastlhssbucketindex],
	      alpha[k],
	      rhssbucketptr[k]->value[rhssbucketindex[k]]);
	  }
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.3. Found wrong subscript for       *
	   *      lhssvector.  Abort.               *
	   ****************************************/
	  
#ifdef _DEBUG
	  else {
	    printf("sspgx error 6: lhssnz=%d, lastlhssindex[]=%d, j=%d\n",
		   lhssvectornonzeroes,
		   lhssbucketptr->index[lastlhssbucketindex],j);
	    abort();
	  }
#endif
	  
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.4. If there is one, get new        *
	   *      nonzero from the rhs_vector     *
	   *      whose nonzero was just          *
	   *      consumed and replace the        *
	   *      heaproot with new entry;        *
	   *      Else, delete heaproot.          *
	   ****************************************/
	  if (++rhssvectorconsumednonzeroes[k] < rhssvectornonzeroes[k]) {
	    if ((rhssbucketindex[k] =
		 MY_MOD(rhssvectorconsumednonzeroes[k],
			sBUCKET_DIM_minus1)) == 0) {
	      if ((rhssbucketptr[k] = rhssbucketptr[k]->nextbucket) == NULL) {
		
#ifdef _DEBUG
		printf("sspgx err 7: unexpected NULL nextbucket: k=%d",
		       k);
		printf("rhssvectorconsumednonzeros=%d, rhssvecnonzeroes=%d,",
		       rhssvectorconsumednonzeroes[k],rhssvectornonzeroes[k]);
		abort();
#endif
		
	      }
	    }     
	    j = rhssbucketptr[k]->index[rhssbucketindex[k]];
	    
	    if (j <= 0) {
#ifdef _DEBUG
	      printf("sspgx err 8: Illegal index: V=%d, subscript=%d\n"
		     ,k,j);
	      abort();
#endif
	      
	    }
	    
	    else
	      heapop(REPLACE,k,j,j);
	  }
	  else
	    heapop(DELETE,DUMMYVECTOR,DUMMYSUBSCRIPT,DUMMYSUBSCRIPT);
	  
	} /*matches     while (heaproot !=NULL) {     */
      } /*matches     else if (governingdstype == SS) {     */
      
      else /* if (governingdstype == CS) */ {
	
	/******************************************************************
	 *                 _COMPRESSED_SUBSCRIPTS AXPY                    *
	 ******************************************************************/
	
	
	/****************************************
	 * Else perform saxpy:                  *
	 * 1. Put in heap the 1st nonzero of    *
	 *    each vector.                      *
	 ****************************************/
	rhssvectorconsumednonzeroranges[firstsaxpyoperand] = 0;
	rhssibucketindex[firstsaxpyoperand] = 0;
	rhssvbucketindex[firstsaxpyoperand] = 0;
	rhssibucketptr[firstsaxpyoperand] =
	  rhssvectorptr[firstsaxpyoperand]->sds.cs.ibucketptr;
	rhssvbucketptr[firstsaxpyoperand] =
	  rhssvectorptr[firstsaxpyoperand]->sds.cs.vbucketptr;
	
#ifdef _DEBUG
	if (rhssibucketptr[firstsaxpyoperand] == NULL ||
	    rhssvbucketptr[firstsaxpyoperand] == NULL) {
	  printf("sspgx error 10: Unexpected NULL rhssbucketptrs: ");
	  printf("vector # %d\n",firstsaxpyoperand);
	  abort();
	}
#endif
	
	j1 = rhssibucketptr[firstsaxpyoperand]->
	  index_RANGE_BEGIN[rhssibucketindex[firstsaxpyoperand]];
	j2 = rhssibucketptr[firstsaxpyoperand]->
	  index_RANGE_END[rhssibucketindex[firstsaxpyoperand]];
	
	if (j1 <= 0 || j2 <= 0 || j2 < j1) {
#ifdef _DEBUG
	  printf("sspgx error 11: Nonsensical Nonzero Range; \n\t\t");
	  printf("vector # %d, subscript 1=%d subscript 2=%d\n",
		 firstsaxpyoperand,j1,j2);
	  abort();
#endif
	}
	
	else
	  heapinitialize(firstsaxpyoperand,j1,j2);
	
	for (k=firstsaxpyoperand+1;k<arity;k++) {

	  /*
	   * rhssvectornonzeroes[k] !=0 will be redundant and
	   *                           unnecessary most of the time
	   *                           but in the code preceding the
	   *                           addition of += module it was
	   *                           needed, so we leave it.
	   */
	
	  if (rhssvectornonzeroes[k] !=0 &&
	      rhssvectornonzeroranges[k] !=0 &&
	      ! X_IS_ZERO(alpha[k])) {
	    rhssvectorconsumednonzeroranges[k] = 0;
	    rhssibucketindex[k] = 0;
	    rhssibucketptr[k] =
	      rhssvectorptr[k]->sds.cs.ibucketptr;
	    rhssvbucketindex[k] = 0;
	    rhssvbucketptr[k] =
	      rhssvectorptr[k]->sds.cs.vbucketptr;
	    
#ifdef _DEBUG
	    if (rhssibucketptr[k] == NULL ||
		rhssvbucketptr[k] == NULL) {
	      printf("sspgx error 12: Unexpected ");
	      printf("NULL rhssibucketptr or rhssvbucketptr: vector # %d\n",k);
	      abort();
	    }
#endif
	    
	    j1 = rhssibucketptr[k]->index_RANGE_BEGIN[rhssibucketindex[k]];
	    j2 = rhssibucketptr[k]->index_RANGE_END[rhssibucketindex[k]];
	    
#ifdef _DEBUG
	    if (j1 <= 0 || j2 <= 0 || j2 < j1) {
	      printf("sspgx err 13: Nonsensical Nonzero Range; \n\t\t");
	      printf("vector # %d, subscript 1=%d subscript 2=%d\n",
		     firstsaxpyoperand,j1,j2);
	      abort();
	    }
#endif
	    
	    heapinsert(k,j1,j2);
	  }
	}
	
	/****************************************
	 * Perform saxpy:                       *
	 * 2. Initialize lhssvector.             *
	 ****************************************/
	sgetibucket(&lhssibucketptr);
	lhssvectorptr->dstype = governingdstype;
	lhssvectorptr->sds.cs.ibucketptr = lhssibucketptr;
	sgetvbucket(&lhssvbucketptr);
	lhssvectorptr->sds.cs.vbucketptr = lhssvbucketptr;
	lhssvectornonzeroranges = 0;
	lhssvectornonzeroes = 0;
	lhssvbucketindex = 0;
	
	/****************************************
	 * Perform saxpy:                       *
	 * 3. Get nonzero by popping the        *
	 *    heaproot; consume it in saxpy;    *
	 *    either replace or delete heaproot.*
	 ****************************************/

	while (heaproot !=NULL) {
	  k = heaproot->vectorindex;
	  j1 = heaproot->rangebeginindex;
	  j2 = heaproot->rangeendindex;
	  rhssibucketindex[k] =
	    MY_MOD(rhssvectorconsumednonzeroranges[k],
		   sIBUCKET_DIM_minus1);
	  lhssibucketindex = MY_MOD(lhssvectornonzeroranges,
					sIBUCKET_DIM_minus1);
	  if ((rhssvbucketindex[k] =
	       MY_MOD(rhssvbucketindex[k],sVBUCKET_DIM_minus1))==0
	      && rhssvectorconsumednonzeroranges[k] !=0) {
	    sHOP_TO_NEXT_RHSVBUCKET(k,&rhssvbucketindex[k],
				   rhssvectorconsumednonzeroranges[k],
				   rhssvectornonzeroranges[k]);
	  }
	  lhssvbucketindex = MY_MOD(lhssvbucketindex,
					sVBUCKET_DIM_minus1);
	  
	  
	  
#ifdef _DEBUG
	  if (rhssibucketptr[k]->errorflag != 9099 ||
	      rhssvbucketptr[k]->errorflag != 9009) {
	    printf("sspgx err 14: mem corruption of sibucketstore ");
	    printf("or svbucketstore.\n");
	    abort();
	  }
#endif      
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.0. Check for startup errors.       *
	   ****************************************/
	  
#ifdef _DEBUG
	  if (lhssvectornonzeroranges == 0) {
	    if (rhssibucketindex[k] != 0 || rhssvbucketindex[k] != 0 ||
		lhssibucketindex != 0 || lhssvbucketindex != 0) {
	      printf("sspgx error 15: Zeroes expected: \n\t");
	      printf("rhssibucketindex[k]=%d, rhssvbucketindex[k]=%d\n\t",
		     rhssibucketindex[k], rhssvbucketindex[k]);
	      printf("lhssibucketindex=%d, lhssvbucketindex=%d\n\t",
		     lhssibucketindex,lhssvbucketindex);
	      abort();
	    }
	  }
#endif      
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.1. Found 1ST or NEW subscript      *
	   *      range (non-overlapping with     *
	   *      previous one, if any) for       *
	   *      lhssvector.  If needed, get new  *
	   *      ibuckets and vbuckets for        *
	   *      lhssvector.                      *
	   ****************************************/
	  if (lhssvectornonzeroranges == 0 ||
	      j1 > lhssibucketptr->index_RANGE_END[lastlhssibucketindex]+1) {/*A.1*/
	    
	    if (lhssibucketindex == 0 && 
		lhssvectornonzeroranges != 0) {
	      sgetibucket(&(lhssibucketptr->nextibucket));
	      lhssibucketptr = lhssibucketptr->nextibucket;
	    }
	    lhssibucketptr->index_RANGE_BEGIN[lhssibucketindex] = j1;
	    lhssibucketptr->index_RANGE_END[lhssibucketindex] = j2;
	    
	    if (lhssvbucketindex == 0 &&
		lhssvectornonzeroranges != 0)
	      sget_and_hop_to_next_vbucket(&lhssvbucketindex);
	    
	    lastlhssvbucketptr = lhssvbucketptr;
	    lastlhssvbucketindex = lhssvbucketindex;
	    
	    totalcurrentnonzerorange = j2-j1+1;
	    
	    lhssvectornonzeroes += totalcurrentnonzerorange;
	    lhssvectornonzeroranges++;
	    
	    
	    sNON_OVER_LAPPING_CS_RANGE_AXPY(COMPLETELY_NEW_RANGE,
					   lhssibucketindex,
					   &lastlhssibucketindex,
					   &lhssvbucketindex,
					   k,
					   &rhssvbucketindex[k],
					   alpha[k],
					   totalcurrentnonzerorange,
					   &unconsumedcurrentnonzerorange,
					   rhssvectorconsumednonzeroranges[k],
					   rhssvectornonzeroranges[k],
					   lhssvectornonzeroranges);
	    

	  } /*A.1*/
	  
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.2. Found old subscript for         *
	   *      lhssvector.  Consume  nonzero in *
	   *      saxpy.                          *
	   *                                      *
	   ****************************************/
	  else if (j1 <= 
		   lhssibucketptr->index_RANGE_END[lastlhssibucketindex]+1 &&
		   j1 >= 
		   lhssibucketptr->index_RANGE_BEGIN[lastlhssibucketindex]) {
	    /*B.1*/
	    
	    templhssvbucketptr = lhssvbucketptr;
	    templhssvbucketindex = lhssvbucketindex;
	    lhssvbucketptr = lastlhssvbucketptr;
	    
	    distancetooverlap =
	      j1 - lhssibucketptr->index_RANGE_BEGIN[lastlhssibucketindex];
	    distancetoendofvbucket = 
	      sVBUCKET_DIM - lastlhssvbucketindex;
	    
	    if (j1 == lhssibucketptr->index_RANGE_END[lastlhssibucketindex]+1) {
	      h1 = 0;
	      h2 = -1;
	    }
	    else if (j2 <= lhssibucketptr->index_RANGE_END[lastlhssibucketindex]) {
	      h1 = j1;
	      h2 = j2;
	      j1 = 0;
	      j2 = -1;
	    }
	    else {
	      h1 = j1;
	      h2 = lhssibucketptr->index_RANGE_END[lastlhssibucketindex];
	      j1 = h2+1;
	    }
	    
	    if (distancetoendofvbucket > distancetooverlap) {
	      lhssvbucketindex =
		MY_MOD(lastlhssvbucketindex + distancetooverlap,
		       sVBUCKET_DIM_minus1);
	    }
	    else if (distancetoendofvbucket < distancetooverlap) {
	      lhssvbucketindex =
		MY_MOD(distancetooverlap - distancetoendofvbucket,
		       sVBUCKET_DIM_minus1);
	    }
	    else {
	      lhssvbucketindex = 0;
	    }
	    
	    while (distancetooverlap >= distancetoendofvbucket &&
		   lhssvbucketptr->nextvbucket != NULL) {
	      lhssvbucketptr = lhssvbucketptr->nextvbucket;
	      distancetooverlap -= distancetoendofvbucket;
	      distancetoendofvbucket = sVBUCKET_DIM;
	    }
	    
	    lhssibucketptr->index_RANGE_END[lastlhssibucketindex] =
	      max(j2, lhssibucketptr->index_RANGE_END[lastlhssibucketindex]);
	    
	    if ((totalcurrentnonzerorange = h2-h1+1) > 0) {
	      
	      
	      sOVERLAPPING_CS_RANGE_AXPY(&lhssvbucketindex,
					k,
					&rhssvbucketindex[k],
					alpha[k],
					totalcurrentnonzerorange,
					&unconsumedcurrentnonzerorange,
					rhssvectorconsumednonzeroranges[k],
					rhssvectornonzeroranges[k],
					lhssvectornonzeroranges);
	
	    }
	      
	    /*
	     * This bit of logic must be performed, whether or not there
	     * is additional nonzeroes to process.
	     */
	    
	    if (j2 != -1 &&
		(lhssvbucketptr != templhssvbucketptr ||
		 (lhssvbucketindex = 
		  MY_MOD(lhssvbucketindex,sVBUCKET_DIM_minus1)) !=
		 templhssvbucketindex)) {
	      
#ifdef _DEBUG
	      printf("sspgx err 50.0: Unexpectd value discrepancies\n");
	      printf("lhssvbucketptr=%d, templhssvbucketptr=%d\n",
		     (int)lhssvbucketptr,(int)templhssvbucketptr);
	      printf("lhssvbucketindex=%d, templhssvbucketindex=%d\n",
		     lhssvbucketindex,templhssvbucketindex);
	      
	      abort();
#endif
	    }
	    else {
	      lhssvbucketptr = templhssvbucketptr;
	      lhssvbucketindex = templhssvbucketindex;
	    }
	    
	    
	    if ((totalcurrentnonzerorange = j2-j1+1) > 0) {
	      lhssvectornonzeroes += totalcurrentnonzerorange;
	      
	      /*
	       * adding new entries to vbucket
	       */
	      
	      if (lhssvbucketindex == 0 &&
		  lhssvectornonzeroranges != 0 &&
		  j2 != -1)
		sget_and_hop_to_next_vbucket(&lhssvbucketindex);
	      
	      if(rhssvbucketindex[k] == sVBUCKET_DIM && j2 != -1) {
		
		sHOP_TO_NEXT_RHSVBUCKET(k,&rhssvbucketindex[k],
				       rhssvectorconsumednonzeroranges[k],
				       rhssvectornonzeroranges[k]);
	      }

	      sNON_OVER_LAPPING_CS_RANGE_AXPY(NEW_ENTRIES_PADDED_TO_OLD_RANGE,
					     lhssibucketindex,
					     &lastlhssibucketindex,
					     &lhssvbucketindex,
					     k,
					     &rhssvbucketindex[k],
					     alpha[k],
					     totalcurrentnonzerorange,
					     &unconsumedcurrentnonzerorange,
					     rhssvectorconsumednonzeroranges[k],
					     rhssvectornonzeroranges[k],
					     lhssvectornonzeroranges);
	      
	    } /*  corresponds to the if ((totalcurrent...) { */
	    
	  } /*B.1*/
	  
	  
	  /****************************************	
	   * Perform saxpy:                       *
	   * 3.3. Found wrong subscript for       *
	   *      lhssvector.  Abort.               *
	   ****************************************/
	  
#ifdef _DEBUG
	  else {
	    printf("sspgx error 60: lhssvectornonzeroranges=%d,",
		   lhssvectornonzeroranges);
	    printf("lastlhssibucketindex=%d, \n\t",
		   lastlhssibucketindex);
	    printf("lhssibucketptr->index[%d] = (%d,%d), j1=%d\n",
		   lastlhssibucketindex,
		   lhssibucketptr->index_RANGE_BEGIN[lastlhssibucketindex],
		   lhssibucketptr->index_RANGE_END[lastlhssibucketindex],j1);
	    abort();
	  }
#endif
	  
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.4. If there is one, get new        *
	   *      nonzero from the rhs_vector     *
	   *      whose nonzero was just          *
	   *      consumed and replace the        *
	   *      heaproot with new entry;        *
	   *      Else, delete heaproot.          *
	   ****************************************/
	  if (++rhssvectorconsumednonzeroranges[k] <
	      rhssvectornonzeroranges[k]) {
	    if ((rhssibucketindex[k] =
		 MY_MOD(rhssvectorconsumednonzeroranges[k],
			sIBUCKET_DIM_minus1)) == 0) {
	      if ((rhssibucketptr[k] = rhssibucketptr[k]->nextibucket) ==
		  NULL) {
		
#ifdef _DEBUG
		printf("sspgx err 61: unexpected NULL nextvbucket: \n");
		printf("\t k=%d, rhssvectorconsumednonzeroranges=%d,\n",
		       k,rhssvectorconsumednonzeroranges[k]);
		printf("\t rhssvectornonzeroranges[k]=%d\n",
		       rhssvectornonzeroranges[k]);
		abort();
#endif
		
	      }
	    }     
	    j1 = rhssibucketptr[k]->index_RANGE_BEGIN[rhssibucketindex[k]];
	    j2 = rhssibucketptr[k]->index_RANGE_END[rhssibucketindex[k]];
	    
	    if (j1 <= 0 || j2 <= 0) {
#ifdef _DEBUG
	      printf("sspgx error 62: Illegal index: vector=%d, ",k);
	      printf("subscript1=%d, subscript2=%d \n",j1,j2);
	      abort();
#endif
	    }
	    
	    else
	      heapop(REPLACE,k,j1,j2);
	  }
	  else
	    heapop(DELETE,DUMMYVECTOR,DUMMYSUBSCRIPT,DUMMYSUBSCRIPT);
	} /*matches     while (heaproot !=NULL) {     */
      } /* matches  else ** if (governingdstype == CS) ** {     */	
      
      saxpy_2_heap_post_processing(destptrptr,
				  rhssvectorptrptrs,
				  alphaptrs,
				  restorerhssvectorptr,
				  arity,
				  lhssvectornonzeroes,
				  lhssvectornonzeroranges,
				  governingdstype);
      
      
    } /* matches    else ** if we are going to use lhssvectorptr ** {  */

#ifdef _PLUSEQUAL
  } /* matches    else ** if (not plusequalsaxpy_flag) ** {    */
#endif /* _PLUSEQUAL */

  /* }  matches    else ** if (not nullsaxpyflag) ** {    */

}

