#define SINGLE
/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#include  <stdlib.h>
#include "sgaxpy_pd_macros_and_data_structures.h"
#include "sgaxpy_pd_global_variables.h"

/****************************************************************************
 *                                                                          *
 *    plusequalsspg2                                                       *
 *                                                                          *
 ****************************************************************************/

void plusequalsspg2(rhsanddestptrptr,
		     otheralphaptr,
                     otherrhssvectorptrptr,
                     status_otherrhssvector)
svectorstruct **rhsanddestptrptr;
float *otheralphaptr;
svectorstruct **otherrhssvectorptrptr;
int status_otherrhssvector;
{
  float *alphaptrs[2];
  
  int arity = 2;
  
  svectorstruct *rhssvectorptr[2];
  svectorstruct *tempvectorptr;
  
  int i;
  

  sibucketstruct *rhssibucketptr[2];
                                      /* Array of pointers to "current"     *
				       * buckets where "current" nonzeroes  *
				       * of the rhs_vectors reside.         */

  sibucketstruct *lhssibucketptr;
                                      /* Pointer to current bucket of       *
				       * lhssvector.                    */

  float alpha[2];          /* Dereferenced alphaptrs[].          */

  int rhssibucketindex[2];        /* Array of indices of current        *
				       * nonzeroes in the current buckets   *
				       * of rhs_vectors.                    */

  int lhssibucketindex;           /* bucketindex of next nonzero to be  *
				       * added to lhssvector.           */

  int lastlhssibucketindex;       /* bucketindex of the latest added    *
				       * nonzero in the lhssvector.  We *
                                       * need this additional index to      *
				       * leap-frog buckets, if need be.     */

  int k;                              /* Counter for arity.                 */

  int lhssvectornonzeroes;        /* Count of # of nonzeroes entered    *
				       * into the lhssvector so far. By *
				       * the end of the saxpy operation,    *
				       * this index equals the total # of   *
				       * nonzeroes in the resultant vector. */

/*
 * These variables were originally for COMPRESSED_SUBSCRIPTS
 */

  svbucketstruct *lastlhssvbucketptr;  
                                      /* Pointer to current vbucket of      *
				       * lhssvector.                    */

  int rhssvectornonzeroranges[2];
                                      /* Array of total # of nonzeroes      *
				       * ranges in each rhs_vector.         */

  int rhssvectorconsumednonzeroranges[2];
                                      /* Array of # of nonzeroes consumed   *
				       * so far in each rhs_vector.         */

  int rhssvbucketindex[2];        /* Array of indices of current        *
				       * nonzeroes in the current buckets   *
				       * of rhs_vectors.                    */

  int lhssvectornonzeroranges;    /* Count of # of nonzeroes entered    *
				       * into the lhssvector so far. By *
				       * the end of the saxpy operation,    *
				       * this index equals the total # of   *
				       * nonzeroes in the resultant vector. */

  int lhssvbucketindex;           /* bucketindex of next nonzero to be  *
				       * added to lhssvector.           */


  int lastlhssvbucketindex;       /* bucketindex of the latest added    *
				       * nonzero in the lhssvector.  We *
				       * need this additional index to frog *
				       * leap-frog buckets, if need be.     */

  svbucketstruct *templhssvbucketptr;

  int templhssvbucketindex;

  int j1;                             /* Index for vectorsubscripts to be   *
				       * entered in (and later popped from) *
				       * the heap.                          */

  int j2;                             /* Index for vectorsubscripts to be   *
				       * entered in (and later popped from) *
				       * the heap.                          */

  int h1;
  int h2;

  int totalcurrentnonzerorange;

#ifdef _DEBUG
  int unconsumedcurrentnonzerorange;
#endif

  int distancetooverlap;
  int distancetoendofvbucket;
  int hopdistance;

  /*
   * These variables were added for union ds
   */
  
  svectorstruct *restorerhssvectorptr[2+1];

  int governingdstype;
  
  /*
   * These variables were added for plusequal code
   */
  
  svectorstruct **rhssvectorptrptrs[MAXARITY];

  int condition_1, condition_2;
  int lastrhssibucketindex[2];
  int lastrhssvbucketindex[2];

  sibucketstruct *lastrhssibucketptr[2];
  svbucketstruct *lastrhssvbucketptr[2];
  svbucketstruct *catchupto_rhsvbucketptr[2];

  sibucketstruct *temprhssibucketptr[2];
  svbucketstruct *temprhssvbucketptr[2];
  int temprhssvbucketindex[2];

  sibucketstruct *restofplusequalibucketptr;
  svbucketstruct *restofplusequalvbucketptr;
  int restofplusequalibucketindex;
  int restofplusequalvbucketindex;
  int restofplusequalvectorconsumednonzeroranges;

#ifdef _TRACE
  int nnzb,nnza;
  svectorstruct *lhssvectoraddress;
  
  svectorstruct **vectorptrptrs[2];
  
  alphaptrs[0] = alphaptr;
  vectorptrptrs[0] = vectoraptrptr;
  
  alphaptrs[1] = betaptr;
  vectorptrptrs[1] = vectorbptrptr;
  
  if(*destptrptr==NULL){ 
    nnzb = 0; 
    lhssvectoraddress= (svectorstruct *) 99900; 
  }
  else {
    nnzb = (*destptrptr)->nonzeroes; 
    lhssvectoraddress = *destptrptr;
  }
#endif
  
  /****************************************
   * Dereferencing.                       *
   ****************************************/
  
  X_PUT_REAL(alpha[0],1.0);

#ifdef _DEBUG
  if (*rhsanddestptrptr == NULL) {
    printf("plusequalsspg2 error #1: Non_initialized lhssvector.\n");
    abort();
  }
#endif

  rhssvectorptr[0] = *rhsanddestptrptr;
  rhssvectornonzeroranges[0] =
    rhssvectorptr[0]->sds.cs.nonzeroranges;

  X_ASSIGN(alpha[1],(*otheralphaptr));

  alphaptrs[0] = &alpha[0];
  alphaptrs[1] = otheralphaptr;

  rhssvectorptr[1] = *otherrhssvectorptrptr;
  rhssvectornonzeroranges[1] =
    rhssvectorptr[1]->sds.cs.nonzeroranges;

  governingdstype = CS;
  k = 1;   
  restorerhssvectorptr[k] = rhssvectorptr[k];

  if (rhssvectorptr[k]->dstype != governingdstype) {
    sconvertdstype((&rhssvectorptr[k]),k,governingdstype);
    rhssvectornonzeroranges[k] =
      rhssvectorptr[k]->sds.cs.nonzeroranges;
  }  
 
  /******************************************************************
   *               _COMPRESSED_SUBSCRIPTS AXPY ONLY                 *
   ******************************************************************/
  
  /****************************************
   * perform saxpy:                       *
   * 1. Put in heap the 1st nonzero of    *
   *    each vector.                      *
   ****************************************/
  for (k=0;k<arity;k++) {
    rhssvectorconsumednonzeroranges[k] = 0;
    rhssibucketindex[k] = 0;
    rhssvbucketindex[k] = 0;
    rhssibucketptr[k] =
      rhssvectorptr[k]->sds.cs.ibucketptr;
    rhssvbucketptr[k] =
      rhssvectorptr[k]->sds.cs.vbucketptr;
    
#ifdef _DEBUG
    if (rhssibucketptr[k] == NULL ||
	rhssvbucketptr[k] == NULL) {
	      printf("plusequalsspg2 err 10: Unexpected NULL rhssbucketptr: ");
	      printf("vector # %d\n",k);
	      abort();
	    }
#endif
  }	  

  
  /****************************************
   * Perform saxpy:                       *
   * 2. Initialize lhssvector.             *
   ****************************************/

  lhssvectornonzeroranges = 0;
  lhssvectornonzeroes = 0;
  lhssvbucketindex = 0;


  /****************************************
   * Perform saxpy:                       *
   * 3. Get nonzero by popping the        *
   *    heaproot; consume it in saxpy;    *
   *    either replace or delete heaproot.*
   ****************************************/
  

  /* while (there are nonzeroranges to be processed AND                      *
   *         (condition_1 OR condition_2 ), where:                           *
   * condition_1 == TRUE, indicates that the range to be processed belongs   *
   *                to the plusequalvector,                                  *
   * condition_2 == TRUE, indicates that the range to be processed belongs   *
   *                to the otherrhssvector, and that this range is       *
   *                entirelyencapsulated in the last processed range of the  *
   *                plusequalvector.                                         */

  lastrhssibucketindex[0] = -1;
  lastrhssibucketptr[0] = NULL;
  
  while ((rhssvectorconsumednonzeroranges[0] <
	  rhssvectornonzeroranges[0] &&
	  rhssvectorconsumednonzeroranges[1] <
	  rhssvectornonzeroranges[1]) &&

	 ((condition_1 = (rhssibucketptr[0]->
			  index_RANGE_BEGIN[rhssibucketindex[0]] <=
			  rhssibucketptr[1]->
			  index_RANGE_BEGIN[rhssibucketindex[1]]))    ||
	  
	  (condition_2 = ((lastrhssibucketindex[0] >= 0) &&
			  (rhssibucketptr[1]->
			   index_RANGE_BEGIN[rhssibucketindex[1]] >=
			   lastrhssibucketptr[0]->
			   index_RANGE_BEGIN[lastrhssibucketindex[0]] &&
			   rhssibucketptr[1]->
			   index_RANGE_END[rhssibucketindex[1]] <=
			   lastrhssibucketptr[0]->
			   index_RANGE_END[lastrhssibucketindex[0]]))))) {

    if (condition_1) {
      k = 0;
      lastrhssibucketindex[0] = rhssibucketindex[0];
      lastrhssvbucketindex[0] = rhssvbucketindex[0];
      lastrhssibucketptr[0] = rhssibucketptr[0];
      lastrhssvbucketptr[0] = rhssvbucketptr[0];
      totalcurrentnonzerorange =
	rhssibucketptr[0]->index_RANGE_END[rhssibucketindex[0]] -
	  rhssibucketptr[0]->index_RANGE_BEGIN[rhssibucketindex[0]]+1;

#ifdef _DEBUG
      unconsumedcurrentnonzerorange = totalcurrentnonzerorange;
#endif

      hopdistance = totalcurrentnonzerorange;
      distancetoendofvbucket = sVBUCKET_DIM - rhssvbucketindex[0];
   
      if (distancetoendofvbucket > hopdistance) {
	rhssvbucketindex[0] = MY_MOD(rhssvbucketindex[0] + hopdistance,
					 sVBUCKET_DIM_minus1);
      }
      else if (distancetoendofvbucket < hopdistance) {
	rhssvbucketindex[0] = MY_MOD(hopdistance - distancetoendofvbucket,
					 sVBUCKET_DIM_minus1);
      }
      else {
	rhssvbucketindex[0] = 0;
      }
      while (hopdistance >= distancetoendofvbucket &&
	     rhssvbucketptr[0]->nextvbucket != NULL) {
	rhssvbucketptr[0] = rhssvbucketptr[0]->nextvbucket;
	hopdistance -= distancetoendofvbucket;
	distancetoendofvbucket = sVBUCKET_DIM;
      }

      lhssvectornonzeroes += totalcurrentnonzerorange;
    }
     
    else if (condition_2) {
      k = 1;
     temprhssibucketptr[0] = rhssibucketptr[0];
      temprhssvbucketptr[0] = rhssvbucketptr[0];
      rhssibucketptr[0] = lastrhssibucketptr[0];
      rhssvbucketptr[0] = lastrhssvbucketptr[0];

      /* The reason rhssvbucketindex[0] is not
       * set to lastrhssvbucketindex[0] is because
       * rhssvbucketindex[0] is computed below from
       * lastrhssvbucketindex[0].
       */

      temprhssvbucketindex[0] = rhssvbucketindex[0];
       
      distancetooverlap =
	rhssibucketptr[1]->index_RANGE_BEGIN[rhssibucketindex[1]] -
	rhssibucketptr[0]->index_RANGE_BEGIN[lastrhssibucketindex[0]];
      distancetoendofvbucket = sVBUCKET_DIM - lastrhssvbucketindex[0];
      
      if (distancetoendofvbucket > distancetooverlap) {
	rhssvbucketindex[0] =
	  MY_MOD(lastrhssvbucketindex[0] + distancetooverlap,
		 sVBUCKET_DIM_minus1);
      }
      else if (distancetoendofvbucket < distancetooverlap) {
	rhssvbucketindex[0] =
	  MY_MOD(distancetooverlap - distancetoendofvbucket,
		 sVBUCKET_DIM_minus1);
      }
      else {
	rhssvbucketindex[0] = 0;
      }
      
      while (distancetooverlap >= distancetoendofvbucket &&
	     rhssvbucketptr[0]->nextvbucket != NULL) {
	rhssvbucketptr[0] = rhssvbucketptr[0]->nextvbucket;
	distancetooverlap -= distancetoendofvbucket;
	distancetoendofvbucket = sVBUCKET_DIM;
      }

      totalcurrentnonzerorange = 
      	rhssibucketptr[1]->index_RANGE_END[rhssibucketindex[1]] -
	  rhssibucketptr[1]->index_RANGE_BEGIN[rhssibucketindex[1]]+1;

      /* Wrapping the call with the two assignments has the effect of
       * EQUIVALENCEing the data structures being pointed two by the
       * two pointers.
       */
    
      lhssvbucketptr = rhssvbucketptr[0];      
      sOVERLAPPING_CS_RANGE_AXPY(&rhssvbucketindex[0],
				1,
				&rhssvbucketindex[1],
				alpha[1],
				totalcurrentnonzerorange,
				&unconsumedcurrentnonzerorange,
				rhssvectorconsumednonzeroranges[1],
				rhssvectornonzeroranges[1],
				rhssvectornonzeroranges[0]);
      rhssvbucketptr[0] = lhssvbucketptr;
      
      rhssibucketptr[0] = temprhssibucketptr[0];
      rhssvbucketptr[0] = temprhssvbucketptr[0];
      rhssvbucketindex[0] = temprhssvbucketindex[0];
      
    } /* matches     else if (condition_2) {  */
  
    /*****************************************
     * Perform saxpy:                        *
     * 3.4. If there is one, get new         *
     *      nonzero from the rhs_vector      *
     *      whose nonzero was just consumed. *
     *****************************************/
    if (++rhssvectorconsumednonzeroranges[k] <
	rhssvectornonzeroranges[k]) {
      if ((rhssibucketindex[k] =
	   MY_MOD(rhssvectorconsumednonzeroranges[k],
		  sIBUCKET_DIM_minus1)) == 0) {
	if ((rhssibucketptr[k] = rhssibucketptr[k]->nextibucket) ==
	    NULL) {
	  
#ifdef _DEBUG
	  printf("plusqualsspg2 error #36: unexpected NULL nextvbucket: \n");
	  printf("\t k=%d, rhssvectorconsumednonzeroranges=%d,\n",
		 k,rhssvectorconsumednonzeroranges[k]);
	  printf("\t rhssvectornonzeroranges[k]=%d\n",
		 rhssvectornonzeroranges[k]);
	  abort();
#endif
	  
	}
      }
    }   

  } /* matches     1st while loop           */

  if (rhssvectorconsumednonzeroranges[1] ==
      rhssvectornonzeroranges[1]) {
    while (rhssvectorconsumednonzeroranges[0] <
	   rhssvectornonzeroranges[0]) {
      totalcurrentnonzerorange =
	rhssibucketptr[0]->index_RANGE_END[rhssibucketindex[0]] -
	  rhssibucketptr[0]->index_RANGE_BEGIN[rhssibucketindex[0]]+1;
      lhssvectornonzeroes += totalcurrentnonzerorange;

      if (++rhssvectorconsumednonzeroranges[0] <
	  rhssvectornonzeroranges[0]) {
	if ((rhssibucketindex[0] =
	     MY_MOD(rhssvectorconsumednonzeroranges[0],
		    sIBUCKET_DIM_minus1)) == 0) {
	  if ((rhssibucketptr[0] = rhssibucketptr[0]->nextibucket) ==
	      NULL) {
	    
#ifdef _DEBUG
	    printf("plusqualsspg2 err #58: unexpected NULL nextvbucket: \n");
	    printf("\t k=%d, rhssvectorconsumednonzeroranges=%d,\n",
		   0,rhssvectorconsumednonzeroranges[0]);
	    printf("\t rhssvectornonzeroranges[0]=%d\n",
		   rhssvectornonzeroranges[0]);
	    abort();
#endif
	    
	  }
	}
      }   
    }
   }

  else if (rhssvectorconsumednonzeroranges[0] ==
	   rhssvectornonzeroranges[0]) {
    while((rhssvectorconsumednonzeroranges[1] <
	   rhssvectornonzeroranges[1]) &&
	  ((lastrhssibucketindex[0] >= 0) &&
	   (rhssibucketptr[1]->
	    index_RANGE_BEGIN[rhssibucketindex[1]] >=
	    lastrhssibucketptr[0]->
	    index_RANGE_BEGIN[lastrhssibucketindex[0]] &&
	    rhssibucketptr[1]->
	    index_RANGE_END[rhssibucketindex[1]] <=
	    lastrhssibucketptr[0]->
	    index_RANGE_END[lastrhssibucketindex[0]]))) {
      
      /* Note: This part of the code is almost a copy of the part above
       * for the case where condition_2 is TRUE.  Therefore, when
       * modifying either code segment, care must be taken so that
       * both parts remain coherent.
       */

      k = 1;
      temprhssibucketptr[0] = rhssibucketptr[0];
      temprhssvbucketptr[0] = rhssvbucketptr[0];
      rhssibucketptr[0] = lastrhssibucketptr[0];
      rhssvbucketptr[0] = lastrhssvbucketptr[0];

      temprhssvbucketindex[0] = rhssvbucketindex[0];
      
      distancetooverlap =
	rhssibucketptr[1]->index_RANGE_BEGIN[rhssibucketindex[1]] -
	rhssibucketptr[0]->index_RANGE_BEGIN[lastrhssibucketindex[0]];
      distancetoendofvbucket = sVBUCKET_DIM - lastrhssvbucketindex[0];
      
      if (distancetoendofvbucket > distancetooverlap) {
	rhssvbucketindex[0] =
	  MY_MOD(lastrhssvbucketindex[0] + distancetooverlap,
		 sVBUCKET_DIM_minus1);
      }
      else if (distancetoendofvbucket < distancetooverlap) {
	rhssvbucketindex[0] =
	  MY_MOD(distancetooverlap - distancetoendofvbucket,
		 sVBUCKET_DIM_minus1);
      }
      else {
	rhssvbucketindex[0] = 0;
      }
      
      while (distancetooverlap >= distancetoendofvbucket &&
	     rhssvbucketptr[0]->nextvbucket != NULL) {
	rhssvbucketptr[0] = rhssvbucketptr[0]->nextvbucket;
	distancetooverlap -= distancetoendofvbucket;
	distancetoendofvbucket = sVBUCKET_DIM;
      }
      
      totalcurrentnonzerorange = 
      	rhssibucketptr[1]->index_RANGE_END[rhssibucketindex[1]] -
	  rhssibucketptr[1]->index_RANGE_BEGIN[rhssibucketindex[1]]+1;
      
      /* Wrapping the call with the two assignments has the effect of
       * EQUIVALENCEing the data structures being pointed two by the
       * two pointers.
       */
      
      lhssvbucketptr = rhssvbucketptr[0];      
      sOVERLAPPING_CS_RANGE_AXPY(&rhssvbucketindex[0],
 				1,
				&rhssvbucketindex[1],
				alpha[1],
				totalcurrentnonzerorange,
				&unconsumedcurrentnonzerorange,
				rhssvectorconsumednonzeroranges[1],
				rhssvectornonzeroranges[1],
				rhssvectornonzeroranges[0]);
      rhssvbucketptr[0] = lhssvbucketptr;
      
      rhssibucketptr[0] = temprhssibucketptr[0];
      rhssvbucketptr[0] = temprhssvbucketptr[0];
      rhssvbucketindex[0] = temprhssvbucketindex[0];
      
      /*****************************************
       * Perform saxpy:                        *
       * 3.4. If there is one, get new         *
       *      nonzero from the rhs_vector      *
       *      whose nonzero was just consumed. *
       *****************************************/

      if (++rhssvectorconsumednonzeroranges[k] <
	  rhssvectornonzeroranges[k]) {
	if ((rhssibucketindex[k] =
	     MY_MOD(rhssvectorconsumednonzeroranges[k],
		    sIBUCKET_DIM_minus1)) == 0) {
	  if ((rhssibucketptr[k] = rhssibucketptr[k]->nextibucket) ==
	      NULL) {
	    
#ifdef _DEBUG
	    printf("plusqualsspg2 err #75: unexpected NULL nextvbucket: \n");
	    printf("\t k=%d, rhssvectorconsumednonzeroranges=%d,\n",
		   k,rhssvectorconsumednonzeroranges[k]);
	    printf("\t rhssvectornonzeroranges[k]=%d\n",
		   rhssvectornonzeroranges[k]);
	    abort();
#endif
	    
	  }
	}
      }   
    }
    
  }

  if (rhssvectorconsumednonzeroranges[0] <
	 rhssvectornonzeroranges[0] ||
	 rhssvectorconsumednonzeroranges[1] <
	 rhssvectornonzeroranges[1]) {

    restofplusequalsvectorptr->nonzeroes = 0;
    restofplusequalsvectorptr->sds.cs.nonzeroranges =
      rhssvectornonzeroranges[0] - rhssvectorconsumednonzeroranges[0];
    restofplusequalvectorconsumednonzeroranges = 0;
      
    if (rhssvectorconsumednonzeroranges[0] <
	rhssvectornonzeroranges[0]) {

      if (rhssibucketindex[0] == 0 &&
	  rhssibucketptr[0] != lastrhssibucketptr[0]) {
	restofplusequalibucketptr = rhssibucketptr[0];
	restofplusequalibucketindex = rhssibucketindex[0];
	restofplusequalsvectorptr->sds.cs.ibucketptr =
	  restofplusequalibucketptr;
	rhssibucketptr[0] = lastrhssibucketptr[0];
      }
      else {
	sgetibucket(&restofplusequalibucketptr);
	restofplusequalsvectorptr->sds.cs.ibucketptr =
	  restofplusequalibucketptr;
	
	/* Note in the 'for' statement below, the remaining
	 * (not-yet-consumed-in-the saxpy) entries of rhssibucketptr[0]
	 * are placed into restofplusequalibucketptr.
	 */
	
	for (i=rhssibucketindex[0];i<sIBUCKET_DIM;i++) {
	  restofplusequalibucketptr->index_RANGE_BEGIN[i] =
	    rhssibucketptr[0]->index_RANGE_BEGIN[i];
	  restofplusequalibucketptr->index_RANGE_END[i] =
	    rhssibucketptr[0]->index_RANGE_END[i];
	}
	restofplusequalibucketindex = rhssibucketindex[0];
	restofplusequalibucketptr->nextibucket =
	  rhssibucketptr[0]->nextibucket;
	
	/* As distinct from the 'if' part above, there's no need here to set
	 * the values for rhssibucketptr[0] and rhssibucketindex[0],
	 * since they're already pointing where they should be pointing.
	 */
	
      }
      rhssibucketptr[0]->nextibucket = NULL;
      
      if (rhssvbucketindex[0] == 0 &&
	  rhssvbucketptr[0] != lastrhssvbucketptr[0]) {
	restofplusequalvbucketptr = rhssvbucketptr[0];
	restofplusequalvbucketindex = rhssvbucketindex[0];
	restofplusequalsvectorptr->sds.cs.vbucketptr =
	  restofplusequalvbucketptr;
	
	catchupto_rhsvbucketptr[0] = lastrhssvbucketptr[0];
	while (catchupto_rhsvbucketptr[0]->nextvbucket !=
	       rhssvbucketptr[0]) {
  
#ifdef _DEBUG
	  if (catchupto_rhsvbucketptr[0]->nextvbucket == NULL) {
	    printf("plusqualsspg2 error 18: unexpected NULL nextvbucket: \n");
	    abort();
	  }
#endif
	  
	  catchupto_rhsvbucketptr[0] = catchupto_rhsvbucketptr[0]->nextvbucket;
	}
	rhssvbucketptr[0] = catchupto_rhsvbucketptr[0];
	
	rhssvbucketindex[0] = sVBUCKET_DIM;
      }
      else {
	sgetvbucket(&restofplusequalvbucketptr);
	restofplusequalsvectorptr->sds.cs.vbucketptr =
	  restofplusequalvbucketptr;
	
	for (i=rhssvbucketindex[0];i<sVBUCKET_DIM;i++) {
	  restofplusequalvbucketptr->value[i] =
	    rhssvbucketptr[0]->value[i];
	}
	restofplusequalvbucketindex = rhssvbucketindex[0];
	restofplusequalvbucketptr->nextvbucket =
	  rhssvbucketptr[0]->nextvbucket;
      }
      rhssvbucketptr[0]->nextvbucket = NULL;
    }
    else {
      restofplusequalsvectorptr->sds.cs.ibucketptr = NULL;
      restofplusequalsvectorptr->sds.cs.vbucketptr = NULL;
    }
	  
    /* Switching vector names in order not to have to change names
     * in the code that follows.  Now, lhssvector will the new name
     * for the plusequalvector(rhssvector[0]), and rhssvector[0] will
     * be the new name for the restofplusequalvector.
     */
    
    tempvectorptr = lhssvectorptr;
    
    lhssvectorptr = rhssvectorptr[0];
    
    lhssvectornonzeroranges = rhssvectorconsumednonzeroranges[0];
    
    lhssibucketptr = rhssibucketptr[0];
    lhssvbucketptr = rhssvbucketptr[0];
    lhssvbucketindex = rhssvbucketindex[0];
    
    /*
     * lhssibucketindex is computed later using
     * lhssvectornonzeroranges.
     */
    
    lastlhssibucketindex = lastrhssibucketindex[0];
    lastlhssvbucketindex = lastrhssvbucketindex[0];
    lastlhssvbucketptr = lastrhssvbucketptr[0];
    
    /*
     * We never make use of lastlhssibucketptr.
     */

    if (rhssvectorconsumednonzeroranges[0] <
	rhssvectornonzeroranges[0]) {

      rhssvectorptr[0] = restofplusequalsvectorptr;
      rhssibucketptr[0] = rhssvectorptr[0]->sds.cs.ibucketptr;
      rhssvbucketptr[0] = rhssvectorptr[0]->sds.cs.vbucketptr;
      rhssibucketindex[0] = restofplusequalibucketindex;
      rhssvbucketindex[0] = restofplusequalvbucketindex;
      
    }

    while (rhssvectorconsumednonzeroranges[0] <
	   rhssvectornonzeroranges[0] ||
	   rhssvectorconsumednonzeroranges[1] <
	   rhssvectornonzeroranges[1]) {
      
      if
	((rhssvectorconsumednonzeroranges[1] ==
	  rhssvectornonzeroranges[1]) ||
	 ((rhssvectorconsumednonzeroranges[0] <
	   rhssvectornonzeroranges[0]) &&
	  (rhssibucketptr[0]->
	   index_RANGE_BEGIN[rhssibucketindex[0]] <=
	   rhssibucketptr[1]->
	   index_RANGE_BEGIN[rhssibucketindex[1]])))
	  k = 0;
      else if
	((rhssvectorconsumednonzeroranges[0] ==
	  rhssvectornonzeroranges[0]) ||
	 ((rhssvectorconsumednonzeroranges[1] <
	   rhssvectornonzeroranges[1]) &&
	  (rhssibucketptr[0]->
	   index_RANGE_BEGIN[rhssibucketindex[0]] >
	   rhssibucketptr[1]->
	   index_RANGE_BEGIN[rhssibucketindex[1]])))
	  k = 1;
#ifdef _DEBUG
      else {
	printf("plusequalsspg2 error #11: BAD LOGIC.\n");
	abort();
      }
#endif
      j1 = rhssibucketptr[k]->index_RANGE_BEGIN[rhssibucketindex[k]];
      j2 = rhssibucketptr[k]->index_RANGE_END[rhssibucketindex[k]];
      rhssibucketindex[k] =
	MY_MOD(rhssvectorconsumednonzeroranges[k],sIBUCKET_DIM_minus1);
      lhssibucketindex =
	MY_MOD(lhssvectornonzeroranges,sIBUCKET_DIM_minus1);
      
      if ((rhssvbucketindex[k] =
	  MY_MOD(rhssvbucketindex[k],sVBUCKET_DIM_minus1))==0
	  && rhssvectorconsumednonzeroranges[k] !=0
	  && (k !=0 || restofplusequalvectorconsumednonzeroranges !=0)) {
	sHOP_TO_NEXT_RHSVBUCKET(k,&rhssvbucketindex[k],
			       rhssvectorconsumednonzeroranges[k],
			       rhssvectornonzeroranges[k]);
      }
      lhssvbucketindex =
	MY_MOD(lhssvbucketindex,sVBUCKET_DIM_minus1);
      
#ifdef _DEBUG
      if (rhssibucketptr[k]->errorflag != 9099 ||
	  rhssvbucketptr[k]->errorflag != 9009) {
	printf("plusequalsspg2 err 14: memory corruption of ibucketstore ");
	printf("or vbucketstore.\n");
	abort();
      }
#endif      
      
      /****************************************
       * Perform saxpy:                       *
       * 3.0. Check for startup errors.       *
       ****************************************/
      
#ifdef _DEBUG
      if (lhssvectornonzeroranges == 0) {
	if (rhssibucketindex[k] != 0 || rhssvbucketindex[k] != 0 ||
	    lhssibucketindex != 0 || lhssvbucketindex != 0) {
	  printf("plusequalsspg2 error 15: Zeroes expected: \n\t");
	  printf("rhssibucketindex[k]=%d, rhssvbucketindex[k]=%d\n\t",
		 rhssibucketindex[k], rhssvbucketindex[k]);
	  printf("lhssibucketindex=%d, lhssvbucketindex=%d\n\t",
		 lhssibucketindex,lhssvbucketindex);
	  abort();
	}
      }
#endif      
      
      /****************************************
       * Perform saxpy:                       *
       * 3.1. Found 1ST or NEW subscript      *
       *      range (non-overlapping with     *
       *      previous one, if any) for       *
       *      lhssvector.  If needed, get new  *
       *      ibuckets and vbuckets for       *
       *      lhssvector.                      *
       ****************************************/
      if (lhssvectornonzeroranges == 0 ||
	  j1 > lhssibucketptr->index_RANGE_END[lastlhssibucketindex]+1) {/*A.1*/
	
	if (lhssibucketindex == 0 && lhssvectornonzeroranges != 0) {
	  sgetibucket(&(lhssibucketptr->nextibucket));
	  lhssibucketptr = lhssibucketptr->nextibucket;
	}
	lhssibucketptr->index_RANGE_BEGIN[lhssibucketindex] = j1;
	lhssibucketptr->index_RANGE_END[lhssibucketindex] = j2;
	
	if (lhssvbucketindex == 0 && lhssvectornonzeroranges != 0)
	  sget_and_hop_to_next_vbucket(&lhssvbucketindex);
	
	lastlhssvbucketptr = lhssvbucketptr;
	lastlhssvbucketindex = lhssvbucketindex;
	
	totalcurrentnonzerorange = j2-j1+1;
	
	lhssvectornonzeroes += totalcurrentnonzerorange;
	lhssvectornonzeroranges++;
	
	
	sNON_OVER_LAPPING_CS_RANGE_AXPY(COMPLETELY_NEW_RANGE,
				       lhssibucketindex,
				       &lastlhssibucketindex,
				       &lhssvbucketindex,
				       k,
				       &rhssvbucketindex[k],
				       alpha[k],
				       totalcurrentnonzerorange,
				       &unconsumedcurrentnonzerorange,
				       rhssvectorconsumednonzeroranges[k],
				       rhssvectornonzeroranges[k],
				       lhssvectornonzeroranges);
	
	
      } /*A.1*/
      
      
      /****************************************
       * Perform saxpy:                       *
       * 3.2. Found old subscript for         *
       *      lhssvector.  Consume  nonzero in *
       *      saxpy.                          *
       *                                      *
       ****************************************/
      else if (j1 <= 
	       lhssibucketptr->index_RANGE_END[lastlhssibucketindex]+1 &&
	       j1 >= 
	       lhssibucketptr->index_RANGE_BEGIN[lastlhssibucketindex]) { /*B.1*/
	
	templhssvbucketptr = lhssvbucketptr;
	templhssvbucketindex = lhssvbucketindex;
	lhssvbucketptr = lastlhssvbucketptr;
	
	distancetooverlap = j1 -
	  lhssibucketptr->index_RANGE_BEGIN[lastlhssibucketindex];
	distancetoendofvbucket = sVBUCKET_DIM -
	  lastlhssvbucketindex;
	
	if (j1 ==
	    lhssibucketptr->index_RANGE_END[lastlhssibucketindex]+1) {
	  h1 = 0;
	  h2 = -1;
	}
	else if (j2 <=
		 lhssibucketptr->index_RANGE_END[lastlhssibucketindex]) {
	  h1 = j1;
	  h2 = j2;
	  j1 = 0;
	  j2 = -1;
	}
	else {
	  h1 = j1;
	  h2 = lhssibucketptr->index_RANGE_END[lastlhssibucketindex];
	  j1 = h2+1;
	}
	
	if (distancetoendofvbucket > distancetooverlap) {
	  lhssvbucketindex =
	    MY_MOD(lastlhssvbucketindex + distancetooverlap,
		   sVBUCKET_DIM_minus1);
	}
	else if (distancetoendofvbucket < distancetooverlap) {
	  lhssvbucketindex =
	    MY_MOD(distancetooverlap - distancetoendofvbucket,
		   sVBUCKET_DIM_minus1);
	}
	else {
	  lhssvbucketindex = 0;
	}
	
	while (distancetooverlap >= distancetoendofvbucket &&
	       lhssvbucketptr->nextvbucket != NULL) {
	  lhssvbucketptr = lhssvbucketptr->nextvbucket;
	  distancetooverlap -= distancetoendofvbucket;
	  distancetoendofvbucket = sVBUCKET_DIM;
	}
	
	lhssibucketptr->index_RANGE_END[lastlhssibucketindex] =
	  max(j2,
	      lhssibucketptr->index_RANGE_END[lastlhssibucketindex]);
	
	if ((totalcurrentnonzerorange = h2-h1+1) > 0) {
	  
	  
	  sOVERLAPPING_CS_RANGE_AXPY(&lhssvbucketindex,
				    k,
				    &rhssvbucketindex[k],
				    alpha[k],
				    totalcurrentnonzerorange,
				    &unconsumedcurrentnonzerorange,
				    rhssvectorconsumednonzeroranges[k],
				    rhssvectornonzeroranges[k],
				    lhssvectornonzeroranges);
	  
	  
	}
	
	/*
	 * This bit of logic must be performed, whether or not there
	 * is additional nonzeroes to process.
	 */
	
	if (j2 != -1 &&
	    (lhssvbucketptr != templhssvbucketptr ||
	     (lhssvbucketindex =
	      MY_MOD(lhssvbucketindex,sVBUCKET_DIM_minus1)) !=
	     templhssvbucketindex)) {
	  
#ifdef _DEBUG
	  printf("plusequalsspg2 err 50.0: Unexpectd value discrepancies\n");
	  printf("lhssvbucketptr=%d, templhssvbucketptr=%d\n",
		 (int)lhssvbucketptr,(int)templhssvbucketptr);
	  printf("lhssvbucketindex=%d, templhssvbucketindex=%d\n",
		 lhssvbucketindex,templhssvbucketindex);
	  
	  abort();
#endif
	}
	else {
	  lhssvbucketptr = templhssvbucketptr;
	  lhssvbucketindex = templhssvbucketindex;
	}
	
	
	if ((totalcurrentnonzerorange = j2-j1+1) > 0) {
	  lhssvectornonzeroes += totalcurrentnonzerorange;
	  
	  /*
	   * adding new entries to vbucket
	   */
	  
	  if (lhssvbucketindex == 0 &&
	      lhssvectornonzeroranges != 0 &&
	      j2 != -1)
	    sget_and_hop_to_next_vbucket(&lhssvbucketindex);
	  
	  if(rhssvbucketindex[k] == sVBUCKET_DIM && j2 != -1) {
	    
	    sHOP_TO_NEXT_RHSVBUCKET(k,&rhssvbucketindex[k],
				   rhssvectorconsumednonzeroranges[k],
				   rhssvectornonzeroranges[k]);
	  }
	  
	  
	  
	  
	  sNON_OVER_LAPPING_CS_RANGE_AXPY(NEW_ENTRIES_PADDED_TO_OLD_RANGE,
					 lhssibucketindex,
					 &lastlhssibucketindex,
					 &lhssvbucketindex,
					 k,
					 &rhssvbucketindex[k],
					 alpha[k],
					 totalcurrentnonzerorange,
					 &unconsumedcurrentnonzerorange,
					 rhssvectorconsumednonzeroranges[k],
					 rhssvectornonzeroranges[k],
					 lhssvectornonzeroranges);
	  
	  
	} /* matches     if ((totalcurrent...) { */
      } /*B.1*/
      
      
      /****************************************	
       * Perform saxpy:                       *
       * 3.3. Found wrong subscript for       *
       *      lhssvector.  Exit.               *
       ****************************************/
      
#ifdef _DEBUG
      else {
	printf("plusequalsspg2 error 60: lhssvectornonzeroranges=%d,",
	       lhssvectornonzeroranges);
	printf("lastlhssibucketindex=%d, \n\t",lastlhssibucketindex);
	printf("lhssibucketptr->index[%d] = (%d,%d), j1=%d\n",
	       lastlhssibucketindex,
	       lhssibucketptr->
	       index_RANGE_BEGIN[lastlhssibucketindex],
	       lhssibucketptr->
	       index_RANGE_END[lastlhssibucketindex],
	       j1);
	abort();
      }
#endif
      
      
      /****************************************
       * Perform saxpy:                       *
       * 3.4. If there is one, get new        *
       *      nonzero from the rhs_vector     *
       *      whose nonzero was just          *
       *      consumed and replace the        *
       *      heaproot with new entry;        *
       *      Else, delete heaproot.          *
       ****************************************/

      if (k==0) restofplusequalvectorconsumednonzeroranges++;

      if (++rhssvectorconsumednonzeroranges[k] <
	  rhssvectornonzeroranges[k]) {
	if ((rhssibucketindex[k] =
	     MY_MOD(rhssvectorconsumednonzeroranges[k],
		    sIBUCKET_DIM_minus1)) == 0) {
	  if ((rhssibucketptr[k] = rhssibucketptr[k]->nextibucket) ==
	      NULL) {
	    
#ifdef _DEBUG
	    printf("plusequalsspg2 err 61: unexpected NULL nextvbucket: \n");
	    printf("\t k=%d, rhssvectorconsumednonzeroranges=%d,\n",
		   k,rhssvectorconsumednonzeroranges[k]);
	    printf("\t rhssvectornonzeroranges[k]=%d\n",
		   rhssvectornonzeroranges[k]);
	    abort();
#endif
	    
	  }
	}
      }  
      
    } /* matches     2nd while loop           */
  } 

  else {

    /* tempvectorptr is used in the if part as temporary storage
     * for the value of lhssvectorptr.  We echo that statement
     * here so that later the reinstatement of the value of
     * lhssvectorptr could be done regardless of the branch taken.
     */

    tempvectorptr = lhssvectorptr;
 
  }

  rhssvectorptrptrs[1] = &rhssvectorptr[1];
  
  splusequal_axpy_postprocessing(rhsanddestptrptr,
				 rhssvectorptrptrs[1],
				 status_otherrhssvector,
				 &restofplusequalsvectorptr,
				 alphaptrs,
				 restorerhssvectorptr,
				 arity,
				 lhssvectornonzeroes,
				 lhssvectornonzeroranges,
				 CS);
  
  
  
  lhssvectorptr =  tempvectorptr;
  
  
#ifdef _TRACE 
  nnza = (*destptrptr)->nonzeroes; 
  sadinfo(arity,nnzb,lhssvectoraddress,nnza,alphaptrs,vectorptrptrs);
#endif

}

