#define SINGLE
/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#include  <stdlib.h>
#include "sgaxpy_pd_macros_and_data_structures.h"
#include "sgaxpy_pd_global_variables.h"

/****************************************************************************
 *                                                                          *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/

void snon_over_lapping_cs_range_axpy(rangedestination,
				    lhssibucketindex,
				    addressoflastlhssibucketindex,
				    addressoflhssvbucketindex,
				    k,
				    addressofrhssvbucketindex_k,
				    alpha_k,
				    totalcurrentnonzerorange
#ifdef _DEBUG
				    ,addressofunconsumedcurrentnonzerorange
				    ,rhssvectorconsumednonzeroranges_k
				    ,rhssvectornonzeroranges_k
				    ,lhssvectornonzeroranges
#endif
				    )
int rangedestination;
int lhssibucketindex;
int *addressoflastlhssibucketindex;
int *addressoflhssvbucketindex;
int k;
int *addressofrhssvbucketindex_k;
float alpha_k;
int totalcurrentnonzerorange;
#ifdef _DEBUG
int *addressofunconsumedcurrentnonzerorange;
int rhssvectorconsumednonzeroranges_k;
int rhssvectornonzeroranges_k;
int lhssvectornonzeroranges;
#endif
{
  int firstloopsize;
  int numberofdoubleforloopiterations;
  int firstiterationloopsize;
  int seconditerationloopsize;
  int remainder1loopsize;
  int remainder2loopsize;

  int i;


  sDETERMINE_LOOP_BOUNDS(totalcurrentnonzerorange,
			*addressoflhssvbucketindex,
			*addressofrhssvbucketindex_k,
			&firstloopsize,
			&numberofdoubleforloopiterations,
			&firstiterationloopsize,
			&seconditerationloopsize,
			&remainder1loopsize,
			&remainder2loopsize,
			addressofunconsumedcurrentnonzerorange);
  sDENSE_AXPY_LOOP(NEW_RANGE,
		  addressoflhssvbucketindex,
		  addressofrhssvbucketindex_k,
		  k,
		  firstloopsize,
		  alpha_k,
		  addressofunconsumedcurrentnonzerorange);
  
    
  /*********************************************
   * Case 1: rhssvbucket has more spaces   *
   *         remaining than lhssvbucket.   *
   *********************************************/
  
  if (*addressofrhssvbucketindex_k <
      *addressoflhssvbucketindex) {
    
    for(i=0;i<numberofdoubleforloopiterations;i++) {
      sDEBUGGING_CODE_1(LEFT,
		       *addressoflhssvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      sget_and_hop_to_next_vbucket(addressoflhssvbucketindex);
      sDENSE_AXPY_LOOP(NEW_RANGE,
		      addressoflhssvbucketindex,
		      addressofrhssvbucketindex_k,
		      k,
		      firstiterationloopsize,alpha_k,
		      addressofunconsumedcurrentnonzerorange);
      sDEBUGGING_CODE_1(RIGHT,
		       *addressofrhssvbucketindex_k,
		       *addressofunconsumedcurrentnonzerorange);
      sHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhssvbucketindex_k,
			     rhssvectorconsumednonzeroranges_k,
			     rhssvectornonzeroranges_k);
      sDENSE_AXPY_LOOP(NEW_RANGE,
		      addressoflhssvbucketindex,
		      addressofrhssvbucketindex_k,
		      k,
		      seconditerationloopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    
    if (remainder1loopsize > 0) {
      sDEBUGGING_CODE_1(LEFT,
		       *addressoflhssvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      sget_and_hop_to_next_vbucket(addressoflhssvbucketindex);
      sDENSE_AXPY_LOOP(NEW_RANGE,
		      addressoflhssvbucketindex,
		      addressofrhssvbucketindex_k,
		      k,
		      remainder1loopsize,alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    if (remainder2loopsize > 0) {
      sDEBUGGING_CODE_1(RIGHT,
		       *addressofrhssvbucketindex_k,
		       *addressofunconsumedcurrentnonzerorange);
      sHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhssvbucketindex_k,
			     rhssvectorconsumednonzeroranges_k,
			     rhssvectornonzeroranges_k);
      sDENSE_AXPY_LOOP(NEW_RANGE,
		      addressoflhssvbucketindex,
		      addressofrhssvbucketindex_k,
		      k,
		      remainder2loopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    DEBUGGING_CODE_2(*addressofunconsumedcurrentnonzerorange);
    if (rangedestination == COMPLETELY_NEW_RANGE) {
      *addressoflastlhssibucketindex = lhssibucketindex;
    }
  }

  /*********************************************
   * Case 2: rhssvbucket has less spaces   *
   * remaining than lhssvbucket.           *
   *********************************************/	  
  else if (*addressofrhssvbucketindex_k >
	   *addressoflhssvbucketindex) {
    
    for(i=0;i<numberofdoubleforloopiterations;i++) {
      sDEBUGGING_CODE_1(RIGHT,
		       *addressofrhssvbucketindex_k,
		       *addressofunconsumedcurrentnonzerorange);
      sHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhssvbucketindex_k,
			     rhssvectorconsumednonzeroranges_k,
			     rhssvectornonzeroranges_k);
      sDENSE_AXPY_LOOP(NEW_RANGE,
		      addressoflhssvbucketindex,
		      addressofrhssvbucketindex_k,
		      k,
		      firstiterationloopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
      sDEBUGGING_CODE_1(LEFT,
		       *addressoflhssvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      sget_and_hop_to_next_vbucket(addressoflhssvbucketindex);
      sDENSE_AXPY_LOOP(NEW_RANGE,
		      addressoflhssvbucketindex,
		      addressofrhssvbucketindex_k,
		      k,
		      seconditerationloopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    
    if (remainder1loopsize > 0) {
      sDEBUGGING_CODE_1(RIGHT,
		       *addressofrhssvbucketindex_k,
		       *addressofunconsumedcurrentnonzerorange);
      sHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhssvbucketindex_k,
			     rhssvectorconsumednonzeroranges_k,
			     rhssvectornonzeroranges_k);
      sDENSE_AXPY_LOOP(NEW_RANGE,
		      addressoflhssvbucketindex,
		      addressofrhssvbucketindex_k,
		      k,
		      remainder1loopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    
    if (remainder2loopsize > 0) {
      sDEBUGGING_CODE_1(LEFT,
		       *addressoflhssvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      sget_and_hop_to_next_vbucket(addressoflhssvbucketindex);
      sDENSE_AXPY_LOOP(NEW_RANGE,
		      addressoflhssvbucketindex,
		      addressofrhssvbucketindex_k,
		      k,
		      remainder2loopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    DEBUGGING_CODE_2(*addressofunconsumedcurrentnonzerorange);
    if (rangedestination == COMPLETELY_NEW_RANGE) {
      *addressoflastlhssibucketindex = lhssibucketindex;
    }
  }
  
   
  /*********************************************
   * Case 3: rhssvbucket has same # of     *
   * spaces as lhssvbucket (i.e. 0).       *
   *********************************************/
  else if (*addressofrhssvbucketindex_k ==
	   *addressoflhssvbucketindex) {
    
    for(i=0;i<numberofdoubleforloopiterations;i++) {
      sDEBUGGING_CODE_1(LEFT,
		       *addressoflhssvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      sget_and_hop_to_next_vbucket(addressoflhssvbucketindex);
      sHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhssvbucketindex_k,
			     rhssvectorconsumednonzeroranges_k,
			     rhssvectornonzeroranges_k);
      sDENSE_AXPY_LOOP(NEW_RANGE,
		      addressoflhssvbucketindex,
		      addressofrhssvbucketindex_k,
		      k,
		      firstiterationloopsize+seconditerationloopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    if (remainder1loopsize+remainder2loopsize > 0) {
      sDEBUGGING_CODE_1(LEFT,
		       *addressoflhssvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      sget_and_hop_to_next_vbucket(addressoflhssvbucketindex);
      sHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhssvbucketindex_k,
			     rhssvectorconsumednonzeroranges_k,
			     rhssvectornonzeroranges_k);
      sDENSE_AXPY_LOOP(NEW_RANGE,
		      addressoflhssvbucketindex,
		      addressofrhssvbucketindex_k,
		      k,
		      remainder1loopsize+remainder2loopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    DEBUGGING_CODE_2(*addressofunconsumedcurrentnonzerorange);
    if (rangedestination == COMPLETELY_NEW_RANGE) {
      *addressoflastlhssibucketindex = lhssibucketindex;
    }
  }
}






/****************************************************************************
 *                                                                          *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/

void soverlapping_cs_range_axpy(addressoflhssvbucketindex,
			       k,
			       addressofrhssvbucketindex_k,
			       alpha_k,
			       totalcurrentnonzerorange
#ifdef _DEBUG
			       ,addressofunconsumedcurrentnonzerorange
			       ,rhssvectorconsumednonzeroranges_k
			       ,rhssvectornonzeroranges_k
			       ,lhssvectornonzeroranges
#endif
			       )
int *addressoflhssvbucketindex;
int k;
int *addressofrhssvbucketindex_k;
float alpha_k;
int totalcurrentnonzerorange;
#ifdef _DEBUG
int *addressofunconsumedcurrentnonzerorange;
int rhssvectorconsumednonzeroranges_k;
int rhssvectornonzeroranges_k;
int lhssvectornonzeroranges;
#endif
{
  int firstloopsize;
  int numberofdoubleforloopiterations;
  int firstiterationloopsize;
  int seconditerationloopsize;
  int remainder1loopsize;
  int remainder2loopsize;

  int i;
  
  sDETERMINE_LOOP_BOUNDS(totalcurrentnonzerorange,
			*addressoflhssvbucketindex,
			*addressofrhssvbucketindex_k,
			&firstloopsize,
			&numberofdoubleforloopiterations,
			&firstiterationloopsize,
			&seconditerationloopsize,
			&remainder1loopsize,
			&remainder2loopsize,
			addressofunconsumedcurrentnonzerorange);
  sDENSE_AXPY_LOOP(OLD_RANGE,
		  addressoflhssvbucketindex,
		  addressofrhssvbucketindex_k,
		  k,
		  firstloopsize,
		  alpha_k,
		  addressofunconsumedcurrentnonzerorange);
  
  
  /*********************************************
   * Case 1: rhssvbucket has more spaces   *
   *         remaining than lhssvbucket.   *
   *********************************************/
  if (*addressofrhssvbucketindex_k <
      *addressoflhssvbucketindex) {
    /*B.2.1*/
    
    for(i=0;i<numberofdoubleforloopiterations;i++) { /*B.3.1*/
      sDEBUGGING_CODE_1(LEFT,
		       *addressoflhssvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      sHOP_TO_NEXT_LHSVBUCKET(addressoflhssvbucketindex,
			     lhssvectornonzeroranges);
      sDENSE_AXPY_LOOP(OLD_RANGE,
		      addressoflhssvbucketindex,
		      addressofrhssvbucketindex_k,
		      k,
		      firstiterationloopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
      sDEBUGGING_CODE_1(RIGHT,
		       *addressofrhssvbucketindex_k,
		       *addressofunconsumedcurrentnonzerorange);
      sHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhssvbucketindex_k,
			     rhssvectorconsumednonzeroranges_k,
			     rhssvectornonzeroranges_k);
      sDENSE_AXPY_LOOP(OLD_RANGE,
		      addressoflhssvbucketindex,
		      addressofrhssvbucketindex_k,
		      k,
		      seconditerationloopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    } /*B.3.1*/
    
    if (remainder1loopsize > 0) {    	    
      sDEBUGGING_CODE_1(LEFT,
		       *addressoflhssvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      sHOP_TO_NEXT_LHSVBUCKET(addressoflhssvbucketindex,
			     lhssvectornonzeroranges);
      sDENSE_AXPY_LOOP(OLD_RANGE,
		      addressoflhssvbucketindex,
		      addressofrhssvbucketindex_k,
		      k,
		      remainder1loopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    
    if (remainder2loopsize > 0) {    
      sDEBUGGING_CODE_1(RIGHT,
		       *addressofrhssvbucketindex_k,
		       *addressofunconsumedcurrentnonzerorange);
      sHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhssvbucketindex_k,
			     rhssvectorconsumednonzeroranges_k,
			     rhssvectornonzeroranges_k);
      sDENSE_AXPY_LOOP(OLD_RANGE,
		      addressoflhssvbucketindex,
		      addressofrhssvbucketindex_k,
		      k,
		      remainder2loopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    DEBUGGING_CODE_2(*addressofunconsumedcurrentnonzerorange);
  } /*B.2.1*/
  
  /*********************************************
   * Case 2: rhssvbucket has less spaces   *
   * remaining than lhssvbucket.           *
   *********************************************/
  else if (*addressofrhssvbucketindex_k >
	   *addressoflhssvbucketindex) {
 /*B.2.2*/
    
    for(i=0;i<numberofdoubleforloopiterations;i++) { /*B.3.2*/
      sDEBUGGING_CODE_1(RIGHT,
		       *addressofrhssvbucketindex_k,
		       *addressofunconsumedcurrentnonzerorange);
      sHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhssvbucketindex_k,
			     rhssvectorconsumednonzeroranges_k,
			     rhssvectornonzeroranges_k);
      sDENSE_AXPY_LOOP(OLD_RANGE,
		      addressoflhssvbucketindex,
		      addressofrhssvbucketindex_k,
		      k,
		      firstiterationloopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
      sDEBUGGING_CODE_1(LEFT,
		       *addressoflhssvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      sHOP_TO_NEXT_LHSVBUCKET(addressoflhssvbucketindex,
			     lhssvectornonzeroranges);
      sDENSE_AXPY_LOOP(OLD_RANGE,
		      addressoflhssvbucketindex,
		      addressofrhssvbucketindex_k,
		      k,
		      seconditerationloopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    } /*B.3.2*/
    
    if (remainder1loopsize > 0) {        
      sDEBUGGING_CODE_1(RIGHT,
		       *addressofrhssvbucketindex_k,
		       *addressofunconsumedcurrentnonzerorange);
      sHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhssvbucketindex_k,
			     rhssvectorconsumednonzeroranges_k,
			     rhssvectornonzeroranges_k);
      sDENSE_AXPY_LOOP(OLD_RANGE,
		      addressoflhssvbucketindex,
		      addressofrhssvbucketindex_k,
		      k,
		      remainder1loopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    if (remainder2loopsize > 0) {    
      sDEBUGGING_CODE_1(LEFT,
		       *addressoflhssvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      sHOP_TO_NEXT_LHSVBUCKET(addressoflhssvbucketindex,
			     lhssvectornonzeroranges);
      sDENSE_AXPY_LOOP(OLD_RANGE,
		      addressoflhssvbucketindex,
		      addressofrhssvbucketindex_k,
		      k,
		      remainder2loopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    DEBUGGING_CODE_2(*addressofunconsumedcurrentnonzerorange);
  } /*B.2.2*/
  
  
  /*********************************************
   * Case 3: rhssvbucket has same # of     *
   * spaces as lhssvbucket (i.e. 0).       *
   *********************************************/
  else if(*addressofrhssvbucketindex_k ==
	  *addressoflhssvbucketindex) {
    /*B.2.3*/
    
    for(i=0;i<numberofdoubleforloopiterations;i++) { /*B.3.3*/
      sDEBUGGING_CODE_1(LEFT,
		       *addressoflhssvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      sHOP_TO_NEXT_LHSVBUCKET(addressoflhssvbucketindex,
			     lhssvectornonzeroranges);
      sHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhssvbucketindex_k,
			     rhssvectorconsumednonzeroranges_k,
			     rhssvectornonzeroranges_k);
      sDENSE_AXPY_LOOP(OLD_RANGE,
		      addressoflhssvbucketindex,
		      addressofrhssvbucketindex_k,
		      k,
		      firstiterationloopsize+seconditerationloopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    } /*B.3.3*/
    
    if (remainder1loopsize+remainder2loopsize > 0) {    
      sDEBUGGING_CODE_1(LEFT,
		       *addressoflhssvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      sHOP_TO_NEXT_LHSVBUCKET(addressoflhssvbucketindex,
			     lhssvectornonzeroranges);
      sHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhssvbucketindex_k,
			     rhssvectorconsumednonzeroranges_k,
			     rhssvectornonzeroranges_k);
      sDENSE_AXPY_LOOP(OLD_RANGE,
		      addressoflhssvbucketindex,
		      addressofrhssvbucketindex_k,
		      k,
		      remainder1loopsize+remainder2loopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    DEBUGGING_CODE_2(*addressofunconsumedcurrentnonzerorange);
  } /*B.2.3*/
}



/****************************************************************************
 *                                                                          *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/

void shop_to_next_rhsvbucket(k,
			    vbucketindex
#ifdef _DEBUG
			    ,rhssvectorconsumednonzeroranges
			    ,rhssvectornonzeroranges
#endif
			    )
int k;
int *vbucketindex;
#ifdef _DEBUG
int rhssvectorconsumednonzeroranges;
int rhssvectornonzeroranges;
#endif
{
  if ((rhssvbucketptr[k] = rhssvbucketptr[k]->nextvbucket) == NULL) {
	      
#ifdef _DEBUG
    printf("shop_to_next_rhsvbucket error: unexpected NULL nextvbucket: \n");
    printf("\t k=%d, rhssvectorconsumednonzeroranges[k]=%d,\n",
	   k,rhssvectorconsumednonzeroranges);
    printf("\t rhssvectornonzeroranges[k]=%d\n",
	   rhssvectornonzeroranges);
    abort();
#endif
  }
  else {
    *vbucketindex = 0;
  }
}





/****************************************************************************
 *                                                                          *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/

void shop_to_next_lhsvbucket(vbucketindex
#ifdef _DEBUG
		      ,lhssvectornonzeroranges
#endif
		      )
int *vbucketindex;
#ifdef _DEBUG
int lhssvectornonzeroranges;
#endif
{
  if ((lhssvbucketptr = lhssvbucketptr->nextvbucket) == NULL) {
    
#ifdef _DEBUG
    printf("shop_to_next_lhsvbucket error: unexpected NULL nextvbucket: \n");
    printf("lhssvectornonzeroranges=%d\n",lhssvectornonzeroranges);
    abort();
#endif
  }
  else {
    *vbucketindex = 0;
  }
}  





/****************************************************************************
 *                                                                          *
 *      sget_and_hop_to_next_vbucket(lhssvbucketindex)                   *
 *                                                                          *
 ****************************************************************************/

void sget_and_hop_to_next_vbucket(lhssvbucketindex)
int *lhssvbucketindex;
{
  sgetvbucket(&(lhssvbucketptr->nextvbucket));
  lhssvbucketptr = lhssvbucketptr->nextvbucket;
  *lhssvbucketindex = 0;
}







/****************************************************************************
 *                                                                          *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/

void sconvertdstype(oldvectorptrptr,k,newvectordstype)
svectorstruct **oldvectorptrptr;
int k;
int newvectordstype;
{
  sbucketstruct *holdingbinbucketptr;
  sbucketstruct *oldbucketptr;
  sibucketstruct *holdingbinibucketptr;
  svbucketstruct *holdingbinvbucketptr;
  int bucketindex;
  int ibucketindex = 0;
  int vbucketindex;
  int j;
  int ij;
  int vj;
  int lastindex;
  svectorstruct *oldvectorptr = (*oldvectorptrptr);
  int oldvectordstype = oldvectorptr->dstype;

  if (oldvectordstype >= newvectordstype || newvectordstype == S1) {
#ifdef _DEBUG
    printf("sconvertdstype error 1: Illegal Data Structure Conversion:\n\t");
    printf("newvectordstype=%d,    oldvectordstype=%d\n",
	   newvectordstype,oldvectordstype);
    abort();
#endif
  }
  else if (oldvectordstype == S1 &&  newvectordstype == SS) {
    holdingbinsvectorptr[k]->nonzeroes = 1;
    holdingbinsvectorptr[k]->dstype = SS;
    sgetbucket(&holdingbinbucketptr);
    holdingbinsvectorptr[k]->sds.ss.bucketptr = holdingbinbucketptr;
    holdingbinbucketptr->index[0] = oldvectorptr->sds.s1.index;
    holdingbinbucketptr->value[0] = oldvectorptr->sds.s1.value;
  }
  else if (oldvectordstype == S1 &&  newvectordstype == CS) {
    holdingbinsvectorptr[k]->nonzeroes = 1;
    holdingbinsvectorptr[k]->dstype = CS;
    sgetibucket(&holdingbinibucketptr);
    sgetvbucket(&holdingbinvbucketptr);
    holdingbinsvectorptr[k]->sds.cs.nonzeroranges = 1;
    holdingbinsvectorptr[k]->sds.cs.ibucketptr = holdingbinibucketptr;
    holdingbinsvectorptr[k]->sds.cs.vbucketptr = holdingbinvbucketptr;
    holdingbinibucketptr->index_RANGE_BEGIN[0] = oldvectorptr->sds.s1.index;
    holdingbinibucketptr->index_RANGE_END[0] = oldvectorptr->sds.s1.index;
    holdingbinvbucketptr->value[0] = oldvectorptr->sds.s1.value;
  }
  else if (oldvectordstype == SS &&  newvectordstype == CS) {
    oldbucketptr = oldvectorptr->sds.ss.bucketptr;
    holdingbinsvectorptr[k]->nonzeroes = oldvectorptr->nonzeroes;
    holdingbinsvectorptr[k]->dstype = CS;
    sgetibucket(&holdingbinibucketptr);
    holdingbinsvectorptr[k]->sds.cs.ibucketptr = holdingbinibucketptr;
    sgetvbucket(&holdingbinvbucketptr);
    holdingbinsvectorptr[k]->sds.cs.vbucketptr = holdingbinvbucketptr;
    holdingbinibucketptr->index_RANGE_BEGIN[0] = oldbucketptr->index[0];
    holdingbinvbucketptr->value[0] = oldbucketptr->value[0];
    lastindex = oldbucketptr->index[0];
    
    for (j=1,ij=0,vj=1;j<oldvectorptr->nonzeroes;j++,vj++) {
      if ((bucketindex = MY_MOD(j,sBUCKET_DIM_minus1)) == 0)
	  oldbucketptr = oldbucketptr->nextbucket;
      if ((vbucketindex = MY_MOD(vj,sVBUCKET_DIM_minus1)) == 0) {
	sgetvbucket(&holdingbinvbucketptr->nextvbucket);
	holdingbinvbucketptr = holdingbinvbucketptr->nextvbucket;
      }
      holdingbinvbucketptr->value[vbucketindex] =
	oldbucketptr->value[bucketindex];

      if (oldbucketptr->index[bucketindex] > lastindex+1) {
	holdingbinibucketptr->index_RANGE_END[ibucketindex] = lastindex;
	ij++;
	if ((ibucketindex = MY_MOD(ij,sIBUCKET_DIM_minus1)) == 0) {
	  sgetibucket(&holdingbinibucketptr->nextibucket);
	  holdingbinibucketptr = holdingbinibucketptr->nextibucket;
	}
	holdingbinibucketptr->index_RANGE_BEGIN[ibucketindex] =
	  oldbucketptr->index[bucketindex];
      }
      else if (oldbucketptr->index[bucketindex] < lastindex+1) {
#ifdef _DEBUG
	printf("sconvertdstype error 2: BAD LOGIC:\n\t");
	printf("newvectordstype=%d,    oldvectordstype=%d\n",
	       newvectordstype,oldvectordstype);
	abort();
#endif
      }
      lastindex = oldbucketptr->index[bucketindex];
    }
    holdingbinibucketptr->index_RANGE_END[ibucketindex] = lastindex;    
    holdingbinsvectorptr[k]->sds.cs.nonzeroranges = ij+1;
  }
  else /* We should never go thru this branch */ {
#ifdef _DEBUG
    printf("sconvertdstype error 3: BAD LOGIC:\n\t");
    printf("newvectordstype=%d,    oldvectordstype=%d\n",
	   newvectordstype,oldvectordstype);
    abort();
#endif
  }
  *oldvectorptrptr =  holdingbinsvectorptr[k];
}








/****************************************************************************
 *                                                                          *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/

void saxpy_0_1_post_processing(destptrptr)
svectorstruct **destptrptr;
{
  svectorstruct *tempvectorptr;
  sbucketstruct *tempbucketptr;
  sibucketstruct *tempibucketptr;
  svbucketstruct *tempvbucketptr;
  
  /*********************************************
   * Switch lhssvectorptr and *destptrptr. *
   *********************************************/
  tempvectorptr = (*destptrptr);
  (*destptrptr) = lhssvectorptr;
  lhssvectorptr = tempvectorptr;
  
  if (tempvectorptr->dstype == SS) {
    /*********************************************
     * If *destptrptr was not empty,             *
     * add its buckets to freedsbucketslist. *
     *********************************************/
    tempbucketptr = tempvectorptr->sds.ss.bucketptr;
    if (tempbucketptr != NULL) {
      while (tempbucketptr->nextbucket != NULL)
	tempbucketptr = tempbucketptr->nextbucket;
      tempbucketptr->nextbucket = freedsbucketslist;
      RESET_freedsbucketslist(tempvectorptr->sds.ss.bucketptr);
      tempvectorptr->sds.ss.bucketptr = NULL;
    }
  }
  else if (tempvectorptr->dstype == CS) {
    /********************************************
     * If *destptrptr was not empty, add        *
     * its ibuckets to freedsibucketslist.  *
     ********************************************/
    
    tempibucketptr = tempvectorptr->sds.cs.ibucketptr;
    tempvbucketptr = tempvectorptr->sds.cs.vbucketptr;
    
    if (tempibucketptr != NULL) {
      while (tempibucketptr->nextibucket != NULL)
	tempibucketptr = tempibucketptr->nextibucket;
      tempibucketptr->nextibucket = freedsibucketslist;
      RESET_freedsibucketslist(tempvectorptr->sds.cs.ibucketptr);
      tempvectorptr->sds.cs.ibucketptr = NULL;
    }
    
    /********************************************
     * If *destptrptr was not empty, add        *
     * its vbuckets to freedsvbucketslist.  *
     ********************************************/
    
    if (tempvbucketptr != NULL) {
      while (tempvbucketptr->nextvbucket != NULL)
	tempvbucketptr = tempvbucketptr->nextvbucket;
      tempvbucketptr->nextvbucket = freedsvbucketslist;
      RESET_freedsvbucketslist(tempvectorptr->sds.cs.vbucketptr);
      tempvectorptr->sds.cs.vbucketptr = NULL;
    }
  } /* matches   else if (tempvectorptr->dstype == CS) {             */
}


/****************************************************************************
 *                                                                          *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/

void saxpy_2_heap_post_processing(destptrptr,
				 rhssvectorptrptrs,
				 alphaptrs,
				 restorerhssvectorptr,
				 arity,
				 lhssvectornonzeroes,
				 lhssvectornonzeroranges,
				 governingdstype)
svectorstruct **destptrptr;
svectorstruct **rhssvectorptrptrs[];
float *alphaptrs[];
svectorstruct *restorerhssvectorptr[MAXARITY+1];
int arity;
int lhssvectornonzeroes;
int lhssvectornonzeroranges;
int governingdstype;
{
  svectorstruct *rhssvectorptr[MAXARITY];
  svectorstruct *tempvectorptr;
  svectorstruct *sstempvectorptr;
  sbucketstruct *tempbucketptr;
  sbucketstruct *sstempbucketptr;
  sibucketstruct *tempibucketptr;
  svbucketstruct *tempvbucketptr;
  int k;

  for (k=0;k<arity;k++)
    rhssvectorptr[k] = *rhssvectorptrptrs[k];

  /****************************************
   * Switch lhssvectorptr and *destptrptr. *
   ****************************************/

  lhssvectorptr->nonzeroes = lhssvectornonzeroes;
  
  if (governingdstype == CS)
    lhssvectorptr->sds.cs.nonzeroranges = lhssvectornonzeroranges;
  
  tempvectorptr = (*destptrptr);
  (*destptrptr) = lhssvectorptr;
  lhssvectorptr = tempvectorptr;

  if (governingdstype == SS) {
    
    if ((INITIALIZED_MAXINDEX &&
	 ((float) (*destptrptr)->nonzeroes / (float) GLOBAL_MAXINDEX) >=
	 SWITCH_NZ_RATIO)     ||
	((*destptrptr)->nonzeroes >= SWITCH_NZ_VALUE)) {
     
       /****************************************
       * Convert SS to CS.                    *
       ****************************************/
      
      sstempvectorptr = (*destptrptr);
      sconvertdstype(destptrptr,MAXARITY,CS);
	  
      sstempbucketptr = sstempvectorptr->sds.ss.bucketptr;
      if (sstempbucketptr != NULL) {
	while (sstempbucketptr->nextbucket != NULL)
	  sstempbucketptr = sstempbucketptr->nextbucket;
	sstempbucketptr->nextbucket = freedsbucketslist;
	RESET_freedsbucketslist(sstempvectorptr->sds.ss.bucketptr);
	sstempvectorptr->sds.ss.bucketptr = NULL;
      }
      holdingbinsvectorptr[MAXARITY] =  sstempvectorptr;
    }
  }
  
  /****************************************
   * If *destptrptr was not empty,        *
   * add its buckets to freedsbucketslist. *
   ****************************************/
  
  if (tempvectorptr->dstype == SS) {
    tempbucketptr = tempvectorptr->sds.ss.bucketptr;
    if (tempbucketptr != NULL) {
      while (tempbucketptr->nextbucket != NULL)
	tempbucketptr = tempbucketptr->nextbucket;
      tempbucketptr->nextbucket = freedsbucketslist;
      RESET_freedsbucketslist(tempvectorptr->sds.ss.bucketptr);
      tempvectorptr->sds.ss.bucketptr = NULL;
    }
  }
  else if (tempvectorptr->dstype == CS) {
    tempibucketptr = tempvectorptr->sds.cs.ibucketptr;
    if (tempibucketptr != NULL) {
      while (tempibucketptr->nextibucket != NULL)
	tempibucketptr = tempibucketptr->nextibucket;
      tempibucketptr->nextibucket = freedsibucketslist;
      RESET_freedsibucketslist(tempvectorptr->sds.cs.ibucketptr);
      tempvectorptr->sds.cs.ibucketptr = NULL;
    }
    tempvbucketptr = tempvectorptr->sds.cs.vbucketptr;
    if (tempvbucketptr != NULL) {
      while (tempvbucketptr->nextvbucket != NULL)
	tempvbucketptr = tempvbucketptr->nextvbucket;
      tempvbucketptr->nextvbucket = freedsvbucketslist;
      RESET_freedsvbucketslist(tempvectorptr->sds.cs.vbucketptr);
      tempvectorptr->sds.cs.vbucketptr = NULL;
    }
  }
  
  /*****************************************
   * If holdingbins were used, add         *
   * their buckets to freedsbucketslist, OR *
   * their ibuckets to freedsibucketslist & *
   * their vbuckets to freedsvbucketslist.  *
   *****************************************/
  for (k=0;k<arity;k++) {
    
    rhssvectorptr[k] = restorerhssvectorptr[k];
    
    if (holdingbinsvectorptr[k]->dstype == SS) {
      tempbucketptr = holdingbinsvectorptr[k]->sds.ss.bucketptr;
      if (tempbucketptr != NULL) {
#ifdef _DEBUG
	if (governingdstype == CS) {
	  printf("sspgx warning 7.4: Unexpcted non-NULL");
	  printf("tempbucketptr.\n");
	}
#endif
	while (tempbucketptr->nextbucket != NULL) {
#ifdef _DEBUG
	  if (governingdstype == SS) {
	    printf("sspgx warning 7.1: tempbucketptr should have");
	    printf(" no trailing nextbuckets\n");
	  }
#endif
	  tempbucketptr = tempbucketptr->nextbucket;
	}
	tempbucketptr->nextbucket = freedsbucketslist;
	RESET_freedsbucketslist(holdingbinsvectorptr[k]->sds.ss.bucketptr);
	holdingbinsvectorptr[k]->sds.ss.bucketptr = NULL;
      }
    }
    else if (holdingbinsvectorptr[k]->dstype == CS) {
      tempibucketptr = holdingbinsvectorptr[k]->sds.cs.ibucketptr;
      if (tempibucketptr != NULL) {
#ifdef _DEBUG
	if (governingdstype == SS) {
	  printf("sspgx warning 7.2: Unexpcted non-NULL");
	  printf(" ibucketptr.\n");
	}
#endif
	while (tempibucketptr->nextibucket != NULL)
	  tempibucketptr = tempibucketptr->nextibucket;
	tempibucketptr->nextibucket = freedsibucketslist;
	RESET_freedsibucketslist(holdingbinsvectorptr[k]->sds.cs.ibucketptr);
	holdingbinsvectorptr[k]->sds.cs.ibucketptr = NULL;
      }
      tempvbucketptr = holdingbinsvectorptr[k]->sds.cs.vbucketptr;
      if (tempvbucketptr != NULL) {
#ifdef _DEBUG
	if (governingdstype == SS) {
	  printf("sspgx warning 7.3: Unexpcted non-NULL");
	  printf(" vbucketptr.\n");
	}
#endif
	while (tempvbucketptr->nextvbucket != NULL)
	  tempvbucketptr = tempvbucketptr->nextvbucket;
	tempvbucketptr->nextvbucket = freedsvbucketslist;
	RESET_freedsvbucketslist(holdingbinsvectorptr[k]->sds.cs.vbucketptr);
	holdingbinsvectorptr[k]->sds.cs.vbucketptr = NULL;
      }
    }
  } /* matches    for (k=0;k<arity;k++) {    */
}








/****************************************************************************
 *                                                                          *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/

void splusequal_axpy_postprocessing(destptrptr,
				   rhssvectorptrptr,
				   status_otherrhssvector,
				   restofplusequalvectorptrptr,
				   alphaptrs,
				   restorerhssvectorptr,
				   arity,
				   lhssvectornonzeroes,
				   lhssvectornonzeroranges,
				   governingdstype)
svectorstruct **destptrptr;
svectorstruct **rhssvectorptrptr;
int status_otherrhssvector;
svectorstruct **restofplusequalvectorptrptr;
float *alphaptrs[2];
svectorstruct *restorerhssvectorptr[2+1];
int arity;
int lhssvectornonzeroes;
int lhssvectornonzeroranges;
int governingdstype;
{
  svectorstruct *rhssvectorptr[2];
  svectorstruct *tempvectorptr;
  sbucketstruct *tempbucketptr;
  sibucketstruct *tempibucketptr;
  svbucketstruct *tempvbucketptr;

  int k;

  k = 1;
    rhssvectorptr[k] = *rhssvectorptrptr;

    /****************************************
     * Switch lhssvectorptr and *destptrptr. *
     ****************************************/

    lhssvectorptr->nonzeroes = lhssvectornonzeroes;
    lhssvectorptr->sds.cs.nonzeroranges = lhssvectornonzeroranges;
  

    /*********************************************************
     * If *restofplusequalvectorptrptr was not empty,        *
     * add its buckets to freedsbucketslist.                  *
     *********************************************************/
  
    tempvectorptr = (*restofplusequalvectorptrptr);
    tempibucketptr = tempvectorptr->sds.cs.ibucketptr;
    if (tempibucketptr != NULL) {
      while (tempibucketptr->nextibucket != NULL)
	tempibucketptr = tempibucketptr->nextibucket;
      tempibucketptr->nextibucket = freedsibucketslist;
      RESET_freedsibucketslist(tempvectorptr->sds.cs.ibucketptr);
      tempvectorptr->sds.cs.ibucketptr = NULL;
    }
    tempvbucketptr = tempvectorptr->sds.cs.vbucketptr;
    if (tempvbucketptr != NULL) {
      while (tempvbucketptr->nextvbucket != NULL)
	tempvbucketptr = tempvbucketptr->nextvbucket;
      tempvbucketptr->nextvbucket = freedsvbucketslist;
      RESET_freedsvbucketslist(tempvectorptr->sds.cs.vbucketptr);
      tempvectorptr->sds.cs.vbucketptr = NULL;
    }

   
  /*****************************************
   * If status_otherrhssvector is TEMP,     *
   * add its buckets to freedsbucketslist,  *
   * and reassign pointers.                *
   *****************************************/

  if (status_otherrhssvector == TEMP_OTHER_VECTOR) {

    tempvectorptr = tempsdestptr;
    tempibucketptr = tempvectorptr->sds.cs.ibucketptr;
    if (tempibucketptr != NULL) {
      while (tempibucketptr->nextibucket != NULL)
	tempibucketptr = tempibucketptr->nextibucket;
      tempibucketptr->nextibucket = freedsibucketslist;
      RESET_freedsibucketslist(tempvectorptr->sds.cs.ibucketptr);
      tempvectorptr->sds.cs.ibucketptr = NULL;
    }
    tempvbucketptr = tempvectorptr->sds.cs.vbucketptr;
    if (tempvbucketptr != NULL) {
      while (tempvbucketptr->nextvbucket != NULL)
	tempvbucketptr = tempvbucketptr->nextvbucket;
      tempvbucketptr->nextvbucket = freedsvbucketslist;
      RESET_freedsvbucketslist(tempvectorptr->sds.cs.vbucketptr);
      tempvectorptr->sds.cs.vbucketptr = NULL;
    }

  }
 
  /*****************************************
   * If holdingbins were used, add         *
   * their buckets to freedsbucketslist, OR *
   * their ibuckets to freedsibucketslist & *
   * their vbuckets to freedsvbucketslist.  *
   *****************************************/

  /*  We could not have put rhssvectorptr[0] in a holdingbin;
   *  but there may be something in it from the routine that called
   *  plusequal (in particular sspg2 (sspgx does it's own 
   *  holdingbin cleaning).
   */

  for (k=0;k<arity;k++) {

    rhssvectorptr[k] = restorerhssvectorptr[k];
    
    if (holdingbinsvectorptr[k]->dstype == SS) {
      tempbucketptr = holdingbinsvectorptr[k]->sds.ss.bucketptr;
      if (tempbucketptr != NULL) {
#ifdef _DEBUG
	if (governingdstype == CS) {
	  printf("sspgx warning 7.4: Unexpcted non-NULL");
	  printf("tempbucketptr.\n");
	}
#endif
	while (tempbucketptr->nextbucket != NULL) {
#ifdef _DEBUG
	  if (governingdstype == SS) {
	    printf("sspgx warning 7.1: tempbucketptr should have");
	    printf(" no trailing nextbuckets\n");
	  }
#endif
	  tempbucketptr = tempbucketptr->nextbucket;
	}
	tempbucketptr->nextbucket = freedsbucketslist;
	RESET_freedsbucketslist(holdingbinsvectorptr[k]->sds.ss.bucketptr);
	holdingbinsvectorptr[k]->sds.ss.bucketptr = NULL;
      }
    }
    else if (holdingbinsvectorptr[k]->dstype == CS) {
      tempibucketptr = holdingbinsvectorptr[k]->sds.cs.ibucketptr;
      if (tempibucketptr != NULL) {
#ifdef _DEBUG
	if (governingdstype == SS) {
	  printf("sspgx warning 7.2: Unexpcted non-NULL");
	  printf(" ibucketptr.\n");
	}
#endif
	while (tempibucketptr->nextibucket != NULL)
	  tempibucketptr = tempibucketptr->nextibucket;
	tempibucketptr->nextibucket = freedsibucketslist;
	RESET_freedsibucketslist(holdingbinsvectorptr[k]->sds.cs.ibucketptr);
	holdingbinsvectorptr[k]->sds.cs.ibucketptr = NULL;
      }
      tempvbucketptr = holdingbinsvectorptr[k]->sds.cs.vbucketptr;
      if (tempvbucketptr != NULL) {
#ifdef _DEBUG
	if (governingdstype == SS) {
	  printf("sspgx warning 7.3: Unexpcted non-NULL");
	  printf(" vbucketptr.\n");
	}
#endif
	while (tempvbucketptr->nextvbucket != NULL)
	  tempvbucketptr = tempvbucketptr->nextvbucket;
	tempvbucketptr->nextvbucket = freedsvbucketslist;
	RESET_freedsvbucketslist(holdingbinsvectorptr[k]->sds.cs.vbucketptr);
	holdingbinsvectorptr[k]->sds.cs.vbucketptr = NULL;
      }
    }
  }
}





/****************************************************************************
 *                                                                          *
 *  sdetermine_loop_bounds                                                   *
 *                                                                          *
 *     Based on how close lhsvbucketindex and rhsvbucketindex pointers are  *
 *     to the end of the vbucket, and based on the number of values left to *
 *     process in the axpy, sdetermine_loop_bounds figures out the various   *
 *     loop values.                                                         *
 *                                                                          *
 ****************************************************************************/


void sdetermine_loop_bounds(totalcurrentnonzerorange,
			   lhsvbucketindex,
			   rhsvbucketindex,
			   firstloopsize,
			   numberofdoubleforloopiterations,
			   firstiterationloopsize,
			   seconditerationloopsize,
			   remainder1loopsize,
			   remainder2loopsize
#ifdef _DEBUG
			   ,addressofunconsumedcurrentnonzerorange
#endif
			   )
int totalcurrentnonzerorange;
int lhsvbucketindex;
int rhsvbucketindex;
int *firstloopsize;
int *numberofdoubleforloopiterations;
int *firstiterationloopsize;
int *seconditerationloopsize;
int *remainder1loopsize;
int *remainder2loopsize;
#ifdef _DEBUG
int *addressofunconsumedcurrentnonzerorange;
#endif
{
  int temprange, tmpmod;

  *firstloopsize =
    min(((sVBUCKET_DIM - max(lhsvbucketindex,rhsvbucketindex))),
	totalcurrentnonzerorange);
  temprange = totalcurrentnonzerorange - *firstloopsize;
  *numberofdoubleforloopiterations = temprange / sVBUCKET_DIM;
  *firstiterationloopsize =
    min((sVBUCKET_DIM -
	 (*firstloopsize + min(lhsvbucketindex,rhsvbucketindex))),temprange);
  *seconditerationloopsize =
    min((sVBUCKET_DIM - *firstiterationloopsize),
	(temprange - *firstiterationloopsize));
  tmpmod = MY_MOD(temprange,sVBUCKET_DIM_minus1);
  *remainder1loopsize = min (tmpmod,*firstiterationloopsize);
  *remainder2loopsize = max ((tmpmod - (*firstiterationloopsize)),0);
  
#ifdef _DEBUG
  *addressofunconsumedcurrentnonzerorange =  totalcurrentnonzerorange;
#endif

}

