#define DOUBLE_COMPLEX
/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#include  <stdlib.h>
#include "sgaxpy_pd_macros_and_data_structures.h"
#include "zgaxpy_pd_macros_and_data_structures.h"
#include "sgaxpy_pd_global_variables.h"
#include "zgaxpy_pd_global_variables.h"


/*
 *  Function Prototyping done locally to grab "TO_pREc_" coorectly.
 */

#define _P(x) ()

void ssp2z _P((zvectorstruct **destptrptr,
				       svectorstruct **realvectorptrptr,
				       svectorstruct **imagvectorptrptr
				       ));

/****************************************************************************
 *                                                                          *
 *    Patterned after sspg2                                                *
 *                                                                          *
 ****************************************************************************/

void ssp2z(destptrptr,realvectorptrptr,imagvectorptrptr)
zvectorstruct **destptrptr;
svectorstruct **realvectorptrptr;
svectorstruct **imagvectorptrptr;
{

  int arity = 2;
  
  svectorstruct *rhsvectorptr[2];
  
  sbucketstruct *rhsbucketptr[2];
  
  zbucketstruct *lhszbucketptr;
  
  int i;
  
  
  sibucketstruct *rhsibucketptr[2];
                                      /* Array of pointers to "current"     *
				       * buckets where "current" nonzeroes  *
				       * of the rhs vectors reside.         */

  zibucketstruct *lhszibucketptr; 
                                      /* Pointer to current bucket of       *
				       * lhszvector.                 */

  int rhsbucketindex[2];              /* Array of indices of current        *
				       * nonzeroes in the current buckets   *
				       * of rhs vectors.                    */

  int lhszbucketindex;         /* bucketindex of next nonzero to be  *
				       * added to lhszvector.        */

  int lastlhszbucketindex;     /* bucketindex of the latest added    *
				       * nonzero in the lhszvector.  *
				       * We need this additional index to   *
				       * leap-frogbuckets, if need be.      */


  int rhsibucketindex[2];             /* Array of indices of current        *
				       * nonzeroes in the current buckets   *
				       * of rhs vectors.                    */

  int lhszibucketindex;        /* bucketindex of next nonzero to be  *
				       * added to lhszvector.        */

  int lastlhszibucketindex;    /* bucketindex of the latest added    *
				       * nonzero in the lhszvector.  *
				       * We need this additional index to   *
				       * leap-frogbuckets, if need be.      */

  int firstsaxpyoperand;              /* Order in the ennumeration of the   *
				       * rhs vectors of the first non-null  *
				       * saxpy vectorop.                    */

  int nullsaxpyflag = 1;              /* == 1 means no non-null vectorops.  */

  int k;                              /* Counter for arity.                 */

  int j;                              /* Index for vectorsubscripts to be   *
				       * entered in (and later popped from) *
				       * the heap.                          */

  int rhsvectornonzeroes[2];          /* Array of total # of nonzeroes in   *
				       * each rhs vector.                   */

  int rhsvectorconsumednonzeroes[2];  /* Array of # of nonzeroes consumed   *
				       * so far in each rhs vector.         */

  int lhszvectornonzeroes;     /* Count of # of nonzeroes entered    *
				       * into the lhszvector so far. *
				       * By the end of the saxpy operation, *
				       * this index equals the total # of   *
				       * nonzeroes in the resultant vector. */

/*
 * These variables were originally for COMPRESSED_SUBSCRIPTS
 */

  svbucketstruct *rhsvbucketptr[2];
                                      /* Array of pointers to "current"     *
				       * vbuckets where "current" nonzeroes *
				       * of the rhs vectors reside.         */


  zvbucketstruct *lastlhszvbucketptr;
                                      /* Pointer to current vbucket of      *
				       * lhszvector.                 */


  int rhsvectornonzeroranges[2];
                                      /* Array of total # of nonzeroes      *
				       * ranges in each rhs vector.         */

  int rhsvectorconsumednonzeroranges[2];
                                      /* Array of # of nonzeroes consumed   *
				       * so far in each rhs vector.         */

  int rhsvbucketindex[2];             /* Array of indices of current        *
				       * nonzeroes in the current buckets   *
				       * of rhs vectors.                    */

  int lhszvectornonzeroranges; /* Count of # of nonzeroes entered    *
				       * into the lhszvector so far. *
				       * By the end of the saxpy operation, *
				       * this index equals the total # of   *
				       * nonzeroes in the resultant vector. */

  int lhszvbucketindex;        /* bucketindex of next nonzero to be  *
				       * added to lhszvector.        */


  int lastlhszvbucketindex;    /* bucketindex of the latest added    *
				       * nonzero in the lhszvector.  *
				       * We need this additional index to   *
				       * leap-frog buckets, if need be.     */

  zvbucketstruct *templhszvbucketptr;
  int templhszvbucketindex;

  int j1;                             /* Index for vectorsubscripts to be   *
				       * entered in (and later popped from) *
				       * the heap.                          */

  int j2;                             /* Index for vectorsubscripts to be   *
				       * entered in (and later popped from) *
				       * the heap.                          */

  int h1;
  int h2;

  int totalcurrentnonzerorange;

#ifdef _DEBUG
  int unconsumedcurrentnonzerorange;
#endif

  int temp1range;
  int distancetooverlap;
  int distancetoendofvbucket;


  /*
   * These variables were added for union ds
   */
  
  svectorstruct *restorerhsvectorptr[2];
  int governingdstype;
  
  /*
   * These variables were added for plusequal code
   */
  
  svectorstruct **temprhsvectorptrptrs[MAXARITY-1];
  
  
  svectorstruct **rhsvectorptrptrs[MAXARITY];
  
  int condition[2];
  
  
  
  /****************************************
   * Initialization of *destptrptr.       *
   ****************************************/
  
  if (*destptrptr == NULL)
    zgetvector(destptrptr);
  
  /****************************************
   * Dereferencing.                       *
   ****************************************/

  rhsvectorptr[0] = *realvectorptrptr;
  rhsvectorptr[1] = *imagvectorptrptr;
 
  for (k=0;k<arity;k++) {
    restorerhsvectorptr[k] = rhsvectorptr[k];
    rhsvectornonzeroes[k] = rhsvectorptr[k]->nonzeroes;
    if (rhsvectorptr[k]->dstype == CS)
      rhsvectornonzeroranges[k] = rhsvectorptr[k]->sds.cs.nonzeroranges;
  }
  
  if (rhsvectorptr[0]->nonzeroes == 0 &&
      rhsvectorptr[1]->nonzeroes == 0) {
    zspzronotca(destptrptr);
  }
  else {
    governingdstype = max(rhsvectorptr[0]->dstype,
			  rhsvectorptr[1]->dstype);
    if (governingdstype == S1 &&
	rhsvectorptr[0]->sds.s1.index != rhsvectorptr[1]->sds.s1.index)
      governingdstype = SS;
    for (k=0;k<arity;k++)
      if (rhsvectorptr[k]->dstype != governingdstype) {
	sconvertdstype((&rhsvectorptr[k]),k,governingdstype);
	if (governingdstype == CS)
	  rhsvectornonzeroranges[k] =
	    rhsvectorptr[k]->sds.cs.nonzeroranges;
      }
    
    if (governingdstype == S1) {
      
      /******************************************************************
       *                      ONE_SUBSCRIPT AXPY                        *
       ******************************************************************/
      
      lhszvectorptr->nonzeroes =  1;
      lhszvectorptr->dstype = governingdstype;
      lhszvectorptr->zds.s1.index =
	rhsvectorptr[0]->sds.s1.index;
      XY_PUT_AND_CAST(lhszvectorptr->zds.s1.value,
		      rhsvectorptr[0]->sds.s1.value,
		      rhsvectorptr[1]->sds.s1.value,
		      double);

      zaxpy_0_1_post_processing(destptrptr);

    }
    else /* if we are going to use lhszvectorptr */ {
      if (governingdstype == SS) {
	
	/****************************************************************
	 *                 _SINGLE_SUBSCRIPTS AXPY                      *
	 ****************************************************************/
	for (k=0;k<arity;k++) {
	  rhsvectorconsumednonzeroes[k]= 0;
	  rhsbucketindex[k] = 0;
	  rhsbucketptr[k] = rhsvectorptr[k]->sds.ss.bucketptr;
	}
	zgetbucket(&lhszbucketptr);
	lhszvectorptr->dstype = governingdstype;
	lhszvectorptr->zds.ss.bucketptr = lhszbucketptr;
	lhszvectornonzeroes = 0;
	
	for (k=0;k<arity;k++)
	  condition[k] = rhsvectorconsumednonzeroes[k] <
	      rhsvectorptr[k]->nonzeroes;
	
	while (condition[0] || condition[1]) {
	  if (condition[0] && condition[1])
	    if (rhsbucketptr[0]->index[rhsbucketindex[0]] <=
		rhsbucketptr[1]->index[rhsbucketindex[1]])
	      k = REAL_PART;
	    else
	      k = IMAG_PART;
	  else if (condition[0])
	    k = REAL_PART;
	  else
	    k = IMAG_PART;
	  
	  j = rhsbucketptr[k]->index[rhsbucketindex[k]];
	  
	  lhszbucketindex = MY_MOD(lhszvectornonzeroes,
					  sBUCKET_DIM_minus1);
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.1. Found 1st or new subscript for  *
	   *      lhszvector.  If needed,  *
	   *      get new bucket for              *
	   *      lhszvector.              *
	   *      Consume nonzero in saxpy.       *
	   ****************************************/
	  if (lhszvectornonzeroes == 0 ||
	      lhszbucketptr->index[lastlhszbucketindex] < j) {
	    if (lhszbucketindex == 0 &&
		lhszvectornonzeroes != 0) {
	      zgetbucket(&(lhszbucketptr->nextbucket));
	      lhszbucketptr = lhszbucketptr->nextbucket;
	    }
	    lhszbucketptr->index[lhszbucketindex] = j;
	    if (k == REAL_PART) {
	      XY_PUT_AND_CAST(lhszbucketptr->value[lhszbucketindex],
			      rhsbucketptr[k]->value[rhsbucketindex[k]],
			      0,
			      double);
	    }
	    else {
	      XY_PUT_AND_CAST(lhszbucketptr->value[lhszbucketindex],
			      0,
			      rhsbucketptr[k]->value[rhsbucketindex[k]],
			      double);
	    }
	    lastlhszbucketindex = lhszbucketindex;
	    lhszvectornonzeroes++;
	  }
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.2. Found old subscript for         *
	   *      lhszvector.                      *
	   *      Consume nonzero in saxpy.       *
	   ****************************************/
	  else if (lhszbucketptr->index[lastlhszbucketindex] == j) {
	    if (k == REAL_PART) {
	      XY_PUT_AND_CAST_PLUS_EQUALS(lhszbucketptr->value[lastlhszbucketindex],
					  rhsbucketptr[k]->value[rhsbucketindex[k]],
					  0,
					  double);
	    }
	    else {
	      XY_PUT_AND_CAST_PLUS_EQUALS(lhszbucketptr->value[lastlhszbucketindex],
					  0,
					  rhsbucketptr[k]->value[rhsbucketindex[k]],
					  double);
	    }

	  }
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.3. Found wrong subscript for       *
	   *      lhszvector.  Abort.       *
	   ****************************************/
	  
#ifdef _DEBUG
	  else {
	    printf("sspg2 error 6: lhsznz=%d, lastlhszindex[]=%d, j=%d\n",
		   lhszvectornonzeroes,
		   lhszbucketptr->index[lastlhszbucketindex],j);
	    abort();
	  }
#endif
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.4. If there is one, get new        *
	   *      nonzero from the rhs vector     *
	   *      whose nonzero was just          *
	   *      consumed and replace the        *
	   *      heaproot with new entry;        *
	   *      Else, delete heaproot.          *
	   ****************************************/
	  rhsvectorconsumednonzeroes[k]++;
	  condition[k] = rhsvectorconsumednonzeroes[k] <
	    rhsvectorptr[k]->nonzeroes;
	  rhsbucketindex[k] =
	    MY_MOD(rhsvectorconsumednonzeroes[k],sBUCKET_DIM_minus1);
	  if (condition[k]) {
	    if (rhsbucketindex[k]  == 0) {
	      if ((rhsbucketptr[k] = rhsbucketptr[k]->nextbucket) 
		  == NULL) {
#ifdef _DEBUG
		printf("sspg2 err: unexpected NULL nextbucket: k=%d",k);
		printf("rhsvectorconsumednonzeros=%d,rhsvecnonzeroes=%d,",
		       rhsvectorconsumednonzeroes[k],
		       rhsvectornonzeroes[k]);
		abort();
#endif
	      }	  
	    }
	  }
	} /* matches     while loop     */
      } /* matches     else if (governingdstype == SS) {      */
      
      else /* if (governingdstype == CS) */ {
	
	/****************************************************************
	 *               _COMPRESSED_SUBSCRIPTS AXPY                    *
	 ****************************************************************/
	
	/****************************************
	 * Else perform saxpy:                  *
	 * 1. Put in heap the 1st nonzero of    *
	 *    each vector.                      *
	 ****************************************/
	for (k=0;k<arity;k++) {
	  rhsvectorconsumednonzeroranges[k] = 0;
	  rhsibucketindex[k] = 0;
	  rhsvbucketindex[k] = 0;
	  rhsibucketptr[k] =
	    rhsvectorptr[k]->sds.cs.ibucketptr;
	  rhsvbucketptr[k] =
	    rhsvectorptr[k]->sds.cs.vbucketptr;
	  
#ifdef _DEBUG
	  if (rhsibucketptr[k] == NULL ||
	      rhsvbucketptr[k] == NULL) {
	    printf("sspg2 error 10: Unexpected NULL rhsbucketptrs: ");
	    printf("vector # %d\n",k);
	    abort();
	  }
#endif
	}	  
	
	/****************************************
	 * Perform saxpy:                       *
	 * 2. Initialize lhszvector.             *
	 ****************************************/
	zgetibucket(&lhszibucketptr);
	lhszvectorptr->dstype = governingdstype;
	lhszvectorptr->zds.cs.ibucketptr = lhszibucketptr;
	zgetvbucket(&lhszvbucketptr);
	lhszvectorptr->zds.cs.vbucketptr = lhszvbucketptr;
	lhszvectornonzeroranges = 0;
	lhszvectornonzeroes = 0;
	lhszvbucketindex = 0;
	
	/****************************************
	 * Perform saxpy:                       *
	 * 3. Get nonzero by popping the        *
	 *    heaproot; consume it in saxpy;    *
	 *    either replace or delete heaproot.*
	 ****************************************/
	
	while (rhsvectorconsumednonzeroranges[0] <
	       rhsvectorptr[0]->sds.cs.nonzeroranges ||
	       rhsvectorconsumednonzeroranges[1] <
	       rhsvectorptr[1]->sds.cs.nonzeroranges) {
	  if
	    ((rhsvectorconsumednonzeroranges[1] ==
	      rhsvectorptr[1]->sds.cs.nonzeroranges) ||
	     ((rhsvectorconsumednonzeroranges[0] <
	       rhsvectorptr[0]->sds.cs.nonzeroranges) &&
	      (rhsibucketptr[0]->index_RANGE_BEGIN[rhsibucketindex[0]] <=
	       rhsibucketptr[1]->index_RANGE_BEGIN[rhsibucketindex[1]])))
	      k = REAL_PART;
	  else if
	    ((rhsvectorconsumednonzeroranges[0] ==
	      rhsvectorptr[0]->sds.cs.nonzeroranges) ||
	     ((rhsvectorconsumednonzeroranges[1] <
	       rhsvectorptr[1]->sds.cs.nonzeroranges) &&
	      (rhsibucketptr[0]->index_RANGE_BEGIN[rhsibucketindex[0]] >
	       rhsibucketptr[1]->index_RANGE_BEGIN[rhsibucketindex[1]])))
	      k = IMAG_PART;
#ifdef _DEBUG
	  else {
	    printf("sspg2 error: BAD LOGIC.\n");
	    abort();
	  }
#endif
	  j1 = rhsibucketptr[k]->index_RANGE_BEGIN[rhsibucketindex[k]];
	  j2 = rhsibucketptr[k]->index_RANGE_END[rhsibucketindex[k]];
	  rhsibucketindex[k] =
	    MY_MOD(rhsvectorconsumednonzeroranges[k],sIBUCKET_DIM_minus1);
	  lhszibucketindex =
	    MY_MOD(lhszvectornonzeroranges,sIBUCKET_DIM_minus1);
	  
	  if ((rhsvbucketindex[k] =
	       MY_MOD(rhsvbucketindex[k],sVBUCKET_DIM_minus1))==0
	      && rhsvectorconsumednonzeroranges[k] !=0) {
	    sHOP_TO_NEXT_RHSVBUCKET(k,&rhsvbucketindex[k],
				   rhsvectorconsumednonzeroranges[k],
				   rhsvectornonzeroranges[k]);
	  }
	  lhszvbucketindex = MY_MOD(lhszvbucketindex,
					   sVBUCKET_DIM_minus1);
	  
#ifdef _DEBUG
	  if (rhsibucketptr[k]->errorflag != 9099 ||
	      rhsvbucketptr[k]->errorflag != 9009) {
	    printf("sspg2 err 14: memory corruption of ibucketstore ");
	    printf("or vbucketstore.\n");
	    abort();
	  }
#endif      
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.0. Check for startup errors.       *
	   ****************************************/
	  
#ifdef _DEBUG
	  if (lhszvectornonzeroranges == 0) {
	    if (rhsibucketindex[k] != 0 || rhsvbucketindex[k] != 0 ||
		lhszibucketindex != 0 || lhszvbucketindex != 0) {
	      printf("sspg2 error 15: Zeroes expected: \n\t");
	      printf("rhsibucketindex[k]=%d, rhsvbucketindex[k]=%d\n\t",
		     rhsibucketindex[k], rhsvbucketindex[k]);
	      printf("lhszibucketindex=%d, lhszvbucketindex=%d\n\t",
		     lhszibucketindex,lhszvbucketindex);
	      abort();
	    }
	  }
#endif      
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.1. Found 1ST or NEW subscript      *
	   *      range (non-overlapping with     *
	   *      previous one, if any) for       *
	   *      lhszvector.  If needed, get new  *
	   *      ibuckets and vbuckets for       *
	   *      lhszvector.                      *
	   ****************************************/
	  if (lhszvectornonzeroranges == 0 ||
	      j1 > lhszibucketptr->index_RANGE_END[lastlhszibucketindex]+1) {/*A.1*/
	    
	    if (lhszibucketindex == 0 &&
		lhszvectornonzeroranges != 0) {
	      zgetibucket(&(lhszibucketptr->nextibucket));
	      lhszibucketptr = lhszibucketptr->nextibucket;
	    }
	    lhszibucketptr->index_RANGE_BEGIN[lhszibucketindex] = j1;
	    lhszibucketptr->index_RANGE_END[lhszibucketindex] = j2;
	    
	    if (lhszvbucketindex == 0 &&
		lhszvectornonzeroranges != 0)
	      sget_and_hop_to_next_to_z_vbucket(&lhszvbucketindex);
	    
	    lastlhszvbucketptr = lhszvbucketptr;
	    lastlhszvbucketindex = lhszvbucketindex;
	    
	    totalcurrentnonzerorange = j2-j1+1;
	    
	    lhszvectornonzeroes += totalcurrentnonzerorange;
	    lhszvectornonzeroranges++;
	    
	    
	    sNON_OVER_LAPPING_CS_RANGE_CONVERT_TO_z(COMPLETELY_NEW_RANGE,
					      lhszibucketindex,
					      &lastlhszibucketindex,
					      &lhszvbucketindex,
					      k,
					      &rhsvbucketindex[k],
					      totalcurrentnonzerorange,
					      &unconsumedcurrentnonzerorange,
					      rhsvectorconsumednonzeroranges[k],
					      rhsvectornonzeroranges[k],
					      lhszvectornonzeroranges);
	    
	    
	  } /*A.1*/
	  
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.2. Found old subscript for         *
	   *      lhszvector.  Consume  nonzero in *
	   *      saxpy.                          *
	   *                                      *
	   ****************************************/
	  else if (j1 <= 
		   lhszibucketptr->index_RANGE_END[lastlhszibucketindex]+1 &&
		   j1 >= 
		   lhszibucketptr->index_RANGE_BEGIN[lastlhszibucketindex]) { /*B.1*/
	    
	    templhszvbucketptr = lhszvbucketptr;
	    templhszvbucketindex = lhszvbucketindex;
	    lhszvbucketptr = lastlhszvbucketptr;
	    
	    distancetooverlap =
	      j1 - lhszibucketptr->index_RANGE_BEGIN[lastlhszibucketindex];
	    distancetoendofvbucket = sVBUCKET_DIM -
	      lastlhszvbucketindex;
	    
	    if (j1 ==
		lhszibucketptr->index_RANGE_END[lastlhszibucketindex]+1) {
	      h1 = 0;
	      h2 = -1;
	    }
	    else if (j2 <=
		     lhszibucketptr->index_RANGE_END[lastlhszibucketindex]) {
	      h1 = j1;
	      h2 = j2;
	      j1 = 0;
	      j2 = -1;
	    }
	    else {
	      h1 = j1;
	      h2 = lhszibucketptr->index_RANGE_END[lastlhszibucketindex];
	      j1 = h2+1;
	    }
	    
	    if (distancetoendofvbucket > distancetooverlap) {
	      lhszvbucketindex =
		MY_MOD(lastlhszvbucketindex + distancetooverlap,
		       sVBUCKET_DIM_minus1);
	    }
	    else if (distancetoendofvbucket < distancetooverlap) {
	      lhszvbucketindex =
		MY_MOD(distancetooverlap - distancetoendofvbucket,
		       sVBUCKET_DIM_minus1);
	    }
	    else {
	      lhszvbucketindex = 0;
	    }
	    
	    while (distancetooverlap >= distancetoendofvbucket &&
		   lhszvbucketptr->nextvbucket != NULL) {
	      lhszvbucketptr = lhszvbucketptr->nextvbucket;
	      distancetooverlap -= distancetoendofvbucket;
	      distancetoendofvbucket = sVBUCKET_DIM;
	    }
	    
	    lhszibucketptr->index_RANGE_END[lastlhszibucketindex] =
	      max(j2,lhszibucketptr->index_RANGE_END[lastlhszibucketindex]);
	    
	    if ((totalcurrentnonzerorange = h2-h1+1) > 0) {
	      sOVERLAPPING_CS_RANGE_CONVERT_TO_z(&lhszvbucketindex,
					   k,
					   &rhsvbucketindex[k],
					   totalcurrentnonzerorange,
					   &unconsumedcurrentnonzerorange,
					   rhsvectorconsumednonzeroranges[k],
					   rhsvectornonzeroranges[k],
					   lhszvectornonzeroranges);
	    }
	    
	    /*
	      
	      This bit of logic must be performed, whether or not there
	      is additional nonzeroes to process.
	      */
	    
	    if (j2 != -1 &&
		(lhszvbucketptr != templhszvbucketptr ||
		 (lhszvbucketindex =
		  MY_MOD(lhszvbucketindex,sVBUCKET_DIM_minus1)) !=
		 templhszvbucketindex)) {
	      
#ifdef _DEBUG
	      printf("sspg2 err 50.0: Unexpectd value discrepancies\n");
	      printf("lhszvbucketptr=%d, templhszvbucketptr=%d\n",
		     (int)lhszvbucketptr,
		     (int)templhszvbucketptr);
	      printf("lhszvbucketindex=%d, templhszvbucketindex=%d\n",
		     lhszvbucketindex,templhszvbucketindex);
	      
	      abort();
#endif
	    }
	    else {
	      lhszvbucketptr = templhszvbucketptr;
	      lhszvbucketindex = templhszvbucketindex;
	    }
	    
	    
	    if ((totalcurrentnonzerorange = j2-j1+1) > 0) {
	      lhszvectornonzeroes += totalcurrentnonzerorange;
	      
	      /*
	       * adding new entries to vbucket
	       */
	      
	      if (lhszvbucketindex == 0 &&
		  lhszvectornonzeroranges != 0 &&
		  j2 != -1)
		sget_and_hop_to_next_to_z_vbucket(&lhszvbucketindex);
	      
	      if(rhsvbucketindex[k] == sVBUCKET_DIM && j2 != -1) {
		
		sHOP_TO_NEXT_RHSVBUCKET(k,&rhsvbucketindex[k],
				       rhsvectorconsumednonzeroranges[k],
				       rhsvectornonzeroranges[k]);
	      }
	      
	      
	      
	      
	      sNON_OVER_LAPPING_CS_RANGE_CONVERT_TO_z(NEW_ENTRIES_PADDED_TO_OLD_RANGE,
						lhszibucketindex,
						&lastlhszibucketindex,
						&lhszvbucketindex,
						k,
						&rhsvbucketindex[k],
						totalcurrentnonzerorange,
						&unconsumedcurrentnonzerorange,
						rhsvectorconsumednonzeroranges[k],
						rhsvectornonzeroranges[k],
						lhszvectornonzeroranges);
	      
	      
	    } /* matches     if ((totalcurrent...) { */
	  } /*B.1*/
	  
	  
	  /****************************************	
	   * Perform saxpy:                       *
	   * 3.3. Found wrong subscript for       *
	   *      lhszvector.  Abort.               *
	   ****************************************/
	  
#ifdef _DEBUG
	  else {
	    printf("sspg2 error 60: lhszvectornonzeroranges=%d,",
		   lhszvectornonzeroranges);
	    printf("lastlhszibucketindex=%d, \n\t",lastlhszibucketindex);
	    printf("lhszibucketptr->index[%d] = (%d,%d), j1=%d\n",
		   lastlhszibucketindex,
		   lhszibucketptr->index_RANGE_BEGIN[lastlhszibucketindex],
		   lhszibucketptr->index_RANGE_END[lastlhszibucketindex],
		   j1);
	    abort();
	  }
#endif
	  
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.4. If there is one, get new        *
	   *      nonzero from the rhs vector     *
	   *      whose nonzero was just          *
	   *      consumed and replace the        *
	   *      heaproot with new entry;        *
	   *      Else, delete heaproot.          *
	   ****************************************/
	  if (++rhsvectorconsumednonzeroranges[k] <
	      rhsvectornonzeroranges[k]) {
	    if ((rhsibucketindex[k] =
		 MY_MOD(rhsvectorconsumednonzeroranges[k],
			sIBUCKET_DIM_minus1)) == 0) {
	      if ((rhsibucketptr[k] = rhsibucketptr[k]->nextibucket) ==
		  NULL) {
		
#ifdef _DEBUG
		printf("sspg2 err 61: unexpected NULL nextvbucket: \n");
		printf("\t k=%d, rhsvectorconsumednonzeroranges=%d,\n",
		       k,rhsvectorconsumednonzeroranges[k]);
		printf("\t rhsvectornonzeroranges[k]=%d\n",
		       rhsvectornonzeroranges[k]);
		abort();
#endif
		
	      }
	    }
	  }   
	} /* matches     while loop          */
      } /* matches     else ** if (governingdstype == CS) ** {     */


      rhsvectorptrptrs[0] = &rhsvectorptr[0];
      rhsvectorptrptrs[1] = &rhsvectorptr[1];
      
      saxpy_2_heap_to_z_post_processing(destptrptr,
					 rhsvectorptrptrs,
					 restorerhsvectorptr,
					 arity,
					 lhszvectornonzeroes,
					 lhszvectornonzeroranges,
					 governingdstype);
      
    } /* matches     if we are going to use lhszvectorptr       */
  }
}





