#define SINGLE
/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#include  <stdlib.h>
#include "sgaxpy_pd_macros_and_data_structures.h"
#include "sgaxpy_pd_global_variables.h"


/****************************************************************************
 *                                                                          *
 *  sadallocate()                                                            *
 *                                                                          *
 *      sadallocate calls the bucket and vector store allocation functions.  *
 *      It should be called by the user once, before any calls to sspsd &  *
 *      adxpyNs.                                                            *
 *                                                                          *
 ****************************************************************************/

void sadallocate()
{
  int k;
  
  sinitialize_pd_global_variables();
  sallocatevectorstore();
  sallocatebucketstore();
  sallocateibucketstore();
  sallocatevbucketstore();

  sgetvector(&lhssvectorptr);
  for (k=0;k<MAXARITY+1;k++)
    sgetvector(&holdingbinsvectorptr[k]);
  sgetvector(&tempsdestptr);
  sgetvector(&restofplusequalsvectorptr);
  restofplusequalsvectorptr->dstype = CS;
}



/****************************************************************************
 *                                                                          *
 *   sadfree()                                                               *
 *                                                                          *
 *      sadfree frees all the dynamically allocated space for a given        *
 *      percision.                                                          *
 *                                                                          *
 ****************************************************************************/

void sadfree()
{
  svectorstorestruct  *remaindervectorstoreptr,  *tempvectorstoreptr;
  sbucketstorestruct  *remainderbucketstoreptr,  *tempbucketstoreptr;
  sibucketstorestruct *remainderibucketstoreptr, *tempibucketstoreptr;
  svbucketstorestruct *remaindervbucketstoreptr, *tempvbucketstoreptr;

  svectorstruct *tempvectorptr;
  sbucketstruct *tempbucketptr;
  sibucketstruct *tempibucketptr;
  svbucketstruct *tempvbucketptr;

  int i;

  for(remaindervectorstoreptr=svectorstorestartptr;
      remaindervectorstoreptr!=NULL;
      remaindervectorstoreptr=tempvectorstoreptr) {

    tempvectorstoreptr = remaindervectorstoreptr->nextstore;
    free(remaindervectorstoreptr);
  }

  for(remainderbucketstoreptr=sbucketstorestartptr;
      remainderbucketstoreptr!=NULL;
      remainderbucketstoreptr=tempbucketstoreptr) {
    tempbucketstoreptr = remainderbucketstoreptr->nextstore;

    /*
     * Freeing based on per BUCKET Dynamic Allocation.
     * Here we free the memory which has been allocated per bucket
     * in sallocatebucketstore().
     */
    for (i=0; i<sBUCKETSTORE_DIM; i++) {
      tempbucketptr = &(remainderbucketstoreptr->bucket[i]);
      free(tempbucketptr->index);
      free(tempbucketptr->value);
    }
  
    /*
     * Now regardless of the allocation method above, we have freed
     * everything hanging off our bucketstore, and we can go ahead and
     * free the bucketstore itself.
     */
    free(remainderbucketstoreptr);
    
  }
  
  for(remainderibucketstoreptr=sibucketstorestartptr;
      remainderibucketstoreptr!=NULL;
      remainderibucketstoreptr=tempibucketstoreptr) {
    tempibucketstoreptr = remainderibucketstoreptr->nextstore;
    
    /*
     * Freeing based on per IBUCKET Dynamic Allocation.
     */
    for (i=0; i<sIBUCKETSTORE_DIM; i++) {
      tempibucketptr = &(remainderibucketstoreptr->ibucket[i]);
      free(tempibucketptr->index_RANGE_BEGIN);
      free(tempibucketptr->index_RANGE_END);
    }
    
    /*
     * Now regardless of the allocation method above, we have freed
     * everything hanging off our ibucketstore, and we can go ahead and
     * free the ibucketstore itself.
     */
    free(remainderibucketstoreptr);
  }
  
  for(remaindervbucketstoreptr=svbucketstorestartptr;
      remaindervbucketstoreptr!=NULL;
      remaindervbucketstoreptr=tempvbucketstoreptr) {
    tempvbucketstoreptr = remaindervbucketstoreptr->nextstore;
    
    /*
     * Freeing based on per VBUCKET Dynamic Allocation.
     */
    for (i=0; i<sVBUCKETSTORE_DIM; i++) {
      tempvbucketptr = &(remaindervbucketstoreptr->vbucket[i]);
      free(tempvbucketptr->value);
    }
    
    /*
     * Now regardless of the allocation method above, we have freed
     * everything hanging off our vbucketstore, and we can go ahead and
     * free the vbucketstore itself.
     */
    free(remaindervbucketstoreptr);
  }
}



/****************************************************************************
 *                                                                          *
 *   sad_free_vectors_buckets(svectorptrptr)                             *
 *                                                                          *
 *       sad_free_vectors_buckets frees the buckets of svectorptrptr.    *
 *       BUT IT DOES NOT free svectorptrptr, itself.                    *
 *       Note that freedsbucketslist, freedsibucketslist, and       *
 *       freedsvbucketslist are global variables included in            *
 *       $(SAXPY_FLAVOR)_global_variables.h.                                *
 *                                                                          *
 *       svectorptrptr  : The vector to be freed.                       *
 *                                                                          *
 ****************************************************************************/

void sad_free_vectors_buckets(svectorptrptr)
svectorstruct **svectorptrptr;
{

  svectorstruct *svectorptr;
  sbucketstruct *sbucketptr;
  sibucketstruct *sibucketptr;
  svbucketstruct *svbucketptr;

  svectorptr = (svectorstruct *) *svectorptrptr;
  
  if (svectorptr != NULL) {
    if (svectorptr->dstype == S1) {
      /* do nothing */
    }
    else if (svectorptr->dstype == SS) {
      /**********************************************************************
       *                        Single Subscript                            *
       **********************************************************************/
      
      sbucketptr = svectorptr->sds.ss.bucketptr;
      if (sbucketptr != NULL) {
	while (sbucketptr->nextbucket != NULL)
	  sbucketptr = sbucketptr->nextbucket;
	sbucketptr->nextbucket = freedsbucketslist;
	RESET_freedsbucketslist(svectorptr->sds.ss.bucketptr);
	svectorptr->sds.ss.bucketptr = NULL;
      }
      
    } else /* svectorptr->sdstype = CS */ {
      /**********************************************************************
       *                      Compressed Subscript                          *
       **********************************************************************/
      
      sibucketptr = svectorptr->sds.cs.ibucketptr;
      if (sibucketptr != NULL) {
	while (sibucketptr->nextibucket != NULL)
	  sibucketptr = sibucketptr->nextibucket;
	sibucketptr->nextibucket = freedsibucketslist;
	RESET_freedsibucketslist(svectorptr->sds.cs.ibucketptr);
	svectorptr->sds.cs.ibucketptr = NULL;
      }
      svbucketptr = svectorptr->sds.cs.vbucketptr;
      if (svbucketptr != NULL) {
	while (svbucketptr->nextvbucket != NULL)
	  svbucketptr = svbucketptr->nextvbucket;
	svbucketptr->nextvbucket = freedsvbucketslist;
	RESET_freedsvbucketslist(svectorptr->sds.cs.vbucketptr);
	svectorptr->sds.cs.vbucketptr = NULL;
      }
    }
  }

/*
 * PMK: begin 12/4/96 edits
 *
 * In order to ensure that svectorptr points to a null vectorstruct,
 * i.e., a vectorstruct whose fields all have zero values, we assign to it
 * the contents of nullsvector which is a global initialized with all
 * zero members in gaxpy_pd_global_variables.h.
 *
 */

  *(svectorptr) = nullsvector;

/*
 * PMK: end 12/4/96 edits
 */

}


/****************************************************************************
 *                                                                          *
 *    sinitialize_pd_global_variables()                                      *
 *                                                                          *
 *    sinitialize_pd_global_variables performs initializations for some      *
 *    precision dependent globals.                                          *
 *                                                                          *
 ****************************************************************************/

void sinitialize_pd_global_variables()
{
  sbucketstorestartptr=NULL;
  freedsbucketslist=NULL;
  sibucketstorestartptr=NULL;
  freedsibucketslist=NULL;
  svbucketstorestartptr=NULL;
  freedsvbucketslist=NULL;
  svectorstorestartptr=NULL;

}

