#define SINGLE
/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#ifndef sGAXPY_PD_GLOBAL_VARIABLES

#define sGAXPY_PD_GLOBAL_VARIABLES

#include "gaxpy_pi_global_variables.h"

#ifdef DEFINE_GLOBALS
#define GLOBAL
#else
#define GLOBAL extern
#endif


/****************************************************************************
 *                                                                          *
 *  Global Variables                                                        *
 *                                                                          *
 ****************************************************************************/

/******************
 *  DIMENSIONS    *
 ******************/

GLOBAL int sBUCKET_DIM;            /* Max. # of (index,value) pairs    *
				        * in an SS bucket.                 */
GLOBAL int sBUCKET_DIM_minus1;     /* sBUCKET_DIM - 1              */

GLOBAL int sVBUCKET_DIM;           /* Max. # of values in a            *
                                        * in an CS vbucket.                */
GLOBAL int sVBUCKET_DIM_minus1;    /* sVBUCKET_DIM - 1             */

GLOBAL int sIBUCKET_DIM;           /* Max. # of                        *
					*(indexrange_begin,indexrange_end) *
				        * pairs in an CS ibucket.          */
GLOBAL int sIBUCKET_DIM_minus1;    /* sIBUCKET_DIM - 1             */

/******************
 *  BUCKET STUFF  *
 ******************/

GLOBAL sbucketstorestruct *sbucketstoreptr;
                                             /* Address of the current      *
                                              * (most recently allocated)   *
                                              * bucketstore.                */

GLOBAL sbucketstorestruct *sbucketstorestartptr;
                                             /* Address of the first        *
                                              * allocated bucketstore.      */

GLOBAL int sbucketstorecounter;          /* Array index of the next     *
                                              * available bucket in current *
                                              * bucketstore.                */

GLOBAL sbucketstruct *freedsbucketslist;
                                             /* Address of the first bucket *
                                              * in the linked list of       *
                                              * previously-used, but        *
                                              * currently-available         *
                                              * buckets.  This is a LIFO    *
					      * list (Last In First Out).   */

/******************
 *  IBUCKET STUFF *
 ******************/

GLOBAL sibucketstorestruct *sibucketstoreptr;
                                            /* Address of the current       *
                                             * (most recently allocated)    *
                                             * ibucketstore.                */

GLOBAL sibucketstorestruct *sibucketstorestartptr;
                                            /* Address of the first         *
                                             * allocated ibucketstore.      */

GLOBAL int sibucketstorecounter;        /* Array index of the next      *
                                             * available ibucket in current *
                                             * ibucketstore.                */

GLOBAL sibucketstruct *freedsibucketslist;
                                            /* Address of the first ibucket *
                                             * in the linked list of        *
                                             * previously-used, but         *
					     * currently-available          *
					     * ibuckets.  This is a LIFO    *
					     * list (Last In First Out).    */


/******************
 *  VBUCKET STUFF *
 ******************/

GLOBAL svbucketstorestruct *svbucketstoreptr;
                                             /* Address of the current      *
                                              * (most recently allocated)   *
                                              * vbucketstore.               */

GLOBAL svbucketstorestruct *svbucketstorestartptr;
                                               /* Address of the first      *
                                                * allocated vbucketstore.   */

GLOBAL int svbucketstorecounter;         /* Array index of the next     *
                                              * available vbucket in        *
                                              * current vbucketstore.       */

GLOBAL svbucketstruct *freedsvbucketslist;
                                             /* Address of the first        *
                                              * vbucket in the linked list  *
                                              * of previously-used, but     *
                                              * currently-available         *
                                              * vbuckets.  This is a LIFO   *
					      * list (Last In First Out).   */

GLOBAL svbucketstruct *rhssvbucketptr[MAXARITY];
                                      /* Array of pointers to "current"     *
                                       * vbuckets where "current" nonzeroes *
                                       * of the rhs vectors reside.         */


GLOBAL svbucketstruct *lhssvbucketptr;
                                      /* Pointer to current vbucket of      *
				       * lhsvector.                         *
				       * This variable is declared globally *
				       * (i.e., cheaply and not a la        *
				       * structural programming), since it  *
				       * is used in sspgx and the "hop" *
				       * routines it calls.  A couple of    *
				       * other routines use it as local     *
				       * variable, namely adpeep and adxpy. */

/******************
 *  VECTOR STUFF  *
 ******************/

GLOBAL svectorstorestruct *svectorstoreptr;
                                             /* Address of the current      *
                                              * vectorstore.                */

GLOBAL svectorstorestruct *svectorstorestartptr;
                                             /* Address of the first        *
                                              * allocated vectorstore.      */

GLOBAL int svectorstorecounter;          /* Array index of the next     *
                                              * available vector in current *
                                              * vectorstore.                */


GLOBAL svectorstruct *lhssvectorptr;
                                      /* The first time thru sspgx,     *
				       * lhsvectorptr is allocated a vector.*
				       * vector.  Subsequently, everytime   *
				       * thru sspgx, lhsvectorptr will  *
				       * house the result of the saxpy      *
				       * operation, then switch places with *
				       * destptrptr.                        */

                                      /* In the fcs/gaxpy versions, since   *
				       * more than one subroutine use       *
				       * lhsvectorptr, it is declared as a  *
				       * global rather than as a static,    *
				       * and also since the first adxpy     *
				       * maybe a non-heap adxpy,            *
				       * lhsvectorptr must be intially      *
				       * malloced wherever it will be used  *
				       * first.                             */


GLOBAL svectorstruct *holdingbinsvectorptr[MAXARITY+1];

GLOBAL svectorstruct *tempsdestptr;

GLOBAL svectorstruct *restofplusequalsvectorptr; 
/*
GLOBAL int initialallocationflag_svectorptrs;
                                             ** Flag used in sspgx and  **
					     ** non-heap adxpys.            **
*/

/*
 * PMK: begin 12/4/96 edits
 */

GLOBAL svectorstruct nullsvector;
                                      /* By declaring nullsvector       *
                                       * in this include file and by virtue *
                                       * of the fact that this declaration  *
                                       * will appear in the header of all   *
                                       * files where nullsvector is     *
                                       * used, nullsvector is a static  *
                                       * variable, and will be initialized  *
                                       * to zero by default (i.e., all its  *
                                       * member fields will be set to zero).*
                                       *                                    *
                                       * Note that nullsvector is not a *
                                       * pointer to a vectorstruct, but is  *
                                       * itself a vectorstruct.             *
                                       *                                    *
                                       * nullsvector is used in the     *
                                       * sad_free_vectors_buckets routine as *
                                       * a way of assigning zero to all     *
                                       * fields of the freed vector.        */

/*
 * PMK: end 12/4/96 edits
 */

#endif   /*  #ifndef sGAXPY_PD_GLOBAL_VARIABLES   */



