#define SINGLE
/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#include <stdlib.h>
#include <string.h>
#include "sgaxpy_pd_macros_and_data_structures.h"
#include "sgaxpy_pd_global_variables.h"


/****************************************************************************
 *                                                                          *
 *  sspsd(destptrptr,indexvector,valuevector,lengthptr)                    *
 *                                                                          *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/

void sspsd(destptrptr,indexvector,valuevector,lengthptr)

float *valuevector;
int *indexvector,*lengthptr;
svectorstruct **destptrptr;
{
  int i,j,tempindex,bucketindex,length=*lengthptr;
  svectorstruct *destptr;
  sbucketstruct *newbucketptr,*tailbucketptr;

  float tempvalue;

  int vbucketindex,nonzeroranges;
  svbucketstruct *newvbucketptr,*tailvbucketptr;
  int last_nz_range[2];

  svectorstruct *sstempvectorptr;
  sbucketstruct *sstempbucketptr;

/*
 *   Error Checking
 */

  if (length < 0) {
    printf("spsd error 1: Illegal vector length, length=%d\n",length);
    abort();
  }
  for (i=0;i<length;i++) {
    if (indexvector[i] <= 0) {
      printf("spsd error 2: Illegal index value, index=%d\n",
	     indexvector[i]);
      abort();
    }
  }

/*
 *   Bubblesort indexvector (and with resultant one-to-one correspondancy,
 *   valuevector).
 */

  for (i=length-1;i>0;i--) {
    for (j=length-1;j>=length-i;j--) {
      if (indexvector[j] < indexvector[j-1]) {
	tempindex = indexvector[j-1];
	X_ASSIGN(tempvalue, valuevector[j-1]);
	indexvector[j-1] = indexvector[j];
	X_ASSIGN(valuevector[j-1], valuevector[j]);
	indexvector[j] = tempindex;
	X_ASSIGN(valuevector[j], tempvalue);
      }
      else if (indexvector[j] == indexvector[j-1]) {
	printf("spsd error 3: Redundant index in indexvector, index=%d\n",
	       indexvector[j]);
	abort();
      }
    }
  }

/*
 *   Common code between _SINGLE_SUBSCRIPTS  and _COMPRESSED_SUBSCRIPTS
 *   data structures.
*/

  if (*destptrptr == NULL) {
    sgetvector(destptrptr);
  }
  else {
    sad_free_vectors_buckets(destptrptr);
  }
  destptr = *destptrptr;
  destptr->nonzeroes = length;

  if (length == 0) {
    destptr->dstype = S1;
  }
  else if (length == 1) {
    destptr->dstype = S1;
    destptr->sds.s1.index = indexvector[0];
    X_ASSIGN(destptr->sds.s1.value,valuevector[0]);
  }
  else {
    destptr->dstype = SS;
    for (i=0;i<length;i++) {
      if ((bucketindex = MY_MOD(i,sBUCKET_DIM_minus1)) == 0) {
	sgetbucket(&newbucketptr);
	if (destptr->sds.ss.bucketptr == NULL)
	  destptr->sds.ss.bucketptr = newbucketptr;
	else 
	  tailbucketptr->nextbucket = newbucketptr;
	tailbucketptr = newbucketptr;
      }
      tailbucketptr->index[bucketindex] = indexvector[i];
      X_ASSIGN(tailbucketptr->value[bucketindex],valuevector[i]);
    }

    /*
     * Harnessed Code, copied from saxpy_2_heap_post_processing.
     * If conversion of SS to CS at spsd time becomes an
     * issue, then this code segment code be unharnessed.


     *****************************************
     *  If necessary, convert SS to CS.     *
     *****************************************
    
    if ((INITIALIZED_MAXINDEX &&
	 ((float) (*destptrptr)->nonzeroes / (float) GLOBAL_MAXINDEX) >=
	 SWITCH_NZ_RATIO)     ||
	((*destptrptr)->nonzeroes >= SWITCH_NZ_VALUE)) {
    
      sstempvectorptr = (*destptrptr);
      sconvertdstype(destptrptr,MAXARITY,CS);
      
      sstempbucketptr = sstempvectorptr->sds.ss.bucketptr;
      if (sstempbucketptr != NULL) {
	while (sstempbucketptr->nextbucket != NULL)
	  sstempbucketptr = sstempbucketptr->nextbucket;
	sstempbucketptr->nextbucket = freedsbucketslist;
	RESET_freedsbucketslist(sstempvectorptr->sds.ss.bucketptr);
	sstempvectorptr->sds.ss.bucketptr = NULL;
      }
      holdingbinsvectorptr[MAXARITY] =  sstempvectorptr;
      }

      *
      * End of harnessed code.
      */

  }
  
  /*
   *   Obsoleted sspsd code for _COMPRESSED_SUBSCRIPTS.
   *
   *   We no longer initialize a vector as CS directly, but rather initialize
   *   the vector as SS, then if necessary, convert SS to CS, as seen above.


#elif _COMPRESSED_SUBSCRIPTS
  bucketindex = 0;
  vbucketindex=0;
  nonzeroranges=0;
  last_nz_range_RANGE_END = 0;

  if (length != 0) {
    last_nz_range_RANGE_BEGIN = indexvector[0];
  }

  destptr->vbucketptr = NULL;

  for (i=0;i<length;i++) {
    if ((bucketindex = MY_MOD(bucketindex,sBUCKET_DIM_minus1)) == 0) {
      sgetbucket(&newbucketptr);
      if (destptr->bucketptr == NULL)
	destptr->bucketptr = newbucketptr;
      else 
	tailbucketptr->nextbucket = newbucketptr;
      tailbucketptr = newbucketptr;
    }

    if ((vbucketindex = MY_MOD(vbucketindex,sVBUCKET_DIM_minus1)) == 0) {
      sgetvbucket(&newvbucketptr);
      if (destptr->vbucketptr == NULL)
	destptr->vbucketptr = newvbucketptr;
      else 
	tailvbucketptr->nextvbucket = newvbucketptr;
      tailvbucketptr = newvbucketptr;
    }

    X_ASSIGN(tailvbucketptr->value[vbucketindex],valuevector[i]);
    
    vbucketindex++;

    if (indexvector[i] > last_nz_range_RANGE_END + 1 ||
	last_nz_range_RANGE_END == 0) {
      last_nz_range_RANGE_BEGIN = indexvector[i];
      last_nz_range_RANGE_END = indexvector[i];
      tailbucketptr->index_RANGE_BEGIN[bucketindex] = indexvector[i];
      tailbucketptr->index_RANGE_END[bucketindex] = indexvector[i];
      bucketindex++;
      nonzeroranges++;
    }
    else if (indexvector[i] == last_nz_range_RANGE_END + 1) {
      last_nz_range_RANGE_END = indexvector[i];
      tailbucketptr->index_RANGE_END[bucketindex] = indexvector[i];
    }

#ifdef _DEBUG
    else {
      printf("spsd error 4: Nonsensical Nonzero_range, index=%d\n",
	     indexvector[j]);
      abort();
    }
#endif

  }

  destptr->nonzeroranges = nonzeroranges;


#endif
  
   *
   *   End of obsoleted sspsd code for _COMPRESSED_SUBSCRIPTS.
   */

}

/****************************************************************************
 *                                                                          *
 *    sspxd(vector,lengthptr,objectptrptr,returned_len,len_info)  *
 *                                                                          *
 *       vector       Pointer to an array containing values of type         *
 *                    float.  On exit, if len_info == 0,         *
 *                    vector(1:returned_len) will contain a dense           *
 *                    representation of objectptrptr.  If len_info != 0     *
 *                    vector is not referenced.                             *
 *                                                                          *
 *       lengthptr    Pointer to an integer specifying the length of the    *
 *                    destination array.                                    *
 *                                                                          *
 *       objectptrptr doubly-indirected pointer to a svectorstruct.     *
 *                                                                          *
 *       returned_len Index of largest nonzero entry in objectptrptr. This  *
 *                    value will always be returned, whether vector is      *
 *                    initialized or not. See description of len_info below.*
 *                                                                          *
 *       len_info     If returned_len is greater than lengthptr, will be set*
 *                    to -1, and vector is not referenced.  Otherwise,      *
 *                    len_info is set to 0, and vector(1:lengthptr)         *
 *                    is initialized to a dense copy of objectptrptr.       *
 *                                                                          *
 *       Dense_adextract takes a sparse vector, pointed to by *objectptrptr,*
 *       and extracts it in a dense format to a Fortran array, pointed to   *
 *       by vector.  All entries in *vector will have a value.  *lengthptr  *
 *       specifies the length of *vector, to ensure that the array is large *
 *       enough to handle all the entries of **objectptrptr.                *
 *                                                                          *
 *       NOTE TO POTENTIAL MODIFIERS OF THIS ROUTINE:                       *
 *         the routines sspxm and sspxa *
 *         are structurally identical to this routine, so make sure such    *
 *         changes, if any, are also reflected there.                       *
 *                                                                          *
 *      PMK: 12/4/96 edits:  Modifications reflect:                         *
 *        1. Taking into account the universe of possible states of inputs: *
 *           a. null sparsevector                : extract a zero vector    *
 *           b. S1 sparsevector with 0 nonzeroes : extract a zero vector    *
 *           c. for all other sparsevectors                                 *
 *               i. if lengthptr >= returned_len : extract vector           * 
 *              ii. if lengthptr <  returned_len : extract nothing          *
 *        2. Implementing a temporary fix for a bug in doing S1 operations. *
 *           The bug manifests itself by adding 1 or more false entries     *
 *           (of index MAXINT and value 0) to the tail of an SS or CS       *
 *           sparsevector. The fix is to ignore (not extract) those entries.*
 *                                                                          *
 ****************************************************************************/


void sspxd(vector,lengthptr,objectptrptr,returned_len,len_info)
svectorstruct **objectptrptr;
float *vector;
int *lengthptr, *returned_len, *len_info;
{
  svectorstruct *objectptr;
  sbucketstruct *bucketptr;
  sibucketstruct *ibucketptr;
  svbucketstruct *vbucketptr;
  int i,j,h,bucketindex,ibucketindex,vbucketindex,largestvectorsubscript;

  /*
   * PMK: 12/4/96 edits:
   *
   * running_nonzero_count is the number of legitimate nonzeroes (i.e.,
   * those whose index is not MAXINT) found in a forward sweep of the
   * sparsevector.
   *
   * number_of_nonzeroes is set to running_nonzero_count if a bad index
   * is found; otherwise is equal to objectptr->nonzeroes.
   *
   * returned_len_update is used to set (*returned_len) to the highest
   * non-false index in the sparsevector.
   *
   * (*returned_len) must always have a value when returned, so at outset,
   * we assign 0 to it.  Likewise, for *len_info.
   *
   */

  int running_nonzero_count;
  int number_of_nonzeroes;
  int returned_len_update;

  (*returned_len) = 0;
  (*len_info) = 0;

  if (*objectptrptr == NULL) {
    for(i=(*lengthptr)-1;i>=0;i--)
	    X_ZERO(vector[i]);
  }
  else {
    objectptr = (svectorstruct *) *objectptrptr;
    number_of_nonzeroes = objectptr->nonzeroes;

    if (objectptr->dstype == S1) {
      if (objectptr->nonzeroes == 0) {
        for(i=(*lengthptr)-1;i>=0;i--)
	    X_ZERO(vector[i]);
      }
      else {
        (*returned_len) = objectptr->sds.s1.index;

#ifdef _DEBUG
        if ( (*returned_len) == MAXINT ) {
	   printf("spxd error: S1 vector with %d nonzeroes has index=%d\n",
                     objectptr->nonzeroes,(*returned_len));
	   abort();
        }
#endif
	
        if ( (*returned_len) > (*lengthptr) ) {
	  (*len_info) = -1;
	  printf("spxd warning: Fortran array dim. (%d) is less than the\n",
	       (*lengthptr) );
	  printf("             largest index (%d) in the nonzero index-set.\n",
	       (*returned_len) );
        }
        else {
	  for(i=(*lengthptr)-1;i>=0;i--)
	    X_ZERO(vector[i]);
          X_ASSIGN(vector[objectptr->sds.s1.index-1],
		   objectptr->sds.s1.value);
        }
      }
    }
    else if (objectptr->dstype == SS) {
      
      /*********************************************************************
       *                _SINGLE_SUBSCRIPTS EXTRACTION                      *
       *********************************************************************/
      
      bucketindex = MY_MOD((objectptr->nonzeroes-1),sBUCKET_DIM_minus1);
      bucketptr = objectptr->sds.ss.bucketptr;
      if (bucketptr != NULL)
	while (bucketptr->nextbucket != NULL)
	  bucketptr = bucketptr->nextbucket;
      (*returned_len) = bucketptr->index[bucketindex];

      if ( (*returned_len) == MAXINT ) {

      /*
       * PMK: 12/4/96 edits:
       *
       * If a bad index (MAXINT) has crept into the index set,
       * then it is possible that more than 1 such entry exists, all of
       * them bunched at the tail of the sparsevector.
       * To fix this, we take a forward sweep through the index set
       * until the first bad index is encountered, and ignore the rest
       * of the entries.  
       *
       */

#ifdef _DEBUG
	printf("spxd warning: corrupted SS vector, index=%d\n",
                   (*returned_len));
#endif

      /*
       *   Forward Sweep: find number of legitimate nonzeroes,
       *                  then rejoin the control flow for extraction.
       */

        bucketptr = objectptr->sds.ss.bucketptr;
        bucketindex = 0;
        running_nonzero_count = 0;
        while (bucketptr->index[bucketindex] != MAXINT) {
          running_nonzero_count++;
	  returned_len_update = bucketptr->index[bucketindex];
          bucketindex++;
	  if((bucketindex = MY_MOD(bucketindex,
				     sBUCKET_DIM_minus1)) == 0) {

#ifdef _DEBUG
            if (bucketptr->nextbucket == NULL) {
              printf("spxd error: Unexpected NULL nextbucket pointer\n");
              abort();
	    }
#endif

	    bucketptr = bucketptr->nextbucket;
	  }
        }

#ifdef _DEBUG
        if (running_nonzero_count >= number_of_nonzeroes) {
           printf("spxd error: running_nonzero_count=%d should be < ",
                   running_nonzero_count);
           printf("number of nonzeroes=%d\n",number_of_nonzeroes);
           abort();
         }
#endif
        number_of_nonzeroes = running_nonzero_count;
	(*returned_len) = returned_len_update;
      }

      if ( (*returned_len) > (*lengthptr) ) {
	(*len_info) = -1;
	printf("spxd warning: The Fortran array dim. (%d) is less than the\n",
	       (*lengthptr) );
	printf("              largest index (%d) in the nonzero index-set.\n",
	       (*returned_len) );
      }
      else {
	for(i=(*lengthptr)-1;i>=0;i--)
	  X_ZERO(vector[i]);
	bucketptr = objectptr->sds.ss.bucketptr;
	for(i=0;i<number_of_nonzeroes;i++) {
	  if(((bucketindex=MY_MOD(i,sBUCKET_DIM_minus1)) == 0) && i!=0)
	    bucketptr = bucketptr->nextbucket;
	  X_ASSIGN(vector[bucketptr->index[bucketindex]-1],
		   bucketptr->value[bucketindex]);
	}
      }
    }
    else /* if (objectptr->dstype == CS) */ { 
      
      /*********************************************************************
       *              _COMPRESSED_SUBSCRIPTS EXTRACTION                    *
       *********************************************************************/
      
      ibucketindex = MY_MOD((objectptr->sds.cs.nonzeroranges-1),
			    sIBUCKET_DIM_minus1);
      ibucketptr = objectptr->sds.cs.ibucketptr;
      if (ibucketptr != NULL)
	while (ibucketptr->nextibucket != NULL)
	  ibucketptr = ibucketptr->nextibucket;
      (*returned_len) = ibucketptr->index_RANGE_END[ibucketindex];

      if ( (*returned_len) == MAXINT ) {

      /*
       * PMK: 12/4/96 edits:
       *
       * If a bad index (MAXINT) has crept into the index set,
       * then it is possible that more than 1 such entry exists, all of
       * them bunched at the tail of the sparsevector.
       * To fix this, we take two forward sweeps through the index set.
       * In the first sweep, we determine the true value of (*returned_len),
       * and if this is <= (*lengthptr), in the second sweep, we extract
       * the vector.
       *
       */

#ifdef _DEBUG
	printf("spxd warning: corrupted CS vector, index_RANGE_END=%d\n",
                   (*returned_len));
#endif

      /*
       *   First Sweep: find true value of (*returned_len)
       */

        ibucketptr = objectptr->sds.cs.ibucketptr;
        ibucketindex = 0;
        while (ibucketptr->index_RANGE_BEGIN[ibucketindex] != MAXINT) {

#ifdef _DEBUG
          if (ibucketptr->index_RANGE_END[ibucketindex] == MAXINT) {
            printf("spxd error: Inconsistantly corrupt CS vector, ");
            printf("index_RANGE_BEGIN=%d,  index_RANGE_END=%d\n",
                    ibucketptr->index_RANGE_BEGIN[ibucketindex],
                    ibucketptr->index_RANGE_END[ibucketindex]);
            abort();
          }
#endif
	  
	  returned_len_update = ibucketptr->index_RANGE_END[ibucketindex];

          ibucketindex++;
          if((ibucketindex = MY_MOD(ibucketindex,
                                     sIBUCKET_DIM_minus1)) == 0) {

#ifdef _DEBUG
            if (ibucketptr->nextibucket == NULL) {
              printf("spxd error: Unexpected NULL nextibucket pointer\n");
              abort();
            }
#endif

            ibucketptr = ibucketptr->nextibucket;
          }
        }

	(*returned_len) = returned_len_update;

      /*
       *   Second Sweep: extract
       */

       if ( (*returned_len) > (*lengthptr) ) {
	  (*len_info) = -1;
	  printf("spxd warning: Fortran array dim. (%d) is less than the\n",
	         (*lengthptr) );
	  printf("              largest index (%d) in the nonzero index set\n",
	         (*returned_len) );
        }
        else {
	  for(i=(*lengthptr)-1;i>=0;i--)
	    X_ZERO(vector[i]);

	  ibucketptr = objectptr->sds.cs.ibucketptr;
	  ibucketindex = 0;
          vbucketptr = objectptr->sds.cs.vbucketptr;
          vbucketindex = 0;
          h = 0;
          while (ibucketptr->index_RANGE_BEGIN[ibucketindex] != MAXINT) {
	    for(j=0; j <= (ibucketptr->index_RANGE_END[ibucketindex] - 
			   ibucketptr->index_RANGE_BEGIN[ibucketindex]);
	        h++,j++,vbucketindex++) {
	      if(((vbucketindex =
		   MY_MOD(vbucketindex,sVBUCKET_DIM_minus1)) == 0) && h!=0)
	        vbucketptr = vbucketptr->nextvbucket;
	      X_ASSIGN(vector[ibucketptr->index_RANGE_BEGIN[ibucketindex]+j-1],
		       vbucketptr->value[vbucketindex]);
	    }
            ibucketindex++;
	    if((ibucketindex=MY_MOD(ibucketindex,
				     sIBUCKET_DIM_minus1)) == 0)
	      ibucketptr = ibucketptr->nextibucket;
	  }
	}
      }

      else { /* no bad (MAXINT) entries in the CS sparsevector */

        if ( (*returned_len) > (*lengthptr) ) {
	  (*len_info) = -1;
	  printf("spxd warning: Fortran array dim. (%d) is less than the\n",
	         (*lengthptr) );
	  printf("              largest index (%d) in the nonzero index set\n",
	         (*returned_len) );
        }
        else {
	  for(i=(*lengthptr)-1;i>=0;i--)
	    X_ZERO(vector[i]);
	  ibucketptr = objectptr->sds.cs.ibucketptr;
	  vbucketptr = objectptr->sds.cs.vbucketptr;
	
	  vbucketindex = 0;
	
	  for(i=0,h=0;i<objectptr->sds.cs.nonzeroranges;i++) {
	    if(((ibucketindex=MY_MOD(i,sIBUCKET_DIM_minus1)) == 0) && i!=0)
	      ibucketptr = ibucketptr->nextibucket;
	    for(j=0; j <= (ibucketptr->index_RANGE_END[ibucketindex] - 
			   ibucketptr->index_RANGE_BEGIN[ibucketindex]);
	        h++,j++,vbucketindex++) {
	      if(((vbucketindex =
		   MY_MOD(vbucketindex,sVBUCKET_DIM_minus1)) == 0) && h!=0)
	        vbucketptr = vbucketptr->nextvbucket;
	    
	      X_ASSIGN(vector[ibucketptr->index_RANGE_BEGIN[ibucketindex]+j-1],
		       vbucketptr->value[vbucketindex]);
	    }
	  }
	}
      }
    }
  }
}



/****************************************************************************
 *                                                                          *
 *    sspxs (indvector,valvector,lengthptr,objectptrptr,         *
 *                      returned_len,len_info)                              *
 *                                                                          *
 *                                                                          *
 *       indvector    Pointer to a Fortran array, to hold indices of        *
 *                    non-zero entries.                                     *
 *                                                                          *
 *       valvector    Pointer to a Fortran array, to hold non-zero          *
 *                    entries.                                              *
 *                                                                          *
 *       lengthptr    Pointer to an integer specifying the length of the    *
 *                    destination arrays indvector and valvector.           *
 *                                                                          *
 *       objectptrptr Doubly-indirected pointer to a svectorstruct.     *
 *                    Points to the array to be extracted.                  * 
 *                                                                          *
 *       returned_len Index of largest nonzero entry in objectptrptr. This  *
 *                    value will always be returned, whether vector is      *
 *                    initialized or not. See description of len_info below.*
 *                                                                          *
 *       len_info     If returned_len is greater than lengthptr, will be set*
 *                    to -1, and vector is not referenced.  Otherwise,      *
 *                    len_info is set to 0, and vector(1:lengthptr)         *
 *                    is initialized to a dense copy of objectptrptr.       *
 *                                                                          *
 *       Sparse_adextract extracts the vector pointed to by *objectptrptr   *
 *       to two Fortran arrays, *valvector and *indvector - an index array  *
 *       and a value array.  *returned_len on exit will hold the number of  *
 *       non-zero entries in the array.  *Lengthptr is an integer passed    *
 *       in as the length of valvector and indvector, for boundary-checking *
 *       purposes.  On exit, indvector holds the indices for the non-zero   *
 *       entries in valvector.                                              *
 *                                                                          *
 ****************************************************************************/

void sspxs(indvector,valvector,lengthptr,objectptrptr,
		      returned_len,len_info) 
svectorstruct **objectptrptr;
float *valvector;
int *lengthptr,*indvector,*returned_len,*len_info;
{
  svectorstruct *objectptr;               /* De-referenced pointers */
  sbucketstruct *bucketptr;              
  svbucketstruct *vbucketptr;            
  sibucketstruct *ibucketptr;
  int h,i,j,bucketindex,ibucketindex,vbucketindex;

  int running_nonzero_count;
  int number_of_nonzeroes;
  int largest_index;

  (*returned_len) = 0;
  (*len_info) = 0;

  if (*objectptrptr == NULL) {
    /* do nothing */
  }
  else {
    objectptr = (svectorstruct *) *objectptrptr;
    number_of_nonzeroes = objectptr->nonzeroes;

    if (objectptr->dstype == S1) {
      if (objectptr->nonzeroes == 0) {
        /* do nothing */
      }
      else {
	(*returned_len) = objectptr->nonzeroes;
        largest_index = objectptr->sds.s1.index;

#ifdef _DEBUG
        if ( largest_index == MAXINT ) {
	   printf("spxs error: S1 vector with %d nonzeroes has index=%d\n",
                     objectptr->nonzeroes,largest_index);
	   abort();
        }
#endif

	if ( (*returned_len) > (*lengthptr) ) {
	  (*len_info) = -1;
	  printf("spxs warning: Fortran array dim. (%d) is less than the\n",
	         (*lengthptr) );
	  printf("              largest index (%d) in the nonzero index-set\n",
	         (*returned_len) );
        }
        else {
	  X_ASSIGN(valvector[0], objectptr->sds.s1.value);
	  indvector[0] = objectptr->sds.s1.index;
        }
      }
    }
    else if (objectptr->dstype == SS) {

      /********************************************************************
       *                _SINGLE_SUBSCRIPTS EXTRACTION                     *
       ********************************************************************/

      bucketindex = MY_MOD((objectptr->nonzeroes-1),sBUCKET_DIM_minus1);
      bucketptr = objectptr->sds.ss.bucketptr;
      if (bucketptr != NULL)
	while (bucketptr->nextbucket != NULL)
	  bucketptr = bucketptr->nextbucket;
      largest_index = bucketptr->index[bucketindex];

      /*
       * PWU: 1/6/97 edits:
       */
      (*returned_len) = number_of_nonzeroes ;

      if ( largest_index == MAXINT ) {

#ifdef _DEBUG
	printf("spxs warning: corrupted SS vector, index=%d\n",
                   largest_index);
#endif

	/*
	 *   Forward Sweep
	 */

        bucketptr = objectptr->sds.ss.bucketptr;
        bucketindex = 0;
	running_nonzero_count = 0;
	while (bucketptr->index[bucketindex] != MAXINT) {
	  running_nonzero_count++;
	  bucketindex++;
	  if((bucketindex = MY_MOD(bucketindex,
				   sBUCKET_DIM_minus1)) == 0) {

#ifdef _DEBUG
	    if (bucketptr->nextbucket == NULL) {
              printf("spxs error: Unexpected NULL nextbucket pointer\n");
              abort();
	    }
#endif

	    bucketptr = bucketptr->nextbucket;
	  }
        }

#ifdef _DEBUG
	if (running_nonzero_count >= number_of_nonzeroes) {
          printf("spxs error: running_nonzero_count=%d should be < ",
                  running_nonzero_count);
          printf("number of nonzeroes=%d\n",number_of_nonzeroes);
          abort();
        }
#endif

	(*returned_len) = running_nonzero_count;
      }

      if ( (*returned_len) > (*lengthptr) ) {
	(*len_info) = -1;
	printf("spxs warning: Fortran array dim. (%d) is less than the\n",
	       (*lengthptr) );
        printf("              largest index (%d) in the nonzero index-set\n",
	       (*returned_len) );
      }
      else {
        bucketptr = objectptr->sds.ss.bucketptr;

        for (i=0; i<(*returned_len);i++) {
          if(((bucketindex=MY_MOD(i,sBUCKET_DIM_minus1)) == 0) && i!=0)
	      bucketptr = bucketptr->nextbucket;
	  X_ASSIGN(valvector[i],bucketptr->value[bucketindex]);
	  indvector[i] = bucketptr->index[bucketindex];
        }
      }
    }
    else /* objectptr-dstype == CS */ {

      /********************************************************************
       *              _COMPRESSED_SUBSCRIPTS EXTRACTION                   *
       ********************************************************************/

      ibucketindex = MY_MOD((objectptr->sds.cs.nonzeroranges-1),
			    sIBUCKET_DIM_minus1);
      ibucketptr = objectptr->sds.cs.ibucketptr;
      if (ibucketptr != NULL)
	while (ibucketptr->nextibucket != NULL)
	  ibucketptr = ibucketptr->nextibucket;
      largest_index = ibucketptr->index_RANGE_END[ibucketindex];

      /*
       * PWU: 1/6/97 edits:
       */
      (*returned_len) = number_of_nonzeroes ;

      if ( largest_index == MAXINT ) {

#ifdef _DEBUG
	printf("spxs warning: corrupted CS vector, index_RANGE_END=%d\n",
                   largest_index);
#endif

      /*
       *   First Sweep
       */

        ibucketptr = objectptr->sds.cs.ibucketptr;
        ibucketindex = 0;
        running_nonzero_count = 0;
        while (ibucketptr->index_RANGE_BEGIN[ibucketindex] != MAXINT) {

#ifdef _DEBUG
          if (ibucketptr->index_RANGE_END[ibucketindex] == MAXINT) {
            printf("spxs error: Inconsistantly corrupt CS vector, ");
            printf("index_RANGE_BEGIN=%d,  index_RANGE_END=%d\n",
                    ibucketptr->index_RANGE_BEGIN[ibucketindex],
                    ibucketptr->index_RANGE_END[ibucketindex]);
            abort();
          }
#endif
	  
	  running_nonzero_count += 1+
	                           ibucketptr->index_RANGE_END[ibucketindex] -
	                           ibucketptr->index_RANGE_BEGIN[ibucketindex];
          ibucketindex++;
          if((ibucketindex = MY_MOD(ibucketindex,
                                     sIBUCKET_DIM_minus1)) == 0) {

#ifdef _DEBUG
            if (ibucketptr->nextibucket == NULL) {
              printf("spxs error: Unexpected NULL nextibucket pointer\n");
              abort();
            }
#endif

            ibucketptr = ibucketptr->nextibucket;
          }
        }

	(*returned_len) = running_nonzero_count;

      /*
       *   Second Sweep: extract
       */

       if ( (*returned_len) > (*lengthptr) ) {
	  (*len_info) = -1;
	  printf("spxs warning: Fortran array dim. (%d) is less than the\n",
	         (*lengthptr) );
	  printf("              largest index (%d) in the nonzero index set\n",
	         (*returned_len) );
        }
        else {
	  ibucketptr = objectptr->sds.cs.ibucketptr;
	  ibucketindex = 0;
          vbucketptr = objectptr->sds.cs.vbucketptr;
          vbucketindex = 0;
          h = 0;
          while (ibucketptr->index_RANGE_BEGIN[ibucketindex] != MAXINT) {
	    for(j=0; j <= (ibucketptr->index_RANGE_END[ibucketindex] - 
			   ibucketptr->index_RANGE_BEGIN[ibucketindex]);
	        h++,j++,vbucketindex++) {
	      if(((vbucketindex =
		   MY_MOD(vbucketindex,sVBUCKET_DIM_minus1)) == 0) && h!=0)
	        vbucketptr = vbucketptr->nextvbucket;
	      X_ASSIGN(valvector[h], vbucketptr->value[vbucketindex]);
	      indvector[h] = ibucketptr->index_RANGE_BEGIN[ibucketindex]+j;
	    }
            ibucketindex++;
	    if((ibucketindex=MY_MOD(ibucketindex,
				     sIBUCKET_DIM_minus1)) == 0)
	      ibucketptr = ibucketptr->nextibucket;
	  }
	}
      }

      else { /* no bad (MAXINT) entries in the CS sparsevector */

        if ( (*returned_len) > (*lengthptr) ) {
	  (*len_info) = -1;
	  printf("spxs warning: Fortran array dim. (%d) is less than the\n",
	         (*lengthptr) );
	  printf("              largest index (%d) in the nonzero index set\n",
	         (*returned_len) );
        }
        else {
	  ibucketptr = objectptr->sds.cs.ibucketptr;
	  vbucketptr = objectptr->sds.cs.vbucketptr;
	  vbucketindex = 0;
	  for(i=0,h=0;i<objectptr->sds.cs.nonzeroranges;i++) {
	    if(((ibucketindex=MY_MOD(i,sIBUCKET_DIM_minus1)) == 0) && i!=0)
	      ibucketptr = ibucketptr->nextibucket;
	    for(j=0; j <= (ibucketptr->index_RANGE_END[ibucketindex] - 
			   ibucketptr->index_RANGE_BEGIN[ibucketindex]);
	        h++,j++,vbucketindex++) {
	      if(((vbucketindex =
		   MY_MOD(vbucketindex,sVBUCKET_DIM_minus1)) == 0) && h!=0)
	        vbucketptr = vbucketptr->nextvbucket;

	      X_ASSIGN(valvector[h], vbucketptr->value[vbucketindex]);
	      indvector[h] = ibucketptr->index_RANGE_BEGIN[ibucketindex]+j;
	    }
	  }
	}
      }
    }
  }
}




/****************************************************************************
 *                                                                          *
 *    sspxm(vector,lengthptr,multiplier,                    *
 *                          objectptrptr,returned_len,len_info)             *
 *                                                                          *
 *       vector       Pointer to an array containing values of type         *
 *                    float.  On exit, if len_info == 0,         *
 *                    vector(i) += multiplier *                             *
 *                                 the value, if any, pointed to by         *
 *                                 *objectptrptr whose index is i, where    *
 *                                 1 <= i <= returned_len.                  *
 *                    If len_info != 0, vector is not modified.             *
 *                                                                          *
 *       lengthptr    Pointer to an integer specifying the length of the    *
 *                    destination array.                                    *
 *                                                                          *
 *       muliplier    See definition of vector.                             *
 *                                                                          *
 *       objectptrptr doubly-indirected pointer to a svectorstruct.     *
 *                                                                          *
 *       returned_len Index of largest nonzero entry in objectptrptr. This  *
 *                    value will always be returned, whether vector is      *
 *                    initialized or not. See description of len_info below.*
 *                                                                          *
 *       len_info     If returned_len is greater than lengthptr, will be set*
 *                    to -1, and vector is not referenced.  Otherwise,      *
 *                    len_info is set to 0, and vector(1:lengthptr)         *
 *                    will be modified as described above.                  *
 *                                                                          *
 *       This routine takes a sparse vector, pointed to by *objectptrptr,   *
 *       and adds the weighted contribution of values pointed to it         *
 *       to the correspondingly-indexed entries in a dense Fortran vector.  *
 *       When called from within a DO loop, looping over the                *
 *       rows of a Jacobian (columns of the corresponding ADFIOR object),   *
 *       this subroutine facilitates the operation, vector-matrix-multiply. *
 *       This is useful, for example, when the computation of the gradient  *
 *       of a partially sep. function is reformulated as a sparse Jacobian  *
 *       problem so that, the accumulation of the gradients of the component*
 *       elements is performed outside of the derivative computation code   *
 *       (see "Computing Gradients of Partially Separable Functions using   *
 *       Automatic Differentiation," by Bischof, Bouaricha, Khademi, More). *
 *                                                                          *
 *       PMK: 12/4/96 edits: See documentation for sspxd routine. *
 *                                                                          *
 ****************************************************************************/


void sspxm(vector,lengthptr,multiplier,
			   objectptrptr,returned_len,len_info)
svectorstruct **objectptrptr;
float *vector, *multiplier;
int *lengthptr, *returned_len, *len_info;
{
  svectorstruct *objectptr;
  sbucketstruct *bucketptr;
  sibucketstruct *ibucketptr;
  svbucketstruct *vbucketptr;
  int i,j,h,bucketindex,ibucketindex,vbucketindex,largestvectorsubscript;

  int running_nonzero_count;
  int number_of_nonzeroes;
  int returned_len_update;

  (*returned_len) = 0;
  (*len_info) = 0;

  if (*objectptrptr == NULL) {
    /* do nothing */
  }
  else {
    objectptr = (svectorstruct *) *objectptrptr;
    number_of_nonzeroes = objectptr->nonzeroes;

    if (objectptr->dstype == S1) {
      if (objectptr->nonzeroes == 0) {
        /* do nothing */
      }
      else {
        (*returned_len) = objectptr->sds.s1.index;

#ifdef _DEBUG
        if ( (*returned_len) == MAXINT ) {
	   printf("spxm error: S1 vector with %d nonzeroes has index=%d\n",
                     objectptr->nonzeroes,(*returned_len));
	   abort();
        }
#endif

	if ( (*returned_len) > (*lengthptr) ) {
	  (*len_info) = -1;
	  printf("spxm warning: Fortran array dim. (%d) is less than the\n",
	       (*lengthptr) );
	  printf("            largest index (%d) in the nonzero index-set.\n",
	       (*returned_len) );
        }
        else {
	  X_PLUS_MULT(vector[objectptr->sds.s1.index-1], *multiplier,
		      objectptr->sds.s1.value);
        }
      }
    }
    else if (objectptr->dstype == SS) {
      
      /*********************************************************************
       *                _SINGLE_SUBSCRIPTS EXTRACTION                      *
       *********************************************************************/
      
      bucketindex = MY_MOD((objectptr->nonzeroes-1),sBUCKET_DIM_minus1);
      bucketptr = objectptr->sds.ss.bucketptr;
      if (bucketptr != NULL)
	while (bucketptr->nextbucket != NULL)
	  bucketptr = bucketptr->nextbucket;
      (*returned_len) = bucketptr->index[bucketindex];

      if ( (*returned_len) == MAXINT ) {

#ifdef _DEBUG
	printf("spxm warning: corrupted SS vector, index=%d\n",
                   (*returned_len));
#endif
	      
      /*
       *   Forward Sweep
       */

        bucketptr = objectptr->sds.ss.bucketptr;
        bucketindex = 0;
        running_nonzero_count = 0;
        while (bucketptr->index[bucketindex] != MAXINT) {
          running_nonzero_count++;
	  returned_len_update = bucketptr->index[bucketindex];
          bucketindex++;
	  if((bucketindex = MY_MOD(bucketindex,
				     sBUCKET_DIM_minus1)) == 0) {

#ifdef _DEBUG
            if (bucketptr->nextbucket == NULL) {
              printf("spxm error: Unexpected NULL nextbucket pointer\n");
              abort();
	    }
#endif

	    bucketptr = bucketptr->nextbucket;
	  }
        }

#ifdef _DEBUG
        if (running_nonzero_count >= number_of_nonzeroes) {
           printf("spxm error: running_nonzero_count=%d should be < ",
                   running_nonzero_count);
           printf("number of nonzeroes=%d\n",number_of_nonzeroes);
           abort();
         }
#endif
        number_of_nonzeroes = running_nonzero_count;
	(*returned_len) = returned_len_update;
      }

      if ( (*returned_len) > (*lengthptr) ) {
	(*len_info) = -1;
	printf("spxm warning: The Fortran array dim. (%d) is less than the\n",
	       (*lengthptr) );
	printf("              largest index (%d) in the nonzero index-set.\n",
	       (*returned_len) );
      }
      else {
	bucketptr = objectptr->sds.ss.bucketptr;
	for(i=0;i<number_of_nonzeroes;i++) {
	  if(((bucketindex=MY_MOD(i,sBUCKET_DIM_minus1)) == 0) && i!=0)
	    bucketptr = bucketptr->nextbucket;
	  X_PLUS_MULT(vector[bucketptr->index[bucketindex]-1], *multiplier,
		      bucketptr->value[bucketindex]);
	}
      }
    }
    else /* if (objectptr->dstype == CS) */ { 
      
      /*********************************************************************
       *              _COMPRESSED_SUBSCRIPTS EXTRACTION                    *
       *********************************************************************/
      
      ibucketindex = MY_MOD((objectptr->sds.cs.nonzeroranges-1),
			    sIBUCKET_DIM_minus1);
      ibucketptr = objectptr->sds.cs.ibucketptr;
      if (ibucketptr != NULL)
	while (ibucketptr->nextibucket != NULL)
	  ibucketptr = ibucketptr->nextibucket;
      (*returned_len) = ibucketptr->index_RANGE_END[ibucketindex];

      if ( (*returned_len) == MAXINT ) {

#ifdef _DEBUG
	printf("spxm warning: corrupted CS vector, index_RANGE_END=%d\n",
                   (*returned_len));
#endif

      /*
       *   First Sweep: find true value of (*returned_len)
       */

        ibucketptr = objectptr->sds.cs.ibucketptr;
        ibucketindex = 0;
        while (ibucketptr->index_RANGE_BEGIN[ibucketindex] != MAXINT) {

#ifdef _DEBUG
          if (ibucketptr->index_RANGE_END[ibucketindex] == MAXINT) {
            printf("spxm error: Inconsistantly corrupt CS vector, ");
            printf("index_RANGE_BEGIN=%d,  index_RANGE_END=%d\n",
                    ibucketptr->index_RANGE_BEGIN[ibucketindex],
                    ibucketptr->index_RANGE_END[ibucketindex]);
            abort();
          }
#endif
          
          returned_len_update = ibucketptr->index_RANGE_END[ibucketindex];

          ibucketindex++;
          if((ibucketindex = MY_MOD(ibucketindex,
                                     sIBUCKET_DIM_minus1)) == 0) {

#ifdef _DEBUG
            if (ibucketptr->nextibucket == NULL) {
              printf("spxm error: Unexpected NULL nextibucket pointer\n");
              abort();
            }
#endif

            ibucketptr = ibucketptr->nextibucket;
          }
        }

        (*returned_len) = returned_len_update;

      /*
       *   Second Sweep: extract
       */

       if ( (*returned_len) > (*lengthptr) ) {
          (*len_info) = -1;
          printf("spxm warning: Fortran array dim. (%d) is less than the\n",
                 (*lengthptr) );
          printf("              largest index (%d) in the nonzero index set\n",
                 (*returned_len) );
        }
        else {
          ibucketptr = objectptr->sds.cs.ibucketptr;
          ibucketindex = 0;
          vbucketptr = objectptr->sds.cs.vbucketptr;
          vbucketindex = 0;
          h = 0;
          while (ibucketptr->index_RANGE_BEGIN[ibucketindex] != MAXINT) {
            for(j=0; j <= (ibucketptr->index_RANGE_END[ibucketindex] - 
                           ibucketptr->index_RANGE_BEGIN[ibucketindex]);
                h++,j++,vbucketindex++) {
              if(((vbucketindex =
                   MY_MOD(vbucketindex,sVBUCKET_DIM_minus1)) == 0) && h!=0)
                vbucketptr = vbucketptr->nextvbucket;
	      X_PLUS_MULT(
		     vector[ibucketptr->index_RANGE_BEGIN[ibucketindex]+j-1],
		     *multiplier, vbucketptr->value[vbucketindex]);
            }
            ibucketindex++;
            if((ibucketindex=MY_MOD(ibucketindex,
                                     sIBUCKET_DIM_minus1)) == 0)
              ibucketptr = ibucketptr->nextibucket;
          }
        }
      }

      else { /* no bad (MAXINT) entries in the CS sparsevector */

        if ( (*returned_len) > (*lengthptr) ) {
	  (*len_info) = -1;
	  printf("spxm warning: Fortran array dim. (%d) is less than the\n",
	       (*lengthptr) );
	  printf("              largest index (%d) in the nonzero index-set\n",
	       (*returned_len) );
        }
        else {
	  ibucketptr = objectptr->sds.cs.ibucketptr;
	  vbucketptr = objectptr->sds.cs.vbucketptr;
	
	  vbucketindex = 0;
	
	  for(i=0,h=0;i<objectptr->sds.cs.nonzeroranges;i++) {
	    if(((ibucketindex =
	         MY_MOD(i,sIBUCKET_DIM_minus1)) == 0) && i!=0)
	      ibucketptr = ibucketptr->nextibucket;
	    for(j=0; j <= (ibucketptr->index_RANGE_END[ibucketindex] - 
			   ibucketptr->index_RANGE_BEGIN[ibucketindex]);
	        h++,j++,vbucketindex++) {
	      if(((vbucketindex =
		   MY_MOD(vbucketindex,sVBUCKET_DIM_minus1)) == 0) && h!=0)
	        vbucketptr = vbucketptr->nextvbucket;
	      X_PLUS_MULT(
		     vector[ibucketptr->index_RANGE_BEGIN[ibucketindex]+j-1],
		     *multiplier, vbucketptr->value[vbucketindex]);
	    }
	  }
	}
      }
    }
  }
}



/****************************************************************************
 *                                                                          *
 *    sspxa(vector,lengthptr,                           *
 *                              objectptrptr,returned_len,len_info)         *
 *                                                                          *
 *       This routine is identical to sspxm, except there   *
 *       is no multiplier (see the documentation for sspxm).*
 *                                                                          *
 *       PMK: 12/4/96 edits: See documentation for sspxd routine. *
 *                                                                          *
 ****************************************************************************/


void sspxa(vector,lengthptr,
			       objectptrptr,returned_len,len_info)
svectorstruct **objectptrptr;
float *vector;
int *lengthptr, *returned_len, *len_info;
{
  svectorstruct *objectptr;
  sbucketstruct *bucketptr;
  sibucketstruct *ibucketptr;
  svbucketstruct *vbucketptr;
  int i,j,h,bucketindex,ibucketindex,vbucketindex,largestvectorsubscript;

  int running_nonzero_count;
  int number_of_nonzeroes;
  int returned_len_update;

  (*returned_len) = 0;
  (*len_info) = 0;

  if (*objectptrptr == NULL) {
    /* do nothing */
  }
  else {
    objectptr = (svectorstruct *) *objectptrptr;
    number_of_nonzeroes = objectptr->nonzeroes;

    if (objectptr->dstype == S1) {
      if (objectptr->nonzeroes == 0) {
        /* do nothing */
      }
      else {
        (*returned_len) = objectptr->sds.s1.index;

#ifdef _DEBUG
        if ( (*returned_len) == MAXINT ) {
           printf("spxa error: S1 vector with %d nonzeroes has index=%d\n",
                     objectptr->nonzeroes,(*returned_len));
           abort();
        }
#endif
  
        if ( (*returned_len) > (*lengthptr) ) {
	 (*len_info) = -1;
	  printf("spxa warning: Fortran array dim. (%d) is less than the\n",
	         (*lengthptr) );
	  printf("            largest index (%d) in the nonzero index-set.\n",
	         (*returned_len) );
        }
        else {
	  X_PLUS_EQUALS(vector[objectptr->sds.s1.index-1],
		      objectptr->sds.s1.value);
        }
      }
    }
    else if (objectptr->dstype == SS) {
      
      /*********************************************************************
       *                _SINGLE_SUBSCRIPTS EXTRACTION                      *
       *********************************************************************/
      
      bucketindex = MY_MOD((objectptr->nonzeroes-1),sBUCKET_DIM_minus1);
      bucketptr = objectptr->sds.ss.bucketptr;
      if (bucketptr != NULL)
	while (bucketptr->nextbucket != NULL)
	  bucketptr = bucketptr->nextbucket;
      (*returned_len) = bucketptr->index[bucketindex];

      if ( (*returned_len) == MAXINT ) {

#ifdef _DEBUG
        printf("spxa warning: corrupted SS vector, index=%d\n",
                   (*returned_len));
#endif
	      
      /*
       *   Forward Sweep
       */

        bucketptr = objectptr->sds.ss.bucketptr;
        bucketindex = 0;
        running_nonzero_count = 0;
        while (bucketptr->index[bucketindex] != MAXINT) {
          running_nonzero_count++;
	  returned_len_update = bucketptr->index[bucketindex];
          bucketindex++;
	  if((bucketindex = MY_MOD(bucketindex,
				     sBUCKET_DIM_minus1)) == 0) {

#ifdef _DEBUG
            if (bucketptr->nextbucket == NULL) {
              printf("spxa error: Unexpected NULL nextbucket pointer\n");
              abort();
	    }
#endif

	    bucketptr = bucketptr->nextbucket;
	  }
        }

#ifdef _DEBUG
        if (running_nonzero_count >= number_of_nonzeroes) {
           printf("spxa error: running_nonzero_count=%d should be < ",
                   running_nonzero_count);
           printf("number of nonzeroes=%d\n",number_of_nonzeroes);
           abort();
         }
#endif
        number_of_nonzeroes = running_nonzero_count;
	(*returned_len) = returned_len_update;
      }

      if ( (*returned_len) > (*lengthptr) ) {
	(*len_info) = -1;
	printf("spxa warning: The Fortran array dim. (%d) is less than the\n",
	       (*lengthptr) );
	printf("              largest index (%d) in the nonzero index-set.\n",
	       (*returned_len) );
      }
      else {
	bucketptr = objectptr->sds.ss.bucketptr;
	for(i=0;i<number_of_nonzeroes;i++) {
	  if(((bucketindex=MY_MOD(i,sBUCKET_DIM_minus1)) == 0) && i!=0)
	    bucketptr = bucketptr->nextbucket;
	  X_PLUS_EQUALS(vector[bucketptr->index[bucketindex]-1],
		      bucketptr->value[bucketindex]);
	}
      }
    }
    else /* if (objectptr->dstype == CS) */ { 
      
      /*********************************************************************
       *              _COMPRESSED_SUBSCRIPTS EXTRACTION                    *
       *********************************************************************/
      
      ibucketindex = MY_MOD((objectptr->sds.cs.nonzeroranges-1),
			    sIBUCKET_DIM_minus1);
      ibucketptr = objectptr->sds.cs.ibucketptr;
      if (ibucketptr != NULL)
	while (ibucketptr->nextibucket != NULL)
	  ibucketptr = ibucketptr->nextibucket;
      (*returned_len) = ibucketptr->index_RANGE_END[ibucketindex];

      if ( (*returned_len) == MAXINT ) {

#ifdef _DEBUG
        printf("spxa warning: corrupted CS vector, index_RANGE_END=%d\n",
                   (*returned_len));
#endif

      /*
       *   First Sweep: find true value of (*returned_len)
       */

        ibucketptr = objectptr->sds.cs.ibucketptr;
        ibucketindex = 0;
        while (ibucketptr->index_RANGE_BEGIN[ibucketindex] != MAXINT) {

#ifdef _DEBUG
          if (ibucketptr->index_RANGE_END[ibucketindex] == MAXINT) {
            printf("spxa error: Inconsistantly corrupt CS vector, ");
            printf("index_RANGE_BEGIN=%d,  index_RANGE_END=%d\n",
                    ibucketptr->index_RANGE_BEGIN[ibucketindex],
                    ibucketptr->index_RANGE_END[ibucketindex]);
            abort();
          }
#endif
          
          returned_len_update = ibucketptr->index_RANGE_END[ibucketindex];

          ibucketindex++;
          if((ibucketindex = MY_MOD(ibucketindex,
                                     sIBUCKET_DIM_minus1)) == 0) {

#ifdef _DEBUG
            if (ibucketptr->nextibucket == NULL) {
              printf("spxa error: Unexpected NULL nextibucket pointer\n");
              abort();
            }
#endif

            ibucketptr = ibucketptr->nextibucket;
          }
        }

        (*returned_len) = returned_len_update;

      /*
       *   Second Sweep: extract
       */

       if ( (*returned_len) > (*lengthptr) ) {
          (*len_info) = -1;
          printf("spxa warning: Fortran array dim. (%d) is less than the\n",
                 (*lengthptr) );
          printf("              largest index (%d) in the nonzero index set\n",
                 (*returned_len) );
        }
        else {
          ibucketptr = objectptr->sds.cs.ibucketptr;
          ibucketindex = 0;
          vbucketptr = objectptr->sds.cs.vbucketptr;
          vbucketindex = 0;
          h = 0;
          while (ibucketptr->index_RANGE_BEGIN[ibucketindex] != MAXINT) {
            for(j=0; j <= (ibucketptr->index_RANGE_END[ibucketindex] - 
                           ibucketptr->index_RANGE_BEGIN[ibucketindex]);
                h++,j++,vbucketindex++) {
              if(((vbucketindex =
                   MY_MOD(vbucketindex,sVBUCKET_DIM_minus1)) == 0) && h!=0)
                vbucketptr = vbucketptr->nextvbucket;
	      X_PLUS_EQUALS(
		     vector[ibucketptr->index_RANGE_BEGIN[ibucketindex]+j-1],
		     vbucketptr->value[vbucketindex]);
            }
            ibucketindex++;
            if((ibucketindex=MY_MOD(ibucketindex,
                                     sIBUCKET_DIM_minus1)) == 0)
              ibucketptr = ibucketptr->nextibucket;
          }
        }
      }

      else { /* no bad (MAXINT) entries in the CS sparsevector */

        if ( (*returned_len) > (*lengthptr) ) {
	  (*len_info) = -1;
	  printf("spxa warning: Fortran array dim. (%d) is less than the\n",
	         (*lengthptr) );
	  printf("              largest index (%d) in the nonzero index-set\n",
	         (*returned_len) );
        }
        else {
	  ibucketptr = objectptr->sds.cs.ibucketptr;
	  vbucketptr = objectptr->sds.cs.vbucketptr;
	
	  vbucketindex = 0;
	
	  for(i=0,h=0;i<objectptr->sds.cs.nonzeroranges;i++) {
	    if(((ibucketindex =
	         MY_MOD(i,sIBUCKET_DIM_minus1)) == 0) && i!=0)
	      ibucketptr = ibucketptr->nextibucket;
	    for(j=0; j <= (ibucketptr->index_RANGE_END[ibucketindex] - 
			   ibucketptr->index_RANGE_BEGIN[ibucketindex]);
	        h++,j++,vbucketindex++) {
	      if(((vbucketindex =
		   MY_MOD(vbucketindex,sVBUCKET_DIM_minus1)) == 0) && h!=0)
	        vbucketptr = vbucketptr->nextvbucket;
	      X_PLUS_EQUALS(
		     vector[ibucketptr->index_RANGE_BEGIN[ibucketindex]+j-1],
		     vbucketptr->value[vbucketindex]);
	    }
	  }
	}
      }
    }
  }
}



/****************************************************************************
 *                                                                          *
 *    ssppr (objectptrptr,unit_number)                                    *
 *                                                                          *
 *       objectptrptr Doubly-indirected pointer to a svectorstruct.     *
 *                                                                          *
 *       unit_number  A pointer to an integer containing information about  *
 *                    where to write output to. If unit_number == 0, stdout *
 *                    is output.  If unit_number is in the range [1,999],   *
 *                    output is to a file                                   *
 *                    whose name as the unit_number appended to 'SPPR.'     *
 *                    For example, if unit_number = 16, then the output     *
 *                    is appended to the file 'SPPR.16' in the current      *
 *                    directory, i.e., if 'SPPR.16' did not already exist   *
 *                    in the directory it is created and written to, else   *
 *                    the output is appended to the end of the 'SPPR.16'.   *
 *                                                                          *
 *       ssppr takes one argument, a pointer to a vector, and prints      *
 *       that vector to stdout or to a file.  The output is formatted as    *
 *       follows (in the case of complex vectors, the value is a tuple):    *
 *                                                                          *
 *             Number of nonzeros = %d.                                     *
 *                                                                          *
 *             Index         Value                                          *
 *             -----         -----                                          *
 *             (Index)       (Value)                                        *
 *                .             .                                           *
 *                .             .                                           *
 *                .             .                                           *
 *                                                                          *
 *       PMK: 12/4/96 edits: See documentation for sspxd routine. *
 *                                                                          *
 ****************************************************************************/

void ssppr (objectptrptr,unit_number)
svectorstruct **objectptrptr;
int *unit_number;
{
  svectorstruct *objectptr;
  sbucketstruct *bucketptr;
  sibucketstruct *ibucketptr;
  svbucketstruct *vbucketptr;
  int i,j,h,bucketindex,ibucketindex,vbucketindex,number_of_nonzeroes;

  int largest_index;      /* PMK: 12/4/96: largest index found in vector */
  int corrupt_index_adjustment = 0;
  int running_nonzero_count;

  FILE *fp;
  char *ext;
  char *filename;

  if ( (*unit_number) > 0 && (*unit_number) < 1000 ) {
    ext = (char *) malloc(4*sizeof(char));
    filename = (char *) malloc(6*sizeof(char));
    strcpy(filename, "SPPR.");
    my_itoa ( (*unit_number), ext );
    strcat (filename, ext);
    if ( (fp = fopen (filename,"a")) == NULL ) {
      printf("sppr error: Couldn't open file.\n");
      abort();
    }
  }
  else if  ( (*unit_number) != 0 ) {
    printf("sppr error: Invalid Unit Number = %d  (must be in [0,1000])\n",
	   (*unit_number) );
    abort();
  }

  if (*objectptrptr == NULL) {
    number_of_nonzeroes = 0;
    if ( (*unit_number) == 0 ) {
      printf ("Number of nonzeros = %d\n",number_of_nonzeroes);
      printf ("    Index    Value\n");
      printf ("    -----    -----\n");
    }
    else {
      fprintf (fp,"Number of nonzeros = %d\n",number_of_nonzeroes);
      fprintf (fp,"    Index    Value\n");
      fprintf (fp,"    -----    -----\n");
    }
  }

  else { /* i.e., if *objectptrptr != NULL */
    objectptr = (svectorstruct *) *objectptrptr;
    number_of_nonzeroes = objectptr->nonzeroes;

    if (objectptr->dstype == S1) {
      if (number_of_nonzeroes != 0 && objectptr->sds.s1.index == MAXINT) {
#ifdef _DEBUG
        printf("sppr error: S1 vector with %d nonzeroes has index=%d\n",
                     number_of_nonzeroes,objectptr->sds.s1.index);
        abort();
#endif
        number_of_nonzeroes = 0;
      }
      if ( (*unit_number) == 0 ) {
        printf ("Number of nonzeros = %d\n",number_of_nonzeroes);
        printf ("    Index    Value\n");
        printf ("    -----    -----\n");
        X_PRINT_D_E(objectptr->sds.s1.index,
		  objectptr->sds.s1.value);
      }
      else {
        fprintf (fp,"Number of nonzeros = %d\n",number_of_nonzeroes);
        fprintf (fp,"    Index    Value\n");
        fprintf (fp,"    -----    -----\n");
        X_FPRINT_D_E(fp,
		   objectptr->sds.s1.index,
		   objectptr->sds.s1.value);
      }
    }

    else if (objectptr->dstype == SS) {

    /*********************************************************************
     *                        Single Subscript                           *
     *********************************************************************/
      
      bucketindex = MY_MOD((objectptr->nonzeroes-1),sBUCKET_DIM_minus1);
      bucketptr = objectptr->sds.ss.bucketptr;
      if (bucketptr != NULL)
        while (bucketptr->nextbucket != NULL)
          bucketptr = bucketptr->nextbucket;
      largest_index = bucketptr->index[bucketindex];

      if ( largest_index == MAXINT ) {

#ifdef _DEBUG
        printf("sppr warning: corrupted SS vector, index=%d\n",
                   largest_index);
#endif

        /*
         *   One Forward Sweep: find number of legitimate nonzeroes
         */

        bucketptr = objectptr->sds.ss.bucketptr;
        bucketindex = 0;
        running_nonzero_count = 0;
        while (bucketptr->index[bucketindex] != MAXINT) {
          running_nonzero_count++;
          bucketindex++;
	  if((bucketindex = MY_MOD(bucketindex,
				     sBUCKET_DIM_minus1)) == 0) {

#ifdef _DEBUG
            if (bucketptr->nextbucket == NULL) {
              printf("sppr error: Unexpected NULL nextbucket pointer\n");
              abort();
	    }
#endif

	    bucketptr = bucketptr->nextbucket;
	  }
        }

#ifdef _DEBUG
        if (running_nonzero_count >= number_of_nonzeroes) {
           printf("sppr error: running_nonzero_count=%d should be < ",
                   running_nonzero_count);
           printf("number of nonzeroes=%d\n",number_of_nonzeroes);
           abort();
         }
#endif
        number_of_nonzeroes = running_nonzero_count;
      }

      if ( (*unit_number) == 0 ) {
        printf ("Number of nonzeros = %d\n",number_of_nonzeroes);
        printf ("    Index    Value\n");
        printf ("    -----    -----\n");
      }
      else {
        fprintf (fp,"Number of nonzeros = %d\n",number_of_nonzeroes);
        fprintf (fp,"    Index    Value\n");
        fprintf (fp,"    -----    -----\n");
      }

      bucketptr = objectptr->sds.ss.bucketptr;
      for (i=0;i<number_of_nonzeroes;i++) {
        if ((bucketindex = MY_MOD(i,sBUCKET_DIM_minus1)) == 0 && i != 0) {
	  bucketptr = bucketptr->nextbucket;
        }
        if ( (*unit_number) == 0 ) {
	  X_PRINT_D_E(bucketptr->index[bucketindex],
		      bucketptr->value[bucketindex]);
        }
        else {
	  X_FPRINT_D_E(fp,
		       bucketptr->index[bucketindex],
		       bucketptr->value[bucketindex]);
        }
      }
    }

    else /* objectptr->dstype = CS */ {

    /*********************************************************************
     *                      Compressed Subscript                         *
     *********************************************************************/

      ibucketindex = MY_MOD((objectptr->sds.cs.nonzeroranges-1),
                            sIBUCKET_DIM_minus1);
      ibucketptr = objectptr->sds.cs.ibucketptr;
      if (ibucketptr != NULL)
        while (ibucketptr->nextibucket != NULL)
          ibucketptr = ibucketptr->nextibucket;
      largest_index = ibucketptr->index_RANGE_END[ibucketindex];

      if ( largest_index == MAXINT ) {

#ifdef _DEBUG
        printf("sppr warning: corrupted CS vector, index_RANGE_END=%d\n",
                   largest_index);
#endif

        /*
         *   First Sweep: find and print number of legitimate nonzeroes
         */

        ibucketptr = objectptr->sds.cs.ibucketptr;
        ibucketindex = 0;
        running_nonzero_count = 0;
        while (ibucketptr->index_RANGE_BEGIN[ibucketindex] != MAXINT) {

#ifdef _DEBUG
          if (ibucketptr->index_RANGE_END[ibucketindex] == MAXINT) {
            printf("sppr error: Inconsistantly corrupt CS vector, ");
            printf("index_RANGE_BEGIN=%d,  index_RANGE_END=%d\n",
                    ibucketptr->index_RANGE_BEGIN[ibucketindex],
                    ibucketptr->index_RANGE_END[ibucketindex]);
            abort();
	  }
#endif

          running_nonzero_count+= 1 +
                                  ibucketptr->index_RANGE_END[ibucketindex] -
                                  ibucketptr->index_RANGE_BEGIN[ibucketindex];
          ibucketindex++;
	  if((ibucketindex = MY_MOD(ibucketindex,
				     sIBUCKET_DIM_minus1)) == 0) {

#ifdef _DEBUG
            if (ibucketptr->nextibucket == NULL) {
              printf("sppr error: Unexpected NULL nextibucket pointer\n");
              abort();
	    }
#endif

	    ibucketptr = ibucketptr->nextibucket;
	  }
        }

        if ( (*unit_number) == 0 ) {
          printf ("Number of nonzeros = %d\n",running_nonzero_count);
          printf ("    Index    Value\n");
          printf ("    -----    -----\n");
        }
        else {
          fprintf (fp,"Number of nonzeros = %d\n",running_nonzero_count);
          fprintf (fp,"    Index    Value\n");
          fprintf (fp,"    -----    -----\n");
        }

        /*
         *   Second Sweep: print legitimate index-value pairs
         */

        ibucketptr = objectptr->sds.cs.ibucketptr;
        ibucketindex = 0;
        vbucketptr = objectptr->sds.cs.vbucketptr;
        vbucketindex=0;
        h = 0;
        while (ibucketptr->index_RANGE_BEGIN[ibucketindex] != MAXINT) {
          for(j=0; j <= (ibucketptr->index_RANGE_END[ibucketindex] - 
                        ibucketptr->index_RANGE_BEGIN[ibucketindex]);
	      h++,j++,vbucketindex++) {
	    if(((vbucketindex =
	         MY_MOD(vbucketindex,sVBUCKET_DIM_minus1))==0) && h!=0) {
	      vbucketptr = vbucketptr->nextvbucket;
	    }
	    if ( (*unit_number) == 0 ) {
	      X_PRINT_D_E(ibucketptr->index_RANGE_BEGIN[ibucketindex]+j,
		          vbucketptr->value[vbucketindex]);
	    }
	    else {
	      X_FPRINT_D_E(fp,
		           ibucketptr->index_RANGE_BEGIN[ibucketindex]+j,
		           vbucketptr->value[vbucketindex]);
	    }
	  }
          ibucketindex++;
	  if((ibucketindex = MY_MOD(ibucketindex,
				     sIBUCKET_DIM_minus1)) == 0) {
	    ibucketptr = ibucketptr->nextibucket;
	  }
        }
      }

      else { /* i.e., if largest_index != MAXINT */
        if ( (*unit_number) == 0 ) {
          printf ("Number of nonzeros = %d\n",number_of_nonzeroes);
          printf ("    Index    Value\n");
          printf ("    -----    -----\n");
        }
        else {
          fprintf (fp,"Number of nonzeros = %d\n",number_of_nonzeroes);
          fprintf (fp,"    Index    Value\n");
          fprintf (fp,"    -----    -----\n");
        }

        ibucketptr = objectptr->sds.cs.ibucketptr;
        vbucketptr = objectptr->sds.cs.vbucketptr;
        vbucketindex=0;
        for(i=0,h=0;i<(objectptr->sds.cs.nonzeroranges);i++) {
          if(((ibucketindex =
	       MY_MOD(i,sIBUCKET_DIM_minus1)) == 0) && i!=0) {
	    ibucketptr = ibucketptr->nextibucket;
          }
          for(j=0; j <= (ibucketptr->index_RANGE_END[ibucketindex] - 
                        ibucketptr->index_RANGE_BEGIN[ibucketindex]);
	      h++,j++,vbucketindex++) {
	    if(((vbucketindex =
	         MY_MOD(vbucketindex,sVBUCKET_DIM_minus1))==00) && h!=0) {
	      vbucketptr = vbucketptr->nextvbucket;
	    }
	    if ( (*unit_number) == 0 ) {
	      X_PRINT_D_E(ibucketptr->index_RANGE_BEGIN[ibucketindex]+j,
		          vbucketptr->value[vbucketindex]);
	    }
	    else {
	      X_FPRINT_D_E(fp,
		           ibucketptr->index_RANGE_BEGIN[ibucketindex]+j,
		           vbucketptr->value[vbucketindex]);
	    }
	  }
	}
      }
    }
  }
  if ( (*unit_number)== 0 ) {
    printf("\n");
  }
  else {
    fprintf(fp,"\n");
    fclose(fp);
  }
}

