/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#ifndef sGAXPY_PD_MACROS_AND_DATA_STRUCTURES

#define sGAXPY_PD_MACROS_AND_DATA_STRUCTURES

#include "gaxpy_pi_macros_and_data_structures.h"


/***************************************************************************
 *                                                                         *
 *  Macro definitions for store dimensions                                 *
 *                                                                         *
 ***************************************************************************/

#define sBUCKETSTORE_DIM  128      /* # of buckets in a bucketstore.   */

#define sVBUCKETSTORE_DIM 128      /* # of vbuckets in a vbucketstore. */

#define sIBUCKETSTORE_DIM 128      /* # of ibuckets in a ibucketstore. */

#define sVECTORSTORE_DIM  128      /* # of vectors in a vectorstore.   */


/***************************************************************************
 *                                                                         *
 *  Macro definitions for subroutines;  DEBUG ON                           *
 *                                                                         *
 ***************************************************************************/

#ifdef _DEBUG

#define sHOP_TO_NEXT_RHSVBUCKET(a,b,c,d) \
        shop_to_next_rhsvbucket(a,b,c,d)

#define sHOP_TO_NEXT_LHSVBUCKET(a,b) \
        shop_to_next_lhsvbucket(a,b)

#define sNON_OVER_LAPPING_CS_RANGE_AXPY(a,b,c,d,e,f,g,h,i,j,k,l) \
        snon_over_lapping_cs_range_axpy(a,b,c,d,e,f,g,h,i,j,k,l)

#define sOVERLAPPING_CS_RANGE_AXPY(a,b,c,d,e,f,g,h,i) \
        soverlapping_cs_range_axpy(a,b,c,d,e,f,g,h,i)

#define sDENSE_PLUSEQUAL_AXPY_LOOP(a,b,c,d,e) \
        sdense_plusequal_axpy_loop(a,b,c,d,e)

#define sNON_OVER_LAPPING_CS_RANGE_CONVERT_TO_c(a,b,c,d,e,f,g,h,i,j,k) \
        snon_over_lapping_cs_range_convert_to_c(a,b,c,d,e,f,g,h,i,j,k)

#define sOVERLAPPING_CS_RANGE_CONVERT_TO_c(a,b,c,d,e,f,g,h) \
        soverlapping_cs_range_convert_to_c(a,b,c,d,e,f,g,h)

#define sDENSE_CONVERT_TO_c_LOOP(a,b,c,d,e,f) \
        sdense_convert_to_c_loop(a,b,c,d,e,f)

#define sNON_OVER_LAPPING_CS_RANGE_CONVERT_TO_z(a,b,c,d,e,f,g,h,i,j,k) \
        snon_over_lapping_cs_range_convert_to_z(a,b,c,d,e,f,g,h,i,j,k)

#define sOVERLAPPING_CS_RANGE_CONVERT_TO_z(a,b,c,d,e,f,g,h) \
        soverlapping_cs_range_convert_to_z(a,b,c,d,e,f,g,h)

#define sDENSE_CONVERT_TO_z_LOOP(a,b,c,d,e,f) \
        sdense_convert_to_z_loop(a,b,c,d,e,f)

#define sDETERMINE_LOOP_BOUNDS(a,b,c,d,e,f,g,h,i,j) \
        sdetermine_loop_bounds(a,b,c,d,e,f,g,h,i,j)

#define sDEBUGGING_CODE_1(a,b,c) \
        sdebugging_code_1(a,b,c)

/*
 *  In-lining sdense_axpy_loop
 */

#define sDENSE_AXPY_LOOP(range,lhssvbucketindex,rhssvbucketindex,k,loopsize,alpha,addressofunconsumedcurrentnonzerorange) \
{ \
  int j; \
  if (range == OLD_RANGE) { \
    for (j=0;j<loopsize;j++) { \
      GAXPY_PLUS_EQUALS(lhssvbucketptr->value[(*lhssvbucketindex)+j], \
	      alpha,rhssvbucketptr[k]->value[(*rhssvbucketindex)+j]); \
    } \
  } \
  else { \
    for (j=0;j<loopsize;j++) { \
      GAXPY_EQUALS(lhssvbucketptr->value[(*lhssvbucketindex)+j], \
	      alpha,rhssvbucketptr[k]->value[(*rhssvbucketindex)+j]); \
    } \
  } \
  (*lhssvbucketindex) += loopsize; \
  (*rhssvbucketindex) += loopsize; \
  (*addressofunconsumedcurrentnonzerorange) -= loopsize; \
}


/***************************************************************************
 *                                                                         *
 *  Macro definitions for subroutines;  DEBUG OFF                          *
 *                                                                         *
 ***************************************************************************/

#else

#define sHOP_TO_NEXT_RHSVBUCKET(a,b,c,d) \
        shop_to_next_rhsvbucket(a,b)

#define sHOP_TO_NEXT_LHSVBUCKET(a,b) \
        shop_to_next_lhsvbucket(a)

#define sNON_OVER_LAPPING_CS_RANGE_AXPY(a,b,c,d,e,f,g,h,i,j,k,l) \
        snon_over_lapping_cs_range_axpy(a,b,c,d,e,f,g,h)

#define sOVERLAPPING_CS_RANGE_AXPY(a,b,c,d,e,f,g,h,i) \
        soverlapping_cs_range_axpy(a,b,c,d,e)

#define sDENSE_PLUSEQUAL_AXPY_LOOP(a,b,c,d,e) \
        sdense_plusequal_axpy_loop(a,b,c,d)

#define sNON_OVER_LAPPING_CS_RANGE_CONVERT_TO_c(a,b,c,d,e,f,g,h,i,j,k) \
        snon_over_lapping_cs_range_convert_to_c(a,b,c,d,e,f,g)

#define sOVERLAPPING_CS_RANGE_CONVERT_TO_c(a,b,c,d,e,f,g,h) \
        soverlapping_cs_range_convert_to_c(a,b,c,d)

#define sDENSE_CONVERT_TO_c_LOOP(a,b,c,d,e,f) \
        sdense_convert_to_c_loop(a,b,c,d,e)

#define sNON_OVER_LAPPING_CS_RANGE_CONVERT_TO_z(a,b,c,d,e,f,g,h,i,j,k) \
        snon_over_lapping_cs_range_convert_to_z(a,b,c,d,e,f,g)

#define sOVERLAPPING_CS_RANGE_CONVERT_TO_z(a,b,c,d,e,f,g,h) \
        soverlapping_cs_range_convert_to_z(a,b,c,d)

#define sDENSE_CONVERT_TO_z_LOOP(a,b,c,d,e,f) \
        sdense_convert_to_z_loop(a,b,c,d,e)

#define sDETERMINE_LOOP_BOUNDS(a,b,c,d,e,f,g,h,i,j) \
        sdetermine_loop_bounds(a,b,c,d,e,f,g,h,i)

#define sDEBUGGING_CODE_1(a,b,c)

/*
 *  In-lining sdense_axpy_loop
 */

#define sDENSE_AXPY_LOOP(range,lhssvbucketindex,rhssvbucketindex,k,loopsize,alpha,addressofunconsumedcurrentnonzerorange) \
{ \
  int j; \
  if (range == OLD_RANGE) { \
    for (j=0;j<loopsize;j++) { \
      GAXPY_PLUS_EQUALS(lhssvbucketptr->value[(*lhssvbucketindex)+j], \
	      alpha,rhssvbucketptr[k]->value[(*rhssvbucketindex)+j]); \
    } \
  } \
  else { \
    for (j=0;j<loopsize;j++) { \
      GAXPY_EQUALS(lhssvbucketptr->value[(*lhssvbucketindex)+j], \
	      alpha,rhssvbucketptr[k]->value[(*rhssvbucketindex)+j]); \
    } \
  } \
  (*lhssvbucketindex) += loopsize; \
  (*rhssvbucketindex) += loopsize; \
}

#endif


/***************************************************************************
 *                                                                         *
 *  Macro definitions for subroutines with the same interface for          *
 * _DEBUG.                                                                 *
 *                                                                         *
 ***************************************************************************/

#define RESET_freedsbucketslist(new_head) \
{ freedsbucketslist = new_head; \
}

#define RESET_freedsibucketslist(new_head) \
{ freedsibucketslist = new_head; \
}

#define RESET_freedsvbucketslist(new_head) \
{ freedsvbucketslist = new_head; \
}



/****************************************************************************
 *                                                                          *
 *  Data Structures                                                         *
 *                                                                          *
 *     Two abstract structures need to be represented: sparse vector and    *
 *     heap.  We will first describe these for _SINGLE_SUBSCRIPTS and then  *
 *     point out the differences that exist between these data structures   *
 *     and an analog set of data structures used for _COMPRESSED_SUBSCRIPTS.*
 *                                                                          *
 *     To represent sparse vectors using _SINGLE_SUBSCRIPTS, we define 4    *
 *     structs:                                                             *
 *                                                                          *
 *                  sbucketstruct contains a max # of pairs of indices  *
 *                                    and the corresponding values of the   *
 *                                    of the nonzeroes in a vector, and a   *
 *                                    pointer to the next bucket if there   *
 *                                    are more nonzeroes in the vector. The *
 *                                    indices are always in ascending order *
 *                                    by virtue of the way they are input   *
 *                                    in sspsd and the fact that the       *
 *                                    heapsaxpy operations construct new    *
 *                                    vectors with the same ordering.       *
 *                                                                          *
 *                  svectorstruct is a sparse representation of a       *
 *                                    vector.  It has a field for the total *
 *                                    # of nonzeroes in the vector and a    *
 *                                    linked list of buckets which contain  *
 *                                    the indices and values of the         *
 *                                    nonzeroes in the vector.              *
 *                                                                          *
 *                  sbucketstorestruct is a long array of buckets, and a*
 *                                    pointer to the next store, if needed. *
 *                                    The basic idea is to allocate space   *
 *                                    in large blocks so that dynamic       *
 *                                    memory allocations will be kept to a  *
 *                                    minimum, yet vectors will be able to  *
 *                                    "grow as they go" by having access to *
 *                                    a memory store of allocated buckets.  *
 *                                                                          *
 *                  svectorstorestruct is a long array of vectors, and a*
 *                                    pointer to the next store, if needed. *
 *                                    The basic idea is the same as in      *
 *                                    sbucketstorestruct.               *
 *                                                                          *
 *     To represent a heap, we define 1 struct:                             *
 *                                                                          *
 *                  heapstruct        represents a node in the binary heap  *
 *                                    tree.  Each node of the heap is a     *
 *                                    placeholder for a nonzero vector      *
 *                                    element.  The nonzero itself is not   *
 *                                    stored in the heap; instead its       *
 *                                    vector index (order of its vector in  *
 *                                    the rhs of the saxpy),  and its       *
 *                                    subscript index (its array index in   *
 *                                    the non-sparse vector) are placed in  *
 *                                    the heap node.  Additionally, each    *
 *                                    heap node contains two pointer fields *
 *                                    containing addresses of the left and  *
 *                                    right children of the node.           *
 *                                    The subscript index serves as the key *
 *                                    for ordering the heap: if x and p(x)  *
 *                                    are a node and its parent, then       *
 *                                    key_p(x) <= key_x.  It follows that   *
 *                                    key_root is the smallest key in the   *
 *                                    tree at all times, and thus by always *
 *                                    poping the root and "consuming" it in *
 *                                    the saxpy, the lhs vector will be     *
 *                                    constructed in the correct (i.e.,     *
 *                                    ascending) order of subscripts.       *
 *                                                                          *
 *     Each typedef below is accompanied by a pictorial depiction.          *
 *                                                                          *
 *     A Note on C syntax: Several of the data structures are defined with  *
 *     the following template:                                              *
 *                                                                          *
 *          typedef struct _Astruct {                                       *
 *             ...                                                          *
 *             struct _Astruct *nextfield;                                  *
 *          } Astruct;                                                      *
 *                                                                          *
 *     _Astruct is the "structure tag".  It's declared explicitly so that   *
 *     the pointer field inside would recognize what it's pointing to as a  *
 *     defined structure.  Then, typedef defines a synonym, "Astruct", for  *
 *     "struct _Astruct".                                                   *
 *                                                                          *
 ****************************************************************************/


typedef struct _sbucketstruct {

  int *index;                        /* Subscript indices of nonzeroes.     */
  float *value;           /* Values of nonzeroes.                */

#ifdef _DEBUG                        /* Debugging Flag Field:               */
  int errorflag;                     /*    to ensure integrity of mallocs.  */
#endif

  struct _sbucketstruct *nextbucket;  /* nsia.                          */
} sbucketstruct;

/*             ___________________bucketstruct___________________           *
 *             |                                                |           *
 *             |                   ___index___   ____value___   |           *
 *             |                   |         |   |          |   |           *
 *             |                 0 |  (int)  |   | (double) |   |           *
 *             |                   |_________|   |__________|   |           *
 *             |                   |         |   |          |   |           *
 *             |                 1 |  (int)  |   | (double) |   |           *
 *             |                   |_________|   |__________|   |           *
 *             |                   |    .    |   |     .    |   |           *
 *             |                   |    .    |   |     .    |   |           *
 *             |                   |    .    |   |     .    |   |           *
 *             |                   |_________|   |__________|   |           *
 *             |                   |         |   |          |   |           *
 *             | sBUCKET_DIM-1 |  (int)  |   | (double) |   |           *
 *             |                   |_________|   |__________|   |           *
 *             |                                                |           *
 *             |             nextbucket *                       |           *
 *             |                        |                       |           *
 *             |________________________|_______________________|           *
 *                                      |                                   *
 *                                      V                                   *
 *                               (sbucketstruct)                        */




typedef struct _sibucketstruct {
  int *index_RANGE_BEGIN;
  int *index_RANGE_END;
                                     /* Subscript index pairs of range of   *
				      * nonzeroes.  Note: since in C, 2-d   *
				      * array elements are stored by row,   *
				      * index_RANGE_BEGIN and               *
				      * index_RANGE_END,                    *
				      * represent a range and are stored    *
				      * contiguously, which is advantageous *
				      * from a cache memory point of view,  *
				      * since they are also accessed in     *
				      * in tandem.                          */

#ifdef _DEBUG                        /* Debugging Flag Field:               */
  int errorflag;                     /*    to ensure integrity of mallocs.  */
#endif

  struct _sibucketstruct *nextibucket;  /* nsia.                        */
} sibucketstruct;

/*     __________________________ibucketstruct__________________________    *
 *     |                                                               |    *
 *     |                       ______index______                       |    *
 *     |                       |       |       |                       |    *
 *     |                     0 | (int) | (int) |                       |    *
 *     |                       |_______|_______|                       |    *
 *     |                       |       |       |                       |    *
 *     |                     1 | (int) | (int) |                       |    *
 *     |                       |_______|_______|                       |    *
 *     |                       |   .   |   .   |                       |    *
 *     |                       |   .   |   .   |                       |    *
 *     |                       |   .   |   .   |                       |    *
 *     |                       |_______|_______|                       |    *
 *     |                       |       |       |                       |    *
 *     |    sIBUCKET_DIM-1 | (int) | (int) |                       |    *
 *     |                       |_______|_______|                       |    *
 *     |                                                               |    *
 *     |                   nextibucket *                               |    *
 *     |                               |                               |    *
 *     |_______________________________|_______________________________|    *
 *                                     |                                    *
 *                                     V                                    *
 *                              (sibucketstruct)                        */



typedef struct _svbucketstruct {

  float *value;           /* Values of nonzeroes.                */

#ifdef _DEBUG                        /* Debugging Flag Field:               */
  int errorflag;                     /*    to ensure integrity of mallocs.  */
#endif

  struct _svbucketstruct *nextvbucket; /* nsia.                         */
} svbucketstruct;

/*                   _____________vbucketstruct_____________                *
 *                   |                                     |                *
 *                   |                    ____value___     |                *
 *                   |                    |          |     |                *
 *                   |                  0 | (double) |     |                *
 *                   |                    |__________|     |                *
 *                   |                    |          |     |                *
 *                   |                  1 | (double) |     |                *
 *                   |                    |__________|     |                *
 *                   |                    |     .    |     |                *
 *                   |                    |     .    |     |                *
 *                   |                    |     .    |     |                *
 *                   |                    |__________|     |                *
 *                   |                    |          |     |                *
 *                   | sVBUCKET_DIM-1 | (double) |     |                *
 *                   |                    |__________|     |                *
 *                   |                                     |                *
 *                   |      nextvbucket *                  |                *
 *                   |                  |                  |                *
 *                   |__________________|__________________|                *
 *                                      |                                   *
 *                                      V                                   *
 *                               (svbucketstruct)                       */


/*
 * Sparse Vector Internal Data Structure (DS) Types: S1, SS, CS
 */

typedef struct _sS1struct {
  int index;

  float value;

} sS1struct;

typedef struct _sSSstruct {
  sbucketstruct *bucketptr;      /* Pointer to linked list of buckets.  */
} sSSstruct;

typedef struct _sCSstruct {
  int nonzeroranges;                 /* Number of contiguous nonzero ranges.*/
  sibucketstruct *ibucketptr;    /* Pointer to linked list of ibuckets. */
  svbucketstruct *vbucketptr;    /* Pointer to linked list of vbuckets. */
} sCSstruct;


typedef struct {
  int nonzeroes;                    /* The total # of                       *
				     * nonzeroes in the vector.             */
  int dstype;                       /* The data-structure type of the       *
				     * svectorstruct will be one of the *
				     * following:
				     *          S1 = 1 SUBSCRIPT            *
				     *          SS = SINGLE SUBSCRIPTS      *
				     *          CS = COMPRESSED SUBSCRIPTS  */
  union _sds {
    sS1struct s1;
    sSSstruct ss;
    sCSstruct cs;
  } sds;

#ifdef _VTRACE

  /* Symbolic identifier */
  char* id;
  /* Numeric identifier */
  int num_id;

  /* Tracked? */
  unsigned char is_tracked;
  /* Named? */
  unsigned char is_named;
  /* Visualized? */
  unsigned char is_visual;

  /* List of all updates over the vector's lifetime */
  pu_struct u_list;

#endif

} svectorstruct;

/*     ___________________________vectorstruct___________________________   *
 *     |                                                                |   *
 *     |          ____nonzeroes____         _____dstype______           |   *
 *     |          |               |         |               |           |   *
 *     |          |     (int)     |         |     (int)     |           |   *
 *     |          |_______________|         |_______________|           |   *
 *     |                                                                |   *
 *     |                               ds                               |   *
 *     |    {~~~~~~~~~~~~~~~~~~~~~~~~~~--~~~~~~~~~~~~~~~~~~~~~~~~~~}    |   *
 *     |    {                                                      }    |   *
 *     |    { EITHER:                                              }    |   *
 *     |    {     _____________________S1_____________________     }    |   *
 *     |    {     |   _____index_____     _____value_____    |     }    |   *
 *     |    {     |   |             |     |             |    |     }    |   *
 *     |    {     |   |    (int)    |     |    (int)    |    |     }    |   *
 *     |    {     |   |_____________|     |_____________|    |     }    |   *
 *     |    {     |__________________________________________|     }    |   *
 *     |    {                                                      }    |   *
 *     |    { OR:                                                  }    |   *
 *     |    {     _____________________SS_____________________     }    |   *
 *     |    {     |                                          |     }    |   *
 *     |    {     |           bucketptr *                    |     }    |   *
 *     |    {     |                     |                    |     }    |   *
 *     |    {     |_____________________|____________________|     }    |   *
 *     |    {                           |                          }    |   *
 *     |    {                           V                          }    |   *
 *     |    {                     (sbucketstruct)              }    |   *
 *     |    {                                                      }    |   *
 *     |    { OR:                                                  }    |   *
 *     |    {     _____________________CS_____________________     }    |   *
 *     |    {     |                                          |     }    |   *
 *     |    {     |              _nonzeroranges_             |     }    |   *
 *     |    {     |              |             |             |     }    |   *
 *     |    {     |              |    (int)    |             |     }    |   *
 *     |    {     |              |_____________|             |     }    |   *
 *     |    {     |                                          |     }    |   *
 *     |    {     |     ibucketptr *        vbucketptr *     |     }    |   *
 *     |    {     |                |                   |     |     }    |   *
 *     |    {     |________________|___________________|_____|     }    |   *
 *     |    {                      |                   |           }    |   *
 *     |    {                      V                   V           }    |   *
 *     |    {          (sibucketstruct)  (svbucketstruct)  }    |   *
 *     |    {                                                      }    |   *
 *     |    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~    |   *
 *     |                                                                |   *
 *     |________________________________________________________________|   */




typedef struct _svectorstorestruct {
  svectorstruct vector[sVECTORSTORE_DIM];  /* Long array of vectors */
  struct _svectorstorestruct *nextstore;       /* nsia.                 */
} svectorstorestruct;

/*             ________________vectorstorestruct________________            *
 *             |                                               |            *
 *             |                        ______vector______     |            *
 *             |                        |                |     |            *
 *             |                      0 | (svectorstruct) ||            *
 *             |                        |________________|     |            *
 *             |                        |                |     |            *
 *             |                      1 | (svectorstruct) ||            *
 *             |                        |________________|     |            *
 *             |                        |        .       |     |            *
 *             |                        |        .       |     |            *
 *             |                        |        .       |     |            *
 *             |                        |________________|     |            *
 *             |                        |                |     |            *
 *             | sVECTORSTORE_DIM-1 | (svectorstruct) ||            *
 *             |                        |________________|     |            *
 *             |                                               |            *
 *             |             nextstore *                       |            *
 *             |                       |                       |            *
 *             |_______________________|_______________________|            *
 *                                     |                                    *
 *                                     V                                    *
 *                            (svectorstorestruct)                      */






typedef struct _sbucketstorestruct {
  sbucketstruct bucket[sBUCKETSTORE_DIM];  /* Long array of buckets.*/
  struct _sbucketstorestruct *nextstore;       /* nsia.                 */
} sbucketstorestruct;

/*             ________________bucketstorestruct________________            *
 *             |                                               |            *
 *             |                        ______bucket______     |            *
 *             |                        |                |     |            *
 *             |                      0 | (sbucketstruct) ||            *
 *             |                        |________________|     |            *
 *             |                        |                |     |            *
 *             |                      1 | (sbucketstruct) ||            *
 *             |                        |________________|     |            *
 *             |                        |        .       |     |            *
 *             |                        |        .       |     |            *
 *             |                        |        .       |     |            *
 *             |                        |________________|     |            *
 *             |                        |                |     |            *
 *             | sBUCKETSTORE_DIM-1 | (sbucketstruct) ||            *
 *             |                        |________________|     |            *
 *             |                                               |            *
 *             |             nextstore *                       |            *
 *             |                       |                       |            *
 *             |_______________________|_______________________|            *
 *                                     |                                    *
 *                                     V                                    *
 *                            (sbucketstorestruct)                      */


typedef struct _sibucketstorestruct {
  sibucketstruct ibucket[sIBUCKETSTORE_DIM];
                                                 /* Long array of ibuckets. */
  struct _sibucketstorestruct *nextstore;    /* nsia.                   */
} sibucketstorestruct;

/*             ________________ibucketstorestruct________________           *
 *             |                                               |            *
 *             |                        ______bucket______     |            *
 *             |                        |                |     |            *
 *             |                      0 |(sibucketstruct) ||            *
 *             |                        |________________|     |            *
 *             |                        |                |     |            *
 *             |                      1 |(sibucketstruct) ||            *
 *             |                        |________________|     |            *
 *             |                        |        .       |     |            *
 *             |                        |        .       |     |            *
 *             |                        |        .       |     |            *
 *             |                        |________________|     |            *
 *             |                        |                |     |            *
 *             | sIBUCKETSTORE_DIM-1| (sibucketstruct)||            *
 *             |                        |________________|     |            *
 *             |                                               |            *
 *             |             nextstore *                       |            *
 *             |                       |                       |            *
 *             |_______________________|_______________________|            *
 *                                     |                                    *
 *                                     V                                    *
 *                            (sibucketstorestruct)                     */



typedef struct _svbucketstorestruct {
  svbucketstruct vbucket[sVBUCKETSTORE_DIM];
                                                 /* Long array of vbuckets. */
  struct _svbucketstorestruct *nextstore;    /* nsia.                   */
} svbucketstorestruct;

/*             ________________vbucketstorestruct_________________          *
 *             |                                                 |          *
 *             |                         ______vbucket______     |          *
 *             |                         |                 |     |          *
 *             |                       0 | (svbucketstruct) ||          *
 *             |                         |_________________|     |          *
 *             |                         |                 |     |          *
 *             |                       1 | (svbucketstruct) ||          *
 *             |                         |_________________|     |          *
 *             |                         |         .       |     |          *
 *             |                         |         .       |     |          *
 *             |                         |         .       |     |          *
 *             |                         |_________________|     |          *
 *             |                         |                 |     |          *
 *             | sVBUCKETSTORE_DIM-1 | (svbucketstruct) ||          *
 *             |                         |_________________|     |          *
 *             |                                                 |          *
 *             |              nextstore *                        |          *
 *             |                        |                        |          *
 *             |________________________|________________________|          *
 *                                      |                                   *
 *                                      V                                   *
 *                            (svbucketstorestruct)                     */





/****************************************************************************
 *                                                                          *
 *  Function Prototyping                                                    *
 *                                                                          *
 ****************************************************************************/

#define _P(x) ()

void sallocatebucketstore _P(());
void sallocateibucketstore _P(());
void sallocatevbucketstore _P(());
void sallocatevectorstore _P(());

void sgetbucket _P((sbucketstruct **bucketptrptr));
void sgetibucket _P((sibucketstruct **ibucketptrptr));
void sgetvbucket _P((svbucketstruct **vbucketptrptr));
void sgetvector _P((svectorstruct **vectorptrptr));

void sspgx _P((svectorstruct **destptrptr,
		   int arity,
		   float *alphaptrs[],
		   svectorstruct **rhsvectorptrptrs[]));

void snon_over_lapping_cs_range_axpy _P((int rangedestination,
				      int lhssibucketindex,
				      int *addressoflastlhssibucketindex,
				      int *addressoflhssvbucketindex,
				      int k,
				      int *addressofrhssvbucketindex_k,
				      float alpha_k,
				      int totalcurrentnonzerorange
#ifdef _DEBUG
				      ,int *addressofunconsumedcurrentnonzerorange
				      ,int rhssvectorconsumednonzeroranges_k
				      ,int rhssvectornonzeroranges_k
				      ,int lhssvectornonzeroranges
#endif
				      ));

void soverlapping_cs_range_axpy _P((int *addressoflhssvbucketindex,
				   int k,
				   int *addressofrhssvbucketindex_k,
				   float alpha_k,
				   int totalcurrentnonzerorange
#ifdef _DEBUG
				   ,int *addressofunconsumedcurrentnonzerorange
				   ,int rhssvectorconsumednonzeroranges_k
				   ,int rhssvectornonzeroranges_k
				   ,int lhssvectornonzeroranges
#endif
				   ));


void sdetermine_loop_bounds _P((int totalcurrentnonzerorange,
			       int lhsvbucketindex,
			       int rhsvbucketindex,
			       int *firstloopsize,
			       int *numberofdoubleforloopiterations,
			       int *firstiterationloopsize,
			       int *seconditerationloopsize,
			       int *remainder1loopsize,
			       int *remainder2loopsize
#ifdef _DEBUG
			       ,int *addressofunconsumedcurrentnonzerorange
#endif
			       ));

void sdebugging_code_1 _P((char *RIGHT_OR_LEFT,
			  int vbucketindex,
			  int unconsumedcurrentnonzerorange));

void shop_to_next_rhsvbucket _P((int k,
				int *vbucketindex
#ifdef _DEBUG
				,int rhssvectorconsumednonzeroranges
				,int rhssvectornonzeroranges
#endif
				));


void shop_to_next_lhsvbucket _P((int *vbucketindex
#ifdef _DEBUG
				,int lhssvectornonzeroranges
#endif
				));

void sget_and_hop_to_next_vbucket _P((int *lhssvbucketindex));

void sconvertdstype _P((svectorstruct **oldvectorptrptr,
		       int k,
		       int newvectordstype));

void saxpy_0_1_post_processing _P((svectorstruct **destptrptr));

void saxpy_2_heap_post_processing _P((svectorstruct **destptrptr,
                                     svectorstruct **rhssvectorptrptrs[],
                                     float *alphaptrs[],
                                     svectorstruct *restorerhssvectorptr[MAXARITY+1],
                                     int arity,
                                     int lhssvectornonzeroes,
                                     int lhssvectornonzeroranges,
                                     int governingdstype));

void splusequal_axpy_postprocessing _P((svectorstruct **destptrptr,
				       svectorstruct **rhssvectorptrptr,
				       int status_otherrhssvector,
				       svectorstruct **restofplusequalvectorptrptr,
				       float *alphaptrs[2],
				       svectorstruct *restorerhssvectorptr[2+1],
				       int arity,
				       int lhssvectornonzeroes,
				       int lhssvectornonzeroranges,
				       int governingdstype
				       ));

#ifdef _PEEP
void sadpeep_pre _P((int vectorindex,
		    svectorstruct **rhsvectorptrptr,
		    float *alphaptr));

void sadpeep_post _P((svectorstruct **destptrptr));
#endif

void sadallocate _P(());
void sadfree _P(());
void sad_free_vectors_buckets _P((svectorstruct **vectorptrptr));
void sinitialize_pd_global_variables _P(());

void sspsd _P((svectorstruct **destptrptr,
		int *indexvector,
		float *valuevector,
		int *lengthptr));

void sspxd _P((float *vector,
			 int *lengthptr,
			 svectorstruct **objectptrptr,
			 int *returned_len,
			 int *len_info));

void sspxm _P((float *vector,
			       int *lengthptr,
			       float *multiplier,
			       svectorstruct **objectptrptr,
			       int *returned_len,
			       int *len_info));

void sspxa _P((float *vector,
				   int *lengthptr,
				   svectorstruct **objectptrptr,
				   int *returned_len,
				   int *len_info));

void sspxs _P((int *indvector,
			  float *valvector,
			  int *lengthptr,
			  svectorstruct **objectptrptr,
			  int *returned_len,
			  int *len_info)); 

void ssppr _P((svectorstruct **objectptrptr, int *unit_number));

void sspzronotca _P((svectorstruct **destptrptr));

void sspcp _P((svectorstruct **destptrptr,
		svectorstruct **oldvectorptrptr));

void sadinfo _P((int arity,
		int nnzb,
		svectorstruct *lhsvectoraddress,
		int nnza,
		float *alphaptrs[],
		svectorstruct **vectorptrptrs[]));

void sspg1 _P((svectorstruct **destptrptr,
		float *alphaptr,
		svectorstruct **vectoraptrptr));
void sspg2 _P((svectorstruct **destptrptr,
		float *alphaptr,
		svectorstruct **vectoraptrptr,
		float *betaptr,
		svectorstruct **vectorbptrptr));
void plusequalsspg1 _P((svectorstruct **rhsanddestptrptr,
			 float *alphaptr));
void plusequalsspg2 _P((svectorstruct **rhsanddestptrptr,
			 float *otheralphaptr,
			 svectorstruct **otherrhsvectorptrptr,
			 int status_otherrhsvector));



#endif /*  #ifndef sGAXPY_PD_MACROS_AND_DATA_STRUCTURES  */
