#define SINGLE
/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#include  <stdlib.h>
#include "sgaxpy_pd_macros_and_data_structures.h"
#include "sgaxpy_pd_global_variables.h"


/****************************************************************************
 *                                                                          *
 *  sallocatebucketstore()                                                   *
 *                                                                          *
 *     1 malloc call gets all the memory for a bucketstore.  If malloc does *
 *     not return NULL, one of the following 2 scenarios will result:       *
 *                                                                          *
 *     A: outcome of sallocatebucketstore the 1st time it's called:          *
 *                                                                          *
 *     bucketstoreptr -----------+                                          *
 *                               |   store                                  *
 *                              _V_____1_____                               *
 *     bucketstorestartptr ---->|    _____  |                               *
 *     bucketstorecounter ------+->0 |___|  |                               *
 *                              |    | . |  |                               *
 *                              |    | . |  |                               *
 *                              |    | . |  |                               *
 *                              |    |___|  |                               *
 *                              |    |___|  |                               *
 *                              |           |                               *
 *                              | nextstore |                               *
 *                              |     *     |                               *
 *                              |_____|_____|                               *
 *                                    |                                     *
 *                                    V                                     *
 *                                   NULL                                   *
 *                                                                          *
 *     B: outcome of sallocatebucketstore the Nth time it's called (N > 1):  *
 *                                                                          *
 *     bucketstorecounter --------------------------------------+           *
 *                                                              |           *
 *     bucketstoreptr ----------------------------------------+ |           *
 *                               store        store           | | store     *
 *                              ___1___      ___2___ ...     _V_|___N_____  *
 *     bucketstorestartptr ---->|     |  +-->|     |     +-->|  V _____  |  *
 *                              |     |  |   |     |     |   |  0 |___|  |  *
 *                              |     |  |   |     |     |   |    | . |  |  *
 *                              |     |  |   |     |     |   |    | . |  |  *
 *                              |     |  |   |     |     |   |    | . |  |  *
 *                              |     |  |   |     |     |   |    |___|  |  *
 *                              |     |  |   |     |     |   |    |___|  |  *
 *                              |     |  |   |     |     |   |           |  *
 *                              |     |  |   |     |     |   | nextstore |  *
 *                              |  *  |  |   |  *  |     |   |     *     |  *
 *                              |__|__|  |   |__|__|     |   |_____|_____|  *
 *                                 |     |      |        |         |        *
 *                                 +-----+      +---...--+         V        *
 *                                                                NULL      *
 *                                                                          *
 ****************************************************************************/

void sallocatebucketstore()
{
  int i,k;
  
  sbucketstorestruct *tempbucketstoreptr;
  sbucketstruct *tempbucketptr;
  
  int *index_store;
  float *value_store;

  int temp_size;

  /*
   * Per BUCKET Dynamic Allocation: constant sBUCKETSTORE_DIM
   *                                variable sBUCKET_DIM
   */
  
  if ((tempbucketstoreptr = (sbucketstorestruct *)
       palloc ((temp_size = (sizeof(sbucketstorestruct)))))==NULL) {
    printf("sallocatebucketstore error: malloc returned NULL.\n");
    abort();
  }
  memory_count(S_EVAL,MEM_BUCKET,temp_size);
  
  for (k=0; k<sBUCKETSTORE_DIM; k++) {
    tempbucketptr = &(tempbucketstoreptr->bucket[k]);
    if ((tempbucketptr->index = (int *)
	 palloc ((temp_size = (sBUCKET_DIM * sizeof(int)))))==NULL) {
      printf("sallocatebucketstore error: malloc returned NULL.\n");
      abort();
    }
    memory_count(S_EVAL,MEM_BUCKET,temp_size);
    if ((tempbucketptr->value = (float *)
	 palloc ((temp_size = (sBUCKET_DIM *
			       sizeof(float)))))==NULL) {
      printf("sallocatebucketstore error: malloc returned NULL.\n");
      abort();
    }
    memory_count(S_EVAL,MEM_BUCKET,temp_size);
  }
      
  if (sbucketstorestartptr == NULL)
    sbucketstorestartptr = tempbucketstoreptr;
  else
    sbucketstoreptr->nextstore = tempbucketstoreptr;
  
  sbucketstoreptr = tempbucketstoreptr;
  sbucketstoreptr->nextstore = NULL;
  sbucketstorecounter = 0;
  
  for (i=0;i<sBUCKETSTORE_DIM;i++) {
    (&(sbucketstoreptr->bucket[i]))->nextbucket = NULL;
#ifdef _DEBUG  
    (&(sbucketstoreptr->bucket[i]))->errorflag = 9999;
#endif
  }
}



/****************************************************************************
 *                                                                          *
 *  sallocateibucketstore()                                                  *
 *                                                                          *
 *     Identical to sallocatebucketstore in terms of control flow.           *
 *                                                                          *
 ****************************************************************************/

void sallocateibucketstore()
{
  int i,k;

  sibucketstorestruct *tempibucketstoreptr;
  sibucketstruct *tempibucketptr;
  
  int *index_RANGE_BEGIN_store;
  int *index_RANGE_END_store;

  int temp_size;
 
  /*
   * Per IBUCKET Dynamic Allocation: constant sIBUCKETSTORE_DIM
   *                                 variable sIBUCKET_DIM
   */
  
  if ((tempibucketstoreptr = (sibucketstorestruct *)
       palloc ((temp_size = (sizeof(sibucketstorestruct)))))==NULL) {
    printf("sallocateibucketstore error: malloc returned NULL.\n");
    abort();
  }
  memory_count(S_EVAL,MEM_IBUCKET,temp_size);
  
  for (k=0; k<sIBUCKETSTORE_DIM; k++) {
    tempibucketptr = &(tempibucketstoreptr->ibucket[k]);
    if ((tempibucketptr->index_RANGE_BEGIN = (int *)
	 palloc ((temp_size = (sIBUCKET_DIM * sizeof(int)))))==NULL) {
      printf("sallocateibucketstore error: malloc returned NULL.\n");
      abort();
    }
    memory_count(S_EVAL,MEM_IBUCKET,temp_size);
    if ((tempibucketptr->index_RANGE_END = (int *)
	 palloc ((temp_size = (sIBUCKET_DIM * sizeof(int)))))==NULL) {
      printf("sallocateibucketstore error: malloc returned NULL.\n");
      abort();
    }
    memory_count(S_EVAL,MEM_IBUCKET,temp_size);
  }
  
  if (sibucketstorestartptr == NULL)
    sibucketstorestartptr = tempibucketstoreptr;
  else
    sibucketstoreptr->nextstore = tempibucketstoreptr;
  
  sibucketstoreptr = tempibucketstoreptr;
  sibucketstoreptr->nextstore = NULL;
  sibucketstorecounter = 0;
  
  for (i=0;i<sIBUCKETSTORE_DIM;i++) {
    (&(sibucketstoreptr->ibucket[i]))->nextibucket = NULL;
#ifdef _DEBUG  
    (&(sibucketstoreptr->ibucket[i]))->errorflag = 9099;
#endif
  }
}




/****************************************************************************
 *                                                                          *
 *  sallocatevbucketstore()                                                  *
 *                                                                          *
 *     Identical to sallocatebucketstore in terms of control flow.           *
 *                                                                          *
 ****************************************************************************/

void sallocatevbucketstore()
{
  int i,k;

  svbucketstorestruct *tempvbucketstoreptr;
  svbucketstruct *tempvbucketptr;
  
  float *value_store;

  int temp_size;
  
  /*
   * Per VBUCKET Dynamic Allocation: constant sVBUCKETSTORE_DIM
   *                                 variable sVBUCKET_DIM
   */
  
  if ((tempvbucketstoreptr = (svbucketstorestruct *)
       palloc ((temp_size = (sizeof(svbucketstorestruct)))))==NULL) {
    printf("sallocatevbucketstore error: malloc returned NULL.\n");
    abort();
  }
  memory_count(S_EVAL,MEM_VBUCKET,temp_size);
  
  for (k=0; k<sVBUCKETSTORE_DIM; k++) {
    tempvbucketptr = &(tempvbucketstoreptr->vbucket[k]);
    if ((tempvbucketptr->value = (float *)
	 palloc ((temp_size = (sVBUCKET_DIM *
			       sizeof(float)))))==NULL) {
      printf("sallocatevbucketstore error: malloc returned NULL.\n");
      abort();
    }
    memory_count(S_EVAL,MEM_VBUCKET,temp_size);
  }
  
  if (svbucketstorestartptr == NULL)
    svbucketstorestartptr = tempvbucketstoreptr;
  else
    svbucketstoreptr->nextstore = tempvbucketstoreptr;
  
  svbucketstoreptr = tempvbucketstoreptr;
  svbucketstoreptr->nextstore = NULL;
  svbucketstorecounter = 0;
  
  for (i=0;i<sVBUCKETSTORE_DIM;i++) {
    (&(svbucketstoreptr->vbucket[i]))->nextvbucket = NULL;
#ifdef _DEBUG  
    (&(svbucketstoreptr->vbucket[i]))->errorflag = 9009;
#endif
  }
}




/****************************************************************************
 *                                                                          *
 *  sallocatevectorstore()                                                   *
 *                                                                          *
 *     sVECTORSTORE_DIM is a constant and in contrast to the buckets    *
 *     there is no VECTOR_DIM.  Therefore there is no need for any          *
 *     allocation scheme requiring the per STORE vs per VECTOR methods.     *
 *                                                                          *
 ****************************************************************************/

void sallocatevectorstore()
{
  int i;
  svectorstorestruct *tempvectorstoreptr;
  svectorstruct *tempvectorptr;

  int temp_size;

  /*
   * Wholesale Dynamic Allocation: constant sVECTORSTORE_DIM
   * This allocation is not obsolete unlike the bucket type stores
   * above, since there is no variable-sized bucket in a vectorstruct.
   */
  if ((tempvectorstoreptr = (svectorstorestruct *)
       palloc ((temp_size = (sizeof(svectorstorestruct)))))==NULL ) {
    printf("sallocatevectorstore error: malloc returned NULL.\n");
    abort();
  }
  memory_count(S_EVAL,MEM_VECTOR,temp_size);
  
  /*
   * End of Wholesale Dynamic Allocation
   */
  
  if (svectorstorestartptr == NULL)
    svectorstorestartptr = tempvectorstoreptr;
  else
    svectorstoreptr->nextstore = tempvectorstoreptr;
  
  svectorstoreptr = tempvectorstoreptr;
  svectorstoreptr->nextstore = NULL;
  svectorstorecounter = 0;
  for (i=0;i<sVECTORSTORE_DIM;i++) {
    tempvectorptr = &(svectorstoreptr->vector[i]);

/*
 *  PMK: 12/4/96 edits
 *
 *  The assignment of the S0 "non-yet-specified" data structure type is
 *  extraneous for now, since we never have checks on S0 (i.e., nowhere
 *  in the code do we say, if (somevectorptr->dstype == S0) ...
 *  Therefore, the following code line is commented out:

    tempvectorptr->dstype = S0;

 *
 */

  }
}



/****************************************************************************
 *                                                                          *
 *  sgetbucket(bucketptrptr)                                                 *
 *                                                                          *
 *     bucketptrptr    is a doubly-indirected pointer to a sbucketstruct*
 *                     Since we want the value of "bucketptr" to be changed *
 *                     by sgetbucket (i.e., *bucketptrptr to get assigned a  *
 *                     pointer to a sbucketstruct), we need to make its *
 *                     address (bucketptrptr) the function argument.  Thus, *
 *                     the need for the double indirection (**).            *
 *                                                                          *
 *     sgetbucket assigns to *bucketptrptr the pointer to the next available *
 *     sbucketstruct from the freedbucketslist or if none there from the*
 *     bucketstore.  In the latter case, if we're at the end of the current *
 *     bucketstore, a new store is allocated before doing the "get".        *
 *                                                                          *
 ****************************************************************************/

void sgetbucket(bucketptrptr)
sbucketstruct **bucketptrptr;
{
  if (freedsbucketslist != NULL) {
    (*bucketptrptr) = freedsbucketslist;
    freedsbucketslist = freedsbucketslist->nextbucket;
  }
  else {
    if (sbucketstorecounter == sBUCKETSTORE_DIM) {
      sallocatebucketstore();
    }
    
#ifdef _DEBUG
    else if (sbucketstorecounter > sBUCKETSTORE_DIM) {
      printf("sgetbucket error: sbucketstorecounter=%d,  STOREDIMENSION=%d\n",
	     sbucketstorecounter,sBUCKETSTORE_DIM);
      abort();
    }
#endif

    /*
     *  The following assignment of a store entry to *bucketptrptr
     *  works regardless of whether we've used per BUCKET or per STORE
     *  allocation method.
     */

    *bucketptrptr = (sbucketstruct *)
                     &(sbucketstoreptr->bucket[sbucketstorecounter++]);
  }
  (*bucketptrptr)->nextbucket = NULL;
  
#ifdef _DEBUG
  if ((*bucketptrptr)->errorflag != 9999) {
    printf("sgetbucket error: memory corruption of sbucketstore\n");
    abort();
  }
#endif
  
}

/****************************************************************************
 *                                                                          *
 *  sgetibucket(ibucketptrptr)                                               *
 *                                                                          *
 *     ibucketptrptr   is a doubly-indirected pointer to a sibucketstruc*
 *                     Since we want the value of "ibucketptr" to be        *
 *                     changed by sgetibucket (i.e., *ibucketptrptr to get   *
 *                     assigned a  pointer to a sibucketstruct), we need*
 *                     to make its address (ibucketptrptr) the function     *
 *                     argument.  Thus, the need for the double indirection *
 *                     (**).                                                *
 *                                                                          *
 *     sgetibucket assigns to *ibucketptrptr the pointer to the next         *
 *     available sibucketstruct from the freedibucketslist or if none   *
 *     from the ibucketstore.  In the latter case, if we're at the end      *
 *     of the current ibucketstore, a new store is allocated before doing   *
 *     the "get".                                                           *
 *                                                                          *
 ****************************************************************************/

void sgetibucket(ibucketptrptr)
sibucketstruct **ibucketptrptr;
{
  if (freedsibucketslist != NULL) {
    (*ibucketptrptr) = freedsibucketslist;
    freedsibucketslist = freedsibucketslist->nextibucket;
  }
  else {
    if (sibucketstorecounter == sIBUCKETSTORE_DIM) {
      sallocateibucketstore();
    }
    
#ifdef _DEBUG
    else if (sibucketstorecounter > sIBUCKETSTORE_DIM) {
      printf("sgetibucket error: sibucketstorecounter=%d,  STOREDIMENSION=%d\n",
	     sibucketstorecounter,sIBUCKETSTORE_DIM);
      abort();
    }
#endif

    *ibucketptrptr = (sibucketstruct *)
              &(sibucketstoreptr->ibucket[sibucketstorecounter++]);
  }
  (*ibucketptrptr)->nextibucket = NULL;
  
#ifdef _DEBUG
  if ((*ibucketptrptr)->errorflag != 9099) {
    printf("sgetibucket error: memory corruption of sibucketstore\n");
    abort();
  }
#endif
  
}


/****************************************************************************
 *                                                                          *
 *  sgetvbucket(vbucketptrptr)                                               *
 *                                                                          *
 *     vbucketptrptr   is a doubly-indirected pointer to a svbucketstruc*
 *                     Since we want the value of "vbucketptr" to be        *
 *                     changed by sgetvbucket (i.e., *vbucketptrptr to get   *
 *                     assigned a  pointer to a svbucketstruct), we need*
 *                     to make its address (vbucketptrptr) the function     *
 *                     argument.  Thus, the need for the double indirection *
 *                     (**).                                                *
 *                                                                          *
 *     sgetvbucket assigns to *vbucketptrptr the pointer to the next         *
 *     available svbucketstruct from the freedvbucketslist or if none   *
 *     there from the vbucketstore.  In the latter case, if we're at the end*
 *     of the current vbucketstore, a new store is allocated before doing   *
 *     the "get".                                                           *
 *                                                                          *
 ****************************************************************************/

void sgetvbucket(vbucketptrptr)
svbucketstruct **vbucketptrptr;
{
  if (freedsvbucketslist != NULL) {
    (*vbucketptrptr) = freedsvbucketslist;
    freedsvbucketslist = freedsvbucketslist->nextvbucket;
  }
  else {
    if (svbucketstorecounter == sVBUCKETSTORE_DIM) {
      sallocatevbucketstore();
    }
    
#ifdef _DEBUG
    else if (svbucketstorecounter > sVBUCKETSTORE_DIM) {
      printf("sgetvbucket error: svbucketstorecounter=%d,  STOREDIMENSION=%d\n",
	     svbucketstorecounter,sVBUCKETSTORE_DIM);
      abort();
    }
#endif

    *vbucketptrptr = (svbucketstruct *)
              &(svbucketstoreptr->vbucket[svbucketstorecounter++]);
  }
  (*vbucketptrptr)->nextvbucket = NULL;
  
#ifdef _DEBUG
  if ((*vbucketptrptr)->errorflag != 9009) {
    printf("sgetvbucket error: memory corruption of svbucketstore\n");
    abort();
  }
#endif
  
}



/****************************************************************************
 *                                                                          *
 *  sgetvector(vectorptrptr)                                                 *
 *                                                                          *
 *     vectorptrptr  is a doubly-indirected pointer to a svectorstruct. *
 *                                                                          *
 *     sgetvector works in the same way as sgetvector, except that there is   *
 *     no freed list of vectors.  It replaces the functionality of "adnew". *
 *                                                                          *
 ****************************************************************************/

void sgetvector(vectorptrptr)
svectorstruct **vectorptrptr;
{
  if (svectorstorecounter == sVECTORSTORE_DIM) {
    sallocatevectorstore();
  }

#ifdef _DEBUG
  else if (svectorstorecounter > sVECTORSTORE_DIM) {
    printf("sgetvector error: svectorstorecount=%d, sVECTORSTORE_DIM=%d\n",
	   svectorstorecounter,sVECTORSTORE_DIM);
    abort();
  }
#endif

  *vectorptrptr = (svectorstruct *)
                    &(svectorstoreptr->vector[svectorstorecounter++]);

#ifdef _TRACE
  sadinfo(-1,0,*vectorptrptr,0,NULL,NULL);
#endif

}

