#define SINGLE
/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#include  <stdlib.h>
#include "sgaxpy_pd_macros_and_data_structures.h"
#include "sgaxpy_pd_global_variables.h"

/****************************************************************************
 *                                                                          *
 *     sspzronotca assigns a zero vector to its argument.                        *
 *                                                                          *
 ****************************************************************************/

void sspzronotca(destptrptr)
svectorstruct **destptrptr;
{
  if (*destptrptr == NULL)
    sgetvector(destptrptr);
  
  lhssvectorptr->nonzeroes = 0;
  lhssvectorptr->dstype = S1;
  lhssvectorptr->sds.s1.index = DUMMYSUBSCRIPT;
  X_ZERO(lhssvectorptr->sds.s1.value);

  saxpy_0_1_post_processing(destptrptr);
}



/****************************************************************************
 *                                                                          *
 *     sspcp assigns to its first vector argument a vector whose values    *
 *     are identical to its second vectpr argument.                         *
 *                                                                          *
 ****************************************************************************/

void sspcp(destptrptr,oldvectorptrptr)
svectorstruct **destptrptr,**oldvectorptrptr;
{
  svectorstruct *oldvectorptr,*newvectorptr;
  sbucketstruct *oldvectorbucketptr,
                    *newvectortailbucketptr,*newbucketptr;
  sibucketstruct *oldvectoribucketptr,
                     *newvectortailibucketptr,*newibucketptr;
  svbucketstruct *oldvectorvbucketptr,
                     *newvectortailvbucketptr,*newvbucketptr;
  int i,bucketindex,ibucketindex,vbucketindex;

#ifdef _PEEP
  float one;
  X_PUT_REAL(one,1.0);
#endif    

  oldvectorptr = *oldvectorptrptr;

  if (*destptrptr == NULL)
     sgetvector(destptrptr);

#ifdef _PEEP
  adpeep_header(0);
  sadpeep_pre(0,oldvectorptrptr,&one);
#endif    
  newvectorptr = lhssvectorptr;

  newvectorptr->nonzeroes =  oldvectorptr->nonzeroes;
  newvectorptr->dstype =  oldvectorptr->dstype;

  if (newvectorptr->dstype == S1) {
    newvectorptr->sds.s1.index =  oldvectorptr->sds.s1.index;
    newvectorptr->sds.s1.value =  oldvectorptr->sds.s1.value;
  }
  else if (newvectorptr->dstype == SS) {
    oldvectorbucketptr = oldvectorptr->sds.ss.bucketptr;

    for (i=0;i<newvectorptr->nonzeroes;i++) {
      if ((bucketindex = MY_MOD(i,sBUCKET_DIM_minus1)) == 0) {
	if (i == 0) {
	  sgetbucket(&newbucketptr);
	  newvectorptr->sds.ss.bucketptr = newbucketptr;
	}
	else {
	  oldvectorbucketptr = oldvectorbucketptr->nextbucket;
	  sgetbucket(&newbucketptr);
	  newvectortailbucketptr->nextbucket = newbucketptr;
	}
	newvectortailbucketptr = newbucketptr;
      }
      newvectortailbucketptr->value[bucketindex] =
        oldvectorbucketptr->value[bucketindex];
      newvectortailbucketptr->index[bucketindex] =
        oldvectorbucketptr->index[bucketindex];
    }
  }
  else /* if (newvectorptr->dstype == CS */ {
    oldvectoribucketptr = oldvectorptr->sds.cs.ibucketptr;
    oldvectorvbucketptr = oldvectorptr->sds.cs.vbucketptr;
    newvectorptr->sds.cs.nonzeroranges =
      oldvectorptr->sds.cs.nonzeroranges;
    
    for (i=0;i<newvectorptr->sds.cs.nonzeroranges;i++) {
      if ((ibucketindex = MY_MOD(i,sIBUCKET_DIM_minus1)) == 0) {
	if (i == 0) {
	  sgetibucket(&newibucketptr);
	  newvectorptr->sds.cs.ibucketptr = newibucketptr;
	}
	else {
	  oldvectoribucketptr = oldvectoribucketptr->nextibucket;
	  sgetibucket(&newibucketptr);
	  newvectortailibucketptr->nextibucket = newibucketptr;
	}
	newvectortailibucketptr = newibucketptr;
      }

      newvectortailibucketptr->index_RANGE_BEGIN[ibucketindex] =
        oldvectoribucketptr->index_RANGE_BEGIN[ibucketindex];
      newvectortailibucketptr->index_RANGE_END[ibucketindex] =
        oldvectoribucketptr->index_RANGE_END[ibucketindex];
    }
    
    for (i=0;i<newvectorptr->nonzeroes;i++) {
      if ((vbucketindex = MY_MOD(i,sVBUCKET_DIM_minus1)) == 0) {
	if (i == 0) {
	  sgetvbucket(&newvbucketptr);
	  newvectorptr->sds.cs.vbucketptr = newvbucketptr;
	}
	else {
	  oldvectorvbucketptr = oldvectorvbucketptr->nextvbucket;
	  sgetvbucket(&newvbucketptr);
	  newvectortailvbucketptr->nextvbucket = newvbucketptr;
	}
	newvectortailvbucketptr = newvbucketptr;
      }
      newvectortailvbucketptr->value[vbucketindex] =
        oldvectorvbucketptr->value[vbucketindex];
    }
  }

  saxpy_0_1_post_processing(destptrptr);

#ifdef _PEEP
  sadpeep_post(destptrptr);
#endif

}

