#define DOUBLE_COMPLEX
/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#include  <stdlib.h>
#include "zgaxpy_pd_macros_and_data_structures.h"
#include "zgaxpy_pd_global_variables.h"

/****************************************************************************
 *                                                                          *
 *     zspg1 performs the arity 1 gaxpy.                                   *
 *                                                                          *
 ****************************************************************************/

void zspg1(destptrptr,alphaptr,vectoraptrptr)
zvectorstruct **vectoraptrptr,**destptrptr;
double_complex *alphaptr;
{

  int arity = 1;

  zvectorstruct *rhszvectorptr[1];
  zvectorstruct *tempvectorptr;
  
  zbucketstruct *rhszbucketptr;
  zbucketstruct *lhszvectortailbucketptr;
  zbucketstruct *lhszbucketptr;
  zbucketstruct *tempbucketptr;
  
  int i,k;
  int bucketindex;
  
  zvbucketstruct *rhszvectorvbucketptr;
  zvbucketstruct *lhszvectortailvbucketptr;
  zvbucketstruct *tempvbucketptr;
  int vbucketindex;
  
  zibucketstruct *rhszvectoribucketptr;
  zibucketstruct *lhszvectortailibucketptr;
  zibucketstruct *tempibucketptr;
  zibucketstruct *lhszibucketptr;
  int ibucketindex;

  double_complex *alphaptrs[1];
  
#ifdef _TRACE
  int nnzb,nnza;
  zvectorstruct *lhszvectoraddress;
  
  
  zvectorstruct **vectorptrptrs[1];
  
  vectorptrptrs[0] = vectoraptrptr;
  
  if(*destptrptr==NULL){ 
    nnzb = 0; 
    lhszvectoraddress= (zvectorstruct *) 99900; 
  }
  else {
    nnzb = (*destptrptr)->nonzeroes; 
    lhszvectoraddress = *destptrptr;
  }
#endif
  
  
  /****************************************
   * Initialization of *destptrptr.       *
   ****************************************/
  
  if (*destptrptr == NULL)
    zgetvector(destptrptr);
  
  rhszvectorptr[0] = *vectoraptrptr;
  alphaptrs[0] = alphaptr;

#ifdef _PEEP
  if (toggle_adxpy_1_peep) {
    adpeep_header(1);
    zadpeep_pre(0,vectoraptrptr,alphaptr);
  }
#endif
  
  if (rhszvectorptr[0]->nonzeroes == 0 || X_IS_ZERO(*alphaptr)) {
    zspzronotca(destptrptr);
  }
  else /* if not ZERO AXPY */ {
    if (rhszvectorptr[0] == *destptrptr &&
	rhszvectorptr[0]->dstype == CS) {
      plusequalzspg1(destptrptr,alphaptr);
    }
    else /* if (not plusequalzspg1 or (dstype is S1 or SS)) */ {
      lhszvectorptr->nonzeroes = rhszvectorptr[0]->nonzeroes;
      lhszvectorptr->dstype = rhszvectorptr[0]->dstype;

      if (lhszvectorptr->dstype == S1) {
	
	/********************************************************************
	 *                       ONE_SUBSCRIPT AXPY                         *
	 ********************************************************************/


	lhszvectorptr->zds.s1.index =
	                rhszvectorptr[0]->zds.s1.index;
	GAXPY_EQUALS(lhszvectorptr->zds.s1.value,*alphaptr,
		        rhszvectorptr[0]->zds.s1.value);
      }
      else if (lhszvectorptr->dstype == SS) {
	
	/******************************************************************
	 *                 _SINGLE_SUBSCRIPTS AXPY                        *
	 ******************************************************************/
	
	rhszbucketptr = rhszvectorptr[0]->zds.ss.bucketptr;
	for (i=0;i<rhszvectorptr[0]->nonzeroes;i++) {
	  if ((bucketindex = MY_MOD(i,zBUCKET_DIM_minus1)) == 0) {
	    if (i == 0) {
	      zgetbucket(&lhszbucketptr);
	      lhszvectorptr->zds.ss.bucketptr = lhszbucketptr;
	    }
	    else {
	      rhszbucketptr = rhszbucketptr->nextbucket;
	      zgetbucket(&lhszbucketptr);
	      lhszvectortailbucketptr->nextbucket = lhszbucketptr;
	    }
	    lhszvectortailbucketptr = lhszbucketptr;
	  }
	  GAXPY_EQUALS(lhszvectortailbucketptr->value[bucketindex],
		       *alphaptr,rhszbucketptr->value[bucketindex]);
	  lhszvectortailbucketptr->index[bucketindex] =
	    rhszbucketptr->index[bucketindex];
	}
      }
      
      else /* if (lhszvectorptr->dstype == CS) */ {
	  
	/******************************************************************
	 *                  _COMPRESSED_SUBSCRIPTS AXPY                   *
	 ******************************************************************/
	
	lhszvectorptr->zds.cs.nonzeroranges =
	  rhszvectorptr[0]->zds.cs.nonzeroranges;
	rhszvectoribucketptr =
	  rhszvectorptr[0]->zds.cs.ibucketptr;
	rhszvectorvbucketptr =
	  rhszvectorptr[0]->zds.cs.vbucketptr;
	
	for (i=0;i<rhszvectorptr[0]->zds.cs.nonzeroranges;i++) {
	  if ((ibucketindex = MY_MOD(i,zIBUCKET_DIM_minus1)) == 0) {
	    if (i == 0) {
	      zgetibucket(&lhszibucketptr);
	      lhszvectorptr->zds.cs.ibucketptr = lhszibucketptr;
	    }
	    else {
	      rhszvectoribucketptr =
		rhszvectoribucketptr->nextibucket;
	      zgetibucket(&lhszibucketptr);
	      lhszvectortailibucketptr->nextibucket = lhszibucketptr;
	    }
	    lhszvectortailibucketptr = lhszibucketptr;
	  }
	  
	  lhszvectortailibucketptr->index_RANGE_BEGIN[ibucketindex] =
	    rhszvectoribucketptr->index_RANGE_BEGIN[ibucketindex];
	  lhszvectortailibucketptr->index_RANGE_END[ibucketindex] =
	    rhszvectoribucketptr->index_RANGE_END[ibucketindex];
	}
	
	for (i=0;i<rhszvectorptr[0]->nonzeroes;i++) {
	  if ((vbucketindex = MY_MOD(i,zVBUCKET_DIM_minus1)) == 0) {
	    if (i == 0) {
	      zgetvbucket(&lhszvbucketptr);
	      lhszvectorptr->zds.cs.vbucketptr = lhszvbucketptr;
	    }
	    else {
	      rhszvectorvbucketptr =
		rhszvectorvbucketptr->nextvbucket;
	      zgetvbucket(&lhszvbucketptr);
	      lhszvectortailvbucketptr->nextvbucket = lhszvbucketptr;
	    }
	    lhszvectortailvbucketptr = lhszvbucketptr;
	  }
	  GAXPY_EQUALS(lhszvectortailvbucketptr->value[vbucketindex],
		       *alphaptr,
		       rhszvectorvbucketptr->value[vbucketindex]);
	}
      } /* matches     else ** if (lhszvectorptr->dstype == CS) ** {  */
    
      zaxpy_0_1_post_processing(destptrptr);

    } /* matches   else ** if (not plusequalzspg1) ** {                   */
  } /* matches   else ** if not ZERO AXPY ** {                             */
  

#ifdef _PEEP
  if (toggle_adxpy_1_peep) {
    zadpeep_post(destptrptr);
  }
#endif


#ifdef _TRACE 
  nnza = (*destptrptr)->nonzeroes; 
  zadinfo(arity,nnzb,lhszvectoraddress,nnza,alphaptrs,vectorptrptrs);
#endif
  
}
  
