#define DOUBLE_COMPLEX
/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#include  <stdlib.h>
#include "zgaxpy_pd_macros_and_data_structures.h"
#include "zgaxpy_pd_global_variables.h"

/****************************************************************************
 *                                                                          *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/

void zspg2(destptrptr,alphaptr,vectoraptrptr,betaptr,vectorbptrptr)
zvectorstruct **destptrptr;
double_complex *alphaptr;
zvectorstruct **vectoraptrptr;
double_complex *betaptr;
zvectorstruct **vectorbptrptr;
{
  
  int arity = 2;
  
  zvectorstruct *rhszvectorptr[2];
  
  zbucketstruct *rhszbucketptr[2];
  
  zbucketstruct *lhszbucketptr;
  
  zibucketstruct *rhszibucketptr[2];
                                      /* Array of pointers to "current"     *
				       * buckets where "current" nonzeroes  *
				       * of the rhs_vectors reside.         */

  zibucketstruct *lhszibucketptr;  /* Pointer to current bucket of       *
				       * lhszvector.                         */

  double_complex alpha[2];          /* Dereferenced alphaptrs[].          */
  double_complex *alphaptrs[2];

  int rhszbucketindex[2];              /* Array of indices of current   *
				       * nonzeroes in the current buckets   *
				       * of rhs_vectors.                    */

  int lhszbucketindex;                 /* bucketindex of next nonzero to be  *
				       * added to lhszvector.                */

  int lastlhszbucketindex;             /* bucketindex of the latest added    *
				       * nonzero in the lhszvector.  We need *
				       * this additional index to leap-frog *
				       * buckets, if need be.               */


  int rhszibucketindex[2];             /* Array of indices of current        *
				       * nonzeroes in the current buckets   *
				       * of rhs_vectors.                    */

  int lhszibucketindex;                /* bucketindex of next nonzero to be  *
				       * added to lhszvector.                */

  int lastlhszibucketindex;            /* bucketindex of the latest added    *
				       * nonzero in the lhszvector.  We need *
				       * this additional index to leap-frog *
				       * buckets, if need be.               */

  int k;                              /* Counter for arity.                 */

  int j;                              /* Index for vectorsubscripts to be   *
				       * entered in (and later popped from) *
				       * the heap.                          */

  int rhszvectornonzeroes[2];          /* Array of total # of nonzeroes in   *
				       * each rhs_vector.                   */

  int rhszvectorconsumednonzeroes[2];  /* Array of # of nonzeroes consumed   *
				       * so far in each rhs_vector.         */

  int lhszvectornonzeroes;             /* Count of # of nonzeroes entered    *
				       * into the lhszvector so far.  By the *
				       * end of the saxpy operation, this   *
				       * index equals the total # of        *
				       * nonzeroes in the resultant vector. */

/*
 * These variables were originally for COMPRESSED_SUBSCRIPTS
 */

  zvbucketstruct *lastlhszvbucketptr; 
                                      /* Pointer to current vbucket of      *
				       * lhszvector.                    */

  int rhszvectornonzeroranges[2];
                                      /* Array of total # of nonzeroes      *
				       * ranges in each rhs_vector.         */

  int rhszvectorconsumednonzeroranges[2];
                                      /* Array of # of nonzeroes consumed   *
				       * so far in each rhs_vector.         */

  int rhszvbucketindex[2];             /* Array of indices of current        *
				       * nonzeroes in the current buckets   *
				       * of rhs_vectors.                    */

  int lhszvectornonzeroranges;         /* Count of # of nonzeroes entered    *
				       * into the lhszvector so far.  By the *
				       * end of the saxpy operation, this   *
				       * index equals the total # of        *
				       * nonzeroes in the resultant vector. */

  int lhszvbucketindex;                /* bucketindex of next nonzero to be  *
				       * added to lhszvector.                */


  int lastlhszvbucketindex;            /* bucketindex of the latest added    *
				       * nonzero in the lhszvector.  We need *
				       * this additional index to leap-frog *
				       * buckets, if need be.               */

  zvbucketstruct *templhszvbucketptr;
  int templhszvbucketindex;

  int j1;                             /* Index for vectorsubscripts to be   *
				       * entered in (and later popped from) *
				       * the heap.                          */

  int j2;                             /* Index for vectorsubscripts to be   *
				       * entered in (and later popped from) *
				       * the heap.                          */

  int h1;
  int h2;

  int totalcurrentnonzerorange;

#ifdef _DEBUG
  int unconsumedcurrentnonzerorange;
#endif

  int distancetooverlap;
  int distancetoendofvbucket;

  double_complex tmp_add,tmp_mult1,tmp_mult2;


  /*
   * These variables were added for union ds
   */
  
  zvectorstruct *restorerhszvectorptr[MAXARITY+1];

  int governingdstype;
  
  /*
   * These variables were added for plusequal code
   */
  
  zvectorstruct **rhszvectorptrptrs[MAXARITY];
  
  int condition[2];
  
#ifdef _PLUSEQUAL  
  int smallest_rhs_index; 
#endif /* _PLUSEQUAL */
  
  
#ifdef _TRACE
  int nnzb,nnza;
  zvectorstruct *lhszvectoraddress;
  
  zvectorstruct **vectorptrptrs[2];
  
  vectorptrptrs[0] = vectoraptrptr;
  
  vectorptrptrs[1] = vectorbptrptr;
  
  if(*destptrptr==NULL){ 
    nnzb = 0; 
    lhszvectoraddress= (zvectorstruct *) 99900; 
  }
  else {
    nnzb = (*destptrptr)->nonzeroes; 
    lhszvectoraddress = *destptrptr;
  }
#endif
  
  
  /****************************************
   * Initialization of *destptrptr.       *
   ****************************************/
  
  if (*destptrptr == NULL)
    zgetvector(destptrptr);
  
  /****************************************
   * Dereferencing.                       *
   ****************************************/
  
  alpha[0] = *alphaptr;
  rhszvectorptr[0] = *vectoraptrptr;
  alphaptrs[0] = alphaptr;
  
#ifdef _PEEP
  adpeep_header(2);
  zadpeep_pre(0,vectoraptrptr,alphaptr);
#endif    

  alpha[1] = *betaptr;
  rhszvectorptr[1] = *vectorbptrptr;
  alphaptrs[1] = betaptr;

#ifdef _PEEP
  zadpeep_pre(1,vectorbptrptr,betaptr);
#endif    
 
  for (k=0;k<arity;k++) {
    restorerhszvectorptr[k] = rhszvectorptr[k];
    rhszvectornonzeroes[k] = rhszvectorptr[k]->nonzeroes;
    if (rhszvectorptr[k]->dstype == CS)
      rhszvectornonzeroranges[k] =
	rhszvectorptr[k]->zds.cs.nonzeroranges;
  }
  
#ifdef _PLUSEQUAL
  smallest_rhs_index = DUMMYSUBSCRIPT;
  for (k=0;k<arity;k++) {
    if (rhszvectorptr[k]->dstype == S1) 
      smallest_rhs_index = min (smallest_rhs_index,
				rhszvectorptr[k]->zds.s1.index);
    else if (rhszvectorptr[k]->dstype == SS)
      smallest_rhs_index =
	min (smallest_rhs_index,
	     rhszvectorptr[k]->zds.ss.bucketptr->index[0]);
    else /* if (rhszvectorptr[k]->dstype == CS) */
      smallest_rhs_index =
	min (smallest_rhs_index,
	     rhszvectorptr[k]->zds.cs.ibucketptr->index_RANGE_BEGIN[0]);
  }
#endif /* _PLUSEQUAL */


  if ((rhszvectorptr[0]->nonzeroes == 0 || X_IS_ZERO(alpha[0])) &&
      (rhszvectorptr[1]->nonzeroes == 0 || X_IS_ZERO(alpha[1]))) {
    zspzronotca(destptrptr);
  }
  else if (rhszvectorptr[0]->nonzeroes == 0 || X_IS_ZERO(alpha[0])) {
    if (rhszvectorptr[1] == *destptrptr &&
	rhszvectorptr[1]->dstype == CS) {
      plusequalzspg1(destptrptr,betaptr);
    }
    /* if (rhszvectorptr[1] != *destptrptr ||
       rhszvectorptr[1]->dstype != CS)           */
    else {
#ifdef _PEEP
      toggle_adxpy_1_peep = 0;
      zspg1(destptrptr,betaptr,vectorbptrptr);
      toggle_adxpy_1_peep = 1;
#else
      zspg1(destptrptr,betaptr,vectorbptrptr);
#endif     
    }
  }
  else if (rhszvectorptr[1]->nonzeroes == 0 || X_IS_ZERO(alpha[1])) {
    if (rhszvectorptr[0] == *destptrptr &&
	rhszvectorptr[0]->dstype == CS) {
      plusequalzspg1(destptrptr,alphaptr);
    }
    /* else if (rhszvectorptr[0] != *destptrptr) ||
       rhszvectorptr[0]->dstype != CS)          */
    else {      
#ifdef _PEEP
      toggle_adxpy_1_peep = 0;
      zspg1(destptrptr,alphaptr,vectoraptrptr);
      toggle_adxpy_1_peep = 1;
#else
      zspg1(destptrptr,alphaptr,vectoraptrptr);
#endif
    }
  }
  else /* neither rhs_operation is a NO-OP */ {
    if (rhszvectorptr[0] == *destptrptr &&
	rhszvectorptr[1] == *destptrptr &&
	rhszvectorptr[0]->dstype == CS) {
#ifdef _DEBUG
      printf("zspg2 warning: lhszvector appears more than ");
      printf("once on rhs.\n");
#endif
      X_ADD(tmp_add,*alphaptr,*betaptr);
      plusequalzspg1(destptrptr,&tmp_add);
    }

#ifdef _PLUSEQUAL
    else if
      (rhszvectorptr[0] == *destptrptr &&
       rhszvectorptr[0]->dstype == CS &&
       rhszvectorptr[0]->zds.cs.ibucketptr->index_RANGE_BEGIN[0] <=
       smallest_rhs_index &&
       X_IS_ONE(alpha[0])) {
	rhszvectorptrptrs[1] = &rhszvectorptr[1];
	if ((*rhszvectorptrptrs[1])->dstype != CS &&
	    (*rhszvectorptrptrs[1])->nonzeroes !=0) {
	  zconvertdstype(rhszvectorptrptrs[1],1,CS);
	}
	plusequalzspg2(destptrptr,
			betaptr,rhszvectorptrptrs[1],ORIG_OTHER_VECTOR);
      }
    else if
      (rhszvectorptr[1] == *destptrptr &&
       rhszvectorptr[1]->dstype == CS &&
       rhszvectorptr[1]->zds.cs.ibucketptr->index_RANGE_BEGIN[0] <=
       smallest_rhs_index &&
       X_IS_ONE(alpha[1])) {
	rhszvectorptrptrs[0] = &rhszvectorptr[0];
	if ((*rhszvectorptrptrs[0])->dstype != CS &&
	    (*rhszvectorptrptrs[0])->nonzeroes !=0) {
	  zconvertdstype(rhszvectorptrptrs[0],0,CS);
	}
	plusequalzspg2(destptrptr,
			alphaptr,rhszvectorptrptrs[0],ORIG_OTHER_VECTOR);
	
      }
    
    /* 
     * else if neither rhs_operation is a NO-OP, and neither
     * rhszvector is (the same as the lhszvector AND has
     * multiplier equal to 1 AND is CS)
     */
    
    else if (rhszvectorptr[0] ==  rhszvectorptr[1]) {
#ifdef _DEBUG
      printf("zspg2 warning: lhszvector appears more than ");
      printf("once on rhs.\n");
#endif
      X_ADD(tmp_add,*alphaptr,*betaptr);
#ifdef _PEEP
      toggle_adxpy_1_peep = 0;
      zspg1(destptrptr,&tmp_add,vectoraptrptr);
      toggle_adxpy_1_peep = 1;
#else
      zspg1(destptrptr,&tmp_add,vectoraptrptr);
#endif
    }
#endif /* _PLUSEQUAL */
    
    else /* do an honest to goodness zspg2 */ {
      
      governingdstype = max(rhszvectorptr[0]->dstype,
			    rhszvectorptr[1]->dstype);
      if (governingdstype == S1 &&
	  rhszvectorptr[0]->zds.s1.index != rhszvectorptr[1]->zds.s1.index)
	governingdstype = SS;
      for (k=0;k<arity;k++)
	if (rhszvectorptr[k]->dstype != governingdstype) {
	  zconvertdstype((&rhszvectorptr[k]),k,governingdstype);
	  if (governingdstype == CS)
	    rhszvectornonzeroranges[k] =
	      rhszvectorptr[k]->zds.cs.nonzeroranges;
	}
      
      if (governingdstype == S1) {
	
	/******************************************************************
	 *                      ONE_SUBSCRIPT AXPY                        *
	 ******************************************************************/

	lhszvectorptr->nonzeroes =  1;
	lhszvectorptr->dstype = governingdstype;
	lhszvectorptr->zds.s1.index =
	  rhszvectorptr[0]->zds.s1.index;
        X_MULT(tmp_mult1,*alphaptr,rhszvectorptr[0]->zds.s1.value);
        X_MULT(tmp_mult2,*betaptr,rhszvectorptr[1]->zds.s1.value);
        X_ADD(lhszvectorptr->zds.s1.value,tmp_mult1,tmp_mult2);

	zaxpy_0_1_post_processing(destptrptr);

      }
      else /* if we are going to use lhszvectorptr */ {
	if (governingdstype == SS) {
	  
	  /****************************************************************
	   *                 _SINGLE_SUBSCRIPTS AXPY                      *
	   ****************************************************************/
	  for (k=0;k<arity;k++) {
	    rhszvectorconsumednonzeroes[k]= 0;
	    rhszbucketindex[k] = 0;
	    rhszbucketptr[k] = rhszvectorptr[k]->zds.ss.bucketptr;
	  }
	  zgetbucket(&lhszbucketptr);
	  lhszvectorptr->dstype = governingdstype;
	  lhszvectorptr->zds.ss.bucketptr = lhszbucketptr;
	  lhszvectornonzeroes = 0;
	  
	  for (k=0;k<arity;k++)
	    condition[k] = rhszvectorconsumednonzeroes[k] <
	      rhszvectorptr[k]->nonzeroes;
	  
	  while (condition[0] || condition[1]) {
	    if (condition[0] && condition[1])
	      if (rhszbucketptr[0]->index[rhszbucketindex[0]] <=
		  rhszbucketptr[1]->index[rhszbucketindex[1]])
		k = 0;
	      else
		k = 1;
	    else if (condition[0])
	      k = 0;
	    else
	      k = 1;
	    
	    j = rhszbucketptr[k]->index[rhszbucketindex[k]];
	    
	    lhszbucketindex = MY_MOD(lhszvectornonzeroes,
					 zBUCKET_DIM_minus1);
	    
	    /****************************************
	     * Perform saxpy:                       *
	     * 3.1. Found 1st or new subscript for  *
	     *      lhszvector.  If needed, get new  *
	     *      bucket for lhszvector.           *
	     *      Consume nonzero in saxpy.       *
	     ****************************************/
	    if (lhszvectornonzeroes == 0 ||
		lhszbucketptr->index[lastlhszbucketindex] < j) {
	      if (lhszbucketindex == 0 && lhszvectornonzeroes != 0) {
		zgetbucket(&(lhszbucketptr->nextbucket));
		lhszbucketptr = lhszbucketptr->nextbucket;
	      }
	      lhszbucketptr->index[lhszbucketindex] = j;
	      GAXPY_EQUALS(
                lhszbucketptr->value[lhszbucketindex],
		alpha[k],
		rhszbucketptr[k]->value[rhszbucketindex[k]]);
	      lastlhszbucketindex = lhszbucketindex;
	      lhszvectornonzeroes++;
	    }
	    
	    /****************************************
	     * Perform saxpy:                       *
	     * 3.2. Found old subscript for         *
	     *      lhszvector.                      *
	     *      Consume nonzero in saxpy.       *
	     ****************************************/
	    else if (lhszbucketptr->index[lastlhszbucketindex] == j) {
	      GAXPY_PLUS_EQUALS(
                lhszbucketptr->value[lastlhszbucketindex],
		alpha[k],
		rhszbucketptr[k]->value[rhszbucketindex[k]]);
	    }
	    
	    /****************************************
	     * Perform saxpy:                       *
	     * 3.3. Found wrong subscript for       *
	     *      lhszvector.  Abort.               *
	     ****************************************/
	    
#ifdef _DEBUG
	    else {
	      printf("zspg2 error 6: lhsznz=%d, lastlhszindex[]=%d, j=%d\n",
		     lhszvectornonzeroes,
		     lhszbucketptr->index[lastlhszbucketindex],j);
	      abort();
	    }
#endif
	    
	    /****************************************
	     * Perform saxpy:                       *
	     * 3.4. If there is one, get new        *
	     *      nonzero from the rhs_vector     *
	     *      whose nonzero was just          *
	     *      consumed and replace the        *
	     *      heaproot with new entry;        *
	     *      Else, delete heaproot.          *
	     ****************************************/
	    rhszvectorconsumednonzeroes[k]++;
	    condition[k] = rhszvectorconsumednonzeroes[k] <
	      rhszvectorptr[k]->nonzeroes;
	    rhszbucketindex[k] = MY_MOD(rhszvectorconsumednonzeroes[k],
					    zBUCKET_DIM_minus1);
	    if (condition[k]) {
	      if (rhszbucketindex[k]  == 0) {
		if ((rhszbucketptr[k] = rhszbucketptr[k]->nextbucket) 
		    == NULL) {
#ifdef _DEBUG
		  printf("zspg2 err: unexpected NULL nextbucket: k=%d",k);
		  printf("rhszvectorconsumednonzeros=%d,rhszvecnonzeroes=%d,",
			 rhszvectorconsumednonzeroes[k],
			 rhszvectornonzeroes[k]);
		  abort();
#endif
		}	  
	      }
	    }
	  } /* matches     while loop     */
	} /* matches     else if (governingdstype == SS) {      */
	
	else /* if (governingdstype == CS) */ {
	  
	  /****************************************************************
	   *               _COMPRESSED_SUBSCRIPTS AXPY                    *
	   ****************************************************************/
	  
	  /****************************************
	   * Else perform saxpy:                  *
	   * 1. Put in heap the 1st nonzero of    *
	   *    each vector.                      *
	   ****************************************/
	  for (k=0;k<arity;k++) {
	    rhszvectorconsumednonzeroranges[k] = 0;
	    rhszibucketindex[k] = 0;
	    rhszvbucketindex[k] = 0;
	    rhszibucketptr[k] =
	      rhszvectorptr[k]->zds.cs.ibucketptr;
	    rhszvbucketptr[k] =
	      rhszvectorptr[k]->zds.cs.vbucketptr;
	    
#ifdef _DEBUG
	    if (rhszibucketptr[k] == NULL ||
		rhszvbucketptr[k] == NULL) {
	      printf("zspg2 error 10: Unexpected NULL rhszbucketptrs: ");
	      printf("vector # %d\n",k);
	      abort();
	    }
#endif
	  }	  
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 2. Initialize lhszvector.             *
	   ****************************************/
	  zgetibucket(&lhszibucketptr);
	  lhszvectorptr->dstype = governingdstype;
	  lhszvectorptr->zds.cs.ibucketptr = lhszibucketptr;
	  zgetvbucket(&lhszvbucketptr);
	  lhszvectorptr->zds.cs.vbucketptr = lhszvbucketptr;
	  lhszvectornonzeroranges = 0;
	  lhszvectornonzeroes = 0;
	  lhszvbucketindex = 0;
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3. Get nonzero by popping the        *
	   *    heaproot; consume it in saxpy;    *
	   *    either replace or delete heaproot.*
	   ****************************************/
	  
	  while (rhszvectorconsumednonzeroranges[0] <
		 rhszvectorptr[0]->zds.cs.nonzeroranges ||
		 rhszvectorconsumednonzeroranges[1] <
		 rhszvectorptr[1]->zds.cs.nonzeroranges) {
	    if
	      ((rhszvectorconsumednonzeroranges[1] ==
		rhszvectorptr[1]->zds.cs.nonzeroranges) ||
	       ((rhszvectorconsumednonzeroranges[0] <
		 rhszvectorptr[0]->zds.cs.nonzeroranges) &&
		(rhszibucketptr[0]->index_RANGE_BEGIN[rhszibucketindex[0]] <=
		 rhszibucketptr[1]->index_RANGE_BEGIN[rhszibucketindex[1]])))
		k = 0;
	    else if
	      ((rhszvectorconsumednonzeroranges[0] ==
		rhszvectorptr[0]->zds.cs.nonzeroranges) ||
	       ((rhszvectorconsumednonzeroranges[1] <
		 rhszvectorptr[1]->zds.cs.nonzeroranges) &&
		(rhszibucketptr[0]->index_RANGE_BEGIN[rhszibucketindex[0]] >
		 rhszibucketptr[1]->index_RANGE_BEGIN[rhszibucketindex[1]])))
		k = 1;
#ifdef _DEBUG
	    else {
	      printf("zspg2 error: BAD LOGIC.\n");
	      abort();
	    }
#endif
	    j1 = rhszibucketptr[k]->index_RANGE_BEGIN[rhszibucketindex[k]];
	    j2 = rhszibucketptr[k]->index_RANGE_END[rhszibucketindex[k]];
	    rhszibucketindex[k] = 
	      MY_MOD(rhszvectorconsumednonzeroranges[k],
		     zIBUCKET_DIM_minus1);
	    lhszibucketindex = MY_MOD(lhszvectornonzeroranges,
					  zIBUCKET_DIM_minus1);
	    
	    if ((rhszvbucketindex[k] = MY_MOD(rhszvbucketindex[k],
						  zVBUCKET_DIM_minus1))==0
		&& rhszvectorconsumednonzeroranges[k] !=0) {
	      zHOP_TO_NEXT_RHSVBUCKET(k,&rhszvbucketindex[k],
				     rhszvectorconsumednonzeroranges[k],
				     rhszvectornonzeroranges[k]);
	    }
	    lhszvbucketindex = MY_MOD(lhszvbucketindex,
					  zVBUCKET_DIM_minus1);
	    
#ifdef _DEBUG
	    if (rhszibucketptr[k]->errorflag != 9099 ||
		rhszvbucketptr[k]->errorflag != 9009) {
	      printf("zspg2 err 14: memory corruption of zibucketstore ");
	      printf("or zvbucketstore.\n");
	      abort();
	    }
#endif      
	    
	    /****************************************
	     * Perform saxpy:                       *
	     * 3.0. Check for startup errors.       *
	     ****************************************/
	    
#ifdef _DEBUG
	    if (lhszvectornonzeroranges == 0) {
	      if (rhszibucketindex[k] != 0 || rhszvbucketindex[k] != 0 ||
		  lhszibucketindex != 0 || lhszvbucketindex != 0) {
		printf("zspg2 error 15: Zeroes expected: \n\t");
		printf("rhszibucketindex[k]=%d, rhszvbucketindex[k]=%d\n\t",
		       rhszibucketindex[k], rhszvbucketindex[k]);
		printf("lhszibucketindex=%d, lhszvbucketindex=%d\n\t",
		       lhszibucketindex,lhszvbucketindex);
		abort();
	      }
	    }
#endif      
	    
	    /****************************************
	     * Perform saxpy:                       *
	     * 3.1. Found 1ST or NEW subscript      *
	     *      range (non-overlapping with     *
	     *      previous one, if any) for       *
	     *      lhszvector.  If needed, get new  *
	     *      ibuckets and vbuckets for       *
	     *      lhszvector.                      *
	     ****************************************/
	    if (lhszvectornonzeroranges == 0 ||
		j1 > lhszibucketptr->index_RANGE_END[lastlhszibucketindex]+1) {/*A.1*/
	      
	      if (lhszibucketindex == 0 && lhszvectornonzeroranges != 0) {
		zgetibucket(&(lhszibucketptr->nextibucket));
		lhszibucketptr = lhszibucketptr->nextibucket;
	      }
	      lhszibucketptr->index_RANGE_BEGIN[lhszibucketindex] = j1;
	      lhszibucketptr->index_RANGE_END[lhszibucketindex] = j2;
	      
	      if (lhszvbucketindex == 0 && lhszvectornonzeroranges != 0)
		zget_and_hop_to_next_vbucket(&lhszvbucketindex);
	      
	      lastlhszvbucketptr = lhszvbucketptr;
	      lastlhszvbucketindex = lhszvbucketindex;
	      
	      totalcurrentnonzerorange = j2-j1+1;
	      
	      lhszvectornonzeroes += totalcurrentnonzerorange;
	      lhszvectornonzeroranges++;
	      
	
	      zNON_OVER_LAPPING_CS_RANGE_AXPY(COMPLETELY_NEW_RANGE,
					     lhszibucketindex,
					     &lastlhszibucketindex,
					     &lhszvbucketindex,
					     k,
					     &rhszvbucketindex[k],
					     alpha[k],
					     totalcurrentnonzerorange,
					     &unconsumedcurrentnonzerorange,
					     rhszvectorconsumednonzeroranges[k],
					     rhszvectornonzeroranges[k],
					     lhszvectornonzeroranges);
	      
	            
	    } /*A.1*/
	    
	    
	    /****************************************
	     * Perform saxpy:                       *
	     * 3.2. Found old subscript for         *
	     *      lhszvector.  Consume  nonzero in *
	     *      saxpy.                          *
	     *                                      *
	     ****************************************/
	    else if (j1 <= lhszibucketptr->index_RANGE_END[lastlhszibucketindex]+1 &&
		     j1 >= lhszibucketptr->index_RANGE_BEGIN[lastlhszibucketindex]
		     ) { /*B.1*/
	      
	      templhszvbucketptr = lhszvbucketptr;
	      templhszvbucketindex = lhszvbucketindex;
	      lhszvbucketptr = lastlhszvbucketptr;
	      
	      distancetooverlap =
		j1 - lhszibucketptr->index_RANGE_BEGIN[lastlhszibucketindex];
	      distancetoendofvbucket = zVBUCKET_DIM -
		lastlhszvbucketindex;
	      
	      if (j1 ==
		  lhszibucketptr->index_RANGE_END[lastlhszibucketindex]+1) {
		h1 = 0;
		h2 = -1;
	      }
	      else if (j2 <= lhszibucketptr->index_RANGE_END[lastlhszibucketindex]
		       ) {
		h1 = j1;
		h2 = j2;
		j1 = 0;
		j2 = -1;
	      }
	      else {
		h1 = j1;
		h2 = lhszibucketptr->index_RANGE_END[lastlhszibucketindex];
		j1 = h2+1;
	      }
	      
	      if (distancetoendofvbucket > distancetooverlap) {
		lhszvbucketindex =
		  MY_MOD(lastlhszvbucketindex+distancetooverlap,
			 zVBUCKET_DIM_minus1);
	      }
	      else if (distancetoendofvbucket < distancetooverlap) {
		lhszvbucketindex =
		  MY_MOD(distancetooverlap - distancetoendofvbucket,
			 zVBUCKET_DIM_minus1);
	      }
	      else {
		lhszvbucketindex = 0;
	      }
	      
	      while (distancetooverlap >= distancetoendofvbucket &&
		     lhszvbucketptr->nextvbucket != NULL) {
		lhszvbucketptr = lhszvbucketptr->nextvbucket;
		distancetooverlap -= distancetoendofvbucket;
		distancetoendofvbucket = zVBUCKET_DIM;
	      }
	      
	      lhszibucketptr->index_RANGE_END[lastlhszibucketindex] =
		max(j2,
		    lhszibucketptr->index_RANGE_END[lastlhszibucketindex]);
	      
	      if ((totalcurrentnonzerorange = h2-h1+1) > 0) {
		

		zOVERLAPPING_CS_RANGE_AXPY(&lhszvbucketindex,
					  k,
					  &rhszvbucketindex[k],
					  alpha[k],
					  totalcurrentnonzerorange,
					  &unconsumedcurrentnonzerorange,
					  rhszvectorconsumednonzeroranges[k],
					  rhszvectornonzeroranges[k],
					  lhszvectornonzeroranges);
		
	      }
	     
	      /*
	       * This bit of logic must be performed, whether or not there
	       * is additional nonzeroes to process.
	       */
	      
	      if (j2 != -1 &&
		  (lhszvbucketptr != templhszvbucketptr ||
		   (lhszvbucketindex = MY_MOD(lhszvbucketindex,
						  zVBUCKET_DIM_minus1)) !=
		   templhszvbucketindex)) {
		
#ifdef _DEBUG
		printf("zspg2 err 50.0: Unexpectd value discrepancies\n");
		printf("lhszvbucketptr=%d, templhszvbucketptr=%d\n",
		       (int)lhszvbucketptr,(int)templhszvbucketptr);
		printf("lhszvbucketindex=%d, templhszvbucketindex=%d\n",
		       lhszvbucketindex,templhszvbucketindex);
		
		abort();
#endif
	      }
	      else {
		lhszvbucketptr = templhszvbucketptr;
		lhszvbucketindex = templhszvbucketindex;
	      }
	      
	      
	      if ((totalcurrentnonzerorange = j2-j1+1) > 0) {
		lhszvectornonzeroes += totalcurrentnonzerorange;
		
		/*
		 * adding new entries to vbucket
		 */
		
		if (lhszvbucketindex == 0 &&
		    lhszvectornonzeroranges != 0 &&
		    j2 != -1)
		  zget_and_hop_to_next_vbucket(&lhszvbucketindex);
		
		if(rhszvbucketindex[k] == zVBUCKET_DIM && j2 != -1) {

		  zHOP_TO_NEXT_RHSVBUCKET(k,&rhszvbucketindex[k],
					 rhszvectorconsumednonzeroranges[k],
					 rhszvectornonzeroranges[k]);
		}
		
		
		
		zNON_OVER_LAPPING_CS_RANGE_AXPY(NEW_ENTRIES_PADDED_TO_OLD_RANGE,
					       lhszibucketindex,
					       &lastlhszibucketindex,
					       &lhszvbucketindex,
					       k,
					       &rhszvbucketindex[k],
					       alpha[k],
					       totalcurrentnonzerorange,
					       &unconsumedcurrentnonzerorange,
					       rhszvectorconsumednonzeroranges[k],
					       rhszvectornonzeroranges[k],
					       lhszvectornonzeroranges);
		
	      } /* matches     if ((totalcurrent...) { */
	    } /*B.1*/
	    
	    
	    /****************************************	
	     * Perform saxpy:                       *
	     * 3.3. Found wrong subscript for       *
	     *      lhszvector.  Abort.               *
	     ****************************************/
	    
#ifdef _DEBUG
	    else {
	      printf("zspg2 error 60: lhszvectornonzeroranges=%d,",
		     lhszvectornonzeroranges);
	      printf("lastlhszibucketindex=%d, \n\t",lastlhszibucketindex);
	      printf("lhszibucketptr->index[%d] = (%d,%d), j1=%d\n",
		     lastlhszibucketindex,
		     lhszibucketptr->index_RANGE_BEGIN[lastlhszibucketindex],
		     lhszibucketptr->index_RANGE_END[lastlhszibucketindex],
		     j1);
	      abort();
	    }
#endif
	    
	    
	    /****************************************
	     * Perform saxpy:                       *
	     * 3.4. If there is one, get new        *
	     *      nonzero from the rhs_vector     *
	     *      whose nonzero was just          *
	     *      consumed and replace the        *
	     *      heaproot with new entry;        *
	     *      Else, delete heaproot.          *
	     ****************************************/
	    if (++rhszvectorconsumednonzeroranges[k] <
		rhszvectornonzeroranges[k]) {
	      if ((rhszibucketindex[k] = 
		   MY_MOD(rhszvectorconsumednonzeroranges[k],
			  zIBUCKET_DIM_minus1)) == 0) {
		if ((rhszibucketptr[k] =
		     rhszibucketptr[k]->nextibucket) == NULL) {
		  
#ifdef _DEBUG
		  printf("zspg2 err 61: unexpected NULL nextvbucket: \n");
		  printf("\t k=%d, rhszvectorconsumednonzeroranges=%d,\n",
			 k,rhszvectorconsumednonzeroranges[k]);
		  printf("\t rhszvectornonzeroranges[k]=%d\n",
			 rhszvectornonzeroranges[k]);
		  abort();
#endif
		  
		}
	      }
	    }   
	  } /* matches     while loop          */
	} /* matches     else ** if (governingdstype == CS) ** {     */
	

	
	
	
	rhszvectorptrptrs[0] = &rhszvectorptr[0];
	rhszvectorptrptrs[1] = &rhszvectorptr[1];
	
	zaxpy_2_heap_post_processing(destptrptr,
				    rhszvectorptrptrs,
				    alphaptrs,
				    restorerhszvectorptr,
				    arity,
				    lhszvectornonzeroes,
				    lhszvectornonzeroranges,
				    governingdstype);
	
      } /* matches     if we are going to use lhszvectorptr           */
    } /* matches     else ** do an honest to goodness zspg2 **  {        */
  } /* matches     else ** neither rhs_operation is a NO-OP  ** {         */
    
  
#ifdef _PEEP
  zadpeep_post(destptrptr);
#endif
  
#ifdef _TRACE 
  nnza = (*destptrptr)->nonzeroes; 
  zadinfo(arity,nnzb,lhszvectoraddress,nnza,alphaptrs,vectorptrptrs);
#endif
  
}

