#define DOUBLE_COMPLEX
/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#include  <stdlib.h>
#include "zgaxpy_pd_macros_and_data_structures.h"
#include "zgaxpy_pd_global_variables.h"

/****************************************************************************
 *                                                                          *
 *  zspgx(destptrptr,arity,alphaptrs,rhszvectorptrptrs)             *
 *                                                                          *
 *     destptrptr      is a doubly-indirected pointer to a zvectorstruct*
 *                     It is passed to zspgx by the adxpyN routines,    *
 *                     where 1 <= N <= MAXARITY.  In turn destptrptr is an  *
 *                     argument of the adxpyN functions which are called by *
 *                     ADIFOR-generated FORTRAN calls.  The purpose of the  *
 *                     double indirection is that FORTRAN calls by          *
 *                     reference, that is it puts the address of say g$x(1) *
 *                     on the stack.  In the FORTRAN code, the value of     *
 *                     g$x(1) represents a pointer to (address of) a sparse *
 *                     vector (i.e., zvectorstruct *).                  *
 *                     The FORTRAN call,                                    *
 *                     CALL adxpyN(g$x(1),...), causes inside C for an      *
 *                     internal copy, say destptrptr, to equal the address  *
 *                     of g$x(1), i.e., destptrptr == &g$x(1).  Therefore,  *
 *                     *destptrptr == g$x(1). But since the value contained *
 *                     in g$x(1) is (as interpreted by C) of type           *
 *                     (zvectorstruct *), destptrptr needs              *
 *                     one more level of                                    *
 *                     indirection, and is declared as                      *
 *                     (zvectorstruct **).                              *
 *                                                                          *
 *     arity           is the # of vectors on the rhs.                      *
 *                                                                          *
 *     alphaptrs       Array of pointers to the scalar multipliers.         *
 *                                                                          *
 *     rhszvectorptrptrs   Array of doubly-indirected pointers to       *
 *                     zvectorstructs.  In this case, since these values*
 *                     are not changed, there is no need for the double-    *
 *                     indirection; we simply want to maintain the g$x      *
 *                     references consistant.                               *
 *                                                                          *
 *     zspgx performs the saxpy operation for all aritys.               *
 *                                                                          *
 ****************************************************************************/

void zspgx(destptrptr,arity,alphaptrs,rhszvectorptrptrs)
zvectorstruct **destptrptr;
int arity;
double_complex *alphaptrs[];
zvectorstruct **rhszvectorptrptrs[];
{

  zvectorstruct *rhszvectorptr[MAXARITY];
                                 /* Dereferenced rhszvectorptrptrs[].   */

  zbucketstruct *rhszbucketptr[MAXARITY];
                                      /* Array of pointers to "current"     *
				       * buckets where "current" nonzeroes  *
				       * of the rhs_vectors reside.         */

  zbucketstruct *lhszbucketptr;
                                      /* Pointer to current bucket of       *
				       * lhszvector.                    */

  zibucketstruct *rhszibucketptr[MAXARITY];
                                      /* Array of pointers to "current"     *
				       * buckets where "current" nonzeroes  *
				       * of the rhs_vectors reside.         */

  zibucketstruct *lhszibucketptr;
                                      /* Pointer to current bucket of       *
				       * lhszvector.                    */

  double_complex alpha[MAXARITY];   /* Dereferenced alphaptrs[].          */

  int rhszbucketindex[MAXARITY];  /* Array of indices of current        *
				       * nonzeroes in the current buckets   *
				       * of rhs_vectors.                    */

  int lhszbucketindex;            /* bucketindex of next nonzero to be  *
				       * added to lhszvector.           */

  int lastlhszbucketindex;        /* bucketindex of the latest added    *
				       * nonzero in the lhszvector.  We *
				       * need this additional index to      *
				       * leap-frog buckets, if need be.     */

  int rhszibucketindex[MAXARITY]; /* Array of indices of current        *
				       * nonzeroes in the current buckets   *
				       * of rhs_vectors.                    */

  int lhszibucketindex;           /* bucketindex of next nonzero to be  *
				       * added to lhszvector.           */

  int lastlhszibucketindex;       /* bucketindex of the latest added    *
				       * nonzero in the lhszvector.  We *
				       * need this additional index to      *
				       * leap-frog buckets, if need be.     */

  int firstsaxpyoperand;              /* Order in the ennumeration of the   *
				       * rhs_vectors of the first non-null  *
				       * saxpy vectorop.                    */
  int k;                              /* Counter for arity.                 */

  int j;                              /* Index for vectorsubscripts to be   *
				       * entered in (and later popped from) *
				       * the heap.                          */

  int rhszvectornonzeroes[MAXARITY];
                                      /* Array of total # of nonzeroes in   *
				       * each rhs_vector.                   */

  int rhszvectorconsumednonzeroes[MAXARITY];
                                      /* Array of # of nonzeroes consumed   *
				       * so far in each rhs_vector.         */

  int lhszvectornonzeroes;        /* Count of # of nonzeroes entered    *
				       * into the lhszvector so far. By *
				       * the end of the saxpy operation     *
				       * this index equals the total # of   *
				       * nonzeroes in the resultant vector. */

/*
 * These variables were originally for COMPRESSED_SUBSCRIPTS
 */

  zvbucketstruct *lastlhszvbucketptr;
                                      /* Pointer to current vbucket of      *
				       * lhszvector.                    */

  int rhszvectornonzeroranges[MAXARITY];
                                      /* Array of total # of nonzeroes      *
				       * ranges in each rhs_vector.         */

  int rhszvectorconsumednonzeroranges[MAXARITY];
                                      /* Array of # of nonzeroes consumed   *
				       * so far in each rhs_vector.         */

  int rhszvbucketindex[MAXARITY]; /* Array of indices of current        *
				       * nonzeroes in the current buckets   *
				       * of rhs_vectors.                    */

  int lhszvectornonzeroranges;    /* Count of # of nonzeroes entered    *
				       * into the lhszvector so far. By *
				       * the end of the saxpy operation     *
				       * this index equals the total # of   *
				       * nonzeroes in the resultant vector. */

  int lhszvbucketindex;           /* bucketindex of next nonzero to be  *
				       * added to lhszvector.           */


  int lastlhszvbucketindex;       /* bucketindex of the latest added    *
				       * nonzero in the lhszvector.  We *
				       * need this additional index to      *
				       * leap-frog buckets, if need be.     */

  zvbucketstruct *templhszvbucketptr;
  int templhszvbucketindex;

  int j1;                             /* Index for vectorsubscripts to be   *
				       * entered in (and later popped from) *
				       * the heap.                          */

  int j2;                             /* Index for vectorsubscripts to be   *
				       * entered in (and later popped from) *
				       * the heap.                          */

  int h1;
  int h2;

  int totalcurrentnonzerorange;

#ifdef _DEBUG
  int unconsumedcurrentnonzerorange;
#endif

  int distancetooverlap;
  int distancetoendofvbucket;


  /*
   * These variables were added for union ds
   */
  
  zvectorstruct *restorerhszvectorptr[MAXARITY+1];

  int governingdstype;
  int rhs_largest_dstype = S1;        /* Largest dstype of any              *
				       * rhszvector ( CS > SS > S1 ).   */
  int s1_index;
  
  double_complex temp_gaxpy;

  /*
   * These variables were added for plusequal code
   */

#ifdef _PLUSEQUAL
  int plusequalsaxpy_flag = 0;
  int plusequalkorder;
  zvectorstruct **tempdestptrptr;
  double_complex *tempalphaptrs[MAXARITY-1];
  zvectorstruct **temprhszvectorptrptrs[MAXARITY-1];
  double_complex one; 
  int smallest_rhs_index;
#endif /* _PLUSEQUAL */

  /****************************************
   * Initialization of *destptrptr.       *
   ****************************************/
  
  if (*destptrptr == NULL)
    zgetvector(destptrptr);
  
  
  /****************************************
   * Dereferencing.                       *
   ****************************************/
  
  for (k=0;k<arity;k++) {
    alpha[k] = *alphaptrs[k];

    rhszvectorptr[k] = *rhszvectorptrptrs[k];
    restorerhszvectorptr[k] = rhszvectorptr[k];
    rhszvectornonzeroes[k] = rhszvectorptr[k]->nonzeroes;
    if (rhszvectorptr[k]->dstype == CS)
      rhszvectornonzeroranges[k] =
	rhszvectorptr[k]->zds.cs.nonzeroranges;
    if (rhszvectorptr[k]->dstype > rhs_largest_dstype)
      rhs_largest_dstype = rhszvectorptr[k]->dstype;

#ifdef _PLUSEQUAL    
    if (rhszvectorptr[k] == *destptrptr &&
	rhszvectorptr[k]->dstype == CS &&
	X_IS_ONE(alpha[k])) {
      plusequalsaxpy_flag++;
      plusequalkorder = k;
    }
#endif /* _PLUSEQUAL */

  }	
  
  /****************************************
   * Determination of 1st nonzero         *
   * rhszvector operand.              *
   ****************************************/
  
  firstsaxpyoperand = 0;

/*
 *   This bit of the code is obsolete, since the routines calling
 *   this routine already have stripped the gaxpy of zero vectors
 *   and zero multipliers.
 *

  while (nullsaxpyflag && firstsaxpyoperand<arity)
    
    if (rhszvectornonzeroes[firstsaxpyoperand] == 0 ||
	X_IS_ZERO(alpha[firstsaxpyoperand]))
      firstsaxpyoperand++;
    else
      nullsaxpyflag = 0;
  
 *
 * If no nonzero vectorops, return a null vector.
 *

  if (nullsaxpyflag) {
    
    (*destptrptr)->nonzeroes = 0;
    (*destptrptr)->dstype = S1;

  }
  else {

*/
    
    /****************************************
     * Predetermine governingdstype and     *
     * appropriate saxpy routine.           *
     ****************************************/

  if (rhs_largest_dstype == S1) {
    s1_index = rhszvectorptr[firstsaxpyoperand]->zds.s1.index;
    for (k=firstsaxpyoperand+1;k<arity;k++)
      if (rhszvectorptr[k]->zds.s1.index !=
	  rhszvectorptr[firstsaxpyoperand]->zds.s1.index &&
	  ! X_IS_ZERO(alpha[k]))
	s1_index = -1;
    if (s1_index > 0)
      governingdstype = S1;
    else
      governingdstype = SS;
  }
  else if (rhs_largest_dstype <= SS)
    governingdstype = SS;
  else /* if (rhs_largest_dstype == CS) */
    governingdstype = CS;

#ifdef _PLUSEQUAL
  smallest_rhs_index = DUMMYSUBSCRIPT;
  for (k=0;k<arity;k++) {
    if (rhszvectorptr[k]->dstype == S1) 
      smallest_rhs_index = min (smallest_rhs_index,
				rhszvectorptr[k]->zds.s1.index);
    else if (rhszvectorptr[k]->dstype == SS)
      smallest_rhs_index = min (smallest_rhs_index,
        rhszvectorptr[k]->zds.ss.bucketptr->index[0]);
    else /* if (rhszvectorptr[k]->dstype == CS) */
      smallest_rhs_index = min (smallest_rhs_index,
	rhszvectorptr[k]->zds.cs.ibucketptr->index_RANGE_BEGIN[0]);
  }

  if (plusequalsaxpy_flag &&
      (*destptrptr)->dstype == CS &&
      (*destptrptr)->zds.cs.ibucketptr->index_RANGE_BEGIN[0] <=
      smallest_rhs_index) {    
#ifdef _DEBUG
    if (plusequalsaxpy_flag > 1) {
      printf("zspgx warning: lhszvector appears more than ");
      printf("once on rhs.\n");
    }
#endif
    if (arity == 1) {
      plusequalzspg1(destptrptr,alphaptrs[plusequalkorder]);
    }
    else {
      for (k=0;k<plusequalkorder;k++) {
	tempalphaptrs[k] = alphaptrs[k];
	temprhszvectorptrptrs[k] = rhszvectorptrptrs[k];
      }
      for (k=plusequalkorder+1;k<arity;k++) {
	tempalphaptrs[k-1] = alphaptrs[k];
	temprhszvectorptrptrs[k-1] = rhszvectorptrptrs[k];
      }
      tempdestptrptr = &tempzdestptr;
      zspgx(tempdestptrptr,arity-1,
		tempalphaptrs,temprhszvectorptrptrs);
      restorerhszvectorptr[MAXARITY] = tempzdestptr;
      if ((*tempdestptrptr)->dstype != governingdstype &&
	  (*tempdestptrptr)->nonzeroes !=0) {
	zconvertdstype(tempdestptrptr,MAXARITY,governingdstype);
      }
      X_PUT_REAL(one,1.0);
      plusequalzspg2(destptrptr,
		      &one,tempdestptrptr,TEMP_OTHER_VECTOR);
      tempzdestptr = restorerhszvectorptr[MAXARITY];
    }    
  }
  else /* if (not plusequalsaxpy_flag) */ {
#endif /* _PLUSEQUAL */

    for (k=0;k<arity;k++) {
      if (rhszvectorptr[k]->dstype != governingdstype &&
	  rhszvectorptr[k]->nonzeroes !=0 && ! X_IS_ZERO(alpha[k])) {
	zconvertdstype((&rhszvectorptr[k]),k,governingdstype);
	if (governingdstype == CS)
	  rhszvectornonzeroranges[k] = 
	    rhszvectorptr[k]->zds.cs.nonzeroranges;
      }
    }
    
    if (governingdstype == S1) {
      
      /********************************************************************
       *                       ONE_SUBSCRIPT AXPY                         *
       ********************************************************************/

      lhszvectorptr->nonzeroes = 1;
      lhszvectorptr->dstype = S1;
      lhszvectorptr->zds.s1.index =
	rhszvectorptr[firstsaxpyoperand]->zds.s1.index;
      X_ZERO(temp_gaxpy);
      for (k=firstsaxpyoperand;k<arity;k++) {
	GAXPY_PLUS_EQUALS(temp_gaxpy,
			  alpha[k], rhszvectorptr[k]->zds.s1.value);
      }
      X_ASSIGN(lhszvectorptr->zds.s1.value,temp_gaxpy);

      zaxpy_0_1_post_processing(destptrptr);

    }
    else /* if we are going to use lhszvectorptr */ {
      if (governingdstype == SS) {
	
	/******************************************************************
	 *                  _SINGLE_SUBSCRIPTS AXPY                       *
	 ******************************************************************/
	
	/****************************************
	 * Else perform saxpy:                  *
	 * 1. Put in heap the 1st nonzero of    *
	 *      each vector.                    *
	 ****************************************/
	rhszvectorconsumednonzeroes[firstsaxpyoperand] = 0;
	rhszbucketindex[firstsaxpyoperand] = 0;
	rhszbucketptr[firstsaxpyoperand] =
	  rhszvectorptr[firstsaxpyoperand]->zds.ss.bucketptr;
	
#ifdef _DEBUG
	if (rhszbucketptr[firstsaxpyoperand] == NULL) {
	  printf("zspgx error 1: Unexpected NULL rhszbucketptr: V# %d\n",
		 firstsaxpyoperand);
	  abort();
	}
#endif
	
	j = rhszbucketptr[firstsaxpyoperand]->
	  index[rhszbucketindex[firstsaxpyoperand]];
	
	if (j <= 0) {
#ifdef _DEBUG
	  printf("zspgx error 2: Illegal index: V # %d, subscript=%d\n",
		 firstsaxpyoperand,j);
	  abort();
#endif
	}
	
	else
	  heapinitialize(firstsaxpyoperand,j,j);
	
	for (k=firstsaxpyoperand+1;k<arity;k++) {
	  if (rhszvectornonzeroes[k] !=0 && ! X_IS_ZERO(alpha[k])) {
	    rhszvectorconsumednonzeroes[k] = 0;
	    rhszbucketindex[k] = 0;
	    rhszbucketptr[k] = rhszvectorptr[k]->zds.ss.bucketptr;
	    
#ifdef _DEBUG
	    if (rhszbucketptr[k] == NULL) {
	      printf("zspgx error 3: Unexpected NULL rhszbucketptr: ");
	      printf("vector # %d\n",k);
	      abort();
	    }
#endif
	    
	    j = rhszbucketptr[k]->index[rhszbucketindex[k]];
	    
#ifdef _DEBUG
	    if (j <= 0) {
	      printf("zspgx error 4: Illegal index: vec.=%d sub.=%d\n",
		     k,j);
	      abort();
	    }
#endif
	    
	    heapinsert(k,j,j);
	  }
	}
	
	/****************************************
	 * Perform saxpy:                       *
	 * 2. Initialize lhszvector.             *
	 ****************************************/
	zgetbucket(&lhszbucketptr);
	lhszvectorptr->dstype = governingdstype;
	lhszvectorptr->zds.ss.bucketptr = lhszbucketptr;
	lhszvectornonzeroes = 0;
	
	/****************************************
	 * Perform saxpy:                       *
	 * 3. Get nonzero by popping the        *
	 *    heaproot; consume it in saxpy;    *
	 *    replace / delete heaproot.        *
	 ****************************************/
	while (heaproot !=NULL) {
	  k = heaproot->vectorindex;
	  j = heaproot->rangebeginindex;
	  rhszbucketindex[k] = MY_MOD(rhszvectorconsumednonzeroes[k],
					  zBUCKET_DIM_minus1);
	  lhszbucketindex = MY_MOD(lhszvectornonzeroes,
				       zBUCKET_DIM_minus1);
	  
#ifdef _DEBUG
	  if (rhszbucketptr[k]->errorflag != 9999) {
	    printf("zspgx err 5: mem. corruption of zbucketstore\n");
	    abort();
	  }
#endif      
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.1. Found 1st or new subscript for  *
	   *      lhszvector.  If needed, get new  *
	   *      bucket for lhszvector.           *
	   *      Consume nonzero in saxpy.       *
	   ****************************************/
	  if (lhszvectornonzeroes == 0 ||
	      lhszbucketptr->index[lastlhszbucketindex] < j) {
	    if (lhszbucketindex == 0 && lhszvectornonzeroes != 0) {
	      zgetbucket(&(lhszbucketptr->nextbucket));
	      lhszbucketptr = lhszbucketptr->nextbucket;
	    }
	    lhszbucketptr->index[lhszbucketindex] = j;
	    GAXPY_EQUALS(lhszbucketptr->value[lhszbucketindex],
			 alpha[k],
			 rhszbucketptr[k]->value[rhszbucketindex[k]]);
	    lastlhszbucketindex = lhszbucketindex;
	    lhszvectornonzeroes++;
	  }
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.2. Found old subscript for         *
	   *      lhszvector.                      *
	   *      Consume nonzero in saxpy.       *
	   ****************************************/
	  else if (lhszbucketptr->index[lastlhszbucketindex] == j) {
	    GAXPY_PLUS_EQUALS(
	      lhszbucketptr->value[lastlhszbucketindex],
	      alpha[k],
	      rhszbucketptr[k]->value[rhszbucketindex[k]]);
	  }
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.3. Found wrong subscript for       *
	   *      lhszvector.  Abort.               *
	   ****************************************/
	  
#ifdef _DEBUG
	  else {
	    printf("zspgx error 6: lhsznz=%d, lastlhszindex[]=%d, j=%d\n",
		   lhszvectornonzeroes,
		   lhszbucketptr->index[lastlhszbucketindex],j);
	    abort();
	  }
#endif
	  
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.4. If there is one, get new        *
	   *      nonzero from the rhs_vector     *
	   *      whose nonzero was just          *
	   *      consumed and replace the        *
	   *      heaproot with new entry;        *
	   *      Else, delete heaproot.          *
	   ****************************************/
	  if (++rhszvectorconsumednonzeroes[k] < rhszvectornonzeroes[k]) {
	    if ((rhszbucketindex[k] =
		 MY_MOD(rhszvectorconsumednonzeroes[k],
			zBUCKET_DIM_minus1)) == 0) {
	      if ((rhszbucketptr[k] = rhszbucketptr[k]->nextbucket) == NULL) {
		
#ifdef _DEBUG
		printf("zspgx err 7: unexpected NULL nextbucket: k=%d",
		       k);
		printf("rhszvectorconsumednonzeros=%d, rhszvecnonzeroes=%d,",
		       rhszvectorconsumednonzeroes[k],rhszvectornonzeroes[k]);
		abort();
#endif
		
	      }
	    }     
	    j = rhszbucketptr[k]->index[rhszbucketindex[k]];
	    
	    if (j <= 0) {
#ifdef _DEBUG
	      printf("zspgx err 8: Illegal index: V=%d, subscript=%d\n"
		     ,k,j);
	      abort();
#endif
	      
	    }
	    
	    else
	      heapop(REPLACE,k,j,j);
	  }
	  else
	    heapop(DELETE,DUMMYVECTOR,DUMMYSUBSCRIPT,DUMMYSUBSCRIPT);
	  
	} /*matches     while (heaproot !=NULL) {     */
      } /*matches     else if (governingdstype == SS) {     */
      
      else /* if (governingdstype == CS) */ {
	
	/******************************************************************
	 *                 _COMPRESSED_SUBSCRIPTS AXPY                    *
	 ******************************************************************/
	
	
	/****************************************
	 * Else perform saxpy:                  *
	 * 1. Put in heap the 1st nonzero of    *
	 *    each vector.                      *
	 ****************************************/
	rhszvectorconsumednonzeroranges[firstsaxpyoperand] = 0;
	rhszibucketindex[firstsaxpyoperand] = 0;
	rhszvbucketindex[firstsaxpyoperand] = 0;
	rhszibucketptr[firstsaxpyoperand] =
	  rhszvectorptr[firstsaxpyoperand]->zds.cs.ibucketptr;
	rhszvbucketptr[firstsaxpyoperand] =
	  rhszvectorptr[firstsaxpyoperand]->zds.cs.vbucketptr;
	
#ifdef _DEBUG
	if (rhszibucketptr[firstsaxpyoperand] == NULL ||
	    rhszvbucketptr[firstsaxpyoperand] == NULL) {
	  printf("zspgx error 10: Unexpected NULL rhszbucketptrs: ");
	  printf("vector # %d\n",firstsaxpyoperand);
	  abort();
	}
#endif
	
	j1 = rhszibucketptr[firstsaxpyoperand]->
	  index_RANGE_BEGIN[rhszibucketindex[firstsaxpyoperand]];
	j2 = rhszibucketptr[firstsaxpyoperand]->
	  index_RANGE_END[rhszibucketindex[firstsaxpyoperand]];
	
	if (j1 <= 0 || j2 <= 0 || j2 < j1) {
#ifdef _DEBUG
	  printf("zspgx error 11: Nonsensical Nonzero Range; \n\t\t");
	  printf("vector # %d, subscript 1=%d subscript 2=%d\n",
		 firstsaxpyoperand,j1,j2);
	  abort();
#endif
	}
	
	else
	  heapinitialize(firstsaxpyoperand,j1,j2);
	
	for (k=firstsaxpyoperand+1;k<arity;k++) {

	  /*
	   * rhszvectornonzeroes[k] !=0 will be redundant and
	   *                           unnecessary most of the time
	   *                           but in the code preceding the
	   *                           addition of += module it was
	   *                           needed, so we leave it.
	   */
	
	  if (rhszvectornonzeroes[k] !=0 &&
	      rhszvectornonzeroranges[k] !=0 &&
	      ! X_IS_ZERO(alpha[k])) {
	    rhszvectorconsumednonzeroranges[k] = 0;
	    rhszibucketindex[k] = 0;
	    rhszibucketptr[k] =
	      rhszvectorptr[k]->zds.cs.ibucketptr;
	    rhszvbucketindex[k] = 0;
	    rhszvbucketptr[k] =
	      rhszvectorptr[k]->zds.cs.vbucketptr;
	    
#ifdef _DEBUG
	    if (rhszibucketptr[k] == NULL ||
		rhszvbucketptr[k] == NULL) {
	      printf("zspgx error 12: Unexpected ");
	      printf("NULL rhszibucketptr or rhszvbucketptr: vector # %d\n",k);
	      abort();
	    }
#endif
	    
	    j1 = rhszibucketptr[k]->index_RANGE_BEGIN[rhszibucketindex[k]];
	    j2 = rhszibucketptr[k]->index_RANGE_END[rhszibucketindex[k]];
	    
#ifdef _DEBUG
	    if (j1 <= 0 || j2 <= 0 || j2 < j1) {
	      printf("zspgx err 13: Nonsensical Nonzero Range; \n\t\t");
	      printf("vector # %d, subscript 1=%d subscript 2=%d\n",
		     firstsaxpyoperand,j1,j2);
	      abort();
	    }
#endif
	    
	    heapinsert(k,j1,j2);
	  }
	}
	
	/****************************************
	 * Perform saxpy:                       *
	 * 2. Initialize lhszvector.             *
	 ****************************************/
	zgetibucket(&lhszibucketptr);
	lhszvectorptr->dstype = governingdstype;
	lhszvectorptr->zds.cs.ibucketptr = lhszibucketptr;
	zgetvbucket(&lhszvbucketptr);
	lhszvectorptr->zds.cs.vbucketptr = lhszvbucketptr;
	lhszvectornonzeroranges = 0;
	lhszvectornonzeroes = 0;
	lhszvbucketindex = 0;
	
	/****************************************
	 * Perform saxpy:                       *
	 * 3. Get nonzero by popping the        *
	 *    heaproot; consume it in saxpy;    *
	 *    either replace or delete heaproot.*
	 ****************************************/

	while (heaproot !=NULL) {
	  k = heaproot->vectorindex;
	  j1 = heaproot->rangebeginindex;
	  j2 = heaproot->rangeendindex;
	  rhszibucketindex[k] =
	    MY_MOD(rhszvectorconsumednonzeroranges[k],
		   zIBUCKET_DIM_minus1);
	  lhszibucketindex = MY_MOD(lhszvectornonzeroranges,
					zIBUCKET_DIM_minus1);
	  if ((rhszvbucketindex[k] =
	       MY_MOD(rhszvbucketindex[k],zVBUCKET_DIM_minus1))==0
	      && rhszvectorconsumednonzeroranges[k] !=0) {
	    zHOP_TO_NEXT_RHSVBUCKET(k,&rhszvbucketindex[k],
				   rhszvectorconsumednonzeroranges[k],
				   rhszvectornonzeroranges[k]);
	  }
	  lhszvbucketindex = MY_MOD(lhszvbucketindex,
					zVBUCKET_DIM_minus1);
	  
	  
	  
#ifdef _DEBUG
	  if (rhszibucketptr[k]->errorflag != 9099 ||
	      rhszvbucketptr[k]->errorflag != 9009) {
	    printf("zspgx err 14: mem corruption of zibucketstore ");
	    printf("or zvbucketstore.\n");
	    abort();
	  }
#endif      
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.0. Check for startup errors.       *
	   ****************************************/
	  
#ifdef _DEBUG
	  if (lhszvectornonzeroranges == 0) {
	    if (rhszibucketindex[k] != 0 || rhszvbucketindex[k] != 0 ||
		lhszibucketindex != 0 || lhszvbucketindex != 0) {
	      printf("zspgx error 15: Zeroes expected: \n\t");
	      printf("rhszibucketindex[k]=%d, rhszvbucketindex[k]=%d\n\t",
		     rhszibucketindex[k], rhszvbucketindex[k]);
	      printf("lhszibucketindex=%d, lhszvbucketindex=%d\n\t",
		     lhszibucketindex,lhszvbucketindex);
	      abort();
	    }
	  }
#endif      
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.1. Found 1ST or NEW subscript      *
	   *      range (non-overlapping with     *
	   *      previous one, if any) for       *
	   *      lhszvector.  If needed, get new  *
	   *      ibuckets and vbuckets for        *
	   *      lhszvector.                      *
	   ****************************************/
	  if (lhszvectornonzeroranges == 0 ||
	      j1 > lhszibucketptr->index_RANGE_END[lastlhszibucketindex]+1) {/*A.1*/
	    
	    if (lhszibucketindex == 0 && 
		lhszvectornonzeroranges != 0) {
	      zgetibucket(&(lhszibucketptr->nextibucket));
	      lhszibucketptr = lhszibucketptr->nextibucket;
	    }
	    lhszibucketptr->index_RANGE_BEGIN[lhszibucketindex] = j1;
	    lhszibucketptr->index_RANGE_END[lhszibucketindex] = j2;
	    
	    if (lhszvbucketindex == 0 &&
		lhszvectornonzeroranges != 0)
	      zget_and_hop_to_next_vbucket(&lhszvbucketindex);
	    
	    lastlhszvbucketptr = lhszvbucketptr;
	    lastlhszvbucketindex = lhszvbucketindex;
	    
	    totalcurrentnonzerorange = j2-j1+1;
	    
	    lhszvectornonzeroes += totalcurrentnonzerorange;
	    lhszvectornonzeroranges++;
	    
	    
	    zNON_OVER_LAPPING_CS_RANGE_AXPY(COMPLETELY_NEW_RANGE,
					   lhszibucketindex,
					   &lastlhszibucketindex,
					   &lhszvbucketindex,
					   k,
					   &rhszvbucketindex[k],
					   alpha[k],
					   totalcurrentnonzerorange,
					   &unconsumedcurrentnonzerorange,
					   rhszvectorconsumednonzeroranges[k],
					   rhszvectornonzeroranges[k],
					   lhszvectornonzeroranges);
	    

	  } /*A.1*/
	  
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.2. Found old subscript for         *
	   *      lhszvector.  Consume  nonzero in *
	   *      saxpy.                          *
	   *                                      *
	   ****************************************/
	  else if (j1 <= 
		   lhszibucketptr->index_RANGE_END[lastlhszibucketindex]+1 &&
		   j1 >= 
		   lhszibucketptr->index_RANGE_BEGIN[lastlhszibucketindex]) {
	    /*B.1*/
	    
	    templhszvbucketptr = lhszvbucketptr;
	    templhszvbucketindex = lhszvbucketindex;
	    lhszvbucketptr = lastlhszvbucketptr;
	    
	    distancetooverlap =
	      j1 - lhszibucketptr->index_RANGE_BEGIN[lastlhszibucketindex];
	    distancetoendofvbucket = 
	      zVBUCKET_DIM - lastlhszvbucketindex;
	    
	    if (j1 == lhszibucketptr->index_RANGE_END[lastlhszibucketindex]+1) {
	      h1 = 0;
	      h2 = -1;
	    }
	    else if (j2 <= lhszibucketptr->index_RANGE_END[lastlhszibucketindex]) {
	      h1 = j1;
	      h2 = j2;
	      j1 = 0;
	      j2 = -1;
	    }
	    else {
	      h1 = j1;
	      h2 = lhszibucketptr->index_RANGE_END[lastlhszibucketindex];
	      j1 = h2+1;
	    }
	    
	    if (distancetoendofvbucket > distancetooverlap) {
	      lhszvbucketindex =
		MY_MOD(lastlhszvbucketindex + distancetooverlap,
		       zVBUCKET_DIM_minus1);
	    }
	    else if (distancetoendofvbucket < distancetooverlap) {
	      lhszvbucketindex =
		MY_MOD(distancetooverlap - distancetoendofvbucket,
		       zVBUCKET_DIM_minus1);
	    }
	    else {
	      lhszvbucketindex = 0;
	    }
	    
	    while (distancetooverlap >= distancetoendofvbucket &&
		   lhszvbucketptr->nextvbucket != NULL) {
	      lhszvbucketptr = lhszvbucketptr->nextvbucket;
	      distancetooverlap -= distancetoendofvbucket;
	      distancetoendofvbucket = zVBUCKET_DIM;
	    }
	    
	    lhszibucketptr->index_RANGE_END[lastlhszibucketindex] =
	      max(j2, lhszibucketptr->index_RANGE_END[lastlhszibucketindex]);
	    
	    if ((totalcurrentnonzerorange = h2-h1+1) > 0) {
	      
	      
	      zOVERLAPPING_CS_RANGE_AXPY(&lhszvbucketindex,
					k,
					&rhszvbucketindex[k],
					alpha[k],
					totalcurrentnonzerorange,
					&unconsumedcurrentnonzerorange,
					rhszvectorconsumednonzeroranges[k],
					rhszvectornonzeroranges[k],
					lhszvectornonzeroranges);
	
	    }
	      
	    /*
	     * This bit of logic must be performed, whether or not there
	     * is additional nonzeroes to process.
	     */
	    
	    if (j2 != -1 &&
		(lhszvbucketptr != templhszvbucketptr ||
		 (lhszvbucketindex = 
		  MY_MOD(lhszvbucketindex,zVBUCKET_DIM_minus1)) !=
		 templhszvbucketindex)) {
	      
#ifdef _DEBUG
	      printf("zspgx err 50.0: Unexpectd value discrepancies\n");
	      printf("lhszvbucketptr=%d, templhszvbucketptr=%d\n",
		     (int)lhszvbucketptr,(int)templhszvbucketptr);
	      printf("lhszvbucketindex=%d, templhszvbucketindex=%d\n",
		     lhszvbucketindex,templhszvbucketindex);
	      
	      abort();
#endif
	    }
	    else {
	      lhszvbucketptr = templhszvbucketptr;
	      lhszvbucketindex = templhszvbucketindex;
	    }
	    
	    
	    if ((totalcurrentnonzerorange = j2-j1+1) > 0) {
	      lhszvectornonzeroes += totalcurrentnonzerorange;
	      
	      /*
	       * adding new entries to vbucket
	       */
	      
	      if (lhszvbucketindex == 0 &&
		  lhszvectornonzeroranges != 0 &&
		  j2 != -1)
		zget_and_hop_to_next_vbucket(&lhszvbucketindex);
	      
	      if(rhszvbucketindex[k] == zVBUCKET_DIM && j2 != -1) {
		
		zHOP_TO_NEXT_RHSVBUCKET(k,&rhszvbucketindex[k],
				       rhszvectorconsumednonzeroranges[k],
				       rhszvectornonzeroranges[k]);
	      }

	      zNON_OVER_LAPPING_CS_RANGE_AXPY(NEW_ENTRIES_PADDED_TO_OLD_RANGE,
					     lhszibucketindex,
					     &lastlhszibucketindex,
					     &lhszvbucketindex,
					     k,
					     &rhszvbucketindex[k],
					     alpha[k],
					     totalcurrentnonzerorange,
					     &unconsumedcurrentnonzerorange,
					     rhszvectorconsumednonzeroranges[k],
					     rhszvectornonzeroranges[k],
					     lhszvectornonzeroranges);
	      
	    } /*  corresponds to the if ((totalcurrent...) { */
	    
	  } /*B.1*/
	  
	  
	  /****************************************	
	   * Perform saxpy:                       *
	   * 3.3. Found wrong subscript for       *
	   *      lhszvector.  Abort.               *
	   ****************************************/
	  
#ifdef _DEBUG
	  else {
	    printf("zspgx error 60: lhszvectornonzeroranges=%d,",
		   lhszvectornonzeroranges);
	    printf("lastlhszibucketindex=%d, \n\t",
		   lastlhszibucketindex);
	    printf("lhszibucketptr->index[%d] = (%d,%d), j1=%d\n",
		   lastlhszibucketindex,
		   lhszibucketptr->index_RANGE_BEGIN[lastlhszibucketindex],
		   lhszibucketptr->index_RANGE_END[lastlhszibucketindex],j1);
	    abort();
	  }
#endif
	  
	  
	  /****************************************
	   * Perform saxpy:                       *
	   * 3.4. If there is one, get new        *
	   *      nonzero from the rhs_vector     *
	   *      whose nonzero was just          *
	   *      consumed and replace the        *
	   *      heaproot with new entry;        *
	   *      Else, delete heaproot.          *
	   ****************************************/
	  if (++rhszvectorconsumednonzeroranges[k] <
	      rhszvectornonzeroranges[k]) {
	    if ((rhszibucketindex[k] =
		 MY_MOD(rhszvectorconsumednonzeroranges[k],
			zIBUCKET_DIM_minus1)) == 0) {
	      if ((rhszibucketptr[k] = rhszibucketptr[k]->nextibucket) ==
		  NULL) {
		
#ifdef _DEBUG
		printf("zspgx err 61: unexpected NULL nextvbucket: \n");
		printf("\t k=%d, rhszvectorconsumednonzeroranges=%d,\n",
		       k,rhszvectorconsumednonzeroranges[k]);
		printf("\t rhszvectornonzeroranges[k]=%d\n",
		       rhszvectornonzeroranges[k]);
		abort();
#endif
		
	      }
	    }     
	    j1 = rhszibucketptr[k]->index_RANGE_BEGIN[rhszibucketindex[k]];
	    j2 = rhszibucketptr[k]->index_RANGE_END[rhszibucketindex[k]];
	    
	    if (j1 <= 0 || j2 <= 0) {
#ifdef _DEBUG
	      printf("zspgx error 62: Illegal index: vector=%d, ",k);
	      printf("subscript1=%d, subscript2=%d \n",j1,j2);
	      abort();
#endif
	    }
	    
	    else
	      heapop(REPLACE,k,j1,j2);
	  }
	  else
	    heapop(DELETE,DUMMYVECTOR,DUMMYSUBSCRIPT,DUMMYSUBSCRIPT);
	} /*matches     while (heaproot !=NULL) {     */
      } /* matches  else ** if (governingdstype == CS) ** {     */	
      
      zaxpy_2_heap_post_processing(destptrptr,
				  rhszvectorptrptrs,
				  alphaptrs,
				  restorerhszvectorptr,
				  arity,
				  lhszvectornonzeroes,
				  lhszvectornonzeroranges,
				  governingdstype);
      
      
    } /* matches    else ** if we are going to use lhszvectorptr ** {  */

#ifdef _PLUSEQUAL
  } /* matches    else ** if (not plusequalsaxpy_flag) ** {    */
#endif /* _PLUSEQUAL */

  /* }  matches    else ** if (not nullsaxpyflag) ** {    */

}

