#define DOUBLE_COMPLEX
/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#include  <stdlib.h>
#include "zgaxpy_pd_macros_and_data_structures.h"
#include "zgaxpy_pd_global_variables.h"

/****************************************************************************
 *                                                                          *
 *    plusequalzspg2                                                       *
 *                                                                          *
 ****************************************************************************/

void plusequalzspg2(rhsanddestptrptr,
		     otheralphaptr,
                     otherrhszvectorptrptr,
                     status_otherrhszvector)
zvectorstruct **rhsanddestptrptr;
double_complex *otheralphaptr;
zvectorstruct **otherrhszvectorptrptr;
int status_otherrhszvector;
{
  double_complex *alphaptrs[2];
  
  int arity = 2;
  
  zvectorstruct *rhszvectorptr[2];
  zvectorstruct *tempvectorptr;
  
  int i;
  

  zibucketstruct *rhszibucketptr[2];
                                      /* Array of pointers to "current"     *
				       * buckets where "current" nonzeroes  *
				       * of the rhs_vectors reside.         */

  zibucketstruct *lhszibucketptr;
                                      /* Pointer to current bucket of       *
				       * lhszvector.                    */

  double_complex alpha[2];          /* Dereferenced alphaptrs[].          */

  int rhszibucketindex[2];        /* Array of indices of current        *
				       * nonzeroes in the current buckets   *
				       * of rhs_vectors.                    */

  int lhszibucketindex;           /* bucketindex of next nonzero to be  *
				       * added to lhszvector.           */

  int lastlhszibucketindex;       /* bucketindex of the latest added    *
				       * nonzero in the lhszvector.  We *
                                       * need this additional index to      *
				       * leap-frog buckets, if need be.     */

  int k;                              /* Counter for arity.                 */

  int lhszvectornonzeroes;        /* Count of # of nonzeroes entered    *
				       * into the lhszvector so far. By *
				       * the end of the saxpy operation,    *
				       * this index equals the total # of   *
				       * nonzeroes in the resultant vector. */

/*
 * These variables were originally for COMPRESSED_SUBSCRIPTS
 */

  zvbucketstruct *lastlhszvbucketptr;  
                                      /* Pointer to current vbucket of      *
				       * lhszvector.                    */

  int rhszvectornonzeroranges[2];
                                      /* Array of total # of nonzeroes      *
				       * ranges in each rhs_vector.         */

  int rhszvectorconsumednonzeroranges[2];
                                      /* Array of # of nonzeroes consumed   *
				       * so far in each rhs_vector.         */

  int rhszvbucketindex[2];        /* Array of indices of current        *
				       * nonzeroes in the current buckets   *
				       * of rhs_vectors.                    */

  int lhszvectornonzeroranges;    /* Count of # of nonzeroes entered    *
				       * into the lhszvector so far. By *
				       * the end of the saxpy operation,    *
				       * this index equals the total # of   *
				       * nonzeroes in the resultant vector. */

  int lhszvbucketindex;           /* bucketindex of next nonzero to be  *
				       * added to lhszvector.           */


  int lastlhszvbucketindex;       /* bucketindex of the latest added    *
				       * nonzero in the lhszvector.  We *
				       * need this additional index to frog *
				       * leap-frog buckets, if need be.     */

  zvbucketstruct *templhszvbucketptr;

  int templhszvbucketindex;

  int j1;                             /* Index for vectorsubscripts to be   *
				       * entered in (and later popped from) *
				       * the heap.                          */

  int j2;                             /* Index for vectorsubscripts to be   *
				       * entered in (and later popped from) *
				       * the heap.                          */

  int h1;
  int h2;

  int totalcurrentnonzerorange;

#ifdef _DEBUG
  int unconsumedcurrentnonzerorange;
#endif

  int distancetooverlap;
  int distancetoendofvbucket;
  int hopdistance;

  /*
   * These variables were added for union ds
   */
  
  zvectorstruct *restorerhszvectorptr[2+1];

  int governingdstype;
  
  /*
   * These variables were added for plusequal code
   */
  
  zvectorstruct **rhszvectorptrptrs[MAXARITY];

  int condition_1, condition_2;
  int lastrhszibucketindex[2];
  int lastrhszvbucketindex[2];

  zibucketstruct *lastrhszibucketptr[2];
  zvbucketstruct *lastrhszvbucketptr[2];
  zvbucketstruct *catchupto_rhsvbucketptr[2];

  zibucketstruct *temprhszibucketptr[2];
  zvbucketstruct *temprhszvbucketptr[2];
  int temprhszvbucketindex[2];

  zibucketstruct *restofplusequalibucketptr;
  zvbucketstruct *restofplusequalvbucketptr;
  int restofplusequalibucketindex;
  int restofplusequalvbucketindex;
  int restofplusequalvectorconsumednonzeroranges;

#ifdef _TRACE
  int nnzb,nnza;
  zvectorstruct *lhszvectoraddress;
  
  zvectorstruct **vectorptrptrs[2];
  
  alphaptrs[0] = alphaptr;
  vectorptrptrs[0] = vectoraptrptr;
  
  alphaptrs[1] = betaptr;
  vectorptrptrs[1] = vectorbptrptr;
  
  if(*destptrptr==NULL){ 
    nnzb = 0; 
    lhszvectoraddress= (zvectorstruct *) 99900; 
  }
  else {
    nnzb = (*destptrptr)->nonzeroes; 
    lhszvectoraddress = *destptrptr;
  }
#endif
  
  /****************************************
   * Dereferencing.                       *
   ****************************************/
  
  X_PUT_REAL(alpha[0],1.0);

#ifdef _DEBUG
  if (*rhsanddestptrptr == NULL) {
    printf("plusequalzspg2 error #1: Non_initialized lhszvector.\n");
    abort();
  }
#endif

  rhszvectorptr[0] = *rhsanddestptrptr;
  rhszvectornonzeroranges[0] =
    rhszvectorptr[0]->zds.cs.nonzeroranges;

  X_ASSIGN(alpha[1],(*otheralphaptr));

  alphaptrs[0] = &alpha[0];
  alphaptrs[1] = otheralphaptr;

  rhszvectorptr[1] = *otherrhszvectorptrptr;
  rhszvectornonzeroranges[1] =
    rhszvectorptr[1]->zds.cs.nonzeroranges;

  governingdstype = CS;
  k = 1;   
  restorerhszvectorptr[k] = rhszvectorptr[k];

  if (rhszvectorptr[k]->dstype != governingdstype) {
    zconvertdstype((&rhszvectorptr[k]),k,governingdstype);
    rhszvectornonzeroranges[k] =
      rhszvectorptr[k]->zds.cs.nonzeroranges;
  }  
 
  /******************************************************************
   *               _COMPRESSED_SUBSCRIPTS AXPY ONLY                 *
   ******************************************************************/
  
  /****************************************
   * perform saxpy:                       *
   * 1. Put in heap the 1st nonzero of    *
   *    each vector.                      *
   ****************************************/
  for (k=0;k<arity;k++) {
    rhszvectorconsumednonzeroranges[k] = 0;
    rhszibucketindex[k] = 0;
    rhszvbucketindex[k] = 0;
    rhszibucketptr[k] =
      rhszvectorptr[k]->zds.cs.ibucketptr;
    rhszvbucketptr[k] =
      rhszvectorptr[k]->zds.cs.vbucketptr;
    
#ifdef _DEBUG
    if (rhszibucketptr[k] == NULL ||
	rhszvbucketptr[k] == NULL) {
	      printf("plusequalzspg2 err 10: Unexpected NULL rhszbucketptr: ");
	      printf("vector # %d\n",k);
	      abort();
	    }
#endif
  }	  

  
  /****************************************
   * Perform saxpy:                       *
   * 2. Initialize lhszvector.             *
   ****************************************/

  lhszvectornonzeroranges = 0;
  lhszvectornonzeroes = 0;
  lhszvbucketindex = 0;


  /****************************************
   * Perform saxpy:                       *
   * 3. Get nonzero by popping the        *
   *    heaproot; consume it in saxpy;    *
   *    either replace or delete heaproot.*
   ****************************************/
  

  /* while (there are nonzeroranges to be processed AND                      *
   *         (condition_1 OR condition_2 ), where:                           *
   * condition_1 == TRUE, indicates that the range to be processed belongs   *
   *                to the plusequalvector,                                  *
   * condition_2 == TRUE, indicates that the range to be processed belongs   *
   *                to the otherrhszvector, and that this range is       *
   *                entirelyencapsulated in the last processed range of the  *
   *                plusequalvector.                                         */

  lastrhszibucketindex[0] = -1;
  lastrhszibucketptr[0] = NULL;
  
  while ((rhszvectorconsumednonzeroranges[0] <
	  rhszvectornonzeroranges[0] &&
	  rhszvectorconsumednonzeroranges[1] <
	  rhszvectornonzeroranges[1]) &&

	 ((condition_1 = (rhszibucketptr[0]->
			  index_RANGE_BEGIN[rhszibucketindex[0]] <=
			  rhszibucketptr[1]->
			  index_RANGE_BEGIN[rhszibucketindex[1]]))    ||
	  
	  (condition_2 = ((lastrhszibucketindex[0] >= 0) &&
			  (rhszibucketptr[1]->
			   index_RANGE_BEGIN[rhszibucketindex[1]] >=
			   lastrhszibucketptr[0]->
			   index_RANGE_BEGIN[lastrhszibucketindex[0]] &&
			   rhszibucketptr[1]->
			   index_RANGE_END[rhszibucketindex[1]] <=
			   lastrhszibucketptr[0]->
			   index_RANGE_END[lastrhszibucketindex[0]]))))) {

    if (condition_1) {
      k = 0;
      lastrhszibucketindex[0] = rhszibucketindex[0];
      lastrhszvbucketindex[0] = rhszvbucketindex[0];
      lastrhszibucketptr[0] = rhszibucketptr[0];
      lastrhszvbucketptr[0] = rhszvbucketptr[0];
      totalcurrentnonzerorange =
	rhszibucketptr[0]->index_RANGE_END[rhszibucketindex[0]] -
	  rhszibucketptr[0]->index_RANGE_BEGIN[rhszibucketindex[0]]+1;

#ifdef _DEBUG
      unconsumedcurrentnonzerorange = totalcurrentnonzerorange;
#endif

      hopdistance = totalcurrentnonzerorange;
      distancetoendofvbucket = zVBUCKET_DIM - rhszvbucketindex[0];
   
      if (distancetoendofvbucket > hopdistance) {
	rhszvbucketindex[0] = MY_MOD(rhszvbucketindex[0] + hopdistance,
					 zVBUCKET_DIM_minus1);
      }
      else if (distancetoendofvbucket < hopdistance) {
	rhszvbucketindex[0] = MY_MOD(hopdistance - distancetoendofvbucket,
					 zVBUCKET_DIM_minus1);
      }
      else {
	rhszvbucketindex[0] = 0;
      }
      while (hopdistance >= distancetoendofvbucket &&
	     rhszvbucketptr[0]->nextvbucket != NULL) {
	rhszvbucketptr[0] = rhszvbucketptr[0]->nextvbucket;
	hopdistance -= distancetoendofvbucket;
	distancetoendofvbucket = zVBUCKET_DIM;
      }

      lhszvectornonzeroes += totalcurrentnonzerorange;
    }
     
    else if (condition_2) {
      k = 1;
     temprhszibucketptr[0] = rhszibucketptr[0];
      temprhszvbucketptr[0] = rhszvbucketptr[0];
      rhszibucketptr[0] = lastrhszibucketptr[0];
      rhszvbucketptr[0] = lastrhszvbucketptr[0];

      /* The reason rhszvbucketindex[0] is not
       * set to lastrhszvbucketindex[0] is because
       * rhszvbucketindex[0] is computed below from
       * lastrhszvbucketindex[0].
       */

      temprhszvbucketindex[0] = rhszvbucketindex[0];
       
      distancetooverlap =
	rhszibucketptr[1]->index_RANGE_BEGIN[rhszibucketindex[1]] -
	rhszibucketptr[0]->index_RANGE_BEGIN[lastrhszibucketindex[0]];
      distancetoendofvbucket = zVBUCKET_DIM - lastrhszvbucketindex[0];
      
      if (distancetoendofvbucket > distancetooverlap) {
	rhszvbucketindex[0] =
	  MY_MOD(lastrhszvbucketindex[0] + distancetooverlap,
		 zVBUCKET_DIM_minus1);
      }
      else if (distancetoendofvbucket < distancetooverlap) {
	rhszvbucketindex[0] =
	  MY_MOD(distancetooverlap - distancetoendofvbucket,
		 zVBUCKET_DIM_minus1);
      }
      else {
	rhszvbucketindex[0] = 0;
      }
      
      while (distancetooverlap >= distancetoendofvbucket &&
	     rhszvbucketptr[0]->nextvbucket != NULL) {
	rhszvbucketptr[0] = rhszvbucketptr[0]->nextvbucket;
	distancetooverlap -= distancetoendofvbucket;
	distancetoendofvbucket = zVBUCKET_DIM;
      }

      totalcurrentnonzerorange = 
      	rhszibucketptr[1]->index_RANGE_END[rhszibucketindex[1]] -
	  rhszibucketptr[1]->index_RANGE_BEGIN[rhszibucketindex[1]]+1;

      /* Wrapping the call with the two assignments has the effect of
       * EQUIVALENCEing the data structures being pointed two by the
       * two pointers.
       */
    
      lhszvbucketptr = rhszvbucketptr[0];      
      zOVERLAPPING_CS_RANGE_AXPY(&rhszvbucketindex[0],
				1,
				&rhszvbucketindex[1],
				alpha[1],
				totalcurrentnonzerorange,
				&unconsumedcurrentnonzerorange,
				rhszvectorconsumednonzeroranges[1],
				rhszvectornonzeroranges[1],
				rhszvectornonzeroranges[0]);
      rhszvbucketptr[0] = lhszvbucketptr;
      
      rhszibucketptr[0] = temprhszibucketptr[0];
      rhszvbucketptr[0] = temprhszvbucketptr[0];
      rhszvbucketindex[0] = temprhszvbucketindex[0];
      
    } /* matches     else if (condition_2) {  */
  
    /*****************************************
     * Perform saxpy:                        *
     * 3.4. If there is one, get new         *
     *      nonzero from the rhs_vector      *
     *      whose nonzero was just consumed. *
     *****************************************/
    if (++rhszvectorconsumednonzeroranges[k] <
	rhszvectornonzeroranges[k]) {
      if ((rhszibucketindex[k] =
	   MY_MOD(rhszvectorconsumednonzeroranges[k],
		  zIBUCKET_DIM_minus1)) == 0) {
	if ((rhszibucketptr[k] = rhszibucketptr[k]->nextibucket) ==
	    NULL) {
	  
#ifdef _DEBUG
	  printf("plusqualzspg2 error #36: unexpected NULL nextvbucket: \n");
	  printf("\t k=%d, rhszvectorconsumednonzeroranges=%d,\n",
		 k,rhszvectorconsumednonzeroranges[k]);
	  printf("\t rhszvectornonzeroranges[k]=%d\n",
		 rhszvectornonzeroranges[k]);
	  abort();
#endif
	  
	}
      }
    }   

  } /* matches     1st while loop           */

  if (rhszvectorconsumednonzeroranges[1] ==
      rhszvectornonzeroranges[1]) {
    while (rhszvectorconsumednonzeroranges[0] <
	   rhszvectornonzeroranges[0]) {
      totalcurrentnonzerorange =
	rhszibucketptr[0]->index_RANGE_END[rhszibucketindex[0]] -
	  rhszibucketptr[0]->index_RANGE_BEGIN[rhszibucketindex[0]]+1;
      lhszvectornonzeroes += totalcurrentnonzerorange;

      if (++rhszvectorconsumednonzeroranges[0] <
	  rhszvectornonzeroranges[0]) {
	if ((rhszibucketindex[0] =
	     MY_MOD(rhszvectorconsumednonzeroranges[0],
		    zIBUCKET_DIM_minus1)) == 0) {
	  if ((rhszibucketptr[0] = rhszibucketptr[0]->nextibucket) ==
	      NULL) {
	    
#ifdef _DEBUG
	    printf("plusqualzspg2 err #58: unexpected NULL nextvbucket: \n");
	    printf("\t k=%d, rhszvectorconsumednonzeroranges=%d,\n",
		   0,rhszvectorconsumednonzeroranges[0]);
	    printf("\t rhszvectornonzeroranges[0]=%d\n",
		   rhszvectornonzeroranges[0]);
	    abort();
#endif
	    
	  }
	}
      }   
    }
   }

  else if (rhszvectorconsumednonzeroranges[0] ==
	   rhszvectornonzeroranges[0]) {
    while((rhszvectorconsumednonzeroranges[1] <
	   rhszvectornonzeroranges[1]) &&
	  ((lastrhszibucketindex[0] >= 0) &&
	   (rhszibucketptr[1]->
	    index_RANGE_BEGIN[rhszibucketindex[1]] >=
	    lastrhszibucketptr[0]->
	    index_RANGE_BEGIN[lastrhszibucketindex[0]] &&
	    rhszibucketptr[1]->
	    index_RANGE_END[rhszibucketindex[1]] <=
	    lastrhszibucketptr[0]->
	    index_RANGE_END[lastrhszibucketindex[0]]))) {
      
      /* Note: This part of the code is almost a copy of the part above
       * for the case where condition_2 is TRUE.  Therefore, when
       * modifying either code segment, care must be taken so that
       * both parts remain coherent.
       */

      k = 1;
      temprhszibucketptr[0] = rhszibucketptr[0];
      temprhszvbucketptr[0] = rhszvbucketptr[0];
      rhszibucketptr[0] = lastrhszibucketptr[0];
      rhszvbucketptr[0] = lastrhszvbucketptr[0];

      temprhszvbucketindex[0] = rhszvbucketindex[0];
      
      distancetooverlap =
	rhszibucketptr[1]->index_RANGE_BEGIN[rhszibucketindex[1]] -
	rhszibucketptr[0]->index_RANGE_BEGIN[lastrhszibucketindex[0]];
      distancetoendofvbucket = zVBUCKET_DIM - lastrhszvbucketindex[0];
      
      if (distancetoendofvbucket > distancetooverlap) {
	rhszvbucketindex[0] =
	  MY_MOD(lastrhszvbucketindex[0] + distancetooverlap,
		 zVBUCKET_DIM_minus1);
      }
      else if (distancetoendofvbucket < distancetooverlap) {
	rhszvbucketindex[0] =
	  MY_MOD(distancetooverlap - distancetoendofvbucket,
		 zVBUCKET_DIM_minus1);
      }
      else {
	rhszvbucketindex[0] = 0;
      }
      
      while (distancetooverlap >= distancetoendofvbucket &&
	     rhszvbucketptr[0]->nextvbucket != NULL) {
	rhszvbucketptr[0] = rhszvbucketptr[0]->nextvbucket;
	distancetooverlap -= distancetoendofvbucket;
	distancetoendofvbucket = zVBUCKET_DIM;
      }
      
      totalcurrentnonzerorange = 
      	rhszibucketptr[1]->index_RANGE_END[rhszibucketindex[1]] -
	  rhszibucketptr[1]->index_RANGE_BEGIN[rhszibucketindex[1]]+1;
      
      /* Wrapping the call with the two assignments has the effect of
       * EQUIVALENCEing the data structures being pointed two by the
       * two pointers.
       */
      
      lhszvbucketptr = rhszvbucketptr[0];      
      zOVERLAPPING_CS_RANGE_AXPY(&rhszvbucketindex[0],
 				1,
				&rhszvbucketindex[1],
				alpha[1],
				totalcurrentnonzerorange,
				&unconsumedcurrentnonzerorange,
				rhszvectorconsumednonzeroranges[1],
				rhszvectornonzeroranges[1],
				rhszvectornonzeroranges[0]);
      rhszvbucketptr[0] = lhszvbucketptr;
      
      rhszibucketptr[0] = temprhszibucketptr[0];
      rhszvbucketptr[0] = temprhszvbucketptr[0];
      rhszvbucketindex[0] = temprhszvbucketindex[0];
      
      /*****************************************
       * Perform saxpy:                        *
       * 3.4. If there is one, get new         *
       *      nonzero from the rhs_vector      *
       *      whose nonzero was just consumed. *
       *****************************************/

      if (++rhszvectorconsumednonzeroranges[k] <
	  rhszvectornonzeroranges[k]) {
	if ((rhszibucketindex[k] =
	     MY_MOD(rhszvectorconsumednonzeroranges[k],
		    zIBUCKET_DIM_minus1)) == 0) {
	  if ((rhszibucketptr[k] = rhszibucketptr[k]->nextibucket) ==
	      NULL) {
	    
#ifdef _DEBUG
	    printf("plusqualzspg2 err #75: unexpected NULL nextvbucket: \n");
	    printf("\t k=%d, rhszvectorconsumednonzeroranges=%d,\n",
		   k,rhszvectorconsumednonzeroranges[k]);
	    printf("\t rhszvectornonzeroranges[k]=%d\n",
		   rhszvectornonzeroranges[k]);
	    abort();
#endif
	    
	  }
	}
      }   
    }
    
  }

  if (rhszvectorconsumednonzeroranges[0] <
	 rhszvectornonzeroranges[0] ||
	 rhszvectorconsumednonzeroranges[1] <
	 rhszvectornonzeroranges[1]) {

    restofplusequalzvectorptr->nonzeroes = 0;
    restofplusequalzvectorptr->zds.cs.nonzeroranges =
      rhszvectornonzeroranges[0] - rhszvectorconsumednonzeroranges[0];
    restofplusequalvectorconsumednonzeroranges = 0;
      
    if (rhszvectorconsumednonzeroranges[0] <
	rhszvectornonzeroranges[0]) {

      if (rhszibucketindex[0] == 0 &&
	  rhszibucketptr[0] != lastrhszibucketptr[0]) {
	restofplusequalibucketptr = rhszibucketptr[0];
	restofplusequalibucketindex = rhszibucketindex[0];
	restofplusequalzvectorptr->zds.cs.ibucketptr =
	  restofplusequalibucketptr;
	rhszibucketptr[0] = lastrhszibucketptr[0];
      }
      else {
	zgetibucket(&restofplusequalibucketptr);
	restofplusequalzvectorptr->zds.cs.ibucketptr =
	  restofplusequalibucketptr;
	
	/* Note in the 'for' statement below, the remaining
	 * (not-yet-consumed-in-the saxpy) entries of rhszibucketptr[0]
	 * are placed into restofplusequalibucketptr.
	 */
	
	for (i=rhszibucketindex[0];i<zIBUCKET_DIM;i++) {
	  restofplusequalibucketptr->index_RANGE_BEGIN[i] =
	    rhszibucketptr[0]->index_RANGE_BEGIN[i];
	  restofplusequalibucketptr->index_RANGE_END[i] =
	    rhszibucketptr[0]->index_RANGE_END[i];
	}
	restofplusequalibucketindex = rhszibucketindex[0];
	restofplusequalibucketptr->nextibucket =
	  rhszibucketptr[0]->nextibucket;
	
	/* As distinct from the 'if' part above, there's no need here to set
	 * the values for rhszibucketptr[0] and rhszibucketindex[0],
	 * since they're already pointing where they should be pointing.
	 */
	
      }
      rhszibucketptr[0]->nextibucket = NULL;
      
      if (rhszvbucketindex[0] == 0 &&
	  rhszvbucketptr[0] != lastrhszvbucketptr[0]) {
	restofplusequalvbucketptr = rhszvbucketptr[0];
	restofplusequalvbucketindex = rhszvbucketindex[0];
	restofplusequalzvectorptr->zds.cs.vbucketptr =
	  restofplusequalvbucketptr;
	
	catchupto_rhsvbucketptr[0] = lastrhszvbucketptr[0];
	while (catchupto_rhsvbucketptr[0]->nextvbucket !=
	       rhszvbucketptr[0]) {
  
#ifdef _DEBUG
	  if (catchupto_rhsvbucketptr[0]->nextvbucket == NULL) {
	    printf("plusqualzspg2 error 18: unexpected NULL nextvbucket: \n");
	    abort();
	  }
#endif
	  
	  catchupto_rhsvbucketptr[0] = catchupto_rhsvbucketptr[0]->nextvbucket;
	}
	rhszvbucketptr[0] = catchupto_rhsvbucketptr[0];
	
	rhszvbucketindex[0] = zVBUCKET_DIM;
      }
      else {
	zgetvbucket(&restofplusequalvbucketptr);
	restofplusequalzvectorptr->zds.cs.vbucketptr =
	  restofplusequalvbucketptr;
	
	for (i=rhszvbucketindex[0];i<zVBUCKET_DIM;i++) {
	  restofplusequalvbucketptr->value[i] =
	    rhszvbucketptr[0]->value[i];
	}
	restofplusequalvbucketindex = rhszvbucketindex[0];
	restofplusequalvbucketptr->nextvbucket =
	  rhszvbucketptr[0]->nextvbucket;
      }
      rhszvbucketptr[0]->nextvbucket = NULL;
    }
    else {
      restofplusequalzvectorptr->zds.cs.ibucketptr = NULL;
      restofplusequalzvectorptr->zds.cs.vbucketptr = NULL;
    }
	  
    /* Switching vector names in order not to have to change names
     * in the code that follows.  Now, lhszvector will the new name
     * for the plusequalvector(rhszvector[0]), and rhszvector[0] will
     * be the new name for the restofplusequalvector.
     */
    
    tempvectorptr = lhszvectorptr;
    
    lhszvectorptr = rhszvectorptr[0];
    
    lhszvectornonzeroranges = rhszvectorconsumednonzeroranges[0];
    
    lhszibucketptr = rhszibucketptr[0];
    lhszvbucketptr = rhszvbucketptr[0];
    lhszvbucketindex = rhszvbucketindex[0];
    
    /*
     * lhszibucketindex is computed later using
     * lhszvectornonzeroranges.
     */
    
    lastlhszibucketindex = lastrhszibucketindex[0];
    lastlhszvbucketindex = lastrhszvbucketindex[0];
    lastlhszvbucketptr = lastrhszvbucketptr[0];
    
    /*
     * We never make use of lastlhszibucketptr.
     */

    if (rhszvectorconsumednonzeroranges[0] <
	rhszvectornonzeroranges[0]) {

      rhszvectorptr[0] = restofplusequalzvectorptr;
      rhszibucketptr[0] = rhszvectorptr[0]->zds.cs.ibucketptr;
      rhszvbucketptr[0] = rhszvectorptr[0]->zds.cs.vbucketptr;
      rhszibucketindex[0] = restofplusequalibucketindex;
      rhszvbucketindex[0] = restofplusequalvbucketindex;
      
    }

    while (rhszvectorconsumednonzeroranges[0] <
	   rhszvectornonzeroranges[0] ||
	   rhszvectorconsumednonzeroranges[1] <
	   rhszvectornonzeroranges[1]) {
      
      if
	((rhszvectorconsumednonzeroranges[1] ==
	  rhszvectornonzeroranges[1]) ||
	 ((rhszvectorconsumednonzeroranges[0] <
	   rhszvectornonzeroranges[0]) &&
	  (rhszibucketptr[0]->
	   index_RANGE_BEGIN[rhszibucketindex[0]] <=
	   rhszibucketptr[1]->
	   index_RANGE_BEGIN[rhszibucketindex[1]])))
	  k = 0;
      else if
	((rhszvectorconsumednonzeroranges[0] ==
	  rhszvectornonzeroranges[0]) ||
	 ((rhszvectorconsumednonzeroranges[1] <
	   rhszvectornonzeroranges[1]) &&
	  (rhszibucketptr[0]->
	   index_RANGE_BEGIN[rhszibucketindex[0]] >
	   rhszibucketptr[1]->
	   index_RANGE_BEGIN[rhszibucketindex[1]])))
	  k = 1;
#ifdef _DEBUG
      else {
	printf("plusequalzspg2 error #11: BAD LOGIC.\n");
	abort();
      }
#endif
      j1 = rhszibucketptr[k]->index_RANGE_BEGIN[rhszibucketindex[k]];
      j2 = rhszibucketptr[k]->index_RANGE_END[rhszibucketindex[k]];
      rhszibucketindex[k] =
	MY_MOD(rhszvectorconsumednonzeroranges[k],zIBUCKET_DIM_minus1);
      lhszibucketindex =
	MY_MOD(lhszvectornonzeroranges,zIBUCKET_DIM_minus1);
      
      if ((rhszvbucketindex[k] =
	  MY_MOD(rhszvbucketindex[k],zVBUCKET_DIM_minus1))==0
	  && rhszvectorconsumednonzeroranges[k] !=0
	  && (k !=0 || restofplusequalvectorconsumednonzeroranges !=0)) {
	zHOP_TO_NEXT_RHSVBUCKET(k,&rhszvbucketindex[k],
			       rhszvectorconsumednonzeroranges[k],
			       rhszvectornonzeroranges[k]);
      }
      lhszvbucketindex =
	MY_MOD(lhszvbucketindex,zVBUCKET_DIM_minus1);
      
#ifdef _DEBUG
      if (rhszibucketptr[k]->errorflag != 9099 ||
	  rhszvbucketptr[k]->errorflag != 9009) {
	printf("plusequalzspg2 err 14: memory corruption of ibucketstore ");
	printf("or vbucketstore.\n");
	abort();
      }
#endif      
      
      /****************************************
       * Perform saxpy:                       *
       * 3.0. Check for startup errors.       *
       ****************************************/
      
#ifdef _DEBUG
      if (lhszvectornonzeroranges == 0) {
	if (rhszibucketindex[k] != 0 || rhszvbucketindex[k] != 0 ||
	    lhszibucketindex != 0 || lhszvbucketindex != 0) {
	  printf("plusequalzspg2 error 15: Zeroes expected: \n\t");
	  printf("rhszibucketindex[k]=%d, rhszvbucketindex[k]=%d\n\t",
		 rhszibucketindex[k], rhszvbucketindex[k]);
	  printf("lhszibucketindex=%d, lhszvbucketindex=%d\n\t",
		 lhszibucketindex,lhszvbucketindex);
	  abort();
	}
      }
#endif      
      
      /****************************************
       * Perform saxpy:                       *
       * 3.1. Found 1ST or NEW subscript      *
       *      range (non-overlapping with     *
       *      previous one, if any) for       *
       *      lhszvector.  If needed, get new  *
       *      ibuckets and vbuckets for       *
       *      lhszvector.                      *
       ****************************************/
      if (lhszvectornonzeroranges == 0 ||
	  j1 > lhszibucketptr->index_RANGE_END[lastlhszibucketindex]+1) {/*A.1*/
	
	if (lhszibucketindex == 0 && lhszvectornonzeroranges != 0) {
	  zgetibucket(&(lhszibucketptr->nextibucket));
	  lhszibucketptr = lhszibucketptr->nextibucket;
	}
	lhszibucketptr->index_RANGE_BEGIN[lhszibucketindex] = j1;
	lhszibucketptr->index_RANGE_END[lhszibucketindex] = j2;
	
	if (lhszvbucketindex == 0 && lhszvectornonzeroranges != 0)
	  zget_and_hop_to_next_vbucket(&lhszvbucketindex);
	
	lastlhszvbucketptr = lhszvbucketptr;
	lastlhszvbucketindex = lhszvbucketindex;
	
	totalcurrentnonzerorange = j2-j1+1;
	
	lhszvectornonzeroes += totalcurrentnonzerorange;
	lhszvectornonzeroranges++;
	
	
	zNON_OVER_LAPPING_CS_RANGE_AXPY(COMPLETELY_NEW_RANGE,
				       lhszibucketindex,
				       &lastlhszibucketindex,
				       &lhszvbucketindex,
				       k,
				       &rhszvbucketindex[k],
				       alpha[k],
				       totalcurrentnonzerorange,
				       &unconsumedcurrentnonzerorange,
				       rhszvectorconsumednonzeroranges[k],
				       rhszvectornonzeroranges[k],
				       lhszvectornonzeroranges);
	
	
      } /*A.1*/
      
      
      /****************************************
       * Perform saxpy:                       *
       * 3.2. Found old subscript for         *
       *      lhszvector.  Consume  nonzero in *
       *      saxpy.                          *
       *                                      *
       ****************************************/
      else if (j1 <= 
	       lhszibucketptr->index_RANGE_END[lastlhszibucketindex]+1 &&
	       j1 >= 
	       lhszibucketptr->index_RANGE_BEGIN[lastlhszibucketindex]) { /*B.1*/
	
	templhszvbucketptr = lhszvbucketptr;
	templhszvbucketindex = lhszvbucketindex;
	lhszvbucketptr = lastlhszvbucketptr;
	
	distancetooverlap = j1 -
	  lhszibucketptr->index_RANGE_BEGIN[lastlhszibucketindex];
	distancetoendofvbucket = zVBUCKET_DIM -
	  lastlhszvbucketindex;
	
	if (j1 ==
	    lhszibucketptr->index_RANGE_END[lastlhszibucketindex]+1) {
	  h1 = 0;
	  h2 = -1;
	}
	else if (j2 <=
		 lhszibucketptr->index_RANGE_END[lastlhszibucketindex]) {
	  h1 = j1;
	  h2 = j2;
	  j1 = 0;
	  j2 = -1;
	}
	else {
	  h1 = j1;
	  h2 = lhszibucketptr->index_RANGE_END[lastlhszibucketindex];
	  j1 = h2+1;
	}
	
	if (distancetoendofvbucket > distancetooverlap) {
	  lhszvbucketindex =
	    MY_MOD(lastlhszvbucketindex + distancetooverlap,
		   zVBUCKET_DIM_minus1);
	}
	else if (distancetoendofvbucket < distancetooverlap) {
	  lhszvbucketindex =
	    MY_MOD(distancetooverlap - distancetoendofvbucket,
		   zVBUCKET_DIM_minus1);
	}
	else {
	  lhszvbucketindex = 0;
	}
	
	while (distancetooverlap >= distancetoendofvbucket &&
	       lhszvbucketptr->nextvbucket != NULL) {
	  lhszvbucketptr = lhszvbucketptr->nextvbucket;
	  distancetooverlap -= distancetoendofvbucket;
	  distancetoendofvbucket = zVBUCKET_DIM;
	}
	
	lhszibucketptr->index_RANGE_END[lastlhszibucketindex] =
	  max(j2,
	      lhszibucketptr->index_RANGE_END[lastlhszibucketindex]);
	
	if ((totalcurrentnonzerorange = h2-h1+1) > 0) {
	  
	  
	  zOVERLAPPING_CS_RANGE_AXPY(&lhszvbucketindex,
				    k,
				    &rhszvbucketindex[k],
				    alpha[k],
				    totalcurrentnonzerorange,
				    &unconsumedcurrentnonzerorange,
				    rhszvectorconsumednonzeroranges[k],
				    rhszvectornonzeroranges[k],
				    lhszvectornonzeroranges);
	  
	  
	}
	
	/*
	 * This bit of logic must be performed, whether or not there
	 * is additional nonzeroes to process.
	 */
	
	if (j2 != -1 &&
	    (lhszvbucketptr != templhszvbucketptr ||
	     (lhszvbucketindex =
	      MY_MOD(lhszvbucketindex,zVBUCKET_DIM_minus1)) !=
	     templhszvbucketindex)) {
	  
#ifdef _DEBUG
	  printf("plusequalzspg2 err 50.0: Unexpectd value discrepancies\n");
	  printf("lhszvbucketptr=%d, templhszvbucketptr=%d\n",
		 (int)lhszvbucketptr,(int)templhszvbucketptr);
	  printf("lhszvbucketindex=%d, templhszvbucketindex=%d\n",
		 lhszvbucketindex,templhszvbucketindex);
	  
	  abort();
#endif
	}
	else {
	  lhszvbucketptr = templhszvbucketptr;
	  lhszvbucketindex = templhszvbucketindex;
	}
	
	
	if ((totalcurrentnonzerorange = j2-j1+1) > 0) {
	  lhszvectornonzeroes += totalcurrentnonzerorange;
	  
	  /*
	   * adding new entries to vbucket
	   */
	  
	  if (lhszvbucketindex == 0 &&
	      lhszvectornonzeroranges != 0 &&
	      j2 != -1)
	    zget_and_hop_to_next_vbucket(&lhszvbucketindex);
	  
	  if(rhszvbucketindex[k] == zVBUCKET_DIM && j2 != -1) {
	    
	    zHOP_TO_NEXT_RHSVBUCKET(k,&rhszvbucketindex[k],
				   rhszvectorconsumednonzeroranges[k],
				   rhszvectornonzeroranges[k]);
	  }
	  
	  
	  
	  
	  zNON_OVER_LAPPING_CS_RANGE_AXPY(NEW_ENTRIES_PADDED_TO_OLD_RANGE,
					 lhszibucketindex,
					 &lastlhszibucketindex,
					 &lhszvbucketindex,
					 k,
					 &rhszvbucketindex[k],
					 alpha[k],
					 totalcurrentnonzerorange,
					 &unconsumedcurrentnonzerorange,
					 rhszvectorconsumednonzeroranges[k],
					 rhszvectornonzeroranges[k],
					 lhszvectornonzeroranges);
	  
	  
	} /* matches     if ((totalcurrent...) { */
      } /*B.1*/
      
      
      /****************************************	
       * Perform saxpy:                       *
       * 3.3. Found wrong subscript for       *
       *      lhszvector.  Exit.               *
       ****************************************/
      
#ifdef _DEBUG
      else {
	printf("plusequalzspg2 error 60: lhszvectornonzeroranges=%d,",
	       lhszvectornonzeroranges);
	printf("lastlhszibucketindex=%d, \n\t",lastlhszibucketindex);
	printf("lhszibucketptr->index[%d] = (%d,%d), j1=%d\n",
	       lastlhszibucketindex,
	       lhszibucketptr->
	       index_RANGE_BEGIN[lastlhszibucketindex],
	       lhszibucketptr->
	       index_RANGE_END[lastlhszibucketindex],
	       j1);
	abort();
      }
#endif
      
      
      /****************************************
       * Perform saxpy:                       *
       * 3.4. If there is one, get new        *
       *      nonzero from the rhs_vector     *
       *      whose nonzero was just          *
       *      consumed and replace the        *
       *      heaproot with new entry;        *
       *      Else, delete heaproot.          *
       ****************************************/

      if (k==0) restofplusequalvectorconsumednonzeroranges++;

      if (++rhszvectorconsumednonzeroranges[k] <
	  rhszvectornonzeroranges[k]) {
	if ((rhszibucketindex[k] =
	     MY_MOD(rhszvectorconsumednonzeroranges[k],
		    zIBUCKET_DIM_minus1)) == 0) {
	  if ((rhszibucketptr[k] = rhszibucketptr[k]->nextibucket) ==
	      NULL) {
	    
#ifdef _DEBUG
	    printf("plusequalzspg2 err 61: unexpected NULL nextvbucket: \n");
	    printf("\t k=%d, rhszvectorconsumednonzeroranges=%d,\n",
		   k,rhszvectorconsumednonzeroranges[k]);
	    printf("\t rhszvectornonzeroranges[k]=%d\n",
		   rhszvectornonzeroranges[k]);
	    abort();
#endif
	    
	  }
	}
      }  
      
    } /* matches     2nd while loop           */
  } 

  else {

    /* tempvectorptr is used in the if part as temporary storage
     * for the value of lhszvectorptr.  We echo that statement
     * here so that later the reinstatement of the value of
     * lhszvectorptr could be done regardless of the branch taken.
     */

    tempvectorptr = lhszvectorptr;
 
  }

  rhszvectorptrptrs[1] = &rhszvectorptr[1];
  
  zplusequal_axpy_postprocessing(rhsanddestptrptr,
				 rhszvectorptrptrs[1],
				 status_otherrhszvector,
				 &restofplusequalzvectorptr,
				 alphaptrs,
				 restorerhszvectorptr,
				 arity,
				 lhszvectornonzeroes,
				 lhszvectornonzeroranges,
				 CS);
  
  
  
  lhszvectorptr =  tempvectorptr;
  
  
#ifdef _TRACE 
  nnza = (*destptrptr)->nonzeroes; 
  zadinfo(arity,nnzb,lhszvectoraddress,nnza,alphaptrs,vectorptrptrs);
#endif

}

