#define DOUBLE_COMPLEX
/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#include  <stdlib.h>
#include "zgaxpy_pd_macros_and_data_structures.h"
#include "zgaxpy_pd_global_variables.h"

/****************************************************************************
 *                                                                          *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/

void znon_over_lapping_cs_range_axpy(rangedestination,
				    lhszibucketindex,
				    addressoflastlhszibucketindex,
				    addressoflhszvbucketindex,
				    k,
				    addressofrhszvbucketindex_k,
				    alpha_k,
				    totalcurrentnonzerorange
#ifdef _DEBUG
				    ,addressofunconsumedcurrentnonzerorange
				    ,rhszvectorconsumednonzeroranges_k
				    ,rhszvectornonzeroranges_k
				    ,lhszvectornonzeroranges
#endif
				    )
int rangedestination;
int lhszibucketindex;
int *addressoflastlhszibucketindex;
int *addressoflhszvbucketindex;
int k;
int *addressofrhszvbucketindex_k;
double_complex alpha_k;
int totalcurrentnonzerorange;
#ifdef _DEBUG
int *addressofunconsumedcurrentnonzerorange;
int rhszvectorconsumednonzeroranges_k;
int rhszvectornonzeroranges_k;
int lhszvectornonzeroranges;
#endif
{
  int firstloopsize;
  int numberofdoubleforloopiterations;
  int firstiterationloopsize;
  int seconditerationloopsize;
  int remainder1loopsize;
  int remainder2loopsize;

  int i;


  zDETERMINE_LOOP_BOUNDS(totalcurrentnonzerorange,
			*addressoflhszvbucketindex,
			*addressofrhszvbucketindex_k,
			&firstloopsize,
			&numberofdoubleforloopiterations,
			&firstiterationloopsize,
			&seconditerationloopsize,
			&remainder1loopsize,
			&remainder2loopsize,
			addressofunconsumedcurrentnonzerorange);
  zDENSE_AXPY_LOOP(NEW_RANGE,
		  addressoflhszvbucketindex,
		  addressofrhszvbucketindex_k,
		  k,
		  firstloopsize,
		  alpha_k,
		  addressofunconsumedcurrentnonzerorange);
  
    
  /*********************************************
   * Case 1: rhszvbucket has more spaces   *
   *         remaining than lhszvbucket.   *
   *********************************************/
  
  if (*addressofrhszvbucketindex_k <
      *addressoflhszvbucketindex) {
    
    for(i=0;i<numberofdoubleforloopiterations;i++) {
      zDEBUGGING_CODE_1(LEFT,
		       *addressoflhszvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      zget_and_hop_to_next_vbucket(addressoflhszvbucketindex);
      zDENSE_AXPY_LOOP(NEW_RANGE,
		      addressoflhszvbucketindex,
		      addressofrhszvbucketindex_k,
		      k,
		      firstiterationloopsize,alpha_k,
		      addressofunconsumedcurrentnonzerorange);
      zDEBUGGING_CODE_1(RIGHT,
		       *addressofrhszvbucketindex_k,
		       *addressofunconsumedcurrentnonzerorange);
      zHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhszvbucketindex_k,
			     rhszvectorconsumednonzeroranges_k,
			     rhszvectornonzeroranges_k);
      zDENSE_AXPY_LOOP(NEW_RANGE,
		      addressoflhszvbucketindex,
		      addressofrhszvbucketindex_k,
		      k,
		      seconditerationloopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    
    if (remainder1loopsize > 0) {
      zDEBUGGING_CODE_1(LEFT,
		       *addressoflhszvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      zget_and_hop_to_next_vbucket(addressoflhszvbucketindex);
      zDENSE_AXPY_LOOP(NEW_RANGE,
		      addressoflhszvbucketindex,
		      addressofrhszvbucketindex_k,
		      k,
		      remainder1loopsize,alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    if (remainder2loopsize > 0) {
      zDEBUGGING_CODE_1(RIGHT,
		       *addressofrhszvbucketindex_k,
		       *addressofunconsumedcurrentnonzerorange);
      zHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhszvbucketindex_k,
			     rhszvectorconsumednonzeroranges_k,
			     rhszvectornonzeroranges_k);
      zDENSE_AXPY_LOOP(NEW_RANGE,
		      addressoflhszvbucketindex,
		      addressofrhszvbucketindex_k,
		      k,
		      remainder2loopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    DEBUGGING_CODE_2(*addressofunconsumedcurrentnonzerorange);
    if (rangedestination == COMPLETELY_NEW_RANGE) {
      *addressoflastlhszibucketindex = lhszibucketindex;
    }
  }

  /*********************************************
   * Case 2: rhszvbucket has less spaces   *
   * remaining than lhszvbucket.           *
   *********************************************/	  
  else if (*addressofrhszvbucketindex_k >
	   *addressoflhszvbucketindex) {
    
    for(i=0;i<numberofdoubleforloopiterations;i++) {
      zDEBUGGING_CODE_1(RIGHT,
		       *addressofrhszvbucketindex_k,
		       *addressofunconsumedcurrentnonzerorange);
      zHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhszvbucketindex_k,
			     rhszvectorconsumednonzeroranges_k,
			     rhszvectornonzeroranges_k);
      zDENSE_AXPY_LOOP(NEW_RANGE,
		      addressoflhszvbucketindex,
		      addressofrhszvbucketindex_k,
		      k,
		      firstiterationloopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
      zDEBUGGING_CODE_1(LEFT,
		       *addressoflhszvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      zget_and_hop_to_next_vbucket(addressoflhszvbucketindex);
      zDENSE_AXPY_LOOP(NEW_RANGE,
		      addressoflhszvbucketindex,
		      addressofrhszvbucketindex_k,
		      k,
		      seconditerationloopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    
    if (remainder1loopsize > 0) {
      zDEBUGGING_CODE_1(RIGHT,
		       *addressofrhszvbucketindex_k,
		       *addressofunconsumedcurrentnonzerorange);
      zHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhszvbucketindex_k,
			     rhszvectorconsumednonzeroranges_k,
			     rhszvectornonzeroranges_k);
      zDENSE_AXPY_LOOP(NEW_RANGE,
		      addressoflhszvbucketindex,
		      addressofrhszvbucketindex_k,
		      k,
		      remainder1loopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    
    if (remainder2loopsize > 0) {
      zDEBUGGING_CODE_1(LEFT,
		       *addressoflhszvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      zget_and_hop_to_next_vbucket(addressoflhszvbucketindex);
      zDENSE_AXPY_LOOP(NEW_RANGE,
		      addressoflhszvbucketindex,
		      addressofrhszvbucketindex_k,
		      k,
		      remainder2loopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    DEBUGGING_CODE_2(*addressofunconsumedcurrentnonzerorange);
    if (rangedestination == COMPLETELY_NEW_RANGE) {
      *addressoflastlhszibucketindex = lhszibucketindex;
    }
  }
  
   
  /*********************************************
   * Case 3: rhszvbucket has same # of     *
   * spaces as lhszvbucket (i.e. 0).       *
   *********************************************/
  else if (*addressofrhszvbucketindex_k ==
	   *addressoflhszvbucketindex) {
    
    for(i=0;i<numberofdoubleforloopiterations;i++) {
      zDEBUGGING_CODE_1(LEFT,
		       *addressoflhszvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      zget_and_hop_to_next_vbucket(addressoflhszvbucketindex);
      zHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhszvbucketindex_k,
			     rhszvectorconsumednonzeroranges_k,
			     rhszvectornonzeroranges_k);
      zDENSE_AXPY_LOOP(NEW_RANGE,
		      addressoflhszvbucketindex,
		      addressofrhszvbucketindex_k,
		      k,
		      firstiterationloopsize+seconditerationloopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    if (remainder1loopsize+remainder2loopsize > 0) {
      zDEBUGGING_CODE_1(LEFT,
		       *addressoflhszvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      zget_and_hop_to_next_vbucket(addressoflhszvbucketindex);
      zHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhszvbucketindex_k,
			     rhszvectorconsumednonzeroranges_k,
			     rhszvectornonzeroranges_k);
      zDENSE_AXPY_LOOP(NEW_RANGE,
		      addressoflhszvbucketindex,
		      addressofrhszvbucketindex_k,
		      k,
		      remainder1loopsize+remainder2loopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    DEBUGGING_CODE_2(*addressofunconsumedcurrentnonzerorange);
    if (rangedestination == COMPLETELY_NEW_RANGE) {
      *addressoflastlhszibucketindex = lhszibucketindex;
    }
  }
}






/****************************************************************************
 *                                                                          *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/

void zoverlapping_cs_range_axpy(addressoflhszvbucketindex,
			       k,
			       addressofrhszvbucketindex_k,
			       alpha_k,
			       totalcurrentnonzerorange
#ifdef _DEBUG
			       ,addressofunconsumedcurrentnonzerorange
			       ,rhszvectorconsumednonzeroranges_k
			       ,rhszvectornonzeroranges_k
			       ,lhszvectornonzeroranges
#endif
			       )
int *addressoflhszvbucketindex;
int k;
int *addressofrhszvbucketindex_k;
double_complex alpha_k;
int totalcurrentnonzerorange;
#ifdef _DEBUG
int *addressofunconsumedcurrentnonzerorange;
int rhszvectorconsumednonzeroranges_k;
int rhszvectornonzeroranges_k;
int lhszvectornonzeroranges;
#endif
{
  int firstloopsize;
  int numberofdoubleforloopiterations;
  int firstiterationloopsize;
  int seconditerationloopsize;
  int remainder1loopsize;
  int remainder2loopsize;

  int i;
  
  zDETERMINE_LOOP_BOUNDS(totalcurrentnonzerorange,
			*addressoflhszvbucketindex,
			*addressofrhszvbucketindex_k,
			&firstloopsize,
			&numberofdoubleforloopiterations,
			&firstiterationloopsize,
			&seconditerationloopsize,
			&remainder1loopsize,
			&remainder2loopsize,
			addressofunconsumedcurrentnonzerorange);
  zDENSE_AXPY_LOOP(OLD_RANGE,
		  addressoflhszvbucketindex,
		  addressofrhszvbucketindex_k,
		  k,
		  firstloopsize,
		  alpha_k,
		  addressofunconsumedcurrentnonzerorange);
  
  
  /*********************************************
   * Case 1: rhszvbucket has more spaces   *
   *         remaining than lhszvbucket.   *
   *********************************************/
  if (*addressofrhszvbucketindex_k <
      *addressoflhszvbucketindex) {
    /*B.2.1*/
    
    for(i=0;i<numberofdoubleforloopiterations;i++) { /*B.3.1*/
      zDEBUGGING_CODE_1(LEFT,
		       *addressoflhszvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      zHOP_TO_NEXT_LHSVBUCKET(addressoflhszvbucketindex,
			     lhszvectornonzeroranges);
      zDENSE_AXPY_LOOP(OLD_RANGE,
		      addressoflhszvbucketindex,
		      addressofrhszvbucketindex_k,
		      k,
		      firstiterationloopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
      zDEBUGGING_CODE_1(RIGHT,
		       *addressofrhszvbucketindex_k,
		       *addressofunconsumedcurrentnonzerorange);
      zHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhszvbucketindex_k,
			     rhszvectorconsumednonzeroranges_k,
			     rhszvectornonzeroranges_k);
      zDENSE_AXPY_LOOP(OLD_RANGE,
		      addressoflhszvbucketindex,
		      addressofrhszvbucketindex_k,
		      k,
		      seconditerationloopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    } /*B.3.1*/
    
    if (remainder1loopsize > 0) {    	    
      zDEBUGGING_CODE_1(LEFT,
		       *addressoflhszvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      zHOP_TO_NEXT_LHSVBUCKET(addressoflhszvbucketindex,
			     lhszvectornonzeroranges);
      zDENSE_AXPY_LOOP(OLD_RANGE,
		      addressoflhszvbucketindex,
		      addressofrhszvbucketindex_k,
		      k,
		      remainder1loopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    
    if (remainder2loopsize > 0) {    
      zDEBUGGING_CODE_1(RIGHT,
		       *addressofrhszvbucketindex_k,
		       *addressofunconsumedcurrentnonzerorange);
      zHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhszvbucketindex_k,
			     rhszvectorconsumednonzeroranges_k,
			     rhszvectornonzeroranges_k);
      zDENSE_AXPY_LOOP(OLD_RANGE,
		      addressoflhszvbucketindex,
		      addressofrhszvbucketindex_k,
		      k,
		      remainder2loopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    DEBUGGING_CODE_2(*addressofunconsumedcurrentnonzerorange);
  } /*B.2.1*/
  
  /*********************************************
   * Case 2: rhszvbucket has less spaces   *
   * remaining than lhszvbucket.           *
   *********************************************/
  else if (*addressofrhszvbucketindex_k >
	   *addressoflhszvbucketindex) {
 /*B.2.2*/
    
    for(i=0;i<numberofdoubleforloopiterations;i++) { /*B.3.2*/
      zDEBUGGING_CODE_1(RIGHT,
		       *addressofrhszvbucketindex_k,
		       *addressofunconsumedcurrentnonzerorange);
      zHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhszvbucketindex_k,
			     rhszvectorconsumednonzeroranges_k,
			     rhszvectornonzeroranges_k);
      zDENSE_AXPY_LOOP(OLD_RANGE,
		      addressoflhszvbucketindex,
		      addressofrhszvbucketindex_k,
		      k,
		      firstiterationloopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
      zDEBUGGING_CODE_1(LEFT,
		       *addressoflhszvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      zHOP_TO_NEXT_LHSVBUCKET(addressoflhszvbucketindex,
			     lhszvectornonzeroranges);
      zDENSE_AXPY_LOOP(OLD_RANGE,
		      addressoflhszvbucketindex,
		      addressofrhszvbucketindex_k,
		      k,
		      seconditerationloopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    } /*B.3.2*/
    
    if (remainder1loopsize > 0) {        
      zDEBUGGING_CODE_1(RIGHT,
		       *addressofrhszvbucketindex_k,
		       *addressofunconsumedcurrentnonzerorange);
      zHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhszvbucketindex_k,
			     rhszvectorconsumednonzeroranges_k,
			     rhszvectornonzeroranges_k);
      zDENSE_AXPY_LOOP(OLD_RANGE,
		      addressoflhszvbucketindex,
		      addressofrhszvbucketindex_k,
		      k,
		      remainder1loopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    if (remainder2loopsize > 0) {    
      zDEBUGGING_CODE_1(LEFT,
		       *addressoflhszvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      zHOP_TO_NEXT_LHSVBUCKET(addressoflhszvbucketindex,
			     lhszvectornonzeroranges);
      zDENSE_AXPY_LOOP(OLD_RANGE,
		      addressoflhszvbucketindex,
		      addressofrhszvbucketindex_k,
		      k,
		      remainder2loopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    DEBUGGING_CODE_2(*addressofunconsumedcurrentnonzerorange);
  } /*B.2.2*/
  
  
  /*********************************************
   * Case 3: rhszvbucket has same # of     *
   * spaces as lhszvbucket (i.e. 0).       *
   *********************************************/
  else if(*addressofrhszvbucketindex_k ==
	  *addressoflhszvbucketindex) {
    /*B.2.3*/
    
    for(i=0;i<numberofdoubleforloopiterations;i++) { /*B.3.3*/
      zDEBUGGING_CODE_1(LEFT,
		       *addressoflhszvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      zHOP_TO_NEXT_LHSVBUCKET(addressoflhszvbucketindex,
			     lhszvectornonzeroranges);
      zHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhszvbucketindex_k,
			     rhszvectorconsumednonzeroranges_k,
			     rhszvectornonzeroranges_k);
      zDENSE_AXPY_LOOP(OLD_RANGE,
		      addressoflhszvbucketindex,
		      addressofrhszvbucketindex_k,
		      k,
		      firstiterationloopsize+seconditerationloopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    } /*B.3.3*/
    
    if (remainder1loopsize+remainder2loopsize > 0) {    
      zDEBUGGING_CODE_1(LEFT,
		       *addressoflhszvbucketindex,
		       *addressofunconsumedcurrentnonzerorange);
      zHOP_TO_NEXT_LHSVBUCKET(addressoflhszvbucketindex,
			     lhszvectornonzeroranges);
      zHOP_TO_NEXT_RHSVBUCKET(k,
			     addressofrhszvbucketindex_k,
			     rhszvectorconsumednonzeroranges_k,
			     rhszvectornonzeroranges_k);
      zDENSE_AXPY_LOOP(OLD_RANGE,
		      addressoflhszvbucketindex,
		      addressofrhszvbucketindex_k,
		      k,
		      remainder1loopsize+remainder2loopsize,
		      alpha_k,
		      addressofunconsumedcurrentnonzerorange);
    }
    DEBUGGING_CODE_2(*addressofunconsumedcurrentnonzerorange);
  } /*B.2.3*/
}



/****************************************************************************
 *                                                                          *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/

void zhop_to_next_rhsvbucket(k,
			    vbucketindex
#ifdef _DEBUG
			    ,rhszvectorconsumednonzeroranges
			    ,rhszvectornonzeroranges
#endif
			    )
int k;
int *vbucketindex;
#ifdef _DEBUG
int rhszvectorconsumednonzeroranges;
int rhszvectornonzeroranges;
#endif
{
  if ((rhszvbucketptr[k] = rhszvbucketptr[k]->nextvbucket) == NULL) {
	      
#ifdef _DEBUG
    printf("zhop_to_next_rhsvbucket error: unexpected NULL nextvbucket: \n");
    printf("\t k=%d, rhszvectorconsumednonzeroranges[k]=%d,\n",
	   k,rhszvectorconsumednonzeroranges);
    printf("\t rhszvectornonzeroranges[k]=%d\n",
	   rhszvectornonzeroranges);
    abort();
#endif
  }
  else {
    *vbucketindex = 0;
  }
}





/****************************************************************************
 *                                                                          *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/

void zhop_to_next_lhsvbucket(vbucketindex
#ifdef _DEBUG
		      ,lhszvectornonzeroranges
#endif
		      )
int *vbucketindex;
#ifdef _DEBUG
int lhszvectornonzeroranges;
#endif
{
  if ((lhszvbucketptr = lhszvbucketptr->nextvbucket) == NULL) {
    
#ifdef _DEBUG
    printf("zhop_to_next_lhsvbucket error: unexpected NULL nextvbucket: \n");
    printf("lhszvectornonzeroranges=%d\n",lhszvectornonzeroranges);
    abort();
#endif
  }
  else {
    *vbucketindex = 0;
  }
}  





/****************************************************************************
 *                                                                          *
 *      zget_and_hop_to_next_vbucket(lhszvbucketindex)                   *
 *                                                                          *
 ****************************************************************************/

void zget_and_hop_to_next_vbucket(lhszvbucketindex)
int *lhszvbucketindex;
{
  zgetvbucket(&(lhszvbucketptr->nextvbucket));
  lhszvbucketptr = lhszvbucketptr->nextvbucket;
  *lhszvbucketindex = 0;
}







/****************************************************************************
 *                                                                          *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/

void zconvertdstype(oldvectorptrptr,k,newvectordstype)
zvectorstruct **oldvectorptrptr;
int k;
int newvectordstype;
{
  zbucketstruct *holdingbinbucketptr;
  zbucketstruct *oldbucketptr;
  zibucketstruct *holdingbinibucketptr;
  zvbucketstruct *holdingbinvbucketptr;
  int bucketindex;
  int ibucketindex = 0;
  int vbucketindex;
  int j;
  int ij;
  int vj;
  int lastindex;
  zvectorstruct *oldvectorptr = (*oldvectorptrptr);
  int oldvectordstype = oldvectorptr->dstype;

  if (oldvectordstype >= newvectordstype || newvectordstype == S1) {
#ifdef _DEBUG
    printf("zconvertdstype error 1: Illegal Data Structure Conversion:\n\t");
    printf("newvectordstype=%d,    oldvectordstype=%d\n",
	   newvectordstype,oldvectordstype);
    abort();
#endif
  }
  else if (oldvectordstype == S1 &&  newvectordstype == SS) {
    holdingbinzvectorptr[k]->nonzeroes = 1;
    holdingbinzvectorptr[k]->dstype = SS;
    zgetbucket(&holdingbinbucketptr);
    holdingbinzvectorptr[k]->zds.ss.bucketptr = holdingbinbucketptr;
    holdingbinbucketptr->index[0] = oldvectorptr->zds.s1.index;
    holdingbinbucketptr->value[0] = oldvectorptr->zds.s1.value;
  }
  else if (oldvectordstype == S1 &&  newvectordstype == CS) {
    holdingbinzvectorptr[k]->nonzeroes = 1;
    holdingbinzvectorptr[k]->dstype = CS;
    zgetibucket(&holdingbinibucketptr);
    zgetvbucket(&holdingbinvbucketptr);
    holdingbinzvectorptr[k]->zds.cs.nonzeroranges = 1;
    holdingbinzvectorptr[k]->zds.cs.ibucketptr = holdingbinibucketptr;
    holdingbinzvectorptr[k]->zds.cs.vbucketptr = holdingbinvbucketptr;
    holdingbinibucketptr->index_RANGE_BEGIN[0] = oldvectorptr->zds.s1.index;
    holdingbinibucketptr->index_RANGE_END[0] = oldvectorptr->zds.s1.index;
    holdingbinvbucketptr->value[0] = oldvectorptr->zds.s1.value;
  }
  else if (oldvectordstype == SS &&  newvectordstype == CS) {
    oldbucketptr = oldvectorptr->zds.ss.bucketptr;
    holdingbinzvectorptr[k]->nonzeroes = oldvectorptr->nonzeroes;
    holdingbinzvectorptr[k]->dstype = CS;
    zgetibucket(&holdingbinibucketptr);
    holdingbinzvectorptr[k]->zds.cs.ibucketptr = holdingbinibucketptr;
    zgetvbucket(&holdingbinvbucketptr);
    holdingbinzvectorptr[k]->zds.cs.vbucketptr = holdingbinvbucketptr;
    holdingbinibucketptr->index_RANGE_BEGIN[0] = oldbucketptr->index[0];
    holdingbinvbucketptr->value[0] = oldbucketptr->value[0];
    lastindex = oldbucketptr->index[0];
    
    for (j=1,ij=0,vj=1;j<oldvectorptr->nonzeroes;j++,vj++) {
      if ((bucketindex = MY_MOD(j,zBUCKET_DIM_minus1)) == 0)
	  oldbucketptr = oldbucketptr->nextbucket;
      if ((vbucketindex = MY_MOD(vj,zVBUCKET_DIM_minus1)) == 0) {
	zgetvbucket(&holdingbinvbucketptr->nextvbucket);
	holdingbinvbucketptr = holdingbinvbucketptr->nextvbucket;
      }
      holdingbinvbucketptr->value[vbucketindex] =
	oldbucketptr->value[bucketindex];

      if (oldbucketptr->index[bucketindex] > lastindex+1) {
	holdingbinibucketptr->index_RANGE_END[ibucketindex] = lastindex;
	ij++;
	if ((ibucketindex = MY_MOD(ij,zIBUCKET_DIM_minus1)) == 0) {
	  zgetibucket(&holdingbinibucketptr->nextibucket);
	  holdingbinibucketptr = holdingbinibucketptr->nextibucket;
	}
	holdingbinibucketptr->index_RANGE_BEGIN[ibucketindex] =
	  oldbucketptr->index[bucketindex];
      }
      else if (oldbucketptr->index[bucketindex] < lastindex+1) {
#ifdef _DEBUG
	printf("zconvertdstype error 2: BAD LOGIC:\n\t");
	printf("newvectordstype=%d,    oldvectordstype=%d\n",
	       newvectordstype,oldvectordstype);
	abort();
#endif
      }
      lastindex = oldbucketptr->index[bucketindex];
    }
    holdingbinibucketptr->index_RANGE_END[ibucketindex] = lastindex;    
    holdingbinzvectorptr[k]->zds.cs.nonzeroranges = ij+1;
  }
  else /* We should never go thru this branch */ {
#ifdef _DEBUG
    printf("zconvertdstype error 3: BAD LOGIC:\n\t");
    printf("newvectordstype=%d,    oldvectordstype=%d\n",
	   newvectordstype,oldvectordstype);
    abort();
#endif
  }
  *oldvectorptrptr =  holdingbinzvectorptr[k];
}








/****************************************************************************
 *                                                                          *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/

void zaxpy_0_1_post_processing(destptrptr)
zvectorstruct **destptrptr;
{
  zvectorstruct *tempvectorptr;
  zbucketstruct *tempbucketptr;
  zibucketstruct *tempibucketptr;
  zvbucketstruct *tempvbucketptr;
  
  /*********************************************
   * Switch lhszvectorptr and *destptrptr. *
   *********************************************/
  tempvectorptr = (*destptrptr);
  (*destptrptr) = lhszvectorptr;
  lhszvectorptr = tempvectorptr;
  
  if (tempvectorptr->dstype == SS) {
    /*********************************************
     * If *destptrptr was not empty,             *
     * add its buckets to freedzbucketslist. *
     *********************************************/
    tempbucketptr = tempvectorptr->zds.ss.bucketptr;
    if (tempbucketptr != NULL) {
      while (tempbucketptr->nextbucket != NULL)
	tempbucketptr = tempbucketptr->nextbucket;
      tempbucketptr->nextbucket = freedzbucketslist;
      RESET_freedzbucketslist(tempvectorptr->zds.ss.bucketptr);
      tempvectorptr->zds.ss.bucketptr = NULL;
    }
  }
  else if (tempvectorptr->dstype == CS) {
    /********************************************
     * If *destptrptr was not empty, add        *
     * its ibuckets to freedzibucketslist.  *
     ********************************************/
    
    tempibucketptr = tempvectorptr->zds.cs.ibucketptr;
    tempvbucketptr = tempvectorptr->zds.cs.vbucketptr;
    
    if (tempibucketptr != NULL) {
      while (tempibucketptr->nextibucket != NULL)
	tempibucketptr = tempibucketptr->nextibucket;
      tempibucketptr->nextibucket = freedzibucketslist;
      RESET_freedzibucketslist(tempvectorptr->zds.cs.ibucketptr);
      tempvectorptr->zds.cs.ibucketptr = NULL;
    }
    
    /********************************************
     * If *destptrptr was not empty, add        *
     * its vbuckets to freedzvbucketslist.  *
     ********************************************/
    
    if (tempvbucketptr != NULL) {
      while (tempvbucketptr->nextvbucket != NULL)
	tempvbucketptr = tempvbucketptr->nextvbucket;
      tempvbucketptr->nextvbucket = freedzvbucketslist;
      RESET_freedzvbucketslist(tempvectorptr->zds.cs.vbucketptr);
      tempvectorptr->zds.cs.vbucketptr = NULL;
    }
  } /* matches   else if (tempvectorptr->dstype == CS) {             */
}


/****************************************************************************
 *                                                                          *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/

void zaxpy_2_heap_post_processing(destptrptr,
				 rhszvectorptrptrs,
				 alphaptrs,
				 restorerhszvectorptr,
				 arity,
				 lhszvectornonzeroes,
				 lhszvectornonzeroranges,
				 governingdstype)
zvectorstruct **destptrptr;
zvectorstruct **rhszvectorptrptrs[];
double_complex *alphaptrs[];
zvectorstruct *restorerhszvectorptr[MAXARITY+1];
int arity;
int lhszvectornonzeroes;
int lhszvectornonzeroranges;
int governingdstype;
{
  zvectorstruct *rhszvectorptr[MAXARITY];
  zvectorstruct *tempvectorptr;
  zvectorstruct *sstempvectorptr;
  zbucketstruct *tempbucketptr;
  zbucketstruct *sstempbucketptr;
  zibucketstruct *tempibucketptr;
  zvbucketstruct *tempvbucketptr;
  int k;

  for (k=0;k<arity;k++)
    rhszvectorptr[k] = *rhszvectorptrptrs[k];

  /****************************************
   * Switch lhszvectorptr and *destptrptr. *
   ****************************************/

  lhszvectorptr->nonzeroes = lhszvectornonzeroes;
  
  if (governingdstype == CS)
    lhszvectorptr->zds.cs.nonzeroranges = lhszvectornonzeroranges;
  
  tempvectorptr = (*destptrptr);
  (*destptrptr) = lhszvectorptr;
  lhszvectorptr = tempvectorptr;

  if (governingdstype == SS) {
    
    if ((INITIALIZED_MAXINDEX &&
	 ((float) (*destptrptr)->nonzeroes / (float) GLOBAL_MAXINDEX) >=
	 SWITCH_NZ_RATIO)     ||
	((*destptrptr)->nonzeroes >= SWITCH_NZ_VALUE)) {
     
       /****************************************
       * Convert SS to CS.                    *
       ****************************************/
      
      sstempvectorptr = (*destptrptr);
      zconvertdstype(destptrptr,MAXARITY,CS);
	  
      sstempbucketptr = sstempvectorptr->zds.ss.bucketptr;
      if (sstempbucketptr != NULL) {
	while (sstempbucketptr->nextbucket != NULL)
	  sstempbucketptr = sstempbucketptr->nextbucket;
	sstempbucketptr->nextbucket = freedzbucketslist;
	RESET_freedzbucketslist(sstempvectorptr->zds.ss.bucketptr);
	sstempvectorptr->zds.ss.bucketptr = NULL;
      }
      holdingbinzvectorptr[MAXARITY] =  sstempvectorptr;
    }
  }
  
  /****************************************
   * If *destptrptr was not empty,        *
   * add its buckets to freedzbucketslist. *
   ****************************************/
  
  if (tempvectorptr->dstype == SS) {
    tempbucketptr = tempvectorptr->zds.ss.bucketptr;
    if (tempbucketptr != NULL) {
      while (tempbucketptr->nextbucket != NULL)
	tempbucketptr = tempbucketptr->nextbucket;
      tempbucketptr->nextbucket = freedzbucketslist;
      RESET_freedzbucketslist(tempvectorptr->zds.ss.bucketptr);
      tempvectorptr->zds.ss.bucketptr = NULL;
    }
  }
  else if (tempvectorptr->dstype == CS) {
    tempibucketptr = tempvectorptr->zds.cs.ibucketptr;
    if (tempibucketptr != NULL) {
      while (tempibucketptr->nextibucket != NULL)
	tempibucketptr = tempibucketptr->nextibucket;
      tempibucketptr->nextibucket = freedzibucketslist;
      RESET_freedzibucketslist(tempvectorptr->zds.cs.ibucketptr);
      tempvectorptr->zds.cs.ibucketptr = NULL;
    }
    tempvbucketptr = tempvectorptr->zds.cs.vbucketptr;
    if (tempvbucketptr != NULL) {
      while (tempvbucketptr->nextvbucket != NULL)
	tempvbucketptr = tempvbucketptr->nextvbucket;
      tempvbucketptr->nextvbucket = freedzvbucketslist;
      RESET_freedzvbucketslist(tempvectorptr->zds.cs.vbucketptr);
      tempvectorptr->zds.cs.vbucketptr = NULL;
    }
  }
  
  /*****************************************
   * If holdingbins were used, add         *
   * their buckets to freedzbucketslist, OR *
   * their ibuckets to freedzibucketslist & *
   * their vbuckets to freedzvbucketslist.  *
   *****************************************/
  for (k=0;k<arity;k++) {
    
    rhszvectorptr[k] = restorerhszvectorptr[k];
    
    if (holdingbinzvectorptr[k]->dstype == SS) {
      tempbucketptr = holdingbinzvectorptr[k]->zds.ss.bucketptr;
      if (tempbucketptr != NULL) {
#ifdef _DEBUG
	if (governingdstype == CS) {
	  printf("zspgx warning 7.4: Unexpcted non-NULL");
	  printf("tempbucketptr.\n");
	}
#endif
	while (tempbucketptr->nextbucket != NULL) {
#ifdef _DEBUG
	  if (governingdstype == SS) {
	    printf("zspgx warning 7.1: tempbucketptr should have");
	    printf(" no trailing nextbuckets\n");
	  }
#endif
	  tempbucketptr = tempbucketptr->nextbucket;
	}
	tempbucketptr->nextbucket = freedzbucketslist;
	RESET_freedzbucketslist(holdingbinzvectorptr[k]->zds.ss.bucketptr);
	holdingbinzvectorptr[k]->zds.ss.bucketptr = NULL;
      }
    }
    else if (holdingbinzvectorptr[k]->dstype == CS) {
      tempibucketptr = holdingbinzvectorptr[k]->zds.cs.ibucketptr;
      if (tempibucketptr != NULL) {
#ifdef _DEBUG
	if (governingdstype == SS) {
	  printf("zspgx warning 7.2: Unexpcted non-NULL");
	  printf(" ibucketptr.\n");
	}
#endif
	while (tempibucketptr->nextibucket != NULL)
	  tempibucketptr = tempibucketptr->nextibucket;
	tempibucketptr->nextibucket = freedzibucketslist;
	RESET_freedzibucketslist(holdingbinzvectorptr[k]->zds.cs.ibucketptr);
	holdingbinzvectorptr[k]->zds.cs.ibucketptr = NULL;
      }
      tempvbucketptr = holdingbinzvectorptr[k]->zds.cs.vbucketptr;
      if (tempvbucketptr != NULL) {
#ifdef _DEBUG
	if (governingdstype == SS) {
	  printf("zspgx warning 7.3: Unexpcted non-NULL");
	  printf(" vbucketptr.\n");
	}
#endif
	while (tempvbucketptr->nextvbucket != NULL)
	  tempvbucketptr = tempvbucketptr->nextvbucket;
	tempvbucketptr->nextvbucket = freedzvbucketslist;
	RESET_freedzvbucketslist(holdingbinzvectorptr[k]->zds.cs.vbucketptr);
	holdingbinzvectorptr[k]->zds.cs.vbucketptr = NULL;
      }
    }
  } /* matches    for (k=0;k<arity;k++) {    */
}








/****************************************************************************
 *                                                                          *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/

void zplusequal_axpy_postprocessing(destptrptr,
				   rhszvectorptrptr,
				   status_otherrhszvector,
				   restofplusequalvectorptrptr,
				   alphaptrs,
				   restorerhszvectorptr,
				   arity,
				   lhszvectornonzeroes,
				   lhszvectornonzeroranges,
				   governingdstype)
zvectorstruct **destptrptr;
zvectorstruct **rhszvectorptrptr;
int status_otherrhszvector;
zvectorstruct **restofplusequalvectorptrptr;
double_complex *alphaptrs[2];
zvectorstruct *restorerhszvectorptr[2+1];
int arity;
int lhszvectornonzeroes;
int lhszvectornonzeroranges;
int governingdstype;
{
  zvectorstruct *rhszvectorptr[2];
  zvectorstruct *tempvectorptr;
  zbucketstruct *tempbucketptr;
  zibucketstruct *tempibucketptr;
  zvbucketstruct *tempvbucketptr;

  int k;

  k = 1;
    rhszvectorptr[k] = *rhszvectorptrptr;

    /****************************************
     * Switch lhszvectorptr and *destptrptr. *
     ****************************************/

    lhszvectorptr->nonzeroes = lhszvectornonzeroes;
    lhszvectorptr->zds.cs.nonzeroranges = lhszvectornonzeroranges;
  

    /*********************************************************
     * If *restofplusequalvectorptrptr was not empty,        *
     * add its buckets to freedzbucketslist.                  *
     *********************************************************/
  
    tempvectorptr = (*restofplusequalvectorptrptr);
    tempibucketptr = tempvectorptr->zds.cs.ibucketptr;
    if (tempibucketptr != NULL) {
      while (tempibucketptr->nextibucket != NULL)
	tempibucketptr = tempibucketptr->nextibucket;
      tempibucketptr->nextibucket = freedzibucketslist;
      RESET_freedzibucketslist(tempvectorptr->zds.cs.ibucketptr);
      tempvectorptr->zds.cs.ibucketptr = NULL;
    }
    tempvbucketptr = tempvectorptr->zds.cs.vbucketptr;
    if (tempvbucketptr != NULL) {
      while (tempvbucketptr->nextvbucket != NULL)
	tempvbucketptr = tempvbucketptr->nextvbucket;
      tempvbucketptr->nextvbucket = freedzvbucketslist;
      RESET_freedzvbucketslist(tempvectorptr->zds.cs.vbucketptr);
      tempvectorptr->zds.cs.vbucketptr = NULL;
    }

   
  /*****************************************
   * If status_otherrhszvector is TEMP,     *
   * add its buckets to freedzbucketslist,  *
   * and reassign pointers.                *
   *****************************************/

  if (status_otherrhszvector == TEMP_OTHER_VECTOR) {

    tempvectorptr = tempzdestptr;
    tempibucketptr = tempvectorptr->zds.cs.ibucketptr;
    if (tempibucketptr != NULL) {
      while (tempibucketptr->nextibucket != NULL)
	tempibucketptr = tempibucketptr->nextibucket;
      tempibucketptr->nextibucket = freedzibucketslist;
      RESET_freedzibucketslist(tempvectorptr->zds.cs.ibucketptr);
      tempvectorptr->zds.cs.ibucketptr = NULL;
    }
    tempvbucketptr = tempvectorptr->zds.cs.vbucketptr;
    if (tempvbucketptr != NULL) {
      while (tempvbucketptr->nextvbucket != NULL)
	tempvbucketptr = tempvbucketptr->nextvbucket;
      tempvbucketptr->nextvbucket = freedzvbucketslist;
      RESET_freedzvbucketslist(tempvectorptr->zds.cs.vbucketptr);
      tempvectorptr->zds.cs.vbucketptr = NULL;
    }

  }
 
  /*****************************************
   * If holdingbins were used, add         *
   * their buckets to freedzbucketslist, OR *
   * their ibuckets to freedzibucketslist & *
   * their vbuckets to freedzvbucketslist.  *
   *****************************************/

  /*  We could not have put rhszvectorptr[0] in a holdingbin;
   *  but there may be something in it from the routine that called
   *  plusequal (in particular zspg2 (zspgx does it's own 
   *  holdingbin cleaning).
   */

  for (k=0;k<arity;k++) {

    rhszvectorptr[k] = restorerhszvectorptr[k];
    
    if (holdingbinzvectorptr[k]->dstype == SS) {
      tempbucketptr = holdingbinzvectorptr[k]->zds.ss.bucketptr;
      if (tempbucketptr != NULL) {
#ifdef _DEBUG
	if (governingdstype == CS) {
	  printf("zspgx warning 7.4: Unexpcted non-NULL");
	  printf("tempbucketptr.\n");
	}
#endif
	while (tempbucketptr->nextbucket != NULL) {
#ifdef _DEBUG
	  if (governingdstype == SS) {
	    printf("zspgx warning 7.1: tempbucketptr should have");
	    printf(" no trailing nextbuckets\n");
	  }
#endif
	  tempbucketptr = tempbucketptr->nextbucket;
	}
	tempbucketptr->nextbucket = freedzbucketslist;
	RESET_freedzbucketslist(holdingbinzvectorptr[k]->zds.ss.bucketptr);
	holdingbinzvectorptr[k]->zds.ss.bucketptr = NULL;
      }
    }
    else if (holdingbinzvectorptr[k]->dstype == CS) {
      tempibucketptr = holdingbinzvectorptr[k]->zds.cs.ibucketptr;
      if (tempibucketptr != NULL) {
#ifdef _DEBUG
	if (governingdstype == SS) {
	  printf("zspgx warning 7.2: Unexpcted non-NULL");
	  printf(" ibucketptr.\n");
	}
#endif
	while (tempibucketptr->nextibucket != NULL)
	  tempibucketptr = tempibucketptr->nextibucket;
	tempibucketptr->nextibucket = freedzibucketslist;
	RESET_freedzibucketslist(holdingbinzvectorptr[k]->zds.cs.ibucketptr);
	holdingbinzvectorptr[k]->zds.cs.ibucketptr = NULL;
      }
      tempvbucketptr = holdingbinzvectorptr[k]->zds.cs.vbucketptr;
      if (tempvbucketptr != NULL) {
#ifdef _DEBUG
	if (governingdstype == SS) {
	  printf("zspgx warning 7.3: Unexpcted non-NULL");
	  printf(" vbucketptr.\n");
	}
#endif
	while (tempvbucketptr->nextvbucket != NULL)
	  tempvbucketptr = tempvbucketptr->nextvbucket;
	tempvbucketptr->nextvbucket = freedzvbucketslist;
	RESET_freedzvbucketslist(holdingbinzvectorptr[k]->zds.cs.vbucketptr);
	holdingbinzvectorptr[k]->zds.cs.vbucketptr = NULL;
      }
    }
  }
}





/****************************************************************************
 *                                                                          *
 *  zdetermine_loop_bounds                                                   *
 *                                                                          *
 *     Based on how close lhsvbucketindex and rhsvbucketindex pointers are  *
 *     to the end of the vbucket, and based on the number of values left to *
 *     process in the axpy, zdetermine_loop_bounds figures out the various   *
 *     loop values.                                                         *
 *                                                                          *
 ****************************************************************************/


void zdetermine_loop_bounds(totalcurrentnonzerorange,
			   lhsvbucketindex,
			   rhsvbucketindex,
			   firstloopsize,
			   numberofdoubleforloopiterations,
			   firstiterationloopsize,
			   seconditerationloopsize,
			   remainder1loopsize,
			   remainder2loopsize
#ifdef _DEBUG
			   ,addressofunconsumedcurrentnonzerorange
#endif
			   )
int totalcurrentnonzerorange;
int lhsvbucketindex;
int rhsvbucketindex;
int *firstloopsize;
int *numberofdoubleforloopiterations;
int *firstiterationloopsize;
int *seconditerationloopsize;
int *remainder1loopsize;
int *remainder2loopsize;
#ifdef _DEBUG
int *addressofunconsumedcurrentnonzerorange;
#endif
{
  int temprange, tmpmod;

  *firstloopsize =
    min(((zVBUCKET_DIM - max(lhsvbucketindex,rhsvbucketindex))),
	totalcurrentnonzerorange);
  temprange = totalcurrentnonzerorange - *firstloopsize;
  *numberofdoubleforloopiterations = temprange / zVBUCKET_DIM;
  *firstiterationloopsize =
    min((zVBUCKET_DIM -
	 (*firstloopsize + min(lhsvbucketindex,rhsvbucketindex))),temprange);
  *seconditerationloopsize =
    min((zVBUCKET_DIM - *firstiterationloopsize),
	(temprange - *firstiterationloopsize));
  tmpmod = MY_MOD(temprange,zVBUCKET_DIM_minus1);
  *remainder1loopsize = min (tmpmod,*firstiterationloopsize);
  *remainder2loopsize = max ((tmpmod - (*firstiterationloopsize)),0);
  
#ifdef _DEBUG
  *addressofunconsumedcurrentnonzerorange =  totalcurrentnonzerorange;
#endif

}

