#define DOUBLE_COMPLEX
/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#include  <stdlib.h>
#include "zgaxpy_pd_macros_and_data_structures.h"
#include "dgaxpy_pd_macros_and_data_structures.h"

#include "dgaxpy_pd_global_variables.h"
#include "zgaxpy_pd_global_variables.h"

/*
 *  Function Prototyping done locally to grab "TO_pREc_" coorectly.
 */

#define _P(x) ()

void zspim _P((dvectorstruct **destptrptr,
			  zvectorstruct **oldvectorptrptr
			  ));

void zspcj _P((zvectorstruct **destptrptr,
		     zvectorstruct **oldvectorptrptr
		     ));

/****************************************************************************
 *                                                                          *
 *                                                                          *
 *                                                                          *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/

void zspim(destptrptr,oldvectorptrptr)
dvectorstruct **destptrptr;
zvectorstruct **oldvectorptrptr;
{
  dvectorstruct *newvectorptr;
  zvectorstruct *oldvectorptr;
  dbucketstruct *newvectortailbucketptr,*newbucketptr;
  zbucketstruct *oldvectorbucketptr;
  dibucketstruct *newvectortailibucketptr,*newibucketptr;
  zibucketstruct *oldvectoribucketptr;
  dvbucketstruct *newvectortailvbucketptr,*newvbucketptr;
  zvbucketstruct *oldvectorvbucketptr;
  int i,bucketindex,ibucketindex,vbucketindex;

  oldvectorptr = *oldvectorptrptr;

  if (*destptrptr == NULL)
     dgetvector(destptrptr);

  newvectorptr = lhsdvectorptr;

  newvectorptr->nonzeroes =  oldvectorptr->nonzeroes;
  newvectorptr->dstype =  oldvectorptr->dstype;

  if (newvectorptr->dstype == S1) {
    newvectorptr->dds.s1.index =  oldvectorptr->zds.s1.index;
    X_GET_AND_CAST_IMAG(newvectorptr->dds.s1.value,
			oldvectorptr->zds.s1.value,double);
  }
  else if (newvectorptr->dstype == SS) {
    oldvectorbucketptr = oldvectorptr->zds.ss.bucketptr;

    for (i=0;i<newvectorptr->nonzeroes;i++) {
      if ((bucketindex = MY_MOD(i,zBUCKET_DIM_minus1)) == 0) {
	if (i == 0) {
	  dgetbucket(&newbucketptr);
	  newvectorptr->dds.ss.bucketptr = newbucketptr;
	}
	else {
	  oldvectorbucketptr = oldvectorbucketptr->nextbucket;
	  dgetbucket(&newbucketptr);
	  newvectortailbucketptr->nextbucket = newbucketptr;
	}
	newvectortailbucketptr = newbucketptr;
      }
      newvectortailbucketptr->index[bucketindex] =
        oldvectorbucketptr->index[bucketindex];
      X_GET_AND_CAST_IMAG(newvectortailbucketptr->value[bucketindex],
			  oldvectorbucketptr->value[bucketindex],double);
    }
  }
  else /* if (newvectorptr->dstype == CS */ {
    oldvectoribucketptr = oldvectorptr->zds.cs.ibucketptr;
    oldvectorvbucketptr = oldvectorptr->zds.cs.vbucketptr;
    newvectorptr->dds.cs.nonzeroranges = oldvectorptr->zds.cs.nonzeroranges;
    
    for (i=0;i<newvectorptr->dds.cs.nonzeroranges;i++) {
      if ((ibucketindex = MY_MOD(i,zIBUCKET_DIM_minus1)) == 0) {
	if (i == 0) {
	  dgetibucket(&newibucketptr);
	  newvectorptr->dds.cs.ibucketptr = newibucketptr;
	}
	else {
	  oldvectoribucketptr = oldvectoribucketptr->nextibucket;
	  dgetibucket(&newibucketptr);
	  newvectortailibucketptr->nextibucket = newibucketptr;
	}
	newvectortailibucketptr = newibucketptr;
      }

      newvectortailibucketptr->index_RANGE_BEGIN[ibucketindex] =
        oldvectoribucketptr->index_RANGE_BEGIN[ibucketindex];
      newvectortailibucketptr->index_RANGE_END[ibucketindex] =
        oldvectoribucketptr->index_RANGE_END[ibucketindex];
    }
    
    for (i=0;i<newvectorptr->nonzeroes;i++) {
      if ((vbucketindex = MY_MOD(i,zVBUCKET_DIM_minus1)) == 0) {
	if (i == 0) {
	  dgetvbucket(&newvbucketptr);
	  newvectorptr->dds.cs.vbucketptr = newvbucketptr;
	}
	else {
	  oldvectorvbucketptr = oldvectorvbucketptr->nextvbucket;
	  dgetvbucket(&newvbucketptr);
	  newvectortailvbucketptr->nextvbucket = newvbucketptr;
	}
	newvectortailvbucketptr = newvbucketptr;
      }
      X_GET_AND_CAST_IMAG(newvectortailvbucketptr->value[vbucketindex],
			  oldvectorvbucketptr->value[vbucketindex],double);
    }
  }

  daxpy_0_1_post_processing(destptrptr);

}


/****************************************************************************
 *                                                                          *
 *                                                                          *
 *                                                                          *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/

void zspcj(destptrptr,oldvectorptrptr)
zvectorstruct **destptrptr;
zvectorstruct **oldvectorptrptr;
{
  zvectorstruct *newvectorptr;
  zvectorstruct *oldvectorptr;
  zbucketstruct *newvectortailbucketptr,*newbucketptr;
  zbucketstruct *oldvectorbucketptr;
  zibucketstruct *newvectortailibucketptr,*newibucketptr;
  zibucketstruct *oldvectoribucketptr;
  zvbucketstruct *newvectortailvbucketptr,*newvbucketptr;
  zvbucketstruct *oldvectorvbucketptr;
  int i,bucketindex,ibucketindex,vbucketindex;

  oldvectorptr = *oldvectorptrptr;

  if (*destptrptr == NULL)
     zgetvector(destptrptr);

  newvectorptr = lhszvectorptr;

  newvectorptr->nonzeroes =  oldvectorptr->nonzeroes;
  newvectorptr->dstype =  oldvectorptr->dstype;

  if (newvectorptr->dstype == S1) {
    newvectorptr->zds.s1.index =  oldvectorptr->zds.s1.index;
    X_CONJUGATE(newvectorptr->zds.s1.value,oldvectorptr->zds.s1.value);
  }
  else if (newvectorptr->dstype == SS) {
    oldvectorbucketptr = oldvectorptr->zds.ss.bucketptr;

    for (i=0;i<newvectorptr->nonzeroes;i++) {
      if ((bucketindex = MY_MOD(i,zBUCKET_DIM_minus1)) == 0) {
	if (i == 0) {
	  zgetbucket(&newbucketptr);
	  newvectorptr->zds.ss.bucketptr = newbucketptr;
	}
	else {
	  oldvectorbucketptr = oldvectorbucketptr->nextbucket;
	  zgetbucket(&newbucketptr);
	  newvectortailbucketptr->nextbucket = newbucketptr;
	}
	newvectortailbucketptr = newbucketptr;
      }
      newvectortailbucketptr->index[bucketindex] =
        oldvectorbucketptr->index[bucketindex];
      X_CONJUGATE(newvectortailbucketptr->value[bucketindex],
		  oldvectorbucketptr->value[bucketindex]);
    }
  }
  else /* if (newvectorptr->dstype == CS */ {
    oldvectoribucketptr = oldvectorptr->zds.cs.ibucketptr;
    oldvectorvbucketptr = oldvectorptr->zds.cs.vbucketptr;
    newvectorptr->zds.cs.nonzeroranges = oldvectorptr->zds.cs.nonzeroranges;
    
    for (i=0;i<newvectorptr->zds.cs.nonzeroranges;i++) {
      if ((ibucketindex = MY_MOD(i,zIBUCKET_DIM_minus1)) == 0) {
	if (i == 0) {
	  zgetibucket(&newibucketptr);
	  newvectorptr->zds.cs.ibucketptr = newibucketptr;
	}
	else {
	  oldvectoribucketptr = oldvectoribucketptr->nextibucket;
	  zgetibucket(&newibucketptr);
	  newvectortailibucketptr->nextibucket = newibucketptr;
	}
	newvectortailibucketptr = newibucketptr;
      }

      newvectortailibucketptr->index_RANGE_BEGIN[ibucketindex] =
        oldvectoribucketptr->index_RANGE_BEGIN[ibucketindex];
      newvectortailibucketptr->index_RANGE_END[ibucketindex] =
        oldvectoribucketptr->index_RANGE_END[ibucketindex];
    }
    
    for (i=0;i<newvectorptr->nonzeroes;i++) {
      if ((vbucketindex = MY_MOD(i,zVBUCKET_DIM_minus1)) == 0) {
	if (i == 0) {
	  zgetvbucket(&newvbucketptr);
	  newvectorptr->zds.cs.vbucketptr = newvbucketptr;
	}
	else {
	  oldvectorvbucketptr = oldvectorvbucketptr->nextvbucket;
	  zgetvbucket(&newvbucketptr);
	  newvectortailvbucketptr->nextvbucket = newvbucketptr;
	}
	newvectortailvbucketptr = newvbucketptr;
      }
      X_CONJUGATE(newvectortailvbucketptr->value[vbucketindex],
		  oldvectorvbucketptr->value[vbucketindex]);
    }
  }

  zaxpy_0_1_post_processing(destptrptr);

}
