#define DOUBLE_COMPLEX
/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#include  <stdlib.h>
#include "zgaxpy_pd_macros_and_data_structures.h"
#include "zgaxpy_pd_global_variables.h"


/****************************************************************************
 *                                                                          *
 *  zadallocate()                                                            *
 *                                                                          *
 *      zadallocate calls the bucket and vector store allocation functions.  *
 *      It should be called by the user once, before any calls to zspsd &  *
 *      adxpyNs.                                                            *
 *                                                                          *
 ****************************************************************************/

void zadallocate()
{
  int k;
  
  zinitialize_pd_global_variables();
  zallocatevectorstore();
  zallocatebucketstore();
  zallocateibucketstore();
  zallocatevbucketstore();

  zgetvector(&lhszvectorptr);
  for (k=0;k<MAXARITY+1;k++)
    zgetvector(&holdingbinzvectorptr[k]);
  zgetvector(&tempzdestptr);
  zgetvector(&restofplusequalzvectorptr);
  restofplusequalzvectorptr->dstype = CS;
}



/****************************************************************************
 *                                                                          *
 *   zadfree()                                                               *
 *                                                                          *
 *      zadfree frees all the dynamically allocated space for a given        *
 *      percision.                                                          *
 *                                                                          *
 ****************************************************************************/

void zadfree()
{
  zvectorstorestruct  *remaindervectorstoreptr,  *tempvectorstoreptr;
  zbucketstorestruct  *remainderbucketstoreptr,  *tempbucketstoreptr;
  zibucketstorestruct *remainderibucketstoreptr, *tempibucketstoreptr;
  zvbucketstorestruct *remaindervbucketstoreptr, *tempvbucketstoreptr;

  zvectorstruct *tempvectorptr;
  zbucketstruct *tempbucketptr;
  zibucketstruct *tempibucketptr;
  zvbucketstruct *tempvbucketptr;

  int i;

  for(remaindervectorstoreptr=zvectorstorestartptr;
      remaindervectorstoreptr!=NULL;
      remaindervectorstoreptr=tempvectorstoreptr) {

    tempvectorstoreptr = remaindervectorstoreptr->nextstore;
    free(remaindervectorstoreptr);
  }

  for(remainderbucketstoreptr=zbucketstorestartptr;
      remainderbucketstoreptr!=NULL;
      remainderbucketstoreptr=tempbucketstoreptr) {
    tempbucketstoreptr = remainderbucketstoreptr->nextstore;

    /*
     * Freeing based on per BUCKET Dynamic Allocation.
     * Here we free the memory which has been allocated per bucket
     * in zallocatebucketstore().
     */
    for (i=0; i<zBUCKETSTORE_DIM; i++) {
      tempbucketptr = &(remainderbucketstoreptr->bucket[i]);
      free(tempbucketptr->index);
      free(tempbucketptr->value);
    }
  
    /*
     * Now regardless of the allocation method above, we have freed
     * everything hanging off our bucketstore, and we can go ahead and
     * free the bucketstore itself.
     */
    free(remainderbucketstoreptr);
    
  }
  
  for(remainderibucketstoreptr=zibucketstorestartptr;
      remainderibucketstoreptr!=NULL;
      remainderibucketstoreptr=tempibucketstoreptr) {
    tempibucketstoreptr = remainderibucketstoreptr->nextstore;
    
    /*
     * Freeing based on per IBUCKET Dynamic Allocation.
     */
    for (i=0; i<zIBUCKETSTORE_DIM; i++) {
      tempibucketptr = &(remainderibucketstoreptr->ibucket[i]);
      free(tempibucketptr->index_RANGE_BEGIN);
      free(tempibucketptr->index_RANGE_END);
    }
    
    /*
     * Now regardless of the allocation method above, we have freed
     * everything hanging off our ibucketstore, and we can go ahead and
     * free the ibucketstore itself.
     */
    free(remainderibucketstoreptr);
  }
  
  for(remaindervbucketstoreptr=zvbucketstorestartptr;
      remaindervbucketstoreptr!=NULL;
      remaindervbucketstoreptr=tempvbucketstoreptr) {
    tempvbucketstoreptr = remaindervbucketstoreptr->nextstore;
    
    /*
     * Freeing based on per VBUCKET Dynamic Allocation.
     */
    for (i=0; i<zVBUCKETSTORE_DIM; i++) {
      tempvbucketptr = &(remaindervbucketstoreptr->vbucket[i]);
      free(tempvbucketptr->value);
    }
    
    /*
     * Now regardless of the allocation method above, we have freed
     * everything hanging off our vbucketstore, and we can go ahead and
     * free the vbucketstore itself.
     */
    free(remaindervbucketstoreptr);
  }
}



/****************************************************************************
 *                                                                          *
 *   zad_free_vectors_buckets(zvectorptrptr)                             *
 *                                                                          *
 *       zad_free_vectors_buckets frees the buckets of zvectorptrptr.    *
 *       BUT IT DOES NOT free zvectorptrptr, itself.                    *
 *       Note that freedzbucketslist, freedzibucketslist, and       *
 *       freedzvbucketslist are global variables included in            *
 *       $(SAXPY_FLAVOR)_global_variables.h.                                *
 *                                                                          *
 *       zvectorptrptr  : The vector to be freed.                       *
 *                                                                          *
 ****************************************************************************/

void zad_free_vectors_buckets(zvectorptrptr)
zvectorstruct **zvectorptrptr;
{

  zvectorstruct *zvectorptr;
  zbucketstruct *zbucketptr;
  zibucketstruct *zibucketptr;
  zvbucketstruct *zvbucketptr;

  zvectorptr = (zvectorstruct *) *zvectorptrptr;
  
  if (zvectorptr != NULL) {
    if (zvectorptr->dstype == S1) {
      /* do nothing */
    }
    else if (zvectorptr->dstype == SS) {
      /**********************************************************************
       *                        Single Subscript                            *
       **********************************************************************/
      
      zbucketptr = zvectorptr->zds.ss.bucketptr;
      if (zbucketptr != NULL) {
	while (zbucketptr->nextbucket != NULL)
	  zbucketptr = zbucketptr->nextbucket;
	zbucketptr->nextbucket = freedzbucketslist;
	RESET_freedzbucketslist(zvectorptr->zds.ss.bucketptr);
	zvectorptr->zds.ss.bucketptr = NULL;
      }
      
    } else /* zvectorptr->zdstype = CS */ {
      /**********************************************************************
       *                      Compressed Subscript                          *
       **********************************************************************/
      
      zibucketptr = zvectorptr->zds.cs.ibucketptr;
      if (zibucketptr != NULL) {
	while (zibucketptr->nextibucket != NULL)
	  zibucketptr = zibucketptr->nextibucket;
	zibucketptr->nextibucket = freedzibucketslist;
	RESET_freedzibucketslist(zvectorptr->zds.cs.ibucketptr);
	zvectorptr->zds.cs.ibucketptr = NULL;
      }
      zvbucketptr = zvectorptr->zds.cs.vbucketptr;
      if (zvbucketptr != NULL) {
	while (zvbucketptr->nextvbucket != NULL)
	  zvbucketptr = zvbucketptr->nextvbucket;
	zvbucketptr->nextvbucket = freedzvbucketslist;
	RESET_freedzvbucketslist(zvectorptr->zds.cs.vbucketptr);
	zvectorptr->zds.cs.vbucketptr = NULL;
      }
    }
  }

/*
 * PMK: begin 12/4/96 edits
 *
 * In order to ensure that zvectorptr points to a null vectorstruct,
 * i.e., a vectorstruct whose fields all have zero values, we assign to it
 * the contents of nullzvector which is a global initialized with all
 * zero members in gaxpy_pd_global_variables.h.
 *
 */

  *(zvectorptr) = nullzvector;

/*
 * PMK: end 12/4/96 edits
 */

}


/****************************************************************************
 *                                                                          *
 *    zinitialize_pd_global_variables()                                      *
 *                                                                          *
 *    zinitialize_pd_global_variables performs initializations for some      *
 *    precision dependent globals.                                          *
 *                                                                          *
 ****************************************************************************/

void zinitialize_pd_global_variables()
{
  zbucketstorestartptr=NULL;
  freedzbucketslist=NULL;
  zibucketstorestartptr=NULL;
  freedzibucketslist=NULL;
  zvbucketstorestartptr=NULL;
  freedzvbucketslist=NULL;
  zvectorstorestartptr=NULL;

}

