#define DOUBLE_COMPLEX
/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#include  <stdlib.h>
#include "zgaxpy_pd_macros_and_data_structures.h"
#include "zgaxpy_pd_global_variables.h"




/****************************************************************************
 *                                                                          *
 *  zdebugging_code_1(RIGHT_OR_LEFT,                                         *
 *                   vbucketindex,                                          *
 *                   unconsumedcurrentnonzerorange)                         *
 *                                                                          *
 *       zdebugging_code_1 prints error messages several quantities expected *
 *       to be zero at some point of the computation are not zero;          *
 *       it then aborts.                                                    *
 *                                                                          *
 *       RIGHT_OR_LEFT                  : "lhs" or "rhs"                    *
 *       vbucketindex                   : nsia                              *
 *       unconsumedcurrentnonzerorange  : nsia                              *
 *                                                                          *
 ****************************************************************************/

void zdebugging_code_1(RIGHT_OR_LEFT,
		      vbucketindex,
		      unconsumedcurrentnonzerorange)
char *RIGHT_OR_LEFT;
int vbucketindex;
int unconsumedcurrentnonzerorange;
{
  if (unconsumedcurrentnonzerorange == 0 ||
      (vbucketindex != 0 &&
       vbucketindex != zVBUCKET_DIM)) {	
    printf("zdebugging_code_1 error: Quantities should be zero or ");
    printf("equal to zVBUCKET_DIM\n\t");
    printf("%3svbucketindex=%d\n ",
	   RIGHT_OR_LEFT,vbucketindex);
    abort();
  }
}

#ifdef _PEEP



/****************************************************************************
 *                                                                          *
 *  zadpeep_pre(index,rhsvectorptrptr,alphaptr)                              *
 *                                                                          *
 *       zadpeep_pre prints an alpha and vector value pair on the            *
 *       right-hand-side of an axpy expression.  Printing is in the         *
 *       following format:                                                  *
 *                                                                          *
 *       alpha 0:                                                           *
 *           ...                                                            *
 *       vector 0:                                                          *
 *           index  value                                                   *
 *           index  value                                                   *
 *           ...    ...                                                     *
 *                                                                          *
 ****************************************************************************/

void zadpeep_pre (vectorindex,rhsvectorptrptr,alphaptr)
int vectorindex;
zvectorstruct **rhsvectorptrptr;
double_complex *alphaptr;
{

  zbucketstruct *rhsbucketptr;
                                      /* Pointer to the "current" bucket    *
				       * where the "current" nonzeroes of   *
				       * the rhs vector reside.             */

  int rhsbucketindex;                 /* Index of the current nonzero in    *
				       * the current bucket of the rhs      *
				       * vector.                            */

  zvbucketstruct *rhsvbucketptr;
                                      /* Pointer to "current" vbucket where *
				       * the "current" nonzero of the rhs   *
				       * vector reside.                     */

  int rhsvbucketindex;                /* Index of current nonzero in the    *
				       * current bucket of the rhs vector   */

  zibucketstruct *rhsibucketptr;
                                      /* Pointer to "current" ibucket where *
				       * the "current" nonzero of the rhs   *
				       * vector resides.                    */

  int rhsibucketindex;                /* Index of the current nonzero in    *
				       * the current bucket of the rhs      *
				       * vector.                            */

  int j;                              /* Index for vectorsubscripts to be   *
				       * entered in (and later popped from) *
				       * the heap.                          */

  int i;

  int h;


  if (fileopenflag) {
    fp = fopen("gaxpy_adxpy.dump","w");     
    fileopenflag = 0;
  }

#ifdef _VERBOSE_DUMP
  
  fprintf(fp,"vector # %d:\t(alpha # %d = %16.9e)\n",
	  vectorindex,vectorindex,*alphaptr);
#ifdef _SCREEN_PEEP
  printf("vector # %d:\t(alpha # %d = %16.9e)\n",
	 vectorindex,vectorindex,*alphaptr);
#endif

  if ((rhsvectorptrptr == NULL) || (*rhsvectorptrptr == NULL)) {
    fprintf(fp,"\t null-vector\n");
#ifdef _SCREEN_PEEP
    printf("\t null-vector\n");
#endif
  }
  else {

    if ((*rhsvectorptrptr)->nonzeroes == 0) {
      fprintf(fp,"\t zero-vector\n");
#ifdef _SCREEN_PEEP
      printf("\t zero-vector\n");
#endif
    }
    else {
      if ((*rhsvectorptrptr)->dstype == S1) {
	fprintf(fp,"\t%d  %16.9e\n", (*rhsvectorptrptr)->zds.s1.index,
		(*rhsvectorptrptr)->zds.s1.value);
#ifdef _SCREEN_PEEP
	printf("\t%d  %16.9e\n", (*rhsvectorptrptr)->zds.s1.index,
	       (*rhsvectorptrptr)->zds.s1.value);
#endif
      }
      else if ((*rhsvectorptrptr)->dstype == SS) {
	rhsbucketptr = (*rhsvectorptrptr)->zds.ss.bucketptr;
	for(i=0;i<(*rhsvectorptrptr)->nonzeroes;i++) {
	  if(((rhsbucketindex=MY_MOD(i,zBUCKET_DIM_minus1)) == 0) && i!=0)
	    rhsbucketptr = rhsbucketptr->nextbucket;
	  fprintf(fp,"\t%d  %16.9e\n",rhsbucketptr->index[rhsbucketindex],
		  rhsbucketptr->value[rhsbucketindex]);
#ifdef _SCREEN_PEEP
	  printf("\t%d  %16.9e\n",rhsbucketptr->index[rhsbucketindex],
		 rhsbucketptr->value[rhsbucketindex]);
#endif
	}
      }
      else /* if ((*rhsvectorptrptr)->dstype == CS) */ {
	rhsibucketptr = (*rhsvectorptrptr)->zds.cs.ibucketptr;
	rhsvbucketptr = (*rhsvectorptrptr)->zds.cs.vbucketptr;    
	rhsvbucketindex = 0;    
	for(i=0,h=0;i< (*rhsvectorptrptr)->zds.cs.nonzeroranges;i++) {
	  if(((rhsibucketindex =
	       MY_MOD(i,zIBUCKET_DIM_minus1)) == 0) && i!=0)
	    rhsibucketptr = rhsibucketptr->nextibucket;
	  for(j=0;
	      j<=(rhsibucketptr->index_RANGE_END[rhsibucketindex] - 
		  rhsibucketptr->index_RANGE_BEGIN[rhsibucketindex]);
	      h++,j++,rhsvbucketindex++) {
	    if(((rhsvbucketindex =
		 MY_MOD(rhsvbucketindex,zVBUCKET_DIM_minus1)) == 0)
	       && h!=0)
	      rhsvbucketptr = rhsvbucketptr->nextvbucket;
	    fprintf(fp,"\t%d  %16.9e\n",
		   rhsibucketptr->index_RANGE_BEGIN[rhsibucketindex]+j,
		   rhsvbucketptr->value[rhsvbucketindex]);
#ifdef _SCREEN_PEEP
	    printf("\t%d  %16.9e\n",
		   rhsibucketptr->index_RANGE_BEGIN[rhsibucketindex]+j,
		   rhsvbucketptr->value[rhsvbucketindex]);
#endif
	  }
	}
      }
    }
  }

#endif /* _VERBOSE_DUMP */

}
  


/****************************************************************************
 *                                                                          *
 *  zadpeep_post(destptrptr)                                                 *
 *                                                                          *
 *       zadpeep_post prints the destination vector on the left-hand side    *
 *       of the axpy operation.  zadpeep_post is called after the axpy is    *
 *       performed.                                                         *
 *                                                                          *
 ****************************************************************************/

void zadpeep_post (destptrptr)
zvectorstruct **destptrptr;

{

  zbucketstruct *lhszbucketptr; /* Pointer to current bucket of     *
				       * lhszvector.                    */

  int lhszbucketindex;            /* bucketindex of next nonzero to be  *
				       * added to lhszvector.           */

  zvbucketstruct *lhszvbucketptr;
                                      /* Pointer to current vbucket of      *
				       * lhszvector.                    */

  int lhszvbucketindex;           /* bucketindex of next nonzero to be  *
				       * added to lhszvector.           */

  zibucketstruct *lhszibucketptr; /* Pointer to current ibucket of  *
				       * lhszvector.                    */

  int lhszibucketindex;           /* bucketindex of next nonzero to be  *
				       * added to lhszvector.           */

  int i, j, h;


  if (fileopenflag) {
    fp = fopen("gaxpy_adxpy.dump","w");     
    fileopenflag = 0;
  }

#ifdef _VERBOSE_DUMP
  fprintf(fp,"destptrptr\n");
#endif

#ifdef _SCREEN_PEEP
  printf("destptrptr\n");
#endif
  if ((*destptrptr)->nonzeroes == 0) {

#ifdef _VERBOSE_DUMP
    fprintf(fp,"\t zero-vector\n");
#endif

#ifdef _SCREEN_PEEP
    printf("\t zero-vector\n");
#endif
  }
  else {
    if ((*destptrptr)->dstype == S1) {
      fprintf(fp,"\t%d  %16.9e\n", (*destptrptr)->zds.s1.index,
	      (*destptrptr)->zds.s1.value);
#ifdef _SCREEN_PEEP
      printf("\t%d  %16.9e\n", (*destptrptr)->zds.s1.index,
	     (*destptrptr)->zds.s1.value);
#endif
    }
    else if ((*destptrptr)->dstype == SS) {
      lhszbucketptr = (*destptrptr)->zds.ss.bucketptr;
      for(i=0;i<(*destptrptr)->nonzeroes;i++) {
	if(((lhszbucketindex =
	     MY_MOD(i,zBUCKET_DIM_minus1)) == 0) && i!=0)
	  lhszbucketptr = lhszbucketptr->nextbucket;
	fprintf(fp,"\t%d  %16.9e\n",lhszbucketptr->index[lhszbucketindex],
		lhszbucketptr->value[lhszbucketindex]);
#ifdef _SCREEN_PEEP
	printf("\t%d  %16.9e\n",lhszbucketptr->index[lhszbucketindex],
	       lhszbucketptr->value[lhszbucketindex]);
#endif
      }
    }
    else /* if ((*destptrptr)->dstype == CS) */ {    
      lhszibucketptr = (*destptrptr)->zds.cs.ibucketptr;
      lhszvbucketptr = (*destptrptr)->zds.cs.vbucketptr;
      lhszvbucketindex = 0;
      for(i=0,h=0;i<(*destptrptr)->zds.cs.nonzeroranges;i++) {
	if(((lhszibucketindex =
	     MY_MOD(i,zIBUCKET_DIM_minus1)) == 0) && i!=0)
	  lhszibucketptr = lhszibucketptr->nextibucket;
	for(j=0; j <= (lhszibucketptr->index_RANGE_END[lhszibucketindex] - 
		       lhszibucketptr->index_RANGE_BEGIN[lhszibucketindex]);
	    h++,j++,lhszvbucketindex++) {
	  if(((lhszvbucketindex =
	       MY_MOD(lhszvbucketindex,zVBUCKET_DIM_minus1)) == 0) &&
	     h!=0)
	    lhszvbucketptr = lhszvbucketptr->nextvbucket;
	  fprintf(fp,"\t%d  %16.9e\n",
		  lhszibucketptr->index_RANGE_BEGIN[lhszibucketindex]+j,
		  lhszvbucketptr->value[lhszvbucketindex]);
#ifdef _SCREEN_PEEP
	  printf("\t%d  %16.9e\n",
		 lhszibucketptr->index_RANGE_BEGIN[lhszibucketindex]+j,
		 lhszvbucketptr->value[lhszvbucketindex]);
#endif
	}
      }
    }
  }
}
 


#endif




/****************************************************************************
 *                                                                          *
 *  zadinfo(arity,nnzb,lhszvectoraddress,nnza,alphaptrs,vectorptrptrs)   *
 *                                                                          *
 *        This routine dumps out information for the Perl analyser.         *  
 *                                                                          *
 *        It was written by Aaron Ross in 1993 and is now obsoloete.        *
 *                                                                          *


void zadinfo(arity,nnzb,lhszvectoraddress,nnza,alphaptrs,vectorptrptrs)
int arity,nnzb,nnza;
double_complex *alphaptrs[];
zvectorstruct **vectorptrptrs[];
zvectorstruct *lhszvectoraddress;
{
  int k;

  printf("<sxpy>  ");  **  This is for perl to filter out non-important
			   lines                                        **
  
  printf("  %d  %d %u        %d    ",arity,nnzb, lhszvectoraddress, nnza);
  if(arity<=0) printf(" \n");
  if(arity>0) {
    for(k=0;k<=arity-2;k++)
      printf(" %10.3f   %u        %d    ",
	     *alphaptrs[k],*vectorptrptrs[k],(*vectorptrptrs[k])->nonzeroes);
    printf(" %10.3f   %u        %d    \n",
	   *alphaptrs[arity-1],*vectorptrptrs[arity-1],
	   (*vectorptrptrs[arity-1])->nonzeroes);    
  }
}

 *                                                                          *
 ****************************************************************************/






