/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#ifndef zGAXPY_PD_MACROS_AND_DATA_STRUCTURES

#define zGAXPY_PD_MACROS_AND_DATA_STRUCTURES

#include "gaxpy_pi_macros_and_data_structures.h"


/***************************************************************************
 *                                                                         *
 *  Macro definitions for store dimensions                                 *
 *                                                                         *
 ***************************************************************************/

#define zBUCKETSTORE_DIM  128      /* # of buckets in a bucketstore.   */

#define zVBUCKETSTORE_DIM 128      /* # of vbuckets in a vbucketstore. */

#define zIBUCKETSTORE_DIM 128      /* # of ibuckets in a ibucketstore. */

#define zVECTORSTORE_DIM  128      /* # of vectors in a vectorstore.   */


/***************************************************************************
 *                                                                         *
 *  Macro definitions for subroutines;  DEBUG ON                           *
 *                                                                         *
 ***************************************************************************/

#ifdef _DEBUG

#define zHOP_TO_NEXT_RHSVBUCKET(a,b,c,d) \
        zhop_to_next_rhsvbucket(a,b,c,d)

#define zHOP_TO_NEXT_LHSVBUCKET(a,b) \
        zhop_to_next_lhsvbucket(a,b)

#define zNON_OVER_LAPPING_CS_RANGE_AXPY(a,b,c,d,e,f,g,h,i,j,k,l) \
        znon_over_lapping_cs_range_axpy(a,b,c,d,e,f,g,h,i,j,k,l)

#define zOVERLAPPING_CS_RANGE_AXPY(a,b,c,d,e,f,g,h,i) \
        zoverlapping_cs_range_axpy(a,b,c,d,e,f,g,h,i)

#define zDENSE_PLUSEQUAL_AXPY_LOOP(a,b,c,d,e) \
        zdense_plusequal_axpy_loop(a,b,c,d,e)

#define zNON_OVER_LAPPING_CS_RANGE_CONVERT_TO_c(a,b,c,d,e,f,g,h,i,j,k) \
        znon_over_lapping_cs_range_convert_to_c(a,b,c,d,e,f,g,h,i,j,k)

#define zOVERLAPPING_CS_RANGE_CONVERT_TO_c(a,b,c,d,e,f,g,h) \
        zoverlapping_cs_range_convert_to_c(a,b,c,d,e,f,g,h)

#define zDENSE_CONVERT_TO_c_LOOP(a,b,c,d,e,f) \
        zdense_convert_to_c_loop(a,b,c,d,e,f)

#define zNON_OVER_LAPPING_CS_RANGE_CONVERT_TO_z(a,b,c,d,e,f,g,h,i,j,k) \
        znon_over_lapping_cs_range_convert_to_z(a,b,c,d,e,f,g,h,i,j,k)

#define zOVERLAPPING_CS_RANGE_CONVERT_TO_z(a,b,c,d,e,f,g,h) \
        zoverlapping_cs_range_convert_to_z(a,b,c,d,e,f,g,h)

#define zDENSE_CONVERT_TO_z_LOOP(a,b,c,d,e,f) \
        zdense_convert_to_z_loop(a,b,c,d,e,f)

#define zDETERMINE_LOOP_BOUNDS(a,b,c,d,e,f,g,h,i,j) \
        zdetermine_loop_bounds(a,b,c,d,e,f,g,h,i,j)

#define zDEBUGGING_CODE_1(a,b,c) \
        zdebugging_code_1(a,b,c)

/*
 *  In-lining zdense_axpy_loop
 */

#define zDENSE_AXPY_LOOP(range,lhszvbucketindex,rhszvbucketindex,k,loopsize,alpha,addressofunconsumedcurrentnonzerorange) \
{ \
  int j; \
  if (range == OLD_RANGE) { \
    for (j=0;j<loopsize;j++) { \
      GAXPY_PLUS_EQUALS(lhszvbucketptr->value[(*lhszvbucketindex)+j], \
	      alpha,rhszvbucketptr[k]->value[(*rhszvbucketindex)+j]); \
    } \
  } \
  else { \
    for (j=0;j<loopsize;j++) { \
      GAXPY_EQUALS(lhszvbucketptr->value[(*lhszvbucketindex)+j], \
	      alpha,rhszvbucketptr[k]->value[(*rhszvbucketindex)+j]); \
    } \
  } \
  (*lhszvbucketindex) += loopsize; \
  (*rhszvbucketindex) += loopsize; \
  (*addressofunconsumedcurrentnonzerorange) -= loopsize; \
}


/***************************************************************************
 *                                                                         *
 *  Macro definitions for subroutines;  DEBUG OFF                          *
 *                                                                         *
 ***************************************************************************/

#else

#define zHOP_TO_NEXT_RHSVBUCKET(a,b,c,d) \
        zhop_to_next_rhsvbucket(a,b)

#define zHOP_TO_NEXT_LHSVBUCKET(a,b) \
        zhop_to_next_lhsvbucket(a)

#define zNON_OVER_LAPPING_CS_RANGE_AXPY(a,b,c,d,e,f,g,h,i,j,k,l) \
        znon_over_lapping_cs_range_axpy(a,b,c,d,e,f,g,h)

#define zOVERLAPPING_CS_RANGE_AXPY(a,b,c,d,e,f,g,h,i) \
        zoverlapping_cs_range_axpy(a,b,c,d,e)

#define zDENSE_PLUSEQUAL_AXPY_LOOP(a,b,c,d,e) \
        zdense_plusequal_axpy_loop(a,b,c,d)

#define zNON_OVER_LAPPING_CS_RANGE_CONVERT_TO_c(a,b,c,d,e,f,g,h,i,j,k) \
        znon_over_lapping_cs_range_convert_to_c(a,b,c,d,e,f,g)

#define zOVERLAPPING_CS_RANGE_CONVERT_TO_c(a,b,c,d,e,f,g,h) \
        zoverlapping_cs_range_convert_to_c(a,b,c,d)

#define zDENSE_CONVERT_TO_c_LOOP(a,b,c,d,e,f) \
        zdense_convert_to_c_loop(a,b,c,d,e)

#define zNON_OVER_LAPPING_CS_RANGE_CONVERT_TO_z(a,b,c,d,e,f,g,h,i,j,k) \
        znon_over_lapping_cs_range_convert_to_z(a,b,c,d,e,f,g)

#define zOVERLAPPING_CS_RANGE_CONVERT_TO_z(a,b,c,d,e,f,g,h) \
        zoverlapping_cs_range_convert_to_z(a,b,c,d)

#define zDENSE_CONVERT_TO_z_LOOP(a,b,c,d,e,f) \
        zdense_convert_to_z_loop(a,b,c,d,e)

#define zDETERMINE_LOOP_BOUNDS(a,b,c,d,e,f,g,h,i,j) \
        zdetermine_loop_bounds(a,b,c,d,e,f,g,h,i)

#define zDEBUGGING_CODE_1(a,b,c)

/*
 *  In-lining zdense_axpy_loop
 */

#define zDENSE_AXPY_LOOP(range,lhszvbucketindex,rhszvbucketindex,k,loopsize,alpha,addressofunconsumedcurrentnonzerorange) \
{ \
  int j; \
  if (range == OLD_RANGE) { \
    for (j=0;j<loopsize;j++) { \
      GAXPY_PLUS_EQUALS(lhszvbucketptr->value[(*lhszvbucketindex)+j], \
	      alpha,rhszvbucketptr[k]->value[(*rhszvbucketindex)+j]); \
    } \
  } \
  else { \
    for (j=0;j<loopsize;j++) { \
      GAXPY_EQUALS(lhszvbucketptr->value[(*lhszvbucketindex)+j], \
	      alpha,rhszvbucketptr[k]->value[(*rhszvbucketindex)+j]); \
    } \
  } \
  (*lhszvbucketindex) += loopsize; \
  (*rhszvbucketindex) += loopsize; \
}

#endif


/***************************************************************************
 *                                                                         *
 *  Macro definitions for subroutines with the same interface for          *
 * _DEBUG.                                                                 *
 *                                                                         *
 ***************************************************************************/

#define RESET_freedzbucketslist(new_head) \
{ freedzbucketslist = new_head; \
}

#define RESET_freedzibucketslist(new_head) \
{ freedzibucketslist = new_head; \
}

#define RESET_freedzvbucketslist(new_head) \
{ freedzvbucketslist = new_head; \
}



/****************************************************************************
 *                                                                          *
 *  Data Structures                                                         *
 *                                                                          *
 *     Two abstract structures need to be represented: sparse vector and    *
 *     heap.  We will first describe these for _SINGLE_SUBSCRIPTS and then  *
 *     point out the differences that exist between these data structures   *
 *     and an analog set of data structures used for _COMPRESSED_SUBSCRIPTS.*
 *                                                                          *
 *     To represent sparse vectors using _SINGLE_SUBSCRIPTS, we define 4    *
 *     structs:                                                             *
 *                                                                          *
 *                  zbucketstruct contains a max # of pairs of indices  *
 *                                    and the corresponding values of the   *
 *                                    of the nonzeroes in a vector, and a   *
 *                                    pointer to the next bucket if there   *
 *                                    are more nonzeroes in the vector. The *
 *                                    indices are always in ascending order *
 *                                    by virtue of the way they are input   *
 *                                    in zspsd and the fact that the       *
 *                                    heapsaxpy operations construct new    *
 *                                    vectors with the same ordering.       *
 *                                                                          *
 *                  zvectorstruct is a sparse representation of a       *
 *                                    vector.  It has a field for the total *
 *                                    # of nonzeroes in the vector and a    *
 *                                    linked list of buckets which contain  *
 *                                    the indices and values of the         *
 *                                    nonzeroes in the vector.              *
 *                                                                          *
 *                  zbucketstorestruct is a long array of buckets, and a*
 *                                    pointer to the next store, if needed. *
 *                                    The basic idea is to allocate space   *
 *                                    in large blocks so that dynamic       *
 *                                    memory allocations will be kept to a  *
 *                                    minimum, yet vectors will be able to  *
 *                                    "grow as they go" by having access to *
 *                                    a memory store of allocated buckets.  *
 *                                                                          *
 *                  zvectorstorestruct is a long array of vectors, and a*
 *                                    pointer to the next store, if needed. *
 *                                    The basic idea is the same as in      *
 *                                    zbucketstorestruct.               *
 *                                                                          *
 *     To represent a heap, we define 1 struct:                             *
 *                                                                          *
 *                  heapstruct        represents a node in the binary heap  *
 *                                    tree.  Each node of the heap is a     *
 *                                    placeholder for a nonzero vector      *
 *                                    element.  The nonzero itself is not   *
 *                                    stored in the heap; instead its       *
 *                                    vector index (order of its vector in  *
 *                                    the rhs of the saxpy),  and its       *
 *                                    subscript index (its array index in   *
 *                                    the non-sparse vector) are placed in  *
 *                                    the heap node.  Additionally, each    *
 *                                    heap node contains two pointer fields *
 *                                    containing addresses of the left and  *
 *                                    right children of the node.           *
 *                                    The subscript index serves as the key *
 *                                    for ordering the heap: if x and p(x)  *
 *                                    are a node and its parent, then       *
 *                                    key_p(x) <= key_x.  It follows that   *
 *                                    key_root is the smallest key in the   *
 *                                    tree at all times, and thus by always *
 *                                    poping the root and "consuming" it in *
 *                                    the saxpy, the lhs vector will be     *
 *                                    constructed in the correct (i.e.,     *
 *                                    ascending) order of subscripts.       *
 *                                                                          *
 *     Each typedef below is accompanied by a pictorial depiction.          *
 *                                                                          *
 *     A Note on C syntax: Several of the data structures are defined with  *
 *     the following template:                                              *
 *                                                                          *
 *          typedef struct _Astruct {                                       *
 *             ...                                                          *
 *             struct _Astruct *nextfield;                                  *
 *          } Astruct;                                                      *
 *                                                                          *
 *     _Astruct is the "structure tag".  It's declared explicitly so that   *
 *     the pointer field inside would recognize what it's pointing to as a  *
 *     defined structure.  Then, typedef defines a synonym, "Astruct", for  *
 *     "struct _Astruct".                                                   *
 *                                                                          *
 ****************************************************************************/


typedef struct _zbucketstruct {

  int *index;                        /* Subscript indices of nonzeroes.     */
  double_complex *value;           /* Values of nonzeroes.                */

#ifdef _DEBUG                        /* Debugging Flag Field:               */
  int errorflag;                     /*    to ensure integrity of mallocs.  */
#endif

  struct _zbucketstruct *nextbucket;  /* nsia.                          */
} zbucketstruct;

/*             ___________________bucketstruct___________________           *
 *             |                                                |           *
 *             |                   ___index___   ____value___   |           *
 *             |                   |         |   |          |   |           *
 *             |                 0 |  (int)  |   | (double) |   |           *
 *             |                   |_________|   |__________|   |           *
 *             |                   |         |   |          |   |           *
 *             |                 1 |  (int)  |   | (double) |   |           *
 *             |                   |_________|   |__________|   |           *
 *             |                   |    .    |   |     .    |   |           *
 *             |                   |    .    |   |     .    |   |           *
 *             |                   |    .    |   |     .    |   |           *
 *             |                   |_________|   |__________|   |           *
 *             |                   |         |   |          |   |           *
 *             | zBUCKET_DIM-1 |  (int)  |   | (double) |   |           *
 *             |                   |_________|   |__________|   |           *
 *             |                                                |           *
 *             |             nextbucket *                       |           *
 *             |                        |                       |           *
 *             |________________________|_______________________|           *
 *                                      |                                   *
 *                                      V                                   *
 *                               (zbucketstruct)                        */




typedef struct _zibucketstruct {
  int *index_RANGE_BEGIN;
  int *index_RANGE_END;
                                     /* Subscript index pairs of range of   *
				      * nonzeroes.  Note: since in C, 2-d   *
				      * array elements are stored by row,   *
				      * index_RANGE_BEGIN and               *
				      * index_RANGE_END,                    *
				      * represent a range and are stored    *
				      * contiguously, which is advantageous *
				      * from a cache memory point of view,  *
				      * since they are also accessed in     *
				      * in tandem.                          */

#ifdef _DEBUG                        /* Debugging Flag Field:               */
  int errorflag;                     /*    to ensure integrity of mallocs.  */
#endif

  struct _zibucketstruct *nextibucket;  /* nsia.                        */
} zibucketstruct;

/*     __________________________ibucketstruct__________________________    *
 *     |                                                               |    *
 *     |                       ______index______                       |    *
 *     |                       |       |       |                       |    *
 *     |                     0 | (int) | (int) |                       |    *
 *     |                       |_______|_______|                       |    *
 *     |                       |       |       |                       |    *
 *     |                     1 | (int) | (int) |                       |    *
 *     |                       |_______|_______|                       |    *
 *     |                       |   .   |   .   |                       |    *
 *     |                       |   .   |   .   |                       |    *
 *     |                       |   .   |   .   |                       |    *
 *     |                       |_______|_______|                       |    *
 *     |                       |       |       |                       |    *
 *     |    zIBUCKET_DIM-1 | (int) | (int) |                       |    *
 *     |                       |_______|_______|                       |    *
 *     |                                                               |    *
 *     |                   nextibucket *                               |    *
 *     |                               |                               |    *
 *     |_______________________________|_______________________________|    *
 *                                     |                                    *
 *                                     V                                    *
 *                              (zibucketstruct)                        */



typedef struct _zvbucketstruct {

  double_complex *value;           /* Values of nonzeroes.                */

#ifdef _DEBUG                        /* Debugging Flag Field:               */
  int errorflag;                     /*    to ensure integrity of mallocs.  */
#endif

  struct _zvbucketstruct *nextvbucket; /* nsia.                         */
} zvbucketstruct;

/*                   _____________vbucketstruct_____________                *
 *                   |                                     |                *
 *                   |                    ____value___     |                *
 *                   |                    |          |     |                *
 *                   |                  0 | (double) |     |                *
 *                   |                    |__________|     |                *
 *                   |                    |          |     |                *
 *                   |                  1 | (double) |     |                *
 *                   |                    |__________|     |                *
 *                   |                    |     .    |     |                *
 *                   |                    |     .    |     |                *
 *                   |                    |     .    |     |                *
 *                   |                    |__________|     |                *
 *                   |                    |          |     |                *
 *                   | zVBUCKET_DIM-1 | (double) |     |                *
 *                   |                    |__________|     |                *
 *                   |                                     |                *
 *                   |      nextvbucket *                  |                *
 *                   |                  |                  |                *
 *                   |__________________|__________________|                *
 *                                      |                                   *
 *                                      V                                   *
 *                               (zvbucketstruct)                       */


/*
 * Sparse Vector Internal Data Structure (DS) Types: S1, SS, CS
 */

typedef struct _zS1struct {
  int index;

  double_complex value;

} zS1struct;

typedef struct _zSSstruct {
  zbucketstruct *bucketptr;      /* Pointer to linked list of buckets.  */
} zSSstruct;

typedef struct _zCSstruct {
  int nonzeroranges;                 /* Number of contiguous nonzero ranges.*/
  zibucketstruct *ibucketptr;    /* Pointer to linked list of ibuckets. */
  zvbucketstruct *vbucketptr;    /* Pointer to linked list of vbuckets. */
} zCSstruct;


typedef struct {
  int nonzeroes;                    /* The total # of                       *
				     * nonzeroes in the vector.             */
  int dstype;                       /* The data-structure type of the       *
				     * zvectorstruct will be one of the *
				     * following:
				     *          S1 = 1 SUBSCRIPT            *
				     *          SS = SINGLE SUBSCRIPTS      *
				     *          CS = COMPRESSED SUBSCRIPTS  */
  union _zds {
    zS1struct s1;
    zSSstruct ss;
    zCSstruct cs;
  } zds;

#ifdef _VTRACE

  /* Symbolic identifier */
  char* id;
  /* Numeric identifier */
  int num_id;

  /* Tracked? */
  unsigned char is_tracked;
  /* Named? */
  unsigned char is_named;
  /* Visualized? */
  unsigned char is_visual;

  /* List of all updates over the vector's lifetime */
  pu_struct u_list;

#endif

} zvectorstruct;

/*     ___________________________vectorstruct___________________________   *
 *     |                                                                |   *
 *     |          ____nonzeroes____         _____dstype______           |   *
 *     |          |               |         |               |           |   *
 *     |          |     (int)     |         |     (int)     |           |   *
 *     |          |_______________|         |_______________|           |   *
 *     |                                                                |   *
 *     |                               ds                               |   *
 *     |    {~~~~~~~~~~~~~~~~~~~~~~~~~~--~~~~~~~~~~~~~~~~~~~~~~~~~~}    |   *
 *     |    {                                                      }    |   *
 *     |    { EITHER:                                              }    |   *
 *     |    {     _____________________S1_____________________     }    |   *
 *     |    {     |   _____index_____     _____value_____    |     }    |   *
 *     |    {     |   |             |     |             |    |     }    |   *
 *     |    {     |   |    (int)    |     |    (int)    |    |     }    |   *
 *     |    {     |   |_____________|     |_____________|    |     }    |   *
 *     |    {     |__________________________________________|     }    |   *
 *     |    {                                                      }    |   *
 *     |    { OR:                                                  }    |   *
 *     |    {     _____________________SS_____________________     }    |   *
 *     |    {     |                                          |     }    |   *
 *     |    {     |           bucketptr *                    |     }    |   *
 *     |    {     |                     |                    |     }    |   *
 *     |    {     |_____________________|____________________|     }    |   *
 *     |    {                           |                          }    |   *
 *     |    {                           V                          }    |   *
 *     |    {                     (zbucketstruct)              }    |   *
 *     |    {                                                      }    |   *
 *     |    { OR:                                                  }    |   *
 *     |    {     _____________________CS_____________________     }    |   *
 *     |    {     |                                          |     }    |   *
 *     |    {     |              _nonzeroranges_             |     }    |   *
 *     |    {     |              |             |             |     }    |   *
 *     |    {     |              |    (int)    |             |     }    |   *
 *     |    {     |              |_____________|             |     }    |   *
 *     |    {     |                                          |     }    |   *
 *     |    {     |     ibucketptr *        vbucketptr *     |     }    |   *
 *     |    {     |                |                   |     |     }    |   *
 *     |    {     |________________|___________________|_____|     }    |   *
 *     |    {                      |                   |           }    |   *
 *     |    {                      V                   V           }    |   *
 *     |    {          (zibucketstruct)  (zvbucketstruct)  }    |   *
 *     |    {                                                      }    |   *
 *     |    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~    |   *
 *     |                                                                |   *
 *     |________________________________________________________________|   */




typedef struct _zvectorstorestruct {
  zvectorstruct vector[zVECTORSTORE_DIM];  /* Long array of vectors */
  struct _zvectorstorestruct *nextstore;       /* nsia.                 */
} zvectorstorestruct;

/*             ________________vectorstorestruct________________            *
 *             |                                               |            *
 *             |                        ______vector______     |            *
 *             |                        |                |     |            *
 *             |                      0 | (zvectorstruct) ||            *
 *             |                        |________________|     |            *
 *             |                        |                |     |            *
 *             |                      1 | (zvectorstruct) ||            *
 *             |                        |________________|     |            *
 *             |                        |        .       |     |            *
 *             |                        |        .       |     |            *
 *             |                        |        .       |     |            *
 *             |                        |________________|     |            *
 *             |                        |                |     |            *
 *             | zVECTORSTORE_DIM-1 | (zvectorstruct) ||            *
 *             |                        |________________|     |            *
 *             |                                               |            *
 *             |             nextstore *                       |            *
 *             |                       |                       |            *
 *             |_______________________|_______________________|            *
 *                                     |                                    *
 *                                     V                                    *
 *                            (zvectorstorestruct)                      */






typedef struct _zbucketstorestruct {
  zbucketstruct bucket[zBUCKETSTORE_DIM];  /* Long array of buckets.*/
  struct _zbucketstorestruct *nextstore;       /* nsia.                 */
} zbucketstorestruct;

/*             ________________bucketstorestruct________________            *
 *             |                                               |            *
 *             |                        ______bucket______     |            *
 *             |                        |                |     |            *
 *             |                      0 | (zbucketstruct) ||            *
 *             |                        |________________|     |            *
 *             |                        |                |     |            *
 *             |                      1 | (zbucketstruct) ||            *
 *             |                        |________________|     |            *
 *             |                        |        .       |     |            *
 *             |                        |        .       |     |            *
 *             |                        |        .       |     |            *
 *             |                        |________________|     |            *
 *             |                        |                |     |            *
 *             | zBUCKETSTORE_DIM-1 | (zbucketstruct) ||            *
 *             |                        |________________|     |            *
 *             |                                               |            *
 *             |             nextstore *                       |            *
 *             |                       |                       |            *
 *             |_______________________|_______________________|            *
 *                                     |                                    *
 *                                     V                                    *
 *                            (zbucketstorestruct)                      */


typedef struct _zibucketstorestruct {
  zibucketstruct ibucket[zIBUCKETSTORE_DIM];
                                                 /* Long array of ibuckets. */
  struct _zibucketstorestruct *nextstore;    /* nsia.                   */
} zibucketstorestruct;

/*             ________________ibucketstorestruct________________           *
 *             |                                               |            *
 *             |                        ______bucket______     |            *
 *             |                        |                |     |            *
 *             |                      0 |(zibucketstruct) ||            *
 *             |                        |________________|     |            *
 *             |                        |                |     |            *
 *             |                      1 |(zibucketstruct) ||            *
 *             |                        |________________|     |            *
 *             |                        |        .       |     |            *
 *             |                        |        .       |     |            *
 *             |                        |        .       |     |            *
 *             |                        |________________|     |            *
 *             |                        |                |     |            *
 *             | zIBUCKETSTORE_DIM-1| (zibucketstruct)||            *
 *             |                        |________________|     |            *
 *             |                                               |            *
 *             |             nextstore *                       |            *
 *             |                       |                       |            *
 *             |_______________________|_______________________|            *
 *                                     |                                    *
 *                                     V                                    *
 *                            (zibucketstorestruct)                     */



typedef struct _zvbucketstorestruct {
  zvbucketstruct vbucket[zVBUCKETSTORE_DIM];
                                                 /* Long array of vbuckets. */
  struct _zvbucketstorestruct *nextstore;    /* nsia.                   */
} zvbucketstorestruct;

/*             ________________vbucketstorestruct_________________          *
 *             |                                                 |          *
 *             |                         ______vbucket______     |          *
 *             |                         |                 |     |          *
 *             |                       0 | (zvbucketstruct) ||          *
 *             |                         |_________________|     |          *
 *             |                         |                 |     |          *
 *             |                       1 | (zvbucketstruct) ||          *
 *             |                         |_________________|     |          *
 *             |                         |         .       |     |          *
 *             |                         |         .       |     |          *
 *             |                         |         .       |     |          *
 *             |                         |_________________|     |          *
 *             |                         |                 |     |          *
 *             | zVBUCKETSTORE_DIM-1 | (zvbucketstruct) ||          *
 *             |                         |_________________|     |          *
 *             |                                                 |          *
 *             |              nextstore *                        |          *
 *             |                        |                        |          *
 *             |________________________|________________________|          *
 *                                      |                                   *
 *                                      V                                   *
 *                            (zvbucketstorestruct)                     */





/****************************************************************************
 *                                                                          *
 *  Function Prototyping                                                    *
 *                                                                          *
 ****************************************************************************/

#define _P(x) ()

void zallocatebucketstore _P(());
void zallocateibucketstore _P(());
void zallocatevbucketstore _P(());
void zallocatevectorstore _P(());

void zgetbucket _P((zbucketstruct **bucketptrptr));
void zgetibucket _P((zibucketstruct **ibucketptrptr));
void zgetvbucket _P((zvbucketstruct **vbucketptrptr));
void zgetvector _P((zvectorstruct **vectorptrptr));

void zspgx _P((zvectorstruct **destptrptr,
		   int arity,
		   double_complex *alphaptrs[],
		   zvectorstruct **rhsvectorptrptrs[]));

void znon_over_lapping_cs_range_axpy _P((int rangedestination,
				      int lhszibucketindex,
				      int *addressoflastlhszibucketindex,
				      int *addressoflhszvbucketindex,
				      int k,
				      int *addressofrhszvbucketindex_k,
				      double_complex alpha_k,
				      int totalcurrentnonzerorange
#ifdef _DEBUG
				      ,int *addressofunconsumedcurrentnonzerorange
				      ,int rhszvectorconsumednonzeroranges_k
				      ,int rhszvectornonzeroranges_k
				      ,int lhszvectornonzeroranges
#endif
				      ));

void zoverlapping_cs_range_axpy _P((int *addressoflhszvbucketindex,
				   int k,
				   int *addressofrhszvbucketindex_k,
				   double_complex alpha_k,
				   int totalcurrentnonzerorange
#ifdef _DEBUG
				   ,int *addressofunconsumedcurrentnonzerorange
				   ,int rhszvectorconsumednonzeroranges_k
				   ,int rhszvectornonzeroranges_k
				   ,int lhszvectornonzeroranges
#endif
				   ));


void zdetermine_loop_bounds _P((int totalcurrentnonzerorange,
			       int lhsvbucketindex,
			       int rhsvbucketindex,
			       int *firstloopsize,
			       int *numberofdoubleforloopiterations,
			       int *firstiterationloopsize,
			       int *seconditerationloopsize,
			       int *remainder1loopsize,
			       int *remainder2loopsize
#ifdef _DEBUG
			       ,int *addressofunconsumedcurrentnonzerorange
#endif
			       ));

void zdebugging_code_1 _P((char *RIGHT_OR_LEFT,
			  int vbucketindex,
			  int unconsumedcurrentnonzerorange));

void zhop_to_next_rhsvbucket _P((int k,
				int *vbucketindex
#ifdef _DEBUG
				,int rhszvectorconsumednonzeroranges
				,int rhszvectornonzeroranges
#endif
				));


void zhop_to_next_lhsvbucket _P((int *vbucketindex
#ifdef _DEBUG
				,int lhszvectornonzeroranges
#endif
				));

void zget_and_hop_to_next_vbucket _P((int *lhszvbucketindex));

void zconvertdstype _P((zvectorstruct **oldvectorptrptr,
		       int k,
		       int newvectordstype));

void zaxpy_0_1_post_processing _P((zvectorstruct **destptrptr));

void zaxpy_2_heap_post_processing _P((zvectorstruct **destptrptr,
                                     zvectorstruct **rhszvectorptrptrs[],
                                     double_complex *alphaptrs[],
                                     zvectorstruct *restorerhszvectorptr[MAXARITY+1],
                                     int arity,
                                     int lhszvectornonzeroes,
                                     int lhszvectornonzeroranges,
                                     int governingdstype));

void zplusequal_axpy_postprocessing _P((zvectorstruct **destptrptr,
				       zvectorstruct **rhszvectorptrptr,
				       int status_otherrhszvector,
				       zvectorstruct **restofplusequalvectorptrptr,
				       double_complex *alphaptrs[2],
				       zvectorstruct *restorerhszvectorptr[2+1],
				       int arity,
				       int lhszvectornonzeroes,
				       int lhszvectornonzeroranges,
				       int governingdstype
				       ));

#ifdef _PEEP
void zadpeep_pre _P((int vectorindex,
		    zvectorstruct **rhsvectorptrptr,
		    double_complex *alphaptr));

void zadpeep_post _P((zvectorstruct **destptrptr));
#endif

void zadallocate _P(());
void zadfree _P(());
void zad_free_vectors_buckets _P((zvectorstruct **vectorptrptr));
void zinitialize_pd_global_variables _P(());

void zspsd _P((zvectorstruct **destptrptr,
		int *indexvector,
		double_complex *valuevector,
		int *lengthptr));

void zspxd _P((double_complex *vector,
			 int *lengthptr,
			 zvectorstruct **objectptrptr,
			 int *returned_len,
			 int *len_info));

void zspxm _P((double_complex *vector,
			       int *lengthptr,
			       double_complex *multiplier,
			       zvectorstruct **objectptrptr,
			       int *returned_len,
			       int *len_info));

void zspxa _P((double_complex *vector,
				   int *lengthptr,
				   zvectorstruct **objectptrptr,
				   int *returned_len,
				   int *len_info));

void zspxs _P((int *indvector,
			  double_complex *valvector,
			  int *lengthptr,
			  zvectorstruct **objectptrptr,
			  int *returned_len,
			  int *len_info)); 

void zsppr _P((zvectorstruct **objectptrptr, int *unit_number));

void zspzronotca _P((zvectorstruct **destptrptr));

void zspcp _P((zvectorstruct **destptrptr,
		zvectorstruct **oldvectorptrptr));

void zadinfo _P((int arity,
		int nnzb,
		zvectorstruct *lhsvectoraddress,
		int nnza,
		double_complex *alphaptrs[],
		zvectorstruct **vectorptrptrs[]));

void zspg1 _P((zvectorstruct **destptrptr,
		double_complex *alphaptr,
		zvectorstruct **vectoraptrptr));
void zspg2 _P((zvectorstruct **destptrptr,
		double_complex *alphaptr,
		zvectorstruct **vectoraptrptr,
		double_complex *betaptr,
		zvectorstruct **vectorbptrptr));
void plusequalzspg1 _P((zvectorstruct **rhsanddestptrptr,
			 double_complex *alphaptr));
void plusequalzspg2 _P((zvectorstruct **rhsanddestptrptr,
			 double_complex *otheralphaptr,
			 zvectorstruct **otherrhsvectorptrptr,
			 int status_otherrhsvector));



#endif /*  #ifndef zGAXPY_PD_MACROS_AND_DATA_STRUCTURES  */
