#define DOUBLE_COMPLEX
/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#include  <stdlib.h>
#include "zgaxpy_pd_macros_and_data_structures.h"
#include "zgaxpy_pd_global_variables.h"


/****************************************************************************
 *                                                                          *
 *  zallocatebucketstore()                                                   *
 *                                                                          *
 *     1 malloc call gets all the memory for a bucketstore.  If malloc does *
 *     not return NULL, one of the following 2 scenarios will result:       *
 *                                                                          *
 *     A: outcome of zallocatebucketstore the 1st time it's called:          *
 *                                                                          *
 *     bucketstoreptr -----------+                                          *
 *                               |   store                                  *
 *                              _V_____1_____                               *
 *     bucketstorestartptr ---->|    _____  |                               *
 *     bucketstorecounter ------+->0 |___|  |                               *
 *                              |    | . |  |                               *
 *                              |    | . |  |                               *
 *                              |    | . |  |                               *
 *                              |    |___|  |                               *
 *                              |    |___|  |                               *
 *                              |           |                               *
 *                              | nextstore |                               *
 *                              |     *     |                               *
 *                              |_____|_____|                               *
 *                                    |                                     *
 *                                    V                                     *
 *                                   NULL                                   *
 *                                                                          *
 *     B: outcome of zallocatebucketstore the Nth time it's called (N > 1):  *
 *                                                                          *
 *     bucketstorecounter --------------------------------------+           *
 *                                                              |           *
 *     bucketstoreptr ----------------------------------------+ |           *
 *                               store        store           | | store     *
 *                              ___1___      ___2___ ...     _V_|___N_____  *
 *     bucketstorestartptr ---->|     |  +-->|     |     +-->|  V _____  |  *
 *                              |     |  |   |     |     |   |  0 |___|  |  *
 *                              |     |  |   |     |     |   |    | . |  |  *
 *                              |     |  |   |     |     |   |    | . |  |  *
 *                              |     |  |   |     |     |   |    | . |  |  *
 *                              |     |  |   |     |     |   |    |___|  |  *
 *                              |     |  |   |     |     |   |    |___|  |  *
 *                              |     |  |   |     |     |   |           |  *
 *                              |     |  |   |     |     |   | nextstore |  *
 *                              |  *  |  |   |  *  |     |   |     *     |  *
 *                              |__|__|  |   |__|__|     |   |_____|_____|  *
 *                                 |     |      |        |         |        *
 *                                 +-----+      +---...--+         V        *
 *                                                                NULL      *
 *                                                                          *
 ****************************************************************************/

void zallocatebucketstore()
{
  int i,k;
  
  zbucketstorestruct *tempbucketstoreptr;
  zbucketstruct *tempbucketptr;
  
  int *index_store;
  double_complex *value_store;

  int temp_size;

  /*
   * Per BUCKET Dynamic Allocation: constant zBUCKETSTORE_DIM
   *                                variable zBUCKET_DIM
   */
  
  if ((tempbucketstoreptr = (zbucketstorestruct *)
       palloc ((temp_size = (sizeof(zbucketstorestruct)))))==NULL) {
    printf("zallocatebucketstore error: malloc returned NULL.\n");
    abort();
  }
  memory_count(Z_EVAL,MEM_BUCKET,temp_size);
  
  for (k=0; k<zBUCKETSTORE_DIM; k++) {
    tempbucketptr = &(tempbucketstoreptr->bucket[k]);
    if ((tempbucketptr->index = (int *)
	 palloc ((temp_size = (zBUCKET_DIM * sizeof(int)))))==NULL) {
      printf("zallocatebucketstore error: malloc returned NULL.\n");
      abort();
    }
    memory_count(Z_EVAL,MEM_BUCKET,temp_size);
    if ((tempbucketptr->value = (double_complex *)
	 palloc ((temp_size = (zBUCKET_DIM *
			       sizeof(double_complex)))))==NULL) {
      printf("zallocatebucketstore error: malloc returned NULL.\n");
      abort();
    }
    memory_count(Z_EVAL,MEM_BUCKET,temp_size);
  }
      
  if (zbucketstorestartptr == NULL)
    zbucketstorestartptr = tempbucketstoreptr;
  else
    zbucketstoreptr->nextstore = tempbucketstoreptr;
  
  zbucketstoreptr = tempbucketstoreptr;
  zbucketstoreptr->nextstore = NULL;
  zbucketstorecounter = 0;
  
  for (i=0;i<zBUCKETSTORE_DIM;i++) {
    (&(zbucketstoreptr->bucket[i]))->nextbucket = NULL;
#ifdef _DEBUG  
    (&(zbucketstoreptr->bucket[i]))->errorflag = 9999;
#endif
  }
}



/****************************************************************************
 *                                                                          *
 *  zallocateibucketstore()                                                  *
 *                                                                          *
 *     Identical to zallocatebucketstore in terms of control flow.           *
 *                                                                          *
 ****************************************************************************/

void zallocateibucketstore()
{
  int i,k;

  zibucketstorestruct *tempibucketstoreptr;
  zibucketstruct *tempibucketptr;
  
  int *index_RANGE_BEGIN_store;
  int *index_RANGE_END_store;

  int temp_size;
 
  /*
   * Per IBUCKET Dynamic Allocation: constant zIBUCKETSTORE_DIM
   *                                 variable zIBUCKET_DIM
   */
  
  if ((tempibucketstoreptr = (zibucketstorestruct *)
       palloc ((temp_size = (sizeof(zibucketstorestruct)))))==NULL) {
    printf("zallocateibucketstore error: malloc returned NULL.\n");
    abort();
  }
  memory_count(Z_EVAL,MEM_IBUCKET,temp_size);
  
  for (k=0; k<zIBUCKETSTORE_DIM; k++) {
    tempibucketptr = &(tempibucketstoreptr->ibucket[k]);
    if ((tempibucketptr->index_RANGE_BEGIN = (int *)
	 palloc ((temp_size = (zIBUCKET_DIM * sizeof(int)))))==NULL) {
      printf("zallocateibucketstore error: malloc returned NULL.\n");
      abort();
    }
    memory_count(Z_EVAL,MEM_IBUCKET,temp_size);
    if ((tempibucketptr->index_RANGE_END = (int *)
	 palloc ((temp_size = (zIBUCKET_DIM * sizeof(int)))))==NULL) {
      printf("zallocateibucketstore error: malloc returned NULL.\n");
      abort();
    }
    memory_count(Z_EVAL,MEM_IBUCKET,temp_size);
  }
  
  if (zibucketstorestartptr == NULL)
    zibucketstorestartptr = tempibucketstoreptr;
  else
    zibucketstoreptr->nextstore = tempibucketstoreptr;
  
  zibucketstoreptr = tempibucketstoreptr;
  zibucketstoreptr->nextstore = NULL;
  zibucketstorecounter = 0;
  
  for (i=0;i<zIBUCKETSTORE_DIM;i++) {
    (&(zibucketstoreptr->ibucket[i]))->nextibucket = NULL;
#ifdef _DEBUG  
    (&(zibucketstoreptr->ibucket[i]))->errorflag = 9099;
#endif
  }
}




/****************************************************************************
 *                                                                          *
 *  zallocatevbucketstore()                                                  *
 *                                                                          *
 *     Identical to zallocatebucketstore in terms of control flow.           *
 *                                                                          *
 ****************************************************************************/

void zallocatevbucketstore()
{
  int i,k;

  zvbucketstorestruct *tempvbucketstoreptr;
  zvbucketstruct *tempvbucketptr;
  
  double_complex *value_store;

  int temp_size;
  
  /*
   * Per VBUCKET Dynamic Allocation: constant zVBUCKETSTORE_DIM
   *                                 variable zVBUCKET_DIM
   */
  
  if ((tempvbucketstoreptr = (zvbucketstorestruct *)
       palloc ((temp_size = (sizeof(zvbucketstorestruct)))))==NULL) {
    printf("zallocatevbucketstore error: malloc returned NULL.\n");
    abort();
  }
  memory_count(Z_EVAL,MEM_VBUCKET,temp_size);
  
  for (k=0; k<zVBUCKETSTORE_DIM; k++) {
    tempvbucketptr = &(tempvbucketstoreptr->vbucket[k]);
    if ((tempvbucketptr->value = (double_complex *)
	 palloc ((temp_size = (zVBUCKET_DIM *
			       sizeof(double_complex)))))==NULL) {
      printf("zallocatevbucketstore error: malloc returned NULL.\n");
      abort();
    }
    memory_count(Z_EVAL,MEM_VBUCKET,temp_size);
  }
  
  if (zvbucketstorestartptr == NULL)
    zvbucketstorestartptr = tempvbucketstoreptr;
  else
    zvbucketstoreptr->nextstore = tempvbucketstoreptr;
  
  zvbucketstoreptr = tempvbucketstoreptr;
  zvbucketstoreptr->nextstore = NULL;
  zvbucketstorecounter = 0;
  
  for (i=0;i<zVBUCKETSTORE_DIM;i++) {
    (&(zvbucketstoreptr->vbucket[i]))->nextvbucket = NULL;
#ifdef _DEBUG  
    (&(zvbucketstoreptr->vbucket[i]))->errorflag = 9009;
#endif
  }
}




/****************************************************************************
 *                                                                          *
 *  zallocatevectorstore()                                                   *
 *                                                                          *
 *     zVECTORSTORE_DIM is a constant and in contrast to the buckets    *
 *     there is no VECTOR_DIM.  Therefore there is no need for any          *
 *     allocation scheme requiring the per STORE vs per VECTOR methods.     *
 *                                                                          *
 ****************************************************************************/

void zallocatevectorstore()
{
  int i;
  zvectorstorestruct *tempvectorstoreptr;
  zvectorstruct *tempvectorptr;

  int temp_size;

  /*
   * Wholesale Dynamic Allocation: constant zVECTORSTORE_DIM
   * This allocation is not obsolete unlike the bucket type stores
   * above, since there is no variable-sized bucket in a vectorstruct.
   */
  if ((tempvectorstoreptr = (zvectorstorestruct *)
       palloc ((temp_size = (sizeof(zvectorstorestruct)))))==NULL ) {
    printf("zallocatevectorstore error: malloc returned NULL.\n");
    abort();
  }
  memory_count(Z_EVAL,MEM_VECTOR,temp_size);
  
  /*
   * End of Wholesale Dynamic Allocation
   */
  
  if (zvectorstorestartptr == NULL)
    zvectorstorestartptr = tempvectorstoreptr;
  else
    zvectorstoreptr->nextstore = tempvectorstoreptr;
  
  zvectorstoreptr = tempvectorstoreptr;
  zvectorstoreptr->nextstore = NULL;
  zvectorstorecounter = 0;
  for (i=0;i<zVECTORSTORE_DIM;i++) {
    tempvectorptr = &(zvectorstoreptr->vector[i]);

/*
 *  PMK: 12/4/96 edits
 *
 *  The assignment of the S0 "non-yet-specified" data structure type is
 *  extraneous for now, since we never have checks on S0 (i.e., nowhere
 *  in the code do we say, if (somevectorptr->dstype == S0) ...
 *  Therefore, the following code line is commented out:

    tempvectorptr->dstype = S0;

 *
 */

  }
}



/****************************************************************************
 *                                                                          *
 *  zgetbucket(bucketptrptr)                                                 *
 *                                                                          *
 *     bucketptrptr    is a doubly-indirected pointer to a zbucketstruct*
 *                     Since we want the value of "bucketptr" to be changed *
 *                     by zgetbucket (i.e., *bucketptrptr to get assigned a  *
 *                     pointer to a zbucketstruct), we need to make its *
 *                     address (bucketptrptr) the function argument.  Thus, *
 *                     the need for the double indirection (**).            *
 *                                                                          *
 *     zgetbucket assigns to *bucketptrptr the pointer to the next available *
 *     zbucketstruct from the freedbucketslist or if none there from the*
 *     bucketstore.  In the latter case, if we're at the end of the current *
 *     bucketstore, a new store is allocated before doing the "get".        *
 *                                                                          *
 ****************************************************************************/

void zgetbucket(bucketptrptr)
zbucketstruct **bucketptrptr;
{
  if (freedzbucketslist != NULL) {
    (*bucketptrptr) = freedzbucketslist;
    freedzbucketslist = freedzbucketslist->nextbucket;
  }
  else {
    if (zbucketstorecounter == zBUCKETSTORE_DIM) {
      zallocatebucketstore();
    }
    
#ifdef _DEBUG
    else if (zbucketstorecounter > zBUCKETSTORE_DIM) {
      printf("zgetbucket error: zbucketstorecounter=%d,  STOREDIMENSION=%d\n",
	     zbucketstorecounter,zBUCKETSTORE_DIM);
      abort();
    }
#endif

    /*
     *  The following assignment of a store entry to *bucketptrptr
     *  works regardless of whether we've used per BUCKET or per STORE
     *  allocation method.
     */

    *bucketptrptr = (zbucketstruct *)
                     &(zbucketstoreptr->bucket[zbucketstorecounter++]);
  }
  (*bucketptrptr)->nextbucket = NULL;
  
#ifdef _DEBUG
  if ((*bucketptrptr)->errorflag != 9999) {
    printf("zgetbucket error: memory corruption of zbucketstore\n");
    abort();
  }
#endif
  
}

/****************************************************************************
 *                                                                          *
 *  zgetibucket(ibucketptrptr)                                               *
 *                                                                          *
 *     ibucketptrptr   is a doubly-indirected pointer to a zibucketstruc*
 *                     Since we want the value of "ibucketptr" to be        *
 *                     changed by zgetibucket (i.e., *ibucketptrptr to get   *
 *                     assigned a  pointer to a zibucketstruct), we need*
 *                     to make its address (ibucketptrptr) the function     *
 *                     argument.  Thus, the need for the double indirection *
 *                     (**).                                                *
 *                                                                          *
 *     zgetibucket assigns to *ibucketptrptr the pointer to the next         *
 *     available zibucketstruct from the freedibucketslist or if none   *
 *     from the ibucketstore.  In the latter case, if we're at the end      *
 *     of the current ibucketstore, a new store is allocated before doing   *
 *     the "get".                                                           *
 *                                                                          *
 ****************************************************************************/

void zgetibucket(ibucketptrptr)
zibucketstruct **ibucketptrptr;
{
  if (freedzibucketslist != NULL) {
    (*ibucketptrptr) = freedzibucketslist;
    freedzibucketslist = freedzibucketslist->nextibucket;
  }
  else {
    if (zibucketstorecounter == zIBUCKETSTORE_DIM) {
      zallocateibucketstore();
    }
    
#ifdef _DEBUG
    else if (zibucketstorecounter > zIBUCKETSTORE_DIM) {
      printf("zgetibucket error: zibucketstorecounter=%d,  STOREDIMENSION=%d\n",
	     zibucketstorecounter,zIBUCKETSTORE_DIM);
      abort();
    }
#endif

    *ibucketptrptr = (zibucketstruct *)
              &(zibucketstoreptr->ibucket[zibucketstorecounter++]);
  }
  (*ibucketptrptr)->nextibucket = NULL;
  
#ifdef _DEBUG
  if ((*ibucketptrptr)->errorflag != 9099) {
    printf("zgetibucket error: memory corruption of zibucketstore\n");
    abort();
  }
#endif
  
}


/****************************************************************************
 *                                                                          *
 *  zgetvbucket(vbucketptrptr)                                               *
 *                                                                          *
 *     vbucketptrptr   is a doubly-indirected pointer to a zvbucketstruc*
 *                     Since we want the value of "vbucketptr" to be        *
 *                     changed by zgetvbucket (i.e., *vbucketptrptr to get   *
 *                     assigned a  pointer to a zvbucketstruct), we need*
 *                     to make its address (vbucketptrptr) the function     *
 *                     argument.  Thus, the need for the double indirection *
 *                     (**).                                                *
 *                                                                          *
 *     zgetvbucket assigns to *vbucketptrptr the pointer to the next         *
 *     available zvbucketstruct from the freedvbucketslist or if none   *
 *     there from the vbucketstore.  In the latter case, if we're at the end*
 *     of the current vbucketstore, a new store is allocated before doing   *
 *     the "get".                                                           *
 *                                                                          *
 ****************************************************************************/

void zgetvbucket(vbucketptrptr)
zvbucketstruct **vbucketptrptr;
{
  if (freedzvbucketslist != NULL) {
    (*vbucketptrptr) = freedzvbucketslist;
    freedzvbucketslist = freedzvbucketslist->nextvbucket;
  }
  else {
    if (zvbucketstorecounter == zVBUCKETSTORE_DIM) {
      zallocatevbucketstore();
    }
    
#ifdef _DEBUG
    else if (zvbucketstorecounter > zVBUCKETSTORE_DIM) {
      printf("zgetvbucket error: zvbucketstorecounter=%d,  STOREDIMENSION=%d\n",
	     zvbucketstorecounter,zVBUCKETSTORE_DIM);
      abort();
    }
#endif

    *vbucketptrptr = (zvbucketstruct *)
              &(zvbucketstoreptr->vbucket[zvbucketstorecounter++]);
  }
  (*vbucketptrptr)->nextvbucket = NULL;
  
#ifdef _DEBUG
  if ((*vbucketptrptr)->errorflag != 9009) {
    printf("zgetvbucket error: memory corruption of zvbucketstore\n");
    abort();
  }
#endif
  
}



/****************************************************************************
 *                                                                          *
 *  zgetvector(vectorptrptr)                                                 *
 *                                                                          *
 *     vectorptrptr  is a doubly-indirected pointer to a zvectorstruct. *
 *                                                                          *
 *     zgetvector works in the same way as zgetvector, except that there is   *
 *     no freed list of vectors.  It replaces the functionality of "adnew". *
 *                                                                          *
 ****************************************************************************/

void zgetvector(vectorptrptr)
zvectorstruct **vectorptrptr;
{
  if (zvectorstorecounter == zVECTORSTORE_DIM) {
    zallocatevectorstore();
  }

#ifdef _DEBUG
  else if (zvectorstorecounter > zVECTORSTORE_DIM) {
    printf("zgetvector error: zvectorstorecount=%d, zVECTORSTORE_DIM=%d\n",
	   zvectorstorecounter,zVECTORSTORE_DIM);
    abort();
  }
#endif

  *vectorptrptr = (zvectorstruct *)
                    &(zvectorstoreptr->vector[zvectorstorecounter++]);

#ifdef _TRACE
  zadinfo(-1,0,*vectorptrptr,0,NULL,NULL);
#endif

}

