#define DOUBLE_COMPLEX
/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#include  <stdlib.h>
#include "zgaxpy_pd_macros_and_data_structures.h"
#include "zgaxpy_pd_global_variables.h"

/****************************************************************************
 *                                                                          *
 *     zspzronotca assigns a zero vector to its argument.                        *
 *                                                                          *
 ****************************************************************************/

void zspzronotca(destptrptr)
zvectorstruct **destptrptr;
{
  if (*destptrptr == NULL)
    zgetvector(destptrptr);
  
  lhszvectorptr->nonzeroes = 0;
  lhszvectorptr->dstype = S1;
  lhszvectorptr->zds.s1.index = DUMMYSUBSCRIPT;
  X_ZERO(lhszvectorptr->zds.s1.value);

  zaxpy_0_1_post_processing(destptrptr);
}



/****************************************************************************
 *                                                                          *
 *     zspcp assigns to its first vector argument a vector whose values    *
 *     are identical to its second vectpr argument.                         *
 *                                                                          *
 ****************************************************************************/

void zspcp(destptrptr,oldvectorptrptr)
zvectorstruct **destptrptr,**oldvectorptrptr;
{
  zvectorstruct *oldvectorptr,*newvectorptr;
  zbucketstruct *oldvectorbucketptr,
                    *newvectortailbucketptr,*newbucketptr;
  zibucketstruct *oldvectoribucketptr,
                     *newvectortailibucketptr,*newibucketptr;
  zvbucketstruct *oldvectorvbucketptr,
                     *newvectortailvbucketptr,*newvbucketptr;
  int i,bucketindex,ibucketindex,vbucketindex;

#ifdef _PEEP
  double_complex one;
  X_PUT_REAL(one,1.0);
#endif    

  oldvectorptr = *oldvectorptrptr;

  if (*destptrptr == NULL)
     zgetvector(destptrptr);

#ifdef _PEEP
  adpeep_header(0);
  zadpeep_pre(0,oldvectorptrptr,&one);
#endif    
  newvectorptr = lhszvectorptr;

  newvectorptr->nonzeroes =  oldvectorptr->nonzeroes;
  newvectorptr->dstype =  oldvectorptr->dstype;

  if (newvectorptr->dstype == S1) {
    newvectorptr->zds.s1.index =  oldvectorptr->zds.s1.index;
    newvectorptr->zds.s1.value =  oldvectorptr->zds.s1.value;
  }
  else if (newvectorptr->dstype == SS) {
    oldvectorbucketptr = oldvectorptr->zds.ss.bucketptr;

    for (i=0;i<newvectorptr->nonzeroes;i++) {
      if ((bucketindex = MY_MOD(i,zBUCKET_DIM_minus1)) == 0) {
	if (i == 0) {
	  zgetbucket(&newbucketptr);
	  newvectorptr->zds.ss.bucketptr = newbucketptr;
	}
	else {
	  oldvectorbucketptr = oldvectorbucketptr->nextbucket;
	  zgetbucket(&newbucketptr);
	  newvectortailbucketptr->nextbucket = newbucketptr;
	}
	newvectortailbucketptr = newbucketptr;
      }
      newvectortailbucketptr->value[bucketindex] =
        oldvectorbucketptr->value[bucketindex];
      newvectortailbucketptr->index[bucketindex] =
        oldvectorbucketptr->index[bucketindex];
    }
  }
  else /* if (newvectorptr->dstype == CS */ {
    oldvectoribucketptr = oldvectorptr->zds.cs.ibucketptr;
    oldvectorvbucketptr = oldvectorptr->zds.cs.vbucketptr;
    newvectorptr->zds.cs.nonzeroranges =
      oldvectorptr->zds.cs.nonzeroranges;
    
    for (i=0;i<newvectorptr->zds.cs.nonzeroranges;i++) {
      if ((ibucketindex = MY_MOD(i,zIBUCKET_DIM_minus1)) == 0) {
	if (i == 0) {
	  zgetibucket(&newibucketptr);
	  newvectorptr->zds.cs.ibucketptr = newibucketptr;
	}
	else {
	  oldvectoribucketptr = oldvectoribucketptr->nextibucket;
	  zgetibucket(&newibucketptr);
	  newvectortailibucketptr->nextibucket = newibucketptr;
	}
	newvectortailibucketptr = newibucketptr;
      }

      newvectortailibucketptr->index_RANGE_BEGIN[ibucketindex] =
        oldvectoribucketptr->index_RANGE_BEGIN[ibucketindex];
      newvectortailibucketptr->index_RANGE_END[ibucketindex] =
        oldvectoribucketptr->index_RANGE_END[ibucketindex];
    }
    
    for (i=0;i<newvectorptr->nonzeroes;i++) {
      if ((vbucketindex = MY_MOD(i,zVBUCKET_DIM_minus1)) == 0) {
	if (i == 0) {
	  zgetvbucket(&newvbucketptr);
	  newvectorptr->zds.cs.vbucketptr = newvbucketptr;
	}
	else {
	  oldvectorvbucketptr = oldvectorvbucketptr->nextvbucket;
	  zgetvbucket(&newvbucketptr);
	  newvectortailvbucketptr->nextvbucket = newvbucketptr;
	}
	newvectortailvbucketptr = newvbucketptr;
      }
      newvectortailvbucketptr->value[vbucketindex] =
        oldvectorvbucketptr->value[vbucketindex];
    }
  }

  zaxpy_0_1_post_processing(destptrptr);

#ifdef _PEEP
  zadpeep_post(destptrptr);
#endif

}

